program sample
  ! ================================================
  use amgs_p_module
  implicit none
  !
  include 'mpif.h'
  integer(kind=4) :: row, i, width, rowsize
  real(kind=8)    :: threshold
  !
  ! crs matrix data
  integer(kind=4), allocatable :: crs_index(:),crs_column(:) 
  real(kind=8),    allocatable ::  crs_val(:),restol(:)
  real(kind=8),    allocatable ::  X(:), B(:)
  
  ! communicater, rank, etc.
  integer(kind=4) :: SOLVER_COMM, my_rank, ierr

  ! the following flag shows whether AMG*() subroutine do setup phase or not.
  logical :: SETUP_AMG=.true., CONV_FL
  ! maxiter limit the maximum iteration number
  integer(kind=4) :: maxiter

  !C-- init. MPI
  CALL MPI_INIT      (ierr)
  CALL MPI_COMM_RANK (MPI_COMM_WORLD, my_rank, ierr )
  CALL MPI_COMM_DUP  (MPI_COMM_WORLD, SOLVER_COMM, ierr)

  if(my_rank == 0) then
     ! -------------------------------------------------------------
     ! *** 2D poisson problem with 5-point stencil in CRS format ***
     ! -------------------------------------------------------------
     width = 500
     rowsize = width*width
     allocate(crs_index(0:rowsize),crs_column(rowsize*5), crs_val(rowsize*5))
     crs_index(0) = 0

     i = 0
     DO row = 1, rowsize
        if(row-width > 0) then
           i=i+1;crs_val(i) = -1.0;     crs_column(i) = row-width;  
        end if
        if(mod((row-1),width) /= 0) then
           i=i+1; crs_val(i) = -1.0;     crs_column(i) = row-1;      
        end if
        i=i+1;crs_val(i) = 4.0;     crs_column(i) = row;        
        if(mod((row+1),width) /= 1) then
           i=i+1; crs_val(i) = -1.0;     crs_column(i) = row+1;      
        end if
        if(row+width <= rowsize) then
           i=i+1; crs_val(i) = -1.0;     crs_column(i) = row+width;  
        end if
        crs_index(row) = i
     END DO

  else
     rowsize = 0
     allocate(crs_index(0:rowsize),crs_column(rowsize*5), crs_val(rowsize*5))
  end if
  
  allocate(B(rowsize), X(rowsize))
  ! right hand side vector is set
  do row = 1, rowsize
     B(row) = 0
     do i = crs_index(row-1)+1, crs_index(row)
        B(row) = B(row)+crs_val(i)
     end do
  end do
  X = 0

  ! --------------------------------------------------------
  ! *** AMG parameters can be set from character string  ***
  ! --------------------------------------------------------

!   call amgs_parameter_set(10, SOLVER_COMM)
   call amgs_parameter_option_file(10, "AMGS_OPTION.dat", SOLVER_COMM)

   ! call amgs_parameter_option(10, SOLVER_COMM,                  &
   !      & "-iter_down=1 -iter_up=1 -strong_con_threshold=0.01   &
   !      & -threshold_reduction_rate=0.8                         &
   !      & -krylov_solver=CG -convergence_check=1                &
   !      & -smoother=mcgs -dist_method=FIX                       &
   !      & -aggregation=coupled_w_lu                             &
   !      & -max_level_size=10 -min_node_size=100")

   !  switching the print contents:
   !             0: parameter list 1: current setting value
   !if(my_rank == 0) call amgs_option_info(1)


  ! -----------------------------------------------------------
  ! *** crs problem matrix is stored in amgs library module *** 
  ! -----------------------------------------------------------
  call amgs_store_problem_matrix(crs_index, crs_column, crs_val, rowsize, SOLVER_COMM)

  ! ------------------------------------------------------------
  ! *** the stored matrix is distributed among MPI processes ***
  ! ------------------------------------------------------------
  call amgs_distribute_matrix()

  ! ---------------------------------
  ! *** terminal condition is set ***
  ! ---------------------------------
  maxiter = 30
  threshold = 1.0E-7
  call amgs_terminal_val_set(threshold)

  ! ----------------------------
  ! *** AMG_solver is called ***
  ! ----------------------------
  call amgs_solver(X, B, maxiter, SETUP_AMG)

  ! -------------------------------------------------------------------
  ! *** clear problem matrix and multi-level data in library module ***
  ! -------------------------------------------------------------------
  call amgs_clear_matrices()

  
  if(my_rank == 0) deallocate(crs_index, crs_column,crs_val)
  deallocate(B,X)
  
  call MPI_FINALIZE(ierr)
  
end program sample
