  !C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
  !C   All rights reserved.
  !C
  !C   Redistribution and use in source and binary forms, with or without
  !C   modification, are permitted provided that the following conditions are met:
  !C   1. Redistributions of source code must retain the above copyright
  !C      notice, this list of conditions and the following disclaimer.
  !C   2. Redistributions in binary form must reproduce the above copyright
  !C      notice, this list of conditions and the following disclaimer in the
  !C      documentation and/or other materials provided with the distribution.
  !C   3. Neither the name of the project nor the names of its contributors
  !C      may be used to endorse or promote products derived from this software
  !C      without specific prior written permission.
  !C
  !C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
  !C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
  !C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  !C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
  !C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
  !C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  !C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  !C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  !C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  !C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  !C   POSSIBILITY OF SUCH DAMAGE.
  
  
#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif
  
  MODULE AMG_cycle
  CONTAINS

    !C B(NP), X(NP), TEMP(NP), WS(WSIZE), WR(WSIZE)
    SUBROUTINE v_cycle_ssi_amg(B, X, TEMP, LEVEL_NUM, SOLVER_COMM, WS, WR, NP, WSIZE)
      IMPLICIT NONE
      include 'mpif.h'
      include 'precision.inc'
      
      INTEGER(kind=kint), intent(in) :: WSIZE, NP

#ifdef C_INTERFACE_FOR_SSI
      REAL   (kind=kreal) :: B(NP), X(NP), TEMP(NP), WS(WSIZE), WR(WSIZE)
#else
      REAL   (kind=kreal) :: B(:), X(:), TEMP(:), WS(:), WR(:)
#endif

      INTEGER(kind=kint), intent(in) :: LEVEL_NUM, SOLVER_COMM
      INTEGER(kind=kint) :: ierr, my_rank, NPROCS
      INTEGER(kind=kint) :: i

      CALL MPI_COMM_SIZE (SOLVER_COMM, NPROCS, ierr) 
      CALL MPI_COMM_RANK (SOLVER_COMM, my_rank, ierr) 

      do i = 1, NP
         X(i) = 0.0d0
      end do

      !C    B(1:NP), X(1:NP), WS(1:WSIZE), WR(1:WSIZE), Temp(1:NP)
      call v_cycle(B, X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, Temp, NPROCS, WSIZE, NP)

    END SUBROUTINE v_cycle_ssi_amg


    !C As for scaling, rhss vec on PRE_LEVEL is assumed to be set before calling this sub
    SUBROUTINE lev1_mod(SOLVER_COMM, my_rank, WS, WR, Temp, &
         &             NPROCS, dim)
      USE comm_switch_mod
      USE data_structure_for_AMG
      USE solver_Gnumbering
      USE count_time_mod
      USE amg_parameter, only: iteration_number_up, iteration_number_down, &
           & iteration_number_bottom, gs_kind, dump_gauss_seidel_weight, W_OUT_LU
      use matrix_type_mod, only: matrix, set_mat_data
      USE chebyshev_smoother, only: cheby_sm_sca


      !$    USE omp_lib

      IMPLICIT NONE
      include 'mpif.h'
      INTEGER(kind=kint), INTENT(in)    :: NPROCS, dim
      INTEGER(kind=kint)                :: N,NP,NPL,NPU

      !C  Temp(1:problem_NP)
      REAL   (kind=kreal)  :: Temp(:)
      !C  WS(1:dim), WR(1:dim)
      REAL   (kind=kreal)  :: WS(:), WR(:)


      !C DIMENSION(NP)
      REAL   (kind=kreal), DIMENSION(:), pointer:: D

      REAL   (kind=kreal), DIMENSION(:), pointer::  B
      !C DIMENSION(NP)
      REAL   (kind=kreal), DIMENSION(:), pointer::  X
      !C DIMENSION(NP)

      REAL   (kind=kreal), DIMENSION(:), pointer::  AU
      REAL   (kind=kreal), DIMENSION(:), pointer::  AL

      INTEGER(kind=kint), DIMENSION(:), pointer ::  INU
      !C DIMENSION(0:N)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  IAU
      !C DIMENSION(NPU)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  INL
      !C DIMENSION(0:N)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  IAL

      INTEGER(kind=kint), intent(in) :: SOLVER_COMM
      INTEGER(kind=kint), intent(in) :: my_rank
      INTEGER(kind=kint) :: ierr

      REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_B
      REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_X

      INTEGER(kind=kint ) :: nth_lev,inod,ieL,isL,is,ie,coarser_N,isU,ieU,coarser_NP
      INTEGER(kind=kint ) :: coarser_NEIBPETOT,i,j,k,l
      REAL   (kind=kreal) :: R_val,B_val,w,X_val,R_norm,GR_norm

      type(INTER_LEVEL_OPERATOR), pointer :: R, P

      INTEGER(kind=kint) :: NEIBPETOT, iNEIBPETOT
      INTEGER(kind=kint), pointer :: NEIBPE(:)
      INTEGER(kind=kint), pointer :: STACK_IMPORT(:),STACK_EXPORT(:)
      INTEGER(kind=kint), pointer :: NOD_IMPORT(:),NOD_EXPORT(:)


      INTEGER(kind=kint) :: color,  omp_switch=0
      INTEGER(kind=kint), pointer :: col_ind(:)
      INTEGER(kind=kint), pointer :: INML(:)
      type(matrix) :: mat

      !C if OpenMP is validated, then ompswitch = 1
      !$ omp_switch = 1
      nth_lev = 1
#ifdef MEASURING
      CALL count_time(1, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

      call set_mat_data(mat, &
           & HIERARCHICAL_DATA(nth_lev) % N,  &
           & HIERARCHICAL_DATA(nth_lev) % NP, &
           & HIERARCHICAL_DATA(nth_lev) % NPL,&
           & HIERARCHICAL_DATA(nth_lev) % NPU,&
           & HIERARCHICAL_DATA(nth_lev) % D,  &
           & HIERARCHICAL_DATA(nth_lev) % AL, &
           & HIERARCHICAL_DATA(nth_lev) % INL,&
           & HIERARCHICAL_DATA(nth_lev) % IAL,&
           & HIERARCHICAL_DATA(nth_lev) % AU, &
           & HIERARCHICAL_DATA(nth_lev) % INU,&
           & HIERARCHICAL_DATA(nth_lev) % IAU,&
           & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE,   &
           & WS, WR, dim, SOLVER_COMM, my_rank)         


      X => HIERARCHICAL_DATA(nth_lev) % X

      !$       color = HIERARCHICAL_DATA(nth_lev) % color
      !$       col_ind => HIERARCHICAL_DATA(nth_lev) % color_index
      !$       INML => HIERARCHICAL_DATA(nth_lev) % INML

      B => HIERARCHICAL_DATA(nth_lev) % B

      CALL count_time(1, SOLVER_COMM,  my_rank,8)

      if(HIERARCHICAL_DATA(nth_lev) % scaling_fl) then
         call scale_mul_vec(X, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % NP)
      end if


      if(gs_kind == 4) then
         l =iteration_number_down
         do j = 4, 2, -1
            k = l / j
            if(k > 0) then
               call cheby_sm_sca(mat, B, X, k, j, &
                    & HIERARCHICAL_DATA(nth_lev) % cheb_h, &
                    & HIERARCHICAL_DATA(nth_lev) % cheb_l, &
                    & HIERARCHICAL_DATA(nth_lev) % cheb_fl)
            end if
            l = l - k*j
         end do
      else
         if(omp_switch == 1 .and. gs_kind >=2) then
            call mcsgs(mat, B, X, iteration_number_down, color, col_ind, INML, &
                 & dump_gauss_seidel_weight, (gs_kind-2))
         else if(gs_kind >=2) then
#ifdef PRINT_REZ
            write(*,*) "mcgs is selected in not OpenMP mode."
#endif
            gs_kind = gs_kind -2
         end if
         CALL sgs(mat, B, X, iteration_number_down, dump_gauss_seidel_weight, gs_kind)
      end if

      CALL count_time(2, SOLVER_COMM, my_rank,8)
      !C barrier in ssor ##############

      if(HIERARCHICAL_DATA(nth_lev) % scaling_fl) then
         call scale_div_vec(X,    HIERARCHICAL_DATA(nth_lev) % scale_D, mat % NP)
         call scale_mul_vec(Temp, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % N)
      end if

#ifdef MEASURING
      CALL count_time(2, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

    END SUBROUTINE lev1_mod



    !C As for scaling, rhss vec on PRE_LEVEL is assumed to be set before calling this sub
    SUBROUTINE v_cycle_mod(LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, Temp, &
         &             NPROCS, dim, PRE_LEVEL, POST_LEVEL)
      USE comm_switch_mod
      USE data_structure_for_AMG
      USE solver_Gnumbering
      USE count_time_mod
      USE amg_parameter, only: iteration_number_up, iteration_number_down, &
           & iteration_number_bottom, gs_kind, dump_gauss_seidel_weight, W_OUT_LU, &
           & iterative_parameter_set_at_each_lev
      use matrix_type_mod, only: matrix, set_mat_data
      USE chebyshev_smoother, only: cheby_sm_sca

      !$    USE omp_lib

      IMPLICIT NONE
      include 'mpif.h'
      INTEGER(kind=kint), INTENT(in)    :: NPROCS, dim, LEVEL_NUM
      INTEGER(kind=kint), INTENT(in)    :: PRE_LEVEL, POST_LEVEL
      INTEGER(kind=kint)                :: N,NP,NPL,NPU

      !C  Temp(1:problem_NP)
      REAL   (kind=kreal)  :: Temp(:)
      !C  WS(1:dim), WR(1:dim)
      REAL   (kind=kreal)  :: WS(:), WR(:)


      !C DIMENSION(NP)
      REAL   (kind=kreal), DIMENSION(:), pointer:: D

      REAL   (kind=kreal), DIMENSION(:), pointer::  B
      !C DIMENSION(NP)
      REAL   (kind=kreal), DIMENSION(:), pointer::  X
      !C DIMENSION(NP)

      REAL   (kind=kreal), DIMENSION(:), pointer::  AU
      REAL   (kind=kreal), DIMENSION(:), pointer::  AL

      INTEGER(kind=kint), DIMENSION(:), pointer ::  INU
      !C DIMENSION(0:N)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  IAU
      !C DIMENSION(NPU)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  INL
      !C DIMENSION(0:N)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  IAL

      INTEGER(kind=kint), intent(in) :: SOLVER_COMM
      INTEGER(kind=kint), intent(in) :: my_rank
      INTEGER(kind=kint) :: ierr

      REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_B
      REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_X

      INTEGER(kind=kint ) :: nth_lev,inod,ieL,isL,is,ie,coarser_N,isU,ieU,coarser_NP
      INTEGER(kind=kint ) :: coarser_NEIBPETOT,i,j,k,l
      REAL   (kind=kreal) :: R_val,B_val,w,X_val,R_norm,GR_norm

      type(INTER_LEVEL_OPERATOR), pointer :: R, P

      INTEGER(kind=kint) :: NEIBPETOT, iNEIBPETOT
      INTEGER(kind=kint), pointer :: NEIBPE(:)
      INTEGER(kind=kint), pointer :: STACK_IMPORT(:),STACK_EXPORT(:)
      INTEGER(kind=kint), pointer :: NOD_IMPORT(:),NOD_EXPORT(:)


      INTEGER(kind=kint) :: color,  omp_switch=0
      INTEGER(kind=kint), pointer :: col_ind(:)
      INTEGER(kind=kint), pointer :: INML(:)
      type(matrix) :: mat
      INTEGER :: INFO
      
      !C if OpenMP is validated, then ompswitch = 1
      !$ omp_switch = 1
      DO nth_lev = PRE_LEVEL, LEVEL_NUM-1

         call iterative_parameter_set_at_each_lev(nth_lev)
#ifdef MEASURING
         CALL count_time(1, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

         call set_mat_data(mat, &
              & HIERARCHICAL_DATA(nth_lev) % N,  &
              & HIERARCHICAL_DATA(nth_lev) % NP, &
              & HIERARCHICAL_DATA(nth_lev) % NPL,&
              & HIERARCHICAL_DATA(nth_lev) % NPU,&
              & HIERARCHICAL_DATA(nth_lev) % D,  &
              & HIERARCHICAL_DATA(nth_lev) % AL, &
              & HIERARCHICAL_DATA(nth_lev) % INL,&
              & HIERARCHICAL_DATA(nth_lev) % IAL,&
              & HIERARCHICAL_DATA(nth_lev) % AU, &
              & HIERARCHICAL_DATA(nth_lev) % INU,&
              & HIERARCHICAL_DATA(nth_lev) % IAU,&
              & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE,   &
              & WS, WR, dim, SOLVER_COMM, my_rank)         
         
         
         X => HIERARCHICAL_DATA(nth_lev) % X

         !$       color = HIERARCHICAL_DATA(nth_lev) % color
         !$       col_ind => HIERARCHICAL_DATA(nth_lev) % color_index
         !$       INML => HIERARCHICAL_DATA(nth_lev) % INML

         B => HIERARCHICAL_DATA(nth_lev) % B
         coarser_B => HIERARCHICAL_DATA(nth_lev+1) % B
         coarser_X => HIERARCHICAL_DATA(nth_lev+1) % X

         CALL count_time(1, SOLVER_COMM,  my_rank,8)
         
         if(HIERARCHICAL_DATA(nth_lev) % scaling_fl) then
            call scale_mul_vec(X, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % NP)
            if(nth_lev /= PRE_LEVEL) then
               call scale_div_vec(B, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % N)
            end if
         end if
         
         
         if(gs_kind == 4) then
            l = iteration_number_down
            do j = 4, 2, -1
               k = l / j
               if(k > 0) then
                  call cheby_sm_sca(mat, B, X, k, j, &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_h, &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_l, &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_fl)
               end if
               l = l - k*j
            end do
         else if(omp_switch == 1 .and. gs_kind >=2) then
            call mcsgs(mat, B, X, iteration_number_down, color, col_ind, INML, &
                 & dump_gauss_seidel_weight, (gs_kind-2))
         else
            if(gs_kind >=2) then
#ifdef PRINT_REZ
               write(*,*) "mcgs is selected in not OpenMP mode."
#endif
               gs_kind = gs_kind -2
            end if
            CALL sgs(mat, B, X, iteration_number_down, dump_gauss_seidel_weight, gs_kind)
         end if

         CALL count_time(2, SOLVER_COMM, my_rank,8)
         !C barrier in ssor ##############
         
         !C calculate residual
         call cal_res_vec_at_nth_lev(B, X, Temp, nth_lev)

         if(HIERARCHICAL_DATA(nth_lev) % scaling_fl) then
            call scale_div_vec(X,    HIERARCHICAL_DATA(nth_lev) % scale_D, mat % NP)
            call scale_mul_vec(Temp, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % N)
         end if

#ifdef MEASURING
         CALL count_time(2, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

#ifdef MEASURING
         CALL count_time(1, SOLVER_COMM,  my_rank, 50+nth_lev)
#endif

         
         !C restrict the residual vector
         call restrict_procedure(nth_lev, Temp, coarser_B, WS, WR, &
              & dim, my_rank, SOLVER_COMM)
         coarser_NP=HIERARCHICAL_DATA(nth_lev+1) % NP
         
         !$omp parallel do       
         DO j=1,coarser_NP
            coarser_X(j)= 0.0
         END DO
         !$omp end parallel do

#ifdef MEASURING
         CALL count_time(2, SOLVER_COMM,  my_rank, 50+nth_lev)
#endif
         
      END DO

#ifdef MEASURING
      CALL count_time(1, SOLVER_COMM,  my_rank, 40+LEVEL_NUM)
#endif

      call iterative_parameter_set_at_each_lev(LEVEL_NUM)

      call set_mat_data(mat, &
           & HIERARCHICAL_DATA(LEVEL_NUM) % N,  &
           & HIERARCHICAL_DATA(LEVEL_NUM) % NP, &
           & HIERARCHICAL_DATA(LEVEL_NUM) % NPL,&
           & HIERARCHICAL_DATA(LEVEL_NUM) % NPU,&
           & HIERARCHICAL_DATA(LEVEL_NUM) % D,  &
           & HIERARCHICAL_DATA(LEVEL_NUM) % AL, &
           & HIERARCHICAL_DATA(LEVEL_NUM) % INL,&
           & HIERARCHICAL_DATA(LEVEL_NUM) % IAL,&
           & HIERARCHICAL_DATA(LEVEL_NUM) % AU, &
           & HIERARCHICAL_DATA(LEVEL_NUM) % INU,&
           & HIERARCHICAL_DATA(LEVEL_NUM) % IAU,&
           & HIERARCHICAL_DATA(LEVEL_NUM) % COMM_TABLE, &
           & WS, WR, dim, SOLVER_COMM, my_rank)         

      X   => HIERARCHICAL_DATA(LEVEL_NUM) % X
      B   => HIERARCHICAL_DATA(LEVEL_NUM) % B

      !$ color= HIERARCHICAL_DATA(LEVEL_NUM) %color
      !$ col_ind => HIERARCHICAL_DATA(LEVEL_NUM) % color_index
      !$ INML => HIERARCHICAL_DATA(LEVEL_NUM) % INML

      CALL count_time(1,SOLVER_COMM,my_rank,8)

      if(HIERARCHICAL_DATA(LEVEL_NUM) % scaling_fl) then
         call scale_mul_vec(X, HIERARCHICAL_DATA(LEVEL_NUM) % scale_D, mat % NP)
         call scale_div_vec(B, HIERARCHICAL_DATA(LEVEL_NUM) % scale_D, mat % N)
      end if

      if(.not.W_OUT_LU) then
         if(mat % N > 0) then
            X = B
!C            call lusub(coarsest_A, mat % N, coarsest_IDX, X)
            call DGETRS('N', mat % N, 1, coarsest_A, mat % N, coarsest_IDX, X, mat % N, INFO)
            if(INFO/=0) stop 'forward/backward substitution error!!'
         end if
      else
         if(gs_kind == 4) then
            call cheby_sm_sca(mat, B, X, iteration_number_bottom, 4, &
                 & HIERARCHICAL_DATA(LEVEL_NUM) % cheb_h, &
                 & HIERARCHICAL_DATA(LEVEL_NUM) % cheb_l, &
                 & HIERARCHICAL_DATA(LEVEL_NUM) % cheb_fl)
         else if(omp_switch == 1 .and. gs_kind >= 2) then
            CALL mcsgs(mat, B, X, iteration_number_bottom, color, col_ind, INML, &
                 & dump_gauss_seidel_weight,0)
         else
            CALL sgs(mat, B, X, iteration_number_bottom, dump_gauss_seidel_weight, 0)
         end if
      end if

      if(HIERARCHICAL_DATA(LEVEL_NUM) % scaling_fl) then
         call scale_div_vec(X, HIERARCHICAL_DATA(LEVEL_NUM) % scale_D, mat % NP)
      end if

      CALL count_time(2,SOLVER_COMM,my_rank,8)

      !C barrier in ssor ########

#ifdef MEASURING
      CALL count_time(2, SOLVER_COMM,  my_rank, 40+LEVEL_NUM)
#endif

      
      DO nth_lev = LEVEL_NUM-1, POST_LEVEL, -1

         call iterative_parameter_set_at_each_lev(nth_lev)

#ifdef MEASURING
         CALL count_time(1, SOLVER_COMM,  my_rank, 50+nth_lev)
#endif


         call set_mat_data(mat, &
              & HIERARCHICAL_DATA(nth_lev) % N,  &
              & HIERARCHICAL_DATA(nth_lev) % NP, &
              & HIERARCHICAL_DATA(nth_lev) % NPL,&
              & HIERARCHICAL_DATA(nth_lev) % NPU,&
              & HIERARCHICAL_DATA(nth_lev) % D,  &
              & HIERARCHICAL_DATA(nth_lev) % AL, &
              & HIERARCHICAL_DATA(nth_lev) % INL,&
              & HIERARCHICAL_DATA(nth_lev) % IAL,&
              & HIERARCHICAL_DATA(nth_lev) % AU, &
              & HIERARCHICAL_DATA(nth_lev) % INU,&
              & HIERARCHICAL_DATA(nth_lev) % IAU,&
              & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE,   &
              & WS, WR, dim, SOLVER_COMM, my_rank)         
         
         X            => HIERARCHICAL_DATA(nth_lev) % X
         B            => HIERARCHICAL_DATA(nth_lev) % B
         
         !$ Color = HIERARCHICAL_DATA(nth_lev) % color
         !$ col_ind => HIERARCHICAL_DATA(nth_lev) % color_index
         !$ INML => HIERARCHICAL_DATA(nth_lev) % INML

         coarser_X    => HIERARCHICAL_DATA(nth_lev+1) % X
         
         call prolong_and_add_procedure(nth_lev, X, coarser_X, WS, WR, dim, my_rank, SOLVER_COMM)
         
#ifdef MEASURING
         CALL count_time(2, SOLVER_COMM,  my_rank, 50+nth_lev)
#endif

#ifdef MEASURING
         CALL count_time(1, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

         CALL count_time(1,SOLVER_COMM,my_rank,8)
         
         if(HIERARCHICAL_DATA(nth_lev) % scaling_fl) then
            call scale_mul_vec(X, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % NP)
         end if

         
         if(gs_kind == 4) then
            l = iteration_number_up
            do j = 4, 2, -1
               k = l / j
               if(k > 0) then
                  call cheby_sm_sca(mat, B, X, k, j,           &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_h,  &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_l,  &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_fl)
               end if
               l = l - k*j
            end do
            
         else if(omp_switch == 1 .and. gs_kind >= 2) then
            !C gs_kind: 0 sgs, 1 gs, 2 gs in reverse order
            CALL mcsgs(mat, B, X, iteration_number_up, color, col_ind, INML, &
                 & dump_gauss_seidel_weight,(gs_kind-2)*2)
         else
            if(gs_kind >=2) then
#ifdef PRINT_REZ
               write(*,*) "mcgs is selected in not OpenMP mode."
#endif
               gs_kind = gs_kind -2
            end if
            CALL sgs(mat, B, X, iteration_number_up, dump_gauss_seidel_weight, gs_kind*2)
            
         end if

         if(HIERARCHICAL_DATA(nth_lev) % scaling_fl) then
            call scale_div_vec(X, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % NP)
         end if
         
         CALL count_time(2,SOLVER_COMM,my_rank,8)
         !C barrier in ssor ########
#ifdef MEASURING
         CALL count_time(2, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

      END DO

    END SUBROUTINE v_cycle_mod

    !C nth_lev+1->nth_lev
    subroutine prolong_and_add_procedure(nth_lev, finer_vec, coarser_vec, &
         & WS, WR, dim, my_rank, SOLVER_COMM)
      USE comm_switch_mod
      USE data_structure_for_AMG
      USE solver_Gnumbering
      !$ USE omp_lib
      
      implicit none

      integer(kind=kint), intent(in) :: my_rank, SOLVER_COMM, nth_lev, dim
      integer(kind=kint) :: j, is, ie, i 
      integer(kind=kint) :: iNEIBPETOT, NEIBPETOT, coarser_NP, inod, N, NP
      real(kind=kreal)   :: X_val
      real(kind=kreal)   :: finer_vec(:), coarser_vec(:), WS(:), WR(:)
      real(kind=kreal),allocatable,save   :: tmp(:)
      integer(kind=kint),save   :: tmp_size=0

      N= HIERARCHICAL_DATA(nth_lev) % N
      NP=HIERARCHICAL_DATA(nth_lev) % NP
      
      iNEIBPETOT = HIERARCHICAL_DATA(nth_lev+1) % INT_LVL_TABLE % NEIBPETOT 

      call comm_sr(SOLVER_COMM,                            &
           & HIERARCHICAL_DATA(nth_lev+1) % INT_LVL_TABLE, &
           & HIERARCHICAL_DATA(nth_lev+1) % NP,            &
           & WS, WR, coarser_vec)
      
      if(.not. HIERARCHICAL_DATA(nth_lev+1) % PRO_AGGRE_OPT) then
         
         !$omp parallel do private(j, is,ie,X_val,i,inod)
         DO j= 1, N
            is= HIERARCHICAL_DATA(nth_lev+1) % P % IN(j-1)+1
            ie= HIERARCHICAL_DATA(nth_lev+1) % P % IN(j)
            X_val=0.0
            DO i=is, ie
               inod = HIERARCHICAL_DATA(nth_lev+1) % P % CN(i)
               X_val= X_val+ &
                    & HIERARCHICAL_DATA(nth_lev+1) % P % V(i) * coarser_vec(inod)
            END DO
            finer_vec(j)=finer_vec(j)+X_val
         END DO
         !$omp end parallel do
      else
         if(NP > tmp_size) then
            if(allocated(tmp)) deallocate(tmp)
            allocate(tmp(NP))
            tmp_size = NP
         end if
         
         tmp=0

         !$omp parallel do private(j, is, ie,X_val,i,inod)
         DO j= 1, NP
            is= HIERARCHICAL_DATA(nth_lev+1) % P % IN(j-1)+1
            ie= HIERARCHICAL_DATA(nth_lev+1) % P % IN(j)
            X_val=0.0
            DO i=is, ie
               inod = HIERARCHICAL_DATA(nth_lev+1) % P % CN(i)
               X_val= X_val+ &
                    & HIERARCHICAL_DATA(nth_lev+1) % P % V(i) * coarser_vec(inod)
            END DO
            tmp(j)=X_val
         END DO
         !$omp end parallel do
         
         call comm_sr_rev_add(SOLVER_COMM,                    &
              & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE,      &
              & NP, WS, WR, tmp)
         
         !$omp parallel do private(j, is, ie,X_val,i,inod)
         DO j= 1, N
            finer_vec(j)=finer_vec(j)+tmp(j)
         END DO
         !$omp end parallel do
         
      end if
      
      NEIBPETOT = HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NEIBPETOT
      call comm_sr(SOLVER_COMM,                            &
           & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE,      &
           & NP, WS, WR, finer_vec)
         
    END subroutine prolong_and_add_procedure


    !C nth_lev->nth_lev+1
    subroutine restrict_procedure(nth_lev, finer_vec, coarser_vec, WS, WR, &
         & dim, my_rank, SOLVER_COMM)
      USE comm_switch_mod
      USE data_structure_for_AMG
      !$ USE omp_lib

      implicit none
      
      integer(kind=kint), intent(in) :: my_rank, SOLVER_COMM, nth_lev, dim
      integer(kind=kint) :: j, is, ie, i, iNEIBPETOT, coarser_NP, inod
      real(kind=kreal)   :: B_val
      real(kind=kreal)   :: finer_vec(:), coarser_vec(:), WS(:), WR(:)

      coarser_NP=HIERARCHICAL_DATA(nth_lev+1) % NP
      if(.not. HIERARCHICAL_DATA(nth_lev+1) % PRO_AGGRE_OPT) then
         
         !$omp parallel private(j,is,ie,B_val,i,inod)
         !$omp do 
         DO j= 1, coarser_NP
            is= HIERARCHICAL_DATA(nth_lev+1) % R % IN(j-1)+1
            ie= HIERARCHICAL_DATA(nth_lev+1) % R % IN(j)
            B_val=0.0
            DO i=is, ie
               inod = HIERARCHICAL_DATA(nth_lev+1) % R % CN(i)
               B_val= B_val + HIERARCHICAL_DATA(nth_lev+1) % R % V(i) * finer_vec(inod)
            END DO
            coarser_vec(j)= B_val
         END DO
         !$omp end do
         !$omp end parallel

      else
         
         call comm_sr(SOLVER_COMM, &
              & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE, &
              & HIERARCHICAL_DATA(nth_lev) % NP,         &
              & WS, WR, finer_vec)
         
         coarser_vec = 0

         !$omp parallel private(j,is,ie,B_val,i,inod)
         !$omp do 
         DO j= 1, HIERARCHICAL_DATA(nth_lev+1) % R % ROW_SIZE
            is= HIERARCHICAL_DATA(nth_lev+1) % R % IN(j-1)+1
            ie= HIERARCHICAL_DATA(nth_lev+1) % R % IN(j)
            B_val=0.0
            DO i=is, ie
               inod = HIERARCHICAL_DATA(nth_lev+1) % R % CN(i)
               B_val= B_val + HIERARCHICAL_DATA(nth_lev+1) % R % V(i) * finer_vec(inod)
            END DO
            coarser_vec(j)= B_val
         END DO
         !$omp end do

         !$omp do
         DO j = HIERARCHICAL_DATA(nth_lev+1) % R % ROW_SIZE+1, coarser_NP
            coarser_vec(j) = 0
         END DO
         !$omp end do
         !$omp end parallel

      end if
      call comm_sr_rev_add(SOLVER_COMM, &
              & HIERARCHICAL_DATA(nth_lev+1) % INT_LVL_TABLE, &
              & coarser_NP, WS, WR, coarser_vec)


    end subroutine restrict_procedure

    subroutine scale_mul_vec(vec, scale_vec, size)
      implicit none
      include 'precision.inc'
      real(kind=kreal) :: vec(:), scale_vec(:)
      integer(kind=kint), intent(in) :: size
      integer(kind=kint) :: i,j

      !$omp parallel do
      DO i =1, size
         vec(i) =  vec(i)*scale_vec(i) 
      END DO
      !$omp end parallel do
      
    end subroutine scale_mul_vec

    subroutine scale_div_vec(vec, scale_vec, size)
      implicit none
      include 'precision.inc'
      real(kind=kreal) :: vec(:), scale_vec(:)
      integer(kind=kint), intent(in) :: size
      integer(kind=kint) :: i,j
      
      !$omp parallel do      
      DO i =1, size
         vec(i) =  vec(i)/scale_vec(i) 
      END DO
      !$omp end parallel do
      
    end subroutine scale_div_vec

    


    SUBROUTINE f_cycle(problem_B, problem_X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, Temp, &
         &             NPROCS, dim, problem_NP)

      USE data_structure_for_AMG
      USE solver_Gnumbering
      USE count_time_mod
      USE chebyshev_smoother, only: cheby_sm_sca
      USE amg_parameter, only: f_first_iteration_number, gs_kind, dump_gauss_seidel_weight
      use matrix_type_mod, only:matrix, set_mat_data
      !$    USE omp_lib

      IMPLICIT NONE
      include 'mpif.h'
      type(matrix) :: mat
      INTEGER(kind=kint), INTENT(in)    :: NPROCS, dim, LEVEL_NUM
      INTEGER(kind=kint), INTENT(in)    :: problem_NP
      INTEGER(kind=kint)                :: N,NP,NPL,NPU

      !C  problem_B(1:problem_NP)
      REAL   (kind=kreal), target::  problem_B(:)
      !C  problem_X(1:problem_NP)   
      REAL   (kind=kreal), target::  problem_X(:)   

      !C  Temp(1:problem_NP)
      REAL   (kind=kreal)  :: Temp(:)
      !C  WS(1:dim), WR(1:dim)
      REAL   (kind=kreal)  :: WS(:), WR(:)


      !C DIMENSION(NP)
      REAL   (kind=kreal), DIMENSION(:), pointer:: D

      REAL   (kind=kreal), DIMENSION(:), pointer::  B
      !C DIMENSION(NP)
      REAL   (kind=kreal), DIMENSION(:), pointer::  X
      !C DIMENSION(NP)

      REAL   (kind=kreal), DIMENSION(:), pointer::  AU
      REAL   (kind=kreal), DIMENSION(:), pointer::  AL

      INTEGER(kind=kint), DIMENSION(:), pointer ::  INU
      !C DIMENSION(0:N)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  IAU
      !C DIMENSION(NPU)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  INL
      !C DIMENSION(0:N)
      INTEGER(kind=kint), DIMENSION(:), pointer ::  IAL

      INTEGER(kind=kint), intent(in) :: SOLVER_COMM
      INTEGER(kind=kint), intent(in) :: my_rank
      INTEGER(kind=kint) :: ierr

      REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_B
      REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_X

      INTEGER(kind=kint ) :: nth_lev,i,inod,ieL,isL,is,ie,coarser_N,isU,ieU,coarser_NP
      INTEGER(kind=kint ) :: coarser_NEIBPETOT,j,k,l
      REAL   (kind=kreal) :: R_val,B_val,w,X_val,R_norm,GR_norm

      type(INTER_LEVEL_OPERATOR), pointer :: R, P

      INTEGER(kind=kint) :: NEIBPETOT, iNEIBPETOT
      INTEGER(kind=kint), pointer :: NEIBPE(:)
      INTEGER(kind=kint), pointer :: STACK_IMPORT(:),STACK_EXPORT(:)
      INTEGER(kind=kint), pointer :: NOD_IMPORT(:),NOD_EXPORT(:)



      INTEGER(kind=kint) :: color, omp_switch=0
      INTEGER(kind=kint), pointer :: col_ind(:)
      INTEGER(kind=kint), pointer :: INML(:)

      call count_time(1, SOLVER_COMM, my_rank, 9)

      

      HIERARCHICAL_DATA(1) % B => problem_B
      HIERARCHICAL_DATA(1) % X => problem_X

      !C if OpenMP is validated, then omp_switch = 1
      !$ omp_switch = 1
      
      DO nth_lev = 1, LEVEL_NUM-1

#ifdef MEASURING
         CALL count_time(1, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

         
         N  = HIERARCHICAL_DATA(nth_lev) % N
         NP = HIERARCHICAL_DATA(nth_lev) % NP
         NPL= HIERARCHICAL_DATA(nth_lev) % NPL
         NPU= HIERARCHICAL_DATA(nth_lev) % NPU
         INU=>HIERARCHICAL_DATA(nth_lev) % INU
         INL=>HIERARCHICAL_DATA(nth_lev) % INL
         IAU=>HIERARCHICAL_DATA(nth_lev) % IAU


         X => HIERARCHICAL_DATA(nth_lev) % X

         !$       color = HIERARCHICAL_DATA(nth_lev) % color
         !$       col_ind => HIERARCHICAL_DATA(nth_lev) % color_index
         !$       INML => HIERARCHICAL_DATA(nth_lev) % INML


         B => HIERARCHICAL_DATA(nth_lev) % B
         D => HIERARCHICAL_DATA(nth_lev) % D
         IAL=>HIERARCHICAL_DATA(nth_lev) % IAL
         AU =>HIERARCHICAL_DATA(nth_lev) % AU
         AL =>HIERARCHICAL_DATA(nth_lev) % AL
         NEIBPETOT    =  HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NEIBPETOT
         STACK_IMPORT => HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % STACK_IMPORT
         STACK_EXPORT => HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % STACK_EXPORT
         NOD_IMPORT   => HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NOD_IMPORT
         NOD_EXPORT   => HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NOD_EXPORT
         NEIBPE       => HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NEIBPE

         call set_mat_data(mat, &
              & HIERARCHICAL_DATA(nth_lev) % N,  &
              & HIERARCHICAL_DATA(nth_lev) % NP, &
              & HIERARCHICAL_DATA(nth_lev) % NPL,&
              & HIERARCHICAL_DATA(nth_lev) % NPU,&
              & HIERARCHICAL_DATA(nth_lev) % D,  &
              & HIERARCHICAL_DATA(nth_lev) % AL, &
              & HIERARCHICAL_DATA(nth_lev) % INL,&
              & HIERARCHICAL_DATA(nth_lev) % IAL,&
              & HIERARCHICAL_DATA(nth_lev) % AU, &
              & HIERARCHICAL_DATA(nth_lev) % INU,&
              & HIERARCHICAL_DATA(nth_lev) % IAU,&
              & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE,   &
              & WS, WR, dim, SOLVER_COMM, my_rank)         
         
         

         coarser_B => HIERARCHICAL_DATA(nth_lev+1) % B
         coarser_X => HIERARCHICAL_DATA(nth_lev+1) % X

         CALL count_time(1,SOLVER_COMM,my_rank,8)

         if(HIERARCHICAL_DATA(nth_lev) % scaling_fl) then
            call scale_mul_vec(X, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % NP)
            call scale_div_vec(B, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % N)
         end if

         
         if(gs_kind == 4) then
            l = f_first_iteration_number
            do j = 4, 2, -1
               k = l / j
               if(k > 0) then
                  call cheby_sm_sca(mat, B, X, k, j,          &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_h, &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_l, &
                       & HIERARCHICAL_DATA(nth_lev) % cheb_fl)
               end if
               l = l - k*j               
            end do

         else if(omp_switch == 1 .and. gs_kind >=2) then
            call mcsgs(mat,B, X, f_first_iteration_number,color,col_ind,INML,&
                 & dump_gauss_seidel_weight,(gs_kind-2))
         else
            if(gs_kind >=2) then
#ifdef PRINT_REZ
               write(*,*) "mcgs is selected in not OpenMP mode."
#endif
               gs_kind = gs_kind -2
            end if

            CALL sgs(mat,B, X,f_first_iteration_number, dump_gauss_seidel_weight, gs_kind)
         end if

         CALL count_time(2,SOLVER_COMM,my_rank,8)
         !C barrier in ssor ##############

         call cal_res_vec_at_nth_lev(B, X, Temp, nth_lev)

         if(HIERARCHICAL_DATA(nth_lev) % scaling_fl) then
            call scale_div_vec(X,    HIERARCHICAL_DATA(nth_lev) % scale_D, mat % NP)
            call scale_mul_vec(Temp, HIERARCHICAL_DATA(nth_lev) % scale_D, mat % N)
         end if
#ifdef MEASURING
         CALL count_time(2, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif


#ifdef MEASURING
         CALL count_time(1, SOLVER_COMM,  my_rank, 50+nth_lev)
#endif


         !C restrict the residual vector
         call restrict_procedure(nth_lev, Temp, coarser_B, WS, WR, &
              & dim, my_rank, SOLVER_COMM)
         coarser_NP = HIERARCHICAL_DATA(nth_lev+1) % NP
         
         !$omp parallel do
         DO j = 1,coarser_NP
            coarser_X(j)= 0.0
         END DO
         !$omp end parallel do

#ifdef MEASURING
         CALL count_time(2, SOLVER_COMM,  my_rank, 50+nth_lev)
#endif

      END DO

      DO nth_lev = LEVEL_NUM-1, 1, -1
         call v_cycle_mod(LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, &
              & Temp, NPROCS, dim, nth_lev+1, nth_lev )
      END DO

      call count_time(2, SOLVER_COMM, my_rank, 9)

    END SUBROUTINE f_cycle


    SUBROUTINE v_cycle(problem_B, problem_X, LEVEL_NUM, SOLVER_COMM,&
         & my_rank, WS, WR, Temp, NPROCS, dim, problem_NP)
      USE data_structure_for_AMG
      USE count_time_mod
      USE amg_parameter, only: iteration_number_up, iteration_number_down, iteration_number_bottom

      IMPLICIT NONE
      include 'mpif.h'
      INTEGER(kind=kint), INTENT(in)    :: NPROCS
      INTEGER(kind=kint), INTENT(in)    :: dim, problem_NP
      INTEGER(kind=kint), INTENT(in)    :: LEVEL_NUM
      INTEGER(kind=kint), intent(in) :: SOLVER_COMM
      INTEGER(kind=kint), intent(in) :: my_rank
      !C  Temp(1:problem_NP)
      REAL   (kind=kreal)  :: Temp(:)
      !C  WS(1:dim), WR(1:dim)
      REAL   (kind=kreal)  :: WS(:), WR(:)
      !C  problem_B(1:problem_NP)
      REAL   (kind=kreal), target::  problem_B(:)
      !C  problem_X(1:problem_NP)   
      REAL   (kind=kreal), target::  problem_X(:)   

      call count_time(1, SOLVER_COMM, my_rank, 9)

      HIERARCHICAL_DATA(1) % B => problem_B
      HIERARCHICAL_DATA(1) % X => problem_X

      call v_cycle_mod(LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, &
           & Temp, NPROCS, dim, 1, 1 )
      call count_time(2, SOLVER_COMM, my_rank, 9)

    END SUBROUTINE v_cycle


    SUBROUTINE smoother_lev_1(problem_B, problem_X, LEVEL_NUM, SOLVER_COMM,&
         & my_rank, WS, WR, Temp, NPROCS, dim, problem_NP)
      USE data_structure_for_AMG
      USE count_time_mod
      USE amg_parameter, only: iteration_number_up, iteration_number_down, iteration_number_bottom

      IMPLICIT NONE
      include 'mpif.h'
      INTEGER(kind=kint), INTENT(in)    :: NPROCS
      INTEGER(kind=kint), INTENT(in)    :: dim, problem_NP
      INTEGER(kind=kint), INTENT(in)    :: LEVEL_NUM
      INTEGER(kind=kint), intent(in) :: SOLVER_COMM
      INTEGER(kind=kint), intent(in) :: my_rank
      !C  Temp(1:problem_NP)
      REAL   (kind=kreal)  :: Temp(:)
      !C  WS(1:dim), WR(1:dim)
      REAL   (kind=kreal)  :: WS(:), WR(:)
      !C  problem_B(1:problem_NP)
      REAL   (kind=kreal), target::  problem_B(:)
      !C  problem_X(1:problem_NP)   
      REAL   (kind=kreal), target::  problem_X(:)   

      call count_time(1, SOLVER_COMM, my_rank, 9)

      HIERARCHICAL_DATA(1) % B => problem_B
      HIERARCHICAL_DATA(1) % X => problem_X

      call lev1_mod(SOLVER_COMM, my_rank, WS, WR, Temp, &
           &             NPROCS, dim)
      
      call count_time(2, SOLVER_COMM, my_rank, 9)

    END SUBROUTINE smoother_lev_1






    SUBROUTINE w_cycle(problem_B, problem_X, LEVEL_NUM,     &
         & SOLVER_COMM, my_rank, WS, WR, Temp, NPROCS, dim, &
         & problem_NP)
      USE data_structure_for_AMG
      USE count_time_mod

      IMPLICIT NONE
      include 'mpif.h'
      INTEGER(kind=kint), INTENT(in)    :: NPROCS
      INTEGER(kind=kint), INTENT(in)    :: dim, problem_NP
      INTEGER(kind=kint), INTENT(in)    :: LEVEL_NUM
      INTEGER(kind=kint), intent(in) :: SOLVER_COMM
      INTEGER(kind=kint), intent(in) :: my_rank
      !C  Temp(1:problem_NP)
      REAL   (kind=kreal)  :: Temp(:)
      !C  WS(1:dim), WR(1:dim)
      REAL   (kind=kreal)  :: WS(:), WR(:)
      !C  problem_B(1:problem_NP)
      REAL   (kind=kreal), target::  problem_B(:)
      !C  problem_X(1:problem_NP)   
      REAL   (kind=kreal), target::  problem_X(:)   
      integer(kind=kint) :: w_lev

      call count_time(1, SOLVER_COMM, my_rank, 9)


      w_lev = LEVEL_NUM/2 + mod(LEVEL_NUM, 2)

      HIERARCHICAL_DATA(1) % B => problem_B
      HIERARCHICAL_DATA(1) % X => problem_X

      call v_cycle_mod(LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, Temp, &
           &             NPROCS, dim, 1, w_lev )
      call v_cycle_mod(LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, Temp, &
           &             NPROCS, dim, w_lev, 1 )

      call count_time(2, SOLVER_COMM, my_rank, 9)

    END SUBROUTINE w_cycle


    SUBROUTINE amg_cycles(problem_B, problem_X, LEVEL_NUM, &
         & SOLVER_COMM, my_rank, WS, WR, Temp, NPROCS, dim, &
         & problem_NP)
      USE amg_parameter, only: cycle_kind

      IMPLICIT NONE
      include 'mpif.h'
      include 'precision.inc'
      INTEGER(kind=kint), INTENT(in)    :: NPROCS
      INTEGER(kind=kint), INTENT(in)    :: dim, problem_NP
      INTEGER(kind=kint), INTENT(in)    :: LEVEL_NUM
      INTEGER(kind=kint), intent(in) :: SOLVER_COMM
      INTEGER(kind=kint), intent(in) :: my_rank
      !C  Temp(1:problem_NP)
      REAL   (kind=kreal)  :: Temp(:)
      !C  WS(1:dim), WR(1:dim)
      REAL   (kind=kreal)  :: WS(:), WR(:)
      !C  problem_B(1:problem_NP)
      REAL   (kind=kreal), target::  problem_B(:)
      !C  problem_X(1:problem_NP)   
      REAL   (kind=kreal), target::  problem_X(:)   

      if(cycle_kind == 0) then
         call v_cycle(problem_B, problem_X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, Temp, &
              &             NPROCS, dim, problem_NP)
      else if(cycle_kind==1) then
         call f_cycle(problem_B, problem_X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, Temp, &
              &             NPROCS, dim, problem_NP)
      else if(cycle_kind==2) then
         call w_cycle(problem_B, problem_X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, Temp, &
              &             NPROCS, dim, problem_NP)
      else if(cycle_kind==3) then
         call smoother_lev_1(problem_B, problem_X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR,&
              &             Temp, NPROCS, dim, problem_NP)
      else 
         write(*,*) "Cycle specified by AMG.dat is set as", cycle_kind
         write(*,*) "Cycle should be chosen among 0:V-cycle, 1:F_cycle, 2:W_cycle"
         stop
      end if

    END SUBROUTINE amg_cycles


    SUBROUTINE ll_slv(lower_mat, b, N)
      IMPLICIT NONE
      include 'precision.inc'
      REAL(kind=kreal),   intent(in) :: lower_mat(:)
      REAL(kind=kreal),   intent(out):: b(:)
      INTEGER(kind=kint), intent(in) :: N
      INTEGER(kind=kint)             ::i,j,k,cj,ci

      DO i = 1, N
         DO j = 1, i - 1
            cj = N * (j - 1) - j * (j - 1) / 2
            b(i) = b(i) - b(j) * lower_mat(i + cj)
         END DO
         b(i) = b(i) * lower_mat(i + N * (i - 1) - i * (i - 1) / 2)
      END DO
      DO i = N, 1, -1
         ci = N * (i - 1) - i * (i - 1) / 2
         DO j = i + 1, N
            b(i) = b(i) - b(j) * lower_mat(j + ci)
         END DO
         b(i) = b(i) * lower_mat(i + ci)
      END DO

    END SUBROUTINE ll_slv


    subroutine cal_res_vec_at_nth_lev(B, X, RES_VEC, nth_lev)
      USE data_structure_for_AMG
!$    USE omp_lib
      implicit none

      integer(kind=kint), intent(in) :: nth_lev
      integer(kind=kint) :: N, NP, NPL, NPU
      REAL   (kind=kreal), DIMENSION(:), pointer:: D,AU,AL
      INTEGER(kind=kint),  DIMENSION(:), pointer :: INU,IAU,INL,IAL
      REAL   (kind=kreal), DIMENSION(:) :: B,X,RES_VEC

      integer(kind=kint) :: i,j,isU,ieU,isL,ieL,inod
      real(kind=kreal) :: R_val

      N  = HIERARCHICAL_DATA(nth_lev) % N
      NP = HIERARCHICAL_DATA(nth_lev) % NP
      NPL= HIERARCHICAL_DATA(nth_lev) % NPL
      NPU= HIERARCHICAL_DATA(nth_lev) % NPU
      INU=>HIERARCHICAL_DATA(nth_lev) % INU
      INL=>HIERARCHICAL_DATA(nth_lev) % INL
      IAU=>HIERARCHICAL_DATA(nth_lev) % IAU
      !C      B  =>HIERARCHICAL_DATA(nth_lev) % B
      D  =>HIERARCHICAL_DATA(nth_lev) % D
      IAL=>HIERARCHICAL_DATA(nth_lev) % IAL
      AU =>HIERARCHICAL_DATA(nth_lev) % AU
      AL =>HIERARCHICAL_DATA(nth_lev) % AL

      !C calculate residual
      !$omp parallel do private(j,isU,ieU,isL,ieL,R_val,i,inod)
      DO j= 1, N
         R_val= B(j) - D(j)*X(j)
         isU= INU(j-1)+1
         ieU= INU(j)
         DO i= isU, ieU
            inod  = IAU(i) + ZERO_ORIGIN
            R_val = R_val - AU(i) * X(inod)
         END DO
         isL= INL(j-1)+1
         ieL= INL(j)
         DO i= isL, ieL
            inod  = IAL(i) + ZERO_ORIGIN
            R_val = R_val - AL(i) * X(inod)
         END DO
         RES_VEC(j)=R_val
      END DO
      !$omp end parallel do
      
    end subroutine cal_res_vec_at_nth_lev


    SUBROUTINE sgs(mat, B, X, count, dump_gauss_seidel_weight, kind)
      USE comm_switch_mod, only: comm_sr
      USE data_structure_for_AMG
      USE solver_Gnumbering
      USE count_time_mod
      USE matrix_type_mod, only: matrix

      IMPLICIT NONE

      type(matrix) :: mat
      !C kind = 0,1,2 ::   sgs, gs, gs in reverse order
      INTEGER(kind=kint), INTENT(in)   :: count,  kind

      REAL   (kind=kreal) :: B(:), X(:), dump_gauss_seidel_weight

      INTEGER(kind=kint)             :: ierr
      
      INTEGER(kind=kint)  :: cnt, isL, ieL, isU, ieU, i, j, inod
      REAL   (kind=kreal) :: R_val, R_norm, GR_norm, v

      REAL(kind=kreal) :: omega

      omega = dump_gauss_seidel_weight

      DO cnt = 1, count
         if(kind==1 .or. kind==0) then
            call count_time(1, mat % SOLVER_COMM, mat % my_rank, 10)
            DO i = 1, mat % N
               v = B(i)

               isL = mat % INL(i - 1) + 1
               ieL = mat % INL(i)
               DO j = isL, ieL
                  inod = mat % IAL(j) + ZERO_ORIGIN
                  v    = v - mat % AL(j) * X(inod)
               END DO

               isU = mat % INU(i - 1) + 1
               ieU = mat % INU(i)
               DO j = isU, ieU
                  inod = mat % IAU(j) + ZERO_ORIGIN
                  v    = v - mat % AU(j) * X(inod)
               END DO

#ifdef DEBUG
               if(mat % D(i) == 0) then
                  write(*,*) mat % my_rank, i,"diagonal part equals 0!"
                  stop "0 in diagonal part in sgs()"
               end if
#endif

               X(i) = omega*v / mat % D(i)+(1-omega)*X(i)
            END DO
            call count_time(2, mat % solver_comm, mat % my_rank, 10)       

            call comm_sr(mat % SOLVER_COMM, mat % comm_table, &
                 & mat % NP, mat % WS, mat % WR, X)
            
         end if

         if(kind==2 .or. kind==0) then
            call count_time(1, mat % solver_comm, mat % my_rank, 10)       
            DO i = mat % N, 1, -1
               v = B(i)
               
               isL = mat % INL(i - 1) + 1
               ieL = mat % INL(i)
               DO j = isL, ieL
                  inod = mat % IAL(j) + ZERO_ORIGIN
                  v = v - mat % AL(j) * X(inod)
               END DO

               isU = mat % INU(i - 1) + 1
               ieU = mat % INU(i)
               DO j = isU, ieU
                  inod = mat % IAU(j) + ZERO_ORIGIN
                  v = v - mat % AU(j) * X(inod)
               END DO

#ifdef DEBUG
               if(mat % D(i) == 0) then
                  write(*,*) mat % my_rank, i,"diagonal part equals 0!"
                  stop "0 in diagonal part in sgs()"
               end if
#endif
               
               X(i) = (omega)*v / mat % D(i)+(1-omega)*X(i)
            END DO
            CALL count_time(2, mat % solver_comm, mat % my_rank, 10)       
            call comm_sr(mat % SOLVER_COMM, mat % comm_table, &
                 & mat % NP, mat % WS, mat % WR, X)

         end if

      END DO

    END SUBROUTINE sgs


    subroutine nonzero_max_rownums(level, SOLVER_COMM, my_rank, rank_sz)
      USE data_structure_for_AMG
      IMPLICIT NONE

      include 'mpif.h'
      integer(kind=kint), intent(in) :: level, SOLVER_COMM, my_rank, rank_sz
      integer(kind=kint) :: i, j, n, nonz, np, row
      integer(kind=kint), allocatable :: nonzeros(:),row_nums(:)

      allocate(nonzeros(rank_sz),row_nums(rank_sz) )
      nonzeros(:)=-1

      n = HIERARCHICAL_DATA(level) % N
      np = HIERARCHICAL_DATA(level) % NP
      if(n > 0) then
         do i = 1, n
            nonz=1+HIERARCHICAL_DATA(level) % INU(i) - HIERARCHICAL_DATA(level) % INU(i-1)+ &
                 & HIERARCHICAL_DATA(level) % INL(i) - HIERARCHICAL_DATA(level) % INL(i-1)

            call insert(nonz, i, nonzeros, row_nums, rank_sz)

         end do

         do i=1, rank_sz
            if(nonzeros(i)>0) then
               write(*,*) "#nonz,rank#", my_rank, i,":", nonzeros(i), row_nums(i)
            end if
         end do
      end if
      
      deallocate(nonzeros, row_nums)

    contains
      subroutine insert(nonz, row, nonzeros, row_nums, rank_sz)
        integer(kind=kint), intent(in) :: nonz, rank_sz, row
        integer(kind=kint) :: nonzeros(:), row_nums(:)
        integer(kind=kint) :: i,j,k

        do i=1, rank_sz
           if(nonz>nonzeros(i)) exit
        end do
        if(i<=rank_sz) then
           do j=rank_sz-1,i+1,-1
              nonzeros(j+1)=nonzeros(j)
              row_nums(j+1)=row_nums(j)
           end do
           nonzeros(i)=nonz
           row_nums(i)=row
        end if
      end subroutine insert
    end subroutine nonzero_max_rownums

    subroutine matrix_counting(LEVEL_NUM, SOLVER_COMM, my_rank)
      USE data_structure_for_AMG
      USE hash_mod
      IMPLICIT NONE

      include 'mpif.h'
      integer(kind=kint), intent(in) :: LEVEL_NUM, SOLVER_COMM, my_rank
      integer(kind=kint) :: i, j, nonzero_l, nonzero_g, level, n, ierr, np
      REAL(kind=kreal), allocatable :: comm_rate(:)
      REAL(kind=kreal) :: rate, crate_l, crate_g

      allocate(comm_rate(LEVEL_NUM))
      DO level = 1, LEVEL_NUM
         n  = HIERARCHICAL_DATA(level) % N
         np = HIERARCHICAL_DATA(level) % NP
         !C in the case of n == 0
         if(np == 0) then
            crate_l = 1
         else
            crate_l = REAL(n) / np
         end if

         CALL MPI_REDUCE(crate_l, crate_g, 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, SOLVER_COMM, ierr)       
         comm_rate(level) = crate_g
      END DO

      nonzero_l=0


      DO level = 1, LEVEL_NUM
         n = HIERARCHICAL_DATA(level) % N
         if(n > 0) then
            nonzero_l = nonzero_l + n + HIERARCHICAL_DATA(level) % INU(n)
            nonzero_l = nonzero_l + HIERARCHICAL_DATA(level) % INL(n)
         end if
      END DO


      nonzero_g = 0
      CALL MPI_REDUCE(nonzero_l, nonzero_g, 1, MPI_INTEGER, MPI_SUM, 0, SOLVER_COMM, ierr)
      i = nonzero_g


      n = HIERARCHICAL_DATA(1) % N

      !C in the case of n == 0
      nonzero_l = 0 

      if(n > 0) then
         nonzero_l = n + HIERARCHICAL_DATA(1) % INU(n) + HIERARCHICAL_DATA(1) % INL(n)
      end if

      CALL MPI_REDUCE(nonzero_l, nonzero_g, 1, MPI_INTEGER, MPI_SUM, 0, SOLVER_COMM, ierr)
      j = nonzero_g
      if(my_rank == 0)then
         rate = REAL(i, kind = kreal) / j
         write(*,*) "nonzeros in all level:", i, "rate of level:", rate
         write(*,*) "worst rate of own nodes of vector"
         write(*,*) comm_rate
      end if
      
      
    end subroutine matrix_counting

    !C Symmetrize the matrix especially the rows: N+1..NP
    !C using the columns N+1..NP of IAU.
    SUBROUTINE matrix_arrange (N,NP,D,AL,INL,IAL,AU,INU,IAU)

      IMPLICIT NONE
      include 'precision.inc'

      INTEGER(kind=kint ), intent(in)    :: N,NP
      REAL   (kind=kreal), intent(inout),DIMENSION(:) :: D
      REAl   (kind=kreal), intent(inout),DIMENSION(:) :: AU,AL
      INTEGER(kind=kint ), intent(in),DIMENSION(:) :: IAU,IAL
      INTEGER(kind=kint ), intent(in),DIMENSION(0:) :: INU,INL

      INTEGER(kind=kint ) :: i,j,isU,ieU,isL,ieL,k,column
      logical :: flag


      DO j=N+1,NP
         isL = INL(j-1)+1
         ieL = INL(j)
         DO i=isL,ieL
            column = IAL(i)+ZERO_ORIGIN
            if(column>N) then
               AL(i)=0
            else if(column>0 .and. column<=N) then
               isU = INU(column-1)+1
               ieU = INU(column)
               flag = .false.
               DO k=isU,ieU
                  if(IAU(k)+ZERO_ORIGIN==j)then
                     AL(i)=AU(k)
                     flag= .true.
                     exit
                  end if
               end DO
               if(.not.flag) then
                  AL(i) =0.0
               end if
            end if
         END DO

         isU = INU(j-1)+1
         ieU = INU(j)
         DO i=isU,ieU
            AU(i)= 0.0
         END DO
      END DO
    END SUBROUTINE matrix_arrange


    SUBROUTINE mcsgs(mat, B, X, count, color, col_ind, INML, dump_gauss_seidel_weight, kind)
      USE comm_switch_mod
      USE data_structure_for_AMG
      USE solver_Gnumbering
      USE count_time_mod
      USE matrix_type_mod, only: matrix
!$    USE omp_lib

      IMPLICIT NONE


      type(matrix) :: mat
      !C kind = 0,1,2 ::   sgs, gs, gs in reverse order
      INTEGER(kind=kint), INTENT(in)  :: count, color, kind
      REAL   (kind=kreal) ::  B(:), X(:), dump_gauss_seidel_weight

      INTEGER(kind=kint) :: col_ind(0:)
      INTEGER(kind=kint) :: INML(:)

      INTEGER(kind=kint)             :: ierr

      INTEGER(kind=kint) cst,ced,ind

      INTEGER(kind=kint)  :: cnt, isL, ieL, isU, ieU
      INTEGER(kind=kint) :: i,j,k,inod
      REAL   (kind=kreal) :: R_val, R_norm, GR_norm, v, omega
      
      omega = dump_gauss_seidel_weight

      DO cnt = 1, count
         if(kind==0 .or. kind==1) then
            call count_time(1, mat % SOLVER_COMM, mat % my_rank, 10)
            DO k = 1,color
               cst = col_ind(k-1) + 1
               ced = col_ind(k)
               !$omp parallel private(i,ind,isL,ieL,inod,j,isU,ieU,v)
               !$omp do 
               DO i = cst,ced
                  ind = INML(i)
                  isL = mat % INL(ind-1) +1
                  ieL = mat % INL(ind)
                  v = B(ind)
                  DO j = isL,ieL
                     inod = mat % IAL(j)
                     v = v - mat % AL(j)*X(inod)
                  END DO
                  isU = mat % INU(ind-1)+1
                  ieU = mat % INU(ind)
                  DO j = isU,ieU
                     inod = mat % IAU(j)
                     v = v - mat % AU(j)*X(inod)
                  END DO
#ifdef DEBUG
                  if(mat % D(ind) == 0) then
                     write(*,*) mat % my_rank, i,"diagonal part equals 0!"
                     stop "0 in diagonal part in sgs()"
                  end if
#endif
                  X(ind) = omega*v / mat % D(ind) + (1-omega)*X(ind)
               END DO
               !$omp end do
               !$omp end parallel
            END DO

            call count_time(2, mat % solver_comm, mat % my_rank, 10)

            call comm_sr(mat % SOLVER_COMM, mat % COMM_TABLE,      &
                 & mat % NP, mat % WS, mat % WR, X)
            
         end if

         
         if(kind==2 .or. kind==0) then
            call count_time(1, mat % solver_comm, mat % my_rank, 10)

            DO k = color, 1, -1
!!$            DO k = 1,color
               cst = col_ind(k-1) + 1
               ced = col_ind(k)
               !$omp parallel private(i,ind,isL,ieL,inod,j,isU,ieU,v)
               !$omp do 
               DO i = cst,ced
                  ind = INML(i)
                  isL = mat % INL(ind-1) +1
                  ieL = mat % INL(ind)
                  v = B(ind)
                  DO j = isL,ieL
                     inod = mat % IAL(j)
                     v = v - mat % AL(j)*X(inod)
                  END DO
                  isU = mat % INU(ind-1)+1
                  ieU = mat % INU(ind)
                  DO j = isU,ieU
                     inod = mat % IAU(j)

                     v = v - mat % AU(j)*X(inod)
                  END DO
#ifdef DEBUG
                  if(mat % D(ind) == 0) then
                     write(*,*) mat % my_rank, i, &
                          & "diagonal part equals 0!"
                     stop "0 in diagonal part in sgs()"
                  end if
#endif
                  X(ind) = (omega)*v / mat % D(ind)+(1-omega)*X(ind)
               END DO
               !$omp end do
               !$omp end parallel
            END DO
            CALL count_time(2, mat % solver_comm, mat % my_rank, 10)

            call comm_sr(mat % SOLVER_COMM, mat % COMM_TABLE,      &
                 & mat % NP, mat % WS, mat % WR, X)


         end if
      END DO

    end subroutine mcsgs

  END MODULE  AMG_cycle

