!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE aggregate_mod
CONTAINS
  subroutine decide_sender_and_receiver(SOLVER_COMM, N, NEIBPETOT, NEIBPE, STACK_IMPORT, &
       & STACK_EXPORT, sender_fl, receiver_fl, partner, NPROCS, my_rank)
    implicit none
    include 'precision.inc'
    include 'mpif.h'
    
    integer :: SOLVER_COMM
    integer(kind=kint), intent(in) :: N, NEIBPETOT, NEIBPE(:), STACK_IMPORT(0:), STACK_EXPORT(0:)
    
    logical, intent(out) :: sender_fl, receiver_fl
    integer(kind=kint) :: NPROCS, my_rank, partner
    integer(kind=kint), pointer :: process_in(:), process_col(:), process_val(:)
    integer(kind=kint), pointer :: process_comminfo(:), process_compinfo(:), pair_list(:)
    
    integer(kind=kint) :: i
    
    real(kind=kreal) :: threshould

    !C initialize
    sender_fl   = .false.
    receiver_fl = .false.
    partner     = -1

#ifdef COUPLE_NEIB_AGGREGATES
    
    call gather_process_neighboring_graph_info(SOLVER_COMM, N, NEIBPETOT, NEIBPE, NPROCS, STACK_IMPORT, &
         & STACK_EXPORT, process_compinfo, process_comminfo, process_in, process_col, process_val)
    
!C    threshould = 10.0
    threshould = 0.5
    call decide_pairs_from_gl_neighb_grph(process_compinfo, process_comminfo, process_in, &
         & process_col, process_val, pair_list, threshould, NPROCS)
    
    if(pair_list(my_rank+1) >= 0) then
       sender_fl = .true.
       partner   = pair_list(my_rank+1)
       
    else if(pair_list(my_rank+1) == -2) then
       receiver_fl = .true.
       do i = 1, NPROCS
          if(pair_list(i) == my_rank) exit
       end do
       if(i > NPROCS) stop 'Error in searching'
       partner = i-1
       
    end if

    deallocate(pair_list, process_compinfo, process_comminfo, process_in, process_col, process_val)
#endif
    
  end subroutine decide_sender_and_receiver
  
  subroutine decide_pairs_from_gl_neighb_grph(process_compinfo, process_comminfo, process_in, &
       & process_col, process_val, pair_list, threshould, NPROCS)
    implicit none
    include 'precision.inc'
    
    integer(kind=kint), intent(in) :: process_compinfo(:), process_comminfo(:), process_in(0:)
    integer(kind=kint), intent(in) :: process_col(:), process_val(:), NPROCS
    
    real(kind=kreal),   intent(in) :: threshould
    
    integer(kind=kint) :: i, j, c, cn, maxr
    
    integer(kind=kint), pointer :: pair_list(:)
    
    allocate(pair_list(NPROCS))
    
    pair_list = -1

    do i = 1, NPROCS
       if(process_compinfo(i) > 0) then
          if( dble(process_comminfo(i)) / process_compinfo(i) > threshould .and. &
               & pair_list(i) == -1) then
             c = 0
             do j  = process_in(i-1)+1, process_in(i)
                cn = process_val(j)
                if( c < cn .and. pair_list(process_col(j)+1) == -1 ) then
                   maxr = process_col(j)
                   c    = cn
                end if
             end do

             if(c > 0) then
                pair_list(i) = maxr
                pair_list(maxr+1) = -2
             end if

          end if
       end if
    end do
    
  end subroutine decide_pairs_from_gl_neighb_grph

  subroutine gather_process_neighboring_graph_info(SOLVER_COMM, N, NEIBPETOT, NEIBPE, NPROCS, STACK_IMPORT, &
       & STACK_EXPORT, process_compinfo, process_comminfo, process_in, process_col, process_val)
    implicit none
    include 'precision.inc'
    include 'mpif.h'
    
    integer(kind=kint), intent(in) :: NPROCS
    integer(kind=kint), intent(in) :: SOLVER_COMM
    integer(kind=kint), intent(in) :: N, NEIBPETOT, NEIBPE(:)
    integer(kind=kint), intent(in) :: STACK_IMPORT(0:), STACK_EXPORT(0:)
    !C # of unknowns, comms of each process
    integer(kind=kint), pointer :: process_compinfo(:), process_comminfo(:)
    !C crs form of neighboring graph. val shows the amount of comm
    integer(kind=kint), pointer :: process_in(:), process_col(:), process_val(:)
    
    integer(kind=kint), allocatable :: temp(:), temp2(:), recvcount(:), temp_comm_weight(:)
    
    integer(kind=kint) :: ierr, i, j, k
    
    allocate(temp(NPROCS*3), temp2(3),  process_in(0:NPROCS), process_compinfo(NPROCS), &
         & process_comminfo(NPROCS)) 
    
    temp2(1) = NEIBPETOT
    temp2(2) = N 
    temp2(3) = STACK_IMPORT(NEIBPETOT)+STACK_EXPORT(NEIBPETOT)

    call MPI_ALLGATHER(temp2(1), 3, MPI_INTEGER, temp(1), 3, MPI_INTEGER, SOLVER_COMM, ierr)

    process_in(0) = 0
    
    do i = 1, NPROCS
       process_in(i) = temp(3*i-2)
       process_compinfo(i) = temp(3*i-1)
       process_comminfo(i) = temp(3*i)
    end do

    allocate(recvcount(NPROCS))
    recvcount(1:NPROCS) = process_in(1:NPROCS)
    
    do i = 1, NPROCS
       process_in(i) = process_in(i-1)+process_in(i)
    end do

    allocate(process_col(process_in(NPROCS)), process_val(process_in(NPROCS)), &
         &   temp_comm_weight(NEIBPETOT+1))

    do i = 1, NEIBPETOT
       temp_comm_weight(i) = (STACK_IMPORT(i)-STACK_IMPORT(i-1))+(STACK_EXPORT(i)-STACK_EXPORT(i-1))
    end do
    
    
    call MPI_ALLGATHERV(NEIBPE(1),           NEIBPETOT, MPI_INTEGER, process_col(1),&
         & recvcount(1), process_in(0), MPI_INTEGER, SOLVER_COMM, ierr)

    call MPI_ALLGATHERV(temp_comm_weight(1), NEIBPETOT, MPI_INTEGER, process_val(1),&
         & recvcount(1), process_in(0), MPI_INTEGER, SOLVER_COMM, ierr)

    deallocate(temp, recvcount, temp_comm_weight, temp2)
    
  end subroutine gather_process_neighboring_graph_info
  
  SUBROUTINE indpdt_agrgt (s,INL,INU,IAL,IAU,NEIBPETOT, STACK_EXPORT, NOD_EXPORT, &
       & PNI,NI,LEVEL_NO,node_index, &
       & in_aggregates_result,aggregates_result, in_aggregates_result_size, &
       & border_flag)
!C    USE data_structure_for_AMG
    USE queue_mod
    IMPLICIT NONE
    INCLUDE  'precision.inc'
    
    INTEGER(kind=kint), INTENT(in) :: NI(:) 
    INTEGER(kind=kint), INTENT(in) :: PNI(0:)
    INTEGER(kind=kint), INTENT(in) :: LEVEL_NO
    INTEGER(kind=kint), INTENT(inout) :: node_index(:)
    INTEGER(kind=kint), pointer :: in_aggregates_result(:)
    INTEGER(kind=kint), pointer :: aggregates_result(:)
    INTEGER(kind=kint), INTENT(out) :: in_aggregates_result_size
    LOGICAL, INTENT(in) :: border_flag

    !C    TYPE(INTER_LEVEL_OPERATOR),POINTER :: P
    INTEGER(kind=kint), ALLOCATABLE :: Temp(:,:)
    
    INTEGER(kind=kint), ALLOCATABLE :: P_queue(:,:)
    INTEGER(kind=kint), ALLOCATABLE :: aggregate_flags(:)

    INTEGER(kind=kint) :: head,tail,queue_size
    
    REAL   (kind=kreal), POINTER :: D(:), AU(:), AL(:)
    INTEGER(kind=kint ), POINTER :: INU(:), IAU(:), INL(:), IAL(:)
    
    INTEGER(kind=kint ), POINTER :: NOD_EXPORT(:)
    INTEGER(kind=kint ), POINTER :: STACK_EXPORT(:)
    INTEGER(kind=kint ) :: NEIBPETOT
    INTEGER(kind=kint ), ALLOCATABLE :: node_record(:)

    INTEGER(kind=kint) :: i, s, s_u, k, f, aggr, i_CN, size_of_aggregates, count, l, j, m, front, back
    INTEGER(kind=kint) :: node_record_counter, node, score, insert_point, max_score

    REAL   (kind=kreal) :: r
    INTEGER(kind=kint), parameter :: aggregate_number=1
    INTEGER(kind=kint), parameter :: aggregate_index=2
    logical :: flag

    ALLOCATE(Temp(0:s, 2))
    ALLOCATE(P_queue(2, s))
    ALLOCATE(aggregate_flags(s))
    ALLOCATE(node_record(s))

    s_u = INU(s)
    


    !C aggregates is stored in P temporarily 
    !C    P=>HIERARCHICAL_DATA(LEVEL_NO) % P    
    
    !C Temp =0
    Temp(1:s, aggregate_number) = 0
    Temp(0:s, aggregate_index) = 0


    DO i = 1, s
       if(node_index(i) == 0) node_index(i) = 1
    END DO
    
    !C initialize Priority queue
    !C P_queue has score and node index. node_index has the index of P_queue
    !C -1 in node_index : selected as part of aggregate
    !C  0 in node_index : not selected as part of aggregate, but it's neighbor is aggregate.
    
    P_queue(1:2, 1:s) = 0
    aggregate_flags(1:s) = 0
    
    
    !C interface part is scored, and enqueued
    !C Temp(,aggregate_number) is used as buffer
    if(NEIBPETOT > 0 .and. border_flag) then
       DO count = 1, STACK_EXPORT(NEIBPETOT)
          j = NOD_EXPORT(count)+ZERO_ORIGIN
          DO k = PNI(2 * j - 2) + 1,PNI(2 * j - 1)
             l = IAL(NI(k))+ZERO_ORIGIN
             Temp(l, aggregate_number) = 1
          ENDDO
          DO k = PNI(2 * j - 1) + 1, PNI(2 * j)
             if(IAU(NI(k))+ZERO_ORIGIN > s) cycle
             l = IAU(NI(k))+ZERO_ORIGIN
             Temp(l, aggregate_number) = 1
          ENDDO
       END DO
       !C interface node is set to 0.
       DO count = 1, STACK_EXPORT(NEIBPETOT)
          j = NOD_EXPORT(count)+ZERO_ORIGIN
          Temp(j, aggregate_number) = 0
       END DO
    end if
    queue_size = 0
    DO count = 1, s
       if(Temp(count, aggregate_number) == 1) then
          
          !C inserting: score is 10,position is queue_size
          queue_size = queue_size + 1
          node_index(count) = queue_size
          P_queue(1, queue_size) = 10
          P_queue(2, queue_size) = count
       END if
    END DO
    
    !C enqueue nodes whose score is 1
    DO count = 1, s
       if(Temp(count, aggregate_number) == 1)then
          Temp(count, aggregate_number) = 0
          !C this case is enqueued
       else if(node_index(count) == 1)then
          queue_size = queue_size + 1
          node_index(count) = queue_size
          P_queue(1, queue_size) = 1
          P_queue(2, queue_size) = count
       end if
    END DO
    
    size_of_aggregates = 0

    if(s > 0) then
       DO while(P_queue(1,1) > 0)

          i = P_queue(2,1)
          size_of_aggregates = size_of_aggregates + 1

          !C----layer1 start
          DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
             j=IAL(NI(k))+ZERO_ORIGIN

             Temp(j, aggregate_number) = size_of_aggregates

             if(node_index(j) > 0) then
                !C dequeue node j
                call dequeue(j, P_queue, node_index, queue_size)
             end if
             node_index(j) = -1
             aggregate_flags(j) = i

          END DO
          DO k = PNI(2 * i - 1) + 1, PNI(2 * i)
             if(IAU(NI(k))+ZERO_ORIGIN > s) cycle
             j = IAU(NI(k))+ZERO_ORIGIN
             Temp(j, aggregate_number) = size_of_aggregates

             if(node_index(j) > 0) then
                !C dequeue node j
                call dequeue(j, P_queue, node_index, queue_size)
             end if
             node_index(j) = -1
             aggregate_flags(j) = i
          END DO

          Temp(i , aggregate_number) = size_of_aggregates

          !C dequeue node i
          call dequeue(i, P_queue, node_index, queue_size)

          node_index(i) = -1
          aggregate_flags(i) = i
          !C----layer1 end

          !C----layer2 and 3 
          call dequeue_scoreup_Layer23(i, s, queue_size, P_queue,  & 
               & node_index, aggregate_flags, node_record, NI, PNI, IAL, IAU)

       END DO
    endif

!!$       !C greedy case
!!$       !C== creating independent aggregate
!!$       size_of_aggregates=0
!!$       DO i=s,1,-1
!!$          IF(node_index(i)==0) THEN
!!$             l=0
!!$             DO k=PNI(2*i-2)+1,PNI(2*i-1)
!!$                j=IAL(NI(k))
!!$                l=l+node_index(j)
!!$             END DO
!!$             DO k=PNI(2*i-1)+1,PNI(2*i)
!!$                j=IAU(NI(k))
!!$                l=l+node_index(j)
!!$             END DO
!!$             IF(l==0) THEN
!!$                size_of_aggregates=size_of_aggregates+1
!!$
!!$                DO k=PNI(2*i-2)+1,PNI(2*i-1)
!!$                   j=IAL(NI(k))
!!$                   Temp(j, aggregate_number) = size_of_aggregates
!!$                   node_index(j)=-1
!!$                END DO
!!$                DO k=PNI(2*i-1)+1,PNI(2*i)
!!$                   j=IAU(NI(k))
!!$                   Temp(j, aggregate_number) = size_of_aggregates
!!$                   node_index(j)=-1
!!$                END DO
!!$
!!$                Temp(i, aggregate_number) = size_of_aggregates
!!$                node_index(i)=-1
!!$
!!$             END IF
!!$          END IF
!!$       END DO
!!$       write(*,*) size_of_aggregates

    
    deallocate(node_record)
    
    !C added aggregate number is set to minus to avoid adding nodes like chain reaction
    aggregate_check: DO count = 1, s
       if(node_index(count) == 0) then
          lower_part: DO k = PNI(2 * count - 2) + 1, PNI(2 * count - 1)
             !C aggr is the aggregate number of next to count
             aggr = Temp(IAL( NI(k) )+ZERO_ORIGIN, aggregate_number)
             IF(aggr > 0) Then
                Temp(count, aggregate_number) = -1*aggr
                exit lower_part
             END IF
          END DO lower_part

          IF(Temp(count, aggregate_number) > 0) Then
             cycle aggregate_check
          END IF
          
          upper_part: DO k = PNI(2 * count - 1) + 1, PNI(2 * count)
             !C aggr is the aggregate number of next to i
             if(IAU(NI(k))+ZERO_ORIGIN > s) cycle
             aggr = Temp(IAU(NI(k))+ZERO_ORIGIN, aggregate_number)
             
             IF(aggr > 0) Then
                Temp(count, aggregate_number) = -1*aggr
                exit upper_part
             END IF
          END DO upper_part
          IF(Temp(count, aggregate_number) == 0) Then
#ifdef DEBUG
             write(*,*) "independent node", node_index(count)
#endif
          END IF
       end if
    END DO aggregate_check

    DEALLOCATE(P_queue)
  
    in_aggregates_result_size = size_of_aggregates
    ALLOCATE(in_aggregates_result(0:size_of_aggregates))
    ALLOCATE(aggregates_result(s))

    in_aggregates_result(0) = 0


    !C minum aggregate number in check part is changed to plus
    DO i = 1, s
       if(Temp(i, aggregate_number) < 0) Temp(i, aggregate_number) = Temp(i, aggregate_number) * -1 
    END DO

    DO i = 1, s
       aggr = Temp(i, aggregate_number)
       if(aggr > 0) then
          Temp(aggr, aggregate_index) = Temp(aggr, aggregate_index) + 1
       end if
    END DO

    !C size_of_aggregates has the size of aggregates
    !C P % IN(aggr)+1 points the place for new data of aggr
    DO i = 1, size_of_aggregates
       Temp(i, aggregate_index) = Temp(i, aggregate_index) + Temp(i - 1, aggregate_index)
       in_aggregates_result(i) = Temp(i - 1, aggregate_index)
    END DO

    !C unneeded nodes belong to aggregate NO 0.
    DO i = 1, s
       aggr = Temp(i, aggregate_number)
       if(aggr > 0) then
          k = in_aggregates_result(aggr) + 1
          in_aggregates_result(aggr) = k
          aggregates_result(k) = i
       end if
    END DO

    i_CN = Temp(size_of_aggregates, aggregate_index)

    DEALLOCATE(Temp)

  END SUBROUTINE indpdt_agrgt


  SUBROUTINE border_aggregate (NP,N,IAL, IAU, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
       &                       STACK_EXPORT, NOD_EXPORT, &
       &                      PNI, NI, LEVEL_NO, node_index, sorted_NEIBPE, my_rank,     &
       &                      PE_num_size,PE_nums,nodes_for_each_PE,in_nodes_for_each_PE, &
       &                      PE_list_size,PE_list,NPROCS,SOLVER_COMM)
    
    USE solver_SR2
    IMPLICIT NONE
    include 'precision.inc'
    INCLUDE  'mpif.h'
    INTEGER(kind=kint), INTENT(in)    :: NI(:)
    INTEGER(kind=kint), INTENT(in)    :: PNI(0:)
    INTEGER(kind=kint), INTENT(in)    :: LEVEL_NO
    INTEGER(kind=kint), INTENT(inout) :: node_index(:)
    INTEGER(kind=kint), INTENT(in)    :: sorted_NEIBPE(:)
    INTEGER(kind=kint), INTENT(in)    :: my_rank
    INTEGER(kind=kint), INTENT(in)    :: SOLVER_COMM
    INTEGER(kind=kint), INTENT(in)     :: PE_num_size
    INTEGER(kind=kint), POINTER :: PE_nums(:)
    INTEGER(kind=kint), POINTER :: nodes_for_each_PE(:)
    INTEGER(kind=kint), POINTER :: in_nodes_for_each_PE(:)
    INTEGER(kind=kint), INTENT(out) :: PE_list_size
    INTEGER(kind=kint), INTENT(out) :: PE_list(:,:)
    INTEGER(kind=kint), INTENT(in)  :: NPROCS
    
    INTEGER(kind=kint)          :: NEIBPETOT
    INTEGER(kind=kint), POINTER :: NEIBPE      (:)
    INTEGER(kind=kint), POINTER :: NOD_IMPORT  (:),STACK_IMPORT(:)
    INTEGER(kind=kint), POINTER :: NOD_EXPORT  (:),STACK_EXPORT(:)
    INTEGER(kind=kint)          :: ierr
    INTEGER(kind=kint), POINTER :: IAU (:)
    INTEGER(kind=kint), POINTER :: IAL (:)

    
    INTEGER(kind=kint) :: i,j,k,l,pe,node,NP,N,sni,pli,nfei,counter
    INTEGER(kind=kint) :: des_pe,nd,nds_sz_for_each_PE,in_s,sz,rbd
    INTEGER(kind=kint) :: index,Neib_index,Pe_num_index,sbi,nd_base
    INTEGER(kind=kint) :: PNI_L,PNI_H,PNI_M,sb_base,nni,sz_in,nd_in
    INTEGER(kind=kint) :: smaller_PEs

    INTEGER(kind=kint) :: inum,istart,recv_count
    INTEGER(kind=kint), ALLOCATABLE :: change_vector(:)
    INTEGER(kind=kint), ALLOCATABLE :: inverse_vector(:)
    INTEGER(kind=kint)              :: inverse_vector_size
    INTEGER(kind=kint)              :: send_buf_size
    INTEGER(kind=kint), ALLOCATABLE :: send_buf(:)
    INTEGER(kind=kint), ALLOCATABLE :: in_send_buf(:)
    INTEGER(kind=kint)              :: recv_buf_size
    INTEGER(kind=kint), ALLOCATABLE :: recv_buf(:)
    INTEGER(kind=kint), ALLOCATABLE :: recv_cnt_ary(:)
    INTEGER(kind=kint), ALLOCATABLE :: recv_buf2(:) 
    INTEGER(kind=kint), ALLOCATABLE :: recv_from_PE(:)
    
    INTEGER(kind=kint), ALLOCATABLE :: req1(:),req2(:),sta1(:,:),sta2(:,:)
    INTEGER(kind=kint), ALLOCATABLE :: PE_in_charge(:), min_rank_in_1_path_left_nds(:)
    INTEGER(kind=kint), ALLOCATABLE :: min_rank_in_1_path(:)
    INTEGER(kind=kint), ALLOCATABLE :: WSI(:),WRI(:)

    INTEGER(kind=kint), ALLOCATABLE :: tmp(:)
    
    LOGICAL :: tf, local_finish, global_finish

    !C PE_list(*,1): PE number
    !C PE_list(*,2): NEIBPE number|
    !C PE_list(*,3): PE_num number|these two has 0 if no correspondings.
    !C-- making PE_list..
    index=1
    Neib_index=1
    Pe_num_index=1

    DO WHILE((Neib_index<=NEIBPETOT) .AND. (Pe_num_index<=PE_num_size))
       i=sorted_NEIBPE(Neib_index)
       IF(PE_nums(Pe_num_index)==NEIBPE(i))THEN
          PE_list(index,1) = PE_nums(Pe_num_index)
          PE_list(index,2) = i
          PE_list(index,3) = Pe_num_index
          Neib_index = Neib_index+1
          PE_num_index = Pe_num_index+1
       ELSE IF(PE_nums(Pe_num_index)<NEIBPE(i))THEN
          PE_list(index,1) = PE_nums(Pe_num_index)
          PE_list(index,2) = 0
          PE_list(index,3) = Pe_num_index
          PE_num_index = PE_num_index+1
       ELSE      
          PE_list(index,1) = NEIBPE(i)
          PE_list(index,2) = i
          PE_list(index,3) = 0
          Neib_index = Neib_index+1
       END IF
       index=index+1
    END DO
    
    IF(Neib_index>NEIBPETOT)THEN
       DO i=PE_num_index,PE_num_size
          PE_list(index,1)=PE_nums(i)
          PE_list(index,2)=0
          PE_list(index,3)=i
          index=index+1
       END DO
    ELSE
       DO i=Neib_index,NEIBPETOT
          j=sorted_NEIBPE(i)
          PE_list(index,1)=NEIBPE(j)
          PE_list(index,2)=j
          PE_list(index,3)=0
          index=index+1
       END DO
    END IF
    PE_list_size = index - 1
    !C-- end: making PE_list..
    
    !C-- PE_in_charge has a rank number which deals with eaaaach unknowns
    ALLOCATE(PE_in_charge(NP), min_rank_in_1_path_left_nds(NP))

    DO i = 1, NP
       min_rank_in_1_path_left_nds(i) = NPROCS
    END DO
    DO i = NEIBPETOT, 1, -1
       j = sorted_NEIBPE(i)
       pe = NEIBPE(j)
       IF(pe > my_rank) pe = my_rank
       DO k = STACK_EXPORT(j - 1) + 1, STACK_EXPORT(j)
          node = NOD_EXPORT(k)+ZERO_ORIGIN
          min_rank_in_1_path_left_nds(node) = pe
       END DO
    END DO
    
    ALLOCATE(WSI(STACK_EXPORT(NEIBPETOT)))
    ALLOCATE(WRI(STACK_IMPORT(NEIBPETOT)))
    CALL SOLVER_SEND_RECV2I(NP, STACK_EXPORT(NEIBPETOT),            &
         &       STACK_IMPORT(NEIBPETOT), NEIBPETOT, NEIBPE,        &
         &       STACK_IMPORT(0:), NOD_IMPORT, STACK_EXPORT(0:),    &
         &       NOD_EXPORT, WSI, WRI, min_rank_in_1_path_left_nds,         &
         &       SOLVER_COMM, my_rank)

    allocate(min_rank_in_1_path(NP))
    min_rank_in_1_path = min_rank_in_1_path_left_nds 
!C    DEALLOCATE(min_rank_in_1_path_left_nds)

    !C-- end: min_rank_in_1_path_left_nds(node) shows minimum rank among 1 path from node,
    
    
    !C-- information of neighbors in COMM tables is exchanged
    !C   * change_vector has serial number of nod_import,
    !C    node_export, nodes_for_each_PE. irrelevant node has 0.
    ALLOCATE(change_vector(NP))
    send_buf_size = NEIBPETOT + STACK_EXPORT(NEIBPETOT) + PNI(2 * N)
    ALLOCATE(send_buf(send_buf_size))
    ALLOCATE(in_send_buf(0:NEIBPETOT))
    
    sbi = 0
    Pli = 1

    in_send_buf(0) = 0
    !C PE_list's PE number is sorted. so the loop is execed 
    !C on PE sorted order.
    DO i = 1, NEIBPETOT
       sni = sorted_NEIBPE(i)
       des_pe = NEIBPE(sni)

       DO WHILE(PE_list(pli, 1) /= des_pe)
          pli = pli + 1
       END DO
       IF(pli > PE_list_size) STOP "unexpected error in border_aggregates"
       nfei = PE_list(pli, 3)
       
       !C== making change_vector
       !C counter is used to create serial number
       !C change_vector: 0 or serial number on extended comm table
       change_vector = 0
       counter = 0
       DO j = STACK_IMPORT(sni - 1) + 1, STACK_IMPORT(sni)
          counter = counter + 1
#ifdef DEBUG
          IF(change_vector(NOD_IMPORT(j)+ZERO_ORIGIN) /= 0) & 
          & STOP "error in change_vector of border_aggregate()"
#endif
          change_vector(NOD_IMPORT(j)+ZERO_ORIGIN) = counter

          
       END DO

       !C nds_sz_for_each_PE: size of nodes to send from my_rank to des_pe
       nds_sz_for_each_PE = 0
       DO j = STACK_EXPORT(sni - 1) + 1, STACK_EXPORT(sni)
          counter = counter + 1
          nd = NOD_EXPORT(j)+ZERO_ORIGIN
#ifdef DEBUG
          IF(change_vector(NOD_EXPORT(j)+ZERO_ORIGIN) /= 0) &
          & STOP "error in change_vector of border_aggregate()"
#endif

          change_vector(nd) = counter
          IF(min_rank_in_1_path_left_nds(nd)==des_pe) nds_sz_for_each_PE=nds_sz_for_each_PE+1
       END DO
       IF(nfei > 0) THEN
          DO j = in_nodes_for_each_PE(nfei - 1) + 1, in_nodes_for_each_PE(nfei)
             counter = counter + 1


#ifdef DEBUG
             IF(change_vector(nodes_for_each_PE(j)+ZERO_ORIGIN)/=0)  &
             & STOP "error in change_vector of border_aggregate()"
#endif
             change_vector(nodes_for_each_PE(j)+ZERO_ORIGIN) = counter
          END DO
       END IF
       !C== end:making change_vector       

       !C== buffering each_node's neighbors for each PE
       !C   sbi: send buffer index
       sbi = sbi + 1

       send_buf(sbi) = nds_sz_for_each_PE
       !C sb_base means to be index of send_buf
       sb_base = sbi + nds_sz_for_each_PE
       DO j = STACK_EXPORT(sni - 1) + 1, STACK_EXPORT(sni)
          k = NOD_EXPORT(j)+ZERO_ORIGIN
          IF(min_rank_in_1_path_left_nds(k) == des_pe) THEN
             PNI_H = PNI(2 * k)
             PNI_M = PNI(2 * k - 1)
             PNI_L = PNI(2 * k - 2)
             !C neighboring info is thrown to the buffer
             sbi = sbi + 1
             send_buf(sbi) = PNI_H - PNI_L
             DO l = PNI_L + 1, PNI_M
                nni = NI(l)
                send_buf(sb_base + l - PNI_L) = change_vector(IAL(nni)+ZERO_ORIGIN)
             END DO
             sb_base = sb_base + PNI_M - PNI_L
             DO l = PNI_M + 1, PNI_H
                nni = NI(l)
                send_buf(sb_base + l - PNI_M) = change_vector(IAU(nni)+ZERO_ORIGIN)
             END DO
             sb_base = sb_base + PNI_H - PNI_M
          END IF
       END DO
       sbi = sb_base
       in_send_buf(i) = sbi 
       !C== end:buffering each_node's neighbors for each PE
    END DO
    DEALLOCATE(change_vector)
    
    !C== count up recv_count, and allcate recv_buf
    !C maybe recv_buf_size is too small..
    !C recv_count is the number of receive times

    recv_count = 0
    DO i = 1, NEIBPETOT
       pe = NEIBPE(i)
       IF(pe > my_rank) recv_count = recv_count + 1
    END DO


    ALLOCATE(recv_from_PE(NEIBPETOT))
    ALLOCATE(recv_cnt_ary(0:recv_count))
    ALLOCATE(req1(PE_list_size))
    ALLOCATE(sta1(MPI_STATUS_SIZE,PE_list_size))
    ALLOCATE(req2(PE_list_size))
    ALLOCATE(sta2(MPI_STATUS_SIZE,PE_list_size))


    !C in_send_buf is changed to have send buffer size 
    DO i = NEIBPETOT, 1, -1
       in_send_buf(i) = in_send_buf(i) - in_send_buf(i-1)
    END DO


    k = 0
    l = 0
    recv_from_PE = -1
    DO i = 1, NEIBPETOT
       sni = sorted_NEIBPE(i)
       des_pe = NEIBPE(sni)
       IF(des_pe > my_rank) THEN
          k = k + 1
          recv_from_PE(sni) = k
          CALL MPI_IRECV(recv_cnt_ary(k),1,MPI_INTEGER,des_pe,0,&
               &        SOLVER_COMM,req2(k),ierr)
       ELSE IF(des_pe<my_rank)THEN
          l = l + 1
!C          in_send_buf is changed to have its own send msg size.
!C          istart = in_send_buf(i-1)
!C          inum = in_send_buf(i)-istart
          CALL MPI_ISEND (in_send_buf(i), 1, MPI_INTEGER, des_pe, 0, &
               &    SOLVER_COMM, req1(l), ierr)
       ELSE
          WRITE(*,*) "error!",des_pe,my_rank
          STOP
       END IF
    END DO
    IF(l > 0) CALL MPI_WAITALL(l, req1, sta1, ierr)
    IF(k > 0) CALL MPI_WAITALL(k, req2, sta2, ierr)
    recv_cnt_ary(0) = 0
    DO i = 1, recv_count
       recv_cnt_ary(i) = recv_cnt_ary(i - 1) + recv_cnt_ary(i)
    END DO
    recv_buf_size = recv_cnt_ary(recv_count)
    ALLOCATE(recv_buf(recv_buf_size))
    !C== end:count up recv_count, and allcate recv_buf    

    !C in_send_buf is changed to index of send buffer
    DO i = 1, NEIBPETOT
       in_send_buf(i) = in_send_buf(i) + in_send_buf(i-1)
    END DO

    k = 0
    l = 0
    DO i = 1, NEIBPETOT
       sni = sorted_NEIBPE(i)
       des_pe = NEIBPE(sni)
       IF(des_pe > my_rank) THEN
          k = k + 1
          istart = recv_cnt_ary(k - 1)
          inum = recv_cnt_ary(k) - istart
          CALL MPI_IRECV(recv_buf(istart+1),inum,MPI_INTEGER,des_pe,0,&
               &        SOLVER_COMM,req2(k),ierr)
       ELSE
          l = l + 1
          istart = in_send_buf(i - 1)
          inum = in_send_buf(i) - istart
          CALL MPI_ISEND (send_buf(istart+1),inum,MPI_INTEGER,des_pe,0,&
               &    SOLVER_COMM,req1(l),ierr)
       END IF
    END DO
    IF(k > 0) CALL MPI_WAITALL(k, req2, sta2, ierr)
    IF(l > 0) CALL MPI_WAITALL(l, req1, sta1, ierr)
    DEALLOCATE(send_buf)
    DEALLOCATE(in_send_buf)
    !C-- end: infomation of neighbors in COMM tables is exchanged


    !C the following arrays is for exchanging the result data
    send_buf_size = (STACK_IMPORT(NEIBPETOT)+1)*2
    ALLOCATE(send_buf(send_buf_size))
    recv_buf_size = (STACK_EXPORT(NEIBPETOT)+1)*2
    ALLOCATE(recv_buf2(recv_buf_size))
    allocate(tmp(2*NP))

    !C this do loop creating aggregates on border consistently
    do
       !C== PE_in_charge is created from min_rank_in_1_path_left_nds
       !C PE_in_charge(i) has minimum rank among 2 path neighboring live nodes of vertex i.
       !C PE_in_charge has info on border domains.
       
       PE_in_charge = min_rank_in_1_path_left_nds

       !C PE_in_charge is changed to have minimum rank among neighboring nodes.
       DO i = 1, NEIBPETOT
          DO j =  STACK_EXPORT(i-1)+1, STACK_EXPORT(i)
             k = NOD_EXPORT(j)
             pe = min_rank_in_1_path_left_nds(k)
             
             !C if pe is NPROCS, then the node is already checked.
             if(pe < NPROCS) then
                DO l = PNI(2*k-2)+1,PNI(2*k-1)
                   node = IAL(NI(l))
                   if(pe > min_rank_in_1_path_left_nds(node)) pe = min_rank_in_1_path_left_nds(node)
                END DO
                DO l = PNI(2*k-1)+1,PNI(2*k)
                   node = IAU(NI(l))
                   if(pe > min_rank_in_1_path_left_nds(node)) pe = min_rank_in_1_path_left_nds(node)
                END DO
                PE_in_charge(k) = pe
             end if
          END DO
       END DO

       CALL SOLVER_SEND_RECV2I(NP, STACK_EXPORT(NEIBPETOT),         &
            &       STACK_IMPORT(NEIBPETOT), NEIBPETOT, NEIBPE,     &
            &       STACK_IMPORT(0:), NOD_IMPORT, STACK_EXPORT(0:), &
            &       NOD_EXPORT, WSI, WRI, PE_in_charge, SOLVER_COMM,&
            &       my_rank)
       !C== end: PE_in_charge is created from min_rank_in_1_path_left_nds

       local_finish = .true.

       !C-- exec aggregation in COMM tables
       pli = 1
       DO i = 1, NEIBPETOT
          sni = sorted_NEIBPE(i)
          des_PE = NEIBPE(sni)
          IF(my_rank < des_PE) THEN
             DO WHILE(PE_list(pli,1)/=des_pe)
                pli=pli+1
             END DO
             IF(pli>PE_list_size) STOP "unexpected error in border_aggregates"
             nfei=PE_list(pli,3)

             !C== inverse_vector creation
             inverse_vector_size=STACK_EXPORT(NEIBPETOT)+STACK_IMPORT(NEIBPETOT)
             IF(PE_num_size/=0) inverse_vector_size=inverse_vector_size+        &
                  & in_nodes_for_each_PE(PE_num_size)

             ALLOCATE(inverse_vector(inverse_vector_size))
             k=0
             DO j=STACK_EXPORT(sni-1)+1,STACK_EXPORT(sni)
                k=k+1
                inverse_vector(k)=NOD_EXPORT(j)+ZERO_ORIGIN
             END DO
             DO j=STACK_IMPORT(sni-1)+1,STACK_IMPORT(sni)
                k=k+1
                inverse_vector(k)=NOD_IMPORT(j)+ZERO_ORIGIN
             END DO
             IF(nfei>0) THEN
                DO j=in_nodes_for_each_PE(nfei-1)+1,in_nodes_for_each_PE(nfei)
                   k=k+1
                   inverse_vector(k)=nodes_for_each_PE(j)+ZERO_ORIGIN
                END DO
             END IF
             !C== end inverse_vector creation


             !C== aggregation on border
             DO j=STACK_EXPORT(sni-1)+1,STACK_EXPORT(sni)
                tf=.TRUE.
                node=NOD_EXPORT(j)+ZERO_ORIGIN
                IF(PE_in_charge(node)==my_rank .AND. node_index(node)==0) THEN
                   l=node_index(node)
                   DO k=PNI(2*node-2)+1,PNI(2*node-1)
                      index=IAL(NI(k))+ZERO_ORIGIN
                      l=l+node_index(index)
                      tf=tf.AND.(my_rank<=PE_in_charge(index))
                   END DO
                   DO k=PNI(2*node-1)+1,PNI(2*node)
                      index=IAU(NI(k))+ZERO_ORIGIN
                      l=l+node_index(index)
                      tf=tf.AND.(my_rank<=PE_in_charge(index))
                   END DO

                   if(tf) then
                      !C mark the node as processed
                      min_rank_in_1_path_left_nds(node) = NPROCS
                   else
                      local_finish = .false.
                   end if


                   IF(l==0 .AND. tf) THEN
                      node_index(node)=-2
                      DO k=PNI(2*node-2)+1,PNI(2*node-1)
                         node_index         (IAL(NI(k))+ZERO_ORIGIN) = -1
                         !C mark the node as processed
                         min_rank_in_1_path_left_nds(IAL(NI(k))+ZERO_ORIGIN) = NPROCS
                      END DO
                      DO k=PNI(2*node-1)+1,PNI(2*node)
                         node_index         (IAU(NI(k))+ZERO_ORIGIN) = -1
                         !C mark the node as processed
                         min_rank_in_1_path_left_nds(IAU(NI(k))+ZERO_ORIGIN) = NPROCS
                      END DO
                   END IF
                END IF
             END DO
             
             k = recv_from_PE(sni)
             IF(k > 0) THEN
                nds_sz_for_each_PE = recv_buf(recv_cnt_ary(k-1)+1)
                nd_base            = nds_sz_for_each_PE+1+recv_cnt_ary(k-1)
!!$             recv_buf(1,k)=0

                nd_in=nd_base
                sz_in=1+recv_cnt_ary(k-1)
                DO j=STACK_IMPORT(sni-1)+1,STACK_IMPORT(sni)
                   node=NOD_IMPORT(j)+ZERO_ORIGIN

                   !C this if shows the this rank has the neighboring info of node.
                   IF(min_rank_in_1_path(node) == my_rank) then
                      in_s  = nd_in+1
                      sz_in = sz_in+1
                      sz    = recv_buf(sz_in)
                      nd_in = nd_in+sz
                      
                      IF(PE_in_charge(node) == my_rank .and. node_index(node) == 0) THEN

                         tf = .TRUE.
                         l = node_index(node)
                         DO index = in_s, nd_in
                            rbd = recv_buf(index)
                            IF(rbd > 0) THEN
                               nd = inverse_vector(rbd)
                               l  = l+node_index(nd)
                               tf = tf.AND.(my_rank<=PE_in_charge(nd))
                            END IF
                         END DO
                         
                         if(tf) then 
                            !C mark the node as processed
                            !C this case includes node_index(node) == 0
                            min_rank_in_1_path_left_nds(node) = NPROCS
                         else
                            local_finish = .false.
                         end if

                         IF(l==0 .AND. tf) THEN
                            node_index(node) = -2

                            DO index=in_s,nd_in
                               rbd = recv_buf(index)
                               IF(rbd > 0) then
                                  node_index(inverse_vector(rbd)) = -1
                                  !C mark the node as processed
                                  min_rank_in_1_path_left_nds(inverse_vector(rbd)) = NPROCS
                               end IF
                            END DO
                         END IF
                      END IF
                   END IF
                END DO
             END IF

             !C== end: aggregation on border
             DEALLOCATE(inverse_vector)
          END IF
       END DO
       
       !C-- end: exec aggregation in COMM tables

       !C== the result of aggregation and min_rank_in_1_path_left_nds are communicated
       k=0
       l=0

       DO i = 1,NEIBPETOT
          des_PE = NEIBPE(i)
          IF(des_pe>my_rank) THEN
             l = l+1
             istart = STACK_IMPORT(i-1)
             inum   = STACK_IMPORT(i)-istart
             DO j = istart+1,istart+inum
                send_buf(istart+j) = node_index(NOD_IMPORT(j)+ZERO_ORIGIN)
             END DO
             DO j = istart+1,istart+inum
                send_buf(istart+inum+j) = min_rank_in_1_path_left_nds(NOD_IMPORT(j)+ZERO_ORIGIN)
             END DO

             !C MPI_ISEND
             CALL MPI_ISEND(send_buf(2*istart+1), 2*inum, MPI_INTEGER, &
                  &  des_PE,0,SOLVER_COMM,req1(l),ierr)
          ELSE
             k = k+1
             istart = STACK_EXPORT(i-1)
             inum   = STACK_EXPORT(i)-istart
             !C MPI_IRECV
             CALL MPI_IRECV(recv_buf2(2*istart+1), 2*inum, MPI_INTEGER,  &
                  &  des_PE, 0, SOLVER_COMM, req2(k), ierr)
          END IF
       END DO
       IF(k>0) CALL MPI_WAITALL(k, req2, sta2, ierr)
       IF(l>0) CALL MPI_WAITALL(l, req1, sta1, ierr)
       
       !C-- assign by large to small number order of PE 
       DO i = NEIBPETOT,1,-1
          sni = sorted_NEIBPE(i)
          IF(NEIBPE(sni) < my_rank) THEN
             istart = STACK_EXPORT(sni-1)
             inum   = STACK_EXPORT(sni)-istart
             DO j = istart+1, istart+inum
                k = NOD_EXPORT(j)+ZERO_ORIGIN
                node_index(k)                  = recv_buf2(istart+j)
                min_rank_in_1_path_left_nds(k) = recv_buf2(istart+inum+j)
             END DO
          END IF
       END DO
       
       tmp(1:NP)      = node_index(1:NP)
       tmp(NP+1:2*NP) = min_rank_in_1_path_left_nds(1:NP)
       
       CALL SOLVER_SEND_RECV2I_n(NP, recv_buf_size, send_buf_size, &
            & NEIBPETOT, NEIBPE, STACK_IMPORT(0:), NOD_IMPORT,     &
            & STACK_EXPORT(0:),  NOD_EXPORT, recv_buf2, send_buf,  &
            & tmp, SOLVER_COMM, my_rank, 2)
       
       node_index(1:NP)                  = tmp(1:NP)
       min_rank_in_1_path_left_nds(1:NP) = tmp(NP+1:2*NP) 
       
       !C== end: the result of aggregation and min_rank_in_1_path_left_nds are communicated
       
       call MPI_allreduce(local_finish, global_finish, 1, MPI_LOGICAL, MPI_LAND, SOLVER_COMM, ierr)

       if(global_finish) exit
    end do

    DEALLOCATE(WSI)
    DEALLOCATE(WRI)
    
    DEALLOCATE(tmp)
    DEALLOCATE(send_buf)    
    DEALLOCATE(recv_buf2)
    DEALLOCATE(recv_buf)
    DEALLOCATE(recv_cnt_ary)
    DEALLOCATE(recv_from_PE)
    
    !C node_index is correct in the own domain, not in the domain of NOD_IMPORT.
    DEALLOCATE(min_rank_in_1_path) 
    DEALLOCATE(PE_in_charge)
    DEALLOCATE(req1)
    DEALLOCATE(req2)
    DEALLOCATE(sta1)
    DEALLOCATE(sta2)

  END SUBROUTINE border_aggregate


  SUBROUTINE aggregate(NP, N, &
       & INL, INU, IAL, IAU, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
       & STACK_EXPORT, NOD_EXPORT, &
       & PNI, NI, LEVEL_NO, node_index, in_aggregates_result, &
       & aggregates_result, in_aggregates_result_size, my_rank,         &
       & aggregate_table_size, aggregate_table_array,                   &
       & aggregate_number_in_table, SOLVER_COMM, NPROCS, GIN_aggregate, &
       & PE_num_size, PE_nums, nodes_for_each_PE, in_nodes_for_each_PE, &
       & PE_list_size, PE_list, kinds)
    
    USE queue_mod
    USE int_vec_module, only: int_vec, int_vec_initialize, check_and_extend
    IMPLICIT NONE
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'

    INTEGER(kind=kint), INTENT(in) :: NI(:),PNI(0:),LEVEL_NO
    INTEGER(kind=kint), INTENT(inout) :: node_index(:)
    INTEGER(kind=kint), pointer :: in_aggregates_result(:)
    INTEGER(kind=kint), pointer :: aggregates_result(:)
    INTEGER(kind=kint), INTENT(out) :: in_aggregates_result_size
    INTEGER(kind=kint), INTENT(in) :: my_rank, PE_list_size
    INTEGER(kind=kint), INTENT(in) :: PE_list(:, :)
    !C kinds = 1,or 2,or 3
    INTEGER(kind=kint), INTENT(in) :: kinds

    REAL   (kind=kreal), POINTER :: AL(:), AU(:), D(:)
    INTEGER(kind=kint), POINTER :: INL(:), IAU(:)
    INTEGER(kind=kint), POINTER :: IAL(:), INU(:)
    INTEGER(kind=kint), POINTER :: NOD_EXPORT(:), NOD_IMPORT(:), NEIBPE(:)
    INTEGER(kind=kint), POINTER :: STACK_EXPORT(:), STACK_IMPORT(:)
    INTEGER(kind=kint) :: NEIBPETOT, N, NP

    INTEGER(kind=kint) :: i, j, k, l, aggr, size_of_aggregates
    INTEGER(kind=kint), allocatable :: Temp(:,:)

    INTEGER(kind=kint),parameter :: aggregate_number = 1
    INTEGER(kind=kint),parameter :: aggregate_index  = 2
    
    INTEGER(kind=kint), ALLOCATABLE :: P_queue(:,:)
    INTEGER(kind=kint), ALLOCATABLE :: aggregate_flags(:)
    INTEGER(kind=kint), ALLOCATABLE :: node_record(:)
    INTEGER(kind=kint) :: queue_size
    
    INTEGER(kind=kint),intent(in)  :: aggregate_table_size, NPROCS
    type(int_vec), pointer :: aggregate_table_array(:)
    INTEGER(kind=kint),intent(out) :: aggregate_number_in_table(:)

    INTEGER(kind=kint),intent(in)  :: SOLVER_COMM
    INTEGER(kind=kint)             :: ierr
    INTEGER(kind=kint),intent(out) :: GIN_aggregate(0:NPROCS)

    INTEGER(kind=kint), intent(in) :: PE_num_size
    INTEGER(kind=kint), pointer    :: PE_nums(:)
    INTEGER(kind=kint), pointer    :: nodes_for_each_PE(:)
    INTEGER(kind=kint), pointer    :: in_nodes_for_each_PE(:)
    
    INTEGER(kind=kint), allocatable :: send_buf(:)
    INTEGER(kind=kint), allocatable :: in_send_buf_for_PEs(:)
    INTEGER(kind=kint), allocatable :: recv_buf(:)
    INTEGER(kind=kint), allocatable :: req1(:),req2(:)
    INTEGER(kind=kint), allocatable :: sta1(:,:), sta2(:,:)

    
    INTEGER(kind=kint) :: inum,istart,maxbufsize,position
    INTEGER(kind=kint) :: Neib_index,PE_num_index,dest,m,lc
    INTEGER(kind=kint), allocatable :: inner_nd(:)

    if(kinds < 1 .or. kinds > 3) stop "error in aggregate()."    

    ALLOCATE(Temp(0:NP, 2))
    Temp = 0    

    if(kinds == 1) then
       !C greedy case
       !C== creating independent aggregate
       size_of_aggregates = 0
       DO i = 1, N
          IF(node_index(i) == -2) THEN
             size_of_aggregates=size_of_aggregates+1
             DO k=PNI(2*i-2)+1,PNI(2*i-1)
                j=IAL(NI(k))+ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates
                node_index(j)=-1
             END DO
             DO k=PNI(2*i-1)+1,PNI(2*i)
                j=IAU(NI(k))+ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates
                node_index(j)=-1
             END DO
             Temp(i, aggregate_number) = size_of_aggregates
          END IF
       END DO
       ALLOCATE(inner_nd(N))
       inner_nd = 1
       DO i = 1, STACK_EXPORT(NEIBPETOT)
          inner_nd(NOD_EXPORT(i)+ZERO_ORIGIN) = 0
       END DO
       DO i = 1, N
          IF(node_index(i) == 0 .AND. inner_nd(i) == 1) THEN
             l = 0
             DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                j = IAL(NI(k))+ZERO_ORIGIN
                l = l + node_index(j)
             END DO
             DO k = PNI(2 * i - 1) + 1, PNI(2 * i)
                j = IAU(NI(k))+ZERO_ORIGIN
                l = l + node_index(j)
             END DO
             IF(l == 0) THEN
                
                size_of_aggregates = size_of_aggregates + 1

                DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                   j = IAL(NI(k))+ZERO_ORIGIN
                   Temp(j, aggregate_number) = size_of_aggregates
                   node_index(j) = -1
                END DO
                
                DO k = PNI(2 * i - 1) + 1, PNI(2 * i)
                   j = IAU(NI(k))+ZERO_ORIGIN
                   Temp(j, aggregate_number) = size_of_aggregates
                   node_index(j) = -1
                END DO

                Temp(i, aggregate_number) = size_of_aggregates
                node_index(i) = -1

             END IF
          END IF
       END DO

       deallocate(inner_nd)
       !C== end: creating independent aggregates
    else if(kinds == 2 .or. kinds == 3 ) then
       !C In these cases, priority queue is utilized

       ALLOCATE(P_queue(2,N))
       ALLOCATE(aggregate_flags(N))
       ALLOCATE(node_record(N))

       P_queue(1:2, 1:N)=0

       !C== enqueue internal nodes using aggregate_flags
       !C== node_index -2 -1 0: 
       !C== root, determined, undtermined
       !C== node_index 0 -> part of the queue data >0
       aggregate_flags(1:N) = 1
       DO i = 1, STACK_EXPORT(NEIBPETOT)
          aggregate_flags(NOD_EXPORT(i)+ZERO_ORIGIN) = 0
       END DO


       
       queue_size = 0

       DO i = 1, N 
          if(aggregate_flags(i) == 1 .and. node_index(i) /= -1) then
             queue_size = queue_size + 1
             node_index(i) = queue_size
             P_queue(1, queue_size) = 1
             P_queue(2, queue_size) = i

             aggregate_flags(i) = 0
          end if
       END DO


       !C== dequeue on nodes around the nodes on EXPORT TABLE with score -1 
       !C== node_index -2 -1 0 >0
       !C== root, determined, undetermined but not root, undetermined and part of queue
       DO l = 1, STACK_EXPORT(NEIBPETOT)
          i = NOD_EXPORT(l)+ZERO_ORIGIN
          if(node_index(i) == -1) then
             DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                j = IAL( NI(k) ) + ZERO_ORIGIN
                if(node_index(j) > 0) then

                   call dequeue(j, P_queue, node_index, queue_size)
                   node_index(j) = 0
                end if
             END DO
             DO k = PNI(2 * i - 1) + 1, PNI(2 * i)
                j=IAU( NI(k) )+ZERO_ORIGIN
                if(node_index(j) > 0) then
                   call dequeue(j, P_queue, node_index,queue_size)
                   node_index(j) = 0
                end if
             END DO
          end if

       END DO
       
       !C== creating aggregates which are predetermined on borders
       !C== dequeue and scoreup on nodes around the nodes on EXPORT TABLE with score -2 
       size_of_aggregates = 0
       DO i = 1, N
          if(node_index(i) == -2) then
             size_of_aggregates = size_of_aggregates + 1
             DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                j=IAL(NI(k)) + ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates
                if(node_index(j) > 0) call dequeue(j, P_queue, node_index,queue_size)
                node_index(j) = -1
             END DO
             DO k = PNI(2 * i - 1) + 1,PNI(2 * i)
                j=IAU(NI(k))+ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates
                if(node_index(j) > 0) call dequeue(j, P_queue, node_index,queue_size)
                node_index(j) = -1
             END DO
             Temp(i, aggregate_number) = size_of_aggregates

             if(kinds==3) then
                !C-- this function weight the nodes around the aggregates on borders
                call dequeue_scoreup_Layer23(i, N, queue_size, P_queue,  & 
                     & node_index, aggregate_flags, node_record, NI, PNI, IAL, IAU)
             end if
          END if
       END DO

       
       !C== create aggregates internal part
       if(N > 0) then
          DO while(P_queue(1, 1) > 0)


             i = P_queue(2, 1)
             size_of_aggregates = size_of_aggregates + 1


             !C----layer1 start
             DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                j = IAL(NI(k)) + ZERO_ORIGIN

                Temp(j, aggregate_number) = size_of_aggregates
                

                if(node_index(j)>0) then
                   !C dequeue node j
                   call dequeue(j, P_queue, node_index, queue_size)
                end if
                node_index(j)=-1
                aggregate_flags(j)=i

             END DO

             DO k=PNI(2*i-1)+1,PNI(2*i)
                j=IAU( NI(k) )+ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates


                if(node_index(j)>0) then
                   !C dequeue node j
                   call dequeue(j, P_queue, node_index,queue_size)
                end if
                node_index(j)=-1
                aggregate_flags(j)=i
             END DO
             Temp(i, aggregate_number) = size_of_aggregates



             !C dequeue node i
             call dequeue(i, P_queue, node_index,queue_size)

             node_index(i) = -1
             aggregate_flags(i) = i
             !C----layer1 end


             !C----layer2 and 3 
             call dequeue_scoreup_Layer23(i, N, queue_size, P_queue,  & 
                  & node_index, aggregate_flags, node_record, NI, PNI, IAL, IAU)

          end DO
       end if
       deallocate(node_record,P_queue,aggregate_flags)
    end if


    !C== sendrecv independent aggregates
    CALL MPI_ALLGATHER(size_of_aggregates,1,MPI_INTEGER, &
         & GIN_aggregate(1),1,MPI_INTEGER,SOLVER_COMM,ierr)
    GIN_aggregate(0)=0
    DO i=2,NPROCS
       GIN_aggregate(i)=GIN_aggregate(i-1)+GIN_aggregate(i)
    END DO
    

    aggregate_number_in_table=0
    maxbufsize = STACK_IMPORT(NEIBPETOT)+STACK_EXPORT(NEIBPETOT)
    if(PE_num_size /= 0) maxbufsize=maxbufsize+in_nodes_for_each_PE(PE_num_size)+1
    
    allocate(req1(PE_list_size))
    allocate(sta1(MPI_STATUS_SIZE,PE_list_size))
    allocate(req2(PE_list_size))
    allocate(sta2(MPI_STATUS_SIZE,PE_list_size))

    allocate(send_buf(maxbufsize))
    allocate(recv_buf(maxbufsize))
    allocate(in_send_buf_for_PEs(0:PE_list_size))
    in_send_buf_for_PEs(0)=0
    position=0
    DO i=1,PE_list_size
       Neib_index=PE_list(i,2)
       PE_num_index=PE_list(i,3)
       if(Neib_index>0) then
          istart= STACK_EXPORT(Neib_index-1)
          inum  = STACK_EXPORT(Neib_index)-istart
          DO j=istart+1,istart+inum
             k=Temp(NOD_EXPORT(j)+ZERO_ORIGIN,aggregate_number)
             if(k>0) k=k+GIN_aggregate(my_rank)
             position=position+1
             send_buf(position)=k
          END DO
          istart= STACK_IMPORT(Neib_index-1)
          inum  = STACK_IMPORT(Neib_index)-istart
          DO j=istart+1,istart+inum
             k=Temp(NOD_IMPORT(j)+ZERO_ORIGIN,aggregate_number)
             if(k>0) k=k+GIN_aggregate(my_rank)
             position=position+1
             send_buf(position)=k
          END DO
       end if
       if(PE_num_index>0) then
          istart= in_nodes_for_each_PE(PE_num_index-1)
          inum = in_nodes_for_each_PE(PE_num_index)-istart
          DO j=istart+1,istart+inum
             k=Temp(nodes_for_each_PE(j)+ZERO_ORIGIN,aggregate_number)
             if(k>0) k=k+GIN_aggregate(my_rank)
             position=position+1
             send_buf(position)=k
          END DO
       end if
       in_send_buf_for_PEs(i)=position
    end DO
    DO i=1,PE_list_size
       dest=PE_list(i,1)
       istart=in_send_buf_for_PEs(i-1)
       inum=in_send_buf_for_PEs(i)-in_send_buf_for_PEs(i-1)
       CALL MPI_ISEND(send_buf(istart+1),inum,MPI_INTEGER, &
            & dest,0,SOLVER_COMM,req1(i),ierr)

       CALL MPI_IRECV(recv_buf(istart+1),inum,MPI_INTEGER, &
            & dest,0,SOLVER_COMM,req2(i),ierr)
    END DO

    CALL MPI_WAITALL(PE_list_size, req2, sta2, ierr)
    CALL MPI_WAITALL(PE_list_size, req1, sta1, ierr)
    deallocate(send_buf)
    deallocate(in_send_buf_for_PEs)

    position = 0
    DO i = 1, PE_list_size
       dest = PE_list(i, 1)
       Neib_index = PE_list(i, 2)
       PE_num_index = PE_list(i, 3)
       
       
       if(Neib_index > 0) then
          istart = STACK_IMPORT(Neib_index - 1)
          inum   = STACK_IMPORT(Neib_index) - STACK_IMPORT(Neib_index - 1)
          DO j = istart + 1, istart + inum
             position = position + 1

             !C== assigning part
             k = recv_buf(position)
             if(k > 0) then
                CALL assign_aggregate_table(aggregate_number_in_table,     &
                     & aggregate_table_array, aggregate_table_size, k, lc, &
                     & i, NPROCS, size_of_aggregates)

                Temp(NOD_IMPORT(j)+ZERO_ORIGIN, aggregate_number) = lc
                node_index(NOD_IMPORT(j)+ZERO_ORIGIN) = -1
             end if
             !C== end:assigning part             
          END DO
          istart = STACK_EXPORT(Neib_index - 1)
          inum   = STACK_EXPORT(Neib_index) - STACK_EXPORT(Neib_index - 1)
          DO j = istart + 1, istart + inum
             position =position + 1
             
             !C== assigning part
             k = recv_buf(position)
             if(k > 0) then
                CALL assign_aggregate_table(aggregate_number_in_table,     &
                     & aggregate_table_array, aggregate_table_size, k, lc, &
                     & i, NPROCS, size_of_aggregates)

                Temp(NOD_EXPORT(j)+ZERO_ORIGIN, aggregate_number) = lc
                node_index(NOD_EXPORT(j)+ZERO_ORIGIN) = -1
             end if
             !C== end:assigning part             
          END DO
       end if
       if(PE_num_index > 0) then
          istart = in_nodes_for_each_PE(PE_num_index - 1)
          inum = in_nodes_for_each_PE(PE_num_index) - istart
          DO j = istart + 1, istart + inum
             position = position + 1

             !C== assigning part
             k = recv_buf(position)
             if(k > 0) then
                CALL assign_aggregate_table(aggregate_number_in_table,     &
                     & aggregate_table_array, aggregate_table_size, k, lc, &
                     & i, NPROCS, size_of_aggregates)

                Temp(nodes_for_each_PE(j)+ZERO_ORIGIN, aggregate_number) = lc
                node_index(nodes_for_each_PE(j)+ZERO_ORIGIN) = -1
             end if
             !C== end:assigning part             

          END DO
       end if

    END DO

    deallocate(recv_buf)
    deallocate(req1)
    deallocate(req2)
    deallocate(sta1)
    deallocate(sta2)

    !C== end: sendrecv independent aggregates

    !C== left nodes are associated to independent aggregates
    DO i = 1, N

       if(node_index(i) == 0) then
          aggr = -1
          k = PNI(2 * i - 2) + 1
          DO while(aggr <= 0 .and. k <= PNI(2 * i - 1))
             aggr = Temp(IAL( NI(k) )+ZERO_ORIGIN, aggregate_number)
             k = k + 1
          END DO
          if(aggr <= 0) then
             k = PNI(2 * i - 1) + 1
             DO while( aggr <= 0 .and. k <= PNI(2 * i))
                aggr = Temp(IAU( NI(k) )+ZERO_ORIGIN, aggregate_number)
                k = k + 1
             END DO
          end if
          !C in case of aggr<0, neiboring node_index is -1 for some reason.
          !C added node's aggregate number is set to minus to avoid adding nodes like chain reaction
          if(aggr > 0) Temp(i, aggregate_number) = -1*aggr
       end if
       
    END DO
    !C==end: left nodes are associated to independent aggregates

    !C added nodes in minus aggregate number is changed to plus
    DO i=1, N
       if(Temp(i,aggregate_number)<0) Temp(i,aggregate_number) = Temp(i,aggregate_number)*-1
    END DO


    !C== Temp is written to the aggregates_results.
    !C   size_of_aggregates is the number of aggregates in 1:NP
    in_aggregates_result_size = size_of_aggregates
    ALLOCATE(in_aggregates_result(0:size_of_aggregates))
    ALLOCATE(aggregates_result(N))

    DO i=1, N
       aggr=Temp(i,aggregate_number)
       if(aggr>0) Temp(aggr,aggregate_index) = Temp(aggr,aggregate_index)+1
    END DO
    in_aggregates_result(0)=0    
    in_aggregates_result_size=size_of_aggregates
    DO i=1,size_of_aggregates
       Temp(i,aggregate_index)=Temp(i,aggregate_index)+Temp(i-1,aggregate_index)
       in_aggregates_result(i)=Temp(i-1,aggregate_index)
    END DO

    DO i=1,N
       aggr=Temp(i,aggregate_number)
       if(aggr>0) then
          k=in_aggregates_result(aggr)+1
          in_aggregates_result(aggr)=k
          aggregates_result(k)=i
       end if
    END DO
    
    DEALLOCATE(Temp)
    !C== end: Temp is written to the aggregates_results.

  END SUBROUTINE aggregate


  SUBROUTINE aggregate_and_transition(NP, N, &
       & INL, INU, IAL, IAU, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
       & STACK_EXPORT, NOD_EXPORT, &
       & PNI, NI, LEVEL_NO, node_index, in_aggregates_result, &
       & aggregates_result, in_aggregates_result_size, my_rank,         &
       & aggregate_table_size, aggregate_table_array,                   &
       & aggregate_number_in_table, SOLVER_COMM, NPROCS, GIN_aggregate, &
       & PE_num_size, PE_nums, nodes_for_each_PE, in_nodes_for_each_PE, &
       & PE_list_size, PE_list, kinds, &
       & sender_fl, receiver_fl, partner)
    
    USE queue_mod
    USE int_vec_module, only: int_vec, int_vec_initialize, check_and_extend
    IMPLICIT NONE
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'

    INTEGER(kind=kint), INTENT(in) :: NI(:),PNI(0:),LEVEL_NO
    INTEGER(kind=kint), INTENT(inout) :: node_index(:)
    INTEGER(kind=kint), pointer :: in_aggregates_result(:)
    INTEGER(kind=kint), pointer :: aggregates_result(:)
    INTEGER(kind=kint), INTENT(out) :: in_aggregates_result_size
    INTEGER(kind=kint), INTENT(in) :: my_rank
    INTEGER(kind=kint), INTENT(inout) :: PE_list_size
    INTEGER(kind=kint), INTENT(inout) :: PE_list(:, :)
    !C kinds = 1,or 2,or 3
    INTEGER(kind=kint), INTENT(in) :: kinds

    REAL   (kind=kreal), POINTER :: AL(:), AU(:), D(:)
    INTEGER(kind=kint), POINTER :: INL(:), IAU(:)
    INTEGER(kind=kint), POINTER :: IAL(:), INU(:)
    INTEGER(kind=kint), POINTER :: NOD_EXPORT(:), NOD_IMPORT(:), NEIBPE(:)
    INTEGER(kind=kint), POINTER :: STACK_EXPORT(:), STACK_IMPORT(:)
    INTEGER(kind=kint) :: NEIBPETOT, N, NP

    INTEGER(kind=kint) :: i, j, k, l, aggr, size_of_aggregates
    INTEGER(kind=kint), allocatable :: Temp(:,:)

    INTEGER(kind=kint),parameter :: aggregate_number = 1
    INTEGER(kind=kint),parameter :: aggregate_index  = 2
    
    INTEGER(kind=kint), ALLOCATABLE :: P_queue(:,:)
    INTEGER(kind=kint), ALLOCATABLE :: aggregate_flags(:)
    INTEGER(kind=kint), ALLOCATABLE :: node_record(:)
    INTEGER(kind=kint) :: queue_size
    
    INTEGER(kind=kint),intent(in)  :: aggregate_table_size, NPROCS
    type(int_vec), pointer :: aggregate_table_array(:)
    INTEGER(kind=kint),intent(out) :: aggregate_number_in_table(:)

    INTEGER(kind=kint),intent(in)  :: SOLVER_COMM
    INTEGER(kind=kint)             :: ierr
    INTEGER(kind=kint),intent(out) :: GIN_aggregate(0:NPROCS)

    INTEGER(kind=kint), intent(in) :: PE_num_size
    INTEGER(kind=kint), pointer    :: PE_nums(:)
    INTEGER(kind=kint), pointer    :: nodes_for_each_PE(:)
    INTEGER(kind=kint), pointer    :: in_nodes_for_each_PE(:)
    
    INTEGER(kind=kint), allocatable :: send_buf(:)
    INTEGER(kind=kint), allocatable :: in_send_buf_for_PEs(:)
    INTEGER(kind=kint), allocatable :: recv_buf(:)
    INTEGER(kind=kint), allocatable :: req1(:),req2(:)
    INTEGER(kind=kint), allocatable :: sta1(:,:), sta2(:,:)

    
    INTEGER(kind=kint) :: inum,istart,maxbufsize,position
    INTEGER(kind=kint) :: Neib_index,PE_num_index,dest,m,lc
    INTEGER(kind=kint), allocatable :: inner_nd(:)
    
    logical, intent(in) :: sender_fl, receiver_fl
    integer(kind=kint), intent(in) :: partner
    integer(kind=kint) :: partner_aggre_size, sta(MPI_STATUS_SIZE)
    
    integer(kind=kint) :: pls, index

    if(kinds < 1 .or. kinds > 3) stop "error in aggregate()."    

    ALLOCATE(Temp(0:NP, 2))
    Temp = 0    

    if(kinds == 1) then
       !C greedy case
       !C== creating independent aggregate
       size_of_aggregates = 0
       DO i = 1, N
          IF(node_index(i) == -2) THEN
             size_of_aggregates=size_of_aggregates+1
             DO k=PNI(2*i-2)+1,PNI(2*i-1)
                j=IAL(NI(k))+ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates
                node_index(j)=-1
             END DO
             DO k=PNI(2*i-1)+1,PNI(2*i)
                j=IAU(NI(k))+ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates
                node_index(j)=-1
             END DO
             Temp(i, aggregate_number) = size_of_aggregates
          END IF
       END DO
       ALLOCATE(inner_nd(N))
       inner_nd = 1
       DO i = 1, STACK_EXPORT(NEIBPETOT)
          inner_nd(NOD_EXPORT(i)+ZERO_ORIGIN) = 0
       END DO
       DO i = 1, N
          IF(node_index(i) == 0 .AND. inner_nd(i) == 1) THEN
             l = 0
             DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                j = IAL(NI(k))+ZERO_ORIGIN
                l = l + node_index(j)
             END DO
             DO k = PNI(2 * i - 1) + 1, PNI(2 * i)
                j = IAU(NI(k))+ZERO_ORIGIN
                l = l + node_index(j)
             END DO
             IF(l == 0) THEN
                
                size_of_aggregates = size_of_aggregates + 1

                DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                   j = IAL(NI(k))+ZERO_ORIGIN
                   Temp(j, aggregate_number) = size_of_aggregates
                   node_index(j) = -1
                END DO
                
                DO k = PNI(2 * i - 1) + 1, PNI(2 * i)
                   j = IAU(NI(k))+ZERO_ORIGIN
                   Temp(j, aggregate_number) = size_of_aggregates
                   node_index(j) = -1
                END DO

                Temp(i, aggregate_number) = size_of_aggregates
                node_index(i) = -1

             END IF
          END IF
       END DO

       deallocate(inner_nd)
       !C== end: creating independent aggregates
    else if(kinds == 2 .or. kinds == 3 ) then
       !C In these cases, priority queue is utilized

       ALLOCATE(P_queue(2,N))
       ALLOCATE(aggregate_flags(N))
       ALLOCATE(node_record(N))

       P_queue(1:2, 1:N)=0

       !C== enqueue internal nodes using aggregate_flags
       !C== node_index -2 -1 0: 
       !C== root, determined, undtermined
       !C== node_index 0 -> part of the queue data >0
       aggregate_flags(1:N) = 1
       DO i = 1, STACK_EXPORT(NEIBPETOT)
          aggregate_flags(NOD_EXPORT(i)+ZERO_ORIGIN) = 0
       END DO


       queue_size = 0

       DO i = 1, N 
          if(aggregate_flags(i) == 1 .and. node_index(i) /= -1) then
             queue_size = queue_size + 1
             node_index(i) = queue_size
             P_queue(1, queue_size) = 1
             P_queue(2, queue_size) = i

             aggregate_flags(i) = 0
          end if
       END DO


       !C== dequeue on nodes around the nodes on EXPORT TABLE with score -1 
       !C== node_index -2 -1 0 >0
       !C== root, determined, undetermined but not root, undetermined and part of queue
       DO l = 1, STACK_EXPORT(NEIBPETOT)
          i = NOD_EXPORT(l)+ZERO_ORIGIN
          if(node_index(i) == -1) then
             DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                j = IAL( NI(k) ) + ZERO_ORIGIN
                if(node_index(j) > 0) then

                   call dequeue(j, P_queue, node_index, queue_size)
                   node_index(j) = 0
                end if
             END DO
             DO k = PNI(2 * i - 1) + 1, PNI(2 * i)
                j=IAU( NI(k) )+ZERO_ORIGIN
                if(node_index(j) > 0) then
                   call dequeue(j, P_queue, node_index,queue_size)
                   node_index(j) = 0
                end if
             END DO
          end if

       END DO
       
       !C== creating aggregates which are predetermined on borders
       !C== dequeue and scoreup on nodes around the nodes on EXPORT TABLE with score -2 
       size_of_aggregates = 0
       DO i = 1, N
          if(node_index(i) == -2) then
             size_of_aggregates = size_of_aggregates + 1
             DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                j=IAL(NI(k)) + ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates
                if(node_index(j) > 0) call dequeue(j, P_queue, node_index,queue_size)
                node_index(j) = -1
             END DO
             DO k = PNI(2 * i - 1) + 1,PNI(2 * i)
                j=IAU(NI(k))+ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates
                if(node_index(j) > 0) call dequeue(j, P_queue, node_index,queue_size)
                node_index(j) = -1
             END DO
             Temp(i, aggregate_number) = size_of_aggregates

             if(kinds==3) then
                !C-- this function weight the nodes around the aggregates on borders
                call dequeue_scoreup_Layer23(i, N, queue_size, P_queue,  & 
                     & node_index, aggregate_flags, node_record, NI, PNI, IAL, IAU)
             end if
          END if
       END DO

       
       !C== create aggregates internal part
       if(N > 0) then
          DO while(P_queue(1, 1) > 0)


             i = P_queue(2, 1)
             size_of_aggregates = size_of_aggregates + 1


             !C----layer1 start
             DO k = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
                j = IAL(NI(k)) + ZERO_ORIGIN

                Temp(j, aggregate_number) = size_of_aggregates
                

                if(node_index(j)>0) then
                   !C dequeue node j
                   call dequeue(j, P_queue, node_index, queue_size)
                end if
                node_index(j)=-1
                aggregate_flags(j)=i

             END DO

             DO k=PNI(2*i-1)+1,PNI(2*i)
                j=IAU( NI(k) )+ZERO_ORIGIN
                Temp(j, aggregate_number) = size_of_aggregates


                if(node_index(j)>0) then
                   !C dequeue node j
                   call dequeue(j, P_queue, node_index,queue_size)
                end if
                node_index(j)=-1
                aggregate_flags(j)=i
             END DO
             Temp(i, aggregate_number) = size_of_aggregates



             !C dequeue node i
             call dequeue(i, P_queue, node_index,queue_size)

             node_index(i) = -1
             aggregate_flags(i) = i
             !C----layer1 end


             !C----layer2 and 3 
             call dequeue_scoreup_Layer23(i, N, queue_size, P_queue,  & 
                  & node_index, aggregate_flags, node_record, NI, PNI, IAL, IAU)

          end DO
       end if
       deallocate(node_record,P_queue,aggregate_flags)
    end if

    call cordinate_GIN_aggregate(sender_fl, receiver_fl, partner, partner_aggre_size,&
         & size_of_aggregates, GIN_aggregate, NPROCS, my_rank, SOLVER_COMM,          &
         & aggregate_number_in_table, aggregate_table_array, aggregate_table_size,   &
         & PE_list, PE_list_size)
    
    
    maxbufsize = STACK_IMPORT(NEIBPETOT)+STACK_EXPORT(NEIBPETOT)
    if(PE_num_size /= 0) maxbufsize=maxbufsize+in_nodes_for_each_PE(PE_num_size)+1
    
    allocate(req1(PE_list_size))
    allocate(sta1(MPI_STATUS_SIZE,PE_list_size))
    allocate(req2(PE_list_size))
    allocate(sta2(MPI_STATUS_SIZE,PE_list_size))

    allocate(send_buf(maxbufsize))
    allocate(recv_buf(maxbufsize))
    allocate(in_send_buf_for_PEs(0:PE_list_size))
    in_send_buf_for_PEs(0)=0
    position=0
    DO i=1,PE_list_size
       Neib_index=PE_list(i,2)
       PE_num_index=PE_list(i,3)
       if(Neib_index>0) then
          istart= STACK_EXPORT(Neib_index-1)
          inum  = STACK_EXPORT(Neib_index)-istart
          DO j=istart+1,istart+inum
             k=Temp(NOD_EXPORT(j)+ZERO_ORIGIN,aggregate_number)
             if(k>0) then
!C III modify number on sender part.                
                if(sender_fl) then
                   k=k+GIN_aggregate(partner)+partner_aggre_size
                else 
                   k=k+GIN_aggregate(my_rank)
                end if
             end if
             position=position+1
             send_buf(position)=k
          END DO
          istart= STACK_IMPORT(Neib_index-1)
          inum  = STACK_IMPORT(Neib_index)-istart
          DO j=istart+1,istart+inum
             k=Temp(NOD_IMPORT(j)+ZERO_ORIGIN,aggregate_number)
             if(k>0) then
!C III modify number on sender part.                
                if(sender_fl) then
                   k=k+GIN_aggregate(partner)+partner_aggre_size
                else 
                   k=k+GIN_aggregate(my_rank)
                end if
             end if

             position=position+1
             send_buf(position)=k
          END DO
       end if
       if(PE_num_index>0) then
          istart= in_nodes_for_each_PE(PE_num_index-1)
          inum = in_nodes_for_each_PE(PE_num_index)-istart
          DO j=istart+1,istart+inum
             k=Temp(nodes_for_each_PE(j)+ZERO_ORIGIN,aggregate_number)
             if(k>0) then
!C III modify number on sender part.                
                if(sender_fl) then
                   k=k+GIN_aggregate(partner)+partner_aggre_size
                else 
                   k=k+GIN_aggregate(my_rank)
                end if
             end if

             position=position+1
             send_buf(position)=k
          END DO
       end if
       in_send_buf_for_PEs(i)=position
    end DO

    DO i=1,PE_list_size
       dest=PE_list(i,1)
       istart=in_send_buf_for_PEs(i-1)
       inum=in_send_buf_for_PEs(i)-in_send_buf_for_PEs(i-1)
       CALL MPI_ISEND(send_buf(istart+1),inum,MPI_INTEGER, &
            & dest,0,SOLVER_COMM,req1(i),ierr)

       CALL MPI_IRECV(recv_buf(istart+1),inum,MPI_INTEGER, &
            & dest,0,SOLVER_COMM,req2(i),ierr)
    END DO


    CALL MPI_WAITALL(PE_list_size, req2, sta2, ierr)
    CALL MPI_WAITALL(PE_list_size, req1, sta1, ierr)
    deallocate(send_buf)
    deallocate(in_send_buf_for_PEs)

    
    !C PE_list_size will be increasing. so communication size is copied to pls.
    pls = PE_list_size

    position = 0
    DO i = 1, pls
       dest = PE_list(i, 1)
       Neib_index = PE_list(i, 2)
       PE_num_index = PE_list(i, 3)
       
       
       if(Neib_index > 0) then
          istart = STACK_IMPORT(Neib_index - 1)
          inum   = STACK_IMPORT(Neib_index) - STACK_IMPORT(Neib_index - 1)
          DO j = istart + 1, istart + inum
             position = position + 1

             !C== assigning part
             k = recv_buf(position)
             if(k > 0) then
                CALL determine_owner_index_frm_gl(k, GIN_aggregate, PE_list, index, PE_list_size, NPROCS, my_rank)
                !C when index == 0 then that is this own aggregates.
                if(index > 0) then
                   CALL assign_aggregate_table(aggregate_number_in_table,     &
                        & aggregate_table_array, aggregate_table_size, k, lc, &
                        & index, NPROCS, size_of_aggregates)
                else
                   !C this case deals with own aggregates coming from neighboring processes
                   lc = k - GIN_aggregate(my_rank)
                end if

                Temp(NOD_IMPORT(j)+ZERO_ORIGIN, aggregate_number) = lc
                node_index(NOD_IMPORT(j)+ZERO_ORIGIN) = -1

             end if
             !C== end:assigning part             
          END DO
          istart = STACK_EXPORT(Neib_index - 1)
          inum   = STACK_EXPORT(Neib_index) - STACK_EXPORT(Neib_index - 1)
          DO j = istart + 1, istart + inum
             position =position + 1
             
             !C== assigning part
             k = recv_buf(position)
             if(k > 0) then
                CALL determine_owner_index_frm_gl(k, GIN_aggregate, PE_list, index, PE_list_size, NPROCS, my_rank)
                if(index > 0) then
                   CALL assign_aggregate_table(aggregate_number_in_table,     &
                        & aggregate_table_array, aggregate_table_size, k, lc, &
                        & index, NPROCS, size_of_aggregates)
                else
                   !C this case deals with own aggregates coming from neighboring processes
                   lc = k - GIN_aggregate(my_rank)
                end if

                Temp(NOD_EXPORT(j)+ZERO_ORIGIN, aggregate_number) = lc
                node_index(NOD_EXPORT(j)+ZERO_ORIGIN) = -1
                
             end if
             !C== end:assigning part             
          END DO
       end if
       if(PE_num_index > 0) then
          istart = in_nodes_for_each_PE(PE_num_index - 1)
          inum = in_nodes_for_each_PE(PE_num_index) - istart
          DO j = istart + 1, istart + inum
             position = position + 1

             !C== assigning part
             k = recv_buf(position)
             if(k > 0) then
                CALL determine_owner_index_frm_gl(k, GIN_aggregate, PE_list, index, PE_list_size, NPROCS, my_rank)
                if(index > 0) then
                   CALL assign_aggregate_table(aggregate_number_in_table,     &
                        & aggregate_table_array, aggregate_table_size, k, lc, &
                        & index, NPROCS, size_of_aggregates)
                else
                   !C this case deals with own aggregates coming from neighboring processes
                   lc = k - GIN_aggregate(my_rank)
                end if

                Temp(nodes_for_each_PE(j)+ZERO_ORIGIN, aggregate_number) = lc
                node_index(nodes_for_each_PE(j)+ZERO_ORIGIN) = -1
             end if
             !C== end:assigning part             

          END DO
       end if

    END DO

    deallocate(recv_buf)
    deallocate(req1)
    deallocate(req2)
    deallocate(sta1)
    deallocate(sta2)
    !C== end: sendrecv independent aggregates


    !C== left nodes are associated to independent aggregates
    !C added aggregate number is set to minus to avoid nodes like chain reaction
    DO i = 1, N

       if(node_index(i) == 0) then
          aggr = -1
          k = PNI(2 * i - 2) + 1
          DO while(aggr <= 0 .and. k <= PNI(2 * i - 1))
             aggr = Temp(IAL( NI(k) )+ZERO_ORIGIN, aggregate_number)
             k = k + 1
          END DO
          if(aggr <= 0) then
             k = PNI(2 * i - 1) + 1
             DO while( aggr <= 0 .and. k <= PNI(2 * i))
                aggr = Temp(IAU( NI(k) )+ZERO_ORIGIN, aggregate_number)
                k = k + 1
             END DO
          end if
          !C in case of aggr<0, neiboring node_index is -1 for some reason.
          if(aggr > 0) Temp(i, aggregate_number) = -1*aggr
       end if
    END DO
    !C==end: left nodes are associated to independent aggregates

    !C== Temp is written to the aggregates_results.
    !C   size_of_aggregates is the number of aggregates in 1:NP
    in_aggregates_result_size = size_of_aggregates
    ALLOCATE(in_aggregates_result(0:size_of_aggregates))

    ALLOCATE(aggregates_result(N))

    DO i=1, N
       aggr=Temp(i,aggregate_number)
       !C minum aggregaate number in check part is changed plus
       if(aggr<0) then
          aggr = -1*aggr
          Temp(i,aggregate_number)=aggr
       end if
       if(aggr>0) Temp(aggr,aggregate_index) = Temp(aggr,aggregate_index)+1
    END DO
    in_aggregates_result(0)=0    
    in_aggregates_result_size=size_of_aggregates
    DO i=1,size_of_aggregates
       Temp(i,aggregate_index)=Temp(i,aggregate_index)+Temp(i-1,aggregate_index)
       in_aggregates_result(i)=Temp(i-1,aggregate_index)
    END DO
    
    DO i=1,N
       aggr=Temp(i,aggregate_number)
       if(aggr>0) then
          k=in_aggregates_result(aggr)+1
          in_aggregates_result(aggr)=k
          aggregates_result(k)=i
       end if
    END DO
    
    DEALLOCATE(Temp)
    !C== end: Temp is written to the aggregates_results.

  END SUBROUTINE aggregate_and_transition
  
  !C out: partner_aggre_size, size_of_aggregates, GIN_aggregate
  !C out: aggregate_table
  !C transferred GIN_aggregate(:) is created and aggregate_table is initialized.
  !C sender process records aggregates in aggregate_table_array
  subroutine cordinate_GIN_aggregate(sender_fl, receiver_fl, partner,            &
       & partner_aggre_size, size_of_aggregates, GIN_aggregate, NPROCS, my_rank, &
       & SOLVER_COMM, aggregate_number_in_table, aggregate_table_array,          &
       & aggregate_table_size, PE_list, PE_list_size)
    USE int_vec_module, only: int_vec, int_vec_initialize, check_and_extend
    implicit none
    include 'precision.inc'
    include 'mpif.h'
    
    integer(kind=kint), intent(out)   :: partner_aggre_size
    integer(kind=kint), intent(inout) :: size_of_aggregates, GIN_aggregate(0:)
    integer(kind=kint), intent(out) :: aggregate_number_in_table(:)
    type(int_vec), pointer :: aggregate_table_array(:)
    integer(kind=kint), intent(inout) :: PE_list(:,:), PE_list_size
    logical, intent(in) :: sender_fl, receiver_fl
    integer(kind=kint), intent(in)    :: partner, NPROCS, my_rank, SOLVER_COMM
    integer(kind=kint), intent(in)    :: aggregate_table_size
    
    integer(kind=kint) :: i, index, ierr, sta(MPI_STATUS_SIZE)
    
    
    !C I size_of_aggregates is sent from sender to receiver
    if(sender_fl .or. receiver_fl) then
       call MPI_SENDRECV(size_of_aggregates, 1, MPI_INTEGER, partner, 0, &
            &  partner_aggre_size, 1, MPI_INTEGER, partner, 0, SOLVER_COMM, &
            &  sta, ierr )
    end if
    
    
    !C== sendrecv independent aggregates

    
    !C II shift the number of nodes.
    if(sender_fl) then
#ifdef PRINT_REZ
       write(*,*) "sender:", my_rank, "to", partner
#endif
       CALL MPI_ALLGATHER(0,1,MPI_INTEGER, &
            & GIN_aggregate(1),1,MPI_INTEGER,SOLVER_COMM,ierr)
    else if(receiver_fl) then
#ifdef PRINT_REZ
       write(*,*) "receiver:", my_rank, "from", partner
#endif
       size_of_aggregates = size_of_aggregates + partner_aggre_size
       CALL MPI_ALLGATHER(size_of_aggregates,1,MPI_INTEGER, &
            & GIN_aggregate(1),1,MPI_INTEGER,SOLVER_COMM,ierr)
    else 
       CALL MPI_ALLGATHER(size_of_aggregates,1,MPI_INTEGER, &
            & GIN_aggregate(1),1,MPI_INTEGER,SOLVER_COMM,ierr)
    end if

    
    GIN_aggregate(0)=0
    DO i=2,NPROCS
       GIN_aggregate(i)=GIN_aggregate(i-1)+GIN_aggregate(i)
    END DO


    aggregate_number_in_table = 0
    
    !C III aggregates should be recorded on the aggregate table of partner.   
    if(sender_fl .and. size_of_aggregates > 0) then
       
       CALL determine_owner_index_frm_gl(GIN_aggregate(partner)+1, GIN_aggregate, &
            & PE_list, index, PE_list_size, NPROCS, my_rank)
       
       do i = 1, size_of_aggregates
          !C this assume all aggregates are not overlapped, 
          !C and all aggregates are recorded on aggregate table.
          call  assign_initial_aggre_for_sender(aggregate_number_in_table,        &
               & aggregate_table_array, aggregate_table_size,                     &
               & i+GIN_aggregate(partner)+partner_aggre_size, i, index)
          
          !C aggregates info is already recorded on Temp(:, aggregate_number),
          !C and node_index(:).
       end do
    end if

  end subroutine cordinate_GIN_aggregate
  

  !C if gl is own aggregate, then index is 0.
  subroutine determine_owner_index_frm_gl(gl, GIN_aggregate, PE_list, index, PE_list_size, NPROCS, my_rank)
    implicit none
    include 'precision.inc'

    integer(kind=kint), intent(in) :: gl, NPROCS, my_rank
    integer(kind=kint), intent(inout) :: GIN_aggregate(0:)
    integer(kind=kint), intent(inout) :: PE_list(:,:), PE_list_size
    integer(kind=kint), intent(out) :: index
    
    integer(kind=kint) :: r

    do r = 1, NPROCS
       if(GIN_aggregate(r) >= gl) exit
    end do
    
    r = r-1

    if(r==my_rank) then
       index = 0
    else
       do index = 1, PE_list_size
          if(PE_list(index,1) == r) exit
       end do
       
       if(index > PE_list_size) then
          PE_list_size = PE_list_size+1
          PE_list(PE_list_size,1) = r
          PE_list(PE_list_size,2) = 0
          PE_list(PE_list_size,3) = 0
          index = PE_list_size
       end if
    end if
    
  end subroutine determine_owner_index_frm_gl

  
  !C-- insert node into aggregate table
  !C-- global_nd -> aggregate_table(:,neib) and local aggregate number is in local_nd

  subroutine assign_initial_aggre_for_sender(aggregate_number_in_table,           &
       & aggregate_table_array, aggregate_table_size, global_nd, local_nd, neib)
    USE int_vec_module, only: int_vec, int_vec_initialize, check_and_extend
    implicit NONE
    include 'precision.inc'
    INTEGER(kind=kint), intent(in)    :: global_nd, aggregate_table_size, neib
    INTEGER(kind=kint), intent(in) :: local_nd
    INTEGER(kind=kint), intent(inout) :: aggregate_number_in_table(:)
    type(int_vec), pointer :: aggregate_table_array(:)
    
    INTEGER(kind=kint) :: l

    l = aggregate_number_in_table(neib)
    l = l+1

    call check_and_extend(aggregate_table_array(neib), 2*l)
    aggregate_table_array(neib) % int_array(2*l - 1) = global_nd
    aggregate_table_array(neib) % int_array(2*l)     = local_nd
    aggregate_number_in_table(neib) = l
    
  end subroutine assign_initial_aggre_for_sender


  SUBROUTINE assign_aggregate_table(aggregate_number_in_table,                      &
       & aggregate_table_array, aggregate_table_size, global_nd, local_nd, neib,    &
       & NPROCS, size_of_aggregates)
    USE int_vec_module, only: int_vec, int_vec_initialize, check_and_extend
    implicit NONE
    include 'precision.inc'
    INTEGER(kind=kint), intent(in)    :: global_nd, NPROCS, aggregate_table_size, neib
    INTEGER(kind=kint), intent(inout) :: local_nd

    INTEGER(kind=kint), intent(inout) :: size_of_aggregates
    INTEGER(kind=kint), intent(inout) :: aggregate_number_in_table(:)
    type(int_vec), pointer :: aggregate_table_array(:)
    
    INTEGER(kind=kint) :: m, l

    m = aggregate_number_in_table(neib)
    DO l = 1, m
       if(aggregate_table_array(neib) % int_array(2*l - 1) == global_nd) exit
    END DO
    if (l > m) then
       size_of_aggregates = size_of_aggregates + 1

       local_nd = size_of_aggregates

       call check_and_extend(aggregate_table_array(neib), 2 * l)
       aggregate_table_array(neib) % int_array(2 * l - 1) = global_nd
       aggregate_table_array(neib) % int_array(2 * l)     = local_nd
       aggregate_number_in_table(neib) = l

    else

       local_nd = aggregate_table_array(neib) % int_array(2 * l)
       
    end if
  END SUBROUTINE assign_aggregate_table
  
  SUBROUTINE make_PE_lists(LEVEL_NO, sorted_NEIBPE, PE_num_size, PE_nums, &
       &                   PE_list_size, PE_list, NPROCS, NEIBPE, NEIBPETOT)
    IMPLICIT NONE
    INCLUDE  'precision.inc'
    INTEGER(kind=kint), INTENT(in)    :: LEVEL_NO, NPROCS
    INTEGER(kind=kint), INTENT(in)    :: sorted_NEIBPE(:)
    INTEGER(kind=kint), INTENT(in)    :: PE_num_size
    INTEGER(kind=kint ), POINTER :: PE_nums(:)
    INTEGER(kind=kint ), INTENT(out) :: PE_list_size
!!$    INTEGER(kind=kint ), INTENT(out) :: PE_list(NPROCS, 3)
    INTEGER(kind=kint ), INTENT(out) :: PE_list(:, :)
    
    INTEGER(kind=kint )          :: NEIBPETOT
    INTEGER(kind=kint ), POINTER :: NEIBPE(:)
    
    INTEGER(kind=kint) :: i, j, k, l, pe, node
    INTEGER(kind=kint) :: index, Neib_index, Pe_num_index
    
    !C PE_list(*,1): PE number
    !C PE_list(*,2): NEIBPE number|
    !C PE_list(*,3): PE_num number|these two has 0 if no correspondings.
    !C-- making PE_list..
    index = 1
    Neib_index = 1
    Pe_num_index = 1

    DO WHILE((Neib_index <= NEIBPETOT) .AND. (Pe_num_index <= PE_num_size))
       i = sorted_NEIBPE(Neib_index)
       IF(PE_nums(Pe_num_index) == NEIBPE(i)) THEN
          PE_list(index, 1) = PE_nums(Pe_num_index)
          PE_list(index, 2) = i
          PE_list(index, 3) = Pe_num_index
          Neib_index = Neib_index + 1
          PE_num_index = Pe_num_index + 1
       ELSE IF(PE_nums(Pe_num_index) < NEIBPE(i)) THEN
          PE_list(index, 1) = PE_nums(Pe_num_index)
          PE_list(index, 2) = 0
          PE_list(index, 3) = Pe_num_index
          PE_num_index = PE_num_index + 1
       ELSE      
          PE_list(index, 1) = NEIBPE(i)
          PE_list(index, 2) = i
          PE_list(index, 3) = 0
          Neib_index = Neib_index + 1
       END IF
       index = index + 1
    END DO
    
    IF(Neib_index > NEIBPETOT) THEN
       DO i = PE_num_index, PE_num_size
          PE_list(index, 1) = PE_nums(i)
          PE_list(index, 2) = 0
          PE_list(index, 3) = i
          index = index + 1
       END DO
    ELSE
       DO i = Neib_index, NEIBPETOT
          j = sorted_NEIBPE(i)
          PE_list(index, 1) = NEIBPE(j)
          PE_list(index, 2) = j
          PE_list(index, 3) = 0
          index = index + 1
       END DO
    END IF
    PE_list_size = index - 1
    !C-- end: making PE_list..
  END SUBROUTINE make_PE_lists


END MODULE aggregate_mod
