!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

module amg_parameter
  include 'precision.inc'
  integer(kind=kint) :: iteration_number_up=1, iteration_number_down=1, iteration_number_bottom=30
  integer(kind=kint) :: f_first_iteration_number=0
  integer(kind=kint) :: cycle_kind=0, number_of_vectors = 5, gs_kind=0, process_aggre_threshold=1000
  integer(kind=kint) :: max_level_size=10, min_node_size=100, comm_opt_for_CGA=0

  real(kind=kreal)   :: theta=0.05, dump_jacobi_weight=0.6666, dump_gauss_seidel_weight=1.0
  real(kind=kreal)   :: theta_reduction_rate=0.5
  character(len=20)  :: solver_string="CG", conv_check_str
  integer(kind=kint) :: aggregation_strategy=1
  !C 1:coupled_w_lu,2:coupled_w_o_lu,3:independent_w_lu
  !C 4:independent_w_o_lu,5:old_independent_w_o_lu
  logical :: INDPDT_AGGRE=.false., PROCESS_AGGRE=.false.
  logical :: NOT_USE_AGGRE_TRANSFER=.false., W_OUT_LU=.false.


  character(len=20)  :: AT_string
  integer(kind=kint) :: AT_interval=100, AT_diff=5, AT_setup=0, AT_init=-1
  logical :: conv_check_fl = .false., vector_problem_fl = .false.
  logical :: scaling_problem_fl = .false.
  integer(kind=kint) :: NB = 1

  INTEGER(kind=kint), parameter :: max_topo_sz=20  
  INTEGER(kind=kint), save :: topo_sz
  INTEGER(kind=kint), save :: topo_X(max_topo_sz),topo_Y(max_topo_sz)
  INTEGER(kind=kint), save :: topo_Z(max_topo_sz),topo_C(max_topo_sz)

  real(kind=kreal), save :: theta_ary(9)=-1
  real(kind=kreal), save :: dump_jacobi_weights(9)=-1
  real(kind=kreal), save :: dump_gauss_seidel_weights(10)=-1
  integer(kind=kint), save:: iter_ups(10)=-1
  integer(kind=kint), save:: iter_downs(10)=-1
  

contains 
  
  subroutine setup_parameter_set_at_each_lev(level, theta_copy)
    integer(kind=kint) :: level
    
    !C theta_copy is used for threshold value for strong connection
    real(kind=kreal) :: theta_copy
    
    if(theta_ary(level-1) >= 0) then
       theta_copy = theta_ary(level-1)
    end if
    
    if(dump_jacobi_weights(level-1) >= 0) then
       dump_jacobi_weight = dump_jacobi_weights(level-1)
    end if
  end subroutine setup_parameter_set_at_each_lev

  subroutine iterative_parameter_set_at_each_lev(level)
    integer(kind=kint) :: level
    if(iter_ups(level)>=0) then
       iteration_number_up   = iter_ups(level)
    end if
    
    if(iter_downs(level)>=0) then
       iteration_number_down = iter_downs(level)
    end if

    if(dump_gauss_seidel_weights(level)>=0) then
       dump_gauss_seidel_weight = dump_gauss_seidel_weights(level)
    end if

  end subroutine iterative_parameter_set_at_each_lev
  
!!$  subroutine amg_parameter_set(my_rank)
!!$    use problem_matrix, only: dst_string
!!$    implicit none
!!$    include 'mpif.h'
!!$    integer(kind=kint) :: Nunit, stat, my_rank
!!$    !C dummy
!!$    character(len=20) :: mstring
!!$
!!$    Nunit = 10
!!$    
!!$    open(Nunit,FILE='AMG.dat',FORM='FORMATTED',STATUS='old', iostat=stat)
!!$    
!!$    if(stat /= 0) then
!!$       if(my_rank == 0) write(*,*)  &
!!$            &"No AMG parameter file: AMG.dat. Default setting is used."
!!$       goto 100
!!$    end if
!!$    
!!$    read(Nunit, *, iostat=stat) cycle_kind
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *, iostat=stat) gs_kind
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *) iteration_number_up
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *) iteration_number_down
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *) iteration_number_bottom
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *) f_first_iteration_number
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *) theta
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *) dump_jacobi_weight
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *) dump_gauss_seidel_weight
!!$    if(stat /=0) goto 110
!!$    read(Nunit, *, iostat=stat) mstring
!!$    if(stat /= 0) goto 110
!!$    read(Nunit, *, iostat=stat) solver_string
!!$    if(stat /= 0) goto 110
!!$    read(Nunit, *, iostat=stat) number_of_vectors
!!$    read(Nunit, *, iostat=stat) dst_string
!!$    !C Here, dst_string is FIX, PARMETIS, or empty.
!!$    !C in the case of empty, FIXED ORDERING is chosen.
!!$    close(Nunit)
!!$
!!$100 return
!!$
!!$110 write(*,*) "AMG parameter file is wrong."
!!$    stop
!!$    
!!$  end subroutine amg_parameter_set
  
end module amg_parameter
