!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University. 
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

module at_mod
  use solver_parameter, only: parameter_type
  include 'precision.inc'

  !C this module is used by parameter_at
  !C public: at_parameter_setting(), offline_tuned_solver(), solver_choice(), update_time(), check()
  !C public: keisoku_fl, setupflag_for_AT, setup_param_changed_fl

  integer(kind=kint), parameter :: max_conv_criterion_size = 10, max_record_size = 1000
  integer(kind=kint), parameter :: time_hist_sz = 3
  
  type parameter_array_type
     type(parameter_type) :: param
     real(kind=kreal)   :: one_iter_time(0:time_hist_sz)
     real(kind=kreal)   :: one_iter_norm_reduction(0:time_hist_sz)
     real(kind=kreal)   :: setup_time(0:time_hist_sz)
     integer(kind=kint) :: conv_count
  end type parameter_array_type
  
  real(kind=kreal)   :: conv_criterion(max_conv_criterion_size)
  integer(kind=kint) :: level_each_criterion(max_conv_criterion_size)
  integer(kind=kint) :: conv_criterion_size = 0
  
  type(parameter_array_type), save :: parameters(max_record_size, max_conv_criterion_size)
  integer(kind=kint) :: record_size = 0 
  integer(kind=kint) :: record_index = 0, record_index2 = 0, conv_index = 0
  
  integer(kind=kint) :: first_level_size = 0, second_level_size = 0, zero_level_size = 0 
  integer(kind=kint) :: level_3_switch_counter(max_conv_criterion_size), counters(max_conv_criterion_size)
  integer(kind=kint) :: level_2_switch_counter(max_conv_criterion_size), level_end_switch_counter(max_conv_criterion_size)
  real(kind=kreal) :: level_switch_1 = 2., level_switch_0 = 1.0, level_switch_1_again = 1.0
!C  real(kind=kreal) :: level_switch_2 = 1.0, level_switch_3 = 8.0
  real(kind=kreal) :: level_switch_2 = 1.0, level_switch_3 = 2.0
  real(kind=kreal) :: level1_p=1.0, level2_p = 1.0, level3_p = 1.0, level4_p=1.0, level0_p = 1.0
  real(kind=kreal) :: differenciate_rate_l4 = 0.05 

  integer(kind=kint) :: check_interval = 100, difference_of_iter = 3, default_index, default_setup_index
  

  logical :: keisoku_fl = .true., setupflag_for_AT=.false., setup_param_changed_fl=.false.  
  integer :: t1, t2, t_max, t_rate

  real(kind=kreal) :: initial_resid
  
contains

  subroutine offline_tuned_solver(X, B, tv, iter, solver_setup_fl)
    use solver, only: switch_solver, conv_state
    use common_mod, only: setup_phase
    use problem_matrix, only: LEVEL_NUM, my_rank, SOLVER_COMM
    use omp_lib
    implicit none

    include 'mpif.h'
    
    logical  :: reach_conv, solver_setup_fl
    integer(kind=kint) :: t0, t1, t2, t_max, t_rate, maxiter, iter
    integer(kind=kint) :: dummy=100
    real(kind=kreal) :: rt1, rt2, rt0, setup_time, iter_time,tv
    real(kind=kreal) :: time
    real(kind=kreal) :: X(:), B(:)
    type(parameter_type) :: param
    
    level_switch_1 = 1.5; level_switch_0 = 1.0
    level1_p = 1.0; level2_p = 1.0; level3_p = 1.0; level4_p = 1.0; level0_p = 1.0
    
    call at_parameter_init()

    level_each_criterion(:) = 0

    !C assumption is made..for offline
    tv=1.0
    
    maxiter = iter
    do 
       X = 0
       !C tv is determined through at_preprocess().
       if(my_rank == 0) then
          call solver_choice(tv, setupflag_for_AT, dummy, param)
       end if
       
       call set_parameter_on_all_PEs(param)
       
       iter = maxiter
       
       !C the following two lines are for measuring correct time
       if(LEVEL_NUM == 0 .or. setup_param_changed_fl) then
          setup_time = -MPI_WTIME()

          call setup_phase()
          
          setup_time = setup_time + MPI_WTIME()

          setup_param_changed_fl = .false.
       end if

       iter_time = -MPI_WTIME()

       !C problem matrix is fixed when offline tuning
       call switch_solver(X, B, iter, .false.)

       call conv_state(reach_conv)

       !C time shows time for iterative part
       iter_time = iter_time + MPI_WTIME()

       !C problem is fixed, so setup part's tuning is set off.
       if(my_rank == 0) then
#ifdef PRINT_REZ
          write(*,*) iter_time, setup_time, iter_time+setup_time, reach_conv
#endif
          call update_time(setup_time, iter_time, iter, tv, reach_conv, solver_setup_fl)
       end if

       if(.not.keisoku_fl) exit
    end do

  end subroutine offline_tuned_solver

  subroutine set_parameter(parameter_set)
    type(parameter_type) :: parameter_set
    call set_setup_parameter(parameter_set)
    call set_iter_parameter(parameter_set)
  end subroutine set_parameter

  subroutine set_parameter_on_all_PEs(param)
    use problem_matrix, only: my_rank, SOLVER_COMM
    implicit none
    include 'mpif.h'

    type(parameter_type) :: param
    real(kind=kreal)  :: i
    character(len=20) :: solver_string
    real(kind=kreal), allocatable :: comm_data(:)
    integer(kind=kint) :: ierr

    allocate(comm_data(13))    

    if(my_rank == 0) then
       comm_data(1) = dble(param % cycle_kind)+0.1
       comm_data(2) = dble(param % number_of_vectors)+0.1
       comm_data(3) = dble(param % iteration_number_up)+0.1
       comm_data(4) = dble(param % iteration_number_down)+0.1
       comm_data(5) = dble(param % iteration_number_bottom)+0.1
       comm_data(6) = dble(param % f_first_iteration_number)+0.1
       comm_data(7) = dble(param % method)+0.1
       comm_data(8) = dble(param % gs_kind)+0.1
       comm_data(9) = param % theta
       comm_data(10) = param % dump_jacobi_weight
       comm_data(11) = param % dump_gauss_seidel_weight

       solver_string = param % solver_string

       if(solver_string(1:1)=='N' .and. solver_string(2:2)=='O') then
          comm_data(12) = 1.1

       else if(solver_string(1:1)=='C' .and. solver_string(2:2)=='G') then
          comm_data(12) = 2.1

       else if(solver_string(1:1)=='B' .and. solver_string(2:2)=='I' .and. &
            &  solver_string(3:3)=='C' .and. solver_string(4:4)=='G' .and. &
            &  solver_string(5:5)=='S' .and. solver_string(6:6)=='T') then
          comm_data(12) = 3.1

       else if(solver_string(1:1)=='R' .and. solver_string(2:2)=='E' .and. &
            &  solver_string(3:3)=='S' .and. solver_string(4:4)=='C' .and. &
            &  solver_string(5:5)=='U' .and. solver_string(6:6)=='T') then
          comm_data(12) = 4.1

       else if(solver_string(1:1)=='G' .and. solver_string(2:2)=='P' .and. &
            &  solver_string(3:3)=='B' .and. solver_string(4:4)=='I' .and. &
            &  solver_string(5:5)=='C' .and. solver_string(6:6)=='G') then
          comm_data(12) = 5.1

       else
          if(my_rank == 0) write(*,*) "No solver is specified by ", solver_string 
          stop
       end if

       if(keisoku_fl) then
          comm_data(13) = 1.1
       else 
          comm_data(13) = 0.1
       end if

       !C comm_data(14) is not needed.
!!$       if(setup_param_changed_fl) then
!!$          comm_data(14) = 1.1
!!$       else 
!!$          comm_data(14) = 0.1
!!$       end if
    end if


    call MPI_BCAST(comm_data, 13, MPI_DOUBLE_PRECISION, 0, SOLVER_COMM, ierr)
    
    call read_comm_data(comm_data)


    deallocate(comm_data)

  contains
    subroutine read_comm_data(comm_data)
      logical :: setupflag_for_AT
      real(kind=kreal) :: comm_data(:)
      integer(kind=kint) :: i
      type(parameter_type) :: p

      p % cycle_kind = int(comm_data(1))
      p % number_of_vectors = int(comm_data(2))
      p % iteration_number_up = int(comm_data(3))
      p % iteration_number_down = int(comm_data(4))
      p % iteration_number_bottom = int(comm_data(5)) 
      p % f_first_iteration_number = int(comm_data(6))
      p % method = int(comm_data(7))
      p % gs_kind = int(comm_data(8))
      p % theta = comm_data(9)
      p % dump_jacobi_weight = comm_data(10)
      p % dump_gauss_seidel_weight = comm_data(11)

      i = int(comm_data(12))

      if(i == 1) then
         p % solver_string = "NO"
      else if(i == 2) then
         p % solver_string = "CG"
      else if(i == 3) then
         p % solver_string = "BICGSTAB"
      else if(i == 4) then
         p % solver_string = "RESCUT" 
      else if(i == 5) then
         p % solver_string = "GPBICG" 
      else 
         write(*,*) "comm_data is collapsed.."
         stop
      end if

      call set_parameter(p)
!!$    call set_iter_parameter(p)
!!$    if(setupflag_for_AT) call set_setup_parameter(p)

      keisoku_fl = (int(comm_data(13)) == 1)
!!$      setup_param_changed_fl = (int(comm_data(14)) == 1)

    end subroutine read_comm_data

  end subroutine set_parameter_on_all_PEs


  subroutine set_setup_parameter(setup_parameter_set)
    use amg_parameter
    implicit none
    real(kind=kreal), save :: theta_bak = -1, dump_jacobi_weight_bak = -1.
    type(parameter_type) :: setup_parameter_set

    !C--
    if(theta_bak == -1) then
       theta_bak = theta
       dump_jacobi_weight_bak = dump_jacobi_weight
    end if

    !C amgs_p has only SA AMG solver.
    !C method = setup_parameter_set % method
    theta = setup_parameter_set % theta
    dump_jacobi_weight = setup_parameter_set % dump_jacobi_weight

    !C set setup_param_changed_fl.
    if(dabs(theta - theta_bak) > 0.0001 .or. dabs(dump_jacobi_weight - dump_jacobi_weight_bak)&
         &  > 0.0001) setup_param_changed_fl = .true.
    
    dump_jacobi_weight_bak = dump_jacobi_weight
    theta_bak = theta
    
  end subroutine set_setup_parameter
  
  subroutine set_iter_parameter(parameter_set)
    use amg_parameter
    implicit none
    
    type(parameter_type) :: parameter_set
    
    cycle_kind               = parameter_set % cycle_kind
    number_of_vectors        = parameter_set % number_of_vectors
    iteration_number_up      = parameter_set % iteration_number_up
    iteration_number_down    = parameter_set % iteration_number_down
    iteration_number_bottom  = parameter_set % iteration_number_bottom
    f_first_iteration_number = parameter_set % f_first_iteration_number
    dump_gauss_seidel_weight = parameter_set % dump_gauss_seidel_weight
    solver_string            = parameter_set % solver_string
    gs_kind                  = parameter_set % gs_kind
    
  end subroutine set_iter_parameter

  subroutine at_parameter_init()
    implicit none
    integer(kind=kint) :: i, j

    do i = 1, record_size
       do j = 1, conv_criterion_size
          parameters(i, j) % one_iter_time(0:) = 0.
          parameters(i, j) % one_iter_norm_reduction(0:) = 0.
          parameters(i, j) % setup_time(0:) = 0.
          
          parameters(i, j) % conv_count = 0
       end do
    end do
    counters = 0
  end subroutine at_parameter_init
  
  !C if inter_val is <=0, offline_fl is set.
  subroutine at_parameter_setting(inter_val, diff_of_iter, init, setup_fl)
!!$  subroutine at_parameter_setting(inter_val, diff_of_iter)
    implicit none
    
    integer(kind=kint) :: i,j,k,l, inter_val, diff_of_iter, init
    logical :: setup_fl 
!!$    logical :: setup_fl = .true.
    
    setupflag_for_AT = setup_fl

    !C set difference of iteration numbers
    difference_of_iter = diff_of_iter
    
    !C set check_interval
    check_interval = inter_val

    !C prepare conv_criterion
    conv_criterion_size = 0 
    conv_criterion_size = conv_criterion_size+1
!!$    conv_criterion(conv_criterion_size) = 0.0
    conv_criterion(conv_criterion_size) = 1.0E-1
    conv_criterion_size = conv_criterion_size+1
    conv_criterion(conv_criterion_size) = 1.0E-3
    conv_criterion_size = conv_criterion_size+1
    conv_criterion(conv_criterion_size) = 1.0E-5
    conv_criterion_size = conv_criterion_size+1
    conv_criterion(conv_criterion_size) = 1.0E-7
    conv_criterion_size = conv_criterion_size+1
    conv_criterion(conv_criterion_size) = 1.0E-9

    !C prepare parameters
    !C NO,CG,BICGSTAB, IDR(?), GMRES(?), RESCUT(?)
    record_size = 0
    
    zero_level_size = 0
    !C0708    if(setup_fl) then
    !C AMGCG with  V-cycle and 1 sgs
    record_size = record_size + 1
    parameters(record_size,1) % param % method = 22
    parameters(record_size,1) % param % theta  = 0.05
    parameters(record_size,1) % param % dump_jacobi_weight  = 0.666
    parameters(record_size,1) % param % cycle_kind = 0
!!$    parameters(record_size,1) % number_of_vectors = 10
    parameters(record_size,1) % param % number_of_vectors = 4
    parameters(record_size,1) % param % iteration_number_up = 1
    parameters(record_size,1) % param % iteration_number_down = 1
    parameters(record_size,1) % param % iteration_number_bottom = 30
    parameters(record_size,1) % param % f_first_iteration_number = 0
    parameters(record_size,1) % param % dump_gauss_seidel_weight = 1.0
    parameters(record_size,1) % param % gs_kind = 0
!!$    parameters(record_size,1) % solver_string = "CG"
    parameters(record_size,1) % param % solver_string = "BICGSTAB"
    default_setup_index = record_size

    
    !C-- theta optimization
    record_size = record_size + 1
    parameters(record_size,1) = parameters(default_setup_index,1)
    parameters(record_size,1) % param % theta  = 0.07

    record_size = record_size + 1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % theta  = 0.09

    record_size = record_size + 1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % theta  = 0.11

    record_size = record_size + 1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % theta  = 0.04

    record_size = record_size + 1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % theta  = 0.03
    !C--

    zero_level_size = record_size
    !C0708    end if

    !C-- solver optimization
    record_size = record_size+1
    parameters(record_size,1) = parameters(default_setup_index,1)    
    
#ifndef ATCG
    parameters(record_size,1) % param % solver_string = "RESCUT"
    parameters(record_size,1) % param % number_of_vectors = 4
    parameters(record_size,1) % param % gs_kind = 0

    record_size = record_size+1
    parameters(record_size,1) = parameters(record_size -1,1)
    parameters(record_size,1) % param % gs_kind = 1

    record_size = record_size+1
    parameters(record_size,1) = parameters(record_size -1,1)
    parameters(record_size,1) % param % number_of_vectors = 2
    parameters(record_size,1) % param % gs_kind = 0

    record_size = record_size+1
    parameters(record_size,1) = parameters(record_size -1,1)
    parameters(record_size,1) % param % gs_kind = 1

    record_size=record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
#endif
    parameters(record_size,1) % param % solver_string = "CG"
    parameters(record_size,1) % param % number_of_vectors = 10
    parameters(record_size,1) % param % gs_kind = 0
    default_index = record_size

    record_size=record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % gs_kind = 1
    
#ifndef ATCG
    record_size=record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % solver_string = "BICGSTAB"
    parameters(record_size,1) % param % gs_kind = 0
    default_index = record_size

    record_size=record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % gs_kind = 1

#endif

!!$    record_size=record_size+1
!!$    parameters(record_size) % number_of_vectors = 2
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % solver_string = "IDR"
!!$
!!$    record_size=record_size+1
!!$    parameters(record_size) % number_of_vectors = 3
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % solver_string = "IDR"

#ifndef ATCG
    record_size=record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % solver_string = "NO"
    parameters(record_size,1) % param % gs_kind = 0

    record_size=record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % gs_kind = 1

#endif
    !C--

    first_level_size = record_size

    !C-- acceleration parametar optimization
    record_size = record_size+1
    parameters(record_size,1) = parameters(default_index,1)
    parameters(record_size,1) % param % dump_gauss_seidel_weight = 1.0

    record_size = record_size+1
    parameters(record_size,1) = parameters(default_index,1)
    parameters(record_size,1) % param % dump_gauss_seidel_weight = 0.8
    
    record_size = record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % dump_gauss_seidel_weight = 1.2

    record_size = record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % dump_gauss_seidel_weight = 1.4

    record_size = record_size+1
    parameters(record_size,1) = parameters(record_size-1,1)
    parameters(record_size,1) % param % dump_gauss_seidel_weight = 1.6

!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % param % dump_gauss_seidel_weight = 1.8

!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % param % dump_gauss_seidel_weight = 2.0
    !C--
    
    second_level_size = record_size

    !C-- cycle optimization
    record_size = record_size+1
    parameters(record_size,1) = parameters(default_index,1)
    parameters(record_size,1) % param % cycle_kind = 0

    record_size = record_size+1
    parameters(record_size,1) = parameters(default_index,1)
    parameters(record_size,1) % param % cycle_kind = 1

    record_size = record_size+1    
    parameters(record_size,1) = parameters(default_index,1)
    parameters(record_size,1) % param % cycle_kind = 2
    !C--


!C    level_each_criterion = 1
!C    if(setup_fl) level_each_criterion = 0
    level_each_criterion = 0 

    !C copy parameters(:,1)->parameters(:,2:conv_criterion_size)
    do i = 2, conv_criterion_size
       parameters(:,i) = parameters(:,1)
    end do
    
    !C0803
    !C level_2_switch_counter is set for setup_fl == .false.
    !C if setup_fl == .true. then this val is set at level 0.
    !C call set_level_switch_counter(int(level_switch_1 * (first_level_size - &
    !C & zero_level_size))+1, 2) 
    
    call at_parameter_init()
    
  end subroutine at_parameter_setting

  subroutine at_parameter_setting_from_file(inter_val, diff_of_iter, init, setup_fl)
    implicit none
    integer(kind=kint) :: i,j,k,l, inter_val, diff_of_iter, init
    logical :: setup_fl 
    
    !C-- valiables for reading AT.dat
    character :: string*20, solver_string*40, default_solver*10
    real(kind=kreal) :: theta_arry(20), omega_arry(20), cycles_arry(3)
    integer(kind=kint) :: theta_s, omega_s, smoother_iter_s, num_of_hist_vecs_s
    integer(kind=kint) :: solver_s, cycles_s, gs_kind_s
    integer(kind=kint) :: smoother_iteration_arry(20), num_hist_vecs(20), gs_kind_arry(10)
    integer(kind=kint) :: def_sol_len

!!$    logical :: setup_fl = .true.

    setupflag_for_AT = setup_fl

    !C set difference of iteration numbers
    difference_of_iter = diff_of_iter

    !C set check_interval
    check_interval = inter_val

    !C prepare conv_criterion
    conv_criterion_size = 0 
    conv_criterion_size = conv_criterion_size+1
!!$    conv_criterion(conv_criterion_size) = 0.0
    conv_criterion(conv_criterion_size) = 1.0E-1
    conv_criterion_size = conv_criterion_size+1
    conv_criterion(conv_criterion_size) = 1.0E-3
    conv_criterion_size = conv_criterion_size+1
    conv_criterion(conv_criterion_size) = 1.0E-5
    conv_criterion_size = conv_criterion_size+1
    conv_criterion(conv_criterion_size) = 1.0E-7
    conv_criterion_size = conv_criterion_size+1
    conv_criterion(conv_criterion_size) = 1.0E-9

    call AT_file_read(11)
    
    call set_default_solver(default_solver)



    !C prepare parameters
    !C NO,CG,BICGSTAB, IDR(?), GMRES(?), RESCUT(?)
    record_size = 0

    zero_level_size = 0
    !C0708    if(setup_fl) then
    !C AMGCG with  V-cycle and 1 sgs
    record_size = record_size + 1
    parameters(record_size,1) % param % method = 22
    parameters(record_size,1) % param % theta  = theta_arry(1)
    parameters(record_size,1) % param % dump_jacobi_weight  = 0.666
    parameters(record_size,1) % param % cycle_kind = cycles_arry(1)
!!$    parameters(record_size,1) % number_of_vectors = 10
    parameters(record_size,1) % param % number_of_vectors = 4
    parameters(record_size,1) % param % iteration_number_up = smoother_iteration_arry(1)
    parameters(record_size,1) % param % iteration_number_down = smoother_iteration_arry(1)
    parameters(record_size,1) % param % iteration_number_bottom = 30
    parameters(record_size,1) % param % f_first_iteration_number = 1
    parameters(record_size,1) % param % dump_gauss_seidel_weight = omega_arry(1)
    parameters(record_size,1) % param % gs_kind = gs_kind_arry(1)
!!$    parameters(record_size,1) % solver_string = "CG"
    parameters(record_size,1) % param % solver_string = default_solver
    default_setup_index = record_size


    !C-- theta optimization
    do i = 2, theta_s
       record_size = record_size + 1       
       parameters(record_size,1) = parameters(default_setup_index,1)       
       parameters(record_size,1) % param % theta  = theta_arry(i)
    end do
    !C--

    zero_level_size = record_size

    def_sol_len = len_trim(default_solver)

    write(*,*) default_solver, def_sol_len, INDEX(default_solver, 'CG'), "default!!"
    !C-- solver optimization

    !C RESCUT--
    j = INDEX(solver_string, 'RESCUT')
    if(j > 0) then
       do i = 1, num_of_hist_vecs_s
          do l = 1, gs_kind_s
             do k = 1, smoother_iter_s
                record_size = record_size+1
                
                if(i == 1 .and. l == 1 .and. k == 1 .and. def_sol_len == 5 .and. &
                     & INDEX(default_solver, 'RESCUT') == 1) default_index = record_size
                
                parameters(record_size,1) = parameters(default_setup_index,1)    
                parameters(record_size,1) % param % solver_string = "RESCUT"
                parameters(record_size,1) % param % number_of_vectors = num_hist_vecs(i)
                parameters(record_size,1) % param % gs_kind = gs_kind_arry(l)
                parameters(record_size,1) % param % iteration_number_up = smoother_iteration_arry(k)
                parameters(record_size,1) % param % iteration_number_down = smoother_iteration_arry(k)
             end do
          end do
       end do
    end if

    !C BICGSTAB--
    j = INDEX(solver_string, 'BICGSTAB')
    if(j > 0) then
       do l = 1, gs_kind_s
          do k = 1, smoother_iter_s
             record_size = record_size+1

             if(l == 1 .and. k == 1 .and. def_sol_len == 8 .and. &
                  & INDEX(default_solver, 'BICGSTAB') == 1) default_index = record_size


             parameters(record_size,1) = parameters(default_setup_index,1)    
             parameters(record_size,1) % param % solver_string = "BICGSTAB"
             parameters(record_size,1) % param % gs_kind = gs_kind_arry(l)
             parameters(record_size,1) % param % iteration_number_up = smoother_iteration_arry(k)
             parameters(record_size,1) % param % iteration_number_down = smoother_iteration_arry(k)
          end do
       end do
    end if

    !C GPBICG--
    j = INDEX(solver_string, 'GPBICG')
    if(j > 0) then
       do l = 1, gs_kind_s
          do k = 1, smoother_iter_s
             record_size = record_size+1

             if(l == 1 .and. k == 1 .and. def_sol_len == 6 .and. &
                  & INDEX(default_solver, 'GPBICG') == 1) default_index = record_size


             parameters(record_size,1) = parameters(default_setup_index,1)    
             parameters(record_size,1) % param % solver_string = "GPBICG"
             parameters(record_size,1) % param % gs_kind = gs_kind_arry(l)
             parameters(record_size,1) % param % iteration_number_up = smoother_iteration_arry(k)
             parameters(record_size,1) % param % iteration_number_down = smoother_iteration_arry(k)
          end do
       end do
    end if


    !C CG--
    j = INDEX(solver_string, 'CG')

    if(j > 1) then
       !C differentiation among GPBICG, BICGSTAB and CG
       if(solver_string(j-1:j-1)=='I') then
          k = INDEX(solver_string(j+1:), 'CG')
          j= j+k

          if(k>0) then
             if(solver_string(j-1:j-1)=='I') then
                k = INDEX(solver_string(j+1:), 'CG')
                j=j+k
             end if
          end if
       end if
    end if

    
    if(j>1) then
       if(solver_string(j-1:j-1)=='I') then
          j=0
       end if
    end if



    if(j > 0) then
       do l = 1, gs_kind_s
          do k = 1, smoother_iter_s
             record_size = record_size+1

             if(l == 1 .and. k == 1 .and. def_sol_len == 2 .and. &
                  & INDEX(default_solver, 'CG') == 1) default_index = record_size
             
             parameters(record_size,1) = parameters(default_setup_index,1)    
             parameters(record_size,1) % param % solver_string = "CG"
             parameters(record_size,1) % param % gs_kind = gs_kind_arry(l)
             parameters(record_size,1) % param % iteration_number_up = smoother_iteration_arry(k)
             parameters(record_size,1) % param % iteration_number_down = smoother_iteration_arry(k)
          end do
       end do
    end if


    !C NO--
    j = INDEX(solver_string, 'NO')
    if(j > 0) then
       do l = 1, gs_kind_s
          do k = 1, smoother_iter_s
             record_size = record_size+1

             if(l == 1 .and. k == 1 .and. def_sol_len == 2 .and. &
                  & INDEX(default_solver, 'NO') == 1) default_index = record_size

             parameters(record_size,1) = parameters(default_setup_index,1)    
             parameters(record_size,1) % param % solver_string = "NO"
             parameters(record_size,1) % param % gs_kind = gs_kind_arry(l)
             parameters(record_size,1) % param % iteration_number_up = smoother_iteration_arry(k)
             parameters(record_size,1) % param % iteration_number_down = smoother_iteration_arry(k)
          end do
       end do
    end if

       
    first_level_size = record_size

    !C-- acceleration parametar optimization
    do i = 1, omega_s
       record_size = record_size+1
       parameters(record_size,1) = parameters(default_setup_index,1)
       parameters(record_size,1) % param % dump_gauss_seidel_weight = omega_arry(i)
    end do
    !C--

    second_level_size = record_size

    !C-- cycle optimization
    do i = 1, cycles_s
       record_size = record_size+1
       parameters(record_size,1) = parameters(default_setup_index,1)
       parameters(record_size,1) % param % cycle_kind = cycles_arry(i)
    end do
    !C--

    if(setup_fl) then
       default_index = default_setup_index
    end if
    

    !C    level_each_criterion = 1
    !C    if(setup_fl) level_each_criterion = 0
    level_each_criterion = 0 

    !C copy parameters(:,1)->parameters(:,2:conv_criterion_size)
    do i = 2, conv_criterion_size
       parameters(:,i) = parameters(:,1)
    end do
    
    call at_parameter_init()
    
    

    
  contains
    subroutine set_default_solver(default_solver)
      character :: default_solver*10
      integer(kind=kint) :: i,j
      i = INDEX(solver_string, 'CG')
      if(i > 1) then
         if(solver_string(i-1:i-1)=='I') then
            i = INDEX(solver_string(i+1:), 'CG')
         end if
      end if
      if(i > 0) default_solver = 'CG'  


      j = INDEX(solver_string, 'RESCUT')
      if(j >0 .and. ((i > 0 .and. j < i) .or. i == 0 )) then
         i = j
         default_solver = 'RESCUT'
      end if

      j = INDEX(solver_string, 'BICGSTAB')
      if(j >0 .and. ((i > 0 .and. j < i) .or. i == 0 )) then
         i = j
         default_solver = 'BICGSTAB'
      end if

      j = INDEX(solver_string, 'GPBICG')
      if(j >0 .and. ((i > 0 .and. j < i) .or. i == 0 )) then
         i = j
         default_solver = 'GPBICG'
      end if

      j = INDEX(solver_string, 'NO')
      if(j >0 .and. ((i > 0 .and. j < i) .or. i == 0 )) then
         i = j
         default_solver = 'NO'
      end if
    end subroutine set_default_solver
    
    subroutine AT_file_read(fio)
      integer(kind=kint) :: fio
      open(fio, File='AT.dat', Form='FORMATTED')
      read(fio, '(A20)') string
      read(fio, *) theta_s, omega_s, solver_s, cycles_s, gs_kind_s, smoother_iter_s, num_of_hist_vecs_s

      read(fio, '(A20)') string    
      read(fio, '(A20)') string
      read(fio, *) (theta_arry(i), i=1, theta_s)

      read(fio, '(A20)') string    
      read(fio, '(A20)') string
      read(fio, *) (omega_arry(i), i=1, omega_s)

      read(fio, '(A20)') string    
      read(fio, '(A20)') string
      read(fio, '(A40)') solver_string

      read(fio, '(A20)') string    
      read(fio, '(A20)') string
      read(fio, *) (cycles_arry(i), i=1, cycles_s)

      read(fio, '(A20)') string    
      read(fio, '(A20)') string
      read(fio, *) (gs_kind_arry(i), i=1, gs_kind_s)

      read(fio, '(A20)') string    
      read(fio, '(A20)') string
      read(fio, *) (smoother_iteration_arry(i), i=1, smoother_iter_s)

      read(fio, '(A20)') string    
      read(fio, '(A20)') string
      read(fio, *) (num_hist_vecs(i), i=1, num_of_hist_vecs_s)

      close(fio)
    end subroutine AT_file_read

  end subroutine at_parameter_setting_from_file



  !C assume this sub is called on level 4
  subroutine check_changed(record_index, conv_index, lev, changed_fl)
    implicit none
    integer(kind=kint) :: record_index, conv_index, lev
    logical :: changed_fl
    
    if(lev == 0) then    
       changed_fl =  parameters(record_index, conv_index) % param % theta /= &
            &        parameters(second_level_size+1, conv_index) % param % theta

    else if(lev == 1) then
       changed_fl = parameters(record_index, conv_index) % param % solver_string /= &
            &       parameters(second_level_size+1, conv_index) % param % solver_string  .or. &
            &       parameters(record_index, conv_index) % param % gs_kind /=   &
            &       parameters(second_level_size+1, conv_index) % param % gs_kind  .or. &
            &       (parameters(record_index, conv_index) % param % solver_string ==   &
            &        parameters(second_level_size+1, conv_index) % param % solver_string  .and. &
            &        parameters(record_index, conv_index) % param % number_of_vectors /=   &
            &        parameters(second_level_size+1, conv_index) % param % number_of_vectors)
       
    else if(lev == 2) then
       changed_fl =  parameters(record_index,conv_index) % param % dump_gauss_seidel_weight /= &
            &        parameters(second_level_size+1, conv_index) % param % dump_gauss_seidel_weight

    else
       write(*,*) "diff_fl: undefined case.."
       
    end if
       
  end subroutine check_changed

  !C this subroutin is called when recorad_index para can't reach convergence.
  subroutine invalidate_aggressive_acceleration(conv_index, record_index)
    implicit none
    integer(kind=kint) :: i, conv_index, record_index
    if(level_each_criterion(conv_index) == 2 .and. record_index > first_level_size) then
       do i = record_index, record_size
          parameters(i, conv_index) % one_iter_time(:) = -1
!C          parameters(i, conv_index) % one_iter_norm_reduction(0:) = 0.
!C          parameters(i, conv_index) % setup_time(0:) = 0.

          parameters(i,conv_index) % conv_count = 0
       end do
    end if
  end subroutine invalidate_aggressive_acceleration



  subroutine reset_time_data(i, conv_index)
    implicit none
    integer(kind=kint) :: i, conv_index
    
    parameters(i, conv_index) % one_iter_time(:) = 0
    parameters(i, conv_index) % one_iter_norm_reduction(0:) = 0
    parameters(i, conv_index) % setup_time(0:) = 0

    parameters(i, conv_index) % conv_count = 0
  end subroutine reset_time_data



  !C level 2 method is changed to parameters(record_index) method.
  subroutine update_method(level, reset_fl, setup_fl, log_tv)
    implicit none
    integer(kind=kint) :: i, level, min_index=0, s, e, ex_iter
    logical, intent(in) :: reset_fl, setup_fl
    real(kind=kreal) :: t, min_t, log_tv

    if(.not.reset_fl) then
       if(level == 1) then 
          s = zero_level_size+1; e = first_level_size
       else if(level == 2) then
          s = first_level_size+1; e = second_level_size
       else if(level == 3) then
          s = second_level_size+1; e = record_size
       end if

       call average_time_for_para(log_tv, s, min_t, ex_iter, setup_fl)
       min_index = s
       do i = s+1, e
          call average_time_for_para(log_tv, i, t, ex_iter, setup_fl)
          if(min_t > t ) then
             min_t = t; min_index = i
          end if
       end do
    end if
    
    if(level == 1) then
       do i = zero_level_size + 1, first_level_size
          !C-- theta @ level0
          parameters(i,conv_index) % param % theta = parameters(record_index, conv_index) % param % theta

          if(reset_fl)  then
             call reset_time_data(i, conv_index)
!!$             parameters(i, conv_index) % one_iter_time(:) = 0
!!$             parameters(i, conv_index) % conv_count = 0
          end if
          
       end do
       
    else if(level == 2) then
       do i = first_level_size+1, second_level_size
          !C-- solver_string, numver_of_vectors, and gs_kind with theta
          !C @ level0
          parameters(i,conv_index) % param % theta = parameters(record_index, conv_index) % param % theta

          !C @ level1
          parameters(i,conv_index) % param % solver_string = parameters(record_index,conv_index) % param % solver_string
          parameters(i,conv_index) % param % number_of_vectors = parameters(record_index,conv_index) % param % number_of_vectors
          parameters(i,conv_index) % param % gs_kind = parameters(record_index,conv_index) % param % gs_kind

          
          if(reset_fl)  then
             call reset_time_data(i, conv_index)
!!$             parameters(i,conv_index) % one_iter_time(:) = 0
!!$             parameters(i,conv_index) % conv_count = 0
          end if
       end do
       
    else if(level == 3) then
       do i = second_level_size+1, record_size
          !C-- dump_gauss_seidel_weight with others.
          !C @ level0
          parameters(i,conv_index) % param % theta = parameters(record_index, conv_index) % param % theta


          !C @ level1
          parameters(i,conv_index) % param % solver_string = parameters(record_index,conv_index) % param % solver_string
          parameters(i,conv_index) % param % number_of_vectors = parameters(record_index,conv_index) % param % number_of_vectors
          parameters(i,conv_index) % param % gs_kind = parameters(record_index,conv_index) % param % gs_kind

          !C @ level2
          parameters(i,conv_index) % param % dump_gauss_seidel_weight = parameters(record_index, conv_index) % &
               & param % dump_gauss_seidel_weight
          
          if(reset_fl)then
             call reset_time_data(i, conv_index)
!!$             parameters(i,conv_index) % one_iter_time(:) = 0
!!$             parameters(i,conv_index) % conv_count = 0
          end if
       end do
    end if

    if(.not.reset_fl) then
       do i = s, e
          call average_time_for_para(log_tv, i, t, ex_iter, setup_fl)
          if(min_t * 2. > t) then
             call reset_time_data(i, conv_index)
!!$             parameters(i,conv_index) % one_iter_time(:) = 0
!!$             parameters(i,conv_index) % conv_count = 0
          end if
       end do
    end if
    
  end subroutine update_method

  subroutine set_level_switch_counter(counter, lev)
    integer(kind=kint) ::counter, lev
    if(lev == 2)then
       level_2_switch_counter(conv_index) = counter
    else if(lev == 3) then
       level_3_switch_counter(conv_index) = counter       
    else if(lev == 4) then
       level_end_switch_counter(conv_index) = counter       
    else
       write(*,*) "unknown lev : in set_level_switch_counter"
    end if
    
  end subroutine set_level_switch_counter

  !C sover time is considered using iteration number expected from log_tv
  subroutine average_time_for_para2(log_tv, p, t, expected_iter, solver_setup_fl)
    implicit none
    logical, intent(in) :: solver_setup_fl
    integer(kind=kint), intent(in) :: p
    real(kind=kreal), intent(in) :: log_tv
    real(kind=kreal), intent(inout) :: t
    integer(kind=kint), intent(inout) :: expected_iter

    real(kind=kreal) :: one_iter_time_av, one_iter_norm_reduction_av, setup_time_av
    
    if(parameters(p, conv_index) % conv_count > 0) then
       one_iter_time_av = parameters(p, conv_index) % one_iter_time(0)
       one_iter_norm_reduction_av = parameters(p, conv_index) % one_iter_norm_reduction(0)
       setup_time_av = parameters(p, conv_index) % setup_time(0)
       
       if(parameters(p, conv_index) % conv_count >= time_hist_sz) then
          one_iter_time_av = one_iter_time_av / time_hist_sz
          one_iter_norm_reduction_av = one_iter_norm_reduction_av / time_hist_sz
          setup_time_av = setup_time_av / time_hist_sz
       else
          one_iter_time_av = one_iter_time_av / parameters(p,conv_index) % conv_count
          one_iter_norm_reduction_av = one_iter_norm_reduction_av / parameters(p,conv_index) % conv_count
          setup_time_av = setup_time_av / parameters(p,conv_index) % conv_count 
       end if

       
       !C log_tv needs to be near the correct value.
       if(solver_setup_fl) then
          expected_iter = int(log_tv / one_iter_norm_reduction_av)+1
          t = setup_time_av + expected_iter * one_iter_time_av             
       else
          expected_iter = int(log_tv / one_iter_norm_reduction_av)+1
          t = expected_iter * one_iter_time_av             
       end if
       
    else 
       !C this case is added for differenciation of divergence.
       t = parameters(p,conv_index) % one_iter_time(0)
    end if

  end subroutine average_time_for_para2

  !C sover time is considered using time for eliminating unit residual norm
  subroutine average_time_for_para(log_tv, p, t, expected_iter, solver_setup_fl)
    implicit none
    logical, intent(in) :: solver_setup_fl
    integer(kind=kint), intent(in) :: p
    real(kind=kreal), intent(in) :: log_tv
    real(kind=kreal), intent(inout) :: t
    integer(kind=kint), intent(inout) :: expected_iter

    real(kind=kreal) :: one_iter_time_av, one_iter_norm_reduction_av, setup_time_av
    
    if(parameters(p, conv_index) % conv_count > 0) then
       one_iter_time_av = parameters(p, conv_index) % one_iter_time(0)
       setup_time_av = parameters(p, conv_index) % setup_time(0)
       
       if(parameters(p, conv_index) % conv_count >= time_hist_sz) then
          one_iter_time_av = one_iter_time_av / time_hist_sz
          setup_time_av = setup_time_av / time_hist_sz
       else
          one_iter_time_av = one_iter_time_av / parameters(p,conv_index) % conv_count
          setup_time_av = setup_time_av / parameters(p,conv_index) % conv_count 
       end if

       
       !C In this case, one_iter_norm_reduction can not be calculated correctly.
       if(solver_setup_fl) then
          t = setup_time_av + one_iter_time_av                          
       else
          t = one_iter_time_av                          
       end if
       
    else 
       !C this case is added for differenciation of divergence.
       t = parameters(p,conv_index) % one_iter_time(0)
    end if

  end subroutine average_time_for_para

  
  subroutine solver_choice(tv,  setup_fl, expected_call_times, param_out)
    implicit none
    type(parameter_type) :: param_out
    logical :: setup_fl, l1_changed_fl, l0_changed_fl, l2_changed_fl
    real(kind=kreal) :: tv, p, q, level_p
    integer(kind=kint) :: i, j, size, c, i_start, expected_call_times
    integer(kind=kint), parameter :: l4 = 15
    integer(kind=kint), save :: l4_counter(max_conv_criterion_size)=l4
    integer(kind=kint), save :: l4_parameter_hist(l4, max_conv_criterion_size) = 0

    integer(kind=kint), parameter :: parameter_hist_sz = 5
    integer(kind=kint), save :: parameter_hist(parameter_hist_sz, max_conv_criterion_size)=0
    
    real(kind=kreal), parameter :: diff_criterion = 0.05
    real(kind=kreal) :: min_time, t, solvers(max_record_size), m, mi, log_tv
    real(kind=kreal) :: l3_min_time = 0.0, diff_rate
    integer(kind=kint) :: l3_min_para, double_check_iteration = 2, expected_iter
    
    !C determin conv_index
    do conv_index = 1, conv_criterion_size
       if(tv > conv_criterion(conv_index)) exit
    end do
    if(conv_index > conv_criterion_size) conv_index = conv_criterion_size

    log_tv = -log10(tv)


    !C parameter switching with 
    call random_number(p)
    
    !C-- determine record_index: 
    if(level_each_criterion(conv_index) == 0) then
       size = zero_level_size
       level_p = level0_p
    else if(level_each_criterion(conv_index) == 1) then
       size = first_level_size
       level_p = level1_p
    else if(level_each_criterion(conv_index) == 2) then
       size = second_level_size
       level_p = level2_p
    else if(level_each_criterion(conv_index) == 3) then
       size = record_size
       level_p = level3_p
    else if(level_each_criterion(conv_index) == 4) then
       size = record_size
       level_p = level4_p
    end if
    
    !C fastest parameter: record_index
    record_index = 0
       
    min_time = -1

    i_start = 1
    
    !C if setup_fl for AT is unset, then i_start starts from zero_level_size+1.
    if(level_each_criterion(conv_index) > 0 .and. .not.setup_fl) i_start = zero_level_size+1
    do i = i_start, size
       call average_time_for_para(log_tv, i, t, expected_iter, setup_fl)
       write(*,*) "parameter #:", i, t,"[s]"
       
       !C performance efficiency is recorded in solvers(:), which is used for selection probability
       if(t > 0 ) then
!C          solvers(i) = (1/t)**2
          solvers(i) = t
       else
          solvers(i) = 0
       end if
       
       
       !C the case t<0 means that the parameter caused divergence 
       if((min_time < 0 .or. min_time > t) .and. t >= 0) then
          min_time = t
          record_index = i
       end if
    end do
    
    !C ------
    if(record_index == 0) then
       write(*,*) "No parameters can get convergence"
       record_index = 1 
    end if

    !C keisoku_fl
    if(keisoku_fl) counters(conv_index) = counters(conv_index) + 1


    !C level 0 -> level 1 
    if(level_each_criterion(conv_index) == 0 .and. &
         & level_switch_0 * zero_level_size + 1 <= counters(conv_index)) then
       level_each_criterion(conv_index) = 1
       call update_method(1, .true., setup_fl, log_tv)
       call set_level_switch_counter(int(level_switch_0 * zero_level_size +  &
            & level_switch_1 * (first_level_size - zero_level_size))+1, 2)
       
       !C level 1 -> level 2
    else if(level_each_criterion(conv_index) == 1 .and. &
         & level_2_switch_counter(conv_index) <= counters(conv_index)) then

       !C recently executed or not?
       do i = 1, parameter_hist_sz
          if(record_index == parameter_hist(i, conv_index)) exit
       end do
       if(i > parameter_hist_sz) then
#ifdef PRINT_REZ          
          !C check more.. 
          write(*,*) "check more 3 times at level 1:"
#endif
          call set_level_switch_counter(counters(conv_index)+3+1, 2)
          
       else
          level_each_criterion(conv_index) = 2
          call update_method(2, .true., setup_fl, log_tv)

          
          call set_level_switch_counter(level_2_switch_counter(conv_index) + &
               & int(level_switch_2*(second_level_size - first_level_size))+1, 3)
       end if
       
       !C level 2 -> level 3       
    else if(level_each_criterion(conv_index) == 2 .and. &
         & level_3_switch_counter(conv_index)  <= counters(conv_index)) then

       !C recently executed or not?
       do i = 1, parameter_hist_sz
          if(record_index == parameter_hist(i, conv_index)) exit
       end do
       if(i > parameter_hist_sz) then
          !C check more.. 
#ifdef PRINT_REZ          
          write(*,*) "check more 3 times at level 2:"
#endif
          call set_level_switch_counter(counters(conv_index)+3+1, 3) 
       else

!!$       l1_changed_fl = .false.
!!$       l0_changed_fl = .false.
!!$       call check_changed(record_index, conv_index, 1, l1_changed_fl)
!!$       call check_changed(record_index, conv_index, 0, l0_changed_fl)

!!$       if(l0_changed_fl .or. l1_changed_fl) then
!!$          !C different solver is called from level2.
!!$          !C do again level2 
!!$
!!$          if(l0_changed_fl) then
!!$             write(*,*) "do again level1", level_3_switch_counter(conv_index), counters(conv_index)
!!$          
!!$             call update_method(1, .true.)
!!$             level_each_criterion(conv_index) = 1
!!$             call set_level_switch_counter(counters(conv_index) + &
!!$                  & int(level_switch_1_again*(first_level_size - zero_level_size))+1, 2)
!!$             
!!$          else if(l1_changed_fl) then 
!!$             write(*,*) "do again level2", level_3_switch_counter(conv_index), counters(conv_index)
!!$             call update_method(2, .true.) 
!!$             call set_level_switch_counter(counters(conv_index) + &
!!$                  & second_level_size - first_level_size + 1, 3)
!!$          end if
          
!!$       else
#ifdef PRINT_REZ          
          write(*,*) "level2 -> level3", level_3_switch_counter(conv_index), counters(conv_index)
#endif
          level_each_criterion(conv_index) = 3
          call update_method(3,  .true., setup_fl, log_tv)
          call set_level_switch_counter(level_3_switch_counter(conv_index) + &
               & int(level_switch_3*(record_size-second_level_size))+1, 4)
          
!!$       end if
       end if
       
    else if(level_each_criterion(conv_index) == 3 .and. &
         & level_end_switch_counter(conv_index) <= counters(conv_index)) then

       !C recently executed or not?
       do i = 1, parameter_hist_sz
          if(record_index == parameter_hist(i, conv_index)) exit
       end do
       if(i > parameter_hist_sz) then
          !C check more.. 
#ifdef PRINT_REZ          
          write(*,*) "check more 3 times at level 3:"
#endif
          call set_level_switch_counter(counters(conv_index)+3+1, 4) 
       else 
          !C after all execution of level 3 solvers, keisoku_fl is unset.
          keisoku_fl = .false.
       end if

    else if(level_each_criterion(conv_index) == 4 .and. &
         & l4_counter(conv_index) == 0) then
       
       call check_changed(record_index, conv_index, 0, l0_changed_fl)
       call check_changed(record_index, conv_index, 1, l1_changed_fl)
       call check_changed(record_index, conv_index, 2, l2_changed_fl)

       !C calculate l3_min_time
       l3_min_time = 0
       do i = 1, l4
          j = l4_parameter_hist(i, conv_index)
          if(j > second_level_size) then
             call average_time_for_para(log_tv, j, t, expected_iter, setup_fl)
             if(l3_min_time == 0 .or. l3_min_time > t) then
                l3_min_time = t
                l3_min_para = j
             end if
          end if
       end do
       
       diff_rate = 1
       if(l3_min_time > 0) then
          diff_rate = (l3_min_time - min_time)/l3_min_time
#ifdef PRINT_REZ          
          write(*,*) "debug3: min_time", min_time, "l3_min_time", l3_min_time
#endif
       end if

       do i = l4-(double_check_iteration-1), l4
          if(record_index == l4_parameter_hist(i, conv_index)) exit
       end do
#ifdef PRINT_REZ                 
       write(*,*) diff_rate, "debug:", "(",record_index, min_time,solvers(record_index),")",i, l4, l1_changed_fl, &
            & l0_changed_fl, "::", l4_parameter_hist(1:l4, conv_index)
#endif
       
       if(i > l4) then
          !C do again level4!
          !C this case doesn't call the chosen solver at level 4
       else if(l0_changed_fl .and. diff_rate > diff_criterion) then
          !C do again level1
          level_each_criterion(conv_index) = 1
          call set_level_switch_counter(int (counters(conv_index) +  &
               & level_switch_1_again * (first_level_size - zero_level_size))+1, 2)
#ifdef PRINT_REZ                    
          write(*,*) "do again level 1", level_2_switch_counter(conv_index), counters(conv_index)
          !C solvers performance information for different theta is not reset.
#endif
          call update_method(1, .true., setup_fl, log_tv)

       else if(l1_changed_fl .and. diff_rate > diff_criterion) then
          !C different solver is called from level2..
          !C do again level2 
          level_each_criterion(conv_index) = 2
          call set_level_switch_counter( counters(conv_index) + second_level_size -first_level_size+1, 3)
#ifdef PRINT_REZ                    
          write(*,*) "do again level 2", level_3_switch_counter(conv_index), counters(conv_index)
#endif
          call update_method(2, .true., setup_fl, log_tv)
          
       else if(l2_changed_fl .and. .not.l0_changed_fl .and. .not.l1_changed_fl .and.  diff_rate > diff_criterion) then
          !C different solver is called from level2..
          !C do again level3 
          level_each_criterion(conv_index) = 3
!C          call set_level_switch_counter( counters(conv_index) + record_size - second_level_size + 1, 4)
          call set_level_switch_counter( counters(conv_index) + int(level_switch_3*(record_size - second_level_size)) + 1, 4)
#ifdef PRINT_REZ          
          write(*,*) "do again level 3", level_end_switch_counter(conv_index), counters(conv_index)
#endif
          call update_method(3, .true., setup_fl, log_tv)

       else 
          !C this operation is set for the case of theta_diff_fl = .true. and diff_rate < 0.25
!!$          if(l0_changed_fl) then
!!$             record_index = l3_min_para 
!!$          end if

          keisoku_fl = .false.
!!$          call set_level_switch_counter( counters(conv_index) + int(level_switch_3*(record_size - second_level_size)) + 1, 4)
          level_each_criterion(conv_index) = 3
       end if
       
       if(level_each_criterion(conv_index) == 4) then
          l4_counter(conv_index) = l4-1
       else
          l4_counter(conv_index) = l4
       end if
       
    else if(level_each_criterion(conv_index) == 4) then
       l4_counter(conv_index) = l4_counter(conv_index) - 1
    end if

    !C at level4, record index is shifted to parameter 
    !C which is not in l4_parameter_hist
    if(level_each_criterion(conv_index) == 4 .and. min_time > 0) then
       if(l4_counter(conv_index) > (double_check_iteration-1)) then
          do i = 1, l4 - l4_counter(conv_index) - 1
             solvers(l4_parameter_hist(i, conv_index)) = 0
          end do
       else
          !C last j times
#ifdef PRINT_REZ          
          write(*,*) "-- double check execution --"
#endif
          do i = l4-(double_check_iteration-1), l4 - l4_counter(conv_index) - 1
             solvers(l4_parameter_hist(i, conv_index)) = 0
          end do
       end if
          
       !C i_start switches setup_fl_AT is true or false.
       m = 0; mi = 0
       do i = i_start, size
          if(m < solvers(i)) then
             m = solvers(i)
             mi = i
          end if
       end do
       if(mi > 0) then
          record_index = mi
       else
#ifdef PRINT_REZ          
          write(*,*) "not change parameter at level 4"
#endif
       end if
    end if

    !C 0708
    !C switching to probabilistic solver selection 
    !C level_p means the rate to select top parameter without probabilistic selection.
!!$    if(p > level_p .and. min_time > 0)  then
!!$       do i = 2, size
!!$          solvers(i) = solvers(i-1)+solvers(i)
!!$       end do
!!$       if(solvers(size) > 0) then
!!$          do i = 1, size
!!$             solvers(i) = solvers(i)/solvers(size)
!!$          end do
!!$       
!!$          call random_number(q)
!!$       
!!$          do i = 1, size
!!$             if(q < solvers(i)) exit
!!$          end do
!!$          
!!$          record_index = i
!!$       end if
!!$    end if
    
    if(record_index > 0) then
       !C record called parameter setting 
       if(level_each_criterion(conv_index) == 4) l4_parameter_hist(l4-l4_counter(conv_index), conv_index) = record_index
       parameter_hist(mod(counters(conv_index),parameter_hist_sz) + 1, conv_index) = record_index

#ifdef PRINT_REZ                 
       write(*,*) "########", record_index, first_level_size, size, level_each_criterion(conv_index), conv_index, "#######"
#endif
       param_out = parameters(record_index,conv_index) % param
       
!!$       call average_time_for_para(log_tv, record_index, t, expected_iter, setup_fl)
!!$       write(*,*) "@@ expected_iter is",  expected_iter, "@@@@"

    end if
    
  end subroutine solver_choice


  !C this subroutin update time information with iteration number.
  subroutine update_time2(setup_time, iter_time, iter, tv, reach_conv, setup_fl)
    implicit none
    real(kind=kreal) :: time, v, tv, setup_time, iter_time, oit, oin
    logical :: setup_fl
    integer(kind=kint) :: c, i, iter
    logical :: reach_conv
    
    if(record_index > 0 .and. record_index <= record_size  .and. &
         & conv_index > 0  .and. conv_index <= conv_criterion_size) then
       if(reach_conv) then
          
          !C tv must be correct value..
          !C if tv changed from the value in at_pre_process(), something strange may occur..
          oin = -log10(tv) / iter
          oit = iter_time / iter
          
          !C one_iter_time(0, conv_index) keep the sum of 
          !C the values of one_iter_time(1:time_hist_sz, conv_index)
          c = parameters(record_index,conv_index) % conv_count
          i = mod(c, time_hist_sz)+1
          parameters(record_index,conv_index) % one_iter_time(0) = &
               &  parameters(record_index,conv_index) % one_iter_time(0) - &
               &  parameters(record_index,conv_index) % one_iter_time(i) + oit
          parameters(record_index,conv_index) % one_iter_norm_reduction(0) = &
               &  parameters(record_index,conv_index) % one_iter_norm_reduction(0) - &
               &  parameters(record_index,conv_index) % one_iter_norm_reduction(i) + oin
          if(setup_fl) then
             parameters(record_index,conv_index) % setup_time(0) = &
                  &  parameters(record_index,conv_index) % setup_time(0) - &
                  &  parameters(record_index,conv_index) % setup_time(i) + setup_time
          end if
          
          !C conv(i,conv_index) have new time data.
          parameters(record_index,conv_index) % one_iter_time(i) = oit
          parameters(record_index,conv_index) % one_iter_norm_reduction(i) = oin
          if(setup_fl) parameters(record_index,conv_index) % setup_time = setup_time
             
          !C count up the counter of that parameter if it reaches convergence
          parameters(record_index,conv_index) % conv_count = c+1  
          
       else 
          if(parameters(record_index,conv_index) % conv_count > 0 .and.  &
               & level_each_criterion(conv_index) /= 4 .and. &
               & level_each_criterion(conv_index) /= 1) then
             !C reset the parameters count and move to level 1             
             call reset_time_data(record_index, conv_index)

             !C in order to goto level 0, counters is reset at 0
             counters(conv_index) = 0
             level_each_criterion(conv_index) = 0
             
          else
             !C invalidate that parameter which fails the first time execution.
             parameters(record_index,conv_index) % one_iter_time(:) = -1

             !C more aggresive acceleration parameter is invalidated.
             call invalidate_aggressive_acceleration(conv_index, record_index)
          end if
       end if
    else
#ifdef PRINT_REZ          
       write(*,*) "parameters are not set: parameter_at.F90"
#endif
    end if
  end subroutine update_time2


  !C one_iter_norm_reduction is not used.
  !C one_iter_time contains the time for reduction of unit norm residual reduction 
  subroutine update_time(setup_time, iter_time, iter, tv, reach_conv, setup_fl)
    implicit none
    real(kind=kreal) :: time, v, tv, setup_time, iter_time, oit, oin
    logical :: setup_fl
    integer(kind=kint) :: c, i, iter
    logical :: reach_conv
    
    if(record_index > 0 .and. record_index <= record_size  .and. &
         & conv_index > 0  .and. conv_index <= conv_criterion_size) then
       if(reach_conv) then
          
          if(conv_index==1) then
             oit = iter_time
          else 
             oit = iter_time / (-log10(tv))
          end if

#ifdef PRINT_REZ                    
          write(*,*) "update time():  time for unit norm reduction : ", iter_time, setup_time, oit, tv
#endif


          !C one_iter_time(0, conv_index) keep the sum of 
          !C the values of one_iter_time(1:time_hist_sz, conv_index)
          c = parameters(record_index,conv_index) % conv_count
          i = mod(c, time_hist_sz)+1
          parameters(record_index,conv_index) % one_iter_time(0) = &
               &  parameters(record_index,conv_index) % one_iter_time(0) - &
               &  parameters(record_index,conv_index) % one_iter_time(i) + oit
          
          if(setup_fl) then
             parameters(record_index,conv_index) % setup_time(0) = &
                  &  parameters(record_index,conv_index) % setup_time(0) - &
                  &  parameters(record_index,conv_index) % setup_time(i) + setup_time
          end if
          
          !C conv(i,conv_index) have new time data.
          parameters(record_index,conv_index) % one_iter_time(i) = oit
          if(setup_fl) parameters(record_index,conv_index) % setup_time(i) = setup_time
             
          !C count up the counter of that parameter if it reaches convergence
          parameters(record_index,conv_index) % conv_count = c+1  
          
       else 
          if(parameters(record_index,conv_index) % conv_count > 0 .and.  &
               & level_each_criterion(conv_index) /= 4 .and. &
               & level_each_criterion(conv_index) /= 1) then
             !C reset the parameters count and move to level 1             
             call reset_time_data(record_index, conv_index)

             !C in order to goto level 0, counters is reset at 0
             counters(conv_index) = 0
             level_each_criterion(conv_index) = 0
             
          else
             !C invalidate that parameter which fails the first time execution.
             parameters(record_index,conv_index) % one_iter_time(:) = -1

             !C more aggresive acceleration parameter is invalidated.
             call invalidate_aggressive_acceleration(conv_index, record_index)
          end if
       end if
    else
#ifdef PRINT_REZ          
       write(*,*) "parameters are not set: parameter_at.F90"
#endif
    end if
  end subroutine update_time



  !C keisoku_fl == .false. ->  it calls check() every iteraion.
  subroutine check(iter, reach_conv, keisoku_fl, offline_fl)
    integer(kind=kint):: iter, i
    integer(kind=kint),save :: counter(max_conv_criterion_size) = 0
    logical :: reach_conv, keisoku_fl, offline_fl, interval_fl
    integer(kind=kint), save :: iter_bak(max_conv_criterion_size) = 0

    counter(conv_index) = counter(conv_index) + 1
    
    !C if offline_fl == .true. then check_interval <= 0. 
    !C to avoid division by zero, interval_fl is set.
    interval_fl = .false.
    if(.not.offline_fl) interval_fl = (mod(counter(conv_index), check_interval) == 0)
#ifdef PRINT_REZ              
    write(*,*) "debug2:", counter(conv_index), interval_fl
#endif
    
    if((iter_bak(conv_index) > 0 .and. iter > iter_bak(conv_index) + difference_of_iter - 1) &
         & .or. .not. reach_conv .or. interval_fl)  then

       keisoku_fl = .true.
       
       if(.not.offline_fl) then
          level_each_criterion(conv_index) = 4
          
          !C default parameter is initialized when it isn't marked as divergent
          if(parameters(default_index,conv_index) % one_iter_time(0) >= 0) then
             call reset_time_data(default_index, conv_index)
             
!!$             parameters(default_index,conv_index) % conv_count = 0
!!$             parameters(default_index,conv_index) % one_iter_time(:) = 0 
          end if
       end if

       !C reset iter_bak
       iter_bak(conv_index) = -1
       
    else
       iter_bak(conv_index) = iter
    end if

  end subroutine check
  
end module at_mod

  !C if inter_val is <=0, offline_fl is set.
!!$  subroutine at_parameter_setting(inter_val, diff_of_iter, setup_fl)
!!$    implicit none
!!$    
!!$    integer(kind=kint) :: i,j,k,l, inter_val, diff_of_iter
!!$    logical :: setup_fl
!!$
!!$    !C set at_mode
!!$    at_flag = .true.
!!$
!!$    setupflag_for_AT = setup_fl
!!$
!!$    !C inter_val < 0 means offline tuning.
!!$    if(inter_val <= 0) then
!!$       offline_fl =.true.
!!$    end if
!!$
!!$    !C set difference of iteration numbers
!!$    difference_of_iter = diff_of_iter
!!$    
!!$    !C set check_interval
!!$    check_interval = inter_val
!!$
!!$    !C prepare conv_criterion
!!$    conv_criterion_size = 0 
!!$    conv_criterion_size = conv_criterion_size+1
!!$    conv_criterion(conv_criterion_size) = 1.0E-1
!!$
!!$    
!!$    !C prepare parameters
!!$    !C NO,CG,BICGSTAB, IDR(?), GMRES(?), RESCUT(?)
!!$    record_size = 0
!!$
!!$    setup_level_size = 0
!!$    if(setup_fl) then
!!$       !C AMGCG with  V-cycle and 1 sgs
!!$       record_size = record_size + 1
!!$       parameters(record_size) % method = 22
!!$       parameters(record_size) % theta  = 0.05
!!$       parameters(record_size) % dump_jacobi_weight  = 0.666
!!$       parameters(record_size) % cycle_kind = 0
!!$       parameters(record_size) % number_of_vectors = 10
!!$       parameters(record_size) % iteration_number_up = 1
!!$       parameters(record_size) % iteration_number_down = 1
!!$       parameters(record_size) % iteration_number_bottom = 30
!!$       parameters(record_size) % f_first_iteration_number = 0
!!$       parameters(record_size) % dump_gauss_seidel_weight = 1.0
!!$       parameters(record_size) % gs_kind = 0
!!$       parameters(record_size) % solver_string = "CG"
!!$       default_index = record_size
!!$       
!!$       record_size = record_size + 1
!!$       parameters(record_size) = parameters(record_size-1)
!!$       parameters(record_size) % theta  = 0.07
!!$
!!$       record_size = record_size + 1
!!$       parameters(record_size) = parameters(record_size-1)
!!$       parameters(record_size) % theta  = 0.09
!!$
!!$       record_size = record_size + 1
!!$       parameters(record_size) = parameters(record_size-1)
!!$       parameters(record_size) % theta  = 0.11
!!$
!!$       record_size = record_size + 1
!!$       parameters(record_size) = parameters(record_size-1)
!!$       parameters(record_size) % theta  = 0.04
!!$
!!$       record_size = record_size + 1
!!$       parameters(record_size) = parameters(record_size-1)
!!$       parameters(record_size) % theta  = 0.03
!!$
!!$       setup_level_size = record_size
!!$       
!!$    end if
!!$       
!!$    record_size = record_size+1
!!$    parameters(record_size) % cycle_kind = 0
!!$    parameters(record_size) % number_of_vectors = 4
!!$    parameters(record_size) % iteration_number_up = 1
!!$    parameters(record_size) % iteration_number_down = 1
!!$    parameters(record_size) % iteration_number_bottom = 30
!!$    parameters(record_size) % f_first_iteration_number = 0
!!$    parameters(record_size) % dump_gauss_seidel_weight = 0.8
!!$    
!!$#ifndef ATCG
!!$    parameters(record_size) % solver_string = "RESCUT"
!!$    parameters(record_size) % gs_kind = 0
!!$
!!$
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size -1)
!!$    parameters(record_size) % gs_kind = 1
!!$
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size -1)
!!$    parameters(record_size) % number_of_vectors = 2
!!$    parameters(record_size) % gs_kind = 0
!!$
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size -1)
!!$    parameters(record_size) % gs_kind = 1
!!$
!!$    record_size=record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$#endif
!!$    parameters(record_size) % solver_string = "CG"
!!$    parameters(record_size) % gs_kind = 0
!!$    !C default_index specifies AMGCG 
!!$    if(.not.setup_fl) default_index = record_size
!!$
!!$    record_size=record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % gs_kind = 1
!!$
!!$#ifndef ATCG
!!$    record_size=record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % solver_string = "BICGSTAB"
!!$    parameters(record_size) % gs_kind = 0
!!$
!!$    record_size=record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % gs_kind = 1
!!$#endif

!!$#ifndef ATCG
!!$    record_size=record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % solver_string = "NO"
!!$    parameters(record_size) % gs_kind = 0
!!$
!!$    record_size=record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % gs_kind = 1
!!$#endif
!!$    
!!$    first_level_size = record_size
!!$    !C -- level 2 parameters
!!$
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % dump_gauss_seidel_weight = 0.8
!!$    
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % dump_gauss_seidel_weight = 1.2
!!$
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % dump_gauss_seidel_weight = 1.4
!!$
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % dump_gauss_seidel_weight = 1.6
!!$
!!$
!!$    second_level_size = record_size
!!$
!!$#ifndef ATCG    
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % cycle_kind = 1
!!$#endif
!!$    
!!$    record_size = record_size+1
!!$    parameters(record_size) = parameters(record_size-1)
!!$    parameters(record_size) % cycle_kind = 2
!!$    
!!$    level_each_criterion = 1
!!$    if(setup_fl) level_each_criterion = 0
!!$    
!!$    !C level_2_switch_counter is set for setup_fl == .false.
!!$    !C if setup_fl == .true. then this val is set at level 0.
!!$    call set_level_2_switch_counter(int(level_switch_1 * first_level_size)) 
!!$    
!!$    call at_parameter_init()
!!$
!!$  end subroutine at_parameter_setting


!!$  subroutine update_time(time, tv, reach_conv, setup_fl)
!!$    implicit none
!!$    real(kind=kreal) :: time, v, tv
!!$    logical :: setup_fl
!!$    integer(kind=kint) :: c, i
!!$    logical :: reach_conv
!!$
!!$    if(record_index > 0 .and. record_index <= record_size  .and. &
!!$         & conv_index > 0   .and. conv_index <= conv_criterion_size) then
!!$       if(reach_conv) then
!!$          
!!$          !C normalize when setup_fl == .false. and tv < 1.0E-1
!!$          if(tv < 1.0E-1 .and. .not. setup_fl) then
!!$             time = time / -log10(tv)
!!$          end if
!!$
!!$          !C conv_time(0, conv_index) keep the sum of 
!!$          !C the values of conv_time(1:time_hist_sz, conv_index)
!!$          c = parameters(record_index) % conv_count(conv_index)
!!$          i = mod(c, time_hist_sz)+1
!!$          parameters(record_index) % conv_time(0, conv_index) = &
!!$               &  parameters(record_index) % conv_time(0, conv_index) - &
!!$               &  parameters(record_index) % conv_time(i, conv_index) + time
!!$          
!!$          !C conv(i,conv_index) have new time data.
!!$          parameters(record_index) % conv_time(i, conv_index) = time          
!!$          
!!$          !C count up the counter of that parameter if it reaches convergence
!!$          parameters(record_index) % conv_count(conv_index) = c+1  
!!$          
!!$       else 
!!$          if(parameters(record_index) % conv_count(conv_index) > 0 .and.  &
!!$               & level_each_criterion(conv_index) /= 4 .and. &
!!$               & level_each_criterion(conv_index) /= 1) then
!!$             !C reset the parameters count and move to level 1             
!!$             parameters(record_index) % conv_time(:, conv_index) = 0
!!$             parameters(record_index) % conv_count(conv_index) = 0
!!$
!!$             !C in order to goto level 1, counters is reset at 0
!!$             counters(conv_index) = 0
!!$             level_each_criterion(conv_index) = 1
!!$             if(setup_fl) level_each_criterion(conv_index) = 0
!!$             
!!$          else
!!$             !C invalidate that parameter which fails the first time execution.
!!$             parameters(record_index) % conv_time(:,conv_index) = -1
!!$
!!$             !C more aggresive acceleration parameter is invalidated.
!!$             if(level_each_criterion(conv_index) == 2 .and. record_index > first_level_size) then
!!$                do i = record_index, second_level_size
!!$                   parameters(i) % conv_time(:,:) = -1
!!$                   parameters(i) % conv_count = 0
!!$                end do
!!$             end if
!!$          end if
!!$       end if
!!$    else
!!$       write(*,*) "parameters are not set: parameter_at.F90"
!!$    end if
!!$  end subroutine update_time

!!$  subroutine solver_choice(tv, keisoku_fl, setup_fl)
!!$    implicit none
!!$    logical :: keisoku_fl, setup_fl, sol_diff_fl, theta_diff_fl
!!$    real(kind=kreal) :: tv, p, q, level_p
!!$    integer(kind=kint) :: i, j, size, c
!!$    integer(kind=kint), parameter :: l4 = 10
!!$    integer(kind=kint), save :: l4_counter=l4
!!$    integer(kind=kint), save :: l4_parameter_hist(l4) = 0
!!$    real(kind=kreal) :: min_time, t, solvers(max_record_size)
!!$    
!!$    !C determin conv_index
!!$    do conv_index = 1, conv_criterion_size
!!$       if(tv > conv_criterion(conv_index)) exit
!!$    end do
!!$    if(conv_index > conv_criterion_size) conv_index = conv_criterion_size
!!$
!!$
!!$    !C parameter switching with 
!!$    call random_number(p)
!!$    
!!$    !C-- determine record_index: 
!!$    if(level_each_criterion(conv_index) == 0) then
!!$       size = setup_level_size
!!$       level_p = level0_p
!!$    else if(level_each_criterion(conv_index) == 1) then
!!$       size = first_level_size
!!$       level_p = level1_p
!!$    else if(level_each_criterion(conv_index) == 2) then
!!$       size = second_level_size
!!$       level_p = level2_p
!!$    else if(level_each_criterion(conv_index) == 3) then
!!$       size = record_size
!!$       level_p = level3_p
!!$    else if(level_each_criterion(conv_index) == 4) then
!!$       size = record_size
!!$       level_p = level4_p
!!$    end if
!!$    
!!$    !C fastest parameter: record_index
!!$    record_index = 0
!!$    min_time = -1
!!$    do i = 1, size
!!$       if(parameters(i) % conv_count(conv_index) > 0) then
!!$          t = parameters(i) % conv_time(0, conv_index)
!!$          if(parameters(i) % conv_count(conv_index) >= time_hist_sz) then
!!$             t = t / time_hist_sz
!!$          else
!!$             t = t / parameters(i) % conv_count(conv_index)
!!$          end if
!!$       else 
!!$          t = parameters(i) % conv_time(0,conv_index)
!!$       end if
!!$       
!!$       !C performance efficiency is recorded in solvers(:), which is used for selection probability
!!$       if(t > 0 ) then
!!$          solvers(i) = (1/t)**2
!!$       else
!!$          solvers(i) = 0
!!$       end if
!!$       
!!$       
!!$       !C the case t<0 means that the parameter caused divergence 
!!$       if((min_time < 0 .or. min_time > t) .and. t >= 0) then
!!$          min_time =t 
!!$          record_index = i
!!$       end if
!!$    end do
!!$    
!!$    !C ------
!!$    if(record_index == 0) then
!!$       write(*,*) "No parameters can get convergence"
!!$       record_index = 1 
!!$    end if
!!$
!!$    !C keisoku_fl
!!$    if(keisoku_fl) counters(conv_index) = counters(conv_index) + 1
!!$
!!$    
!!$    !C level 0 -> level 1 
!!$    if(level_each_criterion(conv_index) == 0 .and. &
!!$         & level_switch_0 * setup_level_size < counters(conv_index)) then
!!$       level_each_criterion(conv_index) = 1
!!$       call update_method(1, .true.)
!!$       call set_level_2_switch_counter(int(level_switch_0 * setup_level_size +  &
!!$            & level_switch_1 * (first_level_size - setup_level_size))) 
!!$       
!!$       !C level 1 -> level 2
!!$    else if(level_each_criterion(conv_index) == 1 .and. &
!!$         & level_2_switch_counter+1 < counters(conv_index)) then
!!$       level_each_criterion(conv_index) = 2
!!$       call update_method(2, .true.)
!!$       
!!$       call set_level_3_switch_counter(level_2_switch_counter + &
!!$            & (second_level_size - first_level_size) + 1)
!!$       
!!$       !C level 2 -> level 3       
!!$    else if(level_each_criterion(conv_index) == 2 .and. &
!!$         & level_3_switch_counter+1  < counters(conv_index)) then
!!$       
!!$       sol_diff_fl = parameters(record_index) % solver_string /=  &
!!$            &        parameters(first_level_size+1) % solver_string .or. &
!!$            &        parameters(record_index) % gs_kind /=  &
!!$            &        parameters(first_level_size+1) % gs_kind
!!$       
!!$       theta_diff_fl = parameters(record_index) % theta /=  &
!!$            &          parameters(first_level_size+1) % theta
!!$       if(sol_diff_fl .or. theta_diff_fl) then
!!$          !C different solver is called from level2.
!!$          !C do again level2 
!!$          
!!$          call set_level_3_switch_counter(counters(conv_index) + &
!!$               & second_level_size - first_level_size + 1)
!!$
!!$          write(*,*) "do again level2", level_3_switch_counter, counters(conv_index)
!!$
!!$          if(sol_diff_fl) then 
!!$             call update_method(2, .true.) 
!!$          else if(theta_diff_fl) then
!!$             call update_method(2, .true.)
!!$          end if
!!$          
!!$       else
!!$          write(*,*) "level2 -> level3", level_3_switch_counter, counters(conv_index)
!!$          level_each_criterion(conv_index) = 3
!!$          call update_method(3,  .true.)
!!$          
!!$       end if
!!$       
!!$    else if(level_each_criterion(conv_index) == 3 .and. &
!!$         & counters(conv_index) > level_3_switch_counter + record_size - second_level_size+2) then
!!$       !C after all execution of level 3 solvers, keisoku_fl is unset.
!!$       keisoku_fl = .false.
!!$
!!$    else if(level_each_criterion(conv_index) == 4 .and. &
!!$         & l4_counter == 0) then
!!$
!!$       do i = 1, l4
!!$          if(record_index == l4_parameter_hist(i)) exit
!!$       end do
!!$
!!$       sol_diff_fl = parameters(record_index) % solver_string /= &
!!$            &        parameters(first_level_size+1) % solver_string .or. &
!!$            &        parameters(record_index) % gs_kind /= parameters(first_level_size+1) % gs_kind
!!$       theta_diff_fl = parameters(record_index) % theta /= &
!!$            &        parameters(first_level_size+1) % theta
!!$       
!!$       if(i > l4) then
!!$          !C do again level4!
!!$          !C this case doesn't call the chosen solver at level 4
!!$       else if(sol_diff_fl .and. .not.theta_diff_fl) then
!!$          !C different solver is called from level2..
!!$          !C do again level2 
!!$          level_each_criterion(conv_index) = 2
!!$          call set_level_3_switch_counter( counters(conv_index) + second_level_size - first_level_size+1)
!!$          
!!$          write(*,*) "do again level2", level_3_switch_counter, counters(conv_index)
!!$          call update_method(2, .true.)
!!$          
!!$       else if(theta_diff_fl) then
!!$          !C different theta is set..
!!$          !C do again level1
!!$          level_each_criterion(conv_index) = 1
!!$          
!!$          call set_level_2_switch_counter(int (counters(conv_index) +  &
!!$               & level_switch_1_again * (first_level_size - setup_level_size)))
!!$          
!!$          write(*,*) "do again level1", level_2_switch_counter, counters(conv_index)
!!$          !C solvers performance information for different theta is not reset.
!!$          call update_method(1, .true.)
!!$          
!!$       else
!!$          keisoku_fl = .false.
!!$          level_each_criterion(conv_index) = 3
!!$       end if
!!$       l4_counter = l4
!!$       
!!$    else if(level_each_criterion(conv_index) == 4) then
!!$       l4_counter = l4_counter -1
!!$    end if
!!$
!!$    !C switching to probabilistic solver selection 
!!$    !C level_p means the rate to select top parameter without probabilistic selection.
!!$
!!$    
!!$    if(p > level_p .and. min_time > 0)  then
!!$       
!!$       !C at level4, parameter in l4_parameter_hist is not chosen again.
!!$       if(level_each_criterion(conv_index) == 4) then
!!$          do i = 1, l4 - l4_counter - 1
!!$             solvers(l4_parameter_hist(i)) = 0
!!$          end do
!!$       end if
!!$       
!!$       do i = 2, size
!!$          solvers(i) = solvers(i-1)+solvers(i)
!!$       end do
!!$       if(solvers(size) > 0) then
!!$          do i = 1, size
!!$             solvers(i) = solvers(i)/solvers(size)
!!$          end do
!!$       
!!$          call random_number(q)
!!$       
!!$          do i = 1, size
!!$             if(q < solvers(i)) exit
!!$          end do
!!$          
!!$          record_index = i
!!$       end if
!!$    end if
!!$    
!!$    if(record_index > 0) then
!!$       !C record called parameter setting 
!!$       if(level_each_criterion(conv_index) == 4) l4_parameter_hist(l4-l4_counter) = record_index
!!$       
!!$       write(*,*) "########", record_index, first_level_size, size, level_each_criterion(conv_index), "#######"
!!$       
!!$       call set_iter_parameter(parameters(record_index))
!!$       if(setup_fl) call set_setup_parameter(parameters(record_index))
!!$    end if
!!$
!!$  end subroutine solver_choice
