module axis_comm
  include 'precision.inc'
  type comm_table
     integer(kind=kint) :: NEIBPETOT
     integer(kind=kint), pointer :: NEIBPE(:)
     integer(kind=kint), pointer :: STACK_IMPORT(:)
     integer(kind=kint), pointer :: STACK_EXPORT(:)
  end type comm_table
  
  type buf_type
     integer(kind=kint) :: msg_sz, msg_num
     !C msg_index(0:msg_num)
     integer(kind=kint), pointer :: msg_index(:)
     !C msg_next_dst(1:msg_num)
     integer(kind=kint), pointer :: msg_next_dst(:)     
     real(kind=kreal), pointer :: buf(:)
  end type buf_type
  
  type comm_info
     type(comm_table) :: comm_X
     type(comm_table) :: comm_Y
     type(comm_table) :: comm_Z
     !C for master
     type(comm_table) :: comm_C
     type(buf_type) :: X_sbuf, X_rbuf
     type(buf_type) :: Y_sbuf, Y_rbuf
     type(buf_type) :: Z_sbuf, Z_rbuf
     type(buf_type) :: C_sbuf, C_rbuf
     
     integer :: SOLVER_COMM, my_rank
     integer :: C,X,Y,Z

     integer :: msg_num_master_self
  end type comm_info
contains
  
  subroutine clear_comm(comm)
    implicit none    
    type(comm_info), target :: comm

    call clear_comm_table(comm % comm_C)
    call clear_buf(comm % C_sbuf)
    call clear_buf(comm % C_rbuf)

    call clear_comm_table(comm % comm_X)
    call clear_buf(comm % X_sbuf)
    call clear_buf(comm % X_rbuf)

    call clear_comm_table(comm % comm_Y)
    call clear_buf(comm % Y_sbuf)
    call clear_buf(comm % Y_rbuf)

    call clear_comm_table(comm % comm_Z)
    call clear_buf(comm % Z_sbuf)
    call clear_buf(comm % Z_rbuf)


  end subroutine clear_comm
  subroutine clear_buf(xbuf)
    implicit none 
    type(buf_type) :: xbuf
    
    if(associated(xbuf % msg_index)) &
         & deallocate(xbuf % msg_index)

    if(associated(xbuf % msg_next_dst)) &
         & deallocate(xbuf % msg_next_dst)
    
    if(associated(xbuf % buf)) &
         & deallocate(xbuf % buf)
    
    
  end subroutine clear_buf
  
  subroutine clear_comm_table(comm_t)
    implicit none
    type(comm_table) :: comm_t
    
    if(associated(comm_t % NEIBPE)) deallocate(comm_t % NEIBPE)
    if(associated(comm_t % STACK_IMPORT)) &
         & deallocate(comm_t % STACK_IMPORT)
    if(associated(comm_t % STACK_EXPORT)) &
         & deallocate(comm_t % STACK_EXPORT)
    
  end subroutine clear_comm_table


  !C reverse version of new_SR2()
  subroutine new_SR2_rev_add(comm, NEIBPETOT, NEIBPE, &
       & STACK_IMPORT, NOD_IMPORT, &
       & STACK_EXPORT, NOD_EXPORT, WR, X)
    implicit none
    type(comm_info) :: comm
    integer(kind=kint) :: NEIBPETOT
    integer(kind=kint) :: NEIBPE(:)
    integer(kind=kint) :: STACK_IMPORT(0:), NOD_IMPORT(:)
    integer(kind=kint) :: STACK_EXPORT(0:), NOD_EXPORT(:)
    real(kind=kreal) :: WR(:), X(:)

    integer(kind=kint) :: i,j, axis, my_rank
    logical :: reverse_fl


    reverse_fl = .true.
    my_rank = comm % my_rank
    if(mod(my_rank, comm % C) > 0) then
       !C C_rbuf % msg_next_dst <- WR <- X
       call recvfrommaster_pkt(comm, NEIBPETOT, &
            & STACK_IMPORT, NOD_IMPORT, X, WR, reverse_fl)

       call sendtomaster_pkt(comm, NEIBPETOT,  &
            & STACK_EXPORT, NOD_EXPORT, X, reverse_fl)

    else
       !C Z_rbuf <- C_sbuf 
       !C C_sbuf(my_rank) with C_sbuf % msg_next_dst, 
       !C                  comm % msg_num_master_self <- WR <- X
       call send_all(comm, NEIBPETOT, STACK_IMPORT, NOD_IMPORT, &
            & X, WR, reverse_fl)

       !C Y_rbuf <- Z_sbuf <- Z_rbuf
       axis = 3
       call sendrecv_all(comm, axis, reverse_fl)

       !C X_rbuf <- Y_sbuf <- Y_rbuf
       axis = axis-1
       call sendrecv_all(comm, axis, reverse_fl)

       !C C_rbuf <- X_sbuf <- X_rbuf
       axis = axis-1
       call sendrecv_all(comm, axis, reverse_fl)

       !C <- C_rbuf
       call recv_all(comm, NEIBPETOT, STACK_EXPORT, &
            & NOD_EXPORT, X, reverse_fl)

    end if

  end subroutine new_SR2_rev_add



  
  !C-- client
  !C sendtomaster_pkt()
  !C recvfrommaster_pkt()
  !C
  !C-- master
  !C recv_all()
  !C sendrecv_all() 3 times X, Y, Z direction.
  !C send_all()
  !C
  subroutine new_SR2(comm, NEIBPETOT, NEIBPE, &
       & STACK_IMPORT, NOD_IMPORT, &
       & STACK_EXPORT, NOD_EXPORT, WR, X)
    implicit none
    
    type(comm_info) :: comm
    integer(kind=kint) :: NEIBPETOT
    integer(kind=kint) :: NEIBPE(:)
    integer(kind=kint) :: STACK_IMPORT(0:), NOD_IMPORT(:)
    integer(kind=kint) :: STACK_EXPORT(0:), NOD_EXPORT(:)
    real(kind=kreal) :: WR(:), X(:)
    
    integer(kind=kint) :: i,j, axis, my_rank
    logical :: reverse_fl
    
    reverse_fl = .false.
    
    my_rank = comm % my_rank
    if(mod(my_rank, comm % C) > 0) then
       call sendtomaster_pkt(comm, NEIBPETOT,  &
            & STACK_EXPORT, NOD_EXPORT, X, reverse_fl)
       
       !C C_rbuf % msg_next_dst -> WR -> X
       call recvfrommaster_pkt(comm, NEIBPETOT, &
            & STACK_IMPORT, NOD_IMPORT, X, WR, reverse_fl)
       
    else
       !    !C -> C_rbuf
       call recv_all(comm, NEIBPETOT, STACK_EXPORT, &
            & NOD_EXPORT, X, reverse_fl)
       
       !C C_rbuf -> X_sbuf -> X_rbuf
       axis = 1
       call sendrecv_all(comm, axis, reverse_fl)

       !C X_rbuf -> Y_sbuf -> Y_rbuf
       axis = axis+1
       call sendrecv_all(comm, axis, reverse_fl)
       
       !C Y_rbuf -> Z_sbuf -> Z_rbuf
       axis = axis+1
       call sendrecv_all(comm, axis, reverse_fl)

       !C Z_rbuf -> C_sbuf 
       !C C_sbuf(my_rank) with C_sbuf % msg_next_dst, 
       !C                  comm % msg_num_master_self-> WR -> X
       call send_all(comm, NEIBPETOT, STACK_IMPORT, NOD_IMPORT, &
            & X, WR, reverse_fl)
    
    end if
    
  end subroutine new_SR2

  subroutine sendtomaster_pkt(comm, NEIBPETOT, &
       & STACK_EXPORT, NOD_EXPORT, X, reverse_fl)
    implicit none
    include 'mpif.h'

    type(comm_info), target :: comm
    integer(kind=kint) :: NEIBPETOT
    integer(kind=kint) :: STACK_EXPORT(0:), NOD_EXPORT(:)
    real(kind=kreal) ::  X(:)
    logical, intent(in) :: reverse_fl
    
    integer(kind=kint) :: i,j,row
    integer :: ierr
    type(buf_type), pointer :: sbuf

    integer, allocatable :: sta(:)
    allocate(sta(MPI_STATUS_SIZE))

    sbuf => comm % C_sbuf

    if(.not.reverse_fl) then
       !$omp parallel do private(row)
       do i = 1,STACK_EXPORT(NEIBPETOT)
          row   = NOD_EXPORT(i)
          sbuf % buf(i) = X(row)
       end do
       !$omp end parallel do

       if(sbuf % msg_sz > 0) then
          call MPI_SEND(sbuf % buf, sbuf % msg_sz, MPI_DOUBLE_PRECISION, &
               & comm % comm_C % NEIBPE(1), 0, comm % SOLVER_COMM, ierr)
       end if
    else
       if(sbuf % msg_sz > 0) then
          call MPI_RECV(sbuf % buf, sbuf % msg_sz, MPI_DOUBLE_PRECISION, &
               & comm % comm_C % NEIBPE(1), 0, comm % SOLVER_COMM, sta, &
               & ierr)
       end if

       do j = 1, NEIBPETOT
          !$omp parallel do private(row)
          do i = STACK_EXPORT(j-1)+1,STACK_EXPORT(j)
             row   = NOD_EXPORT(i)
             X(row) = X(row) + sbuf % buf(i)
          end do
          !$omp end parallel do
       end do

    end if
    
    deallocate(sta)
    
  end subroutine sendtomaster_pkt

  subroutine recvfrommaster_pkt(comm, NEIBPETOT, &
       & STACK_IMPORT, NOD_IMPORT, X, WR, reverse_fl)
    implicit none
    include 'mpif.h'

    type(comm_info), target :: comm
    integer(kind=kint) :: NEIBPETOT
    integer(kind=kint) :: STACK_IMPORT(0:), NOD_IMPORT(:)
    real(kind=kreal) ::  X(:), WR(:)
    logical, intent(in) :: reverse_fl
    
    type(buf_type), pointer ::rbuf
    integer(kind=kint) :: inum, master, i,j,ierr
    
    integer, allocatable :: sta(:)

    allocate(sta(MPI_STATUS_SIZE))

    rbuf => comm % C_rbuf
    inum =  comm % comm_C % STACK_IMPORT(1)
    master= comm % comm_C % NEIBPE(1)
    if(.not. reverse_fl) then
       if(inum > 0) then
          call MPI_RECV(rbuf % buf(1), inum, MPI_DOUBLE_PRECISION, &
               & master, 0, comm % SOLVER_COMM, sta, ierr)
       end if

       !C rbuf -> WR
       call shuffle_wth_next_dst(   &
            & rbuf % msg_num,       &
            & rbuf % msg_index,     &
            & rbuf % buf,           &
            & rbuf % msg_next_dst,  &
            & STACK_IMPORT, WR, reverse_fl)

       !$omp parallel do private(j)
       do i = 1, STACK_IMPORT(NEIBPETOT)
          j = NOD_IMPORT(i)
          X(j) = WR(i)
       end do
       !$omp end parallel do
       
    else
       !$omp parallel do private(j)       
       do i = 1, STACK_IMPORT(NEIBPETOT)
          j = NOD_IMPORT(i)
          WR(i) = X(j) 
       end do
       !$end parallel do
       
       call shuffle_wth_next_dst(   &
            & rbuf % msg_num,       &
            & rbuf % msg_index,     &
            & rbuf % buf,           &
            & rbuf % msg_next_dst,  &
            & STACK_IMPORT, WR, reverse_fl)
       if(inum > 0) then
          call MPI_SEND(rbuf % buf(1), inum, MPI_DOUBLE_PRECISION, &
               & master, 0, comm % SOLVER_COMM, ierr)
       end if
       
    end if
    
  end subroutine recvfrommaster_pkt

  subroutine recv_all(comm, NEIBPETOT, STACK_EXPORT, NOD_EXPORT, X, &
       & reverse_fl)
    implicit none
    include 'mpif.h'
    
    integer(kind=kint) :: i, j, k, row, src, istart, inum
    integer(kind=kint) :: STACK_EXPORT(0:), NOD_EXPORT(:)
    integer(kind=kint) :: NEIBPETOT, my_rank, ierr
    real(kind=kreal)   :: X(:)
    logical, intent(in) :: reverse_fl
    
    type(comm_info),   target :: comm
    type(comm_table), pointer :: comm_C
    type(buf_type),    pointer :: rbuf

    integer, allocatable :: sta(:,:), req1(:)


    comm_C => comm % comm_C
    rbuf   => comm % C_rbuf
    
    my_rank = comm % my_rank

    allocate(sta(MPI_STATUS_SIZE, comm_C % NEIBPETOT), &
         &   req1(comm_C % NEIBPETOT))


    if(.not. reverse_fl) then
       req1 = MPI_REQUEST_NULL
       do i = 1,   comm_C % NEIBPETOT
          src =    comm_C % NEIBPE(i)
          istart = comm_C % STACK_IMPORT(i-1)+1
          inum =   comm_C % STACK_IMPORT(i)-comm_C % STACK_IMPORT(i-1)
          if(my_rank == src) then
             !C for my_rank part
             !$omp parallel do private(row)
             do j = 1, STACK_EXPORT(NEIBPETOT)
                row = NOD_EXPORT(j)
                rbuf % buf(istart+j-1) = X(row)
             end do
             !$omp end parallel do
          else
             if(inum > 0) then
                call MPI_IRECV(rbuf % buf(istart), inum,                 &
                     & MPI_DOUBLE_PRECISION, src, 0, comm % SOLVER_COMM, &
                     & req1(i), ierr)
             end if
          end if
          
       end do
       call MPI_WAITALL(comm_C % NEIBPETOT, req1, sta, ierr)
          
    else
       req1 = MPI_REQUEST_NULL
       do i = 1,   comm_C % NEIBPETOT
          src =    comm_C % NEIBPE(i)
          istart = comm_C % STACK_IMPORT(i-1)+1
          inum =   comm_C % STACK_IMPORT(i)-comm_C % STACK_IMPORT(i-1)
          if(my_rank == src) then
             !C for my_rank part
             do k = 1, NEIBPETOT
                !$omp parallel do private(row)
                do j = STACK_EXPORT(k-1)+1, STACK_EXPORT(k)
                   row = NOD_EXPORT(j)
                   X(row) = X(row)+rbuf % buf(istart+j-1)  
                end do
                !$omp end parallel do
             end do
          else
             if(inum > 0) then
                call MPI_ISEND(rbuf % buf(istart), inum,                  &
                     & MPI_DOUBLE_PRECISION, src, 0, comm % SOLVER_COMM, &
                     & req1(i),ierr)
             end if
          end if
       end do
       call MPI_WAITALL(comm_C % NEIBPETOT, req1, sta, ierr)
       
    end if
  end subroutine recv_all

  subroutine sendrecv_all(comm, axis, reverse_fl)
    implicit none
    include 'mpif.h'
    
    type(comm_info), target :: comm
    integer(kind=kint) :: axis

    logical, intent(in) :: reverse_fl
    
    type(comm_table), pointer :: comm_any
    type(buf_type), pointer :: sbuf, rbuf, older_rbuf
    
    integer(kind=kint) :: my_rank, i,j, pe_rank, ierr
    integer(kind=kint) :: s,e,s2,e2,inum,inum2
    
    integer(kind=kint), allocatable :: req1(:), req2(:), sta(:,:) 


    my_rank = comm % my_rank
    if(axis == 1) then
       older_rbuf => comm % C_rbuf
       sbuf       => comm % X_sbuf
       rbuf       => comm % X_rbuf
       comm_any   => comm % comm_X
    else if(axis == 2) then
       older_rbuf => comm % X_rbuf
       sbuf       => comm % Y_sbuf
       rbuf       => comm % Y_rbuf
       comm_any   => comm % comm_Y
    else if (axis == 3) then
       older_rbuf => comm % Y_rbuf
       sbuf       => comm % Z_sbuf
       rbuf       => comm % Z_rbuf
       comm_any   => comm % comm_Z
    end if

    allocate(req1(comm_any % NEIBPETOT), req2(comm_any % NEIBPETOT), &
         & sta(MPI_STATUS_SIZE, comm_any % NEIBPETOT))

    if(.not. reverse_fl) then
       !C older_rbuf -> sbuf
       call shuffle_wth_next_dst(        &
            & older_rbuf % msg_num,      &
            & older_rbuf % msg_index,    &
            & older_rbuf % buf,          &
            & older_rbuf % msg_next_dst, &
            & sbuf % msg_index,          &
            & sbuf % buf, .false.)


       req1 = MPI_REQUEST_NULL
       req2 = MPI_REQUEST_NULL
       do i = 1, comm_any % NEIBPETOT
          pe_rank = comm_any % NEIBPE(i)
          s = comm_any % STACK_IMPORT(i-1)+1
          e = comm_any % STACK_IMPORT(i)
          inum = e-s+1
          s2 = comm_any % STACK_EXPORT(i-1)+1
          e2 = comm_any % STACK_EXPORT(i)
          inum2 = e2-s2+1
          if(pe_rank == my_rank) then
             rbuf % buf(s:e) = sbuf % buf(s2:e2)
          else
             if(inum > 0) then
                call MPI_IRECV(rbuf % buf(s), inum,       &
                     & MPI_DOUBLE_PRECISION, pe_rank, 0,  &
                     & comm % SOLVER_COMM, req1(i), ierr)
             end if
             if(inum2 > 0) then
                call MPI_ISEND(sbuf % buf(s2), inum2,     &
                     & MPI_DOUBLE_PRECISION, pe_rank, 0,  &
                     & comm % SOLVER_COMM, req2(i), ierr)
             end if
          end if

       end do

       call MPI_WAITALL(comm_any % NEIBPETOT, req2, sta, ierr)
       call MPI_WAITALL(comm_any % NEIBPETOT, req1, sta, ierr)
       
    else
       !C -- reverse_fl== .true.
       req1 = MPI_REQUEST_NULL
       req2 = MPI_REQUEST_NULL
       do i = 1, comm_any % NEIBPETOT
          pe_rank = comm_any % NEIBPE(i)
          s = comm_any % STACK_IMPORT(i-1)+1
          e = comm_any % STACK_IMPORT(i)
          inum = e-s+1
          s2 = comm_any % STACK_EXPORT(i-1)+1
          e2 = comm_any % STACK_EXPORT(i)
          inum2 = e2-s2+1
          if(pe_rank == my_rank) then
             sbuf % buf(s2:e2) = rbuf % buf(s:e)  
          else
             if(inum > 0) then
                call MPI_ISEND(rbuf % buf(s), inum,       &
                     & MPI_DOUBLE_PRECISION, pe_rank, 0,  &
                     & comm % SOLVER_COMM, req1(i), ierr)
             end if
             if(inum2 > 0) then
                call MPI_IRECV(sbuf % buf(s2), inum2,     &
                     & MPI_DOUBLE_PRECISION, pe_rank, 0,  &
                     & comm % SOLVER_COMM, req2(i), ierr)
             end if
          end if

       end do

       call MPI_WAITALL(comm_any % NEIBPETOT, req2, sta, ierr)
       call MPI_WAITALL(comm_any % NEIBPETOT, req1, sta, ierr)

       !C older_rbuf <- sbuf
       call shuffle_wth_next_dst(        &
            & older_rbuf % msg_num,      &
            & older_rbuf % msg_index,    &
            & older_rbuf % buf,          &
            & older_rbuf % msg_next_dst, &
            & sbuf % msg_index,          &
            & sbuf % buf, reverse_fl)
       
    end if
    
  end subroutine sendrecv_all

  subroutine send_all(comm, NEIBPETOT, &
       & STACK_IMPORT, NOD_IMPORT, X, WR, reverse_fl)
    implicit none
    include 'mpif.h'
    
    type(comm_info), target :: comm
    integer(kind=kint) :: NEIBPETOT
    integer(kind=kint) :: STACK_IMPORT(0:), NOD_IMPORT(:)
    real(kind=kreal) ::  X(:), WR(:)
    
    logical, intent(in) :: reverse_fl
    integer(kind=kint) :: i,j,k,s,e,inum, pe_rank, my_rank


    type(comm_table), pointer :: comm_any
    type(buf_type),   pointer :: sbuf, older_rbuf


    integer(kind=kint), allocatable :: req1(:), sta(:,:) 
    integer :: ierr

    my_rank = comm % my_rank
    
    older_rbuf => comm % Z_rbuf
    sbuf       => comm % C_sbuf
    comm_any   => comm % comm_C


    allocate(req1(comm_any % NEIBPETOT),&
         & sta(MPI_STATUS_SIZE, comm_any % NEIBPETOT))


    if(.not. reverse_fl) then
       !C older_rbuf -> sbuf
       call shuffle_wth_next_dst(         &
            & older_rbuf % msg_num,       &
            & older_rbuf % msg_index,     &
            & older_rbuf % buf,           &
            & older_rbuf % msg_next_dst,  &
            & sbuf % msg_index,           &
            & sbuf % buf, reverse_fl)

       req1 = MPI_REQUEST_NULL
       do i = 1, comm_any % NEIBPETOT
          pe_rank = comm_any % NEIBPE(i)
          s = comm_any % STACK_EXPORT(i-1)+1
          e = comm_any % STACK_EXPORT(i)
          inum = e-s+1
          
          if(pe_rank == my_rank) then
             
             !C sbuf for my_rank -> WR
             call shuffle_wth_next_dst(         &
                  & comm % msg_num_master_self, &
                  & sbuf % msg_index,           &
                  & sbuf % buf,                 &
                  & sbuf % msg_next_dst,        &
                  & STACK_IMPORT, WR, reverse_fl)
             
             !$omp parallel do private(k)             
             do j = 1, STACK_IMPORT(NEIBPETOT)
                k = NOD_IMPORT(j)
                X(k) = WR(j)
             end do
             !$omp end parallel do
             
          else
             if(inum > 0) then
                call MPI_ISEND(sbuf % buf(s), inum,      &
                     & MPI_DOUBLE_PRECISION, pe_rank, 0, &
                     & comm % SOLVER_COMM, req1(i), ierr)
             end if
          end if
       end do
       call MPI_WAITALL(comm_any % NEIBPETOT, req1, sta, ierr)
    else
       !C-- reverse operation..--
       req1 = MPI_REQUEST_NULL
       do i = 1, comm_any % NEIBPETOT
          pe_rank = comm_any % NEIBPE(i)
          s = comm_any % STACK_EXPORT(i-1)+1
          e = comm_any % STACK_EXPORT(i)
          inum = e-s+1
          
          if(pe_rank == my_rank) then
             !$omp parallel do private(k)             
             do j = 1, STACK_IMPORT(NEIBPETOT)
                k = NOD_IMPORT(j)
                WR(j) = X(k) 
             end do
             !$omp end parallel do 
             
             !C sbuf for my_rank <- WR
             call shuffle_wth_next_dst(         &
                  & comm % msg_num_master_self, &
                  & sbuf % msg_index,           &
                  & sbuf % buf,                 &
                  & sbuf % msg_next_dst,        &
                  & STACK_IMPORT, WR, reverse_fl)
             
          else
             if(inum > 0) then
                call MPI_IRECV(sbuf % buf(s), inum,      &
                     & MPI_DOUBLE_PRECISION, pe_rank, 0, &
                     & comm % SOLVER_COMM, req1(i), ierr)
             end if
          end if
       end do

       call MPI_WAITALL(comm_any % NEIBPETOT, req1, sta, ierr)

       
       
       !C older_rbuf <- sbuf
       call shuffle_wth_next_dst(         &
            & older_rbuf % msg_num,       &
            & older_rbuf % msg_index,     &
            & older_rbuf % buf,           &
            & older_rbuf % msg_next_dst,  &
            & sbuf % msg_index,           &
            & sbuf % buf, reverse_fl)

    end if
  end subroutine send_all


  subroutine setup_comm_info(NEIBPETOT, NEIBPE, STACK_EXPORT, &
       & STACK_IMPORT, comm)
    implicit none
    include 'mpif.h'

    type(comm_info) :: comm
    integer(kind=kint) :: NEIBPETOT, STACK_EXPORT(0:), NEIBPE(:)
    integer(kind=kint) :: STACK_IMPORT(0:), my_rank

    integer(kind=kint) :: axis, ierr

    integer(kind=kint) :: size, c_size, x_size, y_size, z_size
    integer(kind=kint), pointer :: sz_msgs(:), c_sz_msgs(:)
    integer(kind=kint), pointer :: x_sz_msgs(:), y_sz_msgs(:)
    integer(kind=kint), pointer :: z_sz_msgs(:)

    my_rank = comm % my_rank

!C    C=16, X=4, Y=5, Z=6
!C    C=1; X=2; Y=2; Z=2

    call initialize_comm(comm)

    call sz_msgs_creation(my_rank, NEIBPETOT, NEIBPE, &
         & STACK_EXPORT, sz_msgs, size)
    if(mod(my_rank, comm % C)> 0) then
       call sendtomaster         (comm, sz_msgs, size)
       deallocate(sz_msgs)
       call recvfrommaster(comm, sz_msgs, size, NEIBPETOT, NEIBPE)  
       
    else
       
       call recv_all_and_comm_set(comm, sz_msgs, size, &
            & c_sz_msgs, c_size)
       deallocate(sz_msgs)

       axis = 1
       call sendrecv_all_and_comm_set(comm, c_sz_msgs, c_size, &
            & x_sz_msgs, x_size,  axis)

       axis = axis+1
       call sendrecv_all_and_comm_set(comm, x_sz_msgs, x_size, &
            & y_sz_msgs, y_size,  axis)

       axis = axis+1
       call sendrecv_all_and_comm_set(comm, y_sz_msgs, y_size, &
            & z_sz_msgs, z_size,  axis)

       call send_all_and_comm_set(comm, z_sz_msgs, z_size, &
            & sz_msgs, size)

       !C for master
       !C ordering of sz_msgs is recorded in 
       !C C_sbuf % msg_next_dst(:).
       !C master self msg size is recorded in comm % msg_num_master_self
       comm % msg_num_master_self = size
       allocate(comm % C_sbuf % msg_next_dst(size))
       call buf_ordering(my_rank, sz_msgs, size, &
            & comm % C_sbuf % msg_next_dst, &
            & NEIBPETOT, NEIBPE)

       deallocate(c_sz_msgs, x_sz_msgs, y_sz_msgs, z_sz_msgs)
    end if

    call check(my_rank, sz_msgs, size, NEIBPETOT, NEIBPE, &
         & STACK_IMPORT)
    
    deallocate(sz_msgs)    
    
  end subroutine setup_comm_info
  
  
  subroutine initialize_comm(comm)
    implicit none
    type(comm_info) :: comm
    
    nullify(comm % comm_X % NEIBPE,       &
         &  comm % comm_X % STACK_IMPORT, &
         &  comm % comm_X % STACK_EXPORT) 
    nullify(comm % comm_Y % NEIBPE,       &
         &  comm % comm_Y % STACK_IMPORT, &
         &  comm % comm_Y % STACK_EXPORT) 
    nullify(comm % comm_Z % NEIBPE,       &
         &  comm % comm_Z % STACK_IMPORT, &
         &  comm % comm_Z % STACK_EXPORT) 
    nullify(comm % comm_C % NEIBPE,       &
         &  comm % comm_C % STACK_IMPORT, &
         &  comm % comm_C % STACK_EXPORT) 

    nullify(comm % X_sbuf % msg_index,    &
         &  comm % X_sbuf % msg_next_dst, &
         &  comm % X_sbuf % buf) 
    nullify(comm % X_rbuf % msg_index,    &
         &  comm % X_rbuf % msg_next_dst, &
         &  comm % X_rbuf % buf) 

    nullify(comm % Y_sbuf % msg_index,    &
         &  comm % Y_sbuf % msg_next_dst, &
         &  comm % Y_sbuf % buf) 
    nullify(comm % Y_rbuf % msg_index,    &
         &  comm % Y_rbuf % msg_next_dst, &
         &  comm % Y_rbuf % buf) 

    nullify(comm % Z_sbuf % msg_index,    &
         &  comm % Z_sbuf % msg_next_dst, &
         &  comm % Z_sbuf % buf) 
    nullify(comm % Z_rbuf % msg_index,    &
         &  comm % Z_rbuf % msg_next_dst, &
         &  comm % Z_rbuf % buf) 

    nullify(comm % C_sbuf % msg_index,    &
         &  comm % C_sbuf % msg_next_dst, &
         &  comm % C_sbuf % buf) 
    nullify(comm % C_rbuf % msg_index,    &
         &  comm % C_rbuf % msg_next_dst, &
         &  comm % C_rbuf % buf) 

  end subroutine initialize_comm
  
  subroutine check(my_rank, sz_msgs, size, NEIBPETOT, NEIBPE, &
       & STACK_IMPORT)
    implicit none
    include 'mpif.h'

    integer(kind=kint) :: sz_msgs(:), NEIBPE(:), STACK_IMPORT(0:)
    integer(kind=kint) :: size, NEIBPETOT, my_rank
    
    integer(kind=kint) :: i,j,src

    j = 0
    do i = 1, NEIBPETOT
       if(STACK_IMPORT(i)-STACK_IMPORT(i-1) > 0) j = j+1
    end do

    if(size /= j) write(*,*) "error! not equal to import table pes"
    
    do i = 1, size
       if(sz_msgs(i*3-1) /= my_rank) &
            & write(*,*) "wrong packet info comes.."
       src = sz_msgs(i*3-2)
       
       do j = 1, NEIBPETOT
          if(src == NEIBPE(j)) exit
       end do
       if(j > NEIBPETOT)  write(*,*) "wrong info comes from ", src
       
       if(sz_msgs(i*3) /= STACK_IMPORT(j)-STACK_IMPORT(j-1))  &
            & write(*,*) "wrong size msg comes.sizes are", &
            &  sz_msgs(i*3), &
            & STACK_IMPORT(j)-STACK_IMPORT(j-1)
       
    end do
    
  end subroutine check



  subroutine shuffle_wth_next_dst(msg_num, cmsg_index, cmsg, &
       & msg_next_dst, nmsg_index, nmsg, reverse_fl)
    implicit none
    include 'mpif.h'

    real(kind=kreal) :: cmsg(:), nmsg(:)
    integer(kind=kint) :: msg_num, cmsg_index(0:), nmsg_index(0:)
    integer(kind=kint) :: msg_next_dst(:), next_dst
    logical, intent(in) :: reverse_fl
    integer(kind=kint) :: i, s, e, cs, ce
    
    !$omp parallel do private(next_dst, s,e,cs,ce)
    do i = 1, msg_num
       next_dst = msg_next_dst(i)
       s = nmsg_index(next_dst-1)+1
       e = nmsg_index(next_dst)
       cs = cmsg_index(i-1)+1
       ce = cmsg_index(i)
       if(.not.reverse_fl) then
          nmsg(s:e)   = cmsg(cs:ce)
       else
          cmsg(cs:ce) = nmsg(s:e)  
       end if
       if(e-s /= ce-cs) stop 'error in shuffle_wth_next_dst'
    end do
    !$omp end parallel do
    
  end subroutine shuffle_wth_next_dst



  !C the same code structureas sendrecv_all_and_comm_set()
  subroutine  send_all_and_comm_set(comm, z_sz_msgs, z_size,  &
       & sz_msgs, size)
    implicit none
    include 'mpif.h'

    type(comm_info) :: comm
    integer(kind=kint) :: my_rank, C
    integer(kind=kint) :: z_sz_msgs(:), z_size, size
    integer(kind=kint), pointer :: sz_msgs(:)
    integer(kind=kint), pointer :: sz_msgs_to_send(:)
    integer(kind=kint), pointer :: index_sz_msgs(:)
    
    integer(kind=kint), allocatable :: req2(:), sta(:,:)
    
    integer(kind=kint) :: i, inum, inum2, istart2
    integer(kind=kint) :: pe_rank, ierr

    my_rank = comm % my_rank
    C = comm % C
    
    allocate(comm % Z_rbuf % msg_next_dst(comm % Z_rbuf % msg_num))
    
    call buf_shuffle(my_rank, z_sz_msgs, z_size, sz_msgs_to_send, &
         & index_sz_msgs, comm % Z_rbuf % msg_next_dst, 1, C)

    allocate(req2(C), sta(MPI_STATUS_SIZE, C))

    !C change index to size
    do i = C, 1, -1
       index_sz_msgs(i) = index_sz_msgs(i) - index_sz_msgs(i-1)
    end do
    !C sendrecv size
    req2 = MPI_REQUEST_NULL
    do i = 1, C
       pe_rank = my_rank/C*C + i-1
       
       inum = index_sz_msgs(i)-index_sz_msgs(i-1)
       
       if(pe_rank == my_rank) then
          size = index_sz_msgs(i)
          allocate(sz_msgs(3*size))
       else
!C          if(my_rank < pe_rank) write(*,*) "(", my_rank, pe_rank,")"
          call MPI_ISEND(index_sz_msgs(i), 1, MPI_INTEGER, pe_rank, 0, &
               & comm % SOLVER_COMM, req2(i), ierr)
       end if
       
    end do
    call MPI_WAITALL (C, req2, sta, ierr)
    !C change size to index
    do i = 2, C
       index_sz_msgs(i) = index_sz_msgs(i) + index_sz_msgs(i-1)
    end do



    !C sendrecv size msgs.
    req2 = MPI_REQUEST_NULL
    do i = 1, C
       pe_rank = my_rank/C*C + (i-1)
       istart2 = index_sz_msgs(i-1)*3+1
       inum2   = (index_sz_msgs(i)-index_sz_msgs(i-1))*3

       if(pe_rank == my_rank) then
          if(3*size /= inum2) &
               & stop 'unexpected error in send_all_and_comm_set()'
          sz_msgs(1:3*size) = sz_msgs_to_send(istart2:istart2+inum2-1)
       else
          if(inum2 > 0) then
             call MPI_ISEND(sz_msgs_to_send(istart2), inum2, &
                  & MPI_INTEGER, pe_rank, 0, comm % SOLVER_COMM, &
                  & req2(i), ierr)
          end if
       end if

    end do
    call MPI_WAITALL (C, req2, sta, ierr)
    
    allocate(comm % comm_C % STACK_EXPORT(0:C))
    call set_comm_buf_frm_szmsgs(comm % C_sbuf,        &
         & comm % comm_C % STACK_EXPORT,               &
         & sz_msgs_to_send, index_sz_msgs, C, my_rank)

    deallocate(sz_msgs_to_send, index_sz_msgs)
    
    
    
  end subroutine send_all_and_comm_set

  subroutine sendrecv_all_and_comm_set(comm,  c_sz_msgs, c_size, &
       & x_sz_msgs, x_size, axis)
    implicit none
    include 'mpif.h'

    type(comm_info), target :: comm
    type(comm_table), pointer :: comm_any
    type(buf_type), pointer :: sbuf, rbuf, older_rbuf
    integer(kind=kint) :: my_rank, C, X, Y, Z, axis
    integer(kind=kint) :: c_sz_msgs(:), c_size, x_size
    integer(kind=kint), pointer :: x_sz_msgs(:), msg_next_dst(:)
    
    integer(kind=kint), pointer :: sz_msgs_to_send(:)
    integer(kind=kint), pointer :: index_sz_msgs(:)
    
    integer(kind=kint), allocatable :: index_recv_msgs(:)
    
    integer(kind=kint), allocatable :: req1(:), req2(:), sta(:,:)
    
    integer(kind=kint) :: i,j,k,inum, istart, inum2, istart2
    integer(kind=kint) :: a,b,pe_rank, h_rank, ierr
    
    !C comm exchange among my_rank/ab*ab+a*[0,..,b-1]

    my_rank = comm % my_rank
    C = comm % C; X = comm % X; Y = comm % Y; Z = comm % Z
    
    if(axis == 1)then
       a=C;    b=X;  
       h_rank = my_rank/(C*X)*(C*X)
       
       older_rbuf => comm % C_rbuf
       sbuf       => comm % X_sbuf
       rbuf       => comm % X_rbuf
       comm_any   => comm % comm_X
       
    else if(axis == 2) then
       a=C*X;  b=Y;  
       h_rank = C*(mod(my_rank/C, X) + mod(my_rank/(C*X*Y), Z)*X*Y)
       older_rbuf => comm % X_rbuf
       sbuf       => comm % Y_sbuf
       rbuf       => comm % Y_rbuf
       comm_any   => comm % comm_Y

    else if(axis == 3) then
       a=C*X*Y;b=Z;  
       h_rank = C*(mod(my_rank/C, X) + mod(my_rank/(C*X),Y)*X)
       older_rbuf => comm % Y_rbuf
       sbuf       => comm % Z_sbuf
       rbuf       => comm % Z_rbuf
       comm_any   => comm % comm_Z

    end if

    allocate(older_rbuf % msg_next_dst(older_rbuf % msg_num))

    call buf_shuffle(my_rank, c_sz_msgs, c_size, &
         & sz_msgs_to_send, index_sz_msgs, &
         & older_rbuf % msg_next_dst, a, b)
    
    allocate(index_recv_msgs(0:b))
    index_recv_msgs = 0
    allocate(req1(b), req2(b), sta(MPI_STATUS_SIZE, b))

    !C change index to size
    do i=b,1,-1
       index_sz_msgs(i) = index_sz_msgs(i) - index_sz_msgs(i-1)
    end do

    
    !C sendrecv size
    req1 = MPI_REQUEST_NULL
    req2 = MPI_REQUEST_NULL
    do i = 1, b
       pe_rank = h_rank + (i-1)*a
       
       if(pe_rank == my_rank) then
          index_recv_msgs(i) = index_sz_msgs(i)
       else

!C          if(my_rank < pe_rank) write(*,*) "(", my_rank, pe_rank,")"
          call MPI_IRECV(index_recv_msgs(i), 1, MPI_INTEGER, &
               & pe_rank, 0, comm % SOLVER_COMM, req1(i), ierr)
          call MPI_ISEND(index_sz_msgs(i),   1, MPI_INTEGER, &
               & pe_rank, 0, comm % SOLVER_COMM, req2(i), ierr)
       end if
       
    end do
    call MPI_WAITALL (b, req2, sta, ierr)
    call MPI_WAITALL (b, req1, sta, ierr)

    !C change size to index
    do i=2,b
       index_sz_msgs(i) = index_sz_msgs(i) + index_sz_msgs(i-1)
    end do


    do i = 1, b
       index_recv_msgs(i) = index_recv_msgs(i) + index_recv_msgs(i-1)
    end do
    x_size = index_recv_msgs(b)
    allocate(x_sz_msgs(x_size*3))

    !C sendrecv size msgs.
    req1 = MPI_REQUEST_NULL
    req2 = MPI_REQUEST_NULL
    do i = 1, b
       pe_rank = h_rank + (i-1)*a
       istart = index_recv_msgs(i-1)*3+1
       inum   = (index_recv_msgs(i)-index_recv_msgs(i-1))*3
       istart2 = index_sz_msgs(i-1)*3+1
       inum2   = (index_sz_msgs(i)-index_sz_msgs(i-1))*3

       
       if(pe_rank == my_rank) then
          x_sz_msgs(istart:istart+inum-1) = &
               & sz_msgs_to_send(istart2:istart2+inum2-1)
       else
          if(inum > 0) then
             call MPI_IRECV(x_sz_msgs(istart),         inum,     &
                  & MPI_INTEGER, pe_rank, 0, comm % SOLVER_COMM, &
                  & req1(i), ierr)
          end if
          if(inum2 > 0) then
             call MPI_ISEND(sz_msgs_to_send(istart2), inum2,     &
                  & MPI_INTEGER, pe_rank, 0, comm % SOLVER_COMM, &
                  & req2(i), ierr)
          end if
       end if

    end do
    call MPI_WAITALL (b, req2, sta, ierr)
    call MPI_WAITALL (b, req1, sta, ierr)

       

    !C setting comm_any % NEIBPE
    comm_any % NEIBPETOT = b
    allocate(comm_any % NEIBPE(b), comm_any % STACK_IMPORT(0:b))
    allocate(comm_any % STACK_EXPORT(0:b))
    do i = 1, b
       comm_any % NEIBPE(i) = h_rank + (i-1)*a
    end do
    
    call set_comm_buf_frm_szmsgs(rbuf, comm_any % STACK_IMPORT, &
         & x_sz_msgs,       index_recv_msgs, b, my_rank)

    call set_comm_buf_frm_szmsgs(sbuf, comm_any % STACK_EXPORT, &
         & sz_msgs_to_send, index_sz_msgs,   b, my_rank)
    
    deallocate(sz_msgs_to_send, index_sz_msgs)
    
  end subroutine sendrecv_all_and_comm_set
  
  subroutine buf_ordering(my_rank, sz_msgs, size, msg_next_dst, &
       & NEIBPETOT, NEIBPE)
    implicit none
    
    integer(kind=kint) :: my_rank, sz_msgs(:), size
    integer(kind=kint) :: msg_next_dst(:)
    integer(kind=kint) :: NEIBPETOT, NEIBPE(:)
    
    integer(kind=kint) :: i,j,src

    !C this double loop may need modification!!
    do i = 1, size
       src = sz_msgs(i*3-2)
       do j = 1, NEIBPETOT
          if(NEIBPE(j) == src) exit
       end do
       if(j > NEIBPETOT .and. NEIBPETOT > 0) then
          stop 'error msgs from unknown rank'
       end if
       
       msg_next_dst(i) = j
    end do
    
  end subroutine buf_ordering
  
  subroutine buf_shuffle(my_rank, c_sz_msgs, c_size, &
       & sz_msgs_to_send, index_sz_msgs, msg_next_dst, a, b)
    implicit none

    integer(kind=kint), intent(in) :: my_rank, a, b
    integer(kind=kint) :: c_sz_msgs(:), c_size, msg_next_dst(:)

    integer(kind=kint), pointer :: sz_msgs_to_send(:)
    integer(kind=kint), pointer :: index_sz_msgs(:)
    
    integer(kind=kint) :: i,j,k, src, dest, sz

    allocate(index_sz_msgs(0:b+1))


    index_sz_msgs = 0
    do i = 1, c_size
       src  = c_sz_msgs(i*3-2)
       dest = c_sz_msgs(i*3-1)
       sz   = c_sz_msgs(i*3) 
       
       j = mod(dest/a, b)+1
       
       index_sz_msgs(j+1) = index_sz_msgs(j+1) + 1
       
    end do
    
    do i = 2, b+1
       index_sz_msgs(i) = index_sz_msgs(i) + index_sz_msgs(i-1) 
    end do
    
    allocate(sz_msgs_to_send(index_sz_msgs(b+1)*3))
    
    do i = 1, c_size
       src  = c_sz_msgs(i*3-2)
       dest = c_sz_msgs(i*3-1)
       sz   = c_sz_msgs(i*3) 
       
       j = mod(dest/a, b)+1
       k = index_sz_msgs(j)+1; index_sz_msgs(j) = k
       
       msg_next_dst(i) = k
       
       sz_msgs_to_send(3*k-2:3*k) = c_sz_msgs(3*i-2:3*i)
       
    end do
       
  end subroutine buf_shuffle

  subroutine recvfrommaster(comm, sz_msgs, size, NEIBPETOT, NEIBPE)  
    
    implicit none
    include 'mpif.h'

    integer :: master, my_rank, C, ierr, count
    integer(kind=kint), pointer :: sz_msgs(:)
    integer(kind=kint) :: size
    integer(kind=kint), allocatable :: index_sz_msgs(:)
    type(comm_info) :: comm

    integer(kind=kint) :: NEIBPETOT, NEIBPE(:)

    integer, allocatable :: sta(:)

    allocate(sta(MPI_STATUS_SIZE))
    my_rank = comm % my_rank
    C = comm % C
    
    master = (my_rank/C)*C

    call MPI_RECV(size, 1, MPI_INTEGER, master, 0, &
         & comm % SOLVER_COMM, sta, ierr)

    allocate(index_sz_msgs(0:1))
    index_sz_msgs(0) = 0;     index_sz_msgs(1) = size; 
    
    allocate(sz_msgs(3*size))
    count = size*3
    if(count > 0) then
       call MPI_RECV(sz_msgs(1), count, MPI_INTEGER, master, 0, &
            & comm % SOLVER_COMM, sta, ierr)
    end if
    
    
    deallocate(sta)

    
    !C set comm_C and C_rbuf
    allocate(comm % comm_C % STACK_IMPORT(0:1))
    call set_comm_buf_frm_szmsgs(comm % C_rbuf,  &
         & comm % comm_C % STACK_IMPORT,         &
         & sz_msgs, index_sz_msgs, 1, my_rank)


    !C set C_rbuf % msg_next_dst
    allocate(comm % C_rbuf % msg_next_dst(size))
    call buf_ordering(my_rank, sz_msgs, size, &
         & comm % C_rbuf % msg_next_dst, &
       & NEIBPETOT, NEIBPE)

    
  end subroutine recvfrommaster


  
  subroutine sendtomaster(comm, sz_msgs, size)
    implicit none
    include 'mpif.h'

    integer :: master, my_rank, C, ierr, count
    integer(kind=kint) :: sz_msgs(:), size
    integer(kind=kint), allocatable :: index_sz_msgs(:)
    type(comm_info) :: comm
    
    my_rank = comm % my_rank
    C = comm % C
    
    master = (my_rank/C)*C
    
    allocate(index_sz_msgs(0:1))
    index_sz_msgs(0) = 0;     index_sz_msgs(1) = size; 
    
    call MPI_SEND(size, 1, MPI_INTEGER, master, 0, &
         & comm % SOLVER_COMM, ierr)
    
    if(size > 0) then
       count = size*3
       call MPI_SEND(sz_msgs(1), count, MPI_INTEGER, master, 0, &
            & comm % SOLVER_COMM, ierr)
    end if

    !C set comm_C and C_sbuf
    comm % comm_C % NEIBPETOT = 1
    allocate(comm % comm_C % NEIBPE(1))
    allocate(comm % comm_C % STACK_EXPORT(0:1))
    comm % comm_C % NEIBPE(1) = master
    call set_comm_buf_frm_szmsgs(comm % C_sbuf,  &
         & comm % comm_C % STACK_EXPORT,         &
         & sz_msgs, index_sz_msgs, 1, my_rank)
    
  end subroutine sendtomaster
  
  subroutine recv_all_and_comm_set(comm, sz_msgs, size, &
       & c_sz_msgs, c_size)
    implicit none
    include 'mpif.h'

    type(comm_info) :: comm
    integer(kind=kint) :: C, my_rank
    integer :: ierr
    integer, allocatable :: sta(:)
    integer(kind=kint) :: sz_msgs(:), size, c_size
    integer(kind=kint), pointer :: c_sz_msgs(:)
    integer(kind=kint) :: i, child, istart, inum
    
    integer(kind=kint), allocatable :: size_array(:)
    
    C = comm % C
    my_rank = comm % my_rank
    

    allocate(size_array(0:C), sta(MPI_STATUS_SIZE))
    
    do i = 2, C
       child = (my_rank/C)*C+i-1
       call MPI_RECV(size_array(i), 1, MPI_INTEGER, child, 0, &
            & comm % SOLVER_COMM, sta, ierr)
    end do
    
    size_array(1) = size
    size_array(0) = 0
    do i = 1, C
       size_array(i) = size_array(i-1) + size_array(i)
    end do
    
    allocate(c_sz_msgs(size_array(C) * 3))
    c_size = size_array(C)
    
    do i = 2, C
       if(size_array(i)> size_array(i-1)) then
          child  = (my_rank/C)*C+i-1
          istart = size_array(i-1)*3+1
          inum   = (size_array(i) - size_array(i-1))*3
          if(inum > 0) then
             call MPI_RECV(c_sz_msgs(istart), inum, MPI_INTEGER, &
                  & child, 0, comm % SOLVER_COMM, sta, ierr)
          end if
                 
       end if
    end do
    

    !C copy my_rank's sz_msgs
    do i = 1, size
       c_sz_msgs(i*3-2) = sz_msgs(i*3-2)
       c_sz_msgs(i*3-1) = sz_msgs(i*3-1)
       c_sz_msgs(i*3)   = sz_msgs(i*3)
    end do

    !C setting comm_C
    comm % comm_C % NEIBPETOT = C
    allocate(comm % comm_C % NEIBPE(C))
    allocate(comm % comm_C % STACK_IMPORT(0:C))
    do i = 1, C
       comm % comm_C % NEIBPE(i)=(my_rank/C)*C+i-1
    end do

    call set_comm_buf_frm_szmsgs(comm % C_rbuf, &
         & comm % comm_C % STACK_IMPORT,        &
         & c_sz_msgs, size_array, C, my_rank)

  end subroutine recv_all_and_comm_set
  
  !C sz_msgs has a set of 3-element values
  !C size_array(0:) indicate start position of received msg from 
  !C each PE.
  !C pe_rank_size shows the number of ranks to communicate. 
  !C It includes my_rank.
  subroutine set_comm_buf_frm_szmsgs(buf, stack_port, sz_msgs,  &
       & idx_arry_for_sz_msgs, pe_rank_size, my_rank)
    implicit none

    
    type(buf_type) :: buf
    integer(kind=kint) :: stack_port(0:)
    integer(kind=kint) :: sz_msgs(:), size
    integer(kind=kint) :: idx_arry_for_sz_msgs(0:)
    integer(kind=kint), intent(in) :: pe_rank_size
    
    integer(kind=kint) :: i,j,my_rank
    
    size = idx_arry_for_sz_msgs(pe_rank_size)

    !C for buf
    buf % msg_num = size
    
    allocate(buf % msg_index(0:size))
    buf % msg_index(0) = 0
    
    do i = 1, size
       buf % msg_index(i) = sz_msgs(i*3)
    end do

    do i = 1, size
       buf % msg_index(i) = buf % msg_index(i-1) + buf % msg_index(i)
    end do
    buf % msg_sz = buf % msg_index(size)

    allocate(buf % buf(buf % msg_sz))
    
    !C for stack_port
    stack_port(0) = 0
    do i = 1, pe_rank_size
       j = idx_arry_for_sz_msgs(i)
       stack_port(i)=buf % msg_index(j)
    end do
    
  end subroutine set_comm_buf_frm_szmsgs
  
  subroutine sz_msgs_creation(my_rank, NEIBPETOT, NEIBPE, &
       & STACK_EXPORT, sz_msgs, size)
    implicit none

    integer(kind=kint) :: my_rank
    integer(kind=kint) :: NEIBPETOT, STACK_EXPORT(0:), size
    integer(kind=kint) :: i,j, NEIBPE(:)
    integer(kind=kint), pointer :: sz_msgs(:)

    allocate(sz_msgs(NEIBPETOT*3))
    size = 0
    do i = 1, NEIBPETOT
       if(STACK_EXPORT(i)-STACK_EXPORT(i-1)>0) then
          size = size+1 
          sz_msgs(size*3-2) = my_rank
          sz_msgs(size*3-1) = NEIBPE(i)
          sz_msgs(size*3)   = STACK_EXPORT(i)-STACK_EXPORT(i-1)
       end if
    end do
    
  end subroutine sz_msgs_creation
end module axis_comm
