module blocking_mat
contains
  subroutine blocking_data( &
       & N_in, NP_in, NPL_in, NPU_in, D_in, AL_in, INL_in, IAL_in,        &
       & AU_in, INU_in, IAU_in, NEIBPETOT_in, NEIBPE_in,                  &
       & STACK_IMPORT_in, NOD_IMPORT_in, STACK_EXPORT_in, NOD_EXPORT_in,  &
       & N, NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU,  &
       & NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
       & STACK_EXPORT, NOD_EXPORT,  &
       & my_rank, SOLVER_COMM, NB &
       )

    implicit none
    include 'precision.inc'
    
    integer(kind=kint )                  ::  N, NP, NPL, NPU, NB

    real   (kind=kreal), dimension(:,:,:), pointer ::  AU, AL
    real   (kind=kreal), dimension(:,:,:), pointer ::  D

    integer(kind=kint ), dimension(:), pointer ::  INU, INL
    integer(kind=kint ), dimension(:), pointer ::  IAU, IAL


    integer(kind=kint )                  :: NEIBPETOT
    integer(kind=kint ), pointer         :: NEIBPE      (:)
    integer(kind=kint ), pointer         :: STACK_IMPORT(:)
    integer(kind=kint ), pointer         :: NOD_IMPORT  (:)
    integer(kind=kint ), pointer         :: STACK_EXPORT(:)
    integer(kind=kint ), pointer         :: NOD_EXPORT  (:)

    !C--
    integer(kind=kint )                  ::  N_in, NP_in, NPL_in, NPU_in

    real   (kind=kreal), pointer ::  AU_in(:), AL_in(:)
    real   (kind=kreal), pointer ::  D_in(:)

    integer(kind=kint ), pointer ::  INU_in(:), INL_in(:)
    integer(kind=kint ), pointer ::  IAU_in(:), IAL_in(:)

    integer(kind=kint )                  :: NEIBPETOT_in
    integer(kind=kint ), pointer         :: NEIBPE_in      (:)
    integer(kind=kint ), pointer         :: STACK_IMPORT_in(:)
    integer(kind=kint ), pointer         :: NOD_IMPORT_in  (:)
    integer(kind=kint ), pointer         :: STACK_EXPORT_in(:)
    integer(kind=kint ), pointer         :: NOD_EXPORT_in  (:)

    integer :: SOLVER_COMM, my_rank

    integer(kind=kint) :: i,j,k, col, col_block, deltan, num_of_blk
    integer(kind=kint),allocatable :: temp(:), temp2(:)

    integer(kind=kint), allocatable :: n_np_blk_num(:), n_np_NB_mod(:)

    !C NP_in is assumed to be a multiple of NB.

    !C error check
!!$    if(NEIBPETOT_in > 0) stop 'error in blocking_data(): distributed matrices'
!!$    if(my_rank/=0 .and. N_in > 0)  stop 'error in blocking_data(): distributed matrices'
!!$    if(my_rank==0 .and. mod(N_in, NB)/=0)  stop 'error in blocking_data(): N_in is not a multiple of NB'
!!$    if(my_rank==0 .and. N_in/=NP_in)  stop 'error in blocking_data(): N_in /= NP_in on rank 0'

!!$    if(my_rank == 0) then


    allocate(n_np_blk_num(NP_in-N_in), n_np_NB_mod(NP_in-N_in))

    call create_n_np_arrys(num_of_blk,n_np_NB_mod, n_np_blk_num, N_in,&
         & NP_in, NEIBPETOT_in, NEIBPE_in, STACK_IMPORT_in,           &
         & NOD_IMPORT_in, STACK_EXPORT_in, NOD_EXPORT_in, SOLVER_COMM,&
         & my_rank, NB)

    call comm_table_update(num_of_blk,n_np_NB_mod, n_np_blk_num, N_in,&
         & NP_in, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,        &
         & STACK_EXPORT, NOD_EXPORT, NEIBPETOT_in, NEIBPE_in,         &
         & STACK_IMPORT_in, NOD_IMPORT_in, STACK_EXPORT_in,           &
         & NOD_EXPORT_in, SOLVER_COMM, my_rank, NB)



    !C blocking
    N  = N_in / NB


    !C N is modified to multiples of NB, and (NP-N) is also changed to multiples of NB.
    deltan = 0
    if(mod(N_in, NB) > 0) deltan  = 1


    !C NP is updated to correct number
    NP = N+deltan+num_of_blk

    !C D(NP) and D needs values on boundary.
    allocate(INL(0:N+deltan), INU(0:N+deltan), D(NB,NB,NP), temp(NP))

    INL(:) = 0; INU(:) = 0; NPL = 0; NPU = 0

    !C count up lower and upper part.
    !C and set INL(:), INU(:) pointing to start index of each row
    temp = 0
    DO i = 1, N
       !C-- nonzero counting part for i-th NB block row 
       temp(i) = i
       DO j = 0, NB-1
          DO k = INL_in(NB*(i-1)+j)+1, INL_in(NB*(i-1)+j+1)
             col = IAL_in(k)
             col_block = (col-1)/NB+1

             if(temp(col_block) /= i) then
                NPL = NPL + 1
                temp(col_block) = i
                if(i < N+deltan) INL(i+1) = INL(i+1)+1
             end if
          END DO

          DO k = INU_in(NB*(i-1)+j)+1, INU_in(NB*(i-1)+j+1)
             col = IAU_in(k)

             if(col > N_in) then
                col_block = n_np_blk_num(col-N_in)
             else
                col_block = (col-1)/NB+1
             end if

             if(temp(col_block) /= i) then
                NPU = NPU + 1
                temp(col_block) = i
                if(i < N+deltan) INU(i+1) = INU(i+1)+1
             end if
          END DO
       END DO
       !C--
    END DO

    !C this case is mod(N_in, NB) > 0
    if(deltan > 0) then
       i = N+1
       !C-- nonzero counting part for last NB block row
       temp(i) = i
       DO j = 0, mod(N_in, NB)-1
          DO k = INL_in(NB*(i-1)+j)+1, INL_in(NB*(i-1)+j+1)
             col = IAL_in(k)
             col_block = (col-1)/NB+1

             if(temp(col_block) /= i) then
                NPL = NPL + 1
                temp(col_block) = i
                if(i < N+deltan) INL(i+1) = INL(i+1)+1
             end if

          END DO
          DO k = INU_in(NB*(i-1)+j)+1, INU_in(NB*(i-1)+j+1)
             col = IAU_in(k)

             if(col > N_in) then
                col_block = n_np_blk_num(col-N_in)
             else
                col_block = (col-1)/NB+1
             end if

             if(temp(col_block) /= i) then
                NPU = NPU + 1
                temp(col_block) = i
                if(i < N+deltan) INU(i+1) = INU(i+1)+1
             end if
          END DO
       END DO
       !C--
    END if
    DO i = 1, N+deltan
       INL(i) = INL(i-1)+INL(i)
       INU(i) = INU(i-1)+INU(i)
    END DO



    allocate(IAL(NPL), IAU(NPU), AL(NB,NB,NPL), AU(NB,NB,NPU))

    !C record index of IA?(:) for each column.
    allocate(temp2(NP))



    !C record the AL, IAL. D.
    temp = 0; temp2 = 0
    DO i = 1, N

       !C-- nonzero assigning part for ith NB block row
       D(:,:,i) = 0

       temp(i) = i  !C record diagonal block in temp
       DO j = 0, NB-1

          D(j+1,j+1,i) = D_in(NB*(i-1)+j+1)

          DO k = INL_in(NB*(i-1)+j)+1, INL_in(NB*(i-1)+j+1)
             col = IAL_in(k); col_block = (col-1)/NB+1
             if(col_block == i) then
                D(j+1, mod(col-1, NB)+1, i) = AL_in(k)
             else
                if(temp(col_block) /= i) then
                   temp(col_block) = i
                   INL(i) = INL(i)+1
                   temp2(col_block) = INL(i)
                   AL(:,:,temp2(col_block)) = 0
                   IAL(temp2(col_block)) = col_block
                end if
                AL(j+1, mod(col-1,NB)+1, temp2(col_block)) = AL_in(k)
             end if
          END DO
          DO k = INU_in(NB*(i-1)+j)+1, INU_in(NB*(i-1)+j+1)
             col = IAU_in(k)

             if(col > N_in) then
                col_block = n_np_blk_num(col-N_in)
             else
                col_block = (col-1)/NB+1
             end if

             if(col_block == i) then
                D(j+1, mod(col-1, NB)+1, i) = AU_in(k)
             else
                if(temp(col_block) /= i) then
                   temp(col_block) = i
                   INU(i) = INU(i)+1
                   temp2(col_block) = INU(i)
                   AU(:,:,temp2(col_block)) = 0
                   IAU(temp2(col_block)) = col_block
                end if
                if(col > N_in) then
                   AU(j+1, n_np_NB_mod(col-N_in), temp2(col_block)) = AU_in(k)
                else
                   AU(j+1, mod(col-1,NB)+1, temp2(col_block)) = AU_in(k)
                end if
             end if
          END DO
       END DO
       !C--

    END DO


    !C this case is mod(N_in, NB) > 0
    if(deltan > 0) then
       i = N+1

       !C-- nonzero assigning part for last NB block row
       D(:,:,i) = 0

       temp(i) = i  !C record diagonal block in temp
       DO j = 0, mod(N_in, NB)-1

          D(j+1,j+1,i) = D_in(NB*(i-1)+j+1)

          DO k = INL_in(NB*(i-1)+j)+1, INL_in(NB*(i-1)+j+1)
             col = IAL_in(k); col_block = (col-1)/NB+1
             if(col_block == i) then
                D(j+1, mod(col-1, NB)+1, i) = AL_in(k)
             else
                if(temp(col_block) /= i) then
                   temp(col_block) = i
                   INL(i) = INL(i)+1
                   temp2(col_block) = INL(i)
                   AL(:,:,temp2(col_block)) = 0
                   IAL(temp2(col_block)) = col_block
                end if
                AL(j+1, mod(col-1,NB)+1, temp2(col_block)) = AL_in(k)
             end if
          END DO
          DO k = INU_in(NB*(i-1)+j)+1, INU_in(NB*(i-1)+j+1)
             col = IAU_in(k)
             if(col > N_in) then
                col_block = n_np_blk_num(col-N_in)
             else
                col_block = (col-1)/NB+1
             end if

             if(col_block == i) then
                D(j+1, mod(col-1, NB)+1, i) = AU_in(k)
             else
                if(temp(col_block) /= i) then
                   temp(col_block) = i
                   INU(i) = INU(i)+1
                   temp2(col_block) = INU(i)
                   AU(:,:,temp2(col_block)) = 0
                   IAU(temp2(col_block)) = col_block
                end if

                if(col > N_in) then
                   AU(j+1, n_np_NB_mod(col-N_in), temp2(col_block)) = AU_in(k)
                else
                   AU(j+1, mod(col-1,NB)+1, temp2(col_block)) = AU_in(k)
                end if

             end if
          END DO
       END DO
       !C--

       DO j = mod(N_in, NB), NB-1
          D(j+1,j+1,i) = 1.0
       END DO

    end if


    deallocate(temp, temp2)

    N = N+deltan


    deallocate(INL_in, INU_in, IAL_in, IAU_in, AL_in, AU_in, D_in)

    !C deallocate old data
    deallocate(NEIBPE_in, STACK_IMPORT_in, STACK_EXPORT_in)
    deallocate(NOD_IMPORT_in, NOD_EXPORT_in)

!!$       !C commulnication tables are allocated
!!$       NEIBPETOT = 0
!!$       allocate(NEIBPE(0))
!!$       allocate(STACK_EXPORT(0:NEIBPETOT),STACK_IMPORT(0:NEIBPETOT))
!!$       STACK_EXPORT(0) = 0; STACK_IMPORT(0) = 0
!!$       allocate(NOD_EXPORT(0), NOD_IMPORT(0))

  end subroutine blocking_data
  
  subroutine comm_table_update(num_of_blk, n_np_NB_mod, n_np_blk_num,   &
       & N_in, NP_in, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,      &
       & STACK_EXPORT, NOD_EXPORT, NEIBPETOT_in, NEIBPE_in,             &
       & STACK_IMPORT_in, NOD_IMPORT_in, STACK_EXPORT_in, NOD_EXPORT_in,&
       & SOLVER_COMM, my_rank, NB)
    
    implicit none
    include 'precision.inc'
    
    integer(kind=kint), intent(in) :: N_in, NP_in, num_of_blk
    integer(kind=kint), intent(in) :: NB, my_rank, SOLVER_COMM
    integer(kind=kint), intent(in) :: n_np_NB_mod(:), n_np_blk_num(:)

    integer(kind=kint) :: NEIBPETOT
    integer(kind=kint), pointer :: NEIBPE(:)
    integer(kind=kint), pointer :: STACK_IMPORT(:), NOD_IMPORT(:)
    integer(kind=kint), pointer :: STACK_EXPORT(:), NOD_EXPORT(:)
    
    integer(kind=kint) :: NEIBPETOT_in
    integer(kind=kint) :: NEIBPE_in(:)
    integer(kind=kint) :: STACK_IMPORT_in(0:), NOD_IMPORT_in(:)
    integer(kind=kint) :: STACK_EXPORT_in(0:), NOD_EXPORT_in(:)

    integer(kind=kint), allocatable :: temp(:)
    
    integer(kind=kint) :: i,j,k,l,N, tsize, counter, new_blk_id
    
    NEIBPETOT = NEIBPETOT_in
    
    allocate(NEIBPE(NEIBPETOT), STACK_IMPORT(0:NEIBPETOT), STACK_EXPORT(0:NEIBPETOT))
    STACK_IMPORT(0) = 0; STACK_EXPORT(0)=0

    do i = 1, NEIBPETOT
       NEIBPE(i) = NEIBPE_in(i)
    end do

    !C temp is used for duplication check of send and recv table
    !C size of (NP_in-N_in) is needed for recv table
    tsize = NP_in -N_in 
    !C size of (N_in/NB+1) is needed for send table
    if(tsize < (N_in/NB+1)) tsize = N_in/NB+1
    

    !C NOD_IMPORT size is determined in create_n_np_arrys.
    allocate(temp(tsize), NOD_IMPORT(num_of_blk))
    
    !C recv table
    !C number of recv table is chaged according to n_np_blk_num(:).
    !C And duplication is deleted.
    counter = 0
    temp = 0
    N = N_in/NB + (mod(N_in, NB)+NB-1)/NB
    do i = 1, NEIBPETOT_in
       do j = STACK_IMPORT_in(i-1) + 1, STACK_IMPORT_in(i)
          k = NOD_IMPORT_in(j) - N_in
          new_blk_id = n_np_blk_num(k) 
          
          if(temp(new_blk_id-N) == 0) then
             counter = counter+1
             NOD_IMPORT(counter) = new_blk_id
             temp(new_blk_id-N) = 1
          end if
       end do
       STACK_IMPORT(i) = counter
    end do

    !C send table
    !C first loop is for counting the send table size
    counter = 0
    do i = 1, NEIBPETOT_in
       temp = 0
       do j = STACK_EXPORT_in(i-1) + 1, STACK_EXPORT_in(i)
          new_blk_id = (NOD_EXPORT_in(j)-1)/NB+1
          if(temp(new_blk_id) == 0) then
             counter = counter+1
             temp(new_blk_id) = 1
          end if
       end do
       STACK_EXPORT(i) = counter

    end do

    allocate(NOD_EXPORT(counter))
    !C second loop assigns the data into export table
    counter = 0
    do i = 1, NEIBPETOT_in
       temp = 0
       do j = STACK_EXPORT_in(i-1) + 1, STACK_EXPORT_in(i)
          new_blk_id = (NOD_EXPORT_in(j)-1)/NB+1
          if(temp(new_blk_id) == 0) then
             counter = counter+1
             temp(new_blk_id) = 1
             NOD_EXPORT(counter) = new_blk_id
          end if
       end do
    end do
    
  end subroutine comm_table_update
  
  subroutine create_n_np_arrys(num_of_blk, n_np_NB_mod, n_np_blk_num,  &
       & N_in, NP_in, NEIBPETOT_in, NEIBPE_in, STACK_IMPORT_in,        &
       & NOD_IMPORT_in, STACK_EXPORT_in, NOD_EXPORT_in, SOLVER_COMM,   &
       & my_rank, NB)
    use solver_SR_NB
    use isort
    implicit none
    include 'precision.inc'
    include 'mpif.h'

    integer(kind=kint), intent(out) :: num_of_blk, n_np_blk_num(:), n_np_NB_mod(:)
    integer(kind=kint), intent(in) :: N_in, NP_in, NB 
    integer(kind=kint), intent(in) :: SOLVER_COMM, my_rank
    integer(kind=kint), intent(in) :: NEIBPETOT_in
    integer(kind=kint), pointer    :: NEIBPE_in      (:)
    integer(kind=kint), pointer    :: STACK_IMPORT_in(:)
    integer(kind=kint), pointer    :: NOD_IMPORT_in  (:)
    integer(kind=kint), pointer    :: STACK_EXPORT_in(:)
    integer(kind=kint), pointer    :: NOD_EXPORT_in  (:)

    integer(kind=kint), allocatable :: N_index(:), WSI(:), WRI(:), order(:)
    integer(kind=kint) :: PETOT, i, j, k, l, wr_size, ws_size, N, ierr

    call MPI_COMM_SIZE (SOLVER_COMM, PETOT, ierr)

    !C N_index is calculated
    allocate(N_index(0:PETOT))
    call MPI_ALLGATHER(N_in, 1, MPI_INTEGER, N_index(1), 1, MPI_INTEGER, SOLVER_COMM, ierr)

    !C padding N_index to multiples of NB.
    DO i = 1, PETOT
       if(mod(N_index(i),NB) > 0) N_index(i) = NB * (N_index(i)/NB+1)
    END DO
    N_index(0) = 0

    DO i = 2, PETOT
       N_index(i) = N_index(i-1) + N_index(i)
    END DO

    !C sendrecv
    ws_size = STACK_EXPORT_in(NEIBPETOT_in); wr_size = STACK_IMPORT_in(NEIBPETOT_in)
    allocate(WSI(ws_size), WRI(wr_size))
    
    DO i = 1, NEIBPETOT_in
       DO j = STACK_EXPORT_in(i-1)+1, STACK_EXPORT_in(i)
          WSI(j) = NOD_EXPORT_in(j) +  N_index(my_rank)
       END DO
    END DO


    call send_recv_buf(ws_size, wr_size, NEIBPETOT_in, NEIBPE_in, STACK_IMPORT_in, &
         & NOD_IMPORT_in, STACK_EXPORT_in, NOD_EXPORT_in, WSI, WRI, SOLVER_COMM, my_rank)
    
    DO i = 1, NEIBPETOT_in
       DO j = STACK_IMPORT_in(i-1)+1, STACK_IMPORT_in(i)
          n_np_blk_num(NOD_IMPORT_in(j)-N_in) = WRI(j)
       END DO
    END DO
    
    DO i = 1, NP_in-N_in
       j = n_np_blk_num(i)
       n_np_blk_num(i) = (j-1)/NB+1
       n_np_NB_mod(i)  = mod(j-1, NB)+1
    END DO

    !C remove duplication of n_np_blk_num and count up different global column number.
    allocate(order(NP_in-N_in))
    call qsorti(order, NP_in-N_in, n_np_blk_num)


    !C traverse n_np_blk_num(:) in ascending order and change the numbering from 1 to 
    !C the number of different kinds.
    k = 0
    if(NP_in > N_in) then
       k = 1
       l = n_np_blk_num(order(1))
       n_np_blk_num(order(1)) = 1
    end if
    do i = 2, NP_in-N_in
       j = order(i)
       if(l == n_np_blk_num(j)) then
          n_np_blk_num(j) = k
       else
          l = n_np_blk_num(j)
          k = k+1
          n_np_blk_num(j) = k
       end if
    end do
    num_of_blk = k

    
    N = N_in/NB + (mod(N_in,NB)+NB-1)/NB
    do i = 1, NP_in-N_in
       n_np_blk_num(i) = n_np_blk_num(i) +N
    end do

    deallocate(WSI,WRI,N_index)
    
  end subroutine create_n_np_arrys
  



end module blocking_mat
