module chebyshev_smoother
  
contains
  
  subroutine set_coef_ary(d, coef_ary, h,l)
    implicit none
    include 'precision.inc'
    integer(kind=kint), intent(in) :: d
    real(kind=kreal), intent(out) :: coef_ary(:)
    real(kind=kreal), intent(in) :: h,l
    real(kind=kreal) :: alpha, beta, amp
    
    integer(kind=kint) :: i
    
    if(d>4 .or. d<2) stop 'chebyshev polynomial degree isnot set between 2-4.'
    if(h<l .or. h<0 .or. l<0) stop 'eigen range isnot set between 2-4.'
    
    alpha = h+l
    beta = h-l
    !C  (coef_ary(n) A^n-1 + coef_ary(n-1) A^n-2 + .. + coef_ary(1)) r
    if(d == 2) then
       amp = 2*alpha**2/beta**2-1
       coef_ary(1) = 8*alpha/(beta*beta)/amp
       coef_ary(2) = -8/(beta*beta)/amp
       
    else if(d == 3) then
       amp = -4*alpha**3/beta**3+3*alpha/beta
       coef_ary(1) = (6/beta-24*alpha**2/beta**3)/amp
       coef_ary(2) = 48*alpha/(beta**3)/amp
       coef_ary(3) = -32/(beta**3)/amp
       
    else if(d == 4) then
       amp = 1-8*alpha**2/beta**2+8*alpha**4/beta**4
       coef_ary(1) = (64*alpha**3/beta**4-32*alpha/beta**2)/amp
       coef_ary(2) = (-192*alpha**2/beta**4+32/beta**2)/amp
       coef_ary(3) = 256*alpha/beta**4/amp
       coef_ary(4) = -128/beta**4/amp
       
    end if
    
    !C change coeff to honer format
    DO i = 1, d-1
       coef_ary(i) = coef_ary(i) / coef_ary(d)
    end DO

  end subroutine set_coef_ary



  subroutine matrix_vector(mat, vec_in, vec_out)
    use matrix_type_mod
    USE comm_switch_mod
    implicit none
    type(matrix) :: mat
    real(kind=kreal) ::  vec_in(:), vec_out(:)
    
    integer(kind=kint) :: i,j,is,ie,col
    
    call comm_sr(mat % SOLVER_COMM, mat % COMM_TABLE,      &
         & mat % NP, mat % WS, mat % WR, vec_in)

    !$omp parallel private(i,j,col)
    !$omp do           
    DO i = 1,mat % N
       vec_out(i) = mat % D(i) * vec_in(i)
    end DO
    !$omp end do 

    !$omp do 
    DO i = 1,mat % N
       DO j = mat % INL(i-1)+1, mat % INL(i)
          col = mat % IAL(j)
          vec_out(i) = vec_out(i) + mat % AL(j) * vec_in(col)
       END DO
       DO j = mat % INU(i-1)+1, mat % INU(i)
          col = mat % IAU(j)
          vec_out(i) = vec_out(i) + mat % AU(j) * vec_in(col)
       END DO
    end DO
    !$omp end do
    
    !$omp end parallel
    
  end subroutine matrix_vector

  subroutine res_cal_vec(mat, vec, X, B)
    use matrix_type_mod
    implicit none

    type(matrix) :: mat
    real(kind=kreal) :: vec(:),X(:),B(:)
    
    integer(kind=kint) :: i
    
    call matrix_vector(mat, X, vec)
    !$omp parallel do    
    DO i = 1, mat % N
       vec(i) = B(i) - vec(i)
    END DO
    !$omp end parallel do    
  end subroutine res_cal_vec


  !C if initialize_fl,  set h and l
  !C else uses the values of h and l
  subroutine cheby_sm_sca(mat, B, X, count, deg, h, l, initialize_fl)
    use count_time_mod
    use matrix_type_mod
    USE comm_switch_mod
    implicit none

    type(matrix) :: mat
    real(kind=kreal),   intent(inout) :: x(:)
!!$    real(kind=kreal) :: WS(:), WR(:)
    real(kind=kreal),   intent(in) :: B(:)
    real(kind=kreal),   intent(inout) :: h,l
    integer(kind=kint), intent(in) :: deg, count
!!$    integer(kind=kint), intent(in) :: N,NP,NPL,NPU
!!$    real(kind=kreal),   intent(in), target :: D(:),AL(:),AU(:)
!!$    integer(kind=kint), intent(in), target :: INL(0:),INU(0:),IAL(:),IAU(:)
!!$
!!$    INTEGER(kind=kint), intent(in), target :: NEIBPE(:)
!!$    INTEGER(kind=kint), intent(in), target :: STACK_IMPORT(0:), STACK_EXPORT(0:)
!!$    INTEGER(kind=kint), intent(in), target :: NOD_EXPORT(:), NOD_IMPORT(:)
!!$    INTEGER(kind=kint), intent(in) :: NEIBPETOT, dim

!!$    INTEGER(kind=kint), intent(in) :: SOLVER_COMM, my_rank
    logical :: initialize_fl
    
    real(kind=kreal) :: coef_ary(10)
    real(kind=kreal), allocatable :: vec_1(:), vec_2(:)
    
    integer(kind=kint) :: i, j,iter, iter_for_eigen


!!$    call set_mat_data(mat, N, NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, &
!!$         & NEIBPETOT, NEIBPE, STACK_IMPORT, STACK_EXPORT, NOD_IMPORT,      &
!!$         & NOD_EXPORT, WS, WR, dim, SOLVER_COMM, my_rank)

    if(.not.initialize_fl) then
       call count_time(1, mat % SOLVER_COMM, mat % my_rank, 10)

       iter_for_eigen = 50
       call determine_largest_eigen(h, mat, iter_for_eigen)
       
       l = h / 20.0
       initialize_fl = .true.
#ifdef PRINT_REZ
       if(mat % my_rank == 0) write(*,*)"h,l", h, l, "CG iter:", iter_for_eigen
#endif
       call count_time(2, mat % SOLVER_COMM, mat % my_rank, 10)

    end if
    
    call set_coef_ary(deg, coef_ary, h, l)
    !C (((x+coef(3))x+coef(2))*coef(4)*x+coef(1)) r
    
    allocate(vec_1(mat % NP), vec_2(mat % NP))

    do iter = 1, count

       !C--- residual vector is calculated
       call res_cal_vec(mat, vec_1, X, B)
       !$omp parallel do    
       do i = 1,mat % N
          vec_1(i) = vec_1(i) * coef_ary(deg)
       end do
       !$omp end parallel do    

       do i = 1, deg-1
          if(mod(i,2)==1) then
             call mat_vec_add(mat, vec_1, X, coef_ary(i), vec_2)
          else
             call mat_vec_add(mat, vec_2, X, coef_ary(i), vec_1)
          end if
       end do

       if(mod(deg, 2)==1) then
          !C output is vec_1
          !$omp parallel do    
          do i = 1, mat % N
             X(i) = X(i) + vec_1(i)
          end do
          !$omp end parallel do    
       else
          !C output is vec_2
          !$omp parallel do    
          do i = 1, mat % N
             X(i) = X(i) + vec_2(i)
          end do
          !$omp end parallel do    
       end if
    end do

    call comm_sr(mat % SOLVER_COMM, mat % COMM_TABLE,      &
         & mat % NP, mat % WS, mat % WR, X)

    deallocate(vec_1,vec_2)
    
  end subroutine cheby_sm_sca
  
  subroutine  mat_vec_add(mat, vec_in, sum_vec, alpha, vec_out)
    use matrix_type_mod
    implicit none


    type(matrix) :: mat
    real(kind=kreal),intent(in) :: vec_in(:)
    real(kind=kreal),intent(inout) :: sum_vec(:)
    real(kind=kreal),intent(out) :: vec_out(:)
    real(kind=kreal) :: alpha
    
    integer(kind=kint) :: n, i

    !C--- vec_out = A x vec_in
    call matrix_vector(mat, vec_in, vec_out)

    !$omp parallel do       
    do i = 1, mat % N
       sum_vec(i) = sum_vec(i) + alpha * vec_in(i)
    end do
    !$omp end parallel do       
    
  end subroutine mat_vec_add

  subroutine injection(val, vec1, vec2, n, SOLVER_COMM)
    use matrix_type_mod
    implicit none
    include 'mpif.h'

    real(kind=kreal)   :: val, val0, vec1(:), vec2(:)
    integer(kind=kint) :: n, i, SOLVER_COMM, ierr

    val0 = 0.0
    !$omp parallel do reduction(+:val0)
    do i = 1, n
       val0 = val0 + vec1(i)*vec2(i)
    end do
    !$omp end parallel do       
    
    CALL MPI_ALLREDUCE(val0, val, 1, MPI_DOUBLE_PRECISION, MPI_SUM, SOLVER_COMM, ierr)
    
  end subroutine injection

  subroutine determine_largest_eigen(lambda, mat, iter)
    use matrix_type_mod
    implicit none

    include 'mpif.h'
    
    type(matrix)     :: mat
    real(kind=kreal) :: lambda
    integer(kind=kint), intent(inout) :: iter
    
    real(kind=kreal), allocatable :: tri_diagonal(:,:)
    integer(kind=kint) :: ierr
    
    allocate(tri_diagonal(iter,3))

    
    call tri_diagonal_from_cg(tri_diagonal, mat, iter)
    
    if(mat % my_rank == 0)then
       call power_method_tri_diagonal(lambda, tri_diagonal, iter)
    end if
    
    call MPI_BCAST(lambda, 1, MPI_DOUBLE_PRECISION, 0, mat % SOLVER_COMM, ierr)
    
  end subroutine determine_largest_eigen


  subroutine tri_diagonal_from_cg(tri_diagonal, mat, iter)
    use matrix_type_mod
    implicit none

    type(matrix), intent(in)     :: mat
    real(kind=kreal), intent(out) :: tri_diagonal(:,:)
    integer(kind=kint), intent(inout) :: iter
    integer(kind=kint) :: i,j
    
    real(kind=kreal), allocatable :: b(:), x(:), r(:), p(:), q(:), alpha(:), beta(:)
    real(kind=kreal) :: rho, rho1, v, rho_ini


    allocate(b(mat % NP), x(mat % NP), r(mat % NP), p(mat % NP), q(mat % NP))
    allocate(alpha(iter), beta(iter))
    
    call random_number(b)

    r = b; p = r; x = 0
    call injection(rho, r, r, mat % N, mat % SOLVER_COMM)
    rho_ini = rho
    
    do i = 1, iter
       call matrix_vector(mat, p, q)
       call injection(v, q, p, mat % N, mat % SOLVER_COMM)
       alpha(i) = rho / v
       
       !$omp parallel do       
       do j = 1, mat % N
          x(j) = x(j) + alpha(i)*p(j)
          r(j) = r(j) - alpha(i)*q(j)
       end do
       !$omp end parallel do       

       call injection(rho1, r, r, mat % N, mat % SOLVER_COMM)
       if(dsqrt(rho1/rho_ini) < 1.0E-13) exit
       
       beta(i) = rho1/rho
       !$omp parallel do       
       do j = 1, mat % N
          p(j) = r(j) + beta(i)*p(j)
       end do
       !$omp end parallel do       
       rho = rho1
    end do
    if(i < iter) iter = i
    
    tri_diagonal(1,1) = 0
    tri_diagonal(1,2) = 1/alpha(1)
    do i = 2, iter
       tri_diagonal(i,1) = dsqrt(beta(i-1))/alpha(i-1)
       tri_diagonal(i-1,3) = tri_diagonal(i,1)

       tri_diagonal(i,2) = 1/alpha(i)+beta(i-1)/alpha(i-1)
    end do
    tri_diagonal(iter,3) = 0

    do i =1, iter
       if(dabs(alpha(i)) < 1.0E-9) write(*,*) "small alpha", alpha(i)
    end do

    deallocate(b, x, r, p, q)
    deallocate(alpha, beta)
    
  end subroutine tri_diagonal_from_cg
  
  subroutine power_method_tri_diagonal(lambda, tri_diagonal, n)
    implicit none
    include 'precision.inc'
    integer(kind=kint) :: n
    real(kind=kreal)   :: lambda, tri_diagonal(:,:)

    real(kind=kreal), allocatable :: vec(:), vec2(:)
    
    integer(kind=kint), parameter :: iteration=100
    integer(kind=kint) :: i, iter
    
    real(kind=kreal) :: norm2
    
    allocate(vec(n), vec2(n))
    
    call random_number(vec)
    
    do iter = 1, iteration
       !C-- matvec
       vec2(1) = tri_diagonal(1,2)*vec(1) + tri_diagonal(1,3)*vec(2)
       do i = 2, n-1
          vec2(i) = tri_diagonal(i,2)*vec(i) + tri_diagonal(i,1)*vec(i-1) +tri_diagonal(i,3)*vec(i+1)
       end do
       vec2(n) = tri_diagonal(n,2)*vec(n) + tri_diagonal(n,1)*vec(n-1) 

       !C-- normalization and assignment to vec(:)
       norm2 = 0
       do i = 1, n
          norm2 = norm2 + vec2(i)*vec2(i)
       end do
       norm2 = dsqrt(norm2)
       do i = 1, n
          vec(i) = vec2(i)/norm2
       end do
    end do
    
    lambda = norm2
    
  end subroutine power_method_tri_diagonal

  
end module chebyshev_smoother
