module comm_switch_mod
  !C refers solver_sr2, axis_comm
  
  contains
    subroutine comm_switch(C, X, Y, Z, topo_sz, SOLVER_COMM, comm_tbl, NP)
      use data_structure_for_amg, only: COMMUNICATION_TABLE
      use solver_SR2,             only: solver_send_recv2
      use axis_comm,              only: new_SR2, comm_info, &
           & setup_comm_info, clear_comm, initialize_comm
      implicit none
      include 'mpif.h'
      include 'precision.inc'

      
      type(COMMUNICATION_TABLE) :: comm_tbl
      integer, intent(in) ::SOLVER_COMM
      integer :: ierr, my_rank
      
      integer(kind=kint) :: C(:), X(:), Y(:), Z(:), topo_sz, NP
      real(kind=kreal), allocatable :: test_vec(:), WS(:), WR(:)
      
      real(kind=kreal)   :: t_n, t, tmin
      integer(kind=kint) :: min_topo, i,j,k, tmp(2), max_neibpetot
      type(comm_info), allocatable :: comm_info_ary(:)

      k = 3
      call MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr)

      call initialize_comm(comm_tbl % comm)

      !C compare communication routines only when topo_sz > 0
      if(topo_sz > 0) then
         
         allocate(WS(comm_tbl % STACK_EXPORT(comm_tbl % NEIBPETOT)))
         allocate(WR(comm_tbl % STACK_IMPORT(comm_tbl % NEIBPETOT)))
         allocate(test_vec(NP))

         test_vec = 0
         
         call MPI_allreduce(comm_tbl % NEIBPETOT, max_neibpetot, 1, &
              & MPI_INTEGER, MPI_MAX, SOLVER_COMM, ierr)

         if(max_neibpetot > 60) then

            t_n = -MPI_WTIME(); 

            if(comm_tbl % NEIBPETOT > 0) then
               do j = 1, k
                  call solver_send_recv2(NP,                            &
                       & comm_tbl % STACK_EXPORT(comm_tbl % NEIBPETOT), &
                       & comm_tbl % STACK_IMPORT(comm_tbl % NEIBPETOT), &
                       & comm_tbl % NEIBPETOT,                          &
                       & comm_tbl % NEIBPE,                             &
                       & comm_tbl % STACK_IMPORT,                       &
                       & comm_tbl % NOD_IMPORT,                         &
                       & comm_tbl % STACK_EXPORT,                       &
                       & comm_tbl % NOD_EXPORT,                         &
                       & WS, WR, test_vec, SOLVER_COMM, my_rank)
               end do
            end if
            call MPI_Barrier(SOLVER_COMM, ierr); t_n = t_n+MPI_WTIME()


            allocate(comm_info_ary(topo_sz))

            tmin = t_n
            comm_tbl % switch = 0
            min_topo = 0

#ifdef PRINT_REZ
            if(my_rank ==0 ) write(*,*) "usual comm time[s]:", t_n
            if(my_rank ==0 ) write(*,*) "axis comm time[s]:"
#endif
            do i = 1, topo_sz
               comm_info_ary(i) % C           = C(i)
               comm_info_ary(i) % X           = X(i)
               comm_info_ary(i) % Y           = Y(i)
               comm_info_ary(i) % Z           = Z(i)
               comm_info_ary(i) % my_rank     = my_rank
               comm_info_ary(i) % SOLVER_COMM = SOLVER_COMM

               call setup_comm_info(comm_tbl % NEIBPETOT,    &
                    &               comm_tbl % NEIBPE,       &
                    &               comm_tbl % STACK_EXPORT, &
                    &               comm_tbl % STACK_IMPORT, &
                    &               comm_info_ary(i))

               call MPI_Barrier(SOLVER_COMM, ierr); t = -MPI_WTIME();


               do j = 1, k
                  call new_SR2(comm_info_ary(i),   &
                       &  comm_tbl % NEIBPETOT,    &
                       &  comm_tbl % NEIBPE,       &
                       &  comm_tbl % STACK_IMPORT, &
                       &  comm_tbl % NOD_IMPORT,   &
                       &  comm_tbl % STACK_EXPORT, &
                       &  comm_tbl % NOD_EXPORT,   &
                       &  WR, test_vec);    
               end do

               call MPI_Barrier(SOLVER_COMM, ierr); t = t+MPI_WTIME()
#ifdef PRINT_REZ
               if(my_rank ==0 ) write(*,*) t
#endif

               if(tmin > t .and. my_rank == 0) then
                  comm_tbl % switch = 1
                  tmin              = t
                  min_topo          = i
               end if
            end do

            if(my_rank == 0) then
               tmp(1) = comm_tbl % switch; tmp(2) = min_topo
            end if
            call MPI_BCAST(tmp, 2, MPI_INTEGER, 0, SOLVER_COMM, ierr)
            comm_tbl % switch = tmp(1); min_topo = tmp(2)

            if(comm_tbl % switch == 1) then
               comm_tbl % comm = comm_info_ary(min_topo)
            end if

            do i = 1, topo_sz
               if(i /= min_topo)  call clear_comm(comm_info_ary(i))
            end do

         end if
      end if
      
      
      
    end subroutine comm_switch
    
    subroutine comm_sr(SOLVER_COMM, comm_tbl, NP, WS, WR, vec)
      use solver_SR2,             only: solver_send_recv2
      use axis_comm,              only: new_SR2, comm_info
      use data_structure_for_amg, only: COMMUNICATION_TABLE
      implicit none
      include 'precision.inc'

      type(communication_table) :: comm_tbl
      real(kind=kreal) :: WS(:), WR(:), vec(:)
      integer :: SOLVER_COMM, my_rank, ierr
      integer(kind=kint) :: NP

      call MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr)
      
      if(comm_tbl % switch > 0) then
         call new_SR2(comm_tbl % comm,   &
              & comm_tbl % NEIBPETOT,    &
              & comm_tbl % NEIBPE,       &
              & comm_tbl % STACK_IMPORT, &
              & comm_tbl % NOD_IMPORT,   &
              & comm_tbl % STACK_EXPORT, &
              & comm_tbl % NOD_EXPORT,   &
              & WR, vec)
         
      else
         if(comm_tbl % NEIBPETOT > 0) then
            call solver_send_recv2(NP, &
                 & comm_tbl % STACK_EXPORT(comm_tbl % NEIBPETOT), &
                 & comm_tbl % STACK_IMPORT(comm_tbl % NEIBPETOT), &
                 & comm_tbl % NEIBPETOT,                          &
                 & comm_tbl % NEIBPE,                             &
                 & comm_tbl % STACK_IMPORT,                       &
                 & comm_tbl % NOD_IMPORT,                         &
                 & comm_tbl % STACK_EXPORT,                       &
                 & comm_tbl % NOD_EXPORT,                         &
                 & WS, WR, vec, SOLVER_COMM, my_rank)
         end if
         
      end if
      
    end subroutine comm_sr
    
    subroutine comm_sr_rev_add(SOLVER_COMM, comm_tbl, coarser_NP, WS, WR, vec)
      use solver_Gnumbering,      only: solver_sr_reduce
      use axis_comm,              only: new_SR2_rev_add, comm_info
      use data_structure_for_amg, only: COMMUNICATION_TABLE
      implicit none
      include 'precision.inc'

      type(communication_table) :: comm_tbl
      real(kind=kreal) :: WS(:), WR(:), vec(:)
      integer :: SOLVER_COMM, my_rank, ierr
      integer(kind=kint) :: coarser_NP
      
      call MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr)

      if(comm_tbl % switch > 0) then
         call new_SR2_rev_add(comm_tbl % comm,   &
              & comm_tbl % NEIBPETOT,    &
              & comm_tbl % NEIBPE,       &
              & comm_tbl % STACK_IMPORT, &
              & comm_tbl % NOD_IMPORT,   &
              & comm_tbl % STACK_EXPORT, &
              & comm_tbl % NOD_EXPORT,   &
              & WR, vec)
         
      else
         if(comm_tbl % NEIBPETOT > 0) then
            call solver_sr_reduce(coarser_NP, &
                 & comm_tbl % STACK_EXPORT(comm_tbl % NEIBPETOT), &
                 & comm_tbl % STACK_IMPORT(comm_tbl % NEIBPETOT), &
                 & comm_tbl % NEIBPETOT,                          &
                 & comm_tbl % NEIBPE,                             &
                 & comm_tbl % STACK_IMPORT,                       &
                 & comm_tbl % NOD_IMPORT,                         &
                 & comm_tbl % STACK_EXPORT,                       &
                 & comm_tbl % NOD_EXPORT,                         &
                 & WS, WR, vec, SOLVER_COMM, my_rank)
         end if
         
      end if
      

      
    end subroutine comm_sr_rev_add
end module comm_switch_mod

