module common_mod

contains
  
  SUBROUTINE setup_phase()
    use problem_matrix
    use amg_parameter, only: vector_problem_fl, INDPDT_AGGRE, solver_string
    use data_creation_AMGCG, only: data_creation
    use data_creation_AMGCG_NB, only: data_creation_NB

#ifdef MEASURING
    ! for keisoku
    use count_time_mod
#endif

    IMPLICIT NONE
    

    CHARACTER(len=20)::PRECOND
    integer(kind=kint) :: gpu_sol


    if(INDPDT_AGGRE) then
       PRECOND = "ind1"
    else
       PRECOND = "SHA2"
    end if
    
!!$    PRECOND = "MIX"

    if(vector_problem_fl) then
       !C topology routines is not prepared for vector problems..
       call data_creation_NB                                              &
            &                 (N, NP, NPL, NPU, D_NB, AL_NB, INL, IAL, AU_NB,  &
            &                  INU, IAU, &
            &                  LEVEL_NUM, my_rank, NEIBPETOT, NEIBPE,          &
            &                  STACK_IMPORT, NOD_IMPORT,                       &
            &                  STACK_EXPORT, NOD_EXPORT,                       &
            &                  SOLVER_COMM, PRECOND, NB, k_or_q_vec, NK)     
         
    else

       !C for symmetric matrix case, the last argument should be set as 1
       CALL data_creation(N, NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU,      & 
            & LEVEL_NUM, my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, & 
            & STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, PRECOND, 0)



#ifdef GPU            
       call count_time(0,solver_comm,my_rank,61)
       call count_time(1,solver_comm,my_rank,61)

       if(    solver_string(1:1) == 'G' .and. &
            & solver_string(2:2) == 'P' .and. &
            & solver_string(3:3) == 'U' .and. &
            & solver_string(4:4) == 'B' .and. &
            & solver_string(5:5) == 'I' .and. &
            & solver_string(6:6) == 'C') then
          gpu_sol = 1
          call mat_and_solver_set_on_GPU(LEVEL_NUM, gpu_sol)
       else if(solver_string(1:1) == 'G' .and. &
            & solver_string(2:2) == 'P' .and. &
            & solver_string(3:3) == 'U') then
          gpu_sol = 0
          call mat_and_solver_set_on_GPU(LEVEL_NUM, gpu_sol)
       end if

       call count_time(2,solver_comm,my_rank,61)
#endif

    end if
    
  end SUBROUTINE setup_phase

#ifdef GPU            
  subroutine mat_and_solver_set_on_GPU(LEVEL_NUM, sol)
    USE data_structure_for_AMG, only: HIERARCHICAL_DATA
    USE problem_matrix, only: my_rank
    USE amg_parameter, only: iteration_number_down, iteration_number_bottom, gs_kind, &
         & dump_gauss_seidel_weight

    implicit none
    include 'precision.inc'
    
    !C AMG:sol == 0 AMGBICGSTAB:sol == 1
    INTEGER(kind=kint), intent(in) :: LEVEL_NUM, sol
    INTEGER(kind=kint) :: i, IN_T_N, IN_T_NPLU, EX_T_N, EX_T_NPLU
    INTEGER(kind=kint) :: R_N,P_N,R_NP,P_NP,R_NPLU,P_NPLU
    INTEGER(kind=kint) :: INTER_IN_T_N,INTER_IN_T_NPLU
    INTEGER(kind=kint) :: INTER_EX_T_N,INTER_EX_T_NPLU
    INTEGER(kind=kint) :: iter_down, iter_bottom, gs_at_lev1
    INTEGER(kind=kint),  target, save :: INT_DUM(1)
    REAL   (kind=kreal), target, save :: REAL_DUM(1)
    
    INTEGER(kind=kint), pointer, save :: stack_inptr(:), stack_exptr(:)
    INTEGER(kind=kint), pointer, save :: INT_stack_inptr(:), INT_stack_exptr(:)

    INTEGER(kind=kint) :: scl_fl




    CALL notify_level(LEVEL_NUM, my_rank)
    
    iter_down   = iteration_number_down
    iter_bottom = iteration_number_bottom
    if(gs_kind == 0) then
       !C GS smoother at lev1 and Jacobi at other levels with LU at the coarsest
       gs_at_lev1  = 5 
    else if(gs_kind==1) then
       !C jacobi smoother at all levs with LU at the coarsest
       gs_at_lev1  = 4
    else 
       if(my_rank ==0) write(*,*) "gs_kind for GPU smoother is only 0 or 1. Jacobi smoother is set."
       gs_at_lev1 = 4
       
    end if
    

    call parameter_set_on_GPU(sol, dump_gauss_seidel_weight, iter_down, iter_bottom, gs_at_lev1)
    
    i = 1
    IN_T_N    = HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPETOT
    IN_T_NPLU = HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT(IN_T_N)
    EX_T_N    = HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPETOT
    EX_T_NPLU = HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT(EX_T_N)
    stack_inptr => HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT
    stack_exptr => HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT
    if(IN_T_N == 0) stack_inptr=>INT_DUM
    if(EX_T_N == 0) stack_exptr=>INT_DUM

    if(HIERARCHICAL_DATA(i) % scaling_fl) then
       scl_fl = 1
    else
       scl_fl = 0
       HIERARCHICAL_DATA(i) % SCALE_D =>REAL_DUM
    end if
    
    if(HIERARCHICAL_DATA(i) % N == 0) then
       !C when 0 size problem matrix comes, then dummy data is given to the sub.
       !C if 0 size problem mat at coarser level is ok, because the arrays are
       !C allocated in setup_phase.
       CALL store_mat_on_GPU(i,HIERARCHICAL_DATA(i) % N, &
            & HIERARCHICAL_DATA(i) % NP, &
            & HIERARCHICAL_DATA(i) % NPL,&
            & HIERARCHICAL_DATA(i) % NPU,&
            & INT_DUM,&
            & INT_DUM,&
            & INT_DUM,&
            & REAL_DUM,&
            & INT_DUM,&
            & REAL_DUM, &
            & REAL_DUM,&
            & IN_T_N,IN_T_NPLU,&
            & INT_DUM, &
            & INT_DUM, &
            & INT_DUM, &
            & EX_T_N, EX_T_NPLU, & 
            & INT_DUM, &
            & INT_DUM, &
            & INT_DUM, &
            & INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, &               
            & INT_DUM, INT_DUM, REAL_DUM, &
            & INT_DUM, INT_DUM, REAL_DUM, &
            & INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, &
            & INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, & 
            & scl_fl,  REAL_DUM)
    else
       CALL store_mat_on_GPU(i, HIERARCHICAL_DATA(i) % N, &
            & HIERARCHICAL_DATA(i) % NP, &
            & HIERARCHICAL_DATA(i) % NPL,&
            & HIERARCHICAL_DATA(i) % NPU,&
            & HIERARCHICAL_DATA(i) % INL,&
            & HIERARCHICAL_DATA(i) % INU,&
            & HIERARCHICAL_DATA(i) % IAL,&
            & HIERARCHICAL_DATA(i) % AL,&
            & HIERARCHICAL_DATA(i) % IAU,&
            & HIERARCHICAL_DATA(i) % AU, &
            & HIERARCHICAL_DATA(i) % D,&
            & IN_T_N,IN_T_NPLU,&
            & HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPE, &
!!$            & HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT(1), &
            & stack_inptr(1), &
            & HIERARCHICAL_DATA(i) % COMM_TABLE % NOD_IMPORT, &
            & EX_T_N, EX_T_NPLU, & 
            & HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPE, &
!!$            & HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT(1), &
            & stack_exptr(1), &
            & HIERARCHICAL_DATA(i) % COMM_TABLE % NOD_EXPORT, &
            & INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, &               
            & INT_DUM, INT_DUM, REAL_DUM, &
            & INT_DUM, INT_DUM, REAL_DUM, &
            & INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM,&
            & INT_DUM, INT_DUM, INT_DUM, INT_DUM, INT_DUM, &
            & scl_fl, HIERARCHICAL_DATA(i) % SCALE_D)
    end if
    
    DO i = 2, LEVEL_NUM

       IN_T_N    = HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPETOT
       IN_T_NPLU = HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT(IN_T_N)
       EX_T_N    = HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPETOT
       EX_T_NPLU = HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT(EX_T_N)
       R_N       = HIERARCHICAL_DATA(i)   % N
       P_N       = HIERARCHICAL_DATA(i-1) % N
       R_NP      = HIERARCHICAL_DATA(i) % R % row_size
       P_NP      = HIERARCHICAL_DATA(i) % P % row_size
       if(R_NP > 0) then
          R_NPLU    = HIERARCHICAL_DATA(i) % R % IN(R_NP)
       else
          R_NPLU    = 0
       end if
       if(P_NP > 0) then
          P_NPLU    = HIERARCHICAL_DATA(i) % P % IN(P_NP)
       else
          P_NPLU    = 0
       end if

       stack_inptr => HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT
       stack_exptr => HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT
       if(IN_T_N == 0) stack_inptr=>INT_DUM
       if(EX_T_N == 0) stack_exptr=>INT_DUM

       
       INTER_IN_T_N    = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NEIBPETOT
       INTER_IN_T_NPLU = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_IMPORT(INTER_IN_T_N)
       INTER_EX_T_N    = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NEIBPETOT
       INTER_EX_T_NPLU = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_EXPORT(INTER_EX_T_N)

       INT_stack_inptr => HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_IMPORT
       INT_stack_exptr => HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_EXPORT
       if(INTER_IN_T_N == 0) INT_stack_inptr=>INT_DUM
       if(INTER_EX_T_N == 0) INT_stack_exptr=>INT_DUM


       if(HIERARCHICAL_DATA(i) % scaling_fl) then
          scl_fl = 1
       else
          scl_fl = 0
          HIERARCHICAL_DATA(i) % SCALE_D => REAL_DUM
       end if


       CALL store_mat_on_GPU(i,HIERARCHICAL_DATA(i) % N, &
            & HIERARCHICAL_DATA(i) % NP, &
            & HIERARCHICAL_DATA(i) % NPL,&
            & HIERARCHICAL_DATA(i) % NPU,&
            & HIERARCHICAL_DATA(i) % INL,&
            & HIERARCHICAL_DATA(i) % INU,&
            & HIERARCHICAL_DATA(i) % IAL,&
            & HIERARCHICAL_DATA(i) % AL, &
            & HIERARCHICAL_DATA(i) % IAU,&
            & HIERARCHICAL_DATA(i) % AU, &
            & HIERARCHICAL_DATA(i) % D,  &
            & IN_T_N, IN_T_NPLU, & 
            & HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPE, &
!C            & HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT(1), &
            & stack_inptr(1), &
            & HIERARCHICAL_DATA(i) % COMM_TABLE % NOD_IMPORT, &
            & EX_T_N, EX_T_NPLU, & 
            & HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPE, &
!C            & HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT(1), &
            & stack_exptr(1), &
            & HIERARCHICAL_DATA(i) % COMM_TABLE % NOD_EXPORT, &
            & R_N, P_N, R_NP, P_NP, R_NPLU, P_NPLU, &               
            & HIERARCHICAL_DATA(i) % R % IN,&
            & HIERARCHICAL_DATA(i) % R % CN,&
            & HIERARCHICAL_DATA(i) % R % V, &
            & HIERARCHICAL_DATA(i) % P % IN,&
            & HIERARCHICAL_DATA(i) % P % CN,&
            & HIERARCHICAL_DATA(i) % P % V, &
            & INTER_IN_T_N, INTER_IN_T_NPLU,&
            & HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NEIBPE, &
!C            & HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_IMPORT(1), &
            & INT_stack_inptr(1), &
            & HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NOD_IMPORT, &
            & INTER_EX_T_N, INTER_EX_T_NPLU,&
            & HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NEIBPE, &
!C            & HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_EXPORT(1), &
            & INT_stack_exptr(1), &
            & HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NOD_EXPORT, &
            & scl_fl, HIERARCHICAL_DATA(i) % SCALE_D)
    END DO

  end subroutine mat_and_solver_set_on_GPU
#endif



  subroutine clear_matrix_ssi_amg(LEVEL_NUM)
    IMPLICIT NONE
    include 'precision.inc'
    integer(kind=kint) :: LEVEL_NUM

    call clear_multi_lev_matrix(LEVEL_NUM)

  end subroutine clear_matrix_ssi_amg


  subroutine clear_matrices()
    use problem_matrix
    use amg_parameter, only: vector_problem_fl
    IMPLICIT NONE
    
    call clear_multi_lev_matrix(LEVEL_NUM)
    if(associated(INL)) then
       deallocate(INL, INU, IAL, IAU)
       if(NB > 1 .or. vector_problem_fl) then
          deallocate(AL_NB, AU_NB, D_NB)
       else
          deallocate(AL, AU, D)
       end if
       deallocate(NEIBPE,NOD_IMPORT,NOD_EXPORT,STACK_IMPORT,STACK_EXPORT)
       
       if(associated(order)) then
          deallocate(order, nds_index)
       end if
       if(associated(order_for_dst)) then
          deallocate(order_for_dst,nds_index_aft_dst)
       end if

       if(associated(k_or_q_vec)) then
          deallocate(k_or_q_vec)
       end if
       if(associated(scale_D)) deallocate( scale_D)
       
    end if
    REORDERED_FL = .false.
    BLOCKING_FL = .false.
    SCALING_FL=.false.
    
  end subroutine clear_matrices

  subroutine clear_multi_lev_matrix(LEVEL_NUM)
    use amg_parameter, only: vector_problem_fl
    IMPLICIT NONE
    include 'precision.inc'
    integer(kind=kint) :: LEVEL_NUM

    if(vector_problem_fl) then
       call clear_multi_lev_matrix_VC(LEVEL_NUM)
    else
       call clear_multi_lev_matrix_SC(LEVEL_NUM)
    end if
    
    contains
      SUBROUTINE clear_multi_lev_matrix_SC(LEVEL_NUM)
        use amg_parameter, only: W_OUT_LU
        use data_structure_for_AMG
        use axis_comm, only: clear_comm

        IMPLICIT NONE
        INTEGER(kind=kint) :: i, LEVEL_NUM




        if(LEVEL_NUM >0) then
           HIERARCHICAL_DATA(1:LEVEL_NUM) % cheb_fl    =.false.


           !$ deallocate(HIERARCHICAL_DATA(1) % color_index, HIERARCHICAL_DATA(1) % INML)
           NULLIFY(    HIERARCHICAL_DATA(1) % INU,                        &
                &      HIERARCHICAL_DATA(1) % INL,                        &
                &      HIERARCHICAL_DATA(1) % IAU,                        &
                &      HIERARCHICAL_DATA(1) % IAL,                        &
                &      HIERARCHICAL_DATA(1) % AU,                         &
                &      HIERARCHICAL_DATA(1) % AL,                         &
                &      HIERARCHICAL_DATA(1) % D,                          &
                &      HIERARCHICAL_DATA(1) % X,                          &
                &      HIERARCHICAL_DATA(1) % B,                          &       
                &      HIERARCHICAL_DATA(1) % COMM_TABLE % NEIBPE,        &
                &      HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_IMPORT,  & 
                &      HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_EXPORT,  & 
                &      HIERARCHICAL_DATA(1) % COMM_TABLE % NOD_IMPORT,    &
                &      HIERARCHICAL_DATA(1) % COMM_TABLE % NOD_EXPORT)
           
           call clear_comm(HIERARCHICAL_DATA(1) % COMM_TABLE % comm)


           DO i=2,LEVEL_NUM-1
              deallocate( HIERARCHICAL_DATA(i) % INU,      &
                   &      HIERARCHICAL_DATA(i) % INL,      &
                   &      HIERARCHICAL_DATA(i) % IAU,      &
                   &      HIERARCHICAL_DATA(i) % IAL,      &
                   &      HIERARCHICAL_DATA(i) % AU,       &
                   &      HIERARCHICAL_DATA(i) % AL,       &
                   &      HIERARCHICAL_DATA(i) % D,        &
                   &      HIERARCHICAL_DATA(i) % X,        &
                   &      HIERARCHICAL_DATA(i) % B,        &
                   &      HIERARCHICAL_DATA(i) % R % IN,   &
                   &      HIERARCHICAL_DATA(i) % R % CN,   &
                   &      HIERARCHICAL_DATA(i) % R % V,    &
                   &      HIERARCHICAL_DATA(i) % P % IN,   &
                   &      HIERARCHICAL_DATA(i) % P % CN,   &
                   &      HIERARCHICAL_DATA(i) % P % V        &
                   !$ &  ,HIERARCHICAL_DATA(i) % color_index, &
                   !$ &  HIERARCHICAL_DATA(i) % INML          &
                   &      )

              deallocate( HIERARCHICAL_DATA(i) % P,        &
                   &      HIERARCHICAL_DATA(i) % R)

              deallocate(HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPE,       &
                   &     HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT, & 
                   &     HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT, & 
                   &     HIERARCHICAL_DATA(i) % COMM_TABLE % NOD_IMPORT,   &
                   &     HIERARCHICAL_DATA(i) % COMM_TABLE % NOD_EXPORT)

              call clear_comm(HIERARCHICAL_DATA(i) % COMM_TABLE % comm)

              deallocate(HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NEIBPE, &
                   &     HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_IMPORT, & 
                   &     HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_EXPORT, & 
                   &     HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NOD_IMPORT,   &
                   &     HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NOD_EXPORT)

              call clear_comm(HIERARCHICAL_DATA(i) % INT_LVL_TABLE % comm)
              
              if(HIERARCHICAL_DATA(i) % scaling_fl) deallocate(HIERARCHICAL_DATA(i) % scale_D)

           END DO
           
           deallocate( HIERARCHICAL_DATA(LEVEL_NUM) % INU,         &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % INL,      &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % IAU,      &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % IAL,      &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % AU,       &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % AL,       &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % D,        &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % X,        &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % B,        &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % R % IN,   &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % R % CN,   &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % R % V,    &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % P % IN,   &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % P % CN,   &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % P % V     &
                   )
           
           deallocate(    HIERARCHICAL_DATA(LEVEL_NUM) % P,        &
                   &      HIERARCHICAL_DATA(LEVEL_NUM) % R)

           if(W_OUT_LU) then
              !$  deallocate( &
              !$ &  HIERARCHICAL_DATA(LEVEL_NUM) % color_index, &
              !$ &  HIERARCHICAL_DATA(LEVEL_NUM) % INML &
              !$ & )
           end if

           
           deallocate(HIERARCHICAL_DATA(LEVEL_NUM) % COMM_TABLE % NEIBPE,       &
                &     HIERARCHICAL_DATA(LEVEL_NUM) % COMM_TABLE % STACK_IMPORT, & 
                &     HIERARCHICAL_DATA(LEVEL_NUM) % COMM_TABLE % STACK_EXPORT, & 
                &     HIERARCHICAL_DATA(LEVEL_NUM) % COMM_TABLE % NOD_IMPORT,   &
                &     HIERARCHICAL_DATA(LEVEL_NUM) % COMM_TABLE % NOD_EXPORT)

           call clear_comm(HIERARCHICAL_DATA(LEVEL_NUM) % COMM_TABLE % comm)
           
           deallocate(HIERARCHICAL_DATA(LEVEL_NUM) % INT_LVL_TABLE % NEIBPE, &
                &     HIERARCHICAL_DATA(LEVEL_NUM) % INT_LVL_TABLE % STACK_IMPORT, & 
                &     HIERARCHICAL_DATA(LEVEL_NUM) % INT_LVL_TABLE % STACK_EXPORT, & 
                &     HIERARCHICAL_DATA(LEVEL_NUM) % INT_LVL_TABLE % NOD_IMPORT,   &
                &     HIERARCHICAL_DATA(LEVEL_NUM) % INT_LVL_TABLE % NOD_EXPORT)

           call clear_comm(HIERARCHICAL_DATA(LEVEL_NUM) % INT_LVL_TABLE % comm)
              
           if(HIERARCHICAL_DATA(LEVEL_NUM) % scaling_fl) &
                & deallocate(HIERARCHICAL_DATA(LEVEL_NUM) % scale_D)
           
           HIERARCHICAL_DATA(1:LEVEL_NUM) % scaling_fl =.false.
           LEVEL_NUM = 0
           
           if(associated(coarsest_A)) deallocate(coarsest_A, coarsest_IDX)
           
        end if

      END SUBROUTINE clear_multi_lev_matrix_SC
      

!C   vector version can't use multi-color GS.
      SUBROUTINE clear_multi_lev_matrix_VC(LEVEL_NUM)
        use data_structure_for_AMG_NB

        IMPLICIT NONE
        INTEGER(kind=kint) :: i, LEVEL_NUM


        if(LEVEL_NUM >0) then
!!$           !$ deallocate(HIERARCHICAL_DATA_NB(1) % color_index, HIERARCHICAL_DATA_NB(1) % INML)
           NULLIFY(    HIERARCHICAL_DATA_NB(1) % INU,                        &
                &      HIERARCHICAL_DATA_NB(1) % INL,                        &
                &      HIERARCHICAL_DATA_NB(1) % IAU,                        &
                &      HIERARCHICAL_DATA_NB(1) % IAL,                        &
                &      HIERARCHICAL_DATA_NB(1) % AU,                         &
                &      HIERARCHICAL_DATA_NB(1) % AL,                         &
                &      HIERARCHICAL_DATA_NB(1) % D,                          &
                &      HIERARCHICAL_DATA_NB(1) % X,                          &
                &      HIERARCHICAL_DATA_NB(1) % B,                          &       
                &      HIERARCHICAL_DATA_NB(1) % COMM_TABLE % NEIBPE,        &
                &      HIERARCHICAL_DATA_NB(1) % COMM_TABLE % STACK_IMPORT,  & 
                &      HIERARCHICAL_DATA_NB(1) % COMM_TABLE % STACK_EXPORT,  & 
                &      HIERARCHICAL_DATA_NB(1) % COMM_TABLE % NOD_IMPORT,    &
                &      HIERARCHICAL_DATA_NB(1) % COMM_TABLE % NOD_EXPORT)
           
           deallocate( HIERARCHICAL_DATA_NB(1) % ALU)

           DO i=2,LEVEL_NUM
              deallocate( HIERARCHICAL_DATA_NB(i) % INU,      &
                   &      HIERARCHICAL_DATA_NB(i) % INL,      &
                   &      HIERARCHICAL_DATA_NB(i) % IAU,      &
                   &      HIERARCHICAL_DATA_NB(i) % IAL,      &
                   &      HIERARCHICAL_DATA_NB(i) % AU,       &
                   &      HIERARCHICAL_DATA_NB(i) % AL,       &
                   &      HIERARCHICAL_DATA_NB(i) % ALU,      & 
                   &      HIERARCHICAL_DATA_NB(i) % D,        &
                   &      HIERARCHICAL_DATA_NB(i) % X,        &
                   &      HIERARCHICAL_DATA_NB(i) % B,        &
                   &      HIERARCHICAL_DATA_NB(i) % R % IN,   &
                   &      HIERARCHICAL_DATA_NB(i) % R % CN,   &
                   &      HIERARCHICAL_DATA_NB(i) % R % V,    &
                   &      HIERARCHICAL_DATA_NB(i) % P % IN,   &
                   &      HIERARCHICAL_DATA_NB(i) % P % CN,   &
                   &      HIERARCHICAL_DATA_NB(i) % P % V    &
                   &      )
              deallocate( HIERARCHICAL_DATA_NB(i) % P,        &
                   &      HIERARCHICAL_DATA_NB(i) % R )
              if(associated(HIERARCHICAL_DATA_NB(i) % k_or_q_vec)) deallocate(HIERARCHICAL_DATA_NB(i) % k_or_q_vec)


              deallocate(HIERARCHICAL_DATA_NB(i) % COMM_TABLE % NEIBPE,       &
                   &     HIERARCHICAL_DATA_NB(i) % COMM_TABLE % STACK_IMPORT, & 
                   &     HIERARCHICAL_DATA_NB(i) % COMM_TABLE % STACK_EXPORT, & 
                   &     HIERARCHICAL_DATA_NB(i) % COMM_TABLE % NOD_IMPORT,   &
                   &     HIERARCHICAL_DATA_NB(i) % COMM_TABLE % NOD_EXPORT)

              deallocate(HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % NEIBPE, &
                   &     HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % STACK_IMPORT, & 
                   &     HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % STACK_EXPORT, & 
                   &     HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % NOD_IMPORT,   &
                   &     HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % NOD_EXPORT)

           END DO
           LEVEL_NUM = 0
           
           if(associated(coarsest_A_NB)) deallocate(coarsest_A_NB, coarsest_IDX_NB)
        end if

      END SUBROUTINE clear_multi_lev_matrix_VC
      
  end subroutine clear_multi_lev_matrix

end module common_mod
