!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

MODULE count_time_mod
contains

  SUBROUTINE disp_nonzero_data(op, SOLVER_COMM, my_rank, data_kind, val)
    implicit none
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'

    INTEGER(kind=kint), intent(in) :: op, SOLVER_COMM, my_rank, data_kind
    real(kind=kreal), intent(in) :: val
    INTEGER(kind=kint), parameter :: DATA_KINDS = 60
    INTEGER(kind=kint) :: ierr, NPROCS, i, num_of_kinds
    real(kind=kreal) :: data_table(DATA_KINDS)
    real(kind=kreal) :: send_data_sz, send_data_nonzero
    real(kind=kreal) :: recv_data_sz, recv_data_nonzero
    DATA data_table/DATA_KINDS*0/

    !C op:1 store 2 display data 
    SELECT CASE(op)
    CASE(0)
       !C store N
       data_table(2*data_kind-1) = val
    CASE(1)       
       !C store # of nonzeros
       data_table(2*data_kind) = val
    CASE(2)       
       send_data_sz      = data_table(2*data_kind-1)
       send_data_nonzero = data_table(2*data_kind)
       recv_data_sz = 0
       CALL MPI_REDUCE(send_data_sz, recv_data_sz, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, SOLVER_COMM, ierr)
       recv_data_nonzero = 0
       CALL MPI_REDUCE(send_data_nonzero, recv_data_nonzero, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, SOLVER_COMM, ierr)

       if(recv_data_sz == 0) then
          recv_data_nonzero = 0
       else
          recv_data_nonzero = recv_data_nonzero / recv_data_sz
       end if
       
       IF(my_rank == 0) THEN
          WRITE(*,'("*** level ", i3, " ave # of nonzero per row:", 1pe16.6)') &
               & data_kind, recv_data_nonzero
       END IF
    CASE DEFAULT
       STOP 'error in disp_nonzero_data()'

    END SELECT
    
  END SUBROUTINE disp_nonzero_data


  SUBROUTINE store_and_disp_data(op, SOLVER_COMM, my_rank, data_kind, val)
    implicit none
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'

    INTEGER(kind=kint), intent(in) :: op, SOLVER_COMM, my_rank, data_kind
    real(kind=kreal), intent(in) :: val
    INTEGER(kind=kint), parameter :: DATA_KINDS = 70
    INTEGER(kind=kint) :: ierr, NPROCS, i, num_of_kinds
    real(kind=kreal) :: data_table(DATA_KINDS)
    real(kind=kreal) :: send_data, recv_data, recv_data_sum, recv_data_max, recv_data_min
    DATA data_table/DATA_KINDS*0/

    !C op:1 store 2 display data 
    SELECT CASE(op)
    CASE(0)
       data_table(data_kind) = 0
    CASE(1)       
       data_table(data_kind) = val
    CASE(2)       
       send_data = data_table(data_kind)
       recv_data_sum = 0
       CALL MPI_REDUCE(send_data, recv_data_sum, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, SOLVER_COMM, ierr)
       recv_data_max = 0
       CALL MPI_REDUCE(send_data, recv_data_max, 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, SOLVER_COMM, ierr)
       recv_data_min = 0
       CALL MPI_REDUCE(send_data, recv_data_min, 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, SOLVER_COMM, ierr)
       CALL MPI_COMM_SIZE(SOLVER_COMM, NPROCS, ierr)
       IF(my_rank == 0) THEN
          recv_data_sum = recv_data_sum / NPROCS
          WRITE(*,'("***", i3, " ave:", 1pe16.6," max:", 1pe16.6," min:", 1pe16.6)') &
               & data_kind, recv_data_sum, recv_data_max, recv_data_min
       END IF
    CASE DEFAULT
       STOP 'error in store_and_disp_data()'

    END SELECT
    

    
  END SUBROUTINE store_and_disp_data
  
  SUBROUTINE count_time(op, SOLVER_COMM, my_rank, time_kind)
    IMPLICIT NONE
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'

    INTEGER(kind=kint), INTENT(in) :: op, solver_comm, my_rank, time_kind
    !C 31-40: setup time at each lev,  41-50: v-cycle at each lev,  51-60: v-cycle between levels.
    INTEGER(kind=kint), parameter :: TIME_KINDS = 70
    INTEGER(kind=kint) :: ierr, NPROCS, i, num_of_kinds
    REAL   (kind=kreal), SAVE :: time_table(TIME_KINDS), start_table(TIME_KINDS)
    REAL   (kind=kreal) :: recv_time_sum, recv_time_max, recv_time_min, send_time, out_time
    LOGICAL :: prt_ech_PE(TIME_KINDS)
    DATA time_table/TIME_KINDS*0/
    DATA start_table/TIME_KINDS*0/

    !C num_of_kinds must be less than or equal to TIME_KINDS
    num_of_kinds = 18

    !C prt_ech_PE records time_kind at which each PE's time is needed.
    prt_ech_PE = .FALSE.

    
    !C op:1 start 2 stop 3 print time 0 reset time_table
    SELECT CASE(op)
    CASE(0)
       time_table(time_kind) = 0
       
    CASE(1)
       start_table(time_kind) = MPI_WTIME()

    CASE(2)
       IF(start_table(time_kind) /= 0) THEN
          time_table(time_kind) = time_table(time_kind) + &
               &                  MPI_WTIME() - start_table(time_kind)
          start_table(time_kind) = 0
       ELSE
          WRITE(*,*) "timer error...:", time_kind, start_table(time_kind)
          STOP "timer error!!"
       END IF
    CASE(3)
       send_time = time_table(time_kind)
       recv_time_sum = 0
       CALL MPI_REDUCE(send_time, recv_time_sum, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, SOLVER_COMM, ierr)
       recv_time_max = 0
       CALL MPI_REDUCE(send_time, recv_time_max, 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, SOLVER_COMM, ierr)
       recv_time_min = 0
       CALL MPI_REDUCE(send_time, recv_time_min, 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, SOLVER_COMM, ierr)
       CALL MPI_COMM_SIZE(SOLVER_COMM, NPROCS, ierr)
       IF(my_rank == 0) THEN
          recv_time_sum = recv_time_sum / NPROCS
          WRITE(*,'("***", i3, " ave:", 1pe16.6," max:", 1pe16.6," min:", 1pe16.6, " sec")') &
               & time_kind, recv_time_sum, recv_time_max, recv_time_min
       END IF

    CASE(4)
       CALL MPI_COMM_SIZE(SOLVER_COMM, NPROCS, ierr)
       IF(my_rank == 0) THEN
          WRITE(*,*) "# 1:sr_rep 2:sr_aggre 3:sr_elem 4:RAP 5:neib 6:agr 7:sm_agr   #"
          WRITE(*,*) "# 8:smoother 9:v_cycle 10:sgs calculation part or cheby setup #"
          WRITE(*,*) "# 11:matrix transform 12:ParMetis 13:matrix redistribution    #"
          WRITE(*,*) "# 14:Total time in matrix dist 15:calculation in Redist()     #"
          WRITE(*,*) "# 16:memory allocation deallocation time not included in 4-7  #"
          WRITE(*,*) "kind      ave             max             min"
       END IF
       DO i = 1, num_of_kinds
          send_time = time_table(i)

          CALL MPI_barrier(solver_comm, ierr)
          IF(prt_ech_PE(i)) WRITE(*,'( i3,"--#", i3, ":", 1pe16.6, " sec")') my_rank, i, send_time

          recv_time_sum = 0
          CALL MPI_REDUCE(send_time, recv_time_sum, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, SOLVER_COMM, ierr)
          recv_time_max = 0
          CALL MPI_REDUCE(send_time, recv_time_max, 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, SOLVER_COMM, ierr)
          recv_time_min = 0
          CALL MPI_REDUCE(send_time, recv_time_min, 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, SOLVER_COMM, ierr)

          IF(my_rank == 0) THEN
             recv_time_sum = recv_time_sum / NPROCS
             WRITE(*,'("***", i3, 1pe16.6, 1pe16.6,1pe16.6, " sec")') &
                  & i, recv_time_sum, recv_time_max, recv_time_min
          END IF
       END DO
       
    CASE DEFAULT
       STOP 'error start_time stop time'
    END SELECT
  END SUBROUTINE count_time

  SUBROUTINE count_time_val(op, SOLVER_COMM, my_rank, time_kind, val)
    IMPLICIT NONE
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'

    INTEGER(kind=kint), INTENT(in)   :: op, solver_comm, my_rank, time_kind
    REAL   (kind=kreal), INTENT(out) :: val
    INTEGER(kind=kint), parameter :: TIME_KINDS = 10
    INTEGER(kind=kint) :: ierr, NPROCS, i
    REAL   (kind=kreal), SAVE :: time_table(TIME_KINDS), start_table(TIME_KINDS)
    REAL   (kind=kreal) :: recv_time_sum, recv_time_max, recv_time_min, send_time, out_time
    LOGICAL :: prt_ech_PE(TIME_KINDS)
    DATA time_table/TIME_KINDS*0/
    DATA start_table/TIME_KINDS*0/

    !C prt_ech_PE records time_kind at which each PE's time is needed.
    prt_ech_PE = .FALSE.

    
    !C op:1 start 2 stop 3 print time 4 return with val 0 reset time_table
    SELECT CASE(op)
    CASE(0)
       time_table(time_kind) = 0
       
    CASE(1)
       start_table(time_kind) = MPI_WTIME()

    CASE(2)
       IF(start_table(time_kind) /= 0) THEN
          time_table(time_kind) = time_table(time_kind) + &
               &                  MPI_WTIME() - start_table(time_kind)
          start_table(time_kind) = 0
       ELSE
          WRITE(*,*) "timer error...:", time_kind, start_table(time_kind)
          STOP "timer error!!"
       END IF
       
    CASE(3)
       send_time = time_table(time_kind)
       recv_time_sum = 0
       CALL MPI_REDUCE(send_time, recv_time_sum, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, SOLVER_COMM, ierr)
       recv_time_max = 0
       CALL MPI_REDUCE(send_time, recv_time_max, 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, SOLVER_COMM, ierr)
       recv_time_min = 0
       CALL MPI_REDUCE(send_time, recv_time_min, 1, MPI_DOUBLE_PRECISION, MPI_MIN, 0, SOLVER_COMM, ierr)
       CALL MPI_COMM_SIZE(SOLVER_COMM, NPROCS, ierr)
       IF(my_rank == 0) THEN
          recv_time_sum = recv_time_sum / NPROCS
          WRITE(*,'("***", i3, " ave:", 1pe16.6," max:", 1pe16.6," min:", 1pe16.6, " sec")') &
               & time_kind, recv_time_sum, recv_time_max, recv_time_min
       END IF

    CASE(4)
       val = time_table(time_kind)
       
    CASE DEFAULT
       STOP 'error start_time stop time'
    END SELECT
  END SUBROUTINE count_time_val
  
  subroutine get_solver_time(setup_time,solver_time, SOLVER_COMM, my_rank, disp_fl)
    IMPLICIT NONE
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'

    real(kind=kreal),intent(out) :: solver_time, setup_time
    integer, intent(in) :: my_rank, SOLVER_COMM
    logical :: disp_fl
    
    call count_time_val(4, SOLVER_COMM, my_rank, 1,setup_time)
    call count_time_val(4, SOLVER_COMM, my_rank, 2,solver_time)

    if(disp_fl) then
       if(my_rank == 0) write(*,*) "setup time:"
       call count_time_val(3, SOLVER_COMM, my_rank, 1,setup_time)
       if(my_rank == 0) write(*,*) "solver time:"
       call count_time_val(3, SOLVER_COMM, my_rank, 2,solver_time)
    end if
  end subroutine get_solver_time



END MODULE count_time_mod
