#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#define SP 256


__global__ void bicgstab_func_0_kernel(double *p, double *r, double *v, double beta, double omega, int n);
__global__ void bicgstab_func_1_kernel(double *s, double *r, double *v, double alpha, int n);
__global__ void bicgstab_func_2_kernel(double *x, double *pt, double *st, double alpha, double omega, int n);


void bicgstab_func_0(double *dp, double *dr, double *dv, double beta, double omega, int n)
{
	bicgstab_func_0_kernel<<< n / SP + 1, SP >>>(dp, dr, dv, beta, omega, n);
}


void bicgstab_func_1(double *ds, double *dr, double *dv, double alpha, int n)
{
	bicgstab_func_1_kernel<<< n / SP + 1, SP >>>(ds, dr, dv, alpha, n);
}


void bicgstab_func_2(double *dx, double *dpt, double *dst, double alpha, double omega, int n)
{
	bicgstab_func_2_kernel<<< n / SP + 1, SP >>>(dx, dpt, dst, alpha, omega, n);
}


__global__
void bicgstab_func_0_kernel(double *p, double *r, double *v, double beta, double omega, int n)
{
	const int id = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		p[id] = r[id] + beta * (p[id] - omega * v[id]);
	}
}

__global__
void bicgstab_func_1_kernel(double *s, double *r, double *v, double alpha, int n)
{
	const int id = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		s[id] = r[id] - alpha * v[id];
	}
}

__global__
void bicgstab_func_2_kernel(double *x, double *pt, double *st, double alpha, double omega, int n)
{
	const int id = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < n){
		x[id] += alpha * pt[id] + omega * st[id];
	}
}

