#include "cuda_gauss_seidel.h"

#include <algorithm>
#include <cmath>
#include <cstring>
#include <iostream>
#include <new>
#include <vector>

using std::bad_alloc;
using std::cout;
using std::endl;

using std::vector;
using std::pair;
using std::make_pair;
using std::stable_sort;

int set_color(lu_crs &mat, double alpha, int *color);
bool paint_color(lu_crs &mat, int inc, double alpha, vector< pair<int, int> > &order, int *color);
int set_histogram(int *data, int data_length, int *histogram, int histogram_length);
void print_histogram(int *histogram, int histogram_length);


int multi_color_gauss_seidel_setup(lu_crs &src_mat, lu_icrs &dst_mat, double alpha,  int*b_n, int**b_cnt, int**b_ind, int **order)
{
	try{
		*b_ind = new int[src_mat.n];
		*b_n   = set_color(src_mat, alpha, *b_ind);

		if(*b_n){
			*b_cnt = new int[*b_n];
		
			set_histogram(*b_ind, src_mat.n, *b_cnt, *b_n);
//			print_histogram(*b_cnt, *b_n);
		}else{
			*b_n = 1;
			*b_cnt = new int[*b_n];
			*b_cnt[0] = 0;
		}

		if( convert_lu_crs_to_lu_icrs(src_mat, dst_mat, *b_n, *b_cnt, *b_ind, order) ){
			throw;
		}
	}
	catch(...){
		cout << "multi color gauss seidel setup expection." << endl;
		return 1;
	}

	return 0;
}


int set_color(lu_crs &mat, double alpha, int *color)
{
	int i;
	int edge, inc;

	vector< pair<int, int> > order;

	try{
		memset(color, 0, sizeof(int) * mat.n);

		for(i=0; i<mat.n; i++){
			edge = mat.inl[i+1] + mat.inu[i+1] - mat.inl[i] - mat.inu[i];
			order.push_back(make_pair(edge, i));
		}

		if(alpha != 0.0){
			stable_sort(order.rbegin(), order.rend());
		}

		inc = 1;
		while(paint_color(mat, inc++, alpha, order, color));
	
		for(i=0; i<mat.n; i++){
			color[i]--;
		}
	
		inc = inc - 2;
	
		order.clear();

	}catch(...)
	{
		cout << "color memset expection : " << mat.n << endl;
		throw;
	}

	return inc;
}


bool paint_color(lu_crs &mat, int inc, double alpha, vector< pair<int, int> >  &order, int *color)
{
	int i, j;
	int target, start, end, index;
	double d, val;
	bool colorless_check;

	colorless_check = false;
	for(i=0; i<mat.n; i++){
		target = order[i].second;
		if(color[target] == 0){
			colorless_check = true;

			d = mat.d[target];

			start = mat.inl[ target ];
			end   = mat.inl[target+1];
			for(j=start; j<end; j++){
				index = mat.ial[j];
				val = mat.al[j];

				if(index < mat.n){
					if(color[index] == inc){
						if( -val / d >= alpha || alpha == 0 ){
							goto _next_node;
						}
					}
				}
			}

			start = mat.inu[ target ];
			end   = mat.inu[target+1];
			for(j=start; j<end; j++){
				index = mat.iau[j];
				val = mat.au[j];

				if(index < mat.n){
					if(color[index] == inc){
						if( -val / d >= alpha || alpha == 0 ){
							goto _next_node;
						}
					}
				}
			}

			color[target] = inc;
		}

		_next_node:;
	}

	return colorless_check;
}


int set_histogram(int *data, int data_length, int *histogram, int histogram_length)
{
	int i;

	memset(histogram, 0, sizeof(int) * histogram_length);
	for(i=0; i<data_length; i++){
		histogram[ data[i] ]++;
	}
	return 0;
}


void print_histogram(int *histogram, int histogram_length)
{
	int i, j;
	int max;
	
	cout << "::::::::::::::::::::::::::::::::::::::::::" << endl;
	max = 0;
	for(i=0; i<histogram_length; i++){
		max = histogram[i] > max ? histogram[i] : max;
	}

	for(i=0; i<histogram_length; i++){
		cout << i << ":";
		for(j=0; j<32*histogram[i]/max; j++){
			cout << "*";
		}
		cout << histogram[i] << endl;
	}
}

