#include "boader_zone.h"
#include "lab_matrix.h"
#include "my_mpi_cuda.h"

#include <cuda_runtime.h>
#include <device_launch_parameters.h>

__global__ void multi_color_gauss_seidel_kernel(lu_icrs_part a, double *x, double *b, double omega);

#define SP 256

void multi_color_gauss_seidel(lu_icrs &a, double *x, double *b, double omega)
{
	int i;
	for(i=0; i<a.size(); i++){
		multi_color_gauss_seidel_kernel<<< a[i].n / SP + 1, SP >>>(a[i], x, b, omega);
	}
}


void multi_color_gauss_seidel(lu_icrs &a, double *x, double *b, double omega, boader_zone *area_outer, boader_zone *area_inner)
{
	int i, size;
	size = a.size();
	for(i=size/2; i<size; i++) multi_color_gauss_seidel_kernel<<< a[i].n / SP + 1, SP >>>(a[i], x, b, omega);
	icomm_margin(x, area_outer, area_inner, COMM_MARGIN_EQL);
	for(i=0; i<size/2; i++) multi_color_gauss_seidel_kernel<<< a[i].n / SP + 1, SP >>>(a[i], x, b, omega);
	icomm_margin_wait(x, area_outer, area_inner, COMM_MARGIN_EQL);
}


__global__
void multi_color_gauss_seidel_kernel(lu_icrs_part a, double *x, double *b, double omega)
{
	__shared__ double s_val[SP];
	__shared__ double s_x[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_b, local_x;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_x = x[index];
		local_b = b[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_x[tid] = x[s_col[tid]];

			temp -= s_val[tid] * s_x[tid];
		}

		temp += local_b;
		temp /= local_d;

		index = id + a.row;
		x[index] = local_x + omega * (temp - local_x);
	}
}

