#include "cuda_jacobi.h"
#include "my_mpi_cuda.h"

#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#define SP 256

__global__ void jacobi_kernel(lu_icrs_part a, double *old_x, double *new_x, double *b, double omega);


void jacobi(lu_icrs_part &a, double *old_x, double *new_x, double *b, double omega)
{
	jacobi_kernel<<< a.n / SP + 1, SP >>>(a, old_x, new_x, b, omega);
	return;
}


void jacobi(lu_icrs &a, double *old_x, double *new_x, double *b, double omega)
{
	jacobi_kernel<<< a[0].n / SP + 1, SP >>>(a[0], old_x, new_x, b, omega);
	return;
}


void jacobi(lu_icrs &a, double *old_x, double *new_x, double *b, double omega, boader_zone *area_outer, boader_zone * area_inner)
{
	jacobi_kernel<<< a[1].n / SP + 1, SP >>>(a[1], old_x, new_x, b, omega);
	icomm_margin(new_x, area_outer, area_inner, COMM_MARGIN_EQL);
	jacobi_kernel<<< a[0].n / SP + 1, SP >>>(a[0], old_x, new_x, b, omega);
	icomm_margin_wait(new_x, area_outer, area_inner, COMM_MARGIN_EQL);
	return;
}


__global__
void jacobi_kernel(lu_icrs_part a, double *old_x, double *new_x, double *b, double omega)
{
	__shared__ double s_val[SP];
	__shared__ double s_x[SP];
	__shared__ int    s_col[SP];

	int i, index, local_w;
	double temp, local_d, local_b, local_x;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		index = id + a.row;
		local_w = a.cnt[id];
		local_d = a.d[id];
		local_x = old_x[index];
		local_b = b[index];

		temp = 0.0;

		index = id;
		for(i=0; i<local_w; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_x[tid] = old_x[s_col[tid]];

			//if(isnan(s_x[tid])){
			//	printf("%d:%e\n", s_col[tid], s_x[tid]);
			//}

			temp -= s_val[tid] * s_x[tid];
		}

		temp += local_b;
		temp /= local_d;

		index = id + a.row;
		new_x[index] = local_x + omega * (temp - local_x);
	}
}
