module cycle_shape_tuning
  implicit none
  include 'precision.inc'
  
  integer(kind=kint), parameter :: precision_sz = 6


  type sol_type
     !C sol == 0,1,2  : NO, CG, BICGSTAB
     integer(kind=kint) :: sol  
     integer(kind=kint) :: v_pre_lev
  end type sol_type
  
  type V_type
     !C viter > 0 shows the paraameter's validity
     integer(kind=kint) :: viter
     integer(kind=kint) :: smiter
     !C coarser_level_precision = 0 means it doesn't use the coarser leveels.
     integer(kind=kint) :: coarser_level_precision
     real(kind=kreal)   :: omega
     real(kind=kreal)   :: time
  end type V_type
  
  real(kind=kreal), pointer :: precision_arry(:)
  
  !C  type(V_type), pointer :: record_data(precision_sz, max_lev_sz)
  type(V_type),   pointer :: record_data(:, :)
  
  !C  type(sol_type), pointer :: sol_data(precision_sz)
  type(sol_type), pointer :: sol_data(:)
  
contains
  
  subroutine initialize_record_data(LEVEL_NUM)
    implicit none
    integer(kind=kint), intent(in) :: LEVEL_NUM
    integer(kind=kint) :: i
    
    allocate(record_data(precision_sz, LEVEL_NUM))
    allocate(precision_arry(precision_sz))

!!$    precision_arry(1) = 1.0E-1
!!$    do i = 2, precision_sz
!!$       precision_arry(i) = precision_arry(i-1)*0.333
!!$    end do
    precision_arry(1) = 5.0E-1
    precision_arry(2) = 1.0E-1
    precision_arry(3) = 5.0E-2
    precision_arry(4) = 1.0E-2
    precision_arry(5) = 1.0E-4
    precision_arry(6) = 1.0E-7

    
    !C viter == 0 means there are no parameter combination satisfing the conv criterion
    record_data(:, :) % viter = 0
    record_data(:, :) % time  = 0.0

!!$    do i = 1, LEVEL_NUM
!!$       record_data(1, i) % smiter = i
!!$       record_data(1, i) % coarser_level_precision = 1
!!$       record_data(1, i) % viter = 1
!!$       record_data(1, i) % omega = 0.8
!!$          
!!$       if(i==LEVEL_NUM) record_data(1, i) % coarser_level_precision = 0
!!$    end do

  end subroutine initialize_record_data

  subroutine record_best_cycle(LEVEL_NUM, SOLVER_COMM, Temp, WS, WR, dim, my_rank)
    use data_structure_for_AMG
    implicit none
    integer(kind=kint), intent(in) :: LEVEL_NUM, SOLVER_COMM, dim, my_rank
    real(kind=kreal) :: Temp(:), WS(:), WR(:)
    integer(kind=kint) :: l, p

    !C B and X are not allocated at Level 1.
    allocate(HIERARCHICAL_DATA(1) % B(HIERARCHICAL_DATA(1) % NP))
    allocate(HIERARCHICAL_DATA(1) % X(HIERARCHICAL_DATA(1) % NP))

    do l = LEVEL_NUM, 1, -1
       do p = 1, precision_sz
          if(my_rank == 0) write(*,*) "LEVEL:", l, "precision:", p
          call tune_at_some_lev(l, LEVEL_NUM, p, SOLVER_COMM, Temp, WS, WR, dim, my_rank)
       end do
    end do
    
    deallocate(HIERARCHICAL_DATA(1) % B, HIERARCHICAL_DATA(1) % X)

    
  end subroutine record_best_cycle


  subroutine tune_at_some_lev(nth_lev, LEVEL_NUM, p, SOLVER_COMM, Temp, WS, WR, dim, my_rank)
    use data_structure_for_AMG
    use terminal_condition, only :terminal_cond_val_set, BNRM2_set
    use count_time_mod
    use AMG_cycle, only:  cal_res_vec_at_nth_lev
    USE solver_SR2
      
    implicit none
    
    include 'mpif.h'

    integer(kind=kint), intent(in) :: nth_lev, p, LEVEL_NUM, SOLVER_COMM, dim, my_rank
    integer(kind=kint) :: smiter_dum, smiter, coarser_level_precision_dum, coarser_level_precision
    integer(kind=kint) :: ierr, omega_k, viter, NEIBPETOT
    real(kind=kreal) :: omega, s,e,time, Temp(:), WS(:), WR(:)
    real(kind=kreal),allocatable :: correct_vec(:)
    logical :: CONV_FL, flag
    
    allocate(correct_vec(dim))

!C    call terminal_cond_val_set(precision_arry(p))

    do smiter_dum = 1, 1
       do omega_k = 0,2
          omega = 0.8+0.2*omega_k
!!$          !C for various_v_cycle_BiCGSTAB_solver sub.
!!$          do viter = 1,3

          do coarser_level_precision_dum = 1, precision_sz+1
             
             coarser_level_precision = coarser_level_precision_dum
             if(coarser_level_precision_dum == precision_sz+1) coarser_level_precision = 0

             !C if there is coaser_level's data, then it check the solver using that coarser data.
             flag = .false.
             flag =(coarser_level_precision == 0)
             if(coarser_level_precision > 0 .and. nth_lev < LEVEL_NUM) &
                  & flag = (record_data(coarser_level_precision, nth_lev+1) % viter > 0)
             
             if(flag) then
                call random_number(correct_vec)
                NEIBPETOT = HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NEIBPETOT
                IF(NEIBPETOT > 0) then
                   CALL SOLVER_SEND_RECV2(&
                        & HIERARCHICAL_DATA(nth_lev) % NP, &
                        & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % STACK_EXPORT(NEIBPETOT),  &
                        & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % STACK_IMPORT(NEIBPETOT),  &
                        & NEIBPETOT,               &
                        & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NEIBPE,                   &
                        & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % STACK_IMPORT,             &
                        & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NOD_IMPORT,               &
                        & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % STACK_EXPORT,             &
                        & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NOD_EXPORT,               &
                        &  WS, WR, correct_vec, SOLVER_COMM, my_rank)
                end IF

                HIERARCHICAL_DATA(nth_lev) % X = 0                
                call cal_res_vec_at_nth_lev(HIERARCHICAL_DATA(nth_lev) % X, correct_vec, HIERARCHICAL_DATA(nth_lev) % B, nth_lev)
                
                HIERARCHICAL_DATA(nth_lev) % B = - HIERARCHICAL_DATA(nth_lev) % B 


                smiter = smiter_dum
                !C when coarser_level_precision = 0, it check strong smoother case such as 30 iterations.
                if(coarser_level_precision == 0 .and. smiter == 1) smiter =30
                
                !C- measuring time is also done.
                call MPI_Barrier(SOLVER_COMM, ierr)
                time = MPI_WTIME()

                call various_v_cycle_solver(&
                     & HIERARCHICAL_DATA(nth_lev) % B, &
                     & HIERARCHICAL_DATA(nth_lev) % X, &
                     & Temp, nth_lev, smiter, omega, coarser_level_precision, viter, &
                     & CONV_FL, WS, WR, dim, SOLVER_COMM, my_rank, time, record_data(p,nth_lev) % time, correct_vec, &
                     & p)

!!$                call various_v_cycle_BiCGSTAB_solver(&
!!$                     & HIERARCHICAL_DATA(nth_lev) % B, &
!!$                     & HIERARCHICAL_DATA(nth_lev) % X, &
!!$                     & Temp, nth_lev, smiter, omega, coarser_level_precision, viter, &
!!$                     & CONV_FL, WS, WR, dim, SOLVER_COMM, my_rank, time, record_data(p,nth_lev) % time, correct_vec, &
!!$                     & p, LEVEL_NUM)
                
                if(CONV_FL) then
                   call MPI_Barrier(SOLVER_COMM, ierr)
                   time = MPI_WTIME() - time 
                   
                   call MPI_BCAST(time, 1, MPI_DOUBLE_PRECISION, 0, SOLVER_COMM, ierr)
                   
                   if((time < record_data(p,nth_lev) % time .or. record_data(p,nth_lev) % time == 0) &
                        & .and. CONV_FL) then
                      record_data(p,nth_lev) % time = time
                      
                      record_data(p,nth_lev) % smiter = smiter
                      record_data(p,nth_lev) % omega = omega
                      record_data(p,nth_lev) % coarser_level_precision = coarser_level_precision
                      record_data(p,nth_lev) % viter = viter
                   end if
                end if
             end if
          end do
!!$          end do
       end do
    end do
    
  end subroutine tune_at_some_lev


  subroutine various_v_cycle_solver(B, X, Temp, nth_lev, smiter, omega, coarser_level_precision, viter, &
       & CONV_FL, WS, WR, dim, SOLVER_COMM, my_rank, start_time, min_time, correct_vec, p)
    USE data_structure_for_AMG
    use AMG_cycle, only: cal_res_vec_at_nth_lev
    
    implicit none
    include 'mpif.h'
      
    INTEGER(kind=kint ) :: SOLVER_COMM, my_rank, dim, nth_lev, smiter, coarser_level_precision, viter, iter, ierr, p, i
    REAL   (kind=kreal) :: B(:), X(:), TEMP(:), WS(:), WR(:), omega, RESID, start_time, t, min_time, v, correct_vec(:)
    REAL   (kind=kreal) :: BNRM2, BNRM20, diff_NRM2, diff_NRM20
    
    logical, intent(out) :: CONV_FL 

    integer(kind=kint), parameter :: max_iter=50



    BNRM20 = 0
    DO i = 1, HIERARCHICAL_DATA(nth_lev) % N
       BNRM20 = BNRM20 + correct_vec(i) ** 2
    END DO
    call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
         &                    MPI_SUM, SOLVER_COMM, ierr)


    CONV_FL = .false.
    DO iter = 1, max_iter
       call v_cycle_at_each_lev(B, X, Temp, nth_lev, smiter, omega, coarser_level_precision, 1, &
            & WS, WR, dim, SOLVER_COMM, my_rank)
       
       diff_NRM20 = 0
       DO i = 1, HIERARCHICAL_DATA(nth_lev) % N
          diff_NRM20 = diff_NRM20 + (X(i)-correct_vec(i)) ** 2
       END DO
       call MPI_allREDUCE (diff_NRM20, diff_NRM2, 1, MPI_DOUBLE_PRECISION,       &
            &                    MPI_SUM, SOLVER_COMM, ierr)
       
       !C call cal_res_vec_at_nth_lev(B, X, Temp, nth_lev)
       !Ccall terminal_cond_check(CONV_FL, RESID, Temp, HIERARCHICAL_DATA(nth_lev) % N, SOLVER_COMM)

       RESID = dsqrt(diff_nrm2/BNRM2)
       conv_FL = (RESID < precision_arry(p))
       
       if(CONV_FL) exit
       
       !C every 4 iteration exec time is checked..
!!$       if(mod(iter, 1) == 0 .and. min_time > 0) then
!!$          t =MPI_WTIME() -start_time 
!!$          call MPI_BCAST(t, 1, MPI_DOUBLE_PRECISION, 0, SOLVER_COMM, ierr)
!!$          if(t > min_time) exit
!!$       end if
       
    END DO
    viter=iter
    
  end subroutine various_v_cycle_solver


  subroutine various_v_cycle_BiCGSTAB_solver(B, X, Temp, nth_lev, smiter, smomega, coarser_level_precision, viter, &
       & CONV_FL, WS, WR, dim, SOLVER_COMM, my_rank, start_time, min_time, correct_vec, p, LEVEL_NUM)
    USE data_structure_for_AMG
    use AMG_cycle, only: cal_res_vec_at_nth_lev
    !$ use omp_lib
    
    implicit none
    include 'mpif.h'
      
    INTEGER(kind=kint ) :: SOLVER_COMM, my_rank, dim, nth_lev, smiter, coarser_level_precision, viter, iter, ierr, p
    INTEGER(kind=kint ) :: LEVEL_NUM
    REAL   (kind=kreal) :: B(:), X(:), TEMP(:), WS(:), WR(:), smomega, RESID, start_time, t, min_time, v, correct_vec(:)
    REAL   (kind=kreal) :: BNRM2, BNRM20, diff_NRM2, diff_NRM20
    REAL   (kind=kreal) :: OMEGA,ALPHA,RHO1
    
    logical, intent(out) :: CONV_FL 

    integer(kind=kint), parameter :: max_iter=50

    REAL   (kind=kreal),allocatable,save :: WW(:,:)
    integer(kind=kint) :: NP_max, R,RT,PT,ST
    integer(kind=kint) :: i,j,k
    REAL   (kind=kreal) :: WVAL

    if(.not.allocated(WW)) then
       !C-- NP_max is equal to the max size of unknowns on all levels.
       call max_NP_on_all_levs(NP_max, HIERARCHICAL_DATA(1) % NP, HIERARCHICAL_DATA, LEVEL_NUM)
       allocate (WW(NP_max,8))       
    end if
    
    R=  1;    RT= 2
    PT= 4;    ST= 6
       
    !$omp parallel do private(k,WVAL,i)
    do j= 1, HIERARCHICAL_DATA(nth_lev) % N
       WVAL = B(j) - HIERARCHICAL_DATA(nth_lev) % D(j) * X(j)

       do k= HIERARCHICAL_DATA(nth_lev) % INL(j-1)+1, HIERARCHICAL_DATA(nth_lev) % INL(j)
          i= HIERARCHICAL_DATA(nth_lev) % IAL(k)
          WVAL= WVAL - HIERARCHICAL_DATA(nth_lev) % AL(k)* X(i)
       enddo
       
       do k= HIERARCHICAL_DATA(nth_lev) % INU(j-1)+1, HIERARCHICAL_DATA(nth_lev) % INU(j)
          i= HIERARCHICAL_DATA(nth_lev) % IAU(k)
          WVAL= WVAL - HIERARCHICAL_DATA(nth_lev) % AU(k)* X(i)
       enddo

       WW(j  ,R )= WVAL
       WW(j  ,RT)= WVAL

       WW(j  ,PT)= 0.d0
       WW(j  ,ST)= 0.d0
    enddo
    !$omp end parallel do


    BNRM20 = 0
    DO i = 1, HIERARCHICAL_DATA(nth_lev) % N
       BNRM20 = BNRM20 + B(i) ** 2
    END DO
    call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
         &                    MPI_SUM, SOLVER_COMM, ierr)


    CONV_FL = .false.
    DO iter = 1, max_iter
       CALL AMG_BiCGSTAB_1_iter( &
            &                  HIERARCHICAL_DATA(nth_lev) % N,                 &
            &                  HIERARCHICAL_DATA(nth_lev) % NP,                &
            &                  HIERARCHICAL_DATA(nth_lev) % NPL,               &
            &                  HIERARCHICAL_DATA(nth_lev) % NPU,               &
            &                  HIERARCHICAL_DATA(nth_lev) % D,                 &
            &                  HIERARCHICAL_DATA(nth_lev) % AL,                &
            &                  HIERARCHICAL_DATA(nth_lev) % INL,               &
            &                  HIERARCHICAL_DATA(nth_lev) % IAL,               &
            &                  HIERARCHICAL_DATA(nth_lev) % AU,                &
            &                  HIERARCHICAL_DATA(nth_lev) % INU,               &
            &                  HIERARCHICAL_DATA(nth_lev) % IAU,               &
            &                  B,  X,   iter,                                  &
            &                  my_rank, &
            &                  HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NEIBPETOT,         &
            &                  HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NEIBPE,            &
            &                  HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % STACK_IMPORT,      &
            &                  HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NOD_IMPORT,        &
            &                  HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % STACK_EXPORT,      &
            &                  HIERARCHICAL_DATA(nth_lev) % COMM_TABLE % NOD_EXPORT,        &
            &                  SOLVER_COMM, LEVEL_NUM, WS, WR, WW, dim,        &
            &                  OMEGA,ALPHA,RHO1,                               &
            &                  nth_lev, smiter, smomega,                       &
            &                  coarser_level_precision,viter)     
       
       
       
       diff_NRM20 = 0
       DO i = 1, HIERARCHICAL_DATA(nth_lev) % N
          diff_NRM20 = diff_NRM20 + (X(i)-correct_vec(i)) ** 2
       END DO
       call MPI_allREDUCE (diff_NRM20, diff_NRM2, 1, MPI_DOUBLE_PRECISION,       &
            &                    MPI_SUM, SOLVER_COMM, ierr)
       
       !C call cal_res_vec_at_nth_lev(B, X, Temp, nth_lev)
       !Ccall terminal_cond_check(CONV_FL, RESID, Temp, HIERARCHICAL_DATA(nth_lev) % N, SOLVER_COMM)


       RESID = dsqrt(diff_nrm2/BNRM2)
       conv_FL = (RESID < precision_arry(p))

!C       if(my_rank ==0) write(*,*)iter,":",RESID
       
       if(CONV_FL) exit
       
       !C every 4 iteration exec time is checked..
!!$       if(mod(iter, 1) == 0 .and. min_time > 0) then
!!$          t =MPI_WTIME() -start_time 
!!$          call MPI_BCAST(t, 1, MPI_DOUBLE_PRECISION, 0, SOLVER_COMM, ierr)
!!$          if(t > min_time) exit
!!$       end if
       
    END DO
    
    contains
      subroutine AMG_BiCGSTAB_1_iter( &
           &                  N, NP,  NPL, NPU,                               &
           &                  D,  AL, INL, IAL, AU, INU, IAU,                 &
           &                  B,  X,   ITER,                                  &
           &                  my_rank, NEIBPETOT, NEIBPE,                     &
           &                  STACK_IMPORT, NOD_IMPORT,                       &
           &                  STACK_EXPORT, NOD_EXPORT,                       &
           &                  SOLVER_COMM, LEVEL_NUM, WS, WR, WW, WSIZE,      &
           &                  OMEGA,ALPHA,RHO1,                               &
           &                  nth_lev, smiter, smomega, coarser_level_precision,&
           &                  viter)     
        
        use solver_SR2
        !$ USE omp_lib

        implicit none
!C        include 'mpif.h'
        integer(kind=kint ), intent(in   )::  N, NP, NPU, NPL

        integer(kind=kint ), intent(in   )::  ITER
        integer(kind=kint ), intent(in   )::  my_rank, LEVEL_NUM
        integer            , intent(in)   :: SOLVER_COMM
        real   (kind=kreal),  intent(inout)::  B(:), X(:), AL(:), AU(:), D(:)
        integer(kind=kint ), intent(in) :: INU(0:), INL(0:), IAL(:), IAU(:)
        integer(kind=kint ) :: NEIBPETOT, NEIBPE      (:)
        integer(kind=kint ) :: STACK_IMPORT(0:), NOD_IMPORT  (:)
        integer(kind=kint ) :: STACK_EXPORT(0:), NOD_EXPORT  (:)

        real   (kind=kreal) :: WS(:), WR(:), WW(:,:)
        integer(kind=kint ) :: WSIZE
        
        integer(kind=kint ) :: R,RT,P,PT,S,ST,T,V
        integer(kind=kint ) :: i,j

        integer(kind=kint ) :: nth_lev, smiter, viter, coarser_level_precision
        real   (kind=kreal) :: smomega
        real   (kind=kreal) :: RHO0,RHO,OMEGA,ALPHA,RHO1,BETA,C01,C02,CG(2),C0(2),C20,C2
        
        R=  1;    RT= 2
        P=  3;    PT= 4
        S=  5;    ST= 6
        T=  7;    V=  8

        
        !C
        !C +-----------------+
        !C | RHO= {r}{r_tld} |
        !C +-----------------+
        !C===
        
        RHO0= 0.d0
        !$omp parallel do reduction(+:RHO0)
        do j= 1, N
           RHO0= RHO0+WW(j,RT)*WW(j,R)
        enddo
        !$omp end parallel do

        call MPI_allREDUCE (RHO0, RHO, 1, MPI_DOUBLE_PRECISION,           &
             &                    MPI_SUM, SOLVER_COMM, ierr)
        !C===

        !C
        !C +----------------------------------------+
        !C | BETA= (RHO/RHO1) * (ALPHA/OMEGA)       |
        !C | {p} = {r} + BETA * ( {p} - OMEGA*{v} ) |
        !C +----------------------------------------+
        !C===
        if (iter.gt.1) then
           BETA= (RHO/RHO1) * (ALPHA/OMEGA)
           !$omp parallel do
           do j= 1, N
              WW(j  ,P)= WW(j  ,R)+BETA*(WW(j  ,P)-OMEGA*WW(j  ,V))
           enddo
           !$omp end parallel do
        else
           !$omp parallel do
           do j= 1, N
              WW(j  ,P)= WW(j  ,R)
           enddo
           !$omp end parallel do
        endif
        !C===

        !C
        !C +--------------------+
        !C | {p_tld}= [Minv]{p} |
        !C +--------------------+
        !C===

        !C
        !C-- INTERFACE data EXCHANGE

        call SOLVER_SEND_RECV2                                                   &
             &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
             &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,P) , SOLVER_COMM,      &
             &     my_rank)

        
        !$omp parallel do
        do j= 1, NP
           WW(j,PT) = 0.0d0
        end do
        !$omp end parallel do
        
!!$        CALL amg_cycles(WW(:,P), WW(:,PT), LEVEL_NUM, SOLVER_COMM,&
!!$             & my_rank, WS, WR, WW(:,V), PETOT, WSIZE, NP)
        do j = 1, viter
           call v_cycle_at_each_lev(WW(:,P), WW(:,PT), WW(:,V), nth_lev, smiter, smomega, &
                & coarser_level_precision, viter, WS, WR, WSIZE, SOLVER_COMM, my_rank)
        end do

        !C
        !C +-------------------+
        !C | {v} = [A] {p_tld} |
        !C +-------------------+
        !C===

        call SOLVER_SEND_RECV2                                                   &
             &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
             &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,PT) , SOLVER_COMM,     &
             &     my_rank)

        !C
        !C-- BEGIN calculation
        !$omp parallel do private(WVAL,k,i)
        do j= 1, N
           WVAL= D(j)*WW(j,PT)
           do k= INL(j-1)+1, INL(j)
              i= IAL(k)
              WVAL= WVAL + AL(k)*WW(i,PT)
           enddo

           do k= INU(j-1)+1, INU(j)
              i= IAU(k)
              WVAL= WVAL + AU(k)*WW(i,PT)
           enddo

           WW(j,V)= WVAL
        enddo
        !$omp end parallel do
        !C===

        !C
        !C-- calc. ALPHA
        C20= 0.d0
        !$omp parallel do reduction(+:C20)
        do j= 1, N
           C20= C20 + WW(j,RT)*WW(j,V) 
        enddo
        !$omp end parallel do

        call MPI_allREDUCE (C20, C2, 1, MPI_DOUBLE_PRECISION,             &
             &                    MPI_SUM, SOLVER_COMM, ierr) 
        ALPHA= RHO / C2

        !C
        !C-- {s}= {r} - ALPHA*{V}
        !$omp parallel do 
        do j= 1, N
           WW(j  ,S)= WW(j  ,R) - ALPHA*WW(j  ,V)
        enddo
        !$omp end parallel do 

        !C
        !C +--------------------+
        !C | {s_tld}= [Minv]{s} |
        !C +--------------------+
        !C===

        !C
        !C-- INTERFACE data EXCHANGE

        call SOLVER_SEND_RECV2                                                   &
             &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
             &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,S) , SOLVER_COMM,      &
             &     my_rank)


        !$omp parallel do 
        do j = 1, NP
           WW(j, ST) = 0.0d0
        end do
        !$omp end parallel do 

!!$        CALL amg_cycles(WW(:,S), WW(:,ST), LEVEL_NUM, SOLVER_COMM,&
!!$             & my_rank, WS(1:WSIZE), WR(1:WSIZE), WW(:,T), PETOT, WSIZE, NP)
        do j = 1, viter
           call v_cycle_at_each_lev(WW(:,S), WW(:,ST), WW(:,T), nth_lev, smiter, smomega, &
                & coarser_level_precision, viter, WS, WR, WSIZE, SOLVER_COMM, my_rank)
        end do

        !C
        !C +------------------+
        !C | {t} = [A]{s_tld} |
        !C +------------------+
        !C===

        !C
        !C-- INTERFACE data EXCHANGE
        call SOLVER_SEND_RECV2                                                   &
             &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
             &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,ST) , SOLVER_COMM,     &
             &     my_rank)

        !C
        !C-- BEGIN calculation
        !$omp parallel do private(WVAL,i,k)
        do j= 1, N
           WVAL= D(j)*WW(j  ,ST)

           do k= INL(j-1)+1, INL(j)
              i= IAL(k)
              WVAL= WVAL + AL(k)*WW(i  ,ST)
           enddo

           do k= INU(j-1)+1, INU(j)
              i= IAU(k)
              WVAL= WVAL + AU(k)*WW(i  ,ST)
           enddo

           WW(j  ,T)= WVAL
        enddo
        !$omp end parallel do 
        !C===

        !C
        !C +----------------------------+
        !C | OMEGA= ({t}{s}) / ({t}{t}) |
        !C +----------------------------+
        !C===
        C01= 0.d0
        C02= 0.d0
        CG(1)= 0.d0
        CG(2)= 0.d0

        !$omp parallel do reduction(+:C01,C02)
        do j= 1, N
           C01= C01 + WW(j,T)*WW(j,S)
           C02= C02 + WW(j,T)*WW(j,T)
        enddo
        !$omp end parallel do

        C0(1) = C01; C0(2)=C02
        call MPI_allREDUCE (C0, CG, 2, MPI_DOUBLE_PRECISION,              &
             &                    MPI_SUM, SOLVER_COMM, ierr)
        OMEGA= CG(1) / CG(2)
        !C===

        !C
        !C +----------------+
        !C | update {x},{r} |
        !C +----------------+
        !C===
!C        DNRM20= 0.d0
        !$omp parallel do
        do j= 1, N
           X (j  )= X(j  ) + ALPHA*WW(j  ,PT) + OMEGA*WW(j  ,ST)
           WW(j  ,R)= WW(j  ,S) - OMEGA*WW(j  ,T)
!!$          DNRM20= DNRM20+WW(j,S)**2
        enddo
        !$omp end parallel do

        RHO1= RHO
        
      end subroutine AMG_BiCGSTAB_1_iter
      
  end subroutine various_v_cycle_BiCGSTAB_solver
  

  
  recursive subroutine v_cycle_at_each_lev(B, X, Temp, nth_lev, smiter, omega, coarser_level_precision, viter, &
       & WS, WR, dim, SOLVER_COMM, my_rank)
    use AMG_cycle 
    use matrix_type_mod, only:matrix, set_mat_data

    USE data_structure_for_AMG
    implicit none

    type(matrix) :: mat
      
    INTEGER(kind=kint ) :: SOLVER_COMM, my_rank, dim, nth_lev, smiter, coarser_level_precision, viter
    REAL   (kind=kreal) :: B(:), X(:), TEMP(:), WS(:), WR(:), omega 

    INTEGER(kind=kint ) :: smiter_c, coarser_level_precision_c, viter_c
    REAL   (kind=kreal) :: omega_c

    INTEGER(kind=kint ) :: iter

         call set_mat_data(mat, &
              & HIERARCHICAL_DATA(nth_lev) % N,  &
              & HIERARCHICAL_DATA(nth_lev) % NP, &
              & HIERARCHICAL_DATA(nth_lev) % NPL,&
              & HIERARCHICAL_DATA(nth_lev) % NPU,&
              & HIERARCHICAL_DATA(nth_lev) % D,  &
              & HIERARCHICAL_DATA(nth_lev) % AL, &
              & HIERARCHICAL_DATA(nth_lev) % INL,&
              & HIERARCHICAL_DATA(nth_lev) % IAL,&
              & HIERARCHICAL_DATA(nth_lev) % AU, &
              & HIERARCHICAL_DATA(nth_lev) % INU,&
              & HIERARCHICAL_DATA(nth_lev) % IAU,&
              & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE,   &
              & WS, WR, dim, SOLVER_COMM, my_rank)         
    
    
    
    if(coarser_level_precision == 0) then
       call sgs(mat, B, X, smiter, omega, 0)
      
    else
       do iter = 1, viter
          call sgs(mat, B, X, smiter, omega, 1)
          call cal_res_vec_at_nth_lev(B, X, Temp, nth_lev)

          !C nth_lev -> nth_lev+1          
          call restrict_procedure(nth_lev, Temp,   &
               & HIERARCHICAL_DATA(nth_lev+1) % B, &
               & WS, WR, dim, my_rank, SOLVER_COMM)

          HIERARCHICAL_DATA(nth_lev+1) % X = 0.0
          
          smiter_c                  = record_data(coarser_level_precision, nth_lev+1) % smiter
          coarser_level_precision_c = record_data(coarser_level_precision, nth_lev+1) % coarser_level_precision
          viter_c                   = record_data(coarser_level_precision, nth_lev+1) % viter
          omega_c                   = record_data(coarser_level_precision, nth_lev+1) % omega
          
          call v_cycle_at_each_lev(&
               & HIERARCHICAL_DATA(nth_lev+1) % B, &
               & HIERARCHICAL_DATA(nth_lev+1) % X, &
               & Temp, nth_lev+1, smiter_c, omega_c, coarser_level_precision_c, viter_c, &
               & WS, WR, dim, SOLVER_COMM, my_rank)
          
          !C nth_lev+1 -> nth_lev
          call prolong_and_add_procedure(nth_lev, X, &
               & HIERARCHICAL_DATA(nth_lev+1) % X,   &
               & WS, WR, dim, my_rank, SOLVER_COMM)
          
          call set_mat_data(mat, &
              & HIERARCHICAL_DATA(nth_lev) % N,  &
              & HIERARCHICAL_DATA(nth_lev) % NP, &
              & HIERARCHICAL_DATA(nth_lev) % NPL,&
              & HIERARCHICAL_DATA(nth_lev) % NPU,&
              & HIERARCHICAL_DATA(nth_lev) % D,  &
              & HIERARCHICAL_DATA(nth_lev) % AL, &
              & HIERARCHICAL_DATA(nth_lev) % INL,&
              & HIERARCHICAL_DATA(nth_lev) % IAL,&
              & HIERARCHICAL_DATA(nth_lev) % AU, &
              & HIERARCHICAL_DATA(nth_lev) % INU,&
              & HIERARCHICAL_DATA(nth_lev) % IAU,&
              & HIERARCHICAL_DATA(nth_lev) % COMM_TABLE,   &
              & WS, WR, dim, SOLVER_COMM, my_rank)         
    
          call sgs(mat, B, X, smiter, omega, 2)
       end do
    end if
  end subroutine v_cycle_at_each_lev
  
end module cycle_shape_tuning
