module data_structure_for_AMG_NB
  implicit none
  include 'precision.inc'
  INTEGER(kind=kint),parameter :: SWITCH_SIZE=1700  
  INTEGER(kind=kint),parameter :: LEV_ARRAY_SIZE=10
!  INTEGER(kind=kint),parameter :: MIN_NODE_SIZE=100
  INTEGER(kind=kint),parameter :: NODE_RECORD_SIZE=300
  real (kind=kreal) :: RATE_OF_SPACE=30

  REAL(kind=kreal),   pointer     :: coarsest_A_NB(:,:)
  INTEGER(kind=kint), pointer     :: coarsest_IDX_NB(:)


  type INTER_LEVEL_OPERATOR
     integer(kind=kint) :: ROW_SIZE
     integer(kind=kint ), dimension(:),pointer::IN
     integer(kind=kint ), dimension(:),pointer::CN
     REAL   (kind=kreal), dimension(:),pointer::V
  end type INTER_LEVEL_OPERATOR

  type COMMUNICATION_TABLE
     INTEGER(kind=kint) :: NEIBPETOT
     INTEGER(kind=kint),dimension(:),pointer :: NEIBPE
     INTEGER(kind=kint),dimension(:),pointer :: STACK_IMPORT
     INTEGER(kind=kint),dimension(:),pointer :: STACK_EXPORT
     INTEGER(kind=kint),dimension(:),pointer :: NOD_IMPORT
     INTEGER(kind=kint),dimension(:),pointer :: NOD_EXPORT
  end type COMMUNICATION_TABLE
  
  type DATA_FOR_EACH_LEV
     integer (kind=kint) :: N,NP,NPL,NPU,NB
     integer (kind=kint), dimension(:), pointer :: INU,INL
     integer (kind=kint), dimension(:), pointer :: IAU,IAL
     real (kind=kreal), dimension(:),     pointer :: X,B
     real (kind=kreal), dimension(:,:,:), pointer :: AU,AL,D,ALU
     
     type(COMMUNICATION_TABLE) :: COMM_TABLE
     type(COMMUNICATION_TABLE) :: INT_LVL_TABLE
     type(INTER_LEVEL_OPERATOR), pointer :: R, P
     
     real(kind=kreal), pointer :: k_or_q_vec(:,:)
     integer(kind=kint) :: NK

     logical :: PRO_AGGRE_OPT=.false.

  end type DATA_FOR_EACH_LEV
  
  type (DATA_FOR_EACH_LEV), dimension(LEV_ARRAY_SIZE),save :: HIERARCHICAL_DATA_NB
  
contains
  subroutine max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA, LEVEL_NUM)
    implicit none
    integer(kind=kint) :: NP_max, NP, LEVEL_NUM
    type (DATA_FOR_EACH_LEV) :: HIERARCHICAL_DATA(:)
    
    integer(kind=kint) :: i,j

    NP_max = NP
    DO i = 2, LEVEL_NUM
       j = HIERARCHICAL_DATA(i) % NP
       if(NP_max < j) NP_max = j
    END DO
    
  end subroutine max_NP_on_all_levs
  
  subroutine max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA, LEVEL_NUM)
    implicit none
    integer(kind=kint) :: WSIZE2, LEVEL_NUM
    type (DATA_FOR_EACH_LEV) :: HIERARCHICAL_DATA(:)

    integer(kind=kint) :: i,j,k,l,m, NP, NEIBPETOT,NB
    
    NP = HIERARCHICAL_DATA(1) % NP
    NEIBPETOT = HIERARCHICAL_DATA(1) % COMM_TABLE % NEIBPETOT
    WSIZE2 = 0

    !C level 1
    NB = HIERARCHICAL_DATA(1) % NB
    if(NEIBPETOT > 0) then
       if(    HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_IMPORT(NEIBPETOT) >  &
            & HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_EXPORT(NEIBPETOT)  ) then
          WSIZE2 = HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_IMPORT(NEIBPETOT)*NB
       else
          WSIZE2 = HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_EXPORT(NEIBPETOT)*NB
       end if
    end if
    if( WSIZE2 < NP*NB ) WSIZE2 = NP*NB

    !C level 2..
    m = WSIZE2
    DO i = 2, LEVEL_NUM
       NB = HIERARCHICAL_DATA(i) % NB
       j = HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPETOT 

       if(j>0) then
          k = HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT(j)*NB
          l = HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT(j)*NB
          m = max(k,l,m)
       end if

       j = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NEIBPETOT 

       if(j>0) then
          k = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_IMPORT(j)*NB 
          l = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_EXPORT(j)*NB 
          m = max(k,l,m)
       end if

    END DO

    if(m > WSIZE2) then
       !C-- WS, WR must be reallocated
       WSIZE2 = m
    end if
  end subroutine max_comm_table_sz
  
!!$  subroutine HASH(key,size,tmp)
!!$    implicit none
!!$    integer (kind=kint),intent(in) ::key,size
!!$    integer (kind=kint),intent(out)::tmp
!!$    Real (kind=kreal) :: HASH_KEY
!!$    parameter(HASH_KEY=1)
!!$    tmp= mod(INT(HASH_KEY*key), size)+1
!!$  end subroutine HASH


!!$  subroutine FIND(val,size,tmp,keys,key)
!!$    implicit none
!!$
!!$    integer (kind=kint),intent(in) ::val,size,key
!!$    integer(kind=kint),DIMENSION(:),intent(in) :: keys
!!$    integer (kind=kint),intent(out)::tmp
!!$    integer(kind=kint)::count
!!$
!!$    count=val
!!$    DO while( count<=size .and. keys(count)/=0 .and. keys(count)/=key)
!!$       count=count+1
!!$    end DO
!!$    if(count>size) then
!!$       count=1
!!$       DO while( count<val .and. keys(count)/=0 .and. keys(count)/=key )
!!$          count=count+1
!!$       end DO
!!$       if(count>=val) stop "hash overflow.."
!!$    end if
!!$    tmp=count
!!$  end subroutine FIND

end module data_structure_for_AMG_NB
