!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

module data_structure_for_AMG
  use axis_comm, only: comm_info

  implicit none
  include 'precision.inc'
  
  integer, parameter :: SWITCH_SIZE    = 100
  integer, parameter :: LEV_ARRAY_SIZE = 10
!  integer, parameter :: MIN_NODE_SIZE  = 500
  
  type row_node
     INTEGER(kind=kint) :: column
     REAL(kind=kreal)   :: value
  end type row_node
  
  type INTER_LEVEL_OPERATOR
     integer(kind=kint) :: ROW_SIZE
     integer(kind=kint),  pointer :: IN(:)
     integer(kind=kint),  pointer :: CN(:)
     real   (kind=kreal), pointer :: V(:)
  end type INTER_LEVEL_OPERATOR
  
  type COMMUNICATION_TABLE
     integer(kind=kint) :: NEIBPETOT
     integer(kind=kint), pointer :: NEIBPE(:)
     integer(kind=kint), pointer :: STACK_IMPORT(:)
     integer(kind=kint), pointer :: STACK_EXPORT(:)
     integer(kind=kint), pointer :: NOD_IMPORT(:)
     integer(kind=kint), pointer :: NOD_EXPORT(:)
     
     integer(kind=kint) :: switch
     type(comm_info)    :: comm
  end type COMMUNICATION_TABLE
  
  type DATA_FOR_EACH_LEV
     integer(kind=kint ) :: N,NP,NPL,NPU
     integer(kind=kint ), pointer :: INU(:),INL(:)
     integer(kind=kint ), pointer :: IAU(:),IAL(:)
     real   (kind=kreal), pointer :: AU(:),AL(:),D(:),X(:),B(:)

     !C coloring
     integer(kind=kint), pointer :: INML(:), color_index(:)
     integer(kind=kint) :: color = 0
	     
     type(COMMUNICATION_TABLE) :: COMM_TABLE
     type(COMMUNICATION_TABLE) :: INT_LVL_TABLE
     type(INTER_LEVEL_OPERATOR), pointer :: R, P
     
     real(kind=kreal) :: cheb_h, cheb_l
     logical :: cheb_fl = .false.

     !C LEVEL 2-LEVELNUM is assumed to be scaled..
     logical :: scaling_fl = .false.
     real   (kind=kreal), pointer :: SCALE_D(:)

     !C 
     logical :: PRO_AGGRE_OPT=.false.
     
  end type DATA_FOR_EACH_LEV
  
  type (DATA_FOR_EACH_LEV), save :: HIERARCHICAL_DATA(LEV_ARRAY_SIZE)
  
!C Rate_Of_Space * mesh size is the width of the next coarser 
!C level's restriction

  real (kind=kreal),  parameter :: EPS = 1.E-10
  real (kind=kreal),  parameter :: EPS_COARSEST = 1.E-5
  INTEGER(kind=kint), parameter :: NODE_RECORD_SIZE=300
  LOGICAL :: SYMMETRIC_FLAG

  REAL(kind=kreal),   pointer     :: coarsest_A(:,:)
  INTEGER(kind=kint), pointer     :: coarsest_IDX(:)
  
  INTEGER*8 :: pspcommf
  INTEGER(kind=kint),pointer :: rowdistbx(:)
  INTEGER(kind=kint) :: ldb,ldx,nrhs
  INTEGER(kind=kint) :: ioptions(16)
  REAL(kind=kreal)   :: doptions(16)
  REAL(kind=kreal),pointer        :: coarsest_X(:,:), coarsest_B(:,:)
contains
  subroutine max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA, LEVEL_NUM)
    implicit none
    integer(kind=kint) :: NP_max, NP, LEVEL_NUM
    type (DATA_FOR_EACH_LEV) :: HIERARCHICAL_DATA(:)
    
    integer(kind=kint) :: i,j

    NP_max = NP
    DO i = 2, LEVEL_NUM
       j = HIERARCHICAL_DATA(i) % NP
       if(NP_max < j) NP_max = j
    END DO

    
  end subroutine max_NP_on_all_levs
  
  subroutine max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA, LEVEL_NUM)
    implicit none
    integer(kind=kint) :: WSIZE2, LEVEL_NUM
    type (DATA_FOR_EACH_LEV) :: HIERARCHICAL_DATA(:)

    integer(kind=kint) :: i,j,k,l,m, NP, NEIBPETOT

    NP = HIERARCHICAL_DATA(1) % NP
    NEIBPETOT = HIERARCHICAL_DATA(1) % COMM_TABLE % NEIBPETOT
    WSIZE2 = 0

    !C level 1
    if(NEIBPETOT > 0) then
       if(    HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_IMPORT(NEIBPETOT) >  &
            & HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_EXPORT(NEIBPETOT)  ) then
          WSIZE2 = HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_IMPORT(NEIBPETOT)
       else
          WSIZE2 = HIERARCHICAL_DATA(1) % COMM_TABLE % STACK_EXPORT(NEIBPETOT)
       end if
    end if
    if( WSIZE2 < NP ) WSIZE2 = NP

    !C level 2..
    m = WSIZE2
    DO i = 2, LEVEL_NUM
       j = HIERARCHICAL_DATA(i) % COMM_TABLE % NEIBPETOT 

       if(j>0) then
          k = HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_IMPORT(j)
          l = HIERARCHICAL_DATA(i) % COMM_TABLE % STACK_EXPORT(j)
          m = max(k,l,m)
       end if

       j = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % NEIBPETOT 

       if(j>0) then
          k = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_IMPORT(j)
          l = HIERARCHICAL_DATA(i) % INT_LVL_TABLE % STACK_EXPORT(j)
          m = max(k,l,m)
       end if

    END DO

    if(m > WSIZE2) then
       !C-- WS, WR must be reallocated
       WSIZE2 = m
    end if
  end subroutine max_comm_table_sz


end module data_structure_for_AMG






