/** @file  gpu_solver.h
    @brief GPUAMG&AMG-BiCGSTABˡ
    @author Kosuke Takahashi
 */

#ifndef __GPU_SOLVER_H
#define __GPU_SOLVER_H

extern"C"{

/**
 * @emun  SOLVER_TYPE
 * @brief Фꤹ
 */
enum SOLVER_TYPE
{
	AMG,          /**< AMG */
	AMG_BiCGSTAB, /**< AMG_BiCGSTAB */
	SOLVER_LENGTH
};

/**
 * @emun  SMOOTHER_SET_TYPE
 * @brief ࡼȤ߹碌ꤹ
 */
enum SMOOTHER_SET_TYPE
{
	ALL_JC,              /**< :Jacobi                                */
	MCGS_AND_JC,         /**< Ǿ:Multi Color GS  沼:Jacobi         */
	ALL_MCGS,            /**< :Multi Color GS                        */
	IDR_AND_JC,          /**< Ǿ:IDR Jacobi      沼:Jacobi         */
	ALL_JC_WITH_LU,      /**< :Jacobi          :LU             */
	MCGS_AND_JC_WITH_LU, /**< Ǿ:Multi Color GS  桡:Jacobi :LU */
	ALL_MCGS_WITH_LU,    /**< :Multi Color GS  :LU             */
	IDR_AND_JC_WITH_LU,  /**< Ǿ:IDR Jacobi      桡:Jacobi :LU */
	SMOOTHER_SET_TYPE_LENGTH
};

/**
 * ޥåɤγؿMPIΥץֹϿ
 * ѤǥХꤵ
 * @param [in] Level_NUM ޥåɤγؿ
 * @param [in] rank      MPIΥץֹ
 * @return ʤ
 */
void notify_level_(int *Level_NUM, int *rank);


/**
 * Ѥ륽ФȲ®Ͽ
 * @param [in] select      0:AMG 1:AMG-BiCGSTAB
 * @param [in] omega       ®
 * @param [in] iter_down   ǲذʳȿ
 * @param [in] iter_bottom ǲؤȿ
 * @param [in] gs_at_lev1  0:JC  1:MCGS
 * @return ʤ
 */
void parameter_set_on_gpu_(int *select, double *omega, int* iter_down, int* iter_bottom, int* gs_at_lev1);


/**
 * ࡼstore_mat_on_gpu_Ͽ
 * @param [in] smoother_set 0:MCGS-JC 1:JC 2:MCGS
 * @return ʤ
 */
void smoother_set_on_gpu_(int *smoother_set);


/**
 * ƥ٥ȿͳʥߥ󥰤Ͽ
 * @param [in] iter_type 0:flat 1:bottom_up 2:bottom_down
 * @return ʤ
 */
void iter_type_set_on_gpu_(int *iter_type);
enum ITER_TYPE{FLAT, BOTTOM_UP, BOTTOM_DOWN, NON_ITER, ITER_TYPE_LENGTH};

/**
 * ƥ٥ȿͳʥߥ󥰤Ͽ
 * @param [in] iter_down   ǲذʳȿ
 * @param [in] iter_bottom ǲؤȿ
 * @return ʤ
 */
void iter_set_on_gpu_(int* iter_down, int* iter_bottom);


/**
 * ®ͳʥߥ󥰤Ͽ
 * @param [in] omega ®
 */
void omega_set_on_gpu_(double *omega);


/**
 * ƳؤΥåɤ̿ơ֥Ͽ
 * @param [in] Level
 * @param [in] A
 * @param [in] INPORT_COMM_TABLE
 * @param [in] EXPORT_COMM_TABLE
 * @param [in] R&P
 * @param [in] INTER_INPORT_COMM_TABLE
 * @param [in] INTER_EXPORT_COMM_TABLE
 * @param [in] SCL_FL
 * @param [in] SCL_VEC
 * @return ʤ
 */
void store_mat_on_gpu_(int *Level, 
					  int *N, int *NP, int *NPL, int *NPU, 
					  int *INL, int *INU, 
					  int *IAL, double *AL, 
					  int *IAU, double *AU, 
					  double *D,

					  int *IN_T_N, int *IN_T_NPLU,
					  int *IN_T_NO, int *IN_T_DSP, int *IN_T_COL,
					  int *EX_T_N, int *EX_T_NPLU,
					  int *EX_T_NO, int *EX_T_DSP, int *EX_T_COL,

					  int *R_N, int *P_N, int *R_NP, int *P_NP, int *R_NPLU, int *P_NPLU, 
					  int *R_IN, int *R_CN, double *R_V,
					  int *P_IN, int *P_CN, double *P_V,

					  int *INTER_IN_T_N, int *INTER_IN_T_NPLU,
					  int *INTER_IN_T_NO, int *INTER_IN_T_DSP, int *INTER_IN_T_COL,
					  int *INTER_EX_T_N, int *INTER_EX_T_NPLU,
		                          int *INTER_EX_T_NO, int *INTER_EX_T_DSP, int *INTER_EX_T_COL,
		                          
		                          int *SCL_FL, double *SCL_VEC);



/**
 * «ȽѤͥ٥ȥϿ
 * @param [in] restol  ͥ٥ȥ
 * @param [in] rowsize ͥ٥ȥ륵
 * @return ʤ
 */
void amgs_terminal_vec_set_on_gpu_(double *restol, int *rowsize);


/**
 * «ȽѤͤϿ
 * @param [in] residual 
 * @return ʤ
 */
void amgs_terminal_val_set_on_gpu_(double *residual);


/**
 * Фȿ
 * @param [in,out] x    in:̤ο out:
 * @param [in]     b    ե٥ȥ
 * @param [in,out] iter in:ȿ¿ out:ȿ̿
 * @return ʤ
 */
void solve_on_gpu_(double *x, double *b, int *iter);

/**
 * GPUAMGˡǻѤƲ
 * @return ʤ
 */
void clear_mat_on_gpu_(void);

void *get_amgs_v_nodes();

}

#endif /* __GPU_SOLVER_H */

