#include "icrs.h"
#include "my_cuda.h"

#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#define SP 256


__global__ void restriction_kernel(icrs_part a, double *b, double *c);
__global__ void prolongation_kernel(icrs_part a, double *b, double *c);


void restriction(icrs &a, double *b, double *c)
{
	int i, size;
	size = a.size();
	for(i=0; i<size; i++){
		restriction_kernel<<<a[i].n / SP + 1 , SP >>>(a[i], b, c);
	}
}


void prolongation(icrs &a, double *b, double *c)
{
	int i, size;
	size = a.size();
	for(i=0; i<size; i++){
		prolongation_kernel<<<a[i].n / SP + 1 , SP >>>(a[i], b, c);
	}
}


__global__
void restriction_kernel(icrs_part a, double *b, double *c)
{
	__shared__ double s_val[SP];
	__shared__ double s_b[SP];
	__shared__ int    s_col[SP];

	int i, index, local_cnt;
	double temp;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		local_cnt = a.cnt[id];

		temp = 0.0;

		index = id;
		for(i=0; i<local_cnt; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_b[tid] = b[s_col[tid]];

			temp += s_val[tid] * s_b[tid];
		}

		index = id + a.row;
		c[index] = temp;
	}
}


__global__
void prolongation_kernel(icrs_part a, double *b, double *c)
{
	__shared__ double s_val[SP];
	__shared__ double s_b[SP];
	__shared__ int    s_col[SP];

	int i, index, local_cnt;
	double temp;

	const int tid = threadIdx.x;
	const int id  = blockIdx.x * blockDim.x + threadIdx.x;

	if(id < a.n){
		local_cnt = a.cnt[id];

		temp = 0.0;

		index = id;
		for(i=0; i<local_cnt; i++){
			s_col[tid] = a.col[index];
			s_val[tid] = a.val[index];

			index += a.np;

			s_b[tid] = b[s_col[tid]];

			temp += s_val[tid] * s_b[tid];
		}

		index = id + a.row;
		c[index] += temp;
	}
}
