module int_vec_module
  include 'precision.inc'

  type int_vec
     integer(kind=kint), pointer :: int_array(:)
     integer(kind=kint) :: alloc_size
  end type int_vec

contains
  subroutine int_vec_initialize(int_array_val)
    implicit none
    type(int_vec) :: int_array_val
    
    int_array_val % alloc_size = 0
    nullify(int_array_val % int_array)
  end subroutine int_vec_initialize

  subroutine check_and_extend(int_array_val, a_index)
    implicit none
    type(int_vec) :: int_array_val
    integer(kind=kint) ::  a_index
    
    if(int_array_val % alloc_size < a_index) call check_and_extend_(int_array_val, a_index)
    
  end subroutine check_and_extend

  subroutine check_and_extend_(int_array_val, a_index)
    implicit none
    type(int_vec) :: int_array_val
    integer(kind=kint), pointer, save :: tmp(:)
    integer(kind=kint) :: a_index, alloc_size
    integer(kind=kint) :: s
    integer, parameter :: EXSZ = 1000

    !C allocate, copy, pointer connection
    alloc_size = int_array_val % alloc_size
    s = alloc_size + EXSZ
    if(s < a_index) s = a_index
    allocate(tmp(s))

    if(alloc_size > 0) then
       tmp(1: alloc_size) = int_array_val % int_array(1:alloc_size)
       deallocate(int_array_val % int_array)
    end if

    int_array_val % alloc_size = s

    int_array_val % int_array => tmp
    nullify(tmp)

  end subroutine check_and_extend_
  
  subroutine deallocate_int_vec(int_array_val)
    implicit none
    type(int_vec) :: int_array_val
    if(int_array_val % alloc_size > 0) deallocate(int_array_val % int_array)
  end subroutine deallocate_int_vec
  
end module int_vec_module
