#include "lab_matrix.h"
#include "my_cuda.h"

#include <cstring>
#include <iostream>

using std::cout;
using std::endl;

#define PADDING_BYTE 256
#define PADDING_INT ( PADDING_BYTE / sizeof(int) )


int convert_crs_to_icrs(crs &src_mat, icrs &dst_mat, int b_n, int *b_cnt, int *b_ind, int *order_col)
{
	int i, j;
	int b_length, nnz, index, buffer_length, start, end, b_id;
	int *count, *order, *int_buffer, *ptr;
	double *dfp_buffer;

	try{
		dst_mat.reserve(b_n);

		// set b_length
		b_length = 0;
		for(i=0; i<b_n; i++){
			b_length += b_cnt[i];
		}

		// set icrs n, np, nnz, row.
		for(i=0; i<b_n; i++){
			dst_mat.push_back( icrs_part() );

			if(i < b_n-1){
				dst_mat[i].n   = b_cnt[i];
			}else{
				dst_mat[i].n   = b_cnt[i] + src_mat.n - b_length;
			}
			dst_mat[i].np  = dst_mat[i].n + PADDING_INT - (dst_mat[i].n % PADDING_INT);
			dst_mat[i].nnz = 0;
			dst_mat[i].row = i ? dst_mat[i-1].row + dst_mat[i-1].np : 0;
		}

		for(i=0; i<b_length; i++){
			index = b_ind[i];
			nnz = dst_mat[index].np * (src_mat.ptr[i+1] - src_mat.ptr[i]);
			dst_mat[index].nnz = nnz > dst_mat[index].nnz ? nnz : dst_mat[index].nnz;
		}
		for(i=b_length; i<src_mat.n; i++){
			index = b_n ? b_n - 1 : 0;
			nnz = dst_mat[index].np * (src_mat.ptr[i+1] - src_mat.ptr[i]);
			dst_mat[index].nnz = nnz > dst_mat[index].nnz ? nnz : dst_mat[index].nnz;
		}


		// set order.
		order = new int[src_mat.n];
		count    = new int[b_n];
		for(i=0; i<b_n; i++){
			count[i] = dst_mat[i].row;
		}
		for(i=0; i<b_length; i++){
			index = b_ind[i];
			order[i] = count[index]++;
		}
		for(i=b_length; i<src_mat.n; i++){
			index = b_n ? b_n - 1 : 0;
			order[i] = count[index]++;
		}
		if(count) delete[] count;


		// cudaMalloc icrs & set buffer.
		buffer_length = 0;
		for(i=0; i<b_n; i++){
			dst_mat[i].icrs_part_new(dst_mat[i].n, dst_mat[i].nnz);
			buffer_length += dst_mat[i].n + dst_mat[i].nnz;
		}
		int_buffer = new int[buffer_length];
		dfp_buffer = new double[buffer_length];

		// send icrs cnt.
		memset(int_buffer, 0, buffer_length * sizeof(int));
		memset(dfp_buffer, 0, buffer_length * sizeof(double));
		for(i=0; i<src_mat.n; i++){
			index = order[i];
			int_buffer[index] = src_mat.ptr[i+1] - src_mat.ptr[i];
		}
		for(i=0; i<b_n; i++){
			my_cudaMemcpy(dst_mat[i].cnt, int_buffer + dst_mat[i].row, sizeof(int)    * dst_mat[i].n, cudaMemcpyHostToDevice);
		}

		// send icrs col, val.
		memset(int_buffer, 0, buffer_length * sizeof(int));
		memset(dfp_buffer, 0, buffer_length * sizeof(double));
		ptr = new int[b_n];
		for(i=0; i<b_n; i++){
			ptr[i] = i ? ptr[i-1] + dst_mat[i-1].nnz : 0;
		}
		for(i=0; i<b_length; i++){
			b_id  = b_ind[i];
			index = ptr[b_id] + order[i] - dst_mat[b_id].row;

			start = src_mat.ptr[ i ];
			end   = src_mat.ptr[i+1];
			for(j=start; j<end; j++){
				int_buffer[index] = order_col[ src_mat.col[j] ];
				dfp_buffer[index] = src_mat.val[j];
				index += dst_mat[b_id].np;
			}
		}
		for(i=b_length; i<src_mat.n; i++){
			b_id  = b_n - 1;
			index = ptr[b_id] + order[i] - dst_mat[b_id].row;

			start = src_mat.ptr[ i ];
			end   = src_mat.ptr[i+1];
			for(j=start; j<end; j++){
				int_buffer[index] = order_col[ src_mat.col[j] ];
				dfp_buffer[index] = src_mat.val[j];
				index += dst_mat[b_id].np;
			}
		}
		for(i=0; i<b_n; i++){
			my_cudaMemcpy(dst_mat[i].col, int_buffer + ptr[i], sizeof(int)    * dst_mat[i].nnz, cudaMemcpyHostToDevice);
			my_cudaMemcpy(dst_mat[i].val, dfp_buffer + ptr[i], sizeof(double) * dst_mat[i].nnz, cudaMemcpyHostToDevice);
		}
		if(ptr) delete[] ptr;
                      
		if(dfp_buffer) delete[] dfp_buffer;
		if(int_buffer) delete[] int_buffer;
	}
	catch(...){
		cout << "convert_crs_to_icrs expection." << endl;
		return 1;
	}
	return 0;
}


int convert_lu_crs_to_lu_icrs(lu_crs &src_mat, lu_icrs &dst_mat, int b_n, int *b_cnt, int *b_ind, int **o_t)
{
	int i, j;
	int nnz, index, buffer_length, start, end, b_id;
	int *count, *order, *int_buffer, *ptr;
	double *dfp_buffer;

	try{

		dst_mat.reserve(b_n);

		// set lu_icrs n, np, nnz, row.
		for(i=0; i<b_n; i++){
			dst_mat.push_back( lu_icrs_part() );

			dst_mat[i].n   = b_cnt[i];
			dst_mat[i].np  = dst_mat[i].n + PADDING_INT - (dst_mat[i].n % PADDING_INT);
			dst_mat[i].nnz = 0;
			dst_mat[i].row = i ? dst_mat[i-1].row + dst_mat[i-1].np : 0;
		}
		for(i=0; i<src_mat.n; i++){
			index = b_ind[i];
			nnz = dst_mat[index].np * (src_mat.inl[i+1] - src_mat.inl[i] + src_mat.inu[i+1] -src_mat.inu[i]);
			dst_mat[index].nnz = nnz > dst_mat[index].nnz ? nnz : dst_mat[index].nnz;
		}


		// set order.
		order = *o_t = new int[src_mat.np];
		count    = new int[b_n];
		for(i=0; i<b_n; i++){
			count[i] = dst_mat[i].row;
		}
		for(i=0; i<src_mat.n; i++){
			index = b_ind[i];
			order[i] = count[index]++;
		}
		for(i=src_mat.n; i<src_mat.np; i++){
			index = b_n ? b_n - 1 : 0;
			order[i] = count[index]++;
		}
		if(count) delete[] count;

		// cudaMalloc lu_icrs & set buffer.
		buffer_length = 0;
		for(i=0; i<b_n; i++){
			dst_mat[i].lu_icrs_part_new(dst_mat[i].n, dst_mat[i].nnz);
			buffer_length += dst_mat[i].nnz;
		}
		int_buffer = new int[buffer_length];
		dfp_buffer = new double[buffer_length];

		// send lu_icrs cnt, d.
		memset(int_buffer, 0, buffer_length * sizeof(int));
		memset(dfp_buffer, 0, buffer_length * sizeof(double));
		for(i=0; i<src_mat.n; i++){
			index = order[i];
			int_buffer[index] = src_mat.inl[i+1] - src_mat.inl[i] + src_mat.inu[i+1] -src_mat.inu[i];
			dfp_buffer[index] = src_mat.d[i];
		}
		for(i=0; i<b_n; i++){
			my_cudaMemcpy(dst_mat[i].cnt, int_buffer + dst_mat[i].row, sizeof(int)    * dst_mat[i].n, cudaMemcpyHostToDevice);
			my_cudaMemcpy(dst_mat[i].d  , dfp_buffer + dst_mat[i].row, sizeof(double) * dst_mat[i].n, cudaMemcpyHostToDevice);
		}

		// send lu_icrs col, val.
		memset(int_buffer, 0, buffer_length * sizeof(int));
		memset(dfp_buffer, 0, buffer_length * sizeof(double));
		ptr = new int[b_n];
		for(i=0; i<b_n; i++){
			ptr[i] = i ? ptr[i-1] + dst_mat[i-1].nnz : 0;
		}
		for(i=0; i<src_mat.n; i++){
			b_id  = b_ind[i];
			index = ptr[b_id] + (order[i] - dst_mat[b_id].row);

			start = src_mat.inl[ i ];
			end   = src_mat.inl[i+1];
			for(j=start; j<end; j++){
				int_buffer[index] = order[ src_mat.ial[j] ];
				dfp_buffer[index] = src_mat.al[j];
				index += dst_mat[b_id].np;
			}

			start = src_mat.inu[ i ];
			end   = src_mat.inu[i+1];
			for(j=start; j<end; j++){
				int_buffer[index] = order[ src_mat.iau[j] ];
				dfp_buffer[index] = src_mat.au[j];
				index += dst_mat[b_id].np;
			}
		}
		for(i=0; i<b_n; i++){
			my_cudaMemcpy(dst_mat[i].col, int_buffer + ptr[i], sizeof(int)    * dst_mat[i].nnz, cudaMemcpyHostToDevice);
			my_cudaMemcpy(dst_mat[i].val, dfp_buffer + ptr[i], sizeof(double) * dst_mat[i].nnz, cudaMemcpyHostToDevice);
		}
		if(ptr) delete[] ptr;

		if(dfp_buffer) delete[] dfp_buffer;
		if(int_buffer) delete[] int_buffer;
	}
	catch(...){
		cout << "convert_lu_crs_to_lu_icrs expection." << endl;
		return 1;
	}
	return 0;
}
