#include "lu_crs.h"

#include <cmath>
#include <iostream>
#include <new>

using std::bad_alloc;
using std::cout;
using std::endl;

lu_crs::lu_crs()
{
	n    = 0;
	np   = 0;
	nnzl = 0;
	nnzu = 0;

	inl = 0;
	inu = 0;
	ial = 0;
	iau = 0;

	al = 0;
	au = 0;
	d  = 0;

	new_done = false;
}


lu_crs::~lu_crs()
{
	if( lu_crs_delete() ){
		cout << "exit in lu_crs destructor." << endl;
	}
}


int lu_crs::lu_crs_new(int n, int np, int nnzl, int nnzu)
{
	this->n    = n;
	this->np   = np;
	this->nnzl = nnzl;
	this->nnzu = nnzu;

	try{
		inl = new int[n+1];
		inu = new int[n+1];
		ial = new int[nnzl];
		iau = new int[nnzu];

		al = new double[nnzl];
		au = new double[nnzu];
		d  = new double[n];
	}
	catch(bad_alloc)
	{
		cout << "lu_crs bad alloc expection : ";
		cout << n << " " << nnzl << " " << nnzu;
		cout << endl;
		return 1;
	}
	catch(...)
	{
		cout << "lu_crs new expection." << endl;
		return 1;
	}
	new_done = true;
	return 0;
}


int lu_crs::lu_crs_delete()
{
	try{
		if(new_done){
			delete[] inl;
			delete[] inu;
			delete[] ial;
			delete[] iau;
		
			delete[] al;
			delete[] au;
			delete[] d ;

			new_done = false;
		}

		n    = 0;
		np   = 0;
		nnzl = 0;
		nnzu = 0;

		inl = 0;
		inu = 0;
		ial = 0;
		iau = 0;

		al = 0;
		au = 0;
		d  = 0;
	}
	catch(...)
	{
		cout << "lu_crs delete expection." << endl;
		return 1;
	}

	return 0;
}



double get_two_norm_residual(lu_crs &a, double *x, double *b)
{
	int i, j;
	int start, end;
	double norm, val;

	norm = 0;
	for(i=0; i<a.n; i++){
		val = b[i];

		start = a.inl[ i ];
		end   = a.inl[i+1];
		for(j=start; j<end; j++){
			val -= a.al[j] * x[ a.ial[j] ];
		}

		start = a.inu[ i ];
		end   = a.inu[i+1];
		for(j=start; j<end; j++){
			val -= a.au[j] * x[ a.iau[j] ];
		}

		val -= a.d[i] * x[i];

		norm += val * val;
	}

	return sqrt(norm);
}
