#include "lu_decomposition.h"
#include "my_cuda.h"

#include <algorithm>
#include <cmath>
#include <cstring>
#include <iostream>

using std::cout;
using std::endl;
using std::swap;

#define EPS 1.0E-10
#define MAT(a, col, row, x, y) (a)[(y)*(row) + (x)]


void UnpackLU_CRS(lu_crs *src, double *dst);
int get_pivot(double a[], int n, int column);
void swap_row(double a[], int n, int left, int right);

bool lu_decomposition_setup(lu_crs *a, double *lu, int *p)
{
	int i, j, k;
	int n, pivot;
	double d;

	n  = a->np;

	UnpackLU_CRS(a, lu);
	
	for(i=0; i<n; i++){
		p[i] = i;
	}
	
	for(i=0; i<n; i++){
		pivot = get_pivot(lu, n, i);
		if(pivot != i){
			swap_row(lu, n, i, pivot);
			swap(p[i], p[pivot]);
		}

		d = lu[i*n + i];
		if( fabs(d) < EPS){
			return false;
		}
		for(j=i+1; j<n; j++){
			lu[j*n + i] /= d;

			for(k=i+1; k<n; k++){
				lu[j*n + k] -= lu[j*n + i] * lu[i*n + k];
			}
		}
	}
	
	return false;
}


void lu_decomposition(double a[], double x[], double b[], int p[], int n)
{
	int i, j;
	double val;

	for(i=0; i<n; i++){
		x[i] = 0;
	}

	// Forward Elimination
	for(i=0; i<n; i++){
		val = b[ p[i] ];
		for(j=0; j<i; j++){
			val -= a[i*n + j] * x[j];
		}
		x[i] = val;
	}

	// Backward Substitution
	for(i=n-1; i>=0; i--){
		val = x[i];
		for(j=i+1; j<n; j++){
			val -= a[i*n + j] * x[j];
		}
		x[i] = val / a[i*n + i];
	}
}


void UnpackLU_CRS(lu_crs *lu_a, double *a)
{
	int		i, j, start, end, n, np;
	int		*inl, *inu, *ial, *iau;
	double	*al, *au, *d;

	n   = lu_a->n;
	np  = lu_a->np;
	inl = lu_a->inl;
	inu = lu_a->inu;
	ial = lu_a->ial;
	iau = lu_a->iau;
	al  = lu_a->al;
	au  = lu_a->au;
	d   = lu_a->d;

	for(i=0; i<np; i++){
		for(j=0; j<np; j++){
			MAT(a, np, np, j, i) = 0.0;
		}
		MAT(a, np, np, i, i) = 1.0;
	}

	for(i=0; i<n; i++){

		start = inl[ i ];
		end   = inl[i+1];
		for(j=start; j<end; j++){
			MAT(a, pn, np, ial[j], i) = al[j];
		}

		start = inu[ i ];
		end   = inu[i+1];
		for(j=start; j<end; j++){
			MAT(a, np, np, iau[j], i) = au[j];
		}

		MAT(a, np, np, i, i) = d[i];
	}
}


int get_pivot(double a[], int n, int column)
{
	int i, pivot;

	pivot = column;
	for(i=column+1; i<n; i++){
		if( fabs(a[i*n + column]) > fabs(a[pivot*n + column]) ){
			pivot = i;
		}
	}

	return pivot;
}


void swap_row(double a[], int n, int left, int right)
{
	int i;
	for(i=0; i<n; i++){
		swap(a[left*n + i], a[right*n + i]);
	}
}
