
!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

MODULE matrix_dist_mtom
  
CONTAINS


  !C output data: N,NP,D,IN{U,L},IA{U,L},A{U,L},{STACK,NOD}_{IMPORT,EXPORT}
  !C            : order_o, index_number, new_part(1:NP)
  !C input data : *_i
  SUBROUTINE dist_mtom_main_vec                                              &
       & ( N_i, NP_i, NB_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, B_i,&
       &   STACK_IMPORT_i, STACK_EXPORT_i, NOD_IMPORT_i, NOD_EXPORT_i,       &
       &   NEIBPE_i, NEIBPETOT_i, N, NP, NB, D, INU, INL, IAU, IAL, AU, AL,  &
       &   B, STACK_IMPORT, STACK_EXPORT, NOD_IMPORT, NOD_EXPORT, NEIBPE,    &
       &   NEIBPETOT, SOLVER_COMM, my_rank, PETOT, order_i, order_o,         &
       &   nparts, tpwgts, index_number, new_part, method)
    USE count_time_mod
    IMPLICIT none
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'
    integer(kind=kint ), intent(in) :: SOLVER_COMM, my_rank, PETOT, method
    !C-- ordering 
    integer(kind=kint ), pointer :: order_i(:), order_o(:)

    !C-- initial distributed matrix form
    integer(kind=kint) :: N_i, NP_i, NB_i, NEIBPETOT_i
    real   (kind=KREAL), pointer :: D_i(:,:,:), B_i(:)
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:)
    integer(kind=kint ), pointer :: IAU_i(:), IAL_i(:)
    real   (kind=KREAL), pointer :: AU_i(:,:,:), AL_i(:,:,:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_i(:)
    integer(kind=kint ), pointer :: NEIBPE_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_i(:)
    
    !C-- redistributed matrix form
    integer(kind=kint) :: N, NP, NB, NEIBPETOT
    real   (kind=KREAL), pointer :: D(:,:,:), B(:)
    integer(kind=kint ), pointer :: INU(:), INL(:)
    integer(kind=kint ), pointer :: IAU(:), IAL(:)
    real   (kind=KREAL), pointer :: AU(:,:,:), AL(:,:,:)
    integer(kind=kint ), pointer :: NOD_IMPORT(:)
    integer(kind=kint ), pointer :: NOD_EXPORT(:)
    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:)

    !C-- index arrays for parmetis
    integer(kind=kint),   intent(in) :: nparts
    REAL   (kind=kfloat), intent(in) :: tpwgts(:)
    
    !C-- global number vector for parmetis
    integer(kind=kint ) :: i, j, k, l, ierr, SWITCH_FLAG
    
    integer(kind=kint ) :: index_number(0:)
    integer(kind=kint ) :: new_part(:)
    
    integer(kind=kint) :: sz_export, sz_import, NPL, NPU


    !C method is assumed to be 3 or 4.
    !C in order to return index_number and new_part.
    !C method 1 needs parmetis installation
    !C method 2 needs to be changed for returning index_number and new_part
    call dist_method                                                      &
       & ( N_i, NP_i, INU_i, INL_i, IAU_i, IAL_i, B_i,                    &
       &   STACK_IMPORT_i, STACK_EXPORT_i, NOD_IMPORT_i, NOD_EXPORT_i,    &
       &   NEIBPE_i, NEIBPETOT_i, SOLVER_COMM, nparts, tpwgts, new_part,  &
       &   index_number, SWITCH_FLAG, method)
    
    CALL count_time(1, solver_comm, my_rank, 13)                         

    
    !C communication index arrays are allocated
    allocate(STACK_IMPORT(0:PETOT), STACK_EXPORT(0:PETOT), NEIBPE(PETOT))

    
    !C determine the sizes of the new matrix in N, NP, NPL, NPU
    !C and the communication indices NEIBPETOT, NEIBPE, STACK_IMPORT, STACK_EXPRORT
    call size_new_mat_vec(sz_export, sz_import, N, NP, NB,NPL, NPU, NEIBPETOT, NEIBPE, &
         & STACK_IMPORT, STACK_EXPORT, new_part, index_number,                         &
         & N_i, NP_i, NB_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i,                     &
         & AL_i, NEIBPETOT_i, NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i,                    &
         & STACK_IMPORT_i, STACK_EXPORT_i, B_i, SOLVER_COMM,                           &
         & order_i, SWITCH_FLAG)


    !C new matrix allocation
    allocate(NOD_EXPORT(sz_export), NOD_IMPORT(sz_import),  IAL(NPL),      &
         &   IAU(NPU), INL(0:NP), INU(0:NP), AL(NB,NB,NPL), AU(NB,NB,NPU), &
         &   D(NB,NB,NP), B(NB*NP), order_o(N))

    call assign_new_mat_vec(N, NP, NB, NPL, NPU, D, INU, INL, IAU, IAL, AU, AL, B,    &
         & NEIBPETOT, NEIBPE, NOD_IMPORT, NOD_EXPORT, STACK_IMPORT,                   &
         & STACK_EXPORT, new_part, index_number,                                      &
         & N_i, NP_i, NB_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, NEIBPETOT_i, &
         & NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i, STACK_IMPORT_i, STACK_EXPORT_i,      &
         & B_i, SOLVER_COMM, order_i, order_o)

    CALL count_time(2, solver_comm, my_rank, 13)                         

    
  END SUBROUTINE dist_mtom_main_vec
  



  
  !C output data: N,NP,D,IN{U,L},IA{U,L},A{U,L},{STACK,NOD}_{IMPORT,EXPORT}
  !C            : order_o, index_number, new_part(1:NP)
  !C input data : *_i
  SUBROUTINE dist_mtom_main                                               &
       & ( N_i, NP_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, B_i,   &
       &   STACK_IMPORT_i, STACK_EXPORT_i, NOD_IMPORT_i, NOD_EXPORT_i,    &
       &   NEIBPE_i, NEIBPETOT_i, N,  NP, D, INU, INL, IAU, IAL, AU, AL,  &
       &   B, STACK_IMPORT, STACK_EXPORT, NOD_IMPORT, NOD_EXPORT, NEIBPE, &
       &   NEIBPETOT, SOLVER_COMM, my_rank, PETOT, order_i, order_o,      &
       &   nparts, tpwgts, index_number, new_part, method)
    USE count_time_mod
    IMPLICIT none
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'
    integer(kind=kint ), intent(in) :: SOLVER_COMM, my_rank, PETOT, method
    !C-- ordering 
    integer(kind=kint ), pointer :: order_i(:), order_o(:)

    !C-- initial distributed matrix form
    integer(kind=kint) :: N_i, NP_i, NEIBPETOT_i
    real   (kind=KREAL), pointer :: D_i(:), B_i(:)
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:)
    integer(kind=kint ), pointer :: IAU_i(:), IAL_i(:)
    real   (kind=KREAL), pointer :: AU_i(:), AL_i(:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_i(:)
    integer(kind=kint ), pointer :: NEIBPE_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_i(:)
    
    !C-- redistributed matrix form
    integer(kind=kint) :: N, NP, NEIBPETOT
    real   (kind=KREAL), pointer :: D(:), B(:)
    integer(kind=kint ), pointer :: INU(:), INL(:)
    integer(kind=kint ), pointer :: IAU(:), IAL(:)
    real   (kind=KREAL), pointer :: AU(:), AL(:)
    integer(kind=kint ), pointer :: NOD_IMPORT(:)
    integer(kind=kint ), pointer :: NOD_EXPORT(:)
    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:)
    
    !C-- index arrays for parmetis
    integer(kind=kint),   intent(in) :: nparts
    REAL   (kind=kfloat), intent(in) :: tpwgts(:)
    
    !C-- global number vector for parmetis
    integer(kind=kint ) :: i, j, k, l, ierr, SWITCH_FLAG
    
    integer(kind=kint ) :: index_number(0:)
    integer(kind=kint ) :: new_part(:)
    
    integer(kind=kint) :: sz_export, sz_import, NPL, NPU


    !C method is assumed to be 3 or 4.
    !C in order to return index_number and new_part.
    !C method 1 needs parmetis installation
    !C method 2 needs to be changed for returning index_number and new_part
    call dist_method                                                      &
       & ( N_i, NP_i, INU_i, INL_i, IAU_i, IAL_i, B_i,                    &
       &   STACK_IMPORT_i, STACK_EXPORT_i, NOD_IMPORT_i, NOD_EXPORT_i,    &
       &   NEIBPE_i, NEIBPETOT_i, SOLVER_COMM, nparts, tpwgts, new_part,  &
       &   index_number, SWITCH_FLAG, method)
    
    CALL count_time(1, solver_comm, my_rank, 13)                         

    
    !C communication index arrays are allocated
    allocate(STACK_IMPORT(0:PETOT), STACK_EXPORT(0:PETOT), NEIBPE(PETOT))



    !C determine the sizes of the new matrix in N, NP, NPL, NPU
    !C and the communication indices NEIBPETOT, NEIBPE, STACK_IMPORT, STACK_EXPRORT
    call size_new_mat(sz_export, sz_import, N, NP, NPL, NPU, NEIBPETOT, NEIBPE, &
         & STACK_IMPORT, STACK_EXPORT, new_part, index_number,                  &
         & N_i, NP_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i,                    &
         & AL_i, NEIBPETOT_i, NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i,             &
         & STACK_IMPORT_i, STACK_EXPORT_i, B_i, SOLVER_COMM,                    &
         & order_i, SWITCH_FLAG)


    !C new matrix allocation
    allocate(NOD_EXPORT(sz_export), NOD_IMPORT(sz_import),  IAL(NPL),           &
         &   IAU(NPU), INL(0:NP), INU(0:NP), AL(NPL), AU(NPU), D(NP), B(NP),    &
         &   order_o(N))


    call assign_new_mat(N, NP, NPL, NPU, D, INU, INL, IAU, IAL, AU, AL, B,      &
         & NEIBPETOT, NEIBPE, NOD_IMPORT, NOD_EXPORT, STACK_IMPORT,             &
         & STACK_EXPORT, new_part, index_number,                                &
         & N_i, NP_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, NEIBPETOT_i, &
         & NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i, STACK_IMPORT_i, STACK_EXPORT_i,&
         & B_i, SOLVER_COMM, order_i, order_o)

    CALL count_time(2, solver_comm, my_rank, 13)                         

    
  END SUBROUTINE dist_mtom_main

  !C method 1 metis distribution
  !C method 2 matrix distribution with minimum communication cost
  !C method 3 matrix distribution with no reordering
  SUBROUTINE dist_method                                                  &
!C       & ( N_i, NP_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, B_i,   &
       & ( N_i, NP_i, INU_i, INL_i, IAU_i, IAL_i, B_i,   &
       &   STACK_IMPORT_i, STACK_EXPORT_i, NOD_IMPORT_i, NOD_EXPORT_i,    &
       &   NEIBPE_i, NEIBPETOT_i, SOLVER_COMM, nparts, tpwgts, new_part,  &
       &   index3, SWITCH_FLAG, method)
    USE count_time_mod
    IMPLICIT none
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'
    integer(kind=kint ), intent(in) :: SOLVER_COMM, method
    integer(kind=kint ) ::  my_rank, PETOT
    !C-- ordering 
    
    !C-- initial distributed matrix form
    integer(kind=kint), intent(in) :: N_i, NP_i, NEIBPETOT_i
!C    real   (kind=KREAL), pointer :: D_i(:)
    real   (kind=KREAL), pointer :: B_i(:)
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:)
    integer(kind=kint ), pointer :: IAU_i(:), IAL_i(:)
!C    real   (kind=KREAL), pointer :: AU_i(:), AL_i(:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_i(:)
    integer(kind=kint ), pointer :: NEIBPE_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_i(:)
    
    integer(kind=kint ), intent(out) :: index3(0:)    
    integer(kind=kint ), allocatable :: index2(:)
    integer(kind=kint ), allocatable :: index(:), index_nodes(:)

    
    !C-- index arrays for parmetis
    integer(kind=kint ), pointer :: vtxdist(:), xadj(:)
    integer(kind=kint ), pointer :: vwgt(:), adjwgt(:)
    integer(kind=kint ), pointer :: adjncy(:), part(:)
    integer(kind=kint ) :: wgtflag, numflag, options(0:2), edgecut
    integer(kind=kint ) :: ncon
    REAL   (kind=kfloat):: ubvec
    integer(kind=kint ), intent(in)  :: nparts
    REAL   (kind=kfloat), intent(in) :: tpwgts(:)
    
    !C-- global number vector for parmetis
    integer(kind=kint ), pointer :: global_number_vector(:)
    
    integer(kind=kint ) :: i, j, k, l, order, ierr, ret

    integer(kind=kint ), intent(out) :: SWITCH_FLAG
    integer(kind=kint ), intent(out) :: new_part(:)

    integer(kind=kint ), allocatable :: dist_nodes(:), comm_order(:)

    CHARACTER(len= 4 ) :: penum


    !C my_rank, PETOT
    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 


    if(method == 1) then
#ifndef PARMETIS
       write(*,*) "Error: Compiled without PARMETIS Flag "
       stop
#endif       
#ifdef PARMETIS       
       !C-- matrix transform
       CALL count_time(1, solver_comm, my_rank, 11)
       
       allocate(dist_nodes(PETOT), comm_order(PETOT))
       CALL dist_parmetis_form_mtom(N_i, NP_i, INU_i, INL_i, IAU_i, IAL_i,          &
            & NEIBPETOT_i, NEIBPE_i, STACK_IMPORT_i, STACK_EXPORT_i, NOD_IMPORT_i,  &
            & NOD_EXPORT_i, vtxdist, xadj, adjncy, part, my_rank, PETOT,            &
            & global_number_vector, SOLVER_COMM, dist_nodes, order, comm_order)

       CALL count_time(2, solver_comm, my_rank, 11) 
       
       !C-- default setting
       options(0) = 0;  numflag = 0
       ubvec = 1.05; ncon = 1 

       !C-- weight setting: each vertex has weight equal to the number of nonzeros 
       !C-- weights only on vertices  
       wgtflag = 2
       ALLOCATE(vwgt(0:vtxdist(my_rank + 1) - vtxdist(my_rank) - 1), adjwgt(1))
       DO i = 0, vtxdist(my_rank + 1) - vtxdist(my_rank) - 1
          vwgt(i) = xadj(i + 1) - xadj(i) + 1
       END DO

       !C-- parmetis 
       CALL count_time(1, solver_comm, my_rank, 12)                     

       !C-- part is initialized for nothing done,
       CALL PARMETIS_V3_PARTKWAY(vtxdist, xadj, adjncy, vwgt, adjwgt,  &
            & wgtflag, numflag, ncon, nparts, tpwgts, ubvec, options,  &
            & edgecut, part, SOLVER_COMM, ret)
       
       if(ret > 0) stop 'ParMETIS stops! in matrix_dist_mtom'
       
       CALL dist_reorder_vector(dist_nodes, vtxdist, part, new_part,  &
            & N_i, my_rank, PETOT, SOLVER_COMM, order, comm_order)
       
       DEALLOCATE(dist_nodes, comm_order)
       DEALLOCATE(vwgt, adjwgt, xadj, adjncy, part)
       
       CALL count_time(2, solver_comm, my_rank, 12)                         
       SWITCH_FLAG = 0
#endif
    else if(method == 2) then

       call easy_dist(N_i, NP_i, new_part, my_rank, PETOT, SOLVER_COMM, nparts, tpwgts)
       SWITCH_FLAG = 0
       
    else if(method == 3) then
       
       call easy_dist_no_reordering(N_i, new_part, my_rank, PETOT, SOLVER_COMM, nparts, tpwgts, index3)
       SWITCH_FLAG = 1    

    else if(method == 4) then
       
       call RCM_ordering(N_i, NP_i, INU_i, INL_i, IAU_i, IAL_i, new_part, my_rank, PETOT, SOLVER_COMM, index3)
       SWITCH_FLAG = 1    

    endif
    
  END SUBROUTINE dist_method


  
  SUBROUTINE set_order(nds_index, order, N, SOLVER_COMM)
    implicit none
    include 'mpif.h'
    include 'precision.inc'
    !C- nds_index(0:PETOT), order(N)
    integer(kind=kint) :: nds_index(0:), order(:), i, j, ierr
    integer(kind=kint), intent(in) :: SOLVER_COMM, N
    integer(kind=kint) :: my_rank, PETOT

    !C my_rank, PETOT
    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 
    
    CALL MPI_ALLGATHER( N, 1, MPI_INTEGER, nds_index(1), 1, MPI_INTEGER,  &
         &  SOLVER_COMM, ierr )
     nds_index(0) = 0
     DO i = 1, PETOT
        nds_index(i) = nds_index(i - 1) + nds_index(i)
     END DO

     j = nds_index(my_rank)
     DO i = 1, N 
        order(i) = i + j
     END DO
    
  END SUBROUTINE set_order

  !C OUTPUT data: vec_o(N)
  !C INPUT  data: vec_i(N_i)
  SUBROUTINE distribute_vector(nds_index, order, SOLVER_COMM, vec_i, N_i, vec_o,NB)
    implicit none
    include 'mpif.h'
    include 'precision.inc'
    !C- nds_index(0:PETOT), order(N_i)
    integer(kind=kint ) :: nds_index(0:), order(:), i, j, k, l, N, is, ie,k1
    REAL   (kind=kreal) :: vec_i(:), vec_o(:)
    integer(kind=kint), intent(in) ::  N_i,NB
    integer(kind=kint) ::  PETOT, my_rank, SOLVER_COMM, ierr

    integer(kind=kint) :: count_on_PE_PETOT, send_buf_size
    integer(kind=kint), allocatable :: in_send_buf(:), send_buf(:), recv_buf(:)
    integer(kind=kint), allocatable :: comm_send_sz(:), comm_recv_sz(:)
    REAL (kind=kreal), allocatable :: recv_buf_dbl(:), send_buf_dbl(:)
    
    integer(kind=kint) :: recv_buf_size


    INTEGER, allocatable :: req1(:), req2(:), req3(:), req4(:), sta(:,:)

    !C my_rank, PETOT
    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 
    


    N = nds_index(my_rank + 1) - nds_index(my_rank)
    
    allocate(in_send_buf(0:PETOT))
    in_send_buf = 0

    !C count up for each PE
    count_on_PE_PETOT = 0
    DO i = 1, N_i
       k = order(i)
       DO j = 1, PETOT
          if(k <= nds_index(j)) exit
       END DO
       if(j > PETOT) stop "Unexpected error in distribute_vector()"
       
       if(j + 1 <= PETOT) then
          in_send_buf(j + 1) = in_send_buf(j + 1) + 1
       else
          count_on_PE_PETOT = count_on_PE_PETOT + 1
       end if
    END DO

    DO i = 1, PETOT
       in_send_buf(i) = in_send_buf(i-1) + in_send_buf(i)
    END DO

    !C send_buf_size
    send_buf_size = in_send_buf(PETOT) + count_on_PE_PETOT
    
    allocate(send_buf(send_buf_size), send_buf_dbl(send_buf_size*NB))
    

    !C create send_buf
    DO i = 1, N_i
       k = order(i)
       DO j = 1, PETOT
          if(k <= nds_index(j)) exit
       END DO
       if(j > PETOT) stop "Unexpected error in distribute_vector()"
       
       l = in_send_buf(j) + 1
       in_send_buf(j) = l
       send_buf(l) = k

       DO k1=1,NB
          send_buf_dbl((l-1)*NB+k1) = vec_i((i-1)*NB+k1)
       END DO
    END DO

    !C alltoall send size
    allocate(comm_send_sz(PETOT), comm_recv_sz(PETOT))

    DO i = 1, PETOT
       comm_send_sz(i) = in_send_buf(i) - in_send_buf(i - 1)
    END DO
    
    CALL MPI_ALLTOALL(comm_send_sz, 1, MPI_INTEGER, comm_recv_sz, 1, MPI_INTEGER, &
         &            SOLVER_COMM, ierr)

    
    !C count recv_buf
    recv_buf_size = 0
    DO i = 1, PETOT
       recv_buf_size = recv_buf_size + comm_recv_sz(i)

    END DO


    allocate(recv_buf(recv_buf_size), recv_buf_dbl(recv_buf_size*NB))
    allocate(req1(PETOT), req2(PETOT), req3(PETOT), req4(PETOT), sta(MPI_STATUS_SIZE, PETOT))


    !C recv 
    j = 1; l = 0
    DO i = 1, PETOT
       if(comm_recv_sz(i) > 0 .and. i - 1 /= my_rank) then
          l = l + 1
          CALL MPI_IRECV(recv_buf(j), comm_recv_sz(i), MPI_INTEGER, i - 1, 0, SOLVER_COMM, req1(l), ierr)
          CALL MPI_IRECV(recv_buf_dbl((j-1)*NB+1), comm_recv_sz(i)*NB, MPI_DOUBLE_PRECISION, i - 1, 0, SOLVER_COMM, req2(l), ierr)
       end if
       j = j + comm_recv_sz(i)
    END DO
    
    !C send
    k = 0
    DO i = 1, PETOT
       is = in_send_buf(i - 1) + 1
       j = comm_send_sz(i)
       if(j > 0 .and. i - 1 /= my_rank) then
          k = k + 1
          CALL MPI_ISEND(send_buf(is), j, MPI_INTEGER, i - 1, 0, SOLVER_COMM, req3(k), ierr)
          CALL MPI_ISEND(send_buf_dbl((is-1)*NB+1), j*NB, MPI_DOUBLE_PRECISION, i - 1, 0, SOLVER_COMM, req4(k), ierr)
       end if
    END DO

    !C own PE's nodes copy to recv_buf
    j = 0
    DO i = 1, my_rank 
       j = j + comm_recv_sz(i)
    END DO
    DO i = in_send_buf(my_rank) + 1, in_send_buf(my_rank + 1)
       j = j + 1
       recv_buf(j) = send_buf(i)
       recv_buf_dbl((j-1)*NB+1:j*NB) = send_buf_dbl((i-1)*NB+1:i*NB)
    END DO

    CALL MPI_WAITALL(l, req1,  sta,  ierr)               
    CALL MPI_WAITALL(l, req2,  sta,  ierr)               
    CALL MPI_WAITALL(k, req3,  sta,  ierr)               
    CALL MPI_WAITALL(k, req4,  sta,  ierr)               

    !C owner of the solution
    if(N > 0) then
       vec_o = 0
       DO i = 1, recv_buf_size
          j = recv_buf(i) - nds_index(my_rank)
          
          vec_o((j-1)*NB+1:j*NB) = recv_buf_dbl((i-1)*NB+1:i*NB)
       END DO
    end if
    
    deallocate(req1, req2, req3, req4, sta)
    deallocate(recv_buf, recv_buf_dbl)
    deallocate(comm_send_sz, comm_recv_sz)
    deallocate(in_send_buf)
    deallocate(send_buf, send_buf_dbl)
  END SUBROUTINE distribute_vector
  
  SUBROUTINE parmetis_form_mtom(N_i, NP_i, INU_i, INL_i, IAU_i, IAL_i,        &
       & NEIBPETOT_i, NEIBPE_i, STACK_IMPORT_i, STACK_EXPORT_i, NOD_IMPORT_i, &
       & NOD_EXPORT_i, vtxdist, xadj, adjncy, part, my_rank, PETOT,           &
       & global_number_vector, SOLVER_COMM)
    use solver_SR2
    use solver_Gnumbering
    implicit none
    include 'mpif.h'
    include 'precision.inc'    
    integer(kind=kint) :: N_i, NP_i, NEIBPETOT_i
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:), IAU_i(:), IAL_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:), STACK_EXPORT_i(:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:), NOD_EXPORT_i(:), NEIBPE_i(:)
    integer(kind=kint ), pointer :: vtxdist(:), xadj(:)
    integer(kind=kint ), pointer :: adjncy(:), part(:)
    integer(kind=kint ), intent(in) :: my_rank, PETOT, SOLVER_COMM
    
    integer(kind=kint ), pointer :: global_number_vector(:)
    integer(kind=kint ), allocatable :: WSI(:), WRI(:)
    
    integer(kind=kint) :: NPL, NPU, GN, ierr
    
    integer(kind=kint) :: i, j, k, S, R
    
    !C-- NPL and NPU are different from other subroutines' NPL,NPU
    !C-- which are for rows of 1..N_i
    !C-- part is allocated of NP_i for later calculation
    NPL = INL_i(N_i); NPU = INU_i(N_i)
    if(N_i == 0) then
       NPL = 0
       NPU = 0
       allocate(part(0), adjncy(0))
    else if(N_i > 0) then
       allocate( adjncy(0:NPL + NPU - 1), part(0:NP_i - 1))
    end if
    ALLOCATE(vtxdist(0:PETOT),xadj(0:N_i))
    
    !C-- it is assumed that # of PEs is equal to # of domains
    !C-- vtxdist: vertex number is 0..NODTOT-1
    call MPI_ALLGATHER(N_i, 1, MPI_INTEGER, vtxdist(1), 1, MPI_INTEGER, &
         & SOLVER_COMM, ierr)
    vtxdist(0) = 0
    DO i = 1, PETOT
       vtxdist(i) = vtxdist(i - 1) + vtxdist(i)
    END DO
    GN = vtxdist(PETOT)

    !C-- global number vector for local nodes 1..NP_i
    !C-- global number is 0..GN-1
    j = vtxdist(my_rank)
    allocate(global_number_vector(NP_i))
    DO i = 1, N_i
       global_number_vector(i) =  i + j - 1
    END DO

    if(NEIBPETOT_i > 0) then
       S = STACK_EXPORT_i(NEIBPETOT_i); R = STACK_IMPORT_i(NEIBPETOT_i)
       ALLOCATE(WSI(S), WRI(R)) 
       CALL  SOLVER_SEND_RECV2I                                                   &
            &  (NP_i, S, R, NEIBPETOT_i, NEIBPE_i, STACK_IMPORT_i, NOD_IMPORT_i,  &
            &                          STACK_EXPORT_i, NOD_EXPORT_i,              &
            &    WSI, WRI, global_number_vector, SOLVER_COMM, my_rank)
       DEALLOCATE(WSI, WRI)
    end if

    !C-- xadj, adjncy : row i => adjncy( xadj(i) .. xadj(i + 1) - 1 )
    xadj(0) = 0
    DO i = 1, N_i
       !C row i - 1 of pmetis <=> row i of ISTACK
       xadj(i) = xadj(i - 1) + INL_i(i) - INL_i(i - 1) + INU_i(i) - INU_i(i - 1)
       
       k = xadj(i - 1) - INL_i(i - 1) - 1
       DO j = INL_i(i - 1) + 1, INL_i(i)
          adjncy(k + j) = global_number_vector(IAL_i(j))
       END DO
       k = xadj(i - 1) - 1 + INL_i(i) - INL_i(i - 1) - INU_i(i - 1)  
       DO j = INU_i(i - 1) + 1, INU_i(i)
          adjncy(k + j) = global_number_vector(IAU_i(j))
       END DO
    END DO
    
  END SUBROUTINE parmetis_form_mtom


  SUBROUTINE copy_sub_matrix(buf, INL, INU, IAL, IAU, base, bound, global_number_vector, size)
    IMPLICIT none
    INCLUDE 'precision.inc'
    INTEGER(kind=kint), intent(in)  :: INL(0:), INU(0:), base, bound
    INTEGER(kind=kint), intent(in)  :: IAL(:), IAU(:), global_number_vector(:)
    INTEGER(kind=kint), intent(out) :: buf(:), size
    INTEGER(kind=kint) :: i, j, k, rows

    rows = bound - base
    
    k = 0
    DO i = 1, rows
       DO j = INL(base + i - 1) + 1, INL(base + i)
          k = k + 1
          buf(k) = global_number_vector(IAL(j))
       END DO
       
       DO j = INU(base + i - 1) + 1, INU(base + i)
          k = k + 1
          buf(k) = global_number_vector(IAU(j))
       END DO
    END DO
    size = k
    
  END SUBROUTINE copy_sub_matrix


  SUBROUTINE countup(size, INL, INU, base, bound)
    implicit none
    include 'precision.inc'
    INTEGER(kind=kint), intent(in)  :: INL(0:), INU(0:), base, bound
    INTEGER(kind=kint), intent(out) :: size
    INTEGER(kind=kint) :: i, j, k

    !C initialize the row size for index

    size = 0
    DO i = base + 1, bound
       size = size + INL(i) - INL(i - 1) + INU(i) - INU(i - 1)
    END DO
    
  END SUBROUTINE countup

  SUBROUTINE make_index(newindex, INL, INU, base, bound, buf_size)
    IMPLICIT NONE
    INCLUDE 'precision.inc'
    INTEGER(kind=kint), intent(in)  :: INL(0:), INU(0:), base, bound
    INTEGER(kind=kint), intent(out) :: newindex(:)
    INTEGER(kind=kint), intent(inout) :: buf_size
    INTEGER(kind=kint) :: i, j, k
    
    !C initialize the row size for index
    
    DO i = base + 1, bound
       j = INL(i) - INL(i - 1) + INU(i) - INU(i - 1)
       newindex(i - base) = j
       buf_size = buf_size + j
    END DO
    
  END SUBROUTINE make_index
  
  
  SUBROUTINE dist_parmetis_form_mtom(N_i, NP_i, INU_i, INL_i, IAU_i, IAL_i,   &
       & NEIBPETOT_i, NEIBPE_i, STACK_IMPORT_i, STACK_EXPORT_i, NOD_IMPORT_i, &
       & NOD_EXPORT_i,vtxdist, xadj, adjncy, part, my_rank, PETOT,            &
       & global_number_vector, SOLVER_COMM, dist_nodes, order, comm_order)

    use solver_SR2
    use solver_Gnumbering
    implicit none
    include 'mpif.h'
    include 'precision.inc'    
    INTEGER(kind=kint) :: N_i, NP_i, NEIBPETOT_i
    INTEGER(kind=kint ), pointer :: INU_i(:), INL_i(:), IAU_i(:), IAL_i(:)
    INTEGER(kind=kint ), pointer :: STACK_IMPORT_i(:), STACK_EXPORT_i(:)
    INTEGER(kind=kint ), pointer :: NOD_IMPORT_i(:), NOD_EXPORT_i(:), NEIBPE_i(:)
    INTEGER(kind=kint ), pointer :: vtxdist(:), xadj(:)
    INTEGER(kind=kint ), pointer :: adjncy(:), part(:)
    INTEGER(kind=kint ), intent(in) :: my_rank, PETOT, SOLVER_COMM
    INTEGER(kind=kint ), intent(out):: dist_nodes(:), order, comm_order(:)
    
    INTEGER(kind=kint ), pointer :: global_number_vector(:)
    INTEGER(kind=kint ), allocatable :: WSI(:), WRI(:)
    
    INTEGER(kind=kint) :: NPL, NPU, GN, ierr
    
    INTEGER(kind=kint) :: i, j, k, l, m, S, R, dist_N_i, dist_base, dist_bound, max_pe, min_pe
    INTEGER(kind=kint), allocatable :: nodes_array_to_send(:), dist_base_nodes(:)
    INTEGER(kind=kint), allocatable :: rows_size(:), send_buf(:), size(:)
    INTEGER(kind=kint) :: array_size, send_buf_size, used_size

    INTEGER, allocatable :: req1(:), req2(:), sta(:,:)
    
    ALLOCATE(req1(PETOT), req2(PETOT), sta(MPI_STATUS_SIZE, PETOT), size(PETOT))
    
    ALLOCATE(vtxdist(0:PETOT))
    
    !C-- it is assumed that # of PEs is equal to # of domains
    CALL MPI_ALLGATHER(N_i, 1, MPI_INTEGER, vtxdist(1), 1, MPI_INTEGER, &
         &   SOLVER_COMM, ierr)
    vtxdist(0) = 0

    allocate(dist_base_nodes(PETOT))

    GN = 0
    DO i = 1, PETOT
       GN = GN + vtxdist(i)
    END DO
    
    !C-- dist_base_nodes has node numbers on other PE for nodes to send to that PE.
    !C-- This part will send the nodes after PE's own nodes.
    dist_base_nodes = GN / PETOT
    dist_base_nodes(1) = 0
    
    j = mod(GN, PETOT)
    DO i = PETOT - j + 2, PETOT
       dist_base_nodes(i) = dist_base_nodes(i) + 1 
    END DO
    DO i = 2, PETOT
       dist_base_nodes(i) = dist_base_nodes(i - 1) + dist_base_nodes(i) 
    END DO

    DO i = 1, PETOT
       dist_base_nodes(i) =  vtxdist(i) + dist_base_nodes(i) 
    END DO

    
    !C-- record the nodes size to send in vtxdist array
    DO i = 1, PETOT
       vtxdist(i) = vtxdist(i) -  GN / PETOT
    END DO
    DO i = PETOT - j + 1, PETOT
       vtxdist(i) = vtxdist(i) -  1
    END DO
    

    !C-- In order to check Sender PE, record -1 @ dist_base_nodes(#PE + 1)
    !C-- In order to check do-nothing PE, record -2 @ dist_base_nodes(#PE + 1)
    DO i = 1, PETOT
       if(vtxdist(i) > 0) then
          dist_base_nodes(i) = -1
       else if(vtxdist(i) == 0) then
          dist_base_nodes(i) = -2
       end if
    END DO

    !C-- redistribution nodes with vtxdist(:)
    order = 0
    dist_nodes = 0 
    comm_order = 0
    if(vtxdist(my_rank + 1) > 0) then 
       !C sender PE
       allocate(nodes_array_to_send(PETOT * 2))
       nodes_array_to_send = 0
    end if

    !C-- matching the PEs based on node size
    !C   max_pe and min_pe is 1 + rank # of the PE
    do
       max_pe = 1; min_pe = 1
       do i = 2, PETOT
          if(vtxdist(max_pe) < vtxdist(i)) max_pe = i
          if(vtxdist(min_pe) > vtxdist(i)) min_pe = i
       end do
       
       !C-- all matching is finished
       if(vtxdist(max_pe) == 0) exit
       
       j = vtxdist(max_pe) + vtxdist(min_pe)
       
       if(j >= 0) then
          !C-- all nodes needed for min_pe is allocated from max_pe
          if(my_rank + 1 == max_pe) then
             !C- sender
             nodes_array_to_send(min_pe * 2 - 1) = dist_base_nodes(min_pe) 
             nodes_array_to_send(min_pe * 2)     = dist_base_nodes(min_pe) - vtxdist(min_pe)
             dist_nodes(min_pe) =  -1 * vtxdist(min_pe)
          else if(my_rank + 1 == min_pe) then
             dist_nodes(max_pe) =  -1 * vtxdist(min_pe)
             order = order + 1 
             comm_order(order) = max_pe
          end if

          dist_base_nodes(min_pe) = dist_base_nodes(min_pe) - vtxdist(min_pe)
          
          vtxdist(max_pe) = j
          vtxdist(min_pe) = 0
          
       else if(j < 0) then
          !C-- part of nodes needed with min_pe is allocated from max_pe
          if(my_rank + 1 == max_pe) then
             nodes_array_to_send(min_pe * 2 - 1) = dist_base_nodes(min_pe) 
             nodes_array_to_send(min_pe * 2)     = dist_base_nodes(min_pe) - vtxdist(min_pe) + j
             dist_nodes(min_pe) = j - vtxdist(min_pe)
          else if(my_rank + 1 == min_pe) then
             dist_nodes(max_pe) = j - vtxdist(min_pe)
             order = order + 1
             comm_order(order) = max_pe
          end if
          dist_base_nodes(min_pe) = dist_base_nodes(min_pe) - vtxdist(min_pe) + j
          vtxdist(max_pe) = 0
          vtxdist(min_pe) = j
       end if
       
    end do
    
    !C-- change the vtxdist(:) for new distribution
    vtxdist = GN / PETOT
    vtxdist(0) = 0
    j = mod(GN, PETOT)
    DO i = PETOT - j + 1, PETOT
       vtxdist(i) = vtxdist(i) + 1
    END DO
    DO i = 1, PETOT
       vtxdist(i) = vtxdist(i - 1) + vtxdist(i)
    END DO
    
    !C-- global number vector for local nodes 1..NP_i
    !C-- global number is 0..GN-1
    dist_N_i = vtxdist(my_rank + 1) - vtxdist(my_rank)
    allocate(global_number_vector(NP_i), part(0:dist_N_i - 1), xadj(dist_N_i))


    if(dist_base_nodes(my_rank + 1) /= -1) then
       !C- Not sender PE
       DO i = 1, N_i
          global_number_vector(i) =  i + vtxdist(my_rank) - 1
       END DO
       
    else if(dist_base_nodes(my_rank + 1) == -1) then

       !C- sender PE
       DO i = 1, dist_N_i
          global_number_vector(i) =  i + vtxdist(my_rank) - 1
       END DO

       !C this part allocating rows in order of PE number
       !C it can be other allocation order
       order = 0
       DO i = 1, PETOT
          if(dist_nodes(i) > 0) then
             !C there are some rows to send
             order = order + 1
             comm_order(order) = i
          end if
       END DO
       
       k = dist_N_i
       DO i = 1, order
          l = comm_order(i)
          DO j = nodes_array_to_send(l * 2 - 1) + 1, nodes_array_to_send(l * 2)
             k = k + 1
             global_number_vector(k) =  j - 1
          END DO
       END DO
    end if

    if(NEIBPETOT_i > 0) then
       S = STACK_EXPORT_i(NEIBPETOT_i)
       R = STACK_IMPORT_i(NEIBPETOT_i)
       ALLOCATE(WSI(S), WRI(R)) 
       CALL  SOLVER_SEND_RECV2I                                                   &
            &  (NP_i, S, R, NEIBPETOT_i, NEIBPE_i, STACK_IMPORT_i, NOD_IMPORT_i,  &
            &                          STACK_EXPORT_i, NOD_EXPORT_i,              &
            &    WSI, WRI, global_number_vector, SOLVER_COMM, my_rank)
       DEALLOCATE(WSI, WRI)
    end if

    !C-- communication sizes are informed
    if(dist_base_nodes(my_rank + 1) == -1) then
       !C-- sender
       
       !C allocation of the number of rows
       !C dist_nodes(:) record the size of nodes to send for each PE
       k = 0
       DO i = 1, order
          l = comm_order(i)
          j = dist_nodes(l)
          k = k + j
       END DO
       allocate(rows_size(k))
       
       send_buf_size = 0
       !C index k points the initial row of sendbuf
       k = dist_N_i; m = 1
       DO i = 1, order
          l = comm_order(i)
          j = dist_nodes(l)
          !C pack the number of nonzeros per row
          CALL make_index(rows_size(m:), INL_i, INU_i, k, k + j, send_buf_size)
          CALL MPI_ISEND(rows_size(m), j, MPI_INTEGER, l - 1, 0, SOLVER_COMM, req1(i), ierr)
          k = k + j; m = m + j
       END DO
       
       CALL MPI_WAITALL(order, req1,  sta,  ierr)


       !C allocate send_buf
       allocate(send_buf(send_buf_size))

       !C create & send send_buf
       !C index k points the initial row number of sendbuf
       k = dist_N_i; used_size = 0
       DO i = 1, order
          l = comm_order(i)
          j = dist_nodes(l)
          !C-- countup nonzeros in row k+1..k+j
          
          CALL copy_sub_matrix(send_buf(used_size + 1:), INL_i, INU_i, IAL_i, IAU_i, k, &
               &               k + j, global_number_vector, m)
          k = k + j

          CALL MPI_ISEND(send_buf(used_size + 1), m, MPI_INTEGER, l - 1 , 0, &
               &         SOLVER_COMM, req1(i), ierr)
          used_size = used_size + m
       END DO
       
       !C-- create own matrix
       allocate(xadj(0:dist_N_i))
       xadj(0) = 0 

       CALL make_index(xadj(1:), INL_i, INU_i, 0, dist_N_i, i)
       DO i = 1, dist_N_i
          xadj(i) = xadj(i - 1) + xadj(i)
       END DO

       !C adjncy
       call countup(i, INL_i, INU_i, 0, dist_N_i)

       allocate(adjncy(0:i-1))
       if(dist_N_i > 0) then
          CALL copy_sub_matrix(adjncy(0:), INL_i, INU_i, IAL_i, IAU_i, 0, dist_N_i, global_number_vector, i)
       end if
       
       CALL MPI_WAITALL(order, req1,  sta,  ierr )

       deallocate(nodes_array_to_send)
       deallocate(rows_size, send_buf)       

    else if(dist_base_nodes(my_rank + 1) > 0) then
       !C-- receiver

       !C allocate the number of rows
       allocate(xadj(0:dist_N_i))
       xadj = 0
       
       !C recv index from other PEs
       l = N_i + 1
       DO i = 1, order
          !C dist_nodes(i) has the number of rows sent from PE i-1
          j = comm_order(i)
          CALL MPI_IRECV(xadj(l), dist_nodes(j), MPI_INTEGER, j - 1, 0, SOLVER_COMM, req2(i), ierr)
          l = l + dist_nodes(j)
       END DO
       
       !C own index
       CALL make_index(xadj(1:), INL_i, INU_i, 0, N_i, i)
       
       CALL MPI_WAITALL(order, req2,  sta,  ierr)               
       
       !C size(i) has the number of nonzeros sent from PE i-1
       !C array_size has the number of nonzeros

       array_size = 0
       if(N_i > 0) array_size = INL_i(N_i) + INU_i(N_i)
       size = 0; l = N_i
       DO i = 1, order
          m = comm_order(i)
          k = 0
          DO j = 1, dist_nodes(m)
             l = l + 1
             k = k + xadj(l)
          END DO
          size(m) = k
          array_size = array_size + k
       END DO
       
       !C allocation of the matrix
       allocate(adjncy(0:array_size - 1))
       k = 0
       if(N_i > 0) k = INL_i(N_i) + INU_i(N_i) 


       DO i = 1, order
          j = comm_order(i)
          CALL MPI_IRECV(adjncy(k), size(j), MPI_INTEGER, j - 1, 0, SOLVER_COMM, req2(i), ierr)
          k = k + size(j)
       END DO
       
       CALL MPI_WAITALL(order, req2,  sta,  ierr)        

       IF(N_i > 0 ) then
          CALL copy_sub_matrix(adjncy(0:), INL_i, INU_i, IAL_i, IAU_i, 0, N_i, global_number_vector, m)
       END IF
       
       !C neat the xadj(:)
       DO i = 1, dist_N_i
          xadj(i) = xadj(i - 1) + xadj(i)
       END DO
       
    else if(dist_base_nodes(my_rank + 1) == -2) then
       !C-- do-nothing PE
       !C dist_nodes are set 0
       dist_nodes = 0

       if(N_i == 0) then
          NPL = 0; NPU = 0
       else
          NPL = INL_i(N_i); NPU = INU_i(N_i)
       end if

       ALLOCATE(xadj(0:N_i), adjncy(0:NPL + NPU - 1), part(0:NP_i - 1))

       
       xadj(0) = 0
       DO i = 1, N_i
          !C row i - 1 of pmetis <=> row i of ISTACK
          xadj(i) = xadj(i - 1) + INL_i(i) - INL_i(i - 1) + INU_i(i) - INU_i(i - 1)
          
          k = xadj(i - 1) - INL_i(i - 1) - 1
          DO j = INL_i(i - 1) + 1, INL_i(i)
             adjncy(k + j) = global_number_vector(IAL_i(j))
          END DO
          k = xadj(i - 1) - 1 + INL_i(i) - INL_i(i - 1) - INU_i(i - 1)  
          DO j = INU_i(i - 1) + 1, INU_i(i)
             adjncy(k + j) = global_number_vector(IAU_i(j))
          END DO
       END DO
    else 
       stop 'error in dist_parmetis_form'
    end if

    deallocate(req1, req2, sta, size, dist_base_nodes)    
    deallocate(global_number_vector)

  END SUBROUTINE dist_parmetis_form_mtom
  
  SUBROUTINE RCM_ordering(N_i, NP_i, INU_i, INL_i, IAU_i, IAL_i,  &
       & new_part, my_rank, PETOT, SOLVER_COMM, index3)
    USE isort
    implicit none
    include 'precision.inc'
    include 'mpif.h'
    integer(kind=kint), intent(in) :: N_i, NP_i, my_rank, PETOT, SOLVER_COMM
    integer(kind=kint), pointer    :: INU_i(:), INL_i(:), IAU_i(:), IAL_i(:)
    integer(kind=kint), intent(inout) :: new_part(:), index3(0:)
    
    integer(kind=kint) :: i, j, k, ierr, nd, base
    integer(kind=kint) :: md, nd_index, index
    integer(kind=kint), allocatable :: ORD(:), TEMP_DEG(:), TEMP_NDS(:), CHECK(:)
    
    CALL MPI_ALLGATHER(N_i, 1, MPI_INTEGER, index3(1), 1, MPI_INTEGER, SOLVER_COMM, ierr)

    !C base is the base number..
    base = 0
    DO i = 1, my_rank
       base = base + index3(i)
    END DO
    
    index3(0) = 0
    DO i = 1, PETOT
       index3(i) = index3(i-1) + index3(i)
    END DO

    !C check array 
    allocate(CHECK(N_i))

    !C max number of degrees are evaluated in md
    md = INU_i(1) + INL_i(1)
    CHECK(1) = md
    DO i = 2, N_i
       j = INU_i(i) - INU_i(i-1) + INL_i(i) - INL_i(i-1)
       CHECK(i) = j
       if(j > md) md = j
    END DO
    
    !C create arrays according to the max degrees
    allocate(ORD(md), TEMP_DEG(md), TEMP_NDS(md))
    
    !C pop the node number in ascending order of the edge number
    !C initialize the new_part array
    !C index shows the number of nodes recorded in new_part(:)
    new_part = -1
    index = 0  
    DO i = 1, N_i
       !C i indexes the new_part array in order from N_i to 1
       nd = new_part(N_i + 1 - i)
       if(nd == -1) then
          !C select the initial node again
          DO j = 1, N_i
             IF(CHECK(j) >= 0) EXIT
          END DO
          if(j <= N_i)  then
             nd = j
          else 
             stop "No unselected nodes: unexpected error in RCM_ordering" 
             exit
          end if
          index = index + 1
          new_part(N_i + 1 - i) = nd + base
          CHECK(nd) = -1
       else
          !C change nd to local number
          nd = nd - base
       end if
       
       !C store nd's neighbors in  TEMP_NDS(:)
       nd_index = 0
       DO j = INL_i(nd-1) + 1, INL_i(nd)
          k = IAL_i(j)
          if(CHECK(k) >= 0) then
             nd_index = nd_index + 1
             TEMP_NDS(nd_index) = k
             TEMP_DEG(nd_index) = CHECK(k)
             CHECK(k) = -1
          end if
       END DO
       DO j = INU_i(nd-1) + 1, INU_i(nd)
          k = IAU_i(j)
          if(k <= N_i) then
             if(CHECK(k) >= 0) then
                nd_index = nd_index + 1
                TEMP_NDS(nd_index) = k
                TEMP_DEG(nd_index) = CHECK(k)
                CHECK(k) = -1
             end if
          end if
       END DO
       
       !C sort TEMP_NDS(:) in descending order
       CALL QSORTI(ORD, nd_index, TEMP_DEG)

       !C copy TEMP_NDS to new_part in the descending order of degree
       DO j = nd_index, 1, -1
          nd = TEMP_NDS(ORD(j))
          index = index + 1
          new_part(N_i + 1 - index) = nd + base
       END DO

       if(index >= N_i) exit
    END DO

    deallocate(ORD, TEMP_DEG, TEMP_NDS, CHECK)
    
  END SUBROUTINE RCM_ordering
  
  
  SUBROUTINE easy_dist_no_reordering(N_i, new_part, my_rank, PETOT, SOLVER_COMM, nparts, tpwgts, index3)
    implicit none
    include 'precision.inc'
    include 'mpif.h'

    integer(kind=kint),   intent(in)  :: N_i, my_rank, PETOT, SOLVER_COMM, nparts
    integer(kind=kint),   intent(out) :: new_part(0:), index3(0:)
    REAL   (kind=kfloat), intent(in)  :: tpwgts(:)

    integer(kind=kint) :: i, j, k, ierr
    
    call MPI_ALLGATHER(N_i, 1, MPI_INTEGER, index3(1), 1, MPI_INTEGER, SOLVER_COMM, ierr)

    !C ordering is stored in new_part. k is tha base number.. 
    k = 0 
    DO i = 1, my_rank
       k = k + index3(i)
    END DO
    DO i = 1, N_i
       new_part(i - 1) = k + i
    END DO
    
    !C j is the sum of nodes
    j = 0
    DO i = 1, PETOT
       j = j + index3(i)
    END DO

    !C according to tpwgts, row size is determined
    index3(0) = 0
    DO i = 1, PETOT - 1
       if(i >= nparts) then
          index3(i) =  j
       else
          index3(i) =  index3(i - 1) + tpwgts(i) * j
       end if
    END DO
    index3(PETOT) = j

  END SUBROUTINE easy_dist_no_reordering

  
  SUBROUTINE easy_dist(N_i, NP_i, part, my_rank, PETOT, SOLVER_COMM, nparts, tpwgts)
    implicit none
    include 'mpif.h'
    include 'precision.inc'    
    INTEGER(kind=kint) :: N_i, NP_i
    INTEGER(kind=kint ), intent(in ) :: my_rank, PETOT, SOLVER_COMM
    INTEGER(kind=kint ), intent(out) :: part(:)

    INTEGER(kind=kint) :: GN, ierr
    INTEGER(kind=kint) :: i, j, k, l, dist_N_i, max_pe, min_pe
    INTEGER(kind=kint), allocatable :: vtxdist(:), dist_nodes(:)

    INTEGER(kind=kint), intent(in) :: nparts
    REAL(kind=kfloat), intent(in) :: tpwgts(:)

    REAL(kind=kfloat) :: v
    

    ALLOCATE(dist_nodes(PETOT), vtxdist(0:PETOT))
    
    !C-- it is assumed that # of PEs is equal to # of domains
    CALL MPI_ALLGATHER(N_i, 1, MPI_INTEGER, vtxdist(1), 1, MPI_INTEGER, &
         &   SOLVER_COMM, ierr)
    vtxdist(0) = 0

    GN = 0
    DO i = 1, PETOT
       GN = GN + vtxdist(i)
    END DO
    
    v = 0
    do i = 1, nparts
       v = v + tpwgts(i)
    end do
    k = GN

    !C-- record the nodes size to send in vtxdist array
    DO i = 1, PETOT
       if(i > nparts) then
          j = 0 
       else 
          j = GN * tpwgts(i) / v
       end if
       k = k - j
       vtxdist(i) = vtxdist(i) -  j
    END DO

    !C here we assume PETOT >= nparts
    do
       DO i = 1, nparts
          if(k > 0  .and. tpwgts(i) > 0) then
             vtxdist(i) = vtxdist(i) - 1
             k = k - 1
          end if
       END DO
       if(k == 0) exit
    end do
    

    
    !C-- dist_N_i is set
    dist_N_i = N_i - vtxdist(my_rank + 1)

    
    dist_nodes = 0 

    !C-- matching the PEs based on node size
    !C   max_pe and min_pe is 1 + rank # of the PE
    do
       max_pe = 1; min_pe = 1
       do i = 2, PETOT
          if(vtxdist(max_pe) < vtxdist(i)) max_pe = i
          if(vtxdist(min_pe) > vtxdist(i)) min_pe = i
       end do
       
       !C-- all matching is finished
       if(vtxdist(max_pe) == 0) exit
       
       j = vtxdist(max_pe) + vtxdist(min_pe)
       
       if(j >= 0) then
          !C-- all nodes needed for min_pe is allocated from max_pe
          if(my_rank + 1 == max_pe) then
             dist_nodes(min_pe) =  -1 * vtxdist(min_pe)
          end if

          vtxdist(max_pe) = j
          vtxdist(min_pe) = 0
          
       else if(j < 0) then
          !C-- part of nodes needed with min_pe is allocated from max_pe
          if(my_rank + 1 == max_pe) then
             dist_nodes(min_pe) = j - vtxdist(min_pe)
          end if
          vtxdist(max_pe) = 0
          vtxdist(min_pe) = j
       end if
    end do
    
    if(N_i > dist_N_i) then
       !C sender
       do i = 1, dist_N_i
          part(i) = my_rank
       end do

       k = dist_N_i 
       do i = 1, PETOT
          do j = k + 1, k + dist_nodes(i)
             part(j) = i - 1
          end do
          k = k + dist_nodes(i)
       end do
    else
       !C receiver or do-nothing PE
       do i = 1, N_i
          part(i) = my_rank
       end do
    end if

  END SUBROUTINE easy_dist


  SUBROUTINE dist_reorder_vector(dist_nodes, vtxdist, part, new_part, N_i,  &
       & my_rank, PETOT, SOLVER_COMM, order, comm_order)
    IMPLICIT NONE
    INCLUDE 'precision.inc'
    INCLUDE 'mpif.h'

    INTEGER(kind=kint) :: order, comm_order(:)
    INTEGER(kind=kint) :: dist_nodes(:), vtxdist(0:), part(0:), new_part(0:)
    INTEGER(kind=kint) :: i, j, k, l, dist_N_i
    INTEGER(kind=kint), intent(in) :: PETOT, my_rank, N_i
    INTEGER, allocatable :: req1(:), req2(:), sta(:,:)
    INTEGER ::  SOLVER_COMM, ierr
    
    allocate(req1(PETOT), req2(PETOT), sta(MPI_STATUS_SIZE, PETOT))    
    
    dist_N_i = vtxdist(my_rank + 1) - vtxdist(my_rank)

    if(dist_N_i < N_i) then
       !C sender PE
       j = dist_N_i
       DO i = 1, order
          l = comm_order(i)
          CALL MPI_IRECV(new_part(j), dist_nodes(l), MPI_INTEGER, l - 1, 0, SOLVER_COMM, req1(i), ierr)
          j = j + dist_nodes(l)
       END DO
       
       DO i = 0, dist_N_i - 1
          new_part(i) = part(i)
       END DO
       
       CALL MPI_WAITALL(order, req1,  sta,  ierr )       
       
    else if (dist_N_i > N_i)  then
       !C receiver PE
       j = N_i
       DO i = 1, order
          l = comm_order(i)
          CALL MPI_ISEND(part(j), dist_nodes(l), MPI_INTEGER, l - 1, 0, SOLVER_COMM, req2(i), ierr)
          j = j + dist_nodes(l)
       END DO

       DO i = 0, N_i - 1
          new_part(i) = part(i)
       END DO
       
       CALL MPI_WAITALL(order, req2,  sta,  ierr )
    else
       !C do-nothing PE
       new_part = part
    end if

    deallocate(req1, req2, sta)
    
  END SUBROUTINE dist_reorder_vector
  

  !C Output data: first three rows of the arguments
  !C INPUT data:  Next three rows of the arguments
  !C MPI communicator and distribution description: The last row
  !C # All the arrays of the output data are allocated in this routine.
  SUBROUTINE distribute_matrix_vector(N, NP, D, INU, INL, IAU, IAL, AU, AL,     &
       & NEIBPETOT, NEIBPE, NOD_IMPORT, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, &
       & B, order_o,                                                            & 
       & N_i, NP_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, NEIBPETOT_i,   &
       & NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i, STACK_IMPORT_i, STACK_EXPORT_i,  &
       & B_i, order_i,                                                          &
       & SOLVER_COMM, part, index_number, SWITCH_FLAG)
    implicit none
    include 'precision.inc'
    include 'mpif.h'
    
    integer(kind=kint), intent(in) :: SWITCH_FLAG
    integer(kind=kint) :: part(:), index_number(:)
    
    integer(kind=kint ), pointer :: order_i(:), order_o(:)

    !C-- initial matrix form
    integer(kind=kint) :: N_i, NP_i, NEIBPETOT_i
    real   (kind=KREAL), pointer :: D_i(:),B_i(:)
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:)
    integer(kind=kint ), pointer :: IAU_i(:), IAL_i(:)
    real   (kind=KREAL), pointer :: AU_i(:)
    real   (kind=KREAL), pointer :: AL_i(:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_i(:)
    integer(kind=kint ), pointer :: NEIBPE_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_i(:)
    
    !C-- redistributed matrix form (NPL, NPU isn't refered in other sub.)
    integer(kind=kint) :: N, NP, NEIBPETOT, NPL, NPU
    real   (kind=KREAL), pointer :: D(:), B(:)
    integer(kind=kint ), pointer :: INU(:), INL(:)
    integer(kind=kint ), pointer :: IAU(:), IAL(:)
    real   (kind=KREAL), pointer :: AU(:)
    real   (kind=KREAL), pointer :: AL(:)
    integer(kind=kint ), pointer :: NOD_IMPORT(:)
    integer(kind=kint ), pointer :: NOD_EXPORT(:)
    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:)

    integer(kind=kint), allocatable :: index(:), index2(:), index_nodes(:)
    
    integer(kind=kint), intent(in) :: SOLVER_COMM
    integer(kind=kint) :: PETOT, my_rank, ierr
    
    integer(kind=kint) :: sz_export, sz_import

    
    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 
    
    
    !C-- SWITCH_FLAG : 0 => input data is PART(:) containing PE number
    !C
    !C-- SWITCH_FLAG : 1 => input data is PART(:) containing global number, and
    !C                            index_number(:) containing index information for each PE
    
    allocate(STACK_IMPORT(0:PETOT), STACK_EXPORT(0:PETOT), NEIBPE(PETOT))

    call size_new_mat(sz_export, sz_import, N, NP, NPL, NPU, NEIBPETOT, NEIBPE, &
         & STACK_IMPORT, STACK_EXPORT, part, index_number,            &
         & N_i, NP_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, &
         & AL_i, NEIBPETOT_i, NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i,             &
         & STACK_IMPORT_i, STACK_EXPORT_i, B_i, SOLVER_COMM,    &
         & order_i, SWITCH_FLAG)

    allocate(NOD_EXPORT(sz_export), NOD_IMPORT(sz_import),  IAL(NPL), &
         &   IAU(NPU), INL(0:NP), INU(0:NP), AL(NPL), AU(NPU), D(NP), B(NP), order_o(N))
    
    call assign_new_mat(N, NP, NPL, NPU, D, INU, INL, IAU, IAL, AU, AL, B, NEIBPETOT,  &
         & NEIBPE, NOD_IMPORT, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, part, index_number,       &
         & N_i, NP_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, NEIBPETOT_i,    &
         & NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i, STACK_IMPORT_i,                   &
         & STACK_EXPORT_i, B_i, SOLVER_COMM, order_i, order_o)
    
  END SUBROUTINE distribute_matrix_vector

  
  SUBROUTINE make_indices(part, index3, index, index_nodes, index2, SWITCH_FLAG, N_i, &
       &                  SOLVER_COMM, PETOT)
    implicit none
    include 'precision.inc'
    include 'mpif.h'

    integer(kind=kint), intent(inout) :: index3(0:), part(:)
    integer(kind=kint), intent(in)    :: SWITCH_FLAG, N_i, SOLVER_COMM, PETOT

    integer(kind=kint), intent(out)   :: index(0:), index_nodes(:), index2(0:)

    integer(kind=kint) :: i, j, k, l, ierr
    integer(kind=kint), allocatable :: index4(:), part2(:)

    !C index3(i) means the sum of the number of the nodes in domains i-1..0.
    !C index2(i) means the sum of the number of the associated PE's domain's nodes 
    !C   on PEs i-1..0.
    !C index and index_nodes(:) represent that PE's node number by domains.


    !C-- SWITCH_FLAG : 0 => input data is PART(:) containing PE number
    !C
    !C-- SWITCH_FLAG : 1 => input data is PART(:) containing global number, and
    !C                             index3(:) containing index information for each PE

    !C-- counting node number of each domain, then execute allreduce and alltoall
    !C-- operation in index2 and index3
    
    index = 0
    if(SWITCH_FLAG == 0) then

       DO i = 1, N_i
          j = part(i) + 2
          index(j) = index(j) + 1
       END DO


       index3(0) = 0
       CALL MPI_ALLREDUCE( index(2), index3(1), PETOT, MPI_INTEGER, MPI_SUM,   &
            &  SOLVER_COMM, ierr )

       DO i = 1, PETOT
          index3(i) = index3(i - 1) + index3(i)
       END DO
       
    else if(SWITCH_FLAG == 1) then
       allocate(part2(N_i))
       DO i = 1, N_i 
          DO k = 1, PETOT 
             if(part(i) <= index3(k)) exit
          END DO
          if(k > PETOT) stop 'Error in make_indices()'

          part2(i) = k - 1
          index(k + 1) = index(k + 1) + 1
       END DO
    end if


    index2(0) = 0
    CALL MPI_ALLTOALL( index(2), 1, MPI_INTEGER, index2(1), 1, MPI_INTEGER, &
         &            SOLVER_COMM, ierr )

    DO i = 1, PETOT
       index2(i) = index2(i) + index2(i - 1)
    END DO

    !C-- collecting the nodes in each PE.
    DO i = 1, PETOT
       index(i) = index(i - 1) + index(i)
    END DO


    if(SWITCH_FLAG == 0) then
       DO i = 1, N_i
          j = part(i)  + 1
          k = index(j) + 1
          index(j) = k
          index_nodes(k) = i
       END DO

       allocate(index4(0:PETOT))
       !C index4(i) means the first global node number of the domain i on that PE.
       CALL MPI_ALLTOALL( index2(0), 1, MPI_INTEGER, index4(1), 1, MPI_INTEGER, &
            &             SOLVER_COMM, ierr)

       DO i = 1, PETOT
          index4(i) = index4(i) + index3(i - 1)
       END DO
       DO i = 1, N_i
          j = part(i) + 1
          k = index4(j) + 1
          index4(j) = k
          part(i) = k
       END DO

       !C--
       deallocate(index4)    

    else if(SWITCH_FLAG == 1) then
       DO i = 1, N_i
          j = part2(i) + 1
          k = index(j) + 1
          index(j) = k
          index_nodes(k) = i
       END DO
       deallocate(part2)
    end if

  END SUBROUTINE make_indices


  !C-- SWITCH_FLAG : 0 => input data is PART(:) containing PE number
  !C
  !C-- SWITCH_FLAG : 1 => input data is PART(:) containing global number, and
  !C                             index_number(:) containing index information for each PE

  !C NEIBPE, STACK_IMPORT(0:PETOT), STACK_EXPORT(0:PETOT) should be allocated beforehand.

  !C OUT: sz_export, sz_import, N, NP, NPL, NPU, NEIBPETOT, NEIBPE, STACK_IMPORT, STACK_EXPORT
  !C INOUT: part, index3
  !C IN: *_i, SOLVER_COMM, SWITCH_FLAG
  SUBROUTINE size_new_mat(sz_export, sz_import, N, NP, NPL, NPU, NEIBPETOT,        &
       & NEIBPE, STACK_IMPORT, STACK_EXPORT, part, index3, N_i, NP_i, D_i, INU_i,  &
       & INL_i, IAU_i, IAL_i, AU_i, AL_i, NEIBPETOT_i, NEIBPE_i, NOD_IMPORT_i,     &
       & NOD_EXPORT_i, STACK_IMPORT_i, STACK_EXPORT_i, B_i,                        &
       & SOLVER_COMM, order_i, SWITCH_FLAG)
    
    use solver_SR2
    use count_time_mod
    use redist_mat_mod
    use isort
    implicit none
    include 'mpif.h'
    

    integer(kind=kint), intent(out) :: sz_export, sz_import

    !C-- partition info
    integer(kind=kint) :: part(:)
    integer(kind=kint) :: index3(0:)
    integer(kind=kint) :: PETOT, my_rank

    integer(kind=kint ) :: order_i(:)
    integer(kind=kint ), intent(in) :: SWITCH_FLAG

    !C-- initial matrix form
    integer(kind=kint) :: N_i, NP_i, NEIBPETOT_i
    real   (kind=KREAL), pointer :: D_i(:),B_i(:)
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:)
    integer(kind=kint ), pointer :: IAU_i(:), IAL_i(:)
    real   (kind=KREAL), pointer :: AU_i(:)
    real   (kind=KREAL), pointer :: AL_i(:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_i(:)
    integer(kind=kint ), pointer :: NEIBPE_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_i(:)
    
    !C-- redistributed matrix form (NPL, NPU isn't refered in other sub.)
    integer(kind=kint) :: N, NP, NEIBPETOT, NPL, NPU
    real   (kind=KREAL), pointer :: D(:), B(:)
    integer(kind=kint ), pointer :: INU(:), INL(:)
    integer(kind=kint ), pointer :: IAU(:), IAL(:)
    real   (kind=KREAL), pointer :: AU(:)
    real   (kind=KREAL), pointer :: AL(:)
    integer(kind=kint ), pointer :: NOD_IMPORT(:)
    integer(kind=kint ), pointer :: NOD_EXPORT(:)
    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:)

    integer(kind=kint ) :: i,j,k,l,m,m2,buf_ind,buf_ind_dbl 
    integer(kind=kint ) :: alloc_sz,alloc_sz_dbl,nonzeros
    integer(kind=kint ), allocatable :: sendbuf(:)
    REAL   (kind=kreal), allocatable :: sendbuf_dbl(:)
    
    integer(kind=kint ), allocatable :: sd_buf_sz(:)

    
    !C-- status, and request arrays
    integer(kind=kint ), allocatable :: sta1(:,:), sta11(:,:), sta2(:,:), sta22(:,:)
    integer(kind=kint ), allocatable :: req1(:),req11(:),req2(:),req22(:)              
    integer(kind=kint ) :: ierr, req_cnt1, req_cnt2, req_cnt11,req_cnt22
    integer, intent(in) :: SOLVER_COMM


    !C-- 
    integer(kind=kint ), allocatable :: num_of_ex_nds_for_PE(:)
    integer(kind=kint ), allocatable :: ex_nds_buf(:)
    integer(kind=kint ) :: ex_nds_buf_sz, ex_nds_cnt

    !C altix Qsort
    integer(8) :: qs1,qs2
!!$    integer :: qs1,qs2
    
    integer(kind=kint ), pointer :: ROW_NUM_ptr(:)
    integer(kind=kint ), pointer :: IN_ptr(:), INL_ptr(:), INU_ptr(:)
    integer(kind=kint ), pointer :: IA_ptr(:), IAL_ptr(:), IAU_ptr(:)
    real   (kind=kreal), pointer :: A_ptr(:),  AL_ptr(:), AU_ptr(:)
    integer(kind=kint )          :: N_s, p, m1, L_index, U_index

    !C--
    integer(kind=kint ) :: S, R
    integer(kind=kint ), allocatable :: WS(:), WR(:)

    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 
    
    CALL count_time(1, solver_comm, my_rank, 15)
    
    allocate(index(0:PETOT + 1), index2(0:PETOT), index_nodes(N_i)) 
    CALL make_indices(part, index3, index, index_nodes, index2, SWITCH_FLAG, &
         & N_i, SOLVER_COMM, PETOT)
    
    allocate(sta1(MPI_STATUS_SIZE, PETOT-1), req1(PETOT-1), sd_buf_sz(-3:2 * (PETOT - 1)), &
         & req11(PETOT - 1), sta11(MPI_STATUS_SIZE, PETOT - 1), recv_buf_sz(-1:2 * (PETOT - 1)))
    allocate(req2(PETOT - 1), req22(PETOT - 1), sta2(MPI_STATUS_SIZE, PETOT - 1),     &
         &   sta22(MPI_STATUS_SIZE, PETOT - 1))



    
    !C-- exchange the part data
    if(NEIBPETOT_i > 0) then
       S = STACK_EXPORT_i(NEIBPETOT_i)
       R = STACK_IMPORT_i(NEIBPETOT_i)
       allocate(WS(S), WR(R))
       call  SOLVER_SEND_RECV2I                                                   &
            &  (NP_i, S, R, NEIBPETOT_i, NEIBPE_i, STACK_IMPORT_i, NOD_IMPORT_i,  &
            &                          STACK_EXPORT_i, NOD_EXPORT_i,              &
            &    WS, WR, part, SOLVER_COMM, my_rank)
       deallocate(WS, WR)
    end if

    


    !C-- counting Send Buffer and send the size
    alloc_sz = 0; alloc_sz_dbl = 0

    sd_buf_sz(my_rank * 2 - 1) = 0
    sd_buf_sz(my_rank * 2) = 0
          
    DO i = 1, PETOT
       if( my_rank == i - 1 ) cycle
          
       nonzeros = 0
       DO k = index(i - 1) + 1, index(i)
          
          l = index_nodes(k)
          
          nonzeros = nonzeros + INL_i(l) - INL_i(l - 1) + &
               &     INU_i(l) - INU_i(l - 1)
       END DO
       j = index(i) - index(i - 1)
          
       !C-- number of nonzeros is j+nonzeros
       !    in diagonal part and L and U part
       !    ordering information of j is added.
       if(j == 0) then
          sd_buf_sz((i - 1) * 2 - 1) = 0
          sd_buf_sz((i - 1) * 2) = 0
       else
          sd_buf_sz((i - 1) * 2 - 1) = 1 + 2 * (j + 1) + nonzeros + 2 * j
          sd_buf_sz((i - 1) * 2)   = nonzeros + 2 * j
          alloc_sz     = alloc_sz    + sd_buf_sz((i - 1) * 2 - 1)
          alloc_sz_dbl = alloc_sz_dbl + sd_buf_sz((i - 1) * 2)
       end if
       !C-- send the size to each PE
    END DO


    !C--
    CALL MPI_ALLTOALL(sd_buf_sz(-1), 2, MPI_INTEGER, recv_buf_sz(-1), 2, MPI_INTEGER, &
         &             SOLVER_COMM, ierr)
    
    allocate(sendbuf(alloc_sz))
    allocate(sendbuf_dbl(alloc_sz_dbl))

    
    !C-- recv integer part
    alloc_sz = 0; alloc_sz_dbl = 0
    DO i = 1, PETOT
       if( my_rank == i - 1 ) cycle
       
       alloc_sz     = alloc_sz     + recv_buf_sz((i - 1) * 2 - 1)
       alloc_sz_dbl = alloc_sz_dbl + recv_buf_sz((i - 1) * 2)
    END DO

    allocate( recvbuf( alloc_sz ), recvbuf_dbl( alloc_sz_dbl ) )
    
    buf_ind = 1; req_cnt2 = 0
    DO i = 1, PETOT
       if (my_rank == i - 1 .or. recv_buf_sz((i - 1) * 2 - 1) == 0) cycle
       req_cnt2 = req_cnt2 + 1
       j = recv_buf_sz((i - 1) * 2 - 1)

       CALL MPI_IRECV(recvbuf(buf_ind), j, MPI_INTEGER, i - 1, 0, &
            & SOLVER_COMM, req2(req_cnt2), ierr )
       buf_ind = buf_ind + j
    END DO
    !C-- recv integer part
    

    !C-- creat buffer and send integer part
    !C indexing the array of sd_buf_sz.
    sd_buf_sz(-3) = 0; sd_buf_sz(-2) = 0
    DO i = 0, PETOT - 1
       if(my_rank == i) then
          sd_buf_sz(i * 2)     = sd_buf_sz((i - 1) * 2)
          sd_buf_sz(i * 2 - 1) = sd_buf_sz((i - 1) * 2 - 1)
       else
          sd_buf_sz(i * 2)     = sd_buf_sz(i * 2)     + sd_buf_sz((i - 1) * 2)
          sd_buf_sz(i * 2 - 1) = sd_buf_sz(i * 2 - 1) + sd_buf_sz((i - 1) * 2 - 1)
       end if
    END DO


    
    buf_ind = 0; buf_ind_dbl = 0; req_cnt1 = 0
    DO i = 1, PETOT
       j = sd_buf_sz((i - 1) * 2 - 1) - sd_buf_sz((i - 2) * 2 - 1)
       if( my_rank == i - 1 .or. j == 0 ) cycle
       
       j = index(i) - index(i - 1)
       
       buf_ind = buf_ind + 1
       sendbuf(buf_ind) = j
       
       buf_ind = buf_ind + 1
       sendbuf(buf_ind) = 0
       sendbuf(buf_ind + j + 1) = 0
       DO k = index(i - 1) + 1, index(i)
          m = index_nodes(k)
          buf_ind = buf_ind + 1
          sendbuf(buf_ind) = sendbuf(buf_ind - 1) + INL_i(m) - INL_i(m - 1)
          sendbuf(buf_ind + j + 1) = sendbuf(buf_ind + j) + INU_i(m) - INU_i(m - 1)

          buf_ind_dbl = buf_ind_dbl + 1
          sendbuf_dbl(buf_ind_dbl) = D_i(m)
          sendbuf_dbl(buf_ind_dbl + j) = B_i(m)

       END DO

       buf_ind = buf_ind + j + 1
       buf_ind_dbl = buf_ind_dbl+j
       
       DO k = index(i - 1) + 1, index(i)
          l = index_nodes(k)
          DO m = INL_i(l - 1) + 1,INL_i(l)
             buf_ind = buf_ind + 1
             sendbuf(buf_ind) = part(IAL_i(m))
             
             buf_ind_dbl = buf_ind_dbl + 1
             sendbuf_dbl(buf_ind_dbl) = AL_i(m)
          END DO
       END DO
       DO k = index(i - 1) + 1, index(i)
          l = index_nodes(k)
          DO m = INU_i(l - 1) + 1,INU_i(l)
             buf_ind = buf_ind + 1
             sendbuf(buf_ind) = part(IAU_i(m))
             
             buf_ind_dbl = buf_ind_dbl + 1
             sendbuf_dbl(buf_ind_dbl) = AU_i(m)
          END DO
       END DO
       
       DO k = index(i - 1) + 1, index(i)
          l = index_nodes(k)
          buf_ind = buf_ind + 1
          sendbuf(buf_ind) = order_i(l)
          sendbuf(buf_ind + j) = part(l)
       END DO
       buf_ind = buf_ind + j 

       !C-- SEND the contents
       req_cnt1 = req_cnt1 + 1
       l = sd_buf_sz((i - 1) * 2 - 1) - sd_buf_sz((i - 2) * 2 - 1) 
       k = sd_buf_sz((i - 2) * 2 - 1) + 1
       CALL MPI_ISEND(sendbuf(k), l, MPI_INTEGER, i - 1, 0, SOLVER_COMM,   &
            &         req1(req_cnt1), ierr )
       
    END DO
    !C-- creat buffer and send integer part





    !C-- sendrecv double part
    buf_ind_dbl = 1; req_cnt22 = 0
    DO i = 1, PETOT
       if (recv_buf_sz((i - 1) * 2 - 1) > 0) then
          req_cnt22 = req_cnt22 + 1
          j = recv_buf_sz((i - 1) * 2) 
          CALL MPI_IRECV(recvbuf_dbl(buf_ind_dbl), j, MPI_DOUBLE_PRECISION, &
               &         i - 1, 0, SOLVER_COMM, req22(req_cnt22), ierr )
          buf_ind_dbl = buf_ind_dbl + j
       end if
    END DO


    req_cnt11 = 0
    DO i = 1, PETOT
       j = sd_buf_sz((i - 1) * 2) - sd_buf_sz((i - 2) * 2)
       if(j > 0) then
          
          req_cnt11 = req_cnt11 + 1
          k = sd_buf_sz((i - 2) * 2) + 1
          
          CALL MPI_ISEND(sendbuf_dbl(k), j, MPI_DOUBLE_PRECISION, i - 1, 0, &
               &         SOLVER_COMM, req11(req_cnt11), ierr )
       end if
    END DO
    !C-- sendrecv double part



    CALL count_time(2, solver_comm, my_rank, 15)

    !C-- waitall integer part
    if(req_cnt2>0) CALL MPI_WAITALL( req_cnt2, req2,  sta2,  ierr )
    if(req_cnt1>0) CALL MPI_WAITALL( req_cnt1, req1,  sta1,  ierr )       
    !C-- waitall integer part



    CALL count_time(1, solver_comm, my_rank, 15)    

    deallocate(sendbuf)



    allocate(split_mats(PETOT))    

    !C-- pointer allocation. index of buf_IN? arrays start from 1.
    !C-- NPL and NPU have the number of nonzeros in matrices
    buf_ind = 0

    NPL = 0; NPU = 0 
    DO i = index(my_rank) + 1, index(my_rank + 1) 
       j = index_nodes(i)
       NPL = NPL + INL_i(j) - INL_i(j - 1)
       NPU = NPU + INU_i(j) - INU_i(j - 1) 
    END DO

    DO i = 1, PETOT
       N_s = index2(i) - index2(i - 1)
       split_mats(i) % row_sz = N_s
       
       if( my_rank == i - 1 .or. N_s == 0 ) cycle
       split_mats(i) % INL_s => recvbuf(buf_ind + 2:buf_ind + 2 + N_s)
       split_mats(i) % INU_s => recvbuf(3 + N_s + buf_ind:buf_ind + 3 + 2 * N_s)
       
       buf_NPL = split_mats(i) % INL_s(N_s+1)
       NPL = NPL + buf_NPL
       buf_NPU = split_mats(i) % INU_s(N_s+1)
       NPU = NPU + buf_NPU
       split_mats(i) % IAL_s => recvbuf(buf_ind + 4 + 2 * N_s :                          &
            &                           buf_ind + 3 + 2 * N_s + buf_NPL)
       split_mats(i) % IAU_s => recvbuf(buf_ind + 4 + 2 * N_s + buf_NPL :                &
            &                           buf_ind + 3 + 2 * N_s + buf_NPL + buf_NPU)
       !C ordering information is added
       split_mats(i) % row_num_s => recvbuf(buf_ind + 4 + 3 * N_s + buf_NPL + buf_NPU :  &
            &                               buf_ind + 3 + 4 * N_s + buf_NPL + buf_NPU)

       
       buf_ind     = buf_ind     + recv_buf_sz((i - 1) * 2 - 1)
       
    END DO


    !C-- construction of the distributed matrix for the solver begins
    !C-- 1 pass: count the external aggregate and fix the sizes of arrays
    !C-- buf_IN? start from 1 unlike other index array.

    N = index3( my_rank + 1 ) - index3( my_rank )
    allocate( num_of_ex_nds_for_PE( PETOT ) )
    allocate( chk_rows_mv_LtoU( N ) )
    ex_nds_buf_sz = NPL + NPU
    allocate( ex_nds_buf( ex_nds_buf_sz ) )
    
    ex_nds_cnt = 0
    chk_rows_mv_LtoU = 0
    num_of_ex_nds_in_L = 0
    num_of_ex_nds_for_PE = 0

    nd_range_l = index3(my_rank)
    nd_range_u = index3(my_rank + 1)


    !C count up sub matrix on self PE
    DO l = index(my_rank) + 1, index(my_rank + 1)
       m1 = index_nodes(l)
       k = part(m1) - nd_range_l
       DO i = INL_i(m1 - 1) + 1, INL_i(m1)
          j = part(IAL_i(i))
          call check_mat_L(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
               &           j, k, ex_nds_cnt, ex_nds_buf)
       END DO

       DO i = INU_i(m1 - 1) + 1, INU_i(m1) 
          j = part(IAU_i(i))
          call check_mat_U(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
               &           j, k, ex_nds_cnt, ex_nds_buf)
       END DO
    END DO



    !C count up sub matrices received from other PEs
    DO m = 1, PETOT
       if(my_rank == m - 1) cycle

       N_s    =  split_mats(m) % row_sz
       
       if(N_s>0) then
          ROW_NUM_PTR => split_mats(m) % row_num_s

          INL_PTR      => split_mats(m) % INL_s
          IAL_PTR      => split_mats(m) % IAL_s
          INU_PTR      => split_mats(m) % INU_s
          IAU_PTR      => split_mats(m) % IAU_s
       end if


       DO l = 1, N_s
          k = ROW_NUM_PTR(l) - nd_range_l

          DO i = INL_PTR(l) + 1, INL_PTR(l + 1)
             j = IAL_PTR(i)
             call check_mat_L(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
                  &           j, k, ex_nds_cnt, ex_nds_buf)
          END DO
          
          DO i = INU_PTR(l) + 1, INU_PTR(l + 1) 
             j = IAU_PTR(i)
             call check_mat_U(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
                  &           j, k, ex_nds_cnt, ex_nds_buf)
          END DO

       END DO
    END DO




    if(ex_nds_cnt > 0) then
       qs1 = ex_nds_cnt 
       qs2 = kint
       !C intel intrinsic subroutine
!!$       CALL QSORT(ex_nds_buf, qs1,  qs2, compare)
       call QSORTI_2(ex_nds_cnt, ex_nds_buf)
    end if



    !C-- redundant values are deleted from the array ex_nds_buf
    !C-- count up PE numbers in NEIBPETOT
    !C-- count up number of nodes for each PE
    k = 1;  NEIBPETOT = 0
    if( ex_nds_cnt > 0 ) then
       j = ex_nds_buf(1)
       NEIBPETOT = NEIBPETOT + 1
       
       !C- initial case is considered
       DO l = 1, PETOT
          if( j <= index3(l) ) exit
       END DO
       if( l > PETOT ) stop "unexpected error in REDIST_mtom():1"
       num_of_ex_nds_for_PE(l) = num_of_ex_nds_for_PE(l) + 1
       m = l
       
       !C- redundancy is removed and counting for each PE
       DO i = 2, ex_nds_cnt
          if( j /= ex_nds_buf(i) ) then
             j = ex_nds_buf(i)
             k = k + 1
             ex_nds_buf(k) = j

             !C counting for each PE
             DO l = m, PETOT
                if( j <= index3(l) ) exit
             END DO
             if( l > PETOT ) then
                stop "unexpected error in REDIST_mtom():2"
             end if
             
             if( m /= l ) NEIBPETOT = NEIBPETOT + 1
             m = l
             
             num_of_ex_nds_for_PE(l) = num_of_ex_nds_for_PE(l) + 1
          end if
       END DO
       ex_nds_cnt = k
    end if


    
    !C- NP can be determined only by the number of own external nodes
    NP = N + ex_nds_cnt
    
    !C--NEIBPETOT may be increased and this chages the required size of the
    !C--arrays of STACK_IMPORT and STACK_EXPORT
    allocate(NEIBPE(PETOT), STACK_IMPORT(0:PETOT),              &
         &   STACK_EXPORT(0:PETOT), com_send_buf(ex_nds_cnt))
    
    !C- create the comunication table and send buffer for export table
    !C- containing global numbers.
    DO i = 1, ex_nds_cnt
       com_send_buf(i) = ex_nds_buf(i)
    END DO
    
    deallocate( ex_nds_buf )       
    
    !C-- import table is created
    !C-- external number is attached to STACK_IMPORT in order
    STACK_IMPORT(0) = 0; j = 0
    DO i = 1, PETOT
       if( num_of_ex_nds_for_PE(i) > 0 ) then
          j = j + 1
          NEIBPE(j) = i - 1
          STACK_IMPORT(j) = STACK_IMPORT(j - 1) + num_of_ex_nds_for_PE(i)
       end if
    END DO
    deallocate(num_of_ex_nds_for_PE)


    CALL count_time(2, solver_comm, my_rank, 15)    


    !C------------- late wait call
    !C-- waitall double part
    CALL MPI_WAITALL( req_cnt11, req11, sta11, ierr )
    CALL MPI_WAITALL( req_cnt22, req22, sta22, ierr )
    !C-- waitall double part


    CALL count_time(1, solver_comm, my_rank, 15)    
    
    deallocate(sendbuf_dbl, sd_buf_sz)
    deallocate(sta1, req1, sta11, req11)
    deallocate(sta2, sta22, req2, req22)

    !C-- pointer allocation. index of buf_IN? arrays start from 1.
    !C-- For double part.
    buf_ind_dbl = 0
    DO i = 1, PETOT
       N_s = index2(i) - index2(i - 1)
       if( my_rank == i - 1 .or. N_s == 0 ) cycle
       buf_NPL = split_mats(i) % INL_s(N_s+1)
       buf_NPU = split_mats(i) % INU_s(N_s+1)
       
       split_mats(i) % AL_s => recvbuf_dbl(1 + 2 * N_s + buf_ind_dbl :                   &
            &                              buf_NPL + 2 * N_s + buf_ind_dbl)
       split_mats(i) % AU_s => recvbuf_dbl(1 + 2 * N_s + buf_NPL + buf_ind_dbl :         &
            &                              buf_NPL + buf_NPU + 2 * N_s +buf_ind_dbl)
       buf_ind_dbl = buf_ind_dbl + recv_buf_sz((i - 1) * 2)
    END DO
    !C------------ late wait call


    
    !C----- Export table creation composed of two subroutines.
    !C      These subroutines commonly use  in_recv_buf(0:), pes(:), n_pes, recv_buf_size
    allocate(in_recv_buf(0:PETOT), pes(PETOT))
    call size_export(SOLVER_COMM, PETOT, NEIBPETOT, NEIBPE, STACK_IMPORT,  &
         &  in_recv_buf, pes, n_pes, recv_buf_size)


    buf_NPL = NPL
    NPL = NPL - num_of_ex_nds_in_L
    NPU = NPU + num_of_ex_nds_in_L
    sz_export = recv_buf_size
    sz_import = ex_nds_cnt
    CALL count_time(2, solver_comm, my_rank, 15)    

  end SUBROUTINE size_new_mat

  !C OUT: sz_export, sz_import, N, NP, NPL, NPU, NEIBPETOT, NEIBPE, STACK_IMPORT, STACK_EXPORT
  !C INOUT: part, index3
  !C IN: *_i, SOLVER_COMM, SWITCH_FLAG
  SUBROUTINE size_new_mat_vec(sz_export, sz_import, N, NP, NB, NPL, NPU, NEIBPETOT,        &
       & NEIBPE, STACK_IMPORT, STACK_EXPORT, part, index3, N_i, NP_i, NB_i, D_i, INU_i,&
       & INL_i, IAU_i, IAL_i, AU_i, AL_i, NEIBPETOT_i, NEIBPE_i, NOD_IMPORT_i,     &
       & NOD_EXPORT_i, STACK_IMPORT_i, STACK_EXPORT_i, B_i,                        &
       & SOLVER_COMM, order_i, SWITCH_FLAG)
    
    use solver_SR2
    use count_time_mod
    use redist_mat_mod
    use isort
    implicit none
    include 'mpif.h'
    

    integer(kind=kint), intent(out) :: sz_export, sz_import

    !C-- partition info
    integer(kind=kint) :: part(:)
    integer(kind=kint) :: index3(0:)
    integer(kind=kint) :: PETOT, my_rank

    integer(kind=kint ) :: order_i(:)
    integer(kind=kint ), intent(in) :: SWITCH_FLAG

    !C-- initial matrix form
    integer(kind=kint) :: N_i, NB_i, NP_i, NEIBPETOT_i
    real   (kind=KREAL), pointer :: D_i(:,:,:),B_i(:)
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:)
    integer(kind=kint ), pointer :: IAU_i(:), IAL_i(:)
    real   (kind=KREAL), pointer :: AU_i(:,:,:)
    real   (kind=KREAL), pointer :: AL_i(:,:,:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_i(:)
    integer(kind=kint ), pointer :: NEIBPE_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_i(:)
    
    !C-- redistributed matrix form (NPL, NPU isn't refered in other sub.)
    integer(kind=kint) :: N, NP, NB, NEIBPETOT, NPL, NPU
    real   (kind=KREAL), pointer :: D(:,:,:), B(:)
    integer(kind=kint ), pointer :: INU(:), INL(:)
    integer(kind=kint ), pointer :: IAU(:), IAL(:)
    real   (kind=KREAL), pointer :: AU(:,:,:)
    real   (kind=KREAL), pointer :: AL(:,:,:)
    integer(kind=kint ), pointer :: NOD_IMPORT(:)
    integer(kind=kint ), pointer :: NOD_EXPORT(:)
    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:)

    integer(kind=kint ) :: i,j,k,k1,k2,l,m,m2,buf_ind,buf_ind_dbl 
    integer(kind=kint ) :: alloc_sz,alloc_sz_dbl,nonzeros
    integer(kind=kint ), allocatable :: sendbuf(:)
    REAL   (kind=kreal), allocatable :: sendbuf_dbl(:)
    
    integer(kind=kint ), allocatable :: sd_buf_sz(:)

    
    !C-- status, and request arrays
    integer(kind=kint ), allocatable :: sta1(:,:), sta11(:,:), sta2(:,:), sta22(:,:)
    integer(kind=kint ), allocatable :: req1(:),req11(:),req2(:),req22(:)              
    integer(kind=kint ) :: ierr, req_cnt1, req_cnt2, req_cnt11,req_cnt22
    integer, intent(in) :: SOLVER_COMM


    !C-- 
    integer(kind=kint ), allocatable :: num_of_ex_nds_for_PE(:)
    integer(kind=kint ), allocatable :: ex_nds_buf(:)
    integer(kind=kint ) :: ex_nds_buf_sz, ex_nds_cnt

    !C altix Qsort
    integer(8) :: qs1,qs2
!!$    integer :: qs1,qs2
    
    integer(kind=kint ), pointer :: ROW_NUM_ptr(:)
    integer(kind=kint ), pointer :: IN_ptr(:), INL_ptr(:), INU_ptr(:)
    integer(kind=kint ), pointer :: IA_ptr(:), IAL_ptr(:), IAU_ptr(:)
    real   (kind=kreal), pointer :: A_ptr(:),  AL_ptr(:), AU_ptr(:)
    integer(kind=kint )          :: N_s, p, m1, L_index, U_index

    !C--
    integer(kind=kint ) :: S, R
    integer(kind=kint ), allocatable :: WS(:), WR(:)

    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 
    
    CALL count_time(1, solver_comm, my_rank, 15)
    
    allocate(index(0:PETOT + 1), index2(0:PETOT), index_nodes(N_i)) 
    CALL make_indices(part, index3, index, index_nodes, index2, SWITCH_FLAG, &
         & N_i, SOLVER_COMM, PETOT)
    
    allocate(sta1(MPI_STATUS_SIZE, PETOT-1), req1(PETOT-1), sd_buf_sz(-3:2 * (PETOT - 1)), &
         & req11(PETOT - 1), sta11(MPI_STATUS_SIZE, PETOT - 1), recv_buf_sz(-1:2 * (PETOT - 1)))
    allocate(req2(PETOT - 1), req22(PETOT - 1), sta2(MPI_STATUS_SIZE, PETOT - 1),     &
         &   sta22(MPI_STATUS_SIZE, PETOT - 1))



    
    !C-- exchange the part data
    if(NEIBPETOT_i > 0) then
       S = STACK_EXPORT_i(NEIBPETOT_i)
       R = STACK_IMPORT_i(NEIBPETOT_i)
       allocate(WS(S), WR(R))
       call  SOLVER_SEND_RECV2I                                                   &
            &  (NP_i, S, R, NEIBPETOT_i, NEIBPE_i, STACK_IMPORT_i, NOD_IMPORT_i,  &
            &                          STACK_EXPORT_i, NOD_EXPORT_i,              &
            &    WS, WR, part, SOLVER_COMM, my_rank)
       deallocate(WS, WR)
    end if

    


    !C-- counting Send Buffer and send the size
    alloc_sz = 0; alloc_sz_dbl = 0

    sd_buf_sz(my_rank * 2 - 1) = 0
    sd_buf_sz(my_rank * 2) = 0
          
    DO i = 1, PETOT
       if( my_rank == i - 1 ) cycle
          
       nonzeros = 0
       DO k = index(i - 1) + 1, index(i)
          
          l = index_nodes(k)
          
          nonzeros = nonzeros + INL_i(l) - INL_i(l - 1) + &
               &     INU_i(l) - INU_i(l - 1)
       END DO
       j = index(i) - index(i - 1)
          
       !C-- number of nonzeros is j+nonzeros
       !    in diagonal part and L and U part
       !    ordering information of j is added.
       if(j == 0) then
          sd_buf_sz((i - 1) * 2 - 1) = 0
          sd_buf_sz((i - 1) * 2) = 0
       else
          sd_buf_sz((i - 1) * 2 - 1) = 1 + 2 * (j + 1) + nonzeros + 2 * j
          sd_buf_sz((i - 1) * 2)   = NB_i*NB_i*nonzeros + (NB_i+1)*NB_i*j
          alloc_sz     = alloc_sz    + sd_buf_sz((i - 1) * 2 - 1)
          alloc_sz_dbl = alloc_sz_dbl + sd_buf_sz((i - 1) * 2)
       end if
       !C-- send the size to each PE
    END DO


    !C--
    CALL MPI_ALLTOALL(sd_buf_sz(-1), 2, MPI_INTEGER, recv_buf_sz(-1), 2, MPI_INTEGER, &
         &             SOLVER_COMM, ierr)
    
    allocate(sendbuf(alloc_sz))
    allocate(sendbuf_dbl(alloc_sz_dbl))

    
    !C-- recv integer part
    alloc_sz = 0; alloc_sz_dbl = 0
    DO i = 1, PETOT
       if( my_rank == i - 1 ) cycle
       
       alloc_sz     = alloc_sz     + recv_buf_sz((i - 1) * 2 - 1)
       alloc_sz_dbl = alloc_sz_dbl + recv_buf_sz((i - 1) * 2)
    END DO

    allocate( recvbuf( alloc_sz ), recvbuf_dbl( alloc_sz_dbl ) )
    
    buf_ind = 1; req_cnt2 = 0
    DO i = 1, PETOT
       if (my_rank == i - 1 .or. recv_buf_sz((i - 1) * 2 - 1) == 0) cycle
       req_cnt2 = req_cnt2 + 1
       j = recv_buf_sz((i - 1) * 2 - 1)

       CALL MPI_IRECV(recvbuf(buf_ind), j, MPI_INTEGER, i - 1, 0, &
            & SOLVER_COMM, req2(req_cnt2), ierr )
       buf_ind = buf_ind + j
    END DO
    !C-- recv integer part
    

    !C-- creat buffer and send integer part
    !C indexing the array of sd_buf_sz.
    sd_buf_sz(-3) = 0; sd_buf_sz(-2) = 0
    DO i = 0, PETOT - 1
       if(my_rank == i) then
          sd_buf_sz(i * 2)     = sd_buf_sz((i - 1) * 2)
          sd_buf_sz(i * 2 - 1) = sd_buf_sz((i - 1) * 2 - 1)
       else
          sd_buf_sz(i * 2)     = sd_buf_sz(i * 2)     + sd_buf_sz((i - 1) * 2)
          sd_buf_sz(i * 2 - 1) = sd_buf_sz(i * 2 - 1) + sd_buf_sz((i - 1) * 2 - 1)
       end if
    END DO


    
    buf_ind = 0; buf_ind_dbl = 0; req_cnt1 = 0
    DO i = 1, PETOT
       j = sd_buf_sz((i - 1) * 2 - 1) - sd_buf_sz((i - 2) * 2 - 1)
       if( my_rank == i - 1 .or. j == 0 ) cycle
       
       j = index(i) - index(i - 1)
       
       buf_ind = buf_ind + 1
       sendbuf(buf_ind) = j
       
       buf_ind = buf_ind + 1
       sendbuf(buf_ind) = 0
       sendbuf(buf_ind + j + 1) = 0
       DO k = index(i - 1) + 1, index(i)
          m = index_nodes(k)
          buf_ind = buf_ind + 1
          sendbuf(buf_ind) = sendbuf(buf_ind - 1) + INL_i(m) - INL_i(m - 1)
          sendbuf(buf_ind + j + 1) = sendbuf(buf_ind + j) + INU_i(m) - INU_i(m - 1)
       END DO

       DO k = index(i - 1) + 1, index(i)
          m = index_nodes(k)
          DO k1=1,NB_i
             DO k2=1,NB_i
                buf_ind_dbl = buf_ind_dbl + 1
                sendbuf_dbl(buf_ind_dbl) = D_i(k1,k2,m)
             END DO
          END DO
       END DO

       DO k = index(i - 1) + 1, index(i)
          m = index_nodes(k)
          DO k1=1,NB_i
             buf_ind_dbl = buf_ind_dbl + 1
             sendbuf_dbl(buf_ind_dbl) = B_i((m-1)*NB_i+k1)
          END DO
       END DO

       buf_ind = buf_ind + j + 1
       
       DO k = index(i - 1) + 1, index(i)
          l = index_nodes(k)
          DO m = INL_i(l - 1) + 1,INL_i(l)
             buf_ind = buf_ind + 1
             sendbuf(buf_ind) = part(IAL_i(m))
             DO k1=1,NB_i
                DO k2=1,NB_i
                   sendbuf_dbl(buf_ind_dbl+(k1-1)*NB_i+k2) = AL_i(k1,k2,m)
                END DO
             END DO
             buf_ind_dbl = buf_ind_dbl + NB_i*NB_i
          END DO
       END DO

       DO k = index(i - 1) + 1, index(i)
          l = index_nodes(k)
          DO m = INU_i(l - 1) + 1,INU_i(l)
             buf_ind = buf_ind + 1
             sendbuf(buf_ind) = part(IAU_i(m))
             
             DO k1=1,NB_i
                DO k2=1,NB_i
                   sendbuf_dbl(buf_ind_dbl+(k1-1)*NB_i+k2) = AU_i(k1,k2,m)
                END DO
             END DO
             buf_ind_dbl = buf_ind_dbl + NB_i*NB_i

          END DO
       END DO
       
       DO k = index(i - 1) + 1, index(i)
          l = index_nodes(k)
          buf_ind = buf_ind + 1
          sendbuf(buf_ind) = order_i(l)
          sendbuf(buf_ind + j) = part(l)
       END DO
       buf_ind = buf_ind + j 

       !C-- SEND the contents
       req_cnt1 = req_cnt1 + 1
       l = sd_buf_sz((i - 1) * 2 - 1) - sd_buf_sz((i - 2) * 2 - 1) 
       k = sd_buf_sz((i - 2) * 2 - 1) + 1
       CALL MPI_ISEND(sendbuf(k), l, MPI_INTEGER, i - 1, 0, SOLVER_COMM,   &
            &         req1(req_cnt1), ierr )
       
    END DO
    !C-- creat buffer and send integer part





    !C-- sendrecv double part
    buf_ind_dbl = 1; req_cnt22 = 0
    DO i = 1, PETOT
       if (recv_buf_sz((i - 1) * 2 - 1) > 0) then
          req_cnt22 = req_cnt22 + 1
          j = recv_buf_sz((i - 1) * 2) 
          CALL MPI_IRECV(recvbuf_dbl(buf_ind_dbl), j, MPI_DOUBLE_PRECISION, &
               &         i - 1, 0, SOLVER_COMM, req22(req_cnt22), ierr )
          buf_ind_dbl = buf_ind_dbl + j
       end if
    END DO


    req_cnt11 = 0
    DO i = 1, PETOT
       j = sd_buf_sz((i - 1) * 2) - sd_buf_sz((i - 2) * 2)
       if(j > 0) then
          
          req_cnt11 = req_cnt11 + 1
          k = sd_buf_sz((i - 2) * 2) + 1
          
          CALL MPI_ISEND(sendbuf_dbl(k), j, MPI_DOUBLE_PRECISION, i - 1, 0, &
               &         SOLVER_COMM, req11(req_cnt11), ierr )
       end if
    END DO
    !C-- sendrecv double part



    CALL count_time(2, solver_comm, my_rank, 15)

    !C-- waitall integer part
    if(req_cnt2>0) CALL MPI_WAITALL( req_cnt2, req2,  sta2,  ierr )
    if(req_cnt1>0) CALL MPI_WAITALL( req_cnt1, req1,  sta1,  ierr )       
    !C-- waitall integer part



    CALL count_time(1, solver_comm, my_rank, 15)    

    deallocate(sendbuf)



    allocate(split_mats(PETOT))    

    !C-- pointer allocation. index of buf_IN? arrays start from 1.
    !C-- NPL and NPU have the number of nonzeros in matrices
    buf_ind = 0

    NPL = 0; NPU = 0 
    DO i = index(my_rank) + 1, index(my_rank + 1) 
       j = index_nodes(i)
       NPL = NPL + INL_i(j) - INL_i(j - 1)
       NPU = NPU + INU_i(j) - INU_i(j - 1) 
    END DO

    DO i = 1, PETOT
       N_s = index2(i) - index2(i - 1)
       split_mats(i) % row_sz = N_s
       
       if( my_rank == i - 1 .or. N_s == 0 ) cycle
       split_mats(i) % INL_s => recvbuf(buf_ind + 2:buf_ind + 2 + N_s)
       split_mats(i) % INU_s => recvbuf(3 + N_s + buf_ind:buf_ind + 3 + 2 * N_s)
       
       buf_NPL = split_mats(i) % INL_s(N_s+1)
       NPL = NPL + buf_NPL
       buf_NPU = split_mats(i) % INU_s(N_s+1)
       NPU = NPU + buf_NPU
       split_mats(i) % IAL_s => recvbuf(buf_ind + 4 + 2 * N_s :                          &
            &                           buf_ind + 3 + 2 * N_s + buf_NPL)
       split_mats(i) % IAU_s => recvbuf(buf_ind + 4 + 2 * N_s + buf_NPL :                &
            &                           buf_ind + 3 + 2 * N_s + buf_NPL + buf_NPU)
       !C ordering information is added
       split_mats(i) % row_num_s => recvbuf(buf_ind + 4 + 3 * N_s + buf_NPL + buf_NPU :  &
            &                               buf_ind + 3 + 4 * N_s + buf_NPL + buf_NPU)

       
       buf_ind     = buf_ind     + recv_buf_sz((i - 1) * 2 - 1)
       
    END DO


    !C-- construction of the distributed matrix for the solver begins
    !C-- 1 pass: count the external aggregate and fix the sizes of arrays
    !C-- buf_IN? start from 1 unlike other index array.

    N = index3( my_rank + 1 ) - index3( my_rank )
    allocate( num_of_ex_nds_for_PE( PETOT ) )
    allocate( chk_rows_mv_LtoU( N ) )
    ex_nds_buf_sz = NPL + NPU
    allocate( ex_nds_buf( ex_nds_buf_sz ) )
    
    ex_nds_cnt = 0
    chk_rows_mv_LtoU = 0
    num_of_ex_nds_in_L = 0
    num_of_ex_nds_for_PE = 0

    nd_range_l = index3(my_rank)
    nd_range_u = index3(my_rank + 1)


    !C count up sub matrix on self PE
    DO l = index(my_rank) + 1, index(my_rank + 1)
       m1 = index_nodes(l)
       k = part(m1) - nd_range_l
       DO i = INL_i(m1 - 1) + 1, INL_i(m1)
          j = part(IAL_i(i))
          call check_mat_L(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
               &           j, k, ex_nds_cnt, ex_nds_buf)
       END DO

       DO i = INU_i(m1 - 1) + 1, INU_i(m1) 
          j = part(IAU_i(i))
          call check_mat_U(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
               &           j, k, ex_nds_cnt, ex_nds_buf)
       END DO
    END DO



    !C count up sub matrices received from other PEs
    DO m = 1, PETOT
       if(my_rank == m - 1) cycle

       N_s    =  split_mats(m) % row_sz
       
       if(N_s>0) then
          ROW_NUM_PTR => split_mats(m) % row_num_s

          INL_PTR      => split_mats(m) % INL_s
          IAL_PTR      => split_mats(m) % IAL_s
          INU_PTR      => split_mats(m) % INU_s
          IAU_PTR      => split_mats(m) % IAU_s
       end if


       DO l = 1, N_s
          k = ROW_NUM_PTR(l) - nd_range_l

          DO i = INL_PTR(l) + 1, INL_PTR(l + 1)
             j = IAL_PTR(i)
             call check_mat_L(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
                  &           j, k, ex_nds_cnt, ex_nds_buf)
          END DO
          
          DO i = INU_PTR(l) + 1, INU_PTR(l + 1) 
             j = IAU_PTR(i)
             call check_mat_U(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
                  &           j, k, ex_nds_cnt, ex_nds_buf)
          END DO

       END DO
    END DO




    if(ex_nds_cnt > 0) then
       qs1 = ex_nds_cnt 
       qs2 = kint
       !C intel intrinsic subroutine
!!$       CALL QSORT(ex_nds_buf, qs1,  qs2, compare)
       call QSORTI_2(ex_nds_cnt, ex_nds_buf)
    end if



    !C-- redundant values are deleted from the array ex_nds_buf
    !C-- count up PE numbers in NEIBPETOT
    !C-- count up number of nodes for each PE
    k = 1;  NEIBPETOT = 0
    if( ex_nds_cnt > 0 ) then
       j = ex_nds_buf(1)
       NEIBPETOT = NEIBPETOT + 1
       
       !C- initial case is considered
       DO l = 1, PETOT
          if( j <= index3(l) ) exit
       END DO
       if( l > PETOT ) stop "unexpected error in REDIST_mtom():1"
       num_of_ex_nds_for_PE(l) = num_of_ex_nds_for_PE(l) + 1
       m = l
       
       !C- redundancy is removed and counting for each PE
       DO i = 2, ex_nds_cnt
          if( j /= ex_nds_buf(i) ) then
             j = ex_nds_buf(i)
             k = k + 1
             ex_nds_buf(k) = j

             !C counting for each PE
             DO l = m, PETOT
                if( j <= index3(l) ) exit
             END DO
             if( l > PETOT ) then
                stop "unexpected error in REDIST_mtom():2"
             end if
             
             if( m /= l ) NEIBPETOT = NEIBPETOT + 1
             m = l
             
             num_of_ex_nds_for_PE(l) = num_of_ex_nds_for_PE(l) + 1
          end if
       END DO
       ex_nds_cnt = k
    end if


    
    !C- NP can be determined only by the number of own external nodes
    NP = N + ex_nds_cnt
    NB = NB_i
    
    !C--NEIBPETOT may be increased and this chages the required size of the
    !C--arrays of STACK_IMPORT and STACK_EXPORT
    allocate(NEIBPE(PETOT), STACK_IMPORT(0:PETOT),              &
         &   STACK_EXPORT(0:PETOT), com_send_buf(ex_nds_cnt))
    
    !C- create the comunication table and send buffer for export table
    !C- containing global numbers.
    DO i = 1, ex_nds_cnt
       com_send_buf(i) = ex_nds_buf(i)
    END DO
    
    deallocate( ex_nds_buf )       
    
    !C-- import table is created
    !C-- external number is attached to STACK_IMPORT in order
    STACK_IMPORT(0) = 0; j = 0
    DO i = 1, PETOT
       if( num_of_ex_nds_for_PE(i) > 0 ) then
          j = j + 1
          NEIBPE(j) = i - 1
          STACK_IMPORT(j) = STACK_IMPORT(j - 1) + num_of_ex_nds_for_PE(i)
       end if
    END DO
    deallocate(num_of_ex_nds_for_PE)


    CALL count_time(2, solver_comm, my_rank, 15)    


    !C------------- late wait call
    !C-- waitall double part
    CALL MPI_WAITALL( req_cnt11, req11, sta11, ierr )
    CALL MPI_WAITALL( req_cnt22, req22, sta22, ierr )
    !C-- waitall double part


    CALL count_time(1, solver_comm, my_rank, 15)    
    
    deallocate(sendbuf_dbl, sd_buf_sz)
    deallocate(sta1, req1, sta11, req11)
    deallocate(sta2, sta22, req2, req22)

    !C-- pointer allocation. index of buf_IN? arrays start from 1.
    !C-- For double part.
    buf_ind_dbl = 0
    DO i = 1, PETOT
       N_s = index2(i) - index2(i - 1)
       if( my_rank == i - 1 .or. N_s == 0 ) cycle
       buf_NPL = split_mats(i) % INL_s(N_s+1)
       buf_NPU = split_mats(i) % INU_s(N_s+1)
       
       split_mats(i) % AL_s => recvbuf_dbl(1 + (NB_i+1)*NB_i*N_s + buf_ind_dbl :       &
            &                                  (NB_i+1)*NB_i*N_s + buf_ind_dbl + NB_i*NB_i*buf_NPL)
       split_mats(i) % AU_s => recvbuf_dbl(1 + (NB_i+1)*NB_i*N_s + buf_ind_dbl + NB_i*NB_i*buf_NPL : &
            &                                  (NB_i+1)*NB_i*N_s + buf_ind_dbl + NB_i*NB_i*(buf_NPL+buf_NPU))
       buf_ind_dbl = buf_ind_dbl + recv_buf_sz((i - 1) * 2)
    END DO
    !C------------ late wait call


    
    !C----- Export table creation composed of two subroutines.
    !C      These subroutines commonly use  in_recv_buf(0:), pes(:), n_pes, recv_buf_size
    allocate(in_recv_buf(0:PETOT), pes(PETOT))
    call size_export(SOLVER_COMM, PETOT, NEIBPETOT, NEIBPE, STACK_IMPORT,  &
         &  in_recv_buf, pes, n_pes, recv_buf_size)


    buf_NPL = NPL
    NPL = NPL - num_of_ex_nds_in_L
    NPU = NPU + num_of_ex_nds_in_L
    sz_export = recv_buf_size
    sz_import = ex_nds_cnt
    CALL count_time(2, solver_comm, my_rank, 15)    

  end SUBROUTINE size_new_mat_vec



  !C OUT: NOD_??PORT, IA?, A?, IN?, D, B, order_o
  !C INOUT: NEIBPETOT
  !C IN: part, index3, *_i
  subroutine assign_new_mat(N, NP, NPL, NPU, D, INU, INL, IAU, IAL, AU, AL, B, NEIBPETOT,  &
       & NEIBPE, NOD_IMPORT, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, part, index3,       &
       & N_i, NP_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, NEIBPETOT_i,    &
       & NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i, STACK_IMPORT_i,                   &
       & STACK_EXPORT_i, B_i, SOLVER_COMM, order_i, order_o)

    use count_time_mod
    use redist_mat_mod
    implicit none
    include 'mpif.h'

    integer(kind=kint ) :: i, j, k, l, m, m2, buf_ind_dbl, buf_ind
    integer, intent(in) :: SOLVER_COMM
    integer(kind=kint ), intent(in) :: index3(0:)


    !C-- partition info
    integer(kind=kint) :: part(:)
    integer(kind=kint) :: PETOT, my_rank, ierr

    integer(kind=kint ) :: order_i(:), order_o(:)


    
    !C-- initial matrix form
    integer(kind=kint) :: N_i, NP_i, NEIBPETOT_i
    real   (kind=KREAL), pointer :: D_i(:),B_i(:)
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:)
    integer(kind=kint ), pointer :: IAU_i(:), IAL_i(:)
    real   (kind=KREAL), pointer :: AU_i(:)
    real   (kind=KREAL), pointer :: AL_i(:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_i(:)
    integer(kind=kint ), pointer :: NEIBPE_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_i(:)


    !C-- redistributed matrix form (NPL, NPU isn't refered in other sub.)
    integer(kind=kint), intent(inout) :: NEIBPETOT
    integer(kind=kint), intent(in) :: N, NP, NPL, NPU
    real   (kind=KREAL) :: D(:), B(:)
    integer(kind=kint ) :: INU(0:), INL(0:)
    integer(kind=kint ) :: IAU(:), IAL(:)
    real   (kind=KREAL) :: AU(:)
    real   (kind=KREAL) :: AL(:)
    integer(kind=kint ) :: NOD_IMPORT(:)
    integer(kind=kint ) :: NOD_EXPORT(:)
    integer(kind=kint ) :: NEIBPE(:)
    integer(kind=kint ) :: STACK_IMPORT(0:)
    integer(kind=kint ) :: STACK_EXPORT(0:)

    integer(kind=kint ), pointer :: ROW_NUM_ptr(:)
    integer(kind=kint ), pointer :: IN_ptr(:), INL_ptr(:), INU_ptr(:)
    integer(kind=kint ), pointer :: IA_ptr(:), IAL_ptr(:), IAU_ptr(:)
    real   (kind=kreal), pointer :: A_ptr(:),  AL_ptr(:), AU_ptr(:)
    integer(kind=kint )          :: N_s, p, m1, L_index, U_index

    integer(kind=kint ) :: ex_nds_cnt


    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 

    CALL count_time(1, solver_comm, my_rank, 15)    

    ex_nds_cnt = NP - N

    call make_export (SOLVER_COMM, com_send_buf, index3, my_rank, NEIBPETOT,   &
         & NEIBPE, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, in_recv_buf, pes,   &
         & recv_buf_size, n_pes)
    
    deallocate(in_recv_buf, pes)
    !C-----
    

    DO i = 1, ex_nds_cnt
       NOD_IMPORT(i)   = i + N
    END DO
    

    !C-- 2 pass: neat the matrix

    INL(0) = 0; INU(0) = 0

    k = buf_NPL - num_of_ex_nds_in_L
    DO l = 1, ex_nds_cnt
       INL(l + N) = k 
       INU(l + N) = NPU
    END DO
    
    !C sub matrix of my_rank
    DO m1 = index(my_rank) + 1, index(my_rank + 1)
       k = index_nodes(m1)
       l = part(k) - nd_range_l
       INL(l) = INL_i(k) - INL_i(k - 1) - chk_rows_mv_LtoU(l)
       INU(l) = INU_i(k) - INU_i(k - 1) + chk_rows_mv_LtoU(l)
    END DO
    !C sub matrices received from the other PEs
    DO p = 1, PETOT
       if(my_rank == p - 1) cycle

       N_s = split_mats(p) % row_sz
       if( N_s > 0 ) then
          ROW_NUM_PTR => split_mats(p) % row_num_s
          INL_ptr     => split_mats(p) % INL_s
          INU_ptr     => split_mats(p) % INU_s
          DO m1 = 1, N_s
             l = ROW_NUM_PTR(m1) - nd_range_l

             INL(l) = INL_ptr(m1 + 1) - INL_ptr(m1) - chk_rows_mv_LtoU(l)
             INU(l) = INU_ptr(m1 + 1) - INU_ptr(m1) + chk_rows_mv_LtoU(l)
          END DO
       end if

    END DO
    

    DO i = 1, N
       INL(i) = INL(i) + INL(i - 1)
       INU(i) = INU(i) + INU(i - 1)
    END DO


    !C assign sub matrix on self
    DO m1 = index(my_rank) + 1, index(my_rank + 1)
       L_index = 0
       m2 = index_nodes(m1)
       l = part(m2) - nd_range_l

       !C- copy the buffered matrix on U part
       k = INU(l - 1) + 1
       DO i = INU_i(m2 - 1) + 1, INU_i(m2)
          j = part(IAU_i(i))
          call assign_mat_U(k, j, l, nd_range_l, nd_range_u, N, IAU, AU, AU_i(i), &
               & com_send_buf, STACK_IMPORT(NEIBPETOT), INL, IAL, AL, L_index, chk_rows_mv_LtoU)
       END DO
       U_index = k 
       
       !C- non-zero elements of external nodes in L part is moved to the U part
       k = INL(l - 1) + 1 + L_index
       m = INU(l) - chk_rows_mv_LtoU(l) + 1
       DO i = INL_i(m2 - 1) + 1, INL_i(m2)
          j = part(IAL_i(i))
          call assign_mat_L(k, j, m, l, nd_range_l, nd_range_u, N, IAU, AU, AL_i(i), &
               & com_send_buf, STACK_IMPORT(NEIBPETOT), INL, IAL, AL, U_index, chk_rows_mv_LtoU)
       END DO
    END DO

    !C sub matrices received from other PEs
    DO p = 1, PETOT
       if(my_rank == p - 1) cycle
       N_s = split_mats(p) % row_sz
       if(N_s > 0) then
          ROW_NUM_PTR => split_mats(p) % row_num_s
          INU_ptr     => split_mats(p) % INU_s
          INL_ptr     => split_mats(p) % INL_s
          IAU_ptr     => split_mats(p) % IAU_s
          IAL_ptr     => split_mats(p) % IAL_s
          AU_ptr      => split_mats(p) % AU_s
          AL_ptr      => split_mats(p) % AL_s
       end if
       
       DO m1 = 1, N_s
          L_index = 0
          l = ROW_NUM_PTR(m1) - nd_range_l
          
          !C- copy the buffered matrix on U part
          k = INU(l - 1) + 1
          DO i = INU_ptr(m1) + 1, INU_ptr(m1 + 1)
             j = IAU_ptr(i)
             call assign_mat_U(k, j, l, nd_range_l, nd_range_u, N, IAU, AU, AU_ptr(i), &
                  & com_send_buf, STACK_IMPORT(NEIBPETOT), INL, IAL, AL, L_index, chk_rows_mv_LtoU)
          END DO
          U_index = k
          
          !C- non-zero elements of external nodes in L part is moved to the U part
          k = INL(l - 1) + 1 + L_index
          m = INU(l) - chk_rows_mv_LtoU(l) + 1
          DO i = INL_ptr(m1) + 1, INL_ptr(m1 + 1)
             j = IAL_ptr(i)
             call assign_mat_L(k, j, m, l, nd_range_l, nd_range_u, N, IAU, AU, AL_ptr(i), &
                  & com_send_buf, STACK_IMPORT(NEIBPETOT), INL, IAL, AL, U_index, chk_rows_mv_LtoU)
          END DO
       END DO
       
    END DO

    D = 0

    DO k = index(my_rank) + 1, index(my_rank + 1)
       m = index_nodes(k)
       j = part(m) - nd_range_l
       D(j) = D_i(m)
       B(j) = B_i(m)
       order_o(j) = order_i(m)
    END DO

    buf_ind_dbl = 0; buf_ind = 0
    DO i = 1, PETOT
       if(my_rank == i - 1) cycle
       
       N_s = split_mats(i) % row_sz 
       if( N_s > 0 ) then
          ROW_NUM_ptr => split_mats(i) % row_num_s
          buf_NPL = split_mats(i) % INL_s(N_s + 1)
          buf_NPU = split_mats(i) % INU_s(N_s + 1)
          DO k = 1, N_s
             j = ROW_NUM_ptr(k) - nd_range_l

             D(j)       = recvbuf_dbl(buf_ind_dbl + k)
             B(j)       = recvbuf_dbl(buf_ind_dbl + N_s + k)
             order_o(j) = recvbuf(buf_ind + 2 * N_s + buf_NPL + buf_NPU + 3 + k)

          END DO
          buf_ind_dbl = buf_ind_dbl + recv_buf_sz((i - 1) * 2)
          buf_ind     = buf_ind     + recv_buf_sz((i - 1) * 2 - 1)
          
       end if
    END DO
    
    deallocate( chk_rows_mv_LtoU, com_send_buf )
    deallocate( recvbuf_dbl, recvbuf, recv_buf_sz )
    deallocate( split_mats )
    deallocate(index, index2, index_nodes) 
     
    CALL count_time(2, solver_comm, my_rank, 15)        

  END SUBROUTINE ASSIGN_NEW_MAT


  !C OUT: NOD_??PORT, IA?, A?, IN?, D, B, order_o
  !C INOUT: NEIBPETOT
  !C IN: part, index3, *_i
  subroutine assign_new_mat_vec(N, NP, NB, NPL, NPU, D, INU, INL, IAU, IAL, AU, AL, B, NEIBPETOT,  &
       & NEIBPE, NOD_IMPORT, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, part, index3,       &
       & N_i, NP_i, NB_i, D_i, INU_i, INL_i, IAU_i, IAL_i, AU_i, AL_i, NEIBPETOT_i,    &
       & NEIBPE_i, NOD_IMPORT_i, NOD_EXPORT_i, STACK_IMPORT_i,                   &
       & STACK_EXPORT_i, B_i, SOLVER_COMM, order_i, order_o)

    use count_time_mod
    use redist_mat_mod
    implicit none
    include 'mpif.h'

    integer(kind=kint ) :: i, j, k, l, m, m2, buf_ind_dbl, buf_ind
    integer, intent(in) :: SOLVER_COMM
    integer(kind=kint ), intent(in) :: index3(0:)


    !C-- partition info
    integer(kind=kint) :: part(:)
    integer(kind=kint) :: PETOT, my_rank, ierr

    integer(kind=kint ) :: order_i(:), order_o(:)


    
    !C-- initial matrix form
    integer(kind=kint) :: N_i, NP_i, NB_i, NEIBPETOT_i
    real   (kind=KREAL), pointer :: D_i(:,:,:),B_i(:)
    integer(kind=kint ), pointer :: INU_i(:), INL_i(:)
    integer(kind=kint ), pointer :: IAU_i(:), IAL_i(:)
    real   (kind=KREAL), pointer :: AU_i(:,:,:)
    real   (kind=KREAL), pointer :: AL_i(:,:,:)
    integer(kind=kint ), pointer :: NOD_IMPORT_i(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_i(:)
    integer(kind=kint ), pointer :: NEIBPE_i(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_i(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_i(:)


    !C-- redistributed matrix form (NPL, NPU isn't refered in other sub.)
    integer(kind=kint), intent(inout) :: NEIBPETOT
    integer(kind=kint), intent(in) :: N, NP, NB, NPL, NPU
    real   (kind=KREAL) :: D(:,:,:), B(:)
    integer(kind=kint ) :: INU(0:), INL(0:)
    integer(kind=kint ) :: IAU(:), IAL(:)
    real   (kind=KREAL) :: AU(:,:,:)
    real   (kind=KREAL) :: AL(:,:,:)
    integer(kind=kint ) :: NOD_IMPORT(:)
    integer(kind=kint ) :: NOD_EXPORT(:)
    integer(kind=kint ) :: NEIBPE(:)
    integer(kind=kint ) :: STACK_IMPORT(0:)
    integer(kind=kint ) :: STACK_EXPORT(0:)

    integer(kind=kint ), pointer :: ROW_NUM_ptr(:)
    integer(kind=kint ), pointer :: IN_ptr(:), INL_ptr(:), INU_ptr(:)
    integer(kind=kint ), pointer :: IA_ptr(:), IAL_ptr(:), IAU_ptr(:)
    real   (kind=kreal), pointer :: A_ptr(:),  AL_ptr(:), AU_ptr(:)
    integer(kind=kint )          :: N_s, p, m1, L_index, U_index

    integer(kind=kint ) :: ex_nds_cnt,k1,k2


    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 

    CALL count_time(1, solver_comm, my_rank, 15)    

    ex_nds_cnt = NP - N

    call make_export (SOLVER_COMM, com_send_buf, index3, my_rank, NEIBPETOT,   &
         & NEIBPE, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, in_recv_buf, pes,   &
         & recv_buf_size, n_pes)
    
    deallocate(in_recv_buf, pes)
    !C-----
    

    DO i = 1, ex_nds_cnt
       NOD_IMPORT(i)   = i + N
    END DO
    

    !C-- 2 pass: neat the matrix

    INL(0) = 0; INU(0) = 0

    k = buf_NPL - num_of_ex_nds_in_L
    DO l = 1, ex_nds_cnt
       INL(l + N) = k 
       INU(l + N) = NPU
    END DO
    
    !C sub matrix of my_rank
    DO m1 = index(my_rank) + 1, index(my_rank + 1)
       k = index_nodes(m1)
       l = part(k) - nd_range_l
       INL(l) = INL_i(k) - INL_i(k - 1) - chk_rows_mv_LtoU(l)
       INU(l) = INU_i(k) - INU_i(k - 1) + chk_rows_mv_LtoU(l)
    END DO
    !C sub matrices received from the other PEs
    DO p = 1, PETOT
       if(my_rank == p - 1) cycle

       N_s = split_mats(p) % row_sz
       if( N_s > 0 ) then
          ROW_NUM_PTR => split_mats(p) % row_num_s
          INL_ptr     => split_mats(p) % INL_s
          INU_ptr     => split_mats(p) % INU_s
          DO m1 = 1, N_s
             l = ROW_NUM_PTR(m1) - nd_range_l

             INL(l) = INL_ptr(m1 + 1) - INL_ptr(m1) - chk_rows_mv_LtoU(l)
             INU(l) = INU_ptr(m1 + 1) - INU_ptr(m1) + chk_rows_mv_LtoU(l)
          END DO
       end if

    END DO
    

    DO i = 1, N
       INL(i) = INL(i) + INL(i - 1)
       INU(i) = INU(i) + INU(i - 1)
    END DO


    !C assign sub matrix on self
    DO m1 = index(my_rank) + 1, index(my_rank + 1)
       L_index = 0
       m2 = index_nodes(m1)
       l = part(m2) - nd_range_l

       !C- copy the buffered matrix on U part
       k = INU(l - 1) + 1
       DO i = INU_i(m2 - 1) + 1, INU_i(m2)
          j = part(IAU_i(i))
          call assign_mat_U_vec1(k, j, l, nd_range_l, nd_range_u, N, IAU, AU, AU_i(:,:,i), &
               & com_send_buf, STACK_IMPORT(NEIBPETOT), INL, IAL, AL, L_index, chk_rows_mv_LtoU,NB)
       END DO
       U_index = k 
       
       !C- non-zero elements of external nodes in L part is moved to the U part
       k = INL(l - 1) + 1 + L_index
       m = INU(l) - chk_rows_mv_LtoU(l) + 1
       DO i = INL_i(m2 - 1) + 1, INL_i(m2)
          j = part(IAL_i(i))
          call assign_mat_L_vec1(k, j, m, l, nd_range_l, nd_range_u, N, IAU, AU, AL_i(:,:,i), &
               & com_send_buf, STACK_IMPORT(NEIBPETOT), INL, IAL, AL, U_index, chk_rows_mv_LtoU,NB)
       END DO
    END DO

    !C sub matrices received from other PEs
    DO p = 1, PETOT
       if(my_rank == p - 1) cycle
       N_s = split_mats(p) % row_sz
       if(N_s > 0) then
          ROW_NUM_PTR => split_mats(p) % row_num_s
          INU_ptr     => split_mats(p) % INU_s
          INL_ptr     => split_mats(p) % INL_s
          IAU_ptr     => split_mats(p) % IAU_s
          IAL_ptr     => split_mats(p) % IAL_s
          AU_ptr      => split_mats(p) % AU_s
          AL_ptr      => split_mats(p) % AL_s
       end if
       
       DO m1 = 1, N_s
          L_index = 0
          l = ROW_NUM_PTR(m1) - nd_range_l
          
          !C- copy the buffered matrix on U part
          k = INU(l - 1) + 1
          DO i = INU_ptr(m1) + 1, INU_ptr(m1 + 1)
             j = IAU_ptr(i)
             call assign_mat_U_vec2(k, j, l, nd_range_l, nd_range_u, N, IAU, AU, &
                  & AU_ptr((i-1)*NB*NB+1:i*NB*NB), &
                  & com_send_buf, STACK_IMPORT(NEIBPETOT), INL, IAL, AL, L_index, chk_rows_mv_LtoU,NB)
          END DO
          U_index = k
          
          !C- non-zero elements of external nodes in L part is moved to the U part
          k = INL(l - 1) + 1 + L_index
          m = INU(l) - chk_rows_mv_LtoU(l) + 1
          DO i = INL_ptr(m1) + 1, INL_ptr(m1 + 1)
             j = IAL_ptr(i)
             call assign_mat_L_vec2(k, j, m, l, nd_range_l, nd_range_u, N, IAU, AU, &
                  & AL_ptr((i-1)*NB*NB+1:i*NB*NB), &
                  & com_send_buf, STACK_IMPORT(NEIBPETOT), INL, IAL, AL, U_index, chk_rows_mv_LtoU,NB)
          END DO
       END DO
       
    END DO

    D = 0

    DO k = index(my_rank) + 1, index(my_rank + 1)
       m = index_nodes(k)
       j = part(m) - nd_range_l
       D(:,:,j) = D_i(:,:,m)
       B((j-1)*NB+1:j*NB) = B_i((m-1)*NB+1:m*NB)
       order_o(j) = order_i(m)
    END DO

    buf_ind_dbl = 0; buf_ind = 0
    DO i = 1, PETOT
       if(my_rank == i - 1) cycle
       
       N_s = split_mats(i) % row_sz 
       if( N_s > 0 ) then
          ROW_NUM_ptr => split_mats(i) % row_num_s
          buf_NPL = split_mats(i) % INL_s(N_s + 1)
          buf_NPU = split_mats(i) % INU_s(N_s + 1)
          DO k = 1, N_s
             j = ROW_NUM_ptr(k) - nd_range_l

             DO k1=1,NB
                DO k2=1,NB
                   D(k1,k2,j)       = recvbuf_dbl(buf_ind_dbl + NB*NB*(k-1)+NB*(k1-1)+k2)
                END DO
             END DO
             
             DO k1=1,NB
                B(NB*(j-1)+k1)      = recvbuf_dbl(buf_ind_dbl + NB*NB*N_s + NB*(k-1)+k1)
             END DO
             order_o(j) = recvbuf(buf_ind + 2 * N_s + buf_NPL + buf_NPU + 3 + k)

          END DO
          buf_ind_dbl = buf_ind_dbl + recv_buf_sz((i - 1) * 2)
          buf_ind     = buf_ind     + recv_buf_sz((i - 1) * 2 - 1)
          
       end if
    END DO
    
    deallocate( chk_rows_mv_LtoU, com_send_buf )
    deallocate( recvbuf_dbl, recvbuf, recv_buf_sz )
    deallocate( split_mats )
    deallocate(index, index2, index_nodes) 
    
    CALL count_time(2, solver_comm, my_rank, 15)        

  END SUBROUTINE ASSIGN_NEW_MAT_VEC



  SUBROUTINE assign_mat_L(k, j, m, l, nd_range_l, nd_range_u, N, IAU, AU, v, &
       & com_send_buf, S_I_N, INL, IAL, AL, U_index, chk_rows_mv_LtoU)
    implicit none
    include 'precision.inc'

    integer(kind=kint ), intent(in)    :: nd_range_l, nd_range_u, j, l, com_send_buf(:), N, S_I_N
    integer(kind=kint ), intent(inout) :: k, m, INL(0:), IAL(:), IAU(:), chk_rows_mv_LtoU(:), U_index
    real   (kind=kreal), intent(in)    :: v
    real   (kind=kreal), intent(inout) :: AU(:), AL(:)

    integer(kind=kint ) :: ex_id
    
    !C k: index of AL, IAL
    !C m: index of AU, IAU
    !C j: column number in global
    !C l: row number at local
    !C v: value of the element

    if( nd_range_l >= j .or. j > nd_range_u ) then                
       ex_id = N + search(j, com_send_buf, 1, S_I_N)
       IAU(m) = ex_id
       AU(m)  = v
       m = m + 1

    else if(l < j - nd_range_l) then
       !C this case is that nodes mv from L to U
       !C U_index record the next position in U
       IAU(U_index) = j - nd_range_l
       AU (U_index) = v
       U_index = U_index + 1
    else
       !C ordinary case
       IAL(k) = j - nd_range_l
       AL(k)  = v
       k = k + 1
    end if

  end subroutine assign_mat_L

  !C v is NB x NB two-dimentional array.
  SUBROUTINE assign_mat_L_vec1(k, j, m, l, nd_range_l, nd_range_u, N, IAU, AU, v, &
       & com_send_buf, S_I_N, INL, IAL, AL, U_index, chk_rows_mv_LtoU,NB)
    implicit none
    include 'precision.inc'

    integer(kind=kint ), intent(in)    :: nd_range_l, nd_range_u, j, l, com_send_buf(:), N, S_I_N,NB
    integer(kind=kint ), intent(inout) :: k, m, INL(0:), IAL(:), IAU(:), chk_rows_mv_LtoU(:), U_index
    real   (kind=kreal), intent(in)    :: v(:,:)
    real   (kind=kreal), intent(inout) :: AU(:,:,:), AL(:,:,:)

    integer(kind=kint ) :: ex_id,k1,k2
    
    !C k: index of AL, IAL
    !C m: index of AU, IAU
    !C j: column number in global
    !C l: row number at local
    !C v: value of the element

    if( nd_range_l >= j .or. j > nd_range_u ) then                
       ex_id = N + search(j, com_send_buf, 1, S_I_N)
       IAU(m) = ex_id
       DO k1=1,NB
          DO k2=1,NB
             AU(k1,k2,m)  = v(k1,k2)
          END DO
       END DO
       m = m + 1

    else if(l < j - nd_range_l) then
       !C this case is that nodes mv from L to U
       !C U_index record the next position in U
       IAU(U_index) = j - nd_range_l
       DO k1=1,NB
          DO k2=1,NB
             AU (k1,k2,U_index) = v(k1,k2)
          END DO
       END DO
       U_index = U_index + 1
    else
       !C ordinary case
       IAL(k) = j - nd_range_l
       DO k1=1,NB
          DO k2=1,NB
             AL(k1,k2,k)  = v(k1,k2)
          END DO
       END DO
       k = k + 1
    end if

  end subroutine assign_mat_L_vec1

  !C v is NB x NB one-dimentional array.
  SUBROUTINE assign_mat_L_vec2(k, j, m, l, nd_range_l, nd_range_u, N, IAU, AU, v, &
       & com_send_buf, S_I_N, INL, IAL, AL, U_index, chk_rows_mv_LtoU,NB)
    implicit none
    include 'precision.inc'

    integer(kind=kint ), intent(in)    :: nd_range_l, nd_range_u, j, l, com_send_buf(:), N, S_I_N,NB
    integer(kind=kint ), intent(inout) :: k, m, INL(0:), IAL(:), IAU(:), chk_rows_mv_LtoU(:), U_index
    real   (kind=kreal), intent(in)    :: v(:)
    real   (kind=kreal), intent(inout) :: AU(:,:,:), AL(:,:,:)

    integer(kind=kint ) :: ex_id,k1,k2
    
    
    !C k: index of AL, IAL
    !C m: index of AU, IAU
    !C j: column number in global
    !C l: row number at local
    !C v: value of the element

    if( nd_range_l >= j .or. j > nd_range_u ) then                
       ex_id = N + search(j, com_send_buf, 1, S_I_N)
       IAU(m) = ex_id
       DO k1=1,NB
          DO k2=1,NB
             AU(k1,k2,m)  = v((k1-1)*NB+k2)
          END DO
       END DO
       m = m + 1

    else if(l < j - nd_range_l) then
       !C this case is that nodes mv from L to U
       !C U_index record the next position in U
       IAU(U_index) = j - nd_range_l
       DO k1=1,NB
          DO k2=1,NB
             AU (k1,k2,U_index) = v((k1-1)*NB+k2)
          END DO
       END DO
       U_index = U_index + 1
    else
       !C ordinary case
       IAL(k) = j - nd_range_l
       DO k1=1,NB
          DO k2=1,NB
             AL(k1,k2,k)  = v((k1-1)*NB+k2)
          END DO
       END DO
       k = k + 1
    end if

  end subroutine assign_mat_L_vec2


  
  subroutine assign_mat_U(k, j, l, nd_range_l, nd_range_u, N, IAU, AU, v, &
       & com_send_buf, S_I_N, INL, IAL, AL, L_index, chk_rows_mv_LtoU)
    implicit none
    include 'precision.inc'

    integer(kind=kint), intent(in)    :: nd_range_l, nd_range_u, j, l, com_send_buf(:), N, S_I_N
    integer(kind=kint), intent(inout) :: k, INL(0:), IAL(:), IAU(:), chk_rows_mv_LtoU(:), L_index
    real(kind=kreal), intent(in) :: v
    real(kind=kreal), intent(inout) :: AU(:), AL(:)

    integer(kind=kint) :: ex_id

    !C k: index of AU, IAU
    !C j: column number in global
    !C l: row number at local
    !C v: value of the element
    
    if( nd_range_l >= j .or. j > nd_range_u ) then
       ex_id = N + search(j, com_send_buf, 1, S_I_N) 
       IAU(k) = ex_id
       AU(k)  = v
       k = k + 1

    else if(l > j - nd_range_l) then
       !C this case is that nodes mv from U to L
       !C L_index counts the number of nodes
       L_index = L_index + 1
       IAL(L_index + INL(l - 1))  = j - nd_range_l
       AL (L_index + INL(l - 1))  = v
       chk_rows_mv_LtoU(l)     = chk_rows_mv_LtoU(l) + 1
    else 
       !C ordinary case
       IAU(k) = j - nd_range_l
       AU(k)  = v
       k = k + 1
    end if
  end subroutine assign_mat_U

  !C v is NB x NB two-dimentional array.
  subroutine assign_mat_U_vec1(k, j, l, nd_range_l, nd_range_u, N, IAU, AU, v, &
       & com_send_buf, S_I_N, INL, IAL, AL, L_index, chk_rows_mv_LtoU,NB)
    implicit none
    include 'precision.inc'

    integer(kind=kint), intent(in)    :: nd_range_l, nd_range_u, j, l, com_send_buf(:), N, S_I_N,NB
    integer(kind=kint), intent(inout) :: k, INL(0:), IAL(:), IAU(:), chk_rows_mv_LtoU(:), L_index
    real(kind=kreal), intent(in) :: v(:,:)
    real(kind=kreal), intent(inout) :: AU(:,:,:), AL(:,:,:)

    integer(kind=kint) :: ex_id,k1,k2

    !C k: index of AU, IAU
    !C j: column number in global
    !C l: row number at local
    !C v: value of the element
    
    if( nd_range_l >= j .or. j > nd_range_u ) then
       ex_id = N + search(j, com_send_buf, 1, S_I_N) 
       IAU(k) = ex_id
       DO k1=1,NB
          DO k2=1,NB
             AU(k1,k2,k)  = v(k1,k2)
          END DO
       END DO
       k = k + 1

    else if(l > j - nd_range_l) then
       !C this case is that nodes mv from U to L
       !C L_index counts the number of nodes
       L_index = L_index + 1
       IAL(L_index + INL(l - 1))  = j - nd_range_l
       DO k1=1,NB
          DO k2=1,NB
             AL (k1,k2,L_index + INL(l - 1))  = v(k1,k2)
          END DO
       END DO
       chk_rows_mv_LtoU(l)     = chk_rows_mv_LtoU(l) + 1
    else 
       !C ordinary case
       IAU(k) = j - nd_range_l
       DO k1=1,NB
          DO k2=1,NB
             AU(k1,k2,k)  = v(k1,k2)
          END DO
       END DO
       k = k + 1
    end if
  end subroutine assign_mat_U_vec1

  !C v is NB x NB one-dimentional array.
  subroutine assign_mat_U_vec2(k, j, l, nd_range_l, nd_range_u, N, IAU, AU, v, &
       & com_send_buf, S_I_N, INL, IAL, AL, L_index, chk_rows_mv_LtoU,NB)
    implicit none
    include 'precision.inc'

    integer(kind=kint), intent(in)    :: nd_range_l, nd_range_u, j, l, com_send_buf(:), N, S_I_N,NB
    integer(kind=kint), intent(inout) :: k, INL(0:), IAL(:), IAU(:), chk_rows_mv_LtoU(:), L_index
    real(kind=kreal), intent(in) :: v(:)
    real(kind=kreal), intent(inout) :: AU(:,:,:), AL(:,:,:)

    integer(kind=kint) :: ex_id,k1,k2

    !C k: index of AU, IAU
    !C j: column number in global
    !C l: row number at local
    !C v: value of the element
    
    if( nd_range_l >= j .or. j > nd_range_u ) then
       ex_id = N + search(j, com_send_buf, 1, S_I_N) 
       IAU(k) = ex_id
       DO k1=1,NB
          DO k2=1,NB
             AU(k1,k2,k)  = v((k1-1)*NB+k2)
          END DO
       END DO
       k = k + 1

    else if(l > j - nd_range_l) then
       !C this case is that nodes mv from U to L
       !C L_index counts the number of nodes
       L_index = L_index + 1
       IAL(L_index + INL(l - 1))  = j - nd_range_l
       DO k1=1,NB
          DO k2=1,NB
             AL (k1,k2,L_index + INL(l - 1))  = v((k1-1)*NB+k2)
          END DO
       END DO
       chk_rows_mv_LtoU(l)     = chk_rows_mv_LtoU(l) + 1
    else 
       !C ordinary case
       IAU(k) = j - nd_range_l
       DO k1=1,NB
          DO k2=1,NB
             AU(k1,k2,k)  = v((k1-1)*NB+k2)
          END DO
       END DO
       k = k + 1
    end if
  end subroutine assign_mat_U_vec2


  
  subroutine check_mat_L(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
       &               j, k, ex_nds_cnt, ex_nds_buf)
    include 'precision.inc'
    integer(kind=kint), intent(in)    :: nd_range_l, nd_range_u, j, k
    integer(kind=kint), intent(inout) :: ex_nds_cnt, ex_nds_buf(:), chk_rows_mv_LtoU(:), num_of_ex_nds_in_L
    
    if( nd_range_l >= j .or. j > nd_range_u ) then
       !C external nodes
       chk_rows_mv_LtoU(k) = chk_rows_mv_LtoU(k) + 1
       num_of_ex_nds_in_L  = num_of_ex_nds_in_L  + 1
       ex_nds_cnt = ex_nds_cnt + 1
       ex_nds_buf( ex_nds_cnt ) = j
    else if( k < j - nd_range_l ) then
       chk_rows_mv_LtoU(k) = chk_rows_mv_LtoU(k) + 1
       num_of_ex_nds_in_L  = num_of_ex_nds_in_L  + 1
    end if
  end subroutine check_mat_L

  subroutine check_mat_U(nd_range_l, nd_range_u, chk_rows_mv_LtoU, num_of_ex_nds_in_L, &
       &               j, k, ex_nds_cnt, ex_nds_buf)
    include 'precision.inc'
    integer(kind=kint), intent(in)    :: nd_range_l, nd_range_u, j, k
    integer(kind=kint), intent(inout) :: ex_nds_cnt, ex_nds_buf(:), chk_rows_mv_LtoU(:), num_of_ex_nds_in_L
    
    if( nd_range_l >= j .or. j > nd_range_u ) then
       !C external nodes
       ex_nds_cnt = ex_nds_cnt + 1
       ex_nds_buf( ex_nds_cnt ) = j
    else if( k > j - nd_range_l ) then
       chk_rows_mv_LtoU(k) = chk_rows_mv_LtoU(k) - 1
       num_of_ex_nds_in_L  = num_of_ex_nds_in_L  - 1
    end if
  end subroutine check_mat_U


  RECURSIVE FUNCTION search(v,data,s,e) RESULT(res)
    implicit none
    include 'precision.inc'
    integer(kind=kint) :: res
    integer(kind=kint) :: v, m, s, e 
    integer(kind=kint), dimension(:) :: data
    
    if( s > e ) then
       write(*,*) "error value: ", v
       stop "error : no elements : search()"
    end if
    m = (s+e)/2
    if( v == data(m) ) then
       res = m
    else if( v < data(m) ) then
       res = search(v,data,s,m-1)
    else 
       res = search(v,data,m+1,e)
    end if
  end FUNCTION search
  

  integer(2) function compare(p1,p2)
    include 'precision.inc'
    integer(kind=kint) :: p1,p2

    if(p1.LT.p2) then
       compare = -1
    else if(p1 == p2) then
       compare = 0
    else
       compare = 1
    end if
  end function compare
  
  SUBROUTINE make_export_NOD (SOLVER_COMM, export_global_buffer,            &
       &     GIN_aggregate, my_rank, NPROCS, NEIBPETOT, NEIBPE, NOD_EXPORT, &
       &     STACK_IMPORT, STACK_EXPORT)

    IMPLICIT NONE
    include 'precision.inc'
    INCLUDE  'mpif.h'
    INTEGER(kind=kint), INTENT(in)     :: NPROCS
    INTEGER(kind=kint), INTENT(inout)  :: NEIBPETOT
    INTEGER(kind=kint), INTENT(inout)  :: NEIBPE(:)
    INTEGER(kind=kint), INTENT(inout)  :: STACK_IMPORT(0:),STACK_EXPORT(0:)
    INTEGER(kind=kint), POINTER        :: NOD_EXPORT(:)

    INTEGER(kind=kint),  INTENT(in)    :: export_global_buffer(:)
    INTEGER(kind=kint), ALLOCATABLE    :: pes(:),in_recv_buf(:),recv_buf(:)

    INTEGER(kind=kint) :: pes_s(NPROCS), pes_r(NPROCS)

    INTEGER(kind=kint), INTENT(in)     :: GIN_aggregate(0:)

    INTEGER :: ierr
    INTEGER, INTENT(in) :: SOLVER_COMM
    INTEGER, INTENT(in) :: my_rank
    INTEGER(kind=kint), ALLOCATABLE :: req1(:),req2(:)
    INTEGER(kind=kint), ALLOCATABLE :: sta1(:,:),sta2(:,:)

    INTEGER(kind=kint) :: i, j, k, dest, inum, istart, recv_buf_size, message_size, l
    INTEGER(kind=kint) :: GLOBAL, LOCAL, base_index, n_pes, neib, from_pe, msg_sz
    PARAMETER(GLOBAL = 1, LOCAL = 2)
    LOGICAL :: hash_flag

    neib=NEIBPETOT
    ALLOCATE(sta1(MPI_STATUS_SIZE,neib))
    ALLOCATE(req1(neib))    

    !C-- determine number of recv
    pes_s = -1
    DO i = 1, neib
       dest = NEIBPE(i)
       pes_s(dest + 1)     = STACK_IMPORT(i) - STACK_IMPORT(i - 1)
#ifdef DEBUG
       IF(pes_s(dest + 1) == 0) STOP "0 size tables in make_export_NOD()"
#endif
    END DO
    pes_r = -1
    CALL MPI_ALLTOALL(pes_s, 1, MPI_INTEGER, pes_r, 1, MPI_INTEGER, &
         &            SOLVER_COMM, ierr)
    
    !C-- determine number of recv
    !C RECEIVE
    n_pes = 0
    ALLOCATE(in_recv_buf(0:NPROCS))
    ALLOCATE(pes(NPROCS))
    in_recv_buf(0)=0
    DO i=0,NPROCS-1
       from_pe = i
       msg_sz  = pes_r(i+1)
       IF(msg_sz>=0) THEN
          n_pes              = n_pes+1
          pes(n_pes)         = from_pe
          in_recv_buf(n_pes) = in_recv_buf(n_pes-1)+msg_sz
       END IF
    END DO
    
    ALLOCATE(req2(n_pes))
    ALLOCATE(sta2(MPI_STATUS_SIZE,n_pes))    
    recv_buf_size = in_recv_buf(n_pes)
    ALLOCATE(recv_buf(recv_buf_size))
    ALLOCATE(NOD_EXPORT(recv_buf_size))
    DO i=1,n_pes
       istart = in_recv_buf(i-1)
       inum   = in_recv_buf(i)-in_recv_buf(i-1)
       dest   = pes(i)
       CALL MPI_IRECV(recv_buf(istart+1), inum, MPI_INTEGER, dest ,11,&
            &        SOLVER_COMM, req2(i),ierr)
    END DO

    !C SEND
    DO i = 1, neib
       istart = STACK_IMPORT(i - 1)
       inum   = STACK_IMPORT(i) - istart
       dest = NEIBPE(i)
       CALL MPI_ISEND(export_global_buffer(istart + 1), inum, MPI_INTEGER, dest, 11,  &
            &         SOLVER_COMM, req1(i), ierr)
    END DO
    
    CALL MPI_WAITALL (n_pes, req2, sta2, ierr)
    CALL MPI_WAITALL (neib,  req1, sta1, ierr)

    base_index = GIN_aggregate(my_rank)    
    STACK_EXPORT(0) = 0
    DO i = 1, neib
       dest = NEIBPE(i)
       j = 1
       DO WHILE(j <= n_pes .AND. pes(j) /= dest)
          j = j + 1
       END DO
       IF(j > n_pes) THEN
          !C there is no nodes to send.
          STACK_EXPORT(i) = STACK_EXPORT(i - 1)
       ELSE
          STACK_EXPORT(i) = STACK_EXPORT(i - 1) + in_recv_buf(j) - in_recv_buf(j - 1)
          l = STACK_EXPORT(i - 1) - in_recv_buf(j - 1)
          DO k = in_recv_buf(j - 1) + 1, in_recv_buf(j)
             NOD_EXPORT(l + k) = recv_buf(k) - base_index
#ifdef DEBUG
             IF(pes(j)==-1) STOP "error!!! in make_export_table"
             IF(NOD_EXPORT(k+l)<0 .OR. NOD_EXPORT(k+l)>GIN_aggregate(my_rank+1)-base_index) THEN
                WRITE(*,*) "unexpected error in make_export_nod():", NOD_EXPORT(k+l),my_rank
                STOP
             END IF
#endif            
          END DO
          !C mark which means un-needed 
          pes(j) = -1
       END IF
    END DO

    
    DO i=1,n_pes
       IF(pes(i)/=-1) THEN
          dest=pes(i)
          NEIBPETOT=NEIBPETOT+1
          NEIBPE(NEIBPETOT)=dest
          STACK_IMPORT(NEIBPETOT)=STACK_IMPORT(NEIBPETOT-1)
          STACK_EXPORT(NEIBPETOT)=STACK_EXPORT(NEIBPETOT-1)+in_recv_buf(i)-in_recv_buf(i-1)
          l=STACK_EXPORT(NEIBPETOT-1)-in_recv_buf(i-1)
          DO k=in_recv_buf(i-1)+1,in_recv_buf(i)
             NOD_EXPORT(k+l)=recv_buf(k)-base_index
#ifdef DEBUG
             IF(NOD_EXPORT(k+l)<0 .OR. NOD_EXPORT(k+l)>GIN_aggregate(my_rank+1)-base_index) THEN
                WRITE(*,*) "unexpected error in make_export_nod():", NOD_EXPORT(k + l), my_rank
                STOP
             END IF
#endif            
          END DO
       END IF
    END DO

    DEALLOCATE(pes)    
    DEALLOCATE(in_recv_buf)    
    DEALLOCATE(recv_buf)    
    DEALLOCATE(req1)
    DEALLOCATE(req2)
    DEALLOCATE(sta2)
    DEALLOCATE(sta1)
  END SUBROUTINE make_export_NOD

  
  !C allocate n_pes, in_recv_buf(0:NPROCS), and pes(NPROCS) beforehand
  SUBROUTINE size_export (SOLVER_COMM, NPROCS, NEIBPETOT, NEIBPE, &
       &     STACK_IMPORT, in_recv_buf, pes, n_pes, recv_buf_size)
    
    IMPLICIT NONE
    include 'precision.inc'
    INCLUDE 'mpif.h'
    INTEGER(kind=kint), INTENT(in) :: NPROCS, NEIBPETOT
    INTEGER(kind=kint), INTENT(in) :: NEIBPE(:)
    INTEGER(kind=kint), INTENT(in) :: STACK_IMPORT(0:)
    INTEGER, INTENT(in) :: SOLVER_COMM
    
    INTEGER(kind=kint), INTENT(inout) :: pes(:), in_recv_buf(0:)
    INTEGER(kind=kint), INTENT(out)   :: recv_buf_size, n_pes

    INTEGER(kind=kint) :: pes_s(NPROCS), pes_r(NPROCS)
    
    INTEGER :: ierr
    
    INTEGER(kind=kint) :: i, j, k, dest
    INTEGER(kind=kint) :: neib, from_pe, msg_sz

    neib = NEIBPETOT

    !C-- determine number of recv
    pes_s = -1
    DO i = 1, neib
       dest = NEIBPE(i)
       pes_s(dest + 1)     = STACK_IMPORT(i) - STACK_IMPORT(i - 1)
#ifdef DEBUG
       IF(pes_s(dest + 1) == 0) STOP "0 size tables in make_export_NOD()"
#endif
    END DO
    pes_r = -1
    CALL MPI_ALLTOALL(pes_s, 1, MPI_INTEGER, pes_r, 1, MPI_INTEGER, &
         &            SOLVER_COMM, ierr)
    
    !C-- determine number of recv
    n_pes = 0
    in_recv_buf(0) = 0
    DO i = 0, NPROCS - 1
       from_pe = i
       msg_sz  = pes_r(i + 1)
       IF(msg_sz >= 0) THEN
          n_pes              = n_pes + 1
          pes(n_pes)         = from_pe
          in_recv_buf(n_pes) = in_recv_buf(n_pes - 1) + msg_sz
       END IF
    END DO
    
    recv_buf_size = in_recv_buf(n_pes)
    
  END SUBROUTINE size_export


  !C Allocate NOD_EXPORT(recv_buf_size) beforehand
  SUBROUTINE make_export (SOLVER_COMM, export_global_buffer,                  &
       &   GIN_aggregate, my_rank, NEIBPETOT, NEIBPE, NOD_EXPORT,             &
       &   STACK_IMPORT, STACK_EXPORT, in_recv_buf, pes, recv_buf_size, n_pes)

    IMPLICIT NONE
    include 'precision.inc'
    INCLUDE 'mpif.h'

    INTEGER, INTENT(in) :: SOLVER_COMM
    INTEGER, INTENT(in) :: my_rank
    INTEGER(kind=kint), INTENT(inout) :: NEIBPETOT
    INTEGER(kind=kint), INTENT(inout) :: NEIBPE(:)
    INTEGER(kind=kint), INTENT(inout) :: STACK_IMPORT(0:), STACK_EXPORT(0:)
    INTEGER(kind=kint), INTENT(in)    :: recv_buf_size,  n_pes
!C                                       NOD_EXPORT(recv_buf_size)
    INTEGER(kind=kint), INTENT(out)   :: NOD_EXPORT(:)
    INTEGER(kind=kint), INTENT(in)    :: export_global_buffer(:)
    INTEGER(kind=kint), INTENT(in)    :: GIN_aggregate(0:)
    INTEGER(kind=kint), INTENT(inout) :: pes(:), in_recv_buf(0:)

    INTEGER(kind=kint), ALLOCATABLE   :: recv_buf(:)
    INTEGER :: ierr
    INTEGER(kind=kint), ALLOCATABLE :: req1(:), req2(:)
    INTEGER(kind=kint), ALLOCATABLE :: sta1(:,:), sta2(:,:)

    INTEGER(kind=kint) :: i, j, k, dest, inum, istart,  message_size, l
    INTEGER(kind=kint) :: base_index, neib  

    neib = NEIBPETOT
    ALLOCATE(recv_buf(recv_buf_size))
    ALLOCATE(sta1(MPI_STATUS_SIZE, neib))
    ALLOCATE(req1(neib))    
    ALLOCATE(sta2(MPI_STATUS_SIZE, n_pes))    
    ALLOCATE(req2(n_pes))
    
    DO i = 1, n_pes
       istart = in_recv_buf(i - 1)
       inum   = in_recv_buf(i) - in_recv_buf(i - 1)
       dest   = pes(i)
       CALL MPI_IRECV(recv_buf(istart + 1), inum, MPI_INTEGER, dest ,11,&
            &        SOLVER_COMM, req2(i), ierr)
    END DO

    !C SEND
    DO i = 1, neib
       istart = STACK_IMPORT(i - 1)
       inum   = STACK_IMPORT(i) - istart
       dest = NEIBPE(i)
       CALL MPI_ISEND(export_global_buffer(istart + 1), inum, MPI_INTEGER, dest, 11,  &
            &         SOLVER_COMM, req1(i), ierr)
    END DO
    
    CALL MPI_WAITALL (n_pes, req2, sta2, ierr)
    CALL MPI_WAITALL (neib,  req1, sta1, ierr)

    base_index = GIN_aggregate(my_rank)    
    STACK_EXPORT(0) = 0
    DO i = 1, neib
       dest = NEIBPE(i)
       DO j = 1, n_pes
          if(pes(j) == dest) exit
       end DO

       IF(j > n_pes) THEN
          !C there is no nodes to send.
          STACK_EXPORT(i) = STACK_EXPORT(i - 1)
       ELSE
          STACK_EXPORT(i) = STACK_EXPORT(i - 1) + in_recv_buf(j) - in_recv_buf(j - 1)
          l = STACK_EXPORT(i - 1) - in_recv_buf(j - 1)
          DO k = in_recv_buf(j - 1) + 1, in_recv_buf(j)
             NOD_EXPORT(l + k) = recv_buf(k) - base_index
#ifdef DEBUG
             IF(pes(j)==-1) STOP "error!!! in make_export_table"
             IF(NOD_EXPORT(k+l)<0 .OR. NOD_EXPORT(k+l)>GIN_aggregate(my_rank+1)-base_index) THEN
                WRITE(*,*) "unexpected error in make_export_nod():", NOD_EXPORT(k+l),my_rank
                STOP
             END IF
#endif            
          END DO
          !C mark which means un-needed 
          pes(j) = -1
       END IF
    END DO

    
    DO i = 1, n_pes
       IF(pes(i) /= -1) THEN
          dest = pes(i)
          NEIBPETOT = NEIBPETOT + 1
          NEIBPE(NEIBPETOT) = dest
          STACK_IMPORT(NEIBPETOT) = STACK_IMPORT(NEIBPETOT - 1)
          STACK_EXPORT(NEIBPETOT) = STACK_EXPORT(NEIBPETOT - 1) + in_recv_buf(i) - in_recv_buf(i - 1)
          l = STACK_EXPORT(NEIBPETOT - 1) - in_recv_buf(i - 1)
          DO k = in_recv_buf(i - 1) + 1, in_recv_buf(i)
             NOD_EXPORT(k + l) = recv_buf(k) - base_index
#ifdef DEBUG
             IF(NOD_EXPORT(k+l)<0 .OR. NOD_EXPORT(k+l)>GIN_aggregate(my_rank+1)-base_index) THEN
                WRITE(*,*) "unexpected error in make_export_nod():", NOD_EXPORT(k + l), my_rank
                STOP
             END IF
#endif            
          END DO
       END IF
    END DO

    DEALLOCATE(recv_buf)    
    DEALLOCATE(req1)
    DEALLOCATE(req2)
    DEALLOCATE(sta2)
    DEALLOCATE(sta1)
  END SUBROUTINE make_export

END MODULE  matrix_dist_mtom
