#ifndef __MY_CUDA_H
#define __MY_CUDA_H

#include <cstddef>
#include <cuda_runtime.h>
#include <iostream>

int my_cudaMalloc(void **devptr, size_t count, char str[]);
int my_cudaFree(void *devptr);

int my_cudaMemcpy(void *dst, const void *src, size_t count, cudaMemcpyKind kind);
int my_cudaMemset(void *devptr, int c, size_t count);
int my_cudaSetDevice(int device);
int my_cudaGetLastError(void);

double my_cuda_dot(double *a, double *b, int n);
bool my_cuda_vector_less(double *dx, double *dy, int n);

int my_cudaScale(double *a, double *scl, int n);
int my_cudaInvScale(double *a, double *scl, int n);

template <class type>
int my_cuda_host_to_device_array(type *dst, const type *src, size_t dst_length, size_t src_length, int *order)
{
	size_t i;
	type *buffer;

	buffer = new type[dst_length];
	for(i=0; i<dst_length; i++){
		buffer[i] = 0;
	}	
	for(i=0; i<src_length; i++){
		buffer[ order[i] ] = src[i];
	}
	if( my_cudaMemcpy(dst, buffer, sizeof(type) * dst_length, cudaMemcpyHostToDevice) ){
		std::cout << "my_cuda_host_to_device_array expection." << std::endl;
		return 1;
	}
	delete[] buffer;
	return 0;
}


template <class type>
int my_cuda_device_to_host_array(type *dst, const type *src, size_t dst_length, size_t src_length, int *order)
{
	size_t i;
	type *buffer;

	buffer = new type[src_length];
	for(i=0; i<src_length; i++){
		buffer[i] = 0;
	}
	if( my_cudaMemcpy(buffer, src, sizeof(type) * src_length, cudaMemcpyDeviceToHost) ){
		std::cout << "my_cuda_device_to_host_array expection." << std::endl;
		return 1;
	}
	for(i=0; i<dst_length; i++){
		dst[i] = buffer[ order[i] ];
	}
	delete[] buffer;
	return 0;
}

#endif
