#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#define SP 256
#define SM 32

__global__
void put_into_the_buffer_kernel(double *buffer, const double *src, const int *index, const int n)
{
	unsigned int id = blockIdx.x * blockDim.x + threadIdx.x;
	if(id < n){
		buffer[id] = src[ index[id] ];
	}
}


void put_into_the_buffer(double *buffer, const double *src, const int *index, const int n)
{
	put_into_the_buffer_kernel<<<n / SP + 1, SP>>>(buffer, src, index, n);
}


__global__
void put_into_the_data_kernel(double *data, const double *buffer, const int *index, const int n)
{
	unsigned int id = blockIdx.x * blockDim.x + threadIdx.x;
	if(id < n){
		data[ index[id] ] = buffer[id];
	}
}


void put_into_the_data(double *data, const double *buffer, const int *index, const int n)
{
	put_into_the_data_kernel<<<n / SP + 1, SP>>>(data, buffer, index, n);
}


__global__
void add_into_the_data_kernel(double *data, const double *buffer, const int *index, const int n)
{
	int i;
	for(i=0; i<n; i++){
		data[ index[i] ] += buffer[i];
	}
}


void add_into_the_data(double *data, const double *buffer, const int *index, const int n)
{
	add_into_the_data_kernel<<< 1, 1 >>>(data, buffer, index, n);
}

