#include "output_log.h"

#include <fstream>
#include <sstream>
#include <iomanip>
#include <iostream>
#include <list>

#include <mpi.h>

using std::endl;
using std::ostringstream;
using std::ofstream;
using std::fixed;
using std::setprecision;
using std::list;

typedef struct
{
	int byte;
	double start_time;
	double end_time;
}comm_log_item;

list<comm_log_item> comm_log;


void comm_log_input(int byte, double start_time, double end_time)
{
	comm_log_item item = {byte, start_time, end_time};
	comm_log.push_back(item);
}


void comm_log_clear()
{
	comm_log.clear();
}

void comm_log_output()
{
	ostringstream os;
	ofstream ofs;
	int rank, byte;
	double tick, comm_time, start_time, end_time;

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	os << "comm_" << rank << ".csv";
	ofs.open( os.str().c_str() );
	tick = MPI_Wtick();

	list<comm_log_item>::iterator it;
	for(it=comm_log.begin(); it != comm_log.end(); it++){
		byte       = it->byte;
		start_time = it->start_time;
		end_time   = it->end_time;
		comm_time  = end_time - start_time;
		comm_time  = tick * (int)(comm_time / tick + 0.5);

		ofs << byte << ",";
		ofs << fixed << setprecision(6);
		ofs << comm_time << "," << start_time << "," << end_time << endl;
	}
}


list<double> residual_log;

void residual_log_input(double residual)
{
	residual_log.push_back(residual);
}


void residual_log_clear()
{
	residual_log.clear();
}


void residual_log_output()
{
	ofstream ofs;

	ofs.open("residual.csv");

	list<double>::iterator it;
	for(it=residual_log.begin(); it != residual_log.end(); it++){
		ofs << std::scientific << *it << endl;
	}
}


void result_log_output(double setup_time, double cycle_time, int iter, int iter_down, double omega)
{
	ofstream ofs;
	int rank;

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	if(rank == 0){
		ofs.open("result.csv");
		ofs << setup_time << "," << cycle_time << "," << iter << "," << iter_down <<  "," << omega <<  endl;
		ofs.close();
	}
}

