!C   Copyright (C) 2012 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY HIGH PERFORMANCE COMPUTING LABORATORY IN 
!C   KOGAKUIN UNIVERSITY ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, 
!C   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
!C   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
!C   THE HIGH PERFORMANCE COMPUTING LABORATORY BE LIABLE FOR ANY DIRECT, 
!C   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
!C   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
!C   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
!C   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
!C   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
!C   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

module parameter_at
  include 'precision.inc'

  !C public: at_parameter_setting(), AMG_solver_include_AT()

  !C private:
  real(kind=kreal) :: tv

  logical :: at_flag = .false.

  logical :: offline_fl = .false.

  integer(kind=kint) :: expected_call_times = -1

!C  integer :: t1, t2, t_max, t_rate
  
contains

  SUBROUTINE set_expected_call_times(count)
    implicit none
    integer(kind=kint) :: count
    
    expected_call_times = count
    
  END SUBROUTINE set_expected_call_times

  
  SUBROUTINE AMG_solver(X_i, B_i, iter, SETUP_FL)
    use matrix_dist_mtom
    USE problem_matrix
    ! for keisoku
    use count_time_mod
    use AMG_cycle, only:scale_div_vec, scale_mul_vec

    implicit none
    include 'mpif.h'

    integer(kind=kint), intent(inout) :: iter
    real(kind=kreal),   intent(inout) :: B_i(:), X_i(:)
    real(kind=kreal),   allocatable   :: B(:), X(:)
    real(kind=kreal),   allocatable   :: B_tmp(:), X_tmp(:)
    logical, intent(in)              :: SETUP_FL
    logical :: CONV_FL = .false.
    integer(kind=kint) :: N_i
    real(kind=kreal), save :: t1 = 0.0, t2 = 0.0

    integer(kind=kint) :: i,j,k,l
    

#ifdef MEASURING
    ! for keisoku
    integer(kind=kint) :: ierr

    if(my_rank == 0) write(*,*) "PETOT is", PETOT
    call count_time(0,SOLVER_COMM, my_rank, 20)
#endif

    t1 = t1-MPI_WTIME()
    !C t2 is time for vector redistribution of rank 0.

    if(.not.BLOCKING_FL .and. SCALING_FL) then
       if(REORDERED_FL) then
          i=nds_index(my_rank+1)-nds_index(my_rank)
       else 
          i=N
       end if

       call scale_div_vec(B_i, scale_D, i)
       call scale_mul_vec(X_i, scale_D, i)
    end if
    

    if(REORDERED_FL) then
       t2 = t2-MPI_WTIME()
       N_i = nds_index(my_rank+1) - nds_index(my_rank)
       
       allocate(X(NP*NB),B(NP*NB))
       if(BLOCKING_FL) then
          !C X_i -> X_tmp -> X -> X_i
          allocate(X_tmp(N_i*NB),B_tmp(N_i*NB))
          X_tmp(1:N_BEFORE_BL) = X_i(1:N_BEFORE_BL)
          B_tmp(1:N_BEFORE_BL) = B_i(1:N_BEFORE_BL)

          call distribute_vector(nds_index_aft_dst, order_for_dst, SOLVER_COMM, X_tmp, N_i,  &
               &                 X,NB)    
          call distribute_vector(nds_index_aft_dst, order_for_dst, SOLVER_COMM, B_tmp, N_i,  &
               &                 B,NB)    
       else
          call distribute_vector(nds_index_aft_dst, order_for_dst, SOLVER_COMM, X_i, N_i,  &
               &                 X,NB)    
          call distribute_vector(nds_index_aft_dst, order_for_dst, SOLVER_COMM, B_i, N_i,  &
               &                 B,NB)    
       end if
       t2 = t2+MPI_WTIME()

#ifdef MEASURING
       call MPI_barrier(SOLVER_COMM, ierr)
       call count_time(1,SOLVER_COMM, my_rank, 20)
#endif         
       !C call switch_solver(X, B,  iter, SETUP_FL)
       call switch_solver_include_AT(X, B, iter, SETUP_FL, CONV_FL)

#ifdef MEASURING
       ! for keisoku
       call count_time(2,SOLVER_COMM, my_rank, 20)
#endif
       t2 = t2-MPI_WTIME()
       
       if(BLOCKING_FL) then
          call distribute_vector(nds_index, order, SOLVER_COMM, X, N,  &
               &                 X_tmp,NB)
          X_i(1:N_BEFORE_BL) = X_tmp(1:N_BEFORE_BL)
          deallocate(X_tmp, B_tmp)
       else
          call distribute_vector(nds_index, order, SOLVER_COMM, X, N,  &
               &                 X_i,NB)
       end if
       t2 = t2+MPI_WTIME()
       deallocate(X,B)
       
    else if(BLOCKING_FL) then
       t2 = t2-MPI_WTIME()
       allocate(X(NP*NB),B(NP*NB))
       X(1:N_BEFORE_BL) = X_i(1:N_BEFORE_BL)
       B(1:N_BEFORE_BL) = B_i(1:N_BEFORE_BL)
       t2 = t2+MPI_WTIME()

#ifdef MEASURING
       call MPI_barrier(SOLVER_COMM, ierr)
       call count_time(1,SOLVER_COMM, my_rank, 20)
#endif         
       !C call switch_solver(X, B,  iter, SETUP_FL)
       call switch_solver_include_AT(X, B, iter, SETUP_FL, CONV_FL)

#ifdef MEASURING
       ! for keisoku
       call count_time(2,SOLVER_COMM, my_rank, 20)
#endif
       t2 = t2-MPI_WTIME()
       X_i(1:N_BEFORE_BL) = X(1:N_BEFORE_BL)
       t2 = t2+MPI_WTIME()
       deallocate(X,B)

    else

#ifdef MEASURING
       ! for keisoku
       call MPI_barrier(SOLVER_COMM, ierr)
       call count_time(1,SOLVER_COMM, my_rank, 20)
#endif
       
       call switch_solver_include_AT(X_i, B_i, iter, SETUP_FL, CONV_FL)
       
       
       
#ifdef MEASURING         
       ! for keisoku
       call count_time(2,SOLVER_COMM, my_rank, 20)
#endif         
    end if

    
    if(.not.BLOCKING_FL.and. SCALING_FL) then
       if(REORDERED_FL) then
          i=nds_index(my_rank+1)-nds_index(my_rank)
       else 
          i=N
       end if
       call scale_div_vec(X_i,scale_D,i)
    end if


#ifdef MEASURING               
    if(my_rank == 0) write(*,*) "-Setup Part-"
    call count_time(3, SOLVER_COMM, my_rank, 19)

    if(my_rank == 0) write(*,*) "-Iterative Part-"
    call count_time(3, SOLVER_COMM, my_rank, 20)

#ifdef GPU
    if(my_rank == 0) write(*,*) "-Copy matrices to GPU devices-"
    call count_time(3, SOLVER_COMM, my_rank, 61)
#endif

    call count_time(4, SOLVER_COMM, my_rank, 20)
    if(my_rank == 0) write(*,*) " "

    if(my_rank == 0) write(*,*) "### Setup Part ### lev (num-29)"
    do i = 2, 10
       call count_time(3, SOLVER_COMM, my_rank, 30+i-1)
    end do
    if(my_rank == 0) write(*,*) " "
    
    if(my_rank == 0) write(*,*) "### AMG_cycle time at each level : lev (num-40) ###"
    do i = 1, 10
       call count_time(3, SOLVER_COMM, my_rank, 40+i)
    end do
    if(my_rank == 0) write(*,*) " "

    if(my_rank == 0) write(*,*) "### AMG_cycle time between levels : lev (num-50) - lev (num-49) ###"
    do i = 1, 9
       call count_time(3, SOLVER_COMM, my_rank, 50+i)
    end do
    if(my_rank == 0) write(*,*) " "


    if(my_rank == 0) write(*,*) "### NEIBPETOT at each lev : lev num ###"
    do i = 1, 10
       call store_and_disp_data(2, SOLVER_COMM, my_rank, i, 0.d0)
    end do
    if(my_rank == 0) write(*,*) " "

    if(my_rank == 0) write(*,*) "### COMM size at each lev : lev (num-10) ###"
    do i = 11, 20
       call store_and_disp_data(2, SOLVER_COMM, my_rank, i, 0.d0)
    end do
    if(my_rank == 0) write(*,*) " "

    if(my_rank == 0) write(*,*) "### NEIBPETOT between levs : lev (num-20) - lev (num-19) ###"
    do i = 21, 30
       call store_and_disp_data(2, SOLVER_COMM, my_rank, i, 0.d0)
    end do
    if(my_rank == 0) write(*,*) " "

    if(my_rank == 0) write(*,*) "### COMM size between levs : lev (num-30) - lev (num-29) ###"
    do i = 31, 40
       call store_and_disp_data(2, SOLVER_COMM, my_rank, i, 0.d0)
    end do
    if(my_rank == 0) write(*,*) " "


    if(my_rank == 0) write(*,*) "### NEIBPETOT with axis comm. at each lev : lev (num-30) - lev (num-40) ###"
    do i = 41, 50
       call store_and_disp_data(2, SOLVER_COMM, my_rank, i, 0.d0)
    end do
    if(my_rank == 0) write(*,*) " "

    if(my_rank == 0) write(*,*) "### NEIBPETOT with axis comm. between levs : lev (num-50) - lev (num-49) ###"
    do i = 51, 60
       call store_and_disp_data(2, SOLVER_COMM, my_rank, i, 0.d0)
    end do
    if(my_rank == 0) write(*,*) " "


    do i = 1, 10
       call disp_nonzero_data(2, SOLVER_COMM, my_rank, i, 0.d0)
    end do
    if(my_rank == 0) write(*,*) " "


    
#endif
    
    t1 = t1+MPI_WTIME()
#ifdef PRINT_REZ    
    if(my_rank == 0) write(*,*) "AMG solver time:", t1, "vector dist time:", t2
#endif

  END SUBROUTINE AMG_SOLVER


  subroutine at_preprocess(X, B, solver_setup_fl, setup_time)
    use problem_matrix, only: N, NP, LEVEL_NUM, SOLVER_COMM, my_rank
    use residual_calculation, only: cal_res_vec
    use common_mod, only: setup_phase
    use at_mod, only: keisoku_fl,  setupflag_for_AT, setup_param_changed_fl, &
         & solver_choice, parameter_type, set_parameter_on_all_PEs, initial_resid

    use terminal_condition, only: terminal_kind, conv_criterion_vec, conv_criterion_val, &
         & BNRM2, terminal_cond_check
    implicit none
    include 'mpif.h'

    real(kind=kreal), intent(inout)   :: B(:)
    real(kind=kreal), intent(inout)   :: X(:)
    real(kind=kreal) :: BNRM, v, w, tv2, RNRM,RNRM2, infnrm_data0(2), infnrm_data(2)
    real(kind=kreal) :: setup_time, conv_vec_norm0, conv_vec_norm

    integer :: ierr
    logical :: solver_setup_fl, fl
    integer(kind=kint), save :: X_bak_sz=0
    real(kind=kreal), pointer, save :: X_bak(:)

    type(parameter_type) :: param

    integer(kind=kint) :: i
     
    if(keisoku_fl) then
       if(X_bak_sz == 0) then
          X_bak_sz = NP
          allocate(X_bak(NP))
       else if(X_bak_sz < NP) then
          deallocate(X_bak)
          allocate(X_bak(NP))
          X_bak_sz = NP
       end if
       

       call cal_res_vec(X_bak, X, B)

       !
       if(terminal_kind == 1) then
          
          !C BNRM2 is set as infinity norm of rhs vector
          BNRM = 0
          
          !$omp parallel do reduction(max:BNRM) private(v)
          do i = 1, N
             v = dabs(B(i))
             if(BNRM < v) BNRM = v
          end do
          !$omp end parallel do
          call MPI_ALLREDUCE(BNRM, BNRM2, 1, MPI_DOUBLE_PRECISION, MPI_MAX, SOLVER_COMM, ierr)
          if(BNRM2 == 0) BNRM2 = 1.0

          !C this sub distribute conv_criterion_vec(:) and relative inifinity residual norm at tv.
          call terminal_cond_check(fl, tv, X_bak, N, SOLVER_COMM)

          initial_resid = tv

          !C conv_vec_norm contains infinity norm of conv_criterion_vec(:)
          conv_vec_norm0 = 0
          !$omp parallel do reduction(max:conv_vec_norm0) private(v)
          do i = 1, N
             v = dabs(conv_criterion_vec(i))
             if(conv_vec_norm0 < v) conv_vec_norm0 = v
          end do
          !$omp end parallel do
          call MPI_REDUCE(conv_vec_norm0, conv_vec_norm, 1, MPI_DOUBLE_PRECISION, MPI_MAX, 0, SOLVER_COMM, ierr)
          
          if(my_rank == 0) tv = conv_vec_norm / (tv*BNRM2)
          
       else if(terminal_kind == 0) then
          !C BNRM2 is updated
          BNRM = 0
          !$omp parallel do reduction(+:BNRM)
          do i = 1, N
             BNRM = BNRM+B(i)**2
          end do
          !$omp end parallel do
          call MPI_ALLREDUCE(BNRM, BNRM2, 1, MPI_DOUBLE_PRECISION, MPI_SUM, SOLVER_COMM, ierr)

          call terminal_cond_check(fl, tv, X_bak, N, SOLVER_COMM)
          
          if(my_rank == 0) then
             !C initial_resid has initial relative residual norm
             initial_resid = tv
             tv = conv_criterion_val / tv
          end if
          
       else
          stop 'unknown terminal_kind'
          
       end if

    end if

    ! 4. only my_rank == 0 the following operations..
    if(my_rank == 0)  then
#ifdef PRINT_REZ    
       write(*,*) "******", tv, "******"
#endif
       if(tv > 0.9) tv = 0.9
       if(.not.offline_fl) call solver_choice(tv, setupflag_for_AT, expected_call_times, param)
    end if
    
    if(.not.offline_fl) then
       ! 5. call all_PE_parameter_setting(): Broadcast rank 0 decision and state.
       !    state includes: keisoku_fl, setup_param_changed_fl

       call set_parameter_on_all_PEs(param)


    end if
    

    if(keisoku_fl .and. .not.offline_fl) then
       !C--check if multi-level data is set appropriately.
       !C       solver_setup_fl       -> create multi-level data everytime in AMG_solver()
       !C       .not.solver_setup_fl  -> call setup() only when theta is changed
       
       if(.not.solver_setup_fl .and. (LEVEL_NUM == 0 .or. setup_param_changed_fl) .or. &
            & solver_setup_fl .and. (LEVEL_NUM==0)) then

          setup_time = -MPI_WTIME()
          call setup_phase()
          setup_time = MPI_WTIME()

          !C unset setup_param_changed_fl
          setup_param_changed_fl = .false.
       end if

!!$       if(solver_setup_fl .and. LEVEL_NUM > 0) call clear_multi_lev_matrix(LEVEL_NUM)

       !C unset setup_param_changed_fl
       if(solver_setup_fl) setup_param_changed_fl = .false.
       
    end if


  end subroutine at_preprocess
  

  
  SUBROUTINE switch_solver_include_AT(X, B, iter, solver_setup_fl, CONV_FL)
    use solver, only: switch_solver, conv_state
    use problem_matrix, only: LEVEL_NUM, my_rank, SOLVER_COMM
    use common_mod, only: setup_phase
    use amg_parameter, only: AT_init
    use at_mod, only: keisoku_fl, offline_tuned_solver
    use count_time_mod, only: count_time_val
    implicit none
    
    include 'mpif.h'

    integer(kind=kint), intent(inout) :: iter
    real(kind=kreal), intent(inout)   :: B(:)
    real(kind=kreal), intent(inout)   :: X(:)
    logical, intent(in)               :: solver_setup_fl
    logical :: CONV_FL
    real(kind=kreal) :: setup_time, iter_time, val


    call count_time_val(0, SOLVER_COMM, my_rank, 2, val)
    call count_time_val(1, SOLVER_COMM, my_rank, 2, val)
       
    
    !C when AT_init == 0, at_flag is set.
    !C no AT is used when AT_init is less than 0 at first
    if(AT_init > 0) then
       AT_init = AT_init - 1
    else if(.not.at_flag .and. AT_init == 0) then
       AT_init = AT_init - 1
       at_flag = .true.
    end if

    if(at_flag) then
       call at_preprocess(X, B, solver_setup_fl, setup_time)
    end if


    !C call system_clock(t1, t_rate, t_max)
    iter_time = -MPI_WTIME()
    

    if(at_flag .and. keisoku_fl .and. offline_fl) then
       call offline_tuned_solver(X, B, tv, iter, solver_setup_fl)
    else if(at_flag .and. keisoku_fl) then
       ! 1. AMG_Solver -> switch_solver and the subroutine's arguments are modified

       call switch_solver(X, B, iter, .false.)
    else
       call switch_solver(X, B, iter, solver_setup_fl)
    end if


    !C call system_clock(t1, t_rate, t_max)
    iter_time = iter_time + MPI_WTIME()

    if(at_flag .and. iter/=0) call at_postprocess(iter, X, solver_setup_fl, setup_time, iter_time)

    call conv_state(CONV_FL)

    call count_time_val(2, SOLVER_COMM, my_rank, 2, val)
    
  END SUBROUTINE Switch_solver_include_AT


  !C out: t2, X, keisoku_fl, iter
  subroutine at_postprocess(iter, X, solver_setup_fl, setup_time, iter_time)
    use solver, only: conv_state, resid
    use amg_parameter, only: conv_check_fl
    use problem_matrix, only: NP, N, my_rank, SOLVER_COMM, LEVEL_NUM
    use terminal_condition, only: BNRM2, terminal_kind
    use at_mod, only: keisoku_fl, initial_resid, update_time, check
    use common_mod, only: clear_multi_lev_matrix
    implicit none

    include 'mpif.h'
    
    real(kind=kreal), intent(inout)   :: X(:)
    logical  :: reach_conv, solver_setup_fl
    integer(kind=kint) :: iter, i, ierr
    real(kind=kreal)   ::  setup_time, iter_time 
    

    !C this sub should be included in amgs_solver()
    call conv_state(reach_conv)


    if(keisoku_fl .and. .not.offline_fl) then
       ! 1. only rank 0 process records execution time.
       
       !C following part is for correct reduction of residual vector.
!!$       if(conv_check_fl) then
!!$          !C -- correct tv is calculated..
!!$          if(terminal_kind==0 .and. my_rank == 0) then
!!$             !C BNRM2 contains sum of the squares of rhs vector elements
!!$             tv = resid / initial_resid
!!$          else if(terminal_kind==1) then
!!$             !C BNRM2 contains infinity norm of rhs vector
!!$             tv = resid / initial_resid
!!$          end if
!!$       end if
       
       if(my_rank == 0) call update_time(setup_time, iter_time, iter, tv, reach_conv, solver_setup_fl)
       
    else
       
       
       !C if iteration number increases by some fixed value or reach_conv == .false.,  then
       !C keisoku_fl is set
       if(my_rank == 0) then
          call check(iter, reach_conv, keisoku_fl, offline_fl)
       end if

       call MPI_BCAST(keisoku_fl, 1, MPI_LOGICAL, 0, SOLVER_COMM, ierr)

       
    end if

    expected_call_times = expected_call_times-1
    
    !C-- after this sub, assigning X(:) to PP(:).
    if(solver_setup_fl .and. keisoku_fl) call clear_multi_lev_matrix(LEVEL_NUM)
    
  end subroutine at_postprocess
  
  subroutine at_parameter_finish()
    at_flag = .false.
  end subroutine at_parameter_finish


  subroutine parameter_check()
    use amg_parameter
    implicit none
    if(cycle_kind<0 .or. cycle_kind >3) stop "cycle setting is 0-3"
    if(gs_kind<0 .or. gs_kind >4)       stop "smoother type is 0-4"
    
  end subroutine parameter_check


  subroutine amg_parameter_set_w_val(               &
       & cycle_shape, iter_at_each_lev,             &
       & iter_bottom, F_cycle_iter_down,            &
       & smoother_type, dump_gauss_seidel_coeff,    &
       & strong_connect_thresld, dump_jacobi_coeff, &
       & solver_str, num_of_vecs, dist_str,         &
       & conv_chk_fl )
    
    use amg_parameter, only:  cycle_kind, gs_kind, iteration_number_down, iteration_number_up, &
         &                    iteration_number_bottom, f_first_iteration_number, theta,        &
         &                    dump_jacobi_weight, dump_gauss_seidel_weight, solver_string,     &
         &                    number_of_vectors, conv_check_fl                                 
    use problem_matrix, only: dst_string

    implicit none
    !C vector problem is not considered here.
    integer(kind=kint) :: cycle_shape, iter_at_each_lev, iter_bottom 
    integer(kind=kint) :: smoother_type
    integer(kind=kint) :: F_cycle_iter_down, num_of_vecs, dist_method
    real(kind=kreal)   :: strong_connect_thresld, dump_jacobi_coeff
    real(kind=kreal)   :: dump_gauss_seidel_coeff
    character(*), intent(in)  :: solver_str, dist_str
    logical :: conv_chk_fl

    integer(kind=kint) :: i

    iteration_number_up      = iter_at_each_lev
    iteration_number_down    = iter_at_each_lev
    iteration_number_bottom  = iter_bottom
    f_first_iteration_number = F_cycle_iter_down
    cycle_kind               = cycle_shape
    number_of_vectors        = num_of_vecs
    gs_kind                  = smoother_type
    theta                    = strong_connect_thresld
    dump_jacobi_weight       = dump_jacobi_coeff
    dump_gauss_seidel_weight = dump_gauss_seidel_coeff
    
    i = len_trim(solver_str)
    solver_string(1:i)       = solver_str(1:i)

    i = len_trim(dist_str)
    dst_string(1:i)          = dist_str(1:i)
    
    conv_check_fl            = conv_chk_fl

    call parameter_check()
    
  end subroutine amg_parameter_set_w_val

  subroutine amg_parameter_set_w_val_test(          &
       & iter_at_each_lev,                          &
       & smoother_type, dump_gauss_seidel_coeff,    &
       & strong_connect_thresld, solver_str,        &
       & num_of_vecs)
    use amg_parameter
    use problem_matrix, only: dst_string

    implicit none
    !C vector problem is not considered here.
    integer(kind=kint) :: cycle_shape, iter_at_each_lev, iter_bottom 
    integer(kind=kint) :: smoother_type
    integer(kind=kint) :: F_cycle_iter_down, num_of_vecs, dist_method
    real(kind=kreal)   :: strong_connect_thresld, dump_jacobi_coeff
    real(kind=kreal)   :: dump_gauss_seidel_coeff
    character(*), intent(in)  :: solver_str
    logical :: conv_chk_fl

    integer(kind=kint) :: i


    cycle_shape = 0  !C V-cycle
    iter_bottom = 30 !C LU factorization is used by default
    F_cycle_iter_down = 1
    dump_jacobi_coeff = 0.666 !C aggregate smoothing dump coef.
    conv_chk_fl = .false.

    
    
    call amg_parameter_set_w_val(                   &
       & cycle_shape, iter_at_each_lev,             &
       & iter_bottom, F_cycle_iter_down,            &
       & smoother_type, dump_gauss_seidel_coeff,    &
       & strong_connect_thresld, dump_jacobi_coeff, &
       & solver_str, num_of_vecs, "FIX",            &
       & conv_chk_fl )
    
  end subroutine amg_parameter_set_w_val_test

  subroutine amg_parameter_set_alpha_beta(Nunit, SOLVER_COMM, alpha, beta)
    use amg_parameter
    implicit none
    integer, intent(in) :: SOLVER_COMM
    integer, intent(in) :: Nunit
    real(kind=kreal) :: alpha, beta
    
    call amg_parameter_set(Nunit, SOLVER_COMM)
    
    theta = alpha
    theta_reduction_rate = beta
    
  end subroutine amg_parameter_set_alpha_beta
    

  subroutine parameter_file_input(Nunit, SOLVER_COMM)
    use problem_matrix, only: dst_string
    use amg_parameter,  only: cycle_kind, gs_kind, iteration_number_down, iteration_number_up, &
         &                    iteration_number_bottom, f_first_iteration_number, theta,        &
         &                    dump_jacobi_weight, dump_gauss_seidel_weight, solver_string,     &
         &                    number_of_vectors, conv_check_str, conv_check_fl,                &
         &                    AT_string, AT_init, AT_interval, AT_diff, AT_setup, NB,          &
         &                    vector_problem_fl
    use at_mod, only: at_parameter_setting, at_parameter_setting_from_file
    implicit none

    include 'mpif.h'
    integer :: Nunit, SOLVER_COMM
    integer :: stat, ierr, my_rank
    character(len=20) :: mstring = "SA", vector_scalar_str = "SCALAR"
    
    call MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr)
    !C-- AMG.dat
    if(my_rank == 0) then
       open(Nunit,FILE='AMG.dat',FORM='FORMATTED',STATUS='old', iostat=stat)
       if(stat /= 0) then
          write(*,*) "No AMG parameter file: AMG.dat. Default setting is used."
       else
          read(Nunit, *, iostat=stat) cycle_kind 
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) gs_kind
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) iteration_number_down
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) iteration_number_up
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) iteration_number_bottom
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) f_first_iteration_number
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) theta
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) dump_jacobi_weight
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) dump_gauss_seidel_weight
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) mstring
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) solver_string 
          if(stat /= 0) goto 120
!!$    if(solver_string(1:1)=='G' .and. solver_string(2:2)=='M' .and. &
!!$         &  solver_string(3:3)=='R' .and. solver_string(4:4)=='E' .or. &
!!$         &  solver_string(1:1)=='I' .and. solver_string(2:2)=='D' .and. &
!!$         &  solver_string(3:3)=='R'.or. solver_string(1:1)=='R' .and.&
!!$         &  solver_string(2:2)=='E' .and. solver_string(3:3)=='S' .and. &
!!$         &  solver_string(4:4)=='C') then
          read(Nunit, *, iostat=stat) number_of_vectors
          if(stat /= 0) goto 120
!!$    end if
          read(Nunit, *, iostat=stat) dst_string
          if(stat /= 0) goto 120
          read(Nunit, *, iostat=stat) conv_check_str
          conv_check_fl = conv_check_str(1:1)=='C' .and. conv_check_str(2:2)=='H' .and. &
               &  conv_check_str(3:3)=='E' .and. conv_check_str(4:4)=='C' 

          read(Nunit, *, iostat=stat) AT_string, AT_interval, AT_diff, AT_init, AT_setup
          if(stat /= 0) goto 120

          read(Nunit, *, iostat=stat) vector_scalar_str, NB
          vector_problem_fl = vector_scalar_str(1:1) == 'V' .and. &
               &              vector_scalar_str(2:2) == 'E' .and. &
               &              vector_scalar_str(3:3) == 'C' .and. &
               &              vector_scalar_str(4:4) == 'T'
          if(.not.vector_problem_fl .and. &
               &  (vector_scalar_str(1:1) /= 'S' .or. &
               &   vector_scalar_str(2:2) /= 'C' .or. &
               &   vector_scalar_str(3:3) /= 'A' .or. &
               &   NB < 0 )) goto 120

          if(stat /= 0) goto 120
          
          close(Nunit)
          
          call parameter_check()
       end if
   
    end if

    call MPI_BCAST(cycle_kind              ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(gs_kind                 ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(iteration_number_down   ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(iteration_number_up     ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(iteration_number_bottom ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(f_first_iteration_number,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(theta                   ,  1, MPI_DOUBLE_PRECISION, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(dump_jacobi_weight      ,  1, MPI_DOUBLE_PRECISION, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(dump_gauss_seidel_weight,  1, MPI_DOUBLE_PRECISION, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(mstring                 , 20, MPI_CHARACTER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(solver_string           , 20, MPI_CHARACTER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(number_of_vectors       ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(dst_string              , 20, MPI_CHARACTER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(conv_check_str          , 20, MPI_CHARACTER, 0, SOLVER_COMM, ierr)
    conv_check_fl = conv_check_str(1:1)=='C' .and. conv_check_str(2:2)=='H' .and. &
         &  conv_check_str(3:3)=='E' .and. conv_check_str(4:4)=='C' 
    call MPI_BCAST(AT_string               , 20, MPI_CHARACTER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(AT_interval             ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(AT_diff                 ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(AT_init                 ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(AT_setup                ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(vector_scalar_str       ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    call MPI_BCAST(NB                      ,  1, MPI_INTEGER, 0, SOLVER_COMM, ierr)
    vector_problem_fl = &
         & vector_scalar_str(1:1) == 'V' .and. vector_scalar_str(2:2) == 'E' .and. &
         & vector_scalar_str(3:3) == 'C' .and. vector_scalar_str(4:4) == 'T'
    if(.not.vector_problem_fl .and. &
         & (vector_scalar_str(1:1) /= 'S' .or. vector_scalar_str(2:2) /= 'C' .or. &
         &  vector_scalar_str(3:3) /= 'A'.or. NB < 0 )) goto 120
    
    return

120 write(*,*) "Error in AMG parameter file: AMG.dat"
    close(Nunit)
    stop
    
  end subroutine parameter_file_input

  subroutine parameter_string_input(x)
    implicit none
    character(*) :: x
    character(50) :: key, value
    integer :: h,t,e 

    h = index(x,"-")
    if(h>0) then
       !C X has some string like "-??? = ???"
       do 
          t = h+index(x(h+1: ), "-") - 1
          if(t<h) then
             e = h+index(x(h+1:), "=")
          else
             e = h+index(x(h+1:t), "=")
          end if

          key = trim(adjustl(x(h+1:e-1)))

          if(t<h) then
             value= trim(adjustl(x(e+1:)))
          else
             value= trim(adjustl(x(e+1:t)))
          end if

          call parameter_value_set(key, value)

          if(t<h) exit
          h=t+1
       end do
    end if

  end subroutine parameter_string_input

  subroutine print_current_parameter_set()
    use amg_parameter,  only: cycle_kind, gs_kind, iteration_number_down, iteration_number_up, &
         &                    iteration_number_bottom, f_first_iteration_number, theta,        &
         &                    theta_reduction_rate, dump_jacobi_weight,                        &
         &                    dump_gauss_seidel_weight, solver_string,                         &
         &                    number_of_vectors, conv_check_str, conv_check_fl,                &
         &                    AT_string, AT_init, AT_interval, AT_diff, AT_setup, NB,          &
         &                    vector_problem_fl, aggregation_strategy, INDPDT_AGGRE,           &
         &                    PROCESS_AGGRE, NOT_USE_AGGRE_TRANSFER, W_OUT_LU,                 &
         &                    process_aggre_threshold, max_level_size, min_node_size,          &
         &                    comm_opt_for_CGA
    use problem_matrix, only: dst_string
    use at_mod, only: at_parameter_setting, at_parameter_setting_from_file
    implicit none
    
    integer, parameter :: string_sz=1000
    character(string_sz) :: string_buf
    character(200) :: sbuf
    
    write(sbuf,*) cycle_kind
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -cycle="//sbuf
    string_buf=sbuf
    string_buf=trim(string_buf)

    if(gs_kind==1) then
       sbuf=" -smoother="//"gs"
    else if(gs_kind==0) then
       sbuf=" -smoother="//"sgs"       
    else if(gs_kind==2) then
       sbuf=" -smoother="//"mcsgs"       
    else if(gs_kind==3) then
       sbuf=" -smoother="//"mcgs"       
    else
       sbuf=" -smoother= unknown!"
    end if
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,*) iteration_number_down
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -iter_down="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,*) iteration_number_up
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -iter_up="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,*) iteration_number_bottom
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -iter_bottom="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,*) f_first_iteration_number
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -f_first_iter_down="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,'(f6.3)') theta
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -strong_connection_threshold="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,'(f6.3)') theta_reduction_rate
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -threshold_reduction_rate="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,'(f6.3)') dump_jacobi_weight
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -dump_jacobi_coef="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf,'(f6.3)') dump_gauss_seidel_weight
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -smoother_accel_coef="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    sbuf=adjustl(solver_string); sbuf=trim(sbuf)
    sbuf=" -krylov_solver="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf, *) number_of_vectors
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -restart="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    sbuf=adjustl(dst_string); sbuf=trim(sbuf)
    sbuf=" -dist_method="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    if(conv_check_fl) then
       write(sbuf, *) 1
    else
       write(sbuf, *) 0
    end if
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -convergence_check="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    if(AT_string(1:1)=='A' .and. AT_string(2:2)=='T') then
       write(sbuf, *) "AT"
    else 
       write(sbuf, *) "NOAT"
    end if
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -parameter_at="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf
    if(AT_string(1:1)=='A' .and. AT_string(2:2)=='T') then
       
       write(sbuf, *) AT_init
       sbuf=adjustl(sbuf); sbuf=trim(sbuf)
       sbuf=" -at_init="//sbuf
       string_buf = string_buf(1:len_trim(string_buf))//sbuf

       write(sbuf, *) AT_diff
       sbuf=adjustl(sbuf); sbuf=trim(sbuf)
       sbuf=" -at_diff="//sbuf
       string_buf = string_buf(1:len_trim(string_buf))//sbuf

       write(sbuf, *) AT_interval
       sbuf=adjustl(sbuf); sbuf=trim(sbuf)
       sbuf=" -at_interval="//sbuf
       string_buf = string_buf(1:len_trim(string_buf))//sbuf

       write(sbuf, *) AT_setup
       sbuf=adjustl(sbuf); sbuf=trim(sbuf)
       sbuf=" -at_setup="//sbuf
       string_buf = string_buf(1:len_trim(string_buf))//sbuf
    end if
    
    write(sbuf, *) NB
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -vector_problem="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf
    
    if(aggregation_strategy==1) then
       sbuf = "coupled_w_lu"
    else if(aggregation_strategy==2) then
       sbuf = "coupled_w_o_lu"
    else if(aggregation_strategy==3) then
       sbuf = "independent_w_lu"
    else if(aggregation_strategy==4) then
       sbuf = "independent_w_o_lu"
    else if(aggregation_strategy==5) then
       sbuf = "old_independent_w_lu"
    else if(aggregation_strategy==6) then
       sbuf = "old_independent_w_o_lu"
    end if
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -aggregation="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf, *) process_aggre_threshold
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -process_aggre_threshold="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf, *) min_node_size
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -min_node_size="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf, *) max_level_size
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -max_level_size="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    write(sbuf, *) comm_opt_for_CGA
    sbuf=adjustl(sbuf); sbuf=trim(sbuf)
    sbuf=" -comm_opt_for_CGA="//sbuf
    string_buf = string_buf(1:len_trim(string_buf))//sbuf

    if(string_sz < len_trim(string_buf)) then
       write(*,*) "Error!: string buffer is not sufficient in print_current_parameter_set()"
    end if
    
    write(*,*) string_buf(1:len_trim(string_buf))
    
  end subroutine print_current_parameter_set
    
  subroutine parameter_value_set(key, value)
    use amg_parameter,  only: cycle_kind, gs_kind, iteration_number_down, iteration_number_up, &
         &                    iteration_number_bottom, f_first_iteration_number, theta,        &
         &                    theta_reduction_rate, dump_jacobi_weight,                        &
         &                    dump_gauss_seidel_weight, solver_string,                         &
         &                    number_of_vectors, conv_check_str, conv_check_fl,                &
         &                    AT_string, AT_init, AT_interval, AT_diff, AT_setup, NB,          &
         &                    vector_problem_fl, aggregation_strategy, INDPDT_AGGRE,           &
         &                    PROCESS_AGGRE, NOT_USE_AGGRE_TRANSFER, W_OUT_LU,                 &
         &                    process_aggre_threshold, max_level_size, min_node_size,          &
         &                    comm_opt_for_CGA, iter_ups, iter_downs, theta_ary,               &
         &                    dump_jacobi_weights, dump_gauss_seidel_weights, scaling_problem_fl

    use problem_matrix, only: dst_string
    use at_mod, only: at_parameter_setting, at_parameter_setting_from_file
    implicit none

    character(50) :: key, value

    integer :: i,l
    logical :: flag

    l=len_trim(key)

    if(l==5) then 
       if(key(1:l).eq."cycle") then
!C          write(*,*)"cycle : ", trim(value)
          read(value,*) cycle_kind
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==7) then
       if(key(1:l).eq."iter_up") then
!C          write(*,*)"iter_up : ", value
          read(value,*) iteration_number_up
          do i=1,10
              iter_ups(i)=iteration_number_up
          end do

       else if(key(1:l).eq."at_diff") then
!C          write(*,*)"at_diff : ", value
          read(value,*) AT_diff

       else if(key(1:l).eq."at_init") then
!C          write(*,*)"at_init : ", value
          read(value,*) AT_init
          
       else if(key(1:l).eq."restart") then
!C          write(*,*)"restart : ", value
          read(value,*) number_of_vectors
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==8)then
       if(key(1:l).eq."smoother") then
!C          write(*,*)"smoother : ", trim(value)
          if(value(1:1)=='g' .and. value(2:2)=='s') then
             gs_kind=1
          else if(value(1:1)=='s' .and. &
               &  value(2:2)=='g' .and. &
               &  value(3:3)=='s') then
             gs_kind=0
          else if(value(1:1)=='m' .and. &
               &  value(2:2)=='c' .and. &
               &  value(3:3)=='g' .and. &
               &  value(4:4)=='s') then
             gs_kind=3
          else if(value(1:1)=='m' .and. &
               &  value(2:2)=='c' .and. &
               &  value(3:3)=='s' .and. &
               &  value(4:4)=='g' .and. &
               &  value(5:5)=='s') then
             gs_kind=2
          end if
             
       else if(key(1:l).eq."at_setup") then
!C          write(*,*)"at_setup : ", value
          read(value,*) AT_setup
       else 
          write(*,*) "### unknown option!:", key
       end if


    else if(l==9) then
       if(key(1:l).eq."iter_down") then
!C          write(*,*)"iter_down : ", value
          read(value, *) iteration_number_down
          do i=1,10
              iter_downs(i)=iteration_number_down
          end do

       else if(key(1:l).eq."iter_up_1") then
          read(value, *) iter_ups(1)
       else if(key(1:l).eq."iter_up_2") then
          read(value, *) iter_ups(2)
       else if(key(1:l).eq."iter_up_3") then
          read(value, *) iter_ups(3)
       else if(key(1:l).eq."iter_up_4") then
          read(value, *) iter_ups(4)
       else if(key(1:l).eq."iter_up_5") then
          read(value, *) iter_ups(5)
       else if(key(1:l).eq."iter_up_6") then
          read(value, *) iter_ups(6)
       else if(key(1:l).eq."iter_up_7") then
          read(value, *) iter_ups(7)
       else if(key(1:l).eq."iter_up_8") then
          read(value, *) iter_ups(8)
       else if(key(1:l).eq."iter_up_9") then
          read(value, *) iter_ups(9)
       else
          write(*,*) "### unknown option!:", key
       end if
    else if(l==10) then
       if(key(1:l).eq."iter_up_10") then
          read(value, *) iter_ups(10)
       else
          write(*,*) "### unknown option!:", key
       end if
       

    else if(l==11) then
       if(key(1:l).eq."iter_bottom") then
!C          write(*,*)"iter_bottom : ", value
          read(value,*) iteration_number_bottom
       else if(key(1:l).eq."iter_down_1") then
          read(value, *) iter_downs(1)
       else if(key(1:l).eq."iter_down_2") then
          read(value, *) iter_downs(2)
       else if(key(1:l).eq."iter_down_3") then
          read(value, *) iter_downs(3)
       else if(key(1:l).eq."iter_down_4") then
          read(value, *) iter_downs(4)
       else if(key(1:l).eq."iter_down_5") then
          read(value, *) iter_downs(5)
       else if(key(1:l).eq."iter_down_6") then
          read(value, *) iter_downs(6)
       else if(key(1:l).eq."iter_down_7") then
          read(value, *) iter_downs(7)
       else if(key(1:l).eq."iter_down_8") then
          read(value, *) iter_downs(8)
       else if(key(1:l).eq."iter_down_9") then
          read(value, *) iter_downs(9)

       else if(key(1:l).eq."at_interval") then
!C          write(*,*)"at_interval : ", value
          read(value,*) AT_interval
       else if(key(1:l).eq."dist_method") then
          read(value,*) dst_string
       else if(key(1:l).eq."aggregation") then
          if(     value(1:len_trim("coupled_w_lu")).eq."coupled_w_lu") then
             aggregation_strategy =1
             INDPDT_AGGRE          =.false.
             PROCESS_AGGRE         =.false.
             NOT_USE_AGGRE_TRANSFER=.false.
             W_OUT_LU              =.false.
          else if(value(1:len_trim("coupled_w_o_lu")).eq."coupled_w_o_lu") then
             aggregation_strategy =2
             INDPDT_AGGRE          =.false.
             PROCESS_AGGRE         =.false.
             NOT_USE_AGGRE_TRANSFER=.true.
             W_OUT_LU              =.true.
          else if(value(1:len_trim("independent_w_lu")).eq."independent_w_lu") then
#ifdef PARMETIS
             aggregation_strategy =3
             INDPDT_AGGRE          =.true.
             PROCESS_AGGRE         =.true.
             NOT_USE_AGGRE_TRANSFER=.true.
             W_OUT_LU              =.false.
#else
            write(*,*) "this aggregation strategy needs parmetis"
#endif
          else if(value(1:len_trim("independent_w_o_lu")).eq."independent_w_o_lu") then
#ifdef PARMETIS
             aggregation_strategy =4
             INDPDT_AGGRE          =.true.
             PROCESS_AGGRE         =.true.
             NOT_USE_AGGRE_TRANSFER=.true.
             W_OUT_LU              =.true.
#else
             write(*,*) "this aggregation strategy needs parmetis"
#endif
          else if(value(1:len_trim("old_independent_w_lu")).eq."old_independent_w_lu") then
             aggregation_strategy =5
             INDPDT_AGGRE          =.true.
             PROCESS_AGGRE         =.false.
             NOT_USE_AGGRE_TRANSFER=.false.
             W_OUT_LU              =.false.
          else if(value(1:len_trim("old_independent_w_o_lu")).eq."old_independent_w_o_lu") then
             aggregation_strategy =6
             INDPDT_AGGRE          =.true.
             PROCESS_AGGRE         =.false.
             NOT_USE_AGGRE_TRANSFER=.true.
             W_OUT_LU              =.true.
          else
             write(*,*) "### unknown aggregate_strategy!:", value(1:len_trim(value))
          end if
       else 
          write(*,*) "### unknown option!:", key
       end if
       
    else if(l==12) then
       if(key(1:l) .eq. "parameter_at") then
!C          write(*,*)"parameter_at : ", value
          read(value,*) i
          if(i==1) then
             AT_string="AT"
          else
             AT_string="NOAT"
          end if
       else if(key(1:l).eq."iter_down_10") then
          read(value, *) iter_downs(10)
          
       else 
          write(*,*) "### unknown option!:", key
       end if

    else if(l==13) then
       if(key(1:l).eq."krylov_solver") then
!C          write(*,*)"krylov_solver : ", value
          read(value,*) solver_string
       else if(key(1:l).eq."min_node_size") then
          read(value,*) min_node_size
       else if(key(1:l).eq."scale_problem") then
          read(value,*) i
          if(i==0) then
             scaling_problem_fl=.false.
          else if(i==1) then
             scaling_problem_fl=.true.
          else
             write(*,*) "### value for scale_problem must be 1 or 0."
          end if
          
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==14) then
       if(key(1:l).eq."vector_problem") then
!C          write(*,*)"vector_problem : ", value
          read(value,*) NB
          if(NB > 0) then
             vector_problem_fl = .true.
          else 
             vector_problem_fl = .false.
          end if
       else if(key(1:l).eq."max_level_size") then
          read(value,*) max_level_size
       else 
          write(*,*) "### unknown option!:", key
       end if

    else if(l==16) then
       if(key(1:l).eq."dump_jacobi_coef") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weight
          do i = 1, 9
              dump_jacobi_weights(i)=dump_jacobi_weight
          end do
          
       else if(key(1:l).eq."comm_opt_for_CGA") then
          read(value, *) comm_opt_for_CGA
       else 
          write(*,*) "### unknown option!:", key
       end if
       
    else if(l==17) then
       if(key(1:l).eq."convergence_check") then
!C          write(*,*)"convergence_check : ", value
          read(value,*) i
          if(i /= 0) conv_check_fl=.true.
       else if(key(1:l).eq."f_first_iter_down") then
!C          write(*,*)"f_first_iter_down : ", value
          read(value,*) f_first_iteration_number
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==18) then
       if(key(1:l).eq."dump_jacobi_coef_2") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weights(1)
       else if(key(1:l).eq."dump_jacobi_coef_3") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weights(2)
       else if(key(1:l).eq."dump_jacobi_coef_4") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weights(3)
       else if(key(1:l).eq."dump_jacobi_coef_5") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weights(4)
       else if(key(1:l).eq."dump_jacobi_coef_6") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weights(5)
       else if(key(1:l).eq."dump_jacobi_coef_7") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weights(6)
       else if(key(1:l).eq."dump_jacobi_coef_8") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weights(7)
       else if(key(1:l).eq."dump_jacobi_coef_9") then
!C          write(*,*)"dump_jacobi_coef : ", value
          read(value, *) dump_jacobi_weights(8)
       else
          write(*,*) "### unknown option!:", key
       end if



    else if(l==19) then
       if(key(1:l).eq."smoother_accel_coef") then
!C          write(*,*)"smoother_accel_coef : ", value
          read(value,*) dump_gauss_seidel_weight
          do i=1,10
              dump_gauss_seidel_weights(i)=dump_gauss_seidel_weight
          end do
       else if(key(1:l).eq."dump_jacobi_coef_10") then
          read(value, *) dump_jacobi_weights(9)
       else
          write(*,*) "### unknown option!:", key
       end if

    else if(l==20) then
       if(key(1:l).eq."strong_con_threshold") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta
       else
          write(*,*) "### unknown option!:", key
       end if
    else if(l==21) then
       if(key(1:l).eq."smoother_accel_coef_1") then
          read(value,*) dump_gauss_seidel_weights(1)
       else if(key(1:l).eq."smoother_accel_coef_2") then
          read(value,*) dump_gauss_seidel_weights(2)
       else if(key(1:l).eq."smoother_accel_coef_3") then
          read(value,*) dump_gauss_seidel_weights(3)
       else if(key(1:l).eq."smoother_accel_coef_4") then
          read(value,*) dump_gauss_seidel_weights(4)
       else if(key(1:l).eq."smoother_accel_coef_5") then
          read(value,*) dump_gauss_seidel_weights(5)
       else if(key(1:l).eq."smoother_accel_coef_6") then
          read(value,*) dump_gauss_seidel_weights(6)
       else if(key(1:l).eq."smoother_accel_coef_7") then
          read(value,*) dump_gauss_seidel_weights(7)
       else if(key(1:l).eq."smoother_accel_coef_8") then
          read(value,*) dump_gauss_seidel_weights(8)
       else if(key(1:l).eq."smoother_accel_coef_9") then
          read(value,*) dump_gauss_seidel_weights(9)
       else
          write(*,*) "### unknown option!:", key
       end if


    else if(l==22) then
       if(key(1:l).eq."smoother_accel_coef_10") then
          read(value,*) dump_gauss_seidel_weights(10)

       else if(key(1:l).eq."strong_con_threshold_2") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(1)
       else if(key(1:l).eq."strong_con_threshold_3") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(2)
       else if(key(1:l).eq."strong_con_threshold_4") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(3)
       else if(key(1:l).eq."strong_con_threshold_5") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(4)
       else if(key(1:l).eq."strong_con_threshold_6") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(5)
       else if(key(1:l).eq."strong_con_threshold_7") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(6)
       else if(key(1:l).eq."strong_con_threshold_8") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(7)
       else if(key(1:l).eq."strong_con_threshold_9") then
!C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(8)
       else
          write(*,*) "### unknown option!:", key
       end if
    else if(l==23) then
       if(key(1:l).eq."process_aggre_threshold") then
          read(value,*) process_aggre_threshold
       else if(key(1:l).eq."strong_con_threshold_10") then
          !C          write(*,*)"strong_con_threshold : ", value
          read(value,*) theta_ary(9)

       else
          write(*,*) "### unknown option!:", key
       end if
    else if(l==24) then
       if(key(1:l).eq."threshold_reduction_rate") then
          read(value,*) theta_reduction_rate
       else
          write(*,*) "### unknown option!:", key
       end if
       
    else
       write(*,*) "### unknown option!:", key
    end if
    

  end subroutine parameter_value_set


  subroutine print_amg_option_range(arg)
    implicit none

    integer :: arg
    
    if(arg == 0) then
       !C parameter range is output
       write(*,*) " "
       write(*,*) "----- parameter option list ------"       
       write(*,*) "# means a number, the inital value at each line is default setting."
       write(*,*) "-cycle:                0,1,2 :: V,F,W"
       write(*,*) "-smoother:             sgs,gs,mcsgs,mcgs"
       write(*,*) "-iter_down:            #: 1"
       write(*,*) "-iter_down_?:          #: it can set direct value for the level ?"
       write(*,*) "-iter_up:              #: 1"
       write(*,*) "-iter_up_?:            #: it can set direct value for the level ?"
       write(*,*) "-iter_bottom:          #: 30"
       write(*,*) "-f_first_iter_down:    #"
       write(*,*) "-strong_con_threshold: #: 0.1: 0.0-0.15"
       write(*,*) "-threshold_reduction_rate: #: 0.8: 0.5-1.0"
       write(*,*) "-strong_con_threshold_?: #: it can set direct value for the level ?"
       write(*,*) "-dump_jacobi_coef:     #: 0.66: 0.5-0.8"
       write(*,*) "-dump_jacobi_coef_?:   #: it can set direct value for the level ?"
       write(*,*) "-smoother_accel_coef:  #: 0.8: 0.2-1.4"
       write(*,*) "-smoother_accel_coef_?:#: it can set direct value for the level ?"
       write(*,*) "-krylov_solver:        BICGSTAB, NO, CG, RESCUT, GPBICG"
       write(*,*) "-restart:              #: 2-10"
       write(*,*) "-dist_method:          FIX, PARMETIS"
       write(*,*) "-convergence_check:    #: 1 or 0"
       write(*,*) "-parameter_at:         NOAT, AT"
       write(*,*) "-at_init:              #"
       write(*,*) "-at_diff:              #"
       write(*,*) "-at_interval:          #"
       write(*,*) "-at_setup:             #: 1 or 0"
       write(*,*) "-vector_problem:       #: 1-10: 1 => SCALAR problem"
       write(*,*) "-aggregation:   coupled_w_lu, coupled_w_o_lu"
       write(*,*) "                independent_w_lu, independent_w_o_lu"
       write(*,*) "                old_independent_w_o_lu, old_independent_w_lu"
       write(*,*) "-process_aggre_threshold: #: 1000:100-5000"
       write(*,*) "-min_node_size:        #: 100:100-700"
       write(*,*) "-max_level_size:       #: 10:2-10"
       write(*,*) "-comm_opt_for_CGA:     #: 0:1(yes) or 0(no). Communication optimization for"
       write(*,*) "                       #: coarse grid aggregation. "
       write(*,*) "-scale_problem:        #: 0(no) or 1(yes). diagonal scaling for scalar solver"
       write(*,*) " "
    else if(arg == 1) then
       !C current parameter setting
       write(*,*) " "
       write(*,*) " --- current parameter setting ---"
       call print_current_parameter_set()
       write(*,*) " "
    end if

  end subroutine print_amg_option_range

  subroutine amg_parameter_option(Nunit, SOLVER_COMM, x)
    use amg_parameter, only: topo_sz
    implicit none
    
    character(*) :: x
    integer, intent(in) :: SOLVER_COMM
    integer :: Nunit

    call parameter_string_input(x)

    call parameter_check()
    
    call deal_with_at_parameter(Nunit)

    topo_sz=0
    call topology_setting_from_file(Nunit, SOLVER_COMM)

  end subroutine amg_parameter_option

  subroutine amg_parameter_option_file(Nunit, filename, SOLVER_COMM)
    use amg_parameter, only: topo_sz
    implicit none
    include 'mpif.h'

    character(*) :: filename
    character(len=2048) :: x,xbuf
    character(len=512) :: linebuf
    integer, intent(in) :: SOLVER_COMM
    integer :: Nunit, my_rank, ierr, h
    logical :: NO_FILE_FL


    call MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr)
    NO_FILE_FL=.false.
    if(my_rank==0) then
       x = " "
       open(Nunit, FILE=filename, status='old', &
            & FORM='FORMATTED', err=122)
       do 
          read(Nunit, '(a)', end=998) linebuf
          h = index(linebuf,"#")          
          if(h>0) then
             xbuf = trim(x)//" "//trim(linebuf(1:h-1))
          else if(h==0) then
             xbuf = trim(x)//" "//trim(linebuf)
          else 
             write(*,*) "Unexpected error in amg_parameter_option_file()"
          end if
          x    = xbuf
       end do
    end if

    !C file open error
    if(.false.) then
122    NO_FILE_FL=.true.
       write(*,*) "No option file: ", filename
    end if
998 call MPI_BCAST(NO_FILE_FL,1,MPI_LOGICAL,0,SOLVER_COMM, ierr)


    IF(.not. NO_FILE_FL) then

       !C input optsion string 
       !C if(my_rank == 0) write(*,*) "input option:", trim(x)
       call MPI_BCAST(x,2048,MPI_CHARACTER,0, SOLVER_COMM, ierr)

       call parameter_string_input(x)

       call parameter_check()


       call deal_with_at_parameter(Nunit)


    end IF
    
    topo_sz=0
    call topology_setting_from_file(Nunit, SOLVER_COMM)

    
  end subroutine amg_parameter_option_file




  
  subroutine amg_parameter_set(Nunit, SOLVER_COMM)
    use amg_parameter, only: topo_sz
    implicit none
    
    integer, intent(in) :: SOLVER_COMM
    integer :: Nunit
    
    call parameter_file_input(Nunit, SOLVER_COMM)
    
    call parameter_check()
    
    call deal_with_at_parameter(Nunit)

    topo_sz=0
    call topology_setting_from_file(Nunit, SOLVER_COMM)


  end subroutine amg_parameter_set

  subroutine deal_with_at_parameter(Nunit)
    use amg_parameter, only: AT_string, AT_init, AT_interval, &
         &                   AT_diff, AT_setup
    use at_mod, only: at_parameter_setting_from_file
    implicit none
    
    integer :: Nunit

    if(AT_string(1:1)=='A' .and. AT_string(2:2)=='T') then
       if(AT_init <= 0) then
          at_flag = .true.
          AT_init = -1
       end if
       
       !C inter_val < 0 means offline tuning.
       if(AT_interval <= 0) then
          offline_fl =.true.
       else
          offline_fl =.false.
       end if

       !C all processes must call at_parameter_setting().
       !C setupflag_for_AT and difference_of_iter are set.
       if(AT_setup == 0) then
          call at_parameter_setting_from_file(AT_interval, AT_diff, Nunit, .false.)
       else if(AT_setup == 1) then
          call at_parameter_setting_from_file(AT_interval, AT_diff, Nunit, .true.)
       end if
    else
       !C if AT_string is not equal to AT, then AT_init is reset to negative value.
       AT_init = -1
    end if
    
  end subroutine deal_with_at_parameter

  subroutine topology_setting_from_file(Nunit, SOLVER_COMM)
    implicit none
    include 'mpif.h'
    integer, intent(in) :: Nunit 
    integer(kind=kint), parameter :: linenum=100
    integer :: my_rank, ierr, SOLVER_COMM
    integer(kind=kint) :: cxyz(4, linenum),i,j

    call MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr)
    if(my_rank == 0) then
       open(Nunit, File='topology.dat', status='old', &
            & Form='FORMATTED', err=123)
       do i = 1, linenum
          read(Nunit, *, end=999,err=999) &
               & cxyz(1,i),cxyz(2,i),cxyz(3,i),cxyz(4,i)
       end do
999    close(Nunit)
    end if

    if(.false.) then
123   write(*,*) "no topology setting file."
      i=1
    end if
 
    call MPI_BCAST(i,1,MPI_INTEGER,0,SOLVER_COMM,ierr)
    
    do j=1,i-1
       call MPI_BCAST(cxyz(1,j),4,MPI_INTEGER,0,SOLVER_COMM,ierr)
#ifdef PRINT_REZ       
       if(my_rank==0) write(*,*)"C,X,Y,Z:",cxyz(1:4,j)
#endif
       call suggest_topology(cxyz(1,j),cxyz(2,j), &
            & cxyz(3,j), cxyz(4,j), SOLVER_COMM)
    end do
    
    return
    
  end subroutine topology_setting_from_file
  
  subroutine suggest_topology(C,X,Y,Z, SOLVER_COMM)
    use amg_parameter, only: max_topo_sz, topo_sz, &
         & topo_C, topo_X, topo_Y, topo_Z
    implicit none
    integer(kind=kint), intent(in) :: C,X,Y,Z
    integer, intent(in) :: SOLVER_COMM
    integer :: ierr, my_rank
    
    integer(kind=kint) :: PETOT
    
    
    call MPI_COMM_SIZE(SOLVER_COMM, PETOT,   ierr)
    call MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr)
    
    if(PETOT == C*X*Y*Z) then
       if(topo_sz < max_topo_sz) then
          topo_sz = topo_sz+1
          topo_C(topo_sz)=C
          topo_X(topo_sz)=X
          topo_Y(topo_sz)=Y
          topo_Z(topo_sz)=Z

       else
          if(my_rank==0) write(*,*) "max # of topologies are", max_topo_sz
       end if
    else
       if(my_rank==0) write(*,*) "C*X*Y*Z must be equal to # of processes."
    end if

  end subroutine suggest_topology

!!$  subroutine reset_B_X(B, X, X_bak)
!!$    USE problem_matrix
!!$    USE solver_SR2
!!$
!!$    implicit none
!!$    real(kind=kreal)   :: B(:), X(:), w, X_bak(:)
!!$    real(kind=kreal), allocatable :: WS(:), WR(:)
!!$    integer(kind=kint) :: c,i,j,js, je, esize, isize
!!$
!!$    !
!!$    ! 1.need to be added solver_sendrecv X.. 6/9
!!$    !
!!$    
!!$    if(NEIBPETOT > 0) then
!!$       esize = STACK_EXPORT(NEIBPETOT)
!!$       isize = STACK_IMPORT(NEIBPETOT)
!!$       allocate(WS(esize),WR(isize))
!!$       call SOLVER_SEND_RECV2(NP, esize, isize, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
!!$            & STACK_EXPORT, NOD_EXPORT, WS, WR, X, SOLVER_COMM, my_rank)
!!$       deallocate(WS, WR)
!!$    end if
!!$    
!!$    !
!!$    ! 2.NP -> N
!!$    !
!!$
!!$    do i = 1, N
!!$       w = B(i)
!!$       
!!$       js = INL(i-1)+1
!!$       je = INL(i)
!!$       do j = js, je
!!$          c = IAL(j)
!!$          w = w-AL(j)*X(c)
!!$       end do
!!$       
!!$       w = w-D(i)*X(i)
!!$       
!!$       js = INU(i-1)+1
!!$       je = INU(i)
!!$       do j = js, je
!!$          c = IAU(j)
!!$          w = w-AU(j)*X(c)
!!$       end do
!!$       B(i) = w
!!$    end do
!!$!
!!$    do i = 1, N
!!$       X_bak(i) = X(i)
!!$    end do
!!$    X=0
!!$
!!$  end subroutine reset_B_X
  
end module parameter_at
