!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

MODULE problem_matrix
  IMPLICIT NONE
  INCLUDE 'precision.inc'
  
  INTEGER(kind=kint) :: NB=1, N, NP, NPL, NPU, LEVEL_NUM = 0, NK =0
  INTEGER(kind=kint), pointer :: INL(:), INU(:), IAL(:), IAU(:)
  REAL(kind=kreal), pointer   :: AL(:), AU(:), D(:)
  REAL(kind=kreal), pointer   :: scale_D(:)=>null()
  REAL(kind=kreal), pointer   :: AL_NB(:,:,:), AU_NB(:,:,:), D_NB(:,:,:)
  REAL(kind=kreal), pointer   :: k_or_q_vec(:,:) => null()

  INTEGER(kind=kint) :: NEIBPETOT, SOLVER_COMM, PETOT, my_rank
  INTEGER(kind=kint), pointer :: NEIBPE(:), NOD_IMPORT(:), NOD_EXPORT(:)
  INTEGER(kind=kint), pointer :: STACK_IMPORT(:), STACK_EXPORT(:)

  !C for reverse ordering 
  INTEGER(kind=kint), pointer :: order(:), nds_index(:)
  !C for reordering 
  INTEGER(kind=kint), pointer :: order_for_dst(:), nds_index_aft_dst(:)

  !C Before BLOCKING vector sizes
  INTEGER(kind=kint) :: N_BEFORE_BL, NP_BEFORE_BL
  
  character(len=20), save :: dst_string = "FIX"
  
  !C 
  LOGICAL :: REORDERED_FL = .false., BLOCKING_FL = .false.
  LOGICAL :: SCALING_FL=.false.

  
  private initialize
  private store_problem_matrix_N_NP
  
contains
  
  subroutine blocking_problem_matrix()  
    use blocking_mat

    integer(kind=kint) :: N_o, NP_o, NEIBPETOT_o, NPL_o, NPU_o
    real   (kind=KREAL), pointer :: D_NB_o(:,:,:)
    integer(kind=kint ), pointer :: INU_o(:), INL_o(:)
    integer(kind=kint ), pointer :: IAU_o(:), IAL_o(:)
    real   (kind=KREAL), pointer :: AU_NB_o(:,:,:), AL_NB_o(:,:,:)
    integer(kind=kint ), pointer :: NOD_IMPORT_o(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_o(:)
    integer(kind=kint ), pointer :: NEIBPE_o(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_o(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_o(:)


    BLOCKING_FL =  .true.
    N_BEFORE_BL = N
    NP_BEFORE_BL = NP

    !C the following data is deallocated in this sub.
    !C NOD_IMPORT, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, NEIBPE
    !C INL, IAL, INU, IAU, AL, AU, D

    call blocking_data( &
         & N, NP, NPL, NPU, D, AL, INL, IAL,        &
         & AU, INU, IAU, NEIBPETOT, NEIBPE,                  &
         & STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,  &
         & N_o, NP_o, NPL_o, NPU_o, D_NB_o, AL_NB_o, INL_o, IAL_o, AU_NB_o, INU_o, IAU_o,  &
         & NEIBPETOT_o, NEIBPE_o, STACK_IMPORT_o, NOD_IMPORT_o, &
         & STACK_EXPORT_o, NOD_EXPORT_o,  &
         & my_rank, SOLVER_COMM, NB &
         )
    
    N   = N_o
    NP  = NP_o 
    NPL = INL_o(N_o)
    NPU = INU_o(N_o)
    D_NB   => D_NB_o 
    INU => INU_o 
    INL => INL_o 
    IAU => IAU_o 
    IAL => IAL_o 
    AU_NB  => AU_NB_o 
    AL_NB  => AL_NB_o
    NEIBPETOT    =  NEIBPETOT_o 
    NEIBPE       => NEIBPE_o 
    STACK_IMPORT => STACK_IMPORT_o
    STACK_EXPORT => STACK_EXPORT_o
    NOD_IMPORT   => NOD_IMPORT_o 
    NOD_EXPORT   => NOD_EXPORT_o

  end subroutine blocking_problem_matrix
  
  
!C  subroutine matrix_distribute()
  subroutine distribute_problem_matrix()
    use matrix_dist_mtom
    use count_time_mod
    implicit none
    include 'mpif.h' 


    !C dummy vector
    real(kind=kreal), pointer :: T(:)
    
    integer(kind=kint) :: N_o, NP_o, NEIBPETOT_o, NB_o
    real   (kind=KREAL), pointer :: D_o(:), T_o(:), D_NB_o(:,:,:)
    integer(kind=kint ), pointer :: INU_o(:), INL_o(:)
    integer(kind=kint ), pointer :: IAU_o(:), IAL_o(:)
    real   (kind=KREAL), pointer :: AU_o(:), AL_o(:), AU_NB_o(:,:,:), AL_NB_o(:,:,:)
    integer(kind=kint ), pointer :: NOD_IMPORT_o(:)
    integer(kind=kint ), pointer :: NOD_EXPORT_o(:)
    integer(kind=kint ), pointer :: NEIBPE_o(:)
    integer(kind=kint ), pointer :: STACK_IMPORT_o(:)
    integer(kind=kint ), pointer :: STACK_EXPORT_o(:)
    
    real   (kind=KREAL), pointer :: k_or_q_vec2(:,:)

    integer(kind=kint ), pointer :: order_o(:)
    
    REAL(kind=kfloat),  allocatable :: tpwgts(:) 
    integer(kind=kint) :: nparts,ierr, dst_method
    
    integer(kind=kint) :: i,j

    call count_time(1, SOLVER_COMM, my_rank, 14)
    
    !C nds_index_aft_dst, order_for_dst is needed for reordering X and B

    REORDERED_FL=.true.
    

    
    nparts = PETOT
    if(.not.associated(nds_index_aft_dst)) allocate(nds_index_aft_dst(0:PETOT))
    if(associated(order_for_dst))     deallocate(order_for_dst)
    allocate(tpwgts(nparts), order_for_dst(NP))

    tpwgts = 1.0/nparts
    
    if(dst_string(1:1)=='F' .and. dst_string(2:2)=='I'.and.dst_string(3:3)=='X') then
       dst_method = 3
#ifdef PRINT_REZ    
       if(my_rank == 0) write(*,*) "Ordering is Fixed in distribution process"
#endif
    else if(dst_string(1:1)=='P' .and. dst_string(2:2)=='A'.and.dst_string(3:3)=='R') then
      dst_method = 1
#ifdef PRINT_REZ    
      if(my_rank == 0) write(*,*) "ParMetis ordering is used in distribution process"
#endif
   else
      write(*,*) "unknown matrix distribution method is set."
      write(*,*) dst_string(1:3)
      stop
   end if
   
   if(BLOCKING_FL .eqv. .false.) then
      allocate(T(NP))
      CALL dist_mtom_main                                                   &
           & ( N, NP, D, INU, INL, IAU, IAL, AU,                            &
           &   AL, T, STACK_IMPORT, STACK_EXPORT, NOD_IMPORT,               &
           &   NOD_EXPORT, NEIBPE, NEIBPETOT, N_o, NP_o, D_o, INU_o, INL_o, &
           &   IAU_o, IAL_o, AU_o, AL_o, T_o, STACK_IMPORT_o,               &
           &   STACK_EXPORT_o, NOD_IMPORT_o, NOD_EXPORT_o, NEIBPE_o,        &
           &   NEIBPETOT_o, SOLVER_COMM, my_rank, PETOT, order, order_o,    &
           &   nparts, tpwgts, nds_index_aft_dst, order_for_dst, dst_method)

      DEALLOCATE(NOD_IMPORT, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, NEIBPE)
      DEALLOCATE(INL, IAL, INU, IAU, AL, AU, D, T, T_o)
      DEALLOCATE(order, tpwgts)
      
      D   => D_o 
      AU  => AU_o 
      AL  => AL_o

   else
      allocate(T(NP*NB))
      CALL dist_mtom_main_vec                                               &
           & ( N, NP, NB, D_NB, INU, INL, IAU, IAL, AU_NB, AL_NB, T,        &
           &   STACK_IMPORT, STACK_EXPORT, NOD_IMPORT, NOD_EXPORT,          &
           &   NEIBPE, NEIBPETOT, N_o, NP_o, NB_o, D_NB_o, INU_o, INL_o,    &
           &   IAU_o, IAL_o, AU_NB_o, AL_NB_o, T_o, STACK_IMPORT_o,         &
           &   STACK_EXPORT_o, NOD_IMPORT_o, NOD_EXPORT_o, NEIBPE_o,        &
           &   NEIBPETOT_o, SOLVER_COMM, my_rank, PETOT, order, order_o,    &
           &   nparts, tpwgts, nds_index_aft_dst, order_for_dst, dst_method)

      DEALLOCATE(NOD_IMPORT, NOD_EXPORT, STACK_IMPORT, STACK_EXPORT, NEIBPE)
      DEALLOCATE(INL, IAL, INU, IAU, AL_NB, AU_NB, D_NB, T, T_o)
      DEALLOCATE(order, tpwgts)

      D_NB   => D_NB_o 
      AU_NB  => AU_NB_o 
      AL_NB  => AL_NB_o

      !C k_or_q_vec redistribution
      if(associated(k_or_q_vec)) then
         allocate(k_or_q_vec2(NB_o*N_o,NK))
         !C k_or_q_vec -> k_or_q_vec2
         DO i = 1, NK
            call distribute_vector(nds_index_aft_dst, order_for_dst, SOLVER_COMM,k_or_q_vec(:,i),&
                 &                 N, k_or_q_vec2(:,i), NB)
         END DO
         deallocate(k_or_q_vec)
         k_or_q_vec => k_or_q_vec2
      end if
   end if
   
   N   = N_o
   NP  = NP_o 
   NPL = INL_o(N_o)
   NPU = INU_o(N_o)
   INU => INU_o 
   INL => INL_o 
   IAU => IAU_o 
   IAL => IAL_o 
   NEIBPETOT    =  NEIBPETOT_o 
   NEIBPE       => NEIBPE_o 
   STACK_IMPORT => STACK_IMPORT_o
   STACK_EXPORT => STACK_EXPORT_o
   NOD_IMPORT   => NOD_IMPORT_o 
   NOD_EXPORT   => NOD_EXPORT_o
   order        => order_o
   

   

   call count_time(2, SOLVER_COMM, my_rank, 14)

 end subroutine distribute_problem_matrix

  
  subroutine initialize()
    implicit none
    nullify(order)
    nullify(nds_index)
    nullify(order_for_dst)
    nullify(nds_index_aft_dst)
  end subroutine initialize


  !C store_kernel_vectors must be called before distribute problem matrix
  subroutine store_kernel_vectors(kernel_vecs, len, NumK)
    implicit none
    INCLUDE  'precision.inc'
    !C k_or_q_vec(N*NB, NK)
    integer(kind=kint), intent(in) :: len, NumK
    real(kind=kreal), intent(in)   :: kernel_vecs(:,:)
    integer(kind=kint) :: i,j,k
    integer(kind=kint),save :: init=0
    
    if(init==0) then
       !C initialize k_or_q_vec(:,:) at first.
       nullify(k_or_q_vec)
       init=1
    end if

    if(REORDERED_FL) then
       write(*,*) "amgs_store_kernel_vecs: kernel vectors must be registered &
            & before matrix re-distribution."
       stop
    end if
    
    if(associated(k_or_q_vec)) deallocate(k_or_q_vec)
    allocate(k_or_q_vec(len, NumK))

    NK = NumK
    
    k_or_q_vec = 0
    do j = 1, NK
       do i = 1, len
          k_or_q_vec(i,j) = kernel_vecs(i,j)
       end do
    end do
    
  end subroutine store_kernel_vectors
  

  subroutine store_problem_matrix_N_NP(crs_index, crs_column, crs_val, rowsize,  vectorsize, MPI_COMM)
    use matrix_dist_mtom
    implicit none

    REAL(kind=kreal),   intent(in) :: crs_val(:)
    INTEGER(kind=kint), intent(in) :: crs_column(:), crs_index(0:)
    INTEGER(kind=kint), intent(in) :: rowsize, vectorsize, MPI_COMM
    REAL(kind=kreal)   :: v
    INTEGER(kind=kint) :: i,j,il,iu,row,col,ierr
    
    N = rowsize; NP = vectorsize; NPL=0; NPU=0
    do row = 1,N
       do j = crs_index(row-1)+1, crs_index(row)
          col = crs_column(j)
          if(col .ne. row) then
             if(col < row) then
                NPL=NPL+1
             else
                NPU=NPU+1
             end if
          end if
       end do
    end do

    allocate(INL(0:N),INU(0:N),IAU(NPU),IAL(NPL),AU(NPU),AL(NPL),&
         & D(NP))
    !C debug
!!$    allocate(INL(0:NP),INU(0:NP),IAU(NPU),IAL(NPL),AU(NPU),AL(NPL),&
!!$         & D(NP))

    il=0; iu=0
    INL(0)=0; INU(0)=0
    do row=1,N
       do j = crs_index(row-1)+1,crs_index(row)
          col = crs_column(j)
          v = crs_val(j)
          if( col == row ) then
             D(row) = v
          else if(col < row) then
             il=il+1; IAL(il)=col; AL(il) = v            
          else 
             iu=iu+1; IAU(iu)=col; AU(iu) = v
          end if
       end do
       INL(row) = il; INU(row) = iu

    end do


    !C  communication tables are copied
    SOLVER_COMM = MPI_COMM
    CALL MPI_COMM_SIZE(SOLVER_COMM, PETOT, ierr) 
    CALL MPI_COMM_RANK(SOLVER_COMM, my_rank, ierr) 
    
    
  end subroutine store_problem_matrix_N_NP

  subroutine store_problem_matrix(crs_index, crs_column, crs_val, rowsize, MPI_COMM)
    use amg_parameter, only: vector_problem_fl, amg_parameter_NB => NB, scaling_problem_fl
    use matrix_dist_mtom, only: set_order
    implicit none
    
    REAL(kind=kreal), intent(in)   :: crs_val(:)
    INTEGER(kind=kint), intent(in) :: crs_column(:), crs_index(0:), rowsize, MPI_COMM
    
    INTEGER(kind=kint) :: vectorsize,SISIZE,SESIZE
    INTEGER(kind=kint) :: i,j

    call initialize()

    REORDERED_FL=.false.; BLOCKING_FL = .false.
    vectorsize = rowsize
    call store_problem_matrix_N_NP(crs_index, crs_column, crs_val, rowsize, vectorsize, MPI_COMM)
    
    !C 1PE is running.
    NEIBPETOT = 0
    ALLOCATE (NEIBPE(NEIBPETOT))
    !C GPU solver uses STACK_??PORT(1:).
    ALLOCATE (STACK_IMPORT(0:1))
    ALLOCATE (STACK_EXPORT(0:1))

    STACK_IMPORT(0)=0
    STACK_EXPORT(0)=0
    SISIZE= STACK_IMPORT(NEIBPETOT)
    SESIZE= STACK_EXPORT(NEIBPETOT)
    ALLOCATE (NOD_IMPORT(SISIZE))
    ALLOCATE (NOD_EXPORT(SESIZE))

    if(vector_problem_fl) then
       NB = amg_parameter_NB
       if(.not. associated(k_or_q_vec)) NK = NB
       
       call blocking_problem_matrix()
    end if

    !C record ordering
    allocate(order(N), nds_index(0:PETOT))
    call set_order(nds_index, order, N, MPI_COMM)

    if(.not.vector_problem_fl .and. scaling_problem_fl) then
       allocate(scale_D(NP))
       call scale_mat(N,NP,D,INL,INU,IAL,IAU,AL,AU,scale_D)
       SCALING_FL=.true.
    end if

  end subroutine store_problem_matrix
  
  !C distributed matrix... : allocated data?,  pointer?
  subroutine store_distributed_problem_matrix(crs_index, crs_column, crs_val, rowsize, &
       & vectorsize, neiberingPEsize, neiberingPEs, index_in_table,           &
       & index_out_table, in_table, out_table, MPI_COMM)
    use amg_parameter, only: vector_problem_fl, amg_parameter_NB => NB, scaling_problem_fl
    use solver_SR2, only: solver_send_recv2
    IMPLICIT NONE
    INCLUDE 'mpif.h'
  
    REAL(kind=kreal),   intent(in) :: crs_val(:)
    INTEGER(kind=kint), intent(in) :: crs_column(:), crs_index(0:)
    INTEGER(kind=kint), intent(in) :: index_in_table(0:), index_out_table(0:)
    INTEGER(kind=kint), intent(in) :: in_table(:), out_table(:), neiberingPEs(:)
    INTEGER(kind=kint) :: rowsize, vectorsize, neiberingPEsize,  MPI_COMM, ierr
    REAL(kind=kreal), pointer :: WS(:), WR(:)
    
    REAL(kind=kreal) :: v
    INTEGER(kind=kint) :: i,j,il,iu,row,col,is,os

    call initialize()

    REORDERED_FL=.false.; BLOCKING_FL = .false.
    call store_problem_matrix_N_NP(crs_index, crs_column, crs_val, rowsize, vectorsize, MPI_COMM)

    !C communication tables are copied
    is = 0; os = 0;
    if(neiberingPEsize > 0) then
       is = index_in_table(neiberingPEsize)
       os = index_out_table(neiberingPEsize)
    end if
    NEIBPETOT = neiberingPEsize

    allocate(NEIBPE(NEIBPETOT), STACK_IMPORT(0:NEIBPETOT),            &
         & STACK_EXPORT(0:NEIBPETOT), NOD_IMPORT(is), NOD_EXPORT(os))
    
    do i = 1, is 
       NOD_IMPORT(i) = in_table(i)
    end do
    do i = 1, os 
       NOD_EXPORT(i) = out_table(i)
    end do
    do i = 1, NEIBPETOT
       NEIBPE(i) = neiberingPEs(i)
       STACK_IMPORT(i) = index_in_table(i)
       STACK_EXPORT(i) = index_out_table(i)
    end do
    STACK_IMPORT(0) = 0
    STACK_EXPORT(0) = 0

    if(vector_problem_fl) then
       NB = amg_parameter_NB
       if(.not. associated(k_or_q_vec)) NK = NB
       
       call blocking_problem_matrix()
    end if
    
    if(.not.vector_problem_fl .and. scaling_problem_fl) then
       allocate(WS(NP),WR(NP))
       call SOLVER_SEND_RECV2                                     &
            & ( NP, NP, NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
            &                                  STACK_EXPORT, NOD_EXPORT,&
            &          WS, WR, D, SOLVER_COMM, my_rank)
       deallocate(WS,WR)
       allocate(scale_D(NP))
       call scale_mat(N,NP,D,INL,INU,IAL,IAU,AL,AU,scale_D)
       SCALING_FL=.true.
    end if

  end subroutine store_distributed_problem_matrix
  
  subroutine scale_mat(N,NP,D,INL,INU, IAL, IAU, AL, AU, scale_D)
    implicit none
    REAL   (kind=kreal), POINTER :: D  (:)
    REAL   (kind=kreal), POINTER :: AU (:)
    REAL   (kind=kreal), POINTER :: AL (:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    REAL   (kind=kreal), POINTER :: scale_D(:)
    INTEGER(kind=kint)  :: NP, N, i,j,k
    
    !$omp parallel do
    DO i = 1, NP
       scale_D(i) = dsqrt(dabs(D(i)))

       !C this part deals unassigned diagonal element case.
       if(scale_D(i) == 0) then
          scale_D(i) = 1.0
       end if
    END DO
    !$omp end parallel do


    !$omp parallel do private(j, k)
    DO i = 1, N
       DO j = INL(i-1)+1, INL(i)
          k = IAL(j)
          AL(j) = AL(j)/(scale_D(i)*scale_D(k))
       END DO
       DO j = INU(i-1)+1, INU(i)
          k = IAU(j)
          AU(j) = AU(j)/(scale_D(i)*scale_D(k))
       END DO
       if(D(i) > 0) then
          D(i) = 1.0
       else if(D(i) < 0) then
          D(i) = -1.0
       end if
    END DO
    !$omp end parallel do

    !$omp parallel do
    DO i = N+1, NP
       if(D(i) > 0) then
          D(i) = 1.0
       else if(D(i)<0) then
          D(i) = -1.0
       end if
    END DO
    !$omp end parallel do 
  end subroutine scale_mat
  
  subroutine display_problem_mat()
    implicit none
    integer(kind=kint) :: i,j,k
    DO i = 1, N
       write(*,*) "-------",i,"-------"
       DO j=INL(i-1)+1,INL(i)
          write(*,*) IAL(j),AL(j)
       END DO
       write(*,*) i,D(i)
       DO j=INU(i-1)+1,INU(i)
          write(*,*) IAU(j),AU(j)
       END DO
    END DO
  end subroutine display_problem_mat
end module problem_matrix
