
#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

module process_aggregation
contains
  
  subroutine decide_parallelism(size_of_aggregates, coarser_lev_parallelism, &
       & SOLVER_COMM, process_aggre_threshold)
    implicit none
    include 'precision.inc'
    include 'mpif.h'

    integer(kind=kint), intent(in)  :: size_of_aggregates, SOLVER_COMM
    integer(kind=kint), intent(out) :: coarser_lev_parallelism
    
    integer(kind=kint) :: active, sndbuf(2), rcvbuf(2), ierr
    integer(kind=kint) :: process_aggre_threshold

    if(size_of_aggregates > 0) then
       active=1
    else
       active=0
    end if
    
    sndbuf(1)=size_of_aggregates; sndbuf(2)=active

    call MPI_ALLREDUCE(sndbuf, rcvbuf, 2, MPI_INTEGER, MPI_SUM, SOLVER_COMM, ierr) 
    
    if(rcvbuf(2)==0) then
       !C no process has aggregates!.. This case should stop the solver!
       coarser_lev_parallelism = -1
    else if(rcvbuf(1) / rcvbuf(2) > process_aggre_threshold) then
       coarser_lev_parallelism = -1
    else
       coarser_lev_parallelism = rcvbuf(1)/process_aggre_threshold+1
    end if
    
  end subroutine decide_parallelism


  
  subroutine decide_send_recv_allocation(size_of_aggregates, NEIBPETOT, NEIBPE, &
       & coarser_lev_parallelism, sender_fl, start_nd_num, partner, receiver_fl, &
       & sum_nd_num, SOLVER_COMM, my_rank, NPROCS, k_or_q_vec_c, NK)
    implicit none
    include 'precision.inc'
    include 'mpif.h'

    integer(kind=kint), intent(in) :: NPROCS, NEIBPETOT, NEIBPE(:), SOLVER_COMM
    integer(kind=kint), intent(in) :: size_of_aggregates
    integer(kind=kint), intent(inout) :: coarser_lev_parallelism
    REAL(kind=kreal), pointer :: k_or_q_vec_c(:,:)
    REAL(kind=kreal), pointer :: k_or_q_vec_c_new(:,:)
    integer(kind=kint), intent(in) :: NK
    integer(kind=kint) ::  start_nd_num, partner
    integer(kind=kint)             :: sum_nd_num

    integer(kind=kint), pointer    :: xadj(:), adjncy(:)
    integer(kind=kint), pointer    :: vtxdist(:)
    integer(kind=kint), pointer    :: vwgt(:), adjwgt(:), part(:)
    integer(kind=kint)             :: wgtfl, numfl, options(0:2), edgecut, ncon
    REAL   (kind=kfloat)           :: ubvec

    integer(kind=kint)             :: nparts
    REAL   (kind=kfloat), pointer  :: tpwgts(:)

    integer(kind=kint) :: my_rank, i, j, k, l, dist_N_i, ierr, ret
    integer(kind=kint) :: my_grp, min_rank_of_my_grp, sum_aggre_my_grp

    logical :: sender_fl, receiver_fl

    integer(kind=kint), pointer :: share_aggresize(:), share_part(:)

    INTEGER(kind=kint), ALLOCATABLE :: req(:)
    INTEGER(kind=kint), ALLOCATABLE :: sta(:,:)
    INTEGER(kind=kint), ALLOCATABLE :: recv_index(:)
    real(kind=kreal), ALLOCATABLE   :: buf(:)

    integer(kind=kint) :: inum, neib



    sender_fl = .false.
    receiver_fl = .false.

    sum_nd_num = 0
    start_nd_num = 0
    partner = -1

    if(coarser_lev_parallelism > 0 ) then

       !C   I:  create graph with (V,E). V is size_of_aggregates on each PE, 
       !C       E is connection between processes with NEIBPE(:)

       !C for symmetricity, even if size_of_aggregates == 0, adjncy array is set.

       !C-- xadj make
       allocate(xadj(0:1))
       xadj(0:1)=0
       xadj(1)=NEIBPETOT

       !C-- adjncy make
       allocate(adjncy(0:NEIBPETOT-1))
       adjncy(0:NEIBPETOT-1)=NEIBPE(1:NEIBPETOT)
          
       !C-- vtxdist make
       allocate(vtxdist(0:NPROCS))
       vtxdist(0)=0
       do i=1, NPROCS
          vtxdist(i)=i
       end do

       !C-- make nparts, tpwgts(:)
       nparts=coarser_lev_parallelism
       allocate(tpwgts(nparts))
       tpwgts=1.0/nparts

       !C-- default setting
       options(0)=0
       numfl=0
       ubvec=1.05
       ncon=1

       !C-- make dist_N_i, part(:)
       dist_N_i = vtxdist(my_rank+1) - vtxdist(my_rank)
       allocate(part(0:dist_N_i - 1))

       !C-- make wgtfl, vwgt, adjwgt
       wgtfl=2
       allocate(vwgt(0:dist_N_i-1))
       allocate(adjwgt(1))
       vwgt(0) = size_of_aggregates

       !C  II:  call ParMETIS with coarser_lev_parrallelism
       !C ex.
       !C CALL PARMETIS_V3_PARTKWAY(vtxdist, xadj, adjncy, vwgt, adjwgt,  &
       !C & wgtfl, numfl, ncon, nparts, tpwgts, ubvec, options,  &
       !C & edgecut, part, SOLVER_COMM)
       
       part(0) = my_rank
#ifdef PARMETIS
       CALL PARMETIS_V3_PARTKWAY(vtxdist, xadj, adjncy, vwgt, adjwgt,  &
            & wgtfl, numfl, ncon, nparts, tpwgts, ubvec, options,  &
            & edgecut, part, SOLVER_COMM, ret)

#endif       
       if(ret > 0) stop 'ParMETIS stops! in process_aggregation'
       
#ifdef PRINT_REZ
       if(size_of_aggregates > 0) write(*,*) my_rank,":", size_of_aggregates, " > ", part(:)
#endif
       
       !C III:  allgather domain id and size_of_aggregates.
       !C       Then select the ranks with the same domain id of its own process

       !C 各プロセスの持つアグリゲート数を全プロセスで共有する
       allocate(share_aggresize(0:NPROCS-1))
       share_aggresize(0:NPROCS-1) = 0
       call MPI_ALLGATHER(size_of_aggregates, 1, MPI_INTEGER, share_aggresize(0), 1, MPI_INTEGER, SOLVER_COMM, ierr)

       !       if(my_rank .eq. 0) print *, "share_aggresize:", share_aggresize
       !       call MPI_Barrier(SOLVER_COMM, ierr)

       !C プロセスを纏めるグループ番号を共有する
       allocate(share_part(0:NPROCS-1))
       share_part(0:NPROCS-1) = 0 
       if(size_of_aggregates .eq. 0) part(0) = -1
       call MPI_ALLGATHER(part(0), 1, MPI_INTEGER, share_part(0), 1, MPI_INTEGER, SOLVER_COMM, ierr)

       !       if(my_rank .eq. 0) print *, "share_part:", share_part
       !       call MPI_Barrier(SOLVER_COMM, ierr)

       !C 各プロセスにsender_fl, receiver_flを設定する
       if(size_of_aggregates .ne. 0) then

          coarser_lev_parallelism = 0
          do i = 0, NPROCS-1
             if(coarser_lev_parallelism < share_part(i)) coarser_lev_parallelism = share_part(i)
          end do

!!$         allocate(min_rank(0:coarser_lev_parallelism))
          min_rank_of_my_grp = NPROCS+100

          my_grp = share_part(my_rank)
          do i = 0, my_rank-1
             if(share_part(i) .eq. my_grp) exit
          end do

          if(i < my_rank) then
             !C other smaller rank is in my_grp
             min_rank_of_my_grp = i
          else
             !C no other smaller rank is in my_grp
             do i = my_rank+1, NPROCS-1
                if(share_part(i) .eq. my_grp) exit
             end do
             if(i < NPROCS) then
                !C my group has other processes
                min_rank_of_my_grp = my_rank
                receiver_fl = .true.
             end if
          end if

          !C if min_rank_of_my_grp > NPROCS then it has only one process of my_rank
          if(min_rank_of_my_grp < NPROCS .and. .not.receiver_fl) then
             sender_fl = .true.
          end if

          !C 各ランクのパラメータ(sum_nd_num, start_nd_num, partner)の値を設定する
          sum_aggre_my_grp = 0

          neib = 0
          do i = 0, NPROCS-1
             if(share_part(i) == my_grp) then
                neib = neib+1
                if(my_rank .eq. i) start_nd_num = sum_aggre_my_grp
                sum_aggre_my_grp = sum_aggre_my_grp + share_aggresize(i)
             endif
          enddo


          
          if(sender_fl) then
             partner = min_rank_of_my_grp
          else if(receiver_fl) then
             sum_nd_num = sum_aggre_my_grp
          endif

          !C---- k_or_q_vec_c
          if(associated(k_or_q_vec_c)) then
             if(receiver_fl) then
                !C the number of processes in the same group without receiver rank
                neib = neib-1
             
                !C: copy k_or_q_vec_c to k_or_q_vec_c_new
                allocate(k_or_q_vec_c_new(sum_nd_num*NK+1,NK), req(neib), sta(MPI_STATUS_SIZE, neib))

!C                !$omp parallel do private(i,j)
                do i = 1, size_of_aggregates*NK
                   do j = 1, NK
                      k_or_q_vec_c_new(i,j) = k_or_q_vec_c(i,j)
                   end do
                end do
!C                !$omp end parallel do
                
                deallocate(k_or_q_vec_c)
                k_or_q_vec_c => k_or_q_vec_c_new

                
                !C: receive
                !C recv_index records the number of aggregates.
                allocate(buf(sum_nd_num*NK*NK),recv_index(0:neib))
             

                j = 1
                k = 0
                recv_index(0) = 0 
                do i = my_rank+1, NPROCS-1
                   if(share_part(i) == my_grp) then
                      k = k+1
                      recv_index(k) = recv_index(k-1)+share_aggresize(i)
                      call MPI_IRECV(buf(j), share_aggresize(i)*NK*NK, MPI_DOUBLE_PRECISION, i, 0, &
                           & SOLVER_COMM, req(k), ierr)
                      j = j+share_aggresize(i)*NK*NK
                   endif
                enddo
                call MPI_WAITALL(neib, req, sta, ierr)
             
                !C assign buf(:) to k_or_q_vec_new
                !C !$omp parallel private(i,j,inum,l,k)
                do i = 1, neib
                   j = recv_index(i-1)+1
                   inum = recv_index(i)-recv_index(i-1)
!C                   !$omp do
                   do k = 1, NK
                      do l = 1, inum*NK
                         k_or_q_vec_c(size_of_aggregates*NK+(j-1)*NK+l,k) = &
                              &                                     buf((j-1)*NK*NK+(k-1)*inum*NK+l)
                      end do
                   end do
                  !C !$omp end do
                end do
!C                !$omp end parallel
                
                deallocate(buf,req,sta,recv_index)
             
             else if(sender_fl) then
                allocate(buf(size_of_aggregates*NK*NK), req(1), sta(MPI_STATUS_SIZE,1))
!C                !$omp parallel do private(i,j)
                do j = 1, NK
                   do i = 1, size_of_aggregates*NK
                      buf(i+(j-1)*size_of_aggregates*NK) = k_or_q_vec_c(i,j)
                   end do
                end do
!C                !$omp end parallel do
                
                call MPI_ISEND(buf(1), size_of_aggregates*NK*NK, MPI_DOUBLE_PRECISION, partner, 0, &
                     & SOLVER_COMM, req, ierr)
                !C k_or_q_vec_c is not needed.., because sender PE's aggregates is transferred to partner's domain.
                deallocate(k_or_q_vec_c)
                
                call MPI_WAIT(req, sta(:,1), ierr)
                deallocate(buf,req,sta)
             end if
          
          endif

       !C ==================== OUTPUT ==================== !C

       ! if(my_rank .eq. 0) print *, "########## sender_fl, receiver_fl  ##########"
       ! call MPI_Barrier(SOLVER_COMM, ierr)
       ! print *, "my_rank:", my_rank, "sender_fl:", sender_fl, "receiver_fl:", receiver_fl
       ! call MPI_Barrier(SOLVER_COMM, ierr)
       ! if(my_rank .eq. 0) print *, "########## receiver_fl = .true. ##########"
       ! call MPI_Barrier(SOLVER_COMM, ierr)
       ! print *, "my_rank:", my_rank, "sum_nd_num:", sum_nd_num
       ! call MPI_Barrier(SOLVER_COMM, ierr)
       ! if(my_rank .eq. 0) print *, "########## sender_fl = .true. ##########"
       ! call MPI_Barrier(SOLVER_COMM, ierr)
       ! print *, "my_rank:", my_rank, "start_nd_num:", start_nd_num
       ! call MPI_Barrier(SOLVER_COMM, ierr)
       ! print *, "my_rank:", my_rank, "partner:", partner

       !C ================================================ !C

       end if
    end if
  end subroutine decide_send_recv_allocation

  
  subroutine set_GIN_aggre_and_aggre_tbl(size_of_aggregates, sender_fl, start_nd_num, &
       & partner, receiver_fl, sum_nd_num, GIN_aggregate, NPROCS, aggregate_number_in_table, &
       & aggregate_table_array, aggregate_table_size, PE_list, PE_list_size,&
       & my_rank, SOLVER_COMM, LEVEL_NO, PRO_AGGRE_OPT, vector_problem_fl)
    use data_structure_for_AMG,    only: HIERARCHICAL_DATA
    use data_structure_for_AMG_NB, only: HIERARCHICAL_DATA_NB
    use aggregate_mod, only: determine_owner_index_frm_gl, assign_initial_aggre_for_sender
    use int_vec_module, &
         & only: int_vec, int_vec_initialize, check_and_extend, deallocate_int_vec
    
    
    implicit none
    include 'precision.inc'
    include 'mpif.h'
    

    logical, intent(in) :: vector_problem_fl

    logical :: sender_fl, receiver_fl, PRO_AGGRE_OPT
    integer(kind=kint) :: start_nd_num, partner, sum_nd_num
    integer(kind=kint) :: GIN_aggregate(0:), NPROCS, aggregate_table_size
    type(int_vec), pointer :: aggregate_table_array(:)
    INTEGER(kind=kint),intent(out) :: aggregate_number_in_table(:)
    INTEGER(kind=kint),intent(in) :: my_rank, SOLVER_COMM, LEVEL_NO
    INTEGER(kind=kint),intent(in) :: size_of_aggregates
    INTEGER(kind=kint) :: PE_list_size, ierr
    INTEGER(kind=kint) :: PE_list(:, :)
    
    INTEGER(kind=kint) :: i, index, j, g, l
    INTEGER(kind=kint), allocatable :: export_int_lvl_buffer(:)
    INTEGER(kind=kint), pointer :: newNOD_EXPORT(:) 
    


    !C I size_of_aggregates is sent from sender to receiver
    !C II shift the number of nodes.
    if(sender_fl) then
#ifdef PRINT_REZ
       !C write(*,*) "sender:", my_rank, "to", partner
#endif
       j = 0
       CALL MPI_ALLGATHER(j,1,MPI_INTEGER, &
            & GIN_aggregate(1),1,MPI_INTEGER,SOLVER_COMM,ierr)
    else if(receiver_fl) then
       CALL MPI_ALLGATHER(sum_nd_num,1,MPI_INTEGER, &
            & GIN_aggregate(1),1,MPI_INTEGER,SOLVER_COMM,ierr)
    else 
       CALL MPI_ALLGATHER(size_of_aggregates,1,MPI_INTEGER, &
            & GIN_aggregate(1),1,MPI_INTEGER,SOLVER_COMM,ierr)
    end if
    
    GIN_aggregate(0)=0
    DO i = 2,NPROCS
       GIN_aggregate(i)=GIN_aggregate(i-1)+GIN_aggregate(i)
    END DO

    aggregate_number_in_table = 0
    
    !C III aggregates should be recorded on the aggregate table of partner.   
    if(sender_fl .and. size_of_aggregates > 0) then
       
       CALL determine_owner_index_frm_gl(GIN_aggregate(partner)+1, GIN_aggregate, PE_list, index, &
            & PE_list_size, NPROCS, my_rank)
       
       do i = 1, size_of_aggregates
          !C this assume all aggregates are not overlapped, and all aggregates are recorded on aggregate table.
          call  assign_initial_aggre_for_sender(aggregate_number_in_table, aggregate_table_array, &
               & aggregate_table_size, i+GIN_aggregate(partner)+start_nd_num, i, index)
          
          !C aggregates info is already recorded on share_aggresize(:, aggregate_number), and node_index(:).
       end do
    end if

    if(PRO_AGGRE_OPT) then
       !C--- make inter level communication table for comm optimization
       j = 0
       if(sender_fl) then
          allocate(export_int_lvl_buffer(size_of_aggregates))
          if(vector_problem_fl) then
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(size_of_aggregates))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0:1))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0:1))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(1))
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0) = 0          
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(1) = 0          
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0) = 0
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(1) = size_of_aggregates
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(1) = partner
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT = 1

             do i = 1, size_of_aggregates
                g = i+GIN_aggregate(partner)+start_nd_num
                l = i
                export_int_lvl_buffer(i) = g
                HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(i) = l
             end do

          else
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(size_of_aggregates))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0:1))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0:1))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(1))
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0) = 0          
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(1) = 0          
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0) = 0
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(1) = size_of_aggregates
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(1) = partner
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT = 1

             do i = 1, size_of_aggregates
                g = i+GIN_aggregate(partner)+start_nd_num
                l = i
                export_int_lvl_buffer(i) = g
                HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(i) = l
             end do
          end if

       else if(receiver_fl) then
          allocate(export_int_lvl_buffer(0))
          if(vector_problem_fl) then
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(0))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0:NPROCS))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0:NPROCS))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(NPROCS))
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0) = 0          
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0) = 0
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT = 0

          else
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(0))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0:NPROCS))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0:NPROCS))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(NPROCS))
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0) = 0          
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0) = 0
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT = 0
          end if

       else
          allocate(export_int_lvl_buffer(0))
          if(vector_problem_fl) then
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(0))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0:0))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0:0))
             allocate(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(0))
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0) = 0          
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0) = 0
             HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT = 0

          else
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(0))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0:0))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0:0))
             allocate(HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(0))
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0) = 0          
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0) = 0
             HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT = 0
          end if
          
       end if

       if(vector_problem_fl) then
          !C allocate newNOD_EXPORT and update NEIBPE and STACK_EXPORT
          CALL make_export_NOD &
               & (SOLVER_COMM, export_int_lvl_buffer, GIN_aggregate, my_rank, NPROCS,   &
               &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT,              &
               &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPE,                 &
               &  newNOD_EXPORT,                                                        &
               &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT,           &
               &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT)
          HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NOD_EXPORT => newNOD_EXPORT
       else
          !C allocate newNOD_EXPORT and update NEIBPE and STACK_EXPORT
          CALL make_export_NOD &
               & (SOLVER_COMM, export_int_lvl_buffer, GIN_aggregate, my_rank, NPROCS,&
               &  HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT,           &
               &  HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPE,              &
               &  newNOD_EXPORT,                                                     &
               &  HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT,        &
               &  HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT)
          HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NOD_EXPORT => newNOD_EXPORT
       end if
          
       !C--- make inter level communication table
       
       ! write(*,*) my_rank,"NEIBPETOT:",HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT
       ! i=HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT
       ! if(i>0) then
       !    write(*,*) my_rank,"NEIBPE:",            HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(1:i)
       !    write(*,*) my_rank,"STACK_EXPORT:",      HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0:i)
       !    j = HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(i)
       !    if(j>0) write(*,*) my_rank,"NOD_EXPORT:",HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NOD_EXPORT(1:j)    
       !    write(*,*) my_rank,"STACK_IMPORT:",      HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0:i)    
       !    j = HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(i)
       !    if(j>0) write(*,*) my_rank,"NOD_IMPORT:",HIERARCHICAL_DATA(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(1:j)    
       ! end if
       
    end if

  end subroutine set_GIN_aggre_and_aggre_tbl

  

  SUBROUTINE make_export_NOD (SOLVER_COMM, export_global_buffer,            &
       &     GIN_aggregate, my_rank, NPROCS, NEIBPETOT, NEIBPE, NOD_EXPORT, &
       &     STACK_IMPORT, STACK_EXPORT)

    IMPLICIT NONE
    include 'precision.inc'
    INCLUDE 'mpif.h'
    INTEGER(kind=kint), INTENT(in)     :: NPROCS
    INTEGER(kind=kint), INTENT(inout)  :: NEIBPETOT
!!$      INTEGER(kind=kint), INTENT(inout)  :: NEIBPE(NPROCS)
    INTEGER(kind=kint), INTENT(inout)  :: NEIBPE(:)
    INTEGER(kind=kint), INTENT(inout)  :: STACK_IMPORT(0:), STACK_EXPORT(0:)
    INTEGER(kind=kint), POINTER        :: NOD_EXPORT(:)

    INTEGER(kind=kint),  INTENT(in)    :: export_global_buffer(:)
    INTEGER(kind=kint), ALLOCATABLE    :: pes(:), in_recv_buf(:), recv_buf(:)

    INTEGER(kind=kint) :: pes_s(NPROCS), pes_r(NPROCS)

    INTEGER(kind=kint), INTENT(in)     :: GIN_aggregate(0:)

    INTEGER(kind=kint ) :: ierr
    INTEGER(kind=kint ), INTENT(in) :: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in) :: my_rank
    INTEGER(kind=kint), ALLOCATABLE :: req1(:), req2(:)
    INTEGER(kind=kint), ALLOCATABLE :: sta1(:,:), sta2(:,:)

    INTEGER(kind=kint) :: i, j, k, dest, inum, istart, recv_buf_size, message_size, l
    INTEGER(kind=kint) :: GLOBAL, LOCAL, base_index, n_pes, neib, from_pe, msg_sz
    PARAMETER(GLOBAL=1,LOCAL=2)
    LOGICAL :: hash_flag


    neib = NEIBPETOT


    !C-- determine number of recv
    pes_s = -1
    DO i = 1, neib
       dest = NEIBPE(i)
       pes_s(dest + 1)     = STACK_IMPORT(i) - STACK_IMPORT(i - 1)

#ifdef DEBUG
       IF(pes_s(dest+1)==0) STOP "0 size tables in make_export_NOD()"
#endif

    END DO
    pes_r = -1


    CALL MPI_ALLTOALL(pes_s, 1, MPI_INTEGER, pes_r, 1, MPI_INTEGER, &
         &            SOLVER_COMM, ierr)

    !C-- determine number of recv
    !C RECEIVE
    n_pes = 0
    ALLOCATE(in_recv_buf(0:NPROCS))
    ALLOCATE(pes(NPROCS))
    in_recv_buf(0) = 0
    DO i = 0, NPROCS - 1
       from_pe = i
       msg_sz  = pes_r(i + 1)
       IF(msg_sz >= 0) THEN
          n_pes              = n_pes + 1
          pes(n_pes)         = from_pe
          in_recv_buf(n_pes) = in_recv_buf(n_pes - 1) + msg_sz
       END IF
    END DO

    ALLOCATE(req1(neib))    
    ALLOCATE(sta1(MPI_STATUS_SIZE, neib))
    ALLOCATE(req2(n_pes))
    ALLOCATE(sta2(MPI_STATUS_SIZE, n_pes))    
    recv_buf_size = in_recv_buf(n_pes)
    ALLOCATE(recv_buf(recv_buf_size))
    ALLOCATE(NOD_EXPORT(recv_buf_size))
    DO i=1, n_pes
       istart = in_recv_buf(i - 1)
       inum   = in_recv_buf(i) - istart
       dest   = pes(i)
       CALL MPI_IRECV(recv_buf(istart + 1), inum, MPI_INTEGER, dest, 11, &
            &         SOLVER_COMM, req2(i), ierr)
    END DO

    !C SEND
    DO i = 1, neib
       istart = STACK_IMPORT(i - 1)
       inum   = STACK_IMPORT(i) - istart
       dest = NEIBPE(i)

       CALL MPI_ISEND(export_global_buffer(istart + 1), inum, MPI_INTEGER, dest, 11, & 
            &         SOLVER_COMM, req1(i), ierr)
    END DO

    IF(n_pes > 0) CALL MPI_WAITALL (n_pes, req2, sta2, ierr)
    IF(neib  > 0) CALL MPI_WAITALL (neib,  req1, sta1, ierr)

    base_index = GIN_aggregate(my_rank)    
    STACK_EXPORT(0) = 0
    DO i = 1, neib
       dest = NEIBPE(i)

       DO j = 1, n_pes
          if(pes(j) == dest) exit
       END DO

       IF(j > n_pes) THEN
          !C there is no nodes to send.
          STACK_EXPORT(i) = STACK_EXPORT(i - 1)
       ELSE
          STACK_EXPORT(i) = STACK_EXPORT(i - 1) + in_recv_buf(j) - in_recv_buf(j - 1)
          l = STACK_EXPORT(i - 1) - in_recv_buf(j - 1)
          DO k = in_recv_buf(j - 1) + 1, in_recv_buf(j)
             NOD_EXPORT(l + k) = recv_buf(k) - base_index - ZERO_ORIGIN

#ifdef DEBUG
             IF(pes(j)==-1) STOP "error!!! in make_export_table"
             IF(NOD_EXPORT(k+l)+ZERO_ORIGIN<0 .OR.NOD_EXPORT(k+l)+ZERO_ORIGIN>GIN_aggregate(my_rank+1)-base_index) THEN
                WRITE(*,*) "unexpected error in make_export_nod():", NOD_EXPORT(k+l)+ZERO_ORIGIN,my_rank
                STOP
             END IF
#endif            

          END DO
          !C mark which means un-needed 
          pes(j) = -1
       END IF
    END DO

    DO i = 1, n_pes
       IF(pes(i) /= -1) THEN
          dest = pes(i)
          NEIBPETOT = NEIBPETOT + 1
          NEIBPE(NEIBPETOT) = dest
          STACK_IMPORT(NEIBPETOT) = STACK_IMPORT(NEIBPETOT - 1)
          STACK_EXPORT(NEIBPETOT) = STACK_EXPORT(NEIBPETOT - 1) + in_recv_buf(i) - in_recv_buf(i - 1)
          l = STACK_EXPORT(NEIBPETOT - 1) - in_recv_buf(i - 1)
          DO k = in_recv_buf(i - 1) + 1, in_recv_buf(i)
             NOD_EXPORT(k + l) = recv_buf(k) - base_index-ZERO_ORIGIN

#ifdef DEBUG
             IF(NOD_EXPORT(k+l)+ZERO_ORIGIN<0 .OR. NOD_EXPORT(k+l)+ZERO_ORIGIN>GIN_aggregate(my_rank+1)-base_index) THEN
                WRITE(*,*) "unexpected error in make_export_nod():" ,NOD_EXPORT(k+l)+ZERO_ORIGIN,my_rank
                STOP
             END IF
#endif            
          END DO
       END IF
    END DO
    DEALLOCATE(pes)    
    DEALLOCATE(in_recv_buf)    
    DEALLOCATE(recv_buf)    

    DEALLOCATE(req1)
    DEALLOCATE(req2)
    DEALLOCATE(sta2)
    DEALLOCATE(sta1)
  END SUBROUTINE make_export_NOD
    
  

end module process_aggregation
