module residual_calculation
contains
  subroutine cal_relative_res_norm(res_norm, X, B, WS, WR, WSIZE)
    use solver_SR2      
    use problem_matrix
    use terminal_condition, only: BNRM2
    !$ USE omp_lib
    implicit none
    include 'mpif.h'


    REAL(kind=kreal),intent(out) :: res_norm
    REAL(kind=kreal) :: WVAL, res_norm0, X(:), B(:), WS(:), WR(:)
    INTEGER(kind=kint) :: isU,ieU,isL,ieL, i,j,inod, ierr,WSIZE


!!$    call SOLVER_SEND_RECV2                                       &
!!$         &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT,&
!!$         &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
!!$         &     WS(1:WSIZE), WR(1:WSIZE), X(1:NP) , SOLVER_COMM,  &
!!$         &     my_rank)

    call SOLVER_SEND_RECV2                                       &
         &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT,&
         &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
         &     WS, WR, X, SOLVER_COMM,  &
         &     my_rank)


    res_norm0 = 0.0

    !$omp parallel do private(WVAL, isU, ieU, isL, ieL, i, inod) reduction(+:res_norm0)
    do j= 1, N
       WVAL= B(j) - D(j) * X(j)
       isU= INU(j-1) + 1
       ieU= INU(j  ) 
       do i= isU, ieU
          inod= IAU(i)
          WVAL= WVAL - AU(i) * X(inod)
       enddo

       isL= INL(j-1) + 1
       ieL= INL(j  ) 
       do i= isL, ieL
          inod= IAL(i)
          WVAL= WVAL - AL(i) * X(inod)
       enddo
       !C res(1:N)
       res_norm0 = res_norm0+WVAL*WVAL
    enddo
    !$omp end parallel do

    CALL MPI_ALLREDUCE(res_norm0, res_norm, 1, MPI_DOUBLE_PRECISION, &
         &             MPI_SUM, SOLVER_COMM, ierr)

    res_norm = dsqrt(res_norm/BNRM2)

  end subroutine cal_relative_res_norm


  subroutine cal_relative_res_norm_NB(res_norm, X, B, WS, WR, WSIZE, NB)
    use solver_SR_NB      
    use problem_matrix, only: N,NP,AL_NB,AU_NB,INL,INU,IAL,IAU,D_NB, &
         & NOD_EXPORT, NOD_IMPORT, STACK_EXPORT, STACK_IMPORT, NEIBPE,&
         & NEIBPETOT,my_rank,SOLVER_COMM

    use terminal_condition, only: BNRM2
    !$ USE omp_lib
    implicit none
    include 'mpif.h'
    include 'precision.inc'


    REAL(kind=kreal),intent(out) :: res_norm
    INTEGER(kind=kint), intent(in) :: NB
    REAL(kind=kreal) :: WVAL(NB), res_norm0, X(:), B(:), WS(:), WR(:)
    INTEGER(kind=kint) :: i,j,k,l,col, ierr,WSIZE


    call SOLVER_SEND_RECV_NB                                      &
         &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT,&
         &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
         &     WS, WR, X, SOLVER_COMM,  &
         &     my_rank, NB)


    res_norm0 = 0.0

    !$omp parallel do private(WVAL, i,j,k,l,col) reduction(+:res_norm0)
    do i= 1, N
       do j = 1, NB
          WVAL(j) = B(NB*(i-1)+j)
          do k = 1, NB
             WVAL(j) = WVAL(j) - D_NB(j,k,i)*X(NB*(i-1)+k)
          end do
       end do
       
       do l = INL(i-1)+1, INL(i)
          col = IAL(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AL_NB(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do l = INU(i-1)+1, INU(i)
          col = IAU(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AU_NB(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do
       
       do j = 1, NB
          res_norm0 = res_norm0 + WVAL(j)**2
       end do
    end do
    !$omp end parallel do

    CALL MPI_ALLREDUCE(res_norm0, res_norm, 1, MPI_DOUBLE_PRECISION, &
         &             MPI_SUM, SOLVER_COMM, ierr)

    res_norm = dsqrt(res_norm/BNRM2)
    
  end subroutine cal_relative_res_norm_NB


  subroutine cal_res_vec_NB(res, X, B)
    use problem_matrix, only: NP, NEIBPETOT, STACK_IMPORT, STACK_EXPORT, NB
    implicit none
    include 'precision.inc'
    
    integer(kind=kint) :: WSIZE
    real(kind=kreal) :: res(:), X(:), B(:)
    real(kind=kreal), allocatable :: WS(:), WR(:)
    
    
    WSIZE = STACK_EXPORT(NEIBPETOT)
    if(WSIZE < STACK_IMPORT(NEIBPETOT)) WSIZE = STACK_IMPORT(NEIBPETOT)
    
    allocate(WS(WSIZE*NB), WR(WSIZE*NB))
    
    call cal_res_vec_NB_W(res, X, B, WS, WR, WSIZE, NB)
    
    deallocate(WS, WR)
  end subroutine cal_res_vec_NB

  
  subroutine cal_res_vec_NB_W(res, X, B, WS, WR, WSIZE, NB)
    use solver_SR_NB 
    use problem_matrix, only: N,NP,AL_NB,AU_NB,INL,INU,IAL,IAU,D_NB, &
         & NOD_EXPORT, NOD_IMPORT, STACK_EXPORT, STACK_IMPORT, NEIBPE,&
         & NEIBPETOT,my_rank,SOLVER_COMM
    use terminal_condition, only: BNRM2
    !$ USE omp_lib
    implicit none
    include 'mpif.h'
    include 'precision.inc'

    INTEGER(kind=kint), intent(in) :: NB
    REAL(kind=kreal),intent(out) :: res(:)
    REAL(kind=kreal) :: WVAL(NB), X(:), B(:), WS(:), WR(:)
    INTEGER(kind=kint) :: i,j,k,l,col, ierr,WSIZE


    call SOLVER_SEND_RECV_NB                                      &
         &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT,&
         &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
         &     WS, WR, X, SOLVER_COMM,  &
         &     my_rank, NB)

    
    !$omp parallel do private(WVAL, i,j,k,l,col) 
    do i= 1, N
       do j = 1, NB
          WVAL(j) = B(NB*(i-1)+j)
          do k = 1, NB
             WVAL(j) = WVAL(j) - D_NB(j,k,i)*X(NB*(i-1)+k)
          end do
       end do
       
       do l = INL(i-1)+1, INL(i)
          col = IAL(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AL_NB(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do l = INU(i-1)+1, INU(i)
          col = IAU(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AU_NB(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do
       
       do j = 1, NB
          res((i-1)*NB+j) = WVAL(j)
       end do
    end do
    !$omp end parallel do

  end subroutine cal_res_vec_NB_W


  subroutine cal_res_vec(res, X, B)
    use solver_SR2      
    use problem_matrix
    use terminal_condition, only: BNRM2
    !$ USE omp_lib
    implicit none
    include 'mpif.h'


    REAL(kind=kreal), intent(out) :: res(:)
    REAL(kind=kreal), allocatable :: WS(:), WR(:)
    REAL(kind=kreal) :: WVAL, X(:), B(:)
    INTEGER(kind=kint) :: isU,ieU,isL,ieL, i,j,inod, ierr,WSIZE
    INTEGER(kind=kint) :: esize, isize


    
    if(NEIBPETOT > 0) then
       esize = STACK_EXPORT(NEIBPETOT)
       isize = STACK_IMPORT(NEIBPETOT)
       allocate(WS(esize), WR(isize))
       call SOLVER_SEND_RECV2                                       &
            &   ( NP, esize, isize, NEIBPETOT, NEIBPE, STACK_IMPORT,&
            &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
            &     WS(1:esize), WR(1:isize), X(1:NP) , SOLVER_COMM,  &
            &     my_rank)
       deallocate(WS, WR)
    end if



    !$omp parallel do private(WVAL, isU, ieU, isL, ieL, i, inod) 
    do j= 1, N
       WVAL= B(j) - D(j) * X(j)
       isU= INU(j-1) + 1
       ieU= INU(j  ) 
       do i= isU, ieU
          inod= IAU(i)
          WVAL= WVAL - AU(i) * X(inod)
       enddo

       isL= INL(j-1) + 1
       ieL= INL(j  ) 
       do i= isL, ieL
          inod= IAL(i)
          WVAL= WVAL - AL(i) * X(inod)
       enddo
       
       res(j) = WVAL
    enddo
    !$omp end parallel do

    
  end subroutine cal_res_vec



end module residual_calculation
