!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

  MODULE solver
    include 'precision.inc'
    logical :: REACH_CONV_FL
    real(kind=kreal) :: resid, solver_time
    
  CONTAINS
    SUBROUTINE conv_state(CONV_FL)
      logical, intent(out) :: CONV_FL
      
      CONV_FL = REACH_CONV_FL
    end SUBROUTINE conv_state

    
    
    SUBROUTINE SWITCH_SOLVER(X_x, B_x, iter, SETUP_FL)
      USE problem_matrix
      USE amg_parameter, only: solver_string, number_of_vectors, gs_kind, theta, cycle_kind, &
           & dump_gauss_seidel_weight, iteration_number_down, vector_problem_fl
      use common_mod, only: clear_multi_lev_matrix, setup_phase
      use solver_AMG  
      use solver_AMGCG
      use solver_AMGCG_NB
      use solver_AMG_NB
      use solver_AMGBiCGSTAB
      use solver_AMGBiCGSTAB_NB
      use solver_AMGGPBiCG
      use solver_AMGGPBiCG_NB
      use solver_AMGRESCUT
      use solver_tuning_eval
      use terminal_condition, only: BNRM2_set, terminal_cond_check, BNRM2
      use residual_calculation, only: cal_res_vec_NB, cal_res_vec, cal_relative_res_norm
#ifdef GPU            
      use solver_GPU
#endif

      implicit none
      include  'mpif.h'

      integer(kind=kint), intent(inout) :: iter
      logical, intent(in)              :: SETUP_FL
      real(kind=kreal), intent(inout) :: B_x(:), X_x(:)
      integer(kind=kint) :: gpu_sol
      real(kind=kreal), allocatable :: res_vec(:)
!!$      real(kind=kreal)                 :: resid
      
      integer(kind=kint) :: i, j, WSIZE, ierr
      real(kind=kreal) :: v
      real(kind=kreal) :: res_norm
      real(kind=kreal),allocatable :: WS(:),WR(:)
      
      real(kind=kreal) :: X_x_norm, X_x_norm_gl

      
      !C-- BNRM2 in terminal_cond.F90 is set
      call BNRM2_set(B_x, SOLVER_COMM, N*NB)
      allocate(res_vec(N*NB))
      if(vector_problem_fl) then
         call cal_res_vec_NB(res_vec, X_x, B_x)
      else
         call cal_res_vec(res_vec, X_x, B_x)
      end if
      call terminal_cond_check(REACH_CONV_FL, resid, res_vec, N*NB, SOLVER_COMM)

      deallocate(res_vec)
      if(REACH_CONV_FL) then
         iter = 0
#ifdef PRINT_REZ
         if(my_rank == 0) write(*,*) "reach convergence without calling solver"
#endif         
         return
      end if
      

      if(SETUP_FL) then
         if(LEVEL_NUM > 0) call clear_multi_lev_matrix(LEVEL_NUM)
      else
         if(LEVEL_NUM == 0) then
            call setup_phase()
         end if
      end if


      if(vector_problem_fl) then
         if(solver_string(1:1)=='N' .and. solver_string(2:2)=='O') then
#ifdef PRINT_REZ    
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            CALL AMG_NB(N, NP,  NPL, NPU, D_NB,  AL_NB, INL, IAL,     &
                 & AU_NB, INU, IAU, B_x, X_x, RESID, ITER,            &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,          &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL, NB,     &
                 & k_or_q_vec, NK)
            
         else if(solver_string(1:1)=='C' .and. solver_string(2:2)=='G') then

#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            CALL AMGCG_NB(N, NP,  NPL, NPU, D_NB,  AL_NB, INL, IAL,   &
                 & AU_NB, INU, IAU, B_x, X_x, RESID, ITER,            &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,          &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL, NB,     &
                 & k_or_q_vec, NK)

         else if(solver_string(1:1)=='B' .and. solver_string(2:2)=='I' .and. &
              &  solver_string(3:3)=='C' .and. solver_string(4:4)=='G' .and. &
              &  solver_string(5:5)=='S' .and. solver_string(6:6)=='T') then
#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            call AMGBiCGSTAB_NB(N, NP,  NPL, NPU,                  &
                 &   D_NB,  AL_NB, INL, IAL, AU_NB, INU, IAU,      &
                 &   B_x,  X_x, resid, iter, my_rank, NEIBPETOT,   &
                 & NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, &
                 & NOD_EXPORT, SOLVER_COMM, PETOT, SETUP_FL,       &
                 & LEVEL_NUM, REACH_CONV_FL, NB, k_or_q_vec, NK)

         else if(solver_string(1:1)=='G' .and. solver_string(2:2)=='P' .and. &
              &  solver_string(3:3)=='B' .and. solver_string(4:4)=='I' .and. &
              &  solver_string(5:5)=='C' .and. solver_string(6:6)=='G') then
            
            CALL AMGGPBiCG_NB(N, NP,  NPL, NPU, D_NB,  AL_NB, INL, IAL,   &
                 & AU_NB, INU, IAU, B_x, X_x, RESID, ITER,                &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,              &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM,     &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL, NB,         &
                 & k_or_q_vec, NK)
            
         end if
      else

         
         if(solver_string(1:1)=='N' .and. solver_string(2:2)=='O') then
#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            CALL AMG(N, NP,  NPL, NPU, D,  AL, INL, IAL,              &
                 & AU, INU, IAU, B_x, X_x, RESID, ITER,               &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,          &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)

         else if(solver_string(1:1)=='C' .and. solver_string(2:2)=='G') then
#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            CALL AMGCG(N, NP,  NPL, NPU, D,  AL, INL, IAL,            &
                 & AU, INU, IAU, B_x, X_x, RESID, ITER,               &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,          &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)

         else if(solver_string(1:1)=='B' .and. solver_string(2:2)=='I' .and. &
              &  solver_string(3:3)=='C' .and. solver_string(4:4)=='G' .and. &
              &  solver_string(5:5)=='S' .and. solver_string(6:6)=='T') then
#ifdef PRINT_REZ
            if(my_rank==0) then

               call print_parameter()
            end if
#endif
            v = 0
            DO i = 1, N
               v = v+B_x(i)
            END DO



            call AMGBiCGSTAB(N, NP,  NPL, NPU,                     &
                 &   D,  AL, INL, IAL, AU, INU, IAU,               &
                 &   B_x,  X_x, resid, iter, my_rank, NEIBPETOT,   &
                 & NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, &
                 & NOD_EXPORT, SOLVER_COMM, PETOT, SETUP_FL,       &
                 & LEVEL_NUM, REACH_CONV_FL)

         else if(solver_string(1:1)=='R' .and. solver_string(2:2)=='E' .and. &
              &  solver_string(3:3)=='S' .and. solver_string(4:4)=='C' .and. &
              &  solver_string(5:5)=='U' .and. solver_string(6:6)=='T') then
#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            call AMGRESCUT                                                   &
                 &                 (number_of_vectors, N, NP,  NPL, NPU,     &
                 &                  D,  AL, INL, IAL, AU, INU, IAU,          &
                 &                  B_x,  X_x,                               &
                 &                  RESID,  ITER,                            &
                 &                  my_rank, NEIBPETOT, NEIBPE,              &
                 &                  STACK_IMPORT, NOD_IMPORT,                &
                 &                  STACK_EXPORT, NOD_EXPORT,                &
                 &                  SOLVER_COMM, PETOT, SETUP_FL, LEVEL_NUM, &
                 &                  REACH_CONV_FL )     

         else if(solver_string(1:1)=='G' .and. solver_string(2:2)=='P' .and. &
              &  solver_string(3:3)=='B' .and. solver_string(4:4)=='I' .and. &
              &  solver_string(5:5)=='C' .and. solver_string(6:6)=='G') then
#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif
            
            CALL AMGGPBiCG(N, NP,  NPL, NPU, D,  AL, INL, IAL,        &
                 & AU, INU, IAU, B_x, X_x, RESID, ITER,               &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,          &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)
            
         else if(solver_string(1:1)=='T' .and. solver_string(2:2)=='U' .and. &
              &  solver_string(3:3)=='N' .and. solver_string(4:4)=='I' .and. &
              &  solver_string(5:5)=='N' .and. solver_string(6:6)=='G') then
#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            CALL TUNING_EVAL(N, NP,  NPL, NPU, D,  AL, INL, IAL,      &
                 & AU, INU, IAU, B_x, X_x, RESID, ITER,               &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,          &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL)
#ifdef GPU            
         else if(solver_string(1:1)=='G' .and. solver_string(2:2)=='P' .and. &
              &  solver_string(3:3)=='U' .and. solver_string(4:4)=='B' .and. &
              &  solver_string(5:5)=='I' .and. solver_string(6:6)=='C') then
            
            gpu_sol = 1

#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            
            CALL solver_on_GPU (N, NP,  NPL, NPU, D,  AL, INL, IAL,   &
                 & AU, INU, IAU, B_x, X_x, RESID, ITER,               &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,          &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL, gpu_sol,&
                 & dump_gauss_seidel_weight)

            

         else if(solver_string(1:1)=='G' .and. solver_string(2:2)=='P' .and. &
              &  solver_string(3:3)=='U' ) then
            gpu_sol = 0

#ifdef PRINT_REZ
            if(my_rank==0) then
               call print_parameter()
            end if
#endif

            CALL solver_on_GPU (N, NP,  NPL, NPU, D,  AL, INL, IAL,   &
                 & AU, INU, IAU, B_x, X_x, RESID, ITER,               &
                 & my_rank, NEIBPETOT, NEIBPE, STACK_IMPORT,          &
                 & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
                 & PETOT, SETUP_FL, LEVEL_NUM, REACH_CONV_FL, gpu_sol,&
                 & dump_gauss_seidel_weight)
#endif
            
         else
            if(my_rank == 0) write(*,*) "No solver is specified by ", solver_string 
         end if
         
      end if

    END SUBROUTINE SWITCH_SOLVER

    SUBROUTINE print_parameter()
      USE amg_parameter, only: solver_string, number_of_vectors, gs_kind, &
           & theta, theta_reduction_rate, cycle_kind,                     &
           & dump_gauss_seidel_weight, iteration_number_down,             &
           & iteration_number_up, iteration_number_bottom,                &
           & vector_problem_fl, aggregation_strategy,                     &
           & process_aggre_threshold, min_node_size, max_level_size,      &
           & dump_jacobi_weight
      implicit none
      
      write(*,*) "Krylov solver::",solver_string, " restart::", number_of_vectors
      write(*,*)"Smoother 0:sgs 1:gs 2 mcsgs 3:mcgs ::", gs_kind
      write(*,*)"iter down:", iteration_number_down, &
           &  "iter up:", iteration_number_up, "iter bottom:", iteration_number_bottom,     &
           &  "Smoother acceleration coeff. :: ", dump_gauss_seidel_weight
      write(*,*) "CYCLE: 0:V,1:F,2:W :: ",  cycle_kind
      write(*,*) "dump jacobi weight:", dump_jacobi_weight
      write(*,*) "Strong connection threshold:", theta
      write(*,*) "Threshold reduction rate:", theta_reduction_rate
      if(aggregation_strategy == 1) then
         write(*,*) "aggregation strategy: coupled_w_lu"
      else if(aggregation_strategy == 2) then
         write(*,*) "aggregation strategy: coupled_w_o_lu"
      else if(aggregation_strategy == 3) then
         write(*,*) "aggregation strategy: independent_w_lu"
      else if(aggregation_strategy == 4) then
         write(*,*) "aggregation strategy: independent_w_o_lu"
      else if(aggregation_strategy == 5) then
         write(*,*) "aggregation strategy: old_independent_w_lu"
      else if(aggregation_strategy == 6) then
         write(*,*) "aggregation strategy: old_independent_w_o_lu"
      else
         write(*,*) "unknown aggregation strategy!!:", aggregation_strategy
      end if
      
      write(*,*) "process_aggregation_threshold:", process_aggre_threshold
      write(*,*) "max_level_size:", max_level_size
      write(*,*) "min_node_size:",  min_node_size
      
    END SUBROUTINE print_parameter

    
  END MODULE  solver
