module solver_AMGBiCGSTAB_NB
contains
  subroutine AMGBiCGSTAB_NB                                               &
       &                 (N, NP,  NPL, NPU,                               &
       &                  D,  AL, INL, IAL, AU, INU, IAU,                 &
       &                  B,  X,                                          &
       &                  RESID,  ITER,                                   &
       &                  my_rank, NEIBPETOT, NEIBPE,                     &
       &                  STACK_IMPORT, NOD_IMPORT,                       &
       &                  STACK_EXPORT, NOD_EXPORT,                       &
       &                  SOLVER_COMM, PETOT, SETUP_FL, LEVEL_NUM,        &
       &                  REACH_CONV_FL, NB, k_or_q_vec, NK)     

    use solver_SR_NB
    use count_time_mod
    use data_creation_AMGCG_NB, only: data_creation_NB
    use data_structure_for_AMG_NB
    use solver_AMGCG_NB
    USE terminal_condition, only: terminal_cond_check, BNRM2, terminal_kind
    USE residual_calculation, only: cal_relative_res_norm_NB
    USE amg_parameter, only: conv_check_fl, INDPDT_AGGRE
    !$ USE omp_lib    
    
    implicit NONE
!C    include  'precision.inc'
    include  'mpif.h'

    integer(kind=kint ), intent(in):: N, NP, NPU, NPL, my_rank, NB
    integer(kind=kint ), intent(in):: NEIBPETOT, PETOT
    integer(kind=kint ), intent(in):: SOLVER_COMM
    logical, intent(in) :: SETUP_FL
    LOGICAL, intent(inout) :: REACH_CONV_FL


    integer(kind=kint ), intent(inout):: ITER, LEVEL_NUM
    real   (kind=kreal), intent(inout):: RESID

    real(kind=kreal), dimension(:)   , intent(inout):: B, X
    real(kind=kreal), dimension(:,:,:), intent(inout):: AL
    real(kind=kreal), dimension(:,:,:), intent(inout):: AU
    real(kind=kreal), dimension(:,:,:), intent(inout):: D

!!$    integer(kind=kint ), dimension(0:NP) ,intent(in) :: INU, INL
!!$    integer(kind=kint ), dimension(  NPL),intent(in) :: IAL
!!$    integer(kind=kint ), dimension(  NPU),intent(in) :: IAU
    integer(kind=kint ), dimension(0:) ,intent(in) :: INU, INL
    integer(kind=kint ), dimension(:),intent(in) :: IAL
    integer(kind=kint ), dimension(:),intent(in) :: IAU

    INTEGER(kind=kint ), INTENT(in) :: NK
    real(kind=kreal), pointer :: k_or_q_vec(:,:)


    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:), NOD_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:), NOD_EXPORT(:)

    real(kind=kreal), dimension(:),    allocatable       :: WS, WR
    real(kind=kreal), dimension(:,:),  allocatable       :: WW

    real(kind=kreal), dimension(:,:,:), allocatable :: AUlu0, ALlu0
    real(kind=kreal), dimension(:,:,:), allocatable :: Dlu0

    integer(kind=kint), dimension(:), allocatable :: inumFI1L, FI1L
    integer(kind=kint), dimension(:), allocatable :: inumFI1U, FI1U

    real   (kind=kreal), dimension(2)                :: C0, CG
    real   (kind=kreal) :: WVAL(NB)

    integer(kind=kint ) :: R, RT, P, PT, S, ST, T, V, MAXIT
    integer(kind=kint ) :: ierr, i, j, k, l, m, NPLf1, NPLF2, icou, iSk, jj, iEj, kk, iEk
    integer(kind=kint ) :: iSj, NPUf1, icouU, icouL1, icouL2, icouL3, icouU1, icouL, kk1, iconIK, iconKJ
    integer(kind=kint ) :: icouU2, icouU3, icou0, ik, ij0, jj1, NPUf2, iAS, ieU, ieL,isU, isL, ip
    real   (kind=kreal) :: TOL, W, SS, RHO, RHO0, ALPHA, BETA, OMEGA, SW1, SW2, DNROM20,C20, C2
    real   (kind=kreal) :: DNRM2, RHO1 
    real   (kind=kreal) :: DNRM20, BNRM20, VAL, C01, C02

    logical,save :: initialize = .true.
    logical :: CONV_FL
    integer(kind=kint) :: NP_old, SISIZE, SESIZE, WSIZE, WSIZE2, col, NP_max

    !C following part is to be modified..
    logical :: amgprecond
    CHARACTER(len=20) :: PRECOND 

    
    !C
    !C-- INIT.

    amgprecond = .true.

    if(SETUP_FL .or. initialize) then
       i = max(8,NB)
       allocate (WW(NB*NP, i))
       NP_old = NP
       !C- decision of the allocation size for WS, WR
       WSIZE = 0
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)*NB
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)*NB
          end if
       end if
       if( WSIZE < NP*NB ) WSIZE = NP*NB

       ALLOCATE ( WS(WSIZE),WR(WSIZE) )
       initialize = .false.
    end if

    R= 1
    RT= 2
    P= 3
    PT= 4
    S= 5
    ST= 6
    T= 7
    V= 8

    MAXIT = ITER
!C    TOL   = RESID



    SISIZE = STACK_IMPORT(NEIBPETOT)
    SESIZE = STACK_EXPORT(NEIBPETOT)

    WSIZE2 = 0
    if(amgprecond .and. SETUP_FL ) then

       !C data creation for AMG
       if(INDPDT_AGGRE) then
          PRECOND ="IND1"
       else
          PRECOND ="SHA2"
       end if
       call data_creation_NB                                                   &
            &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU,           &
            &                  INU, IAU,                                       &
            &                  LEVEL_NUM, my_rank, NEIBPETOT, NEIBPE,          &
            &                  STACK_IMPORT, NOD_IMPORT,                       &
            &                  STACK_EXPORT, NOD_EXPORT,                       &
            &                  SOLVER_COMM, PRECOND, NB, k_or_q_vec, NK)     

    end if
    
    
    if(LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA_NB, LEVEL_NUM)
       
    else
       !C in the case, LEVEL_NUM == 0
       j = max(SISIZE, SESIZE)
       if(WSIZE < j) WSIZE2 = j*NB
    end if


    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA_NB, LEVEL_NUM)
    
    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2), WR(WSIZE2))
       WSIZE = WSIZE2
    END if
    if(NP_max > NP_old) then
!       i = max(8,NB)
       i = 8
       deallocate(WW)
       allocate (WW(NB*NP,i))
       NP_old = NP_max
    end if

    !C
    !C +-----------------------+
    !C | {r0}= {b} - [A]{xini} |
    !C +-----------------------+
    !C===

    !C
    !C-- INTERFACE data EXCHANGE
    if(NEIBPETOT > 0) then
       call SOLVER_SEND_RECV_NB                                                    &
            &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,   &
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,my_rank, NB)
    end if
    !C
    !C-- BEGIN calculation
    !$omp parallel do private(i,j,k,l,col,WVAL)
    do i= 1, N
       do j = 1, NB
          WVAL(j) = B(NB*(i-1)+j)
          do k = 1, NB
             WVAL(j) = WVAL(j) - D(j,k,i)*X(NB*(i-1)+k)
          end do
       end do
       
       do l = INL(i-1)+1, INL(i)
          col = IAL(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AL(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do l = INU(i-1)+1, INU(i)
          col = IAU(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AU(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do j = 1, NB
          WW(NB*(i-1)+j, R)  = WVAL(j)
          WW(NB*(i-1)+j, RT) = WVAL(j)
          WW(NB*(i-1)+j, PT) = 0.d0
          WW(NB*(i-1)+j, ST) = 0.d0
       end do
    enddo
    !$omp end parallel do    
    
!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(+:BNRM20)
!!$       do i= 1, NB*N
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       !$omp end parallel do
!!$
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$    else if(terminal_kind == 1) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(max:BNRM20)
!!$       do i= 1, NB*N
!!$          VAL = dabs(B(i))
!!$          if(BNRM20 < VAL) BNRM20 = VAL
!!$       enddo
!!$       !$omp end parallel do
!!$       
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_MAX, SOLVER_COMM, ierr)
!!$    else
!!$       stop 'unknown terminal_kind: solver_AMGBICGSTAB_NB.F90'
!!$    end if
!!$
!!$    if (BNRM2.eq.0.d0) BNRM2= 1.d0

    iter= 0
    !C===

    !C
    !C*************************************************************** iterative procedures
    !C
    do iter= 1, MAXIT
       !C
       !C +-----------------+
       !C | RHO= {r}{r_tld} |
       !C +-----------------+
       !C===

       RHO0= 0.d0
       
       !$omp parallel do reduction(+:RHO0) private(j,k)
       do j= 1, N
          do k = 1, NB
             RHO0 = RHO0 + WW(NB*(j-1)+k,RT)*WW(NB*(j-1)+k,R)
          end do
       enddo
       !$omp end parallel do
       
       call MPI_allREDUCE (RHO0, RHO, 1, MPI_DOUBLE_PRECISION,           &
            &                    MPI_SUM, SOLVER_COMM, ierr)

       !C===

       !C
       !C +----------------------------------------+
       !C | BETA= (RHO/RHO1) * (ALPHA/OMEGA)       |
       !C | {p} = {r} + BETA * ( {p} - OMEGA*{v} ) |
       !C +----------------------------------------+
       !C===
       if (iter.gt.1) then
          BETA= (RHO/RHO1) * (ALPHA/OMEGA)
          !$omp parallel do private(j,k)
          do j= 1, N
             do k = 1, NB
                WW(NB*(j-1)+k, P) = WW(NB*(j-1)+k, R)+BETA*(WW(NB*(j-1)+k,P)-OMEGA*WW(NB*(j-1)+k,V))
             end do
          end do
          !$omp end parallel do 
       else
          !$omp parallel do private(j,k)
          do j= 1, N
             do k = 1, NB
                WW(NB*(j-1)+k, P) = WW(NB*(j-1)+k,R)
             end do
          end do
          !$omp end parallel do 
       endif
       !C===

       !C
       !C +--------------------+
       !C | {p_tld}= [Minv]{p} |
       !C +--------------------+
       !C===
       !C
       !C-- INTERFACE data EXCHANGE

       call SOLVER_SEND_RECV_NB                                                 &
            &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,P) , SOLVER_COMM,      &
            &     my_rank, NB)


       if(amgprecond) then
          !C WW(:,V) is used as temporary space

          !$omp parallel do private(j)
          do j= 1, NB*NP
             WW(j,PT) = 0.0d0
          end do
          !$omp end parallel do
          
          call count_time(1,SOLVER_COMM,my_rank,9)          
          CALL v_cycle(WW(:,P), WW(:,PT), LEVEL_NUM, SOLVER_COMM,&
               & my_rank, WS, WR, WW(:,V), WSIZE)
          call count_time(2,SOLVER_COMM,my_rank,9)          
          
       else
          !$omp parallel do private(j)
          do j= 1, NB*NP
             WW(j  ,PT)= WW(j  ,P)
          enddo
          !$omp end parallel do
       end if


       !C
       !C +-------------------+
       !C | {v} = [A] {p_tld} |
       !C +-------------------+
       !C===

       !C
       !C-- INTERFACE data EXCHANGE

       call SOLVER_SEND_RECV_NB                                                 &
            &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,PT) , SOLVER_COMM,     &
            &     my_rank, NB)

       !C
       !C-- BEGIN calculation
       !$omp parallel do private(i,j,k,l,m)
       do j= 1, N
          do k = 1, NB
             WW(NB*(j-1)+k, V) = 0.d0
             do l = 1, NB
                WW(NB*(j-1)+k, V) = WW(NB*(j-1)+k, V)+D(k,l,j)*WW(NB*(j-1)+l, PT)
             end do
          end do

          do k= INL(j-1)+1, INL(j)
             i= IAL(k)
             do l = 1, NB
                do m = 1, NB
                   WW(NB*(j-1)+l, V) = WW(NB*(j-1)+l, V)+AL(l,m,k)*WW(NB*(i-1)+m, PT)
                end do
             end do
          enddo

          do k= INU(j-1)+1, INU(j)
             i= IAU(k)
             do l = 1, NB
                do m = 1, NB
                   WW(NB*(j-1)+l, V) = WW(NB*(j-1)+l, V)+AU(l,m,k)*WW(NB*(i-1)+m, PT)
                end do
             end do
          enddo
       enddo
       !$omp end parallel do 
       
       !C===

       !C
       !C-- calc. ALPHA

       C20= 0.d0
       !$omp parallel do reduction(+:C20) private(i,j)
       do j= 1, N
          do i = 1, NB
             C20= C20 + WW(NB*(j-1)+i,RT)*WW(NB*(j-1)+i,V)
          end do
       end do
       !$omp end parallel do

       call MPI_allREDUCE (C20, C2, 1, MPI_DOUBLE_PRECISION,             &
            &                    MPI_SUM, SOLVER_COMM, ierr) 
       ALPHA= RHO / C2

       !C
       !C-- {s}= {r} - ALPHA*{V}
       !$omp parallel do private(i,j)
       do j= 1, N
          do i = 1, NB
             WW(NB*(j-1)+i,S) = WW(NB*(j-1)+i,R) - ALPHA*WW(NB*(j-1)+i,V)
          end do
       enddo
       !$omp end parallel do 


       !C
       !C +--------------------+
       !C | {s_tld}= [Minv]{s} |
       !C +--------------------+
       !C===

       call SOLVER_SEND_RECV_NB                                                 &
            &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,S) , SOLVER_COMM,      &
            &     my_rank, NB)


       if(amgprecond) then
          !C WW(:,T) is used as temporary space
!!$          CALL v_cycle_ssi_amg(WW(:,S), WW(:,ST), WW(:,T), LEVEL_NUM, SOLVER_COMM, WS, &
!!$               & WR, NP, WSIZE)

          !$omp parallel do private(j)
          do j = 1, NB*NP
             WW(j, ST) = 0.0d0
          end do
          
          !$omp end parallel do

          call count_time(1,SOLVER_COMM,my_rank,9)          
          
          CALL v_cycle(WW(:,S), WW(:,ST), LEVEL_NUM, SOLVER_COMM, &
               & my_rank, WS, WR, WW(:,T), WSIZE)
          
          call count_time(2,SOLVER_COMM,my_rank,9)          
                    

       else
          !$omp parallel do private(j)
          do j= 1, NB*NP
             WW(j  ,ST)= WW(j  ,S)
          enddo
          !$omp end parallel do
       end if




       !C
       !C +------------------+
       !C | {t} = [A]{s_tld} |
       !C +------------------+
       !C===

       !C
       !C-- INTERFACE data EXCHANGE
       call SOLVER_SEND_RECV_NB                                            &
            &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,ST) , SOLVER_COMM,      &
            &     my_rank, NB)

       !C
       !C-- BEGIN calculation
       !$omp parallel do private(i,j,k,l,m)
       do j= 1, N
          do k = 1, NB
             WW(NB*(j-1)+k, T) = 0.d0
             do l = 1, NB
                WW(NB*(j-1)+k, T) = WW(NB*(j-1)+k, T)+D(k,l,j)*WW(NB*(j-1)+l,ST)
             end do
          end do

          do k= INL(j-1)+1, INL(j)
             i= IAL(k)
             do l = 1, NB
                do m = 1, NB
                   WW(NB*(j-1)+l, T) = WW(NB*(j-1)+l, T)+AL(l,m,k)*WW(NB*(i-1)+m,ST)
                end do
             end do
          enddo

          do k= INU(j-1)+1, INU(j)
             i= IAU(k)
             do l = 1, NB
                do m = 1, NB
                   WW(NB*(j-1)+l, T) = WW(NB*(j-1)+l, T)+AU(l,m,k)*WW(NB*(i-1)+m,ST)
                end do
             end do
          enddo
       enddo
       !$omp end parallel do 


       !C===

       !C
       !C +----------------------------+
       !C | OMEGA= ({t}{s}) / ({t}{t}) |
       !C +----------------------------+
       !C===
       C01 = 0.d0
       C02 = 0.d0
       CG(1)= 0.d0
       CG(2)= 0.d0

       !$omp parallel do reduction(+:C01,C02) private(i,j)
       do j= 1, N
          do i = 1, NB
             C01= C01 + WW(NB*(j-1)+i,T)*WW(NB*(j-1)+i,S)
             C02= C02 + WW(NB*(j-1)+i,T)*WW(NB*(j-1)+i,T)
          end do
       end do
       !$omp end parallel do
       
       C0(1) = C01; C0(2) = C02
       call MPI_allREDUCE (C0, CG, 2, MPI_DOUBLE_PRECISION,              &
            &                    MPI_SUM, SOLVER_COMM, ierr)
       OMEGA= CG(1) / CG(2)
       !C===

       !C
       !C +----------------+
       !C | update {x},{r} |
       !C +----------------+
       !C===
!!$       DNRM20= 0.d0
       !$omp parallel do private(j,k) 
       do j= 1, N
          do k = 1, NB
             X(NB*(j-1)+k) = X(NB*(j-1)+k) +ALPHA*WW(NB*(j-1)+k,PT)+OMEGA*WW(NB*(j-1)+k,ST)
             WW(NB*(j-1)+k,R) = WW(NB*(j-1)+k,S) - OMEGA*WW(NB*(j-1)+k,T)
!!$             DNRM20 = DNRM20 + WW(NB*(j-1)+k,S)**2
          end do
       enddo
       !$omp end parallel do


       RHO1= RHO

!!$       call MPI_allREDUCE  (DNRM20, DNRM2, 1, MPI_DOUBLE_PRECISION,      &
!!$            &                     MPI_SUM, SOLVER_COMM, ierr)
!!$       RESID= dsqrt(DNRM2/BNRM2)
       
       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N, SOLVER_COMM)
       
#ifdef PRINT_REZ
       if (my_rank.eq.0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif

!!$       if (RESID.le.TOL   ) exit
       if(CONV_FL) then
          REACH_CONV_FL = .true.
          exit
       end if
       !C===

    enddo

    !C
    !C-- INTERFACE data EXCHANGE
    call SOLVER_SEND_RECV_NB                                                  &
         &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
         &     STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,my_rank, NB)
    
    !C differentiate divergence cases from non-convergent cases
    if(conv_check_fl)then
       if(.not.CONV_FL) then
          call cal_relative_res_norm_NB(RESID, X, B, WS, WR, WSIZE, NB)
          if(RESID > 1.0) then
             iter = -1
          end if
       end if
    end if

    
    if(SETUP_FL) then
       deallocate (WW)
       deallocate (WR)
       deallocate (WS)
       initialize = .true.
    end if
    
  end subroutine        AMGBiCGSTAB_NB
end module     solver_AMGBiCGSTAB_NB

