!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

module solver_AMGBiCGSTAB
contains
  subroutine AMGBiCGSTAB                                                  &
       &                 (N, NP,  NPL, NPU,                               &
       &                  D,  AL, INL, IAL, AU, INU, IAU,                 &
       &                  B,  X,                                          &
       &                  RESID,  ITER,                                   &
       &                  my_rank, NEIBPETOT, NEIBPE,                     &
       &                  STACK_IMPORT, NOD_IMPORT,                       &
       &                  STACK_EXPORT, NOD_EXPORT,                       &
       &                  SOLVER_COMM, PETOT, SETUP_FL, LEVEL_NUM,        &
       &                  REACH_CONV_FL)     

    use solver_SR2
    use data_creation_AMGCG
    USE common_mod, only: clear_matrix_ssi_amg
    use AMG_cycle
    use data_structure_for_AMG
    USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
         & terminal_cond_check, BNRM2, terminal_kind, conv_criterion_val
    USE residual_calculation
    USE amg_parameter, only: conv_check_fl
    !$ USE omp_lib


    implicit none
    include  'mpif.h'

    integer(kind=kint ), intent(in   )::  N
    integer(kind=kint ), intent(in   )::  NP
    integer(kind=kint ), intent(in   )::  NPU, NPL

    real   (kind=kreal), intent(inout)::  RESID
    integer(kind=kint ), intent(inout)::  ITER
    integer(kind=kint ), intent(in   )::  my_rank

    integer            , intent(in)   :: SOLVER_COMM
    integer(kind=kint ), intent(in   )::  PETOT
    logical, intent(in)  :: SETUP_FL
    logical, intent(inout) :: REACH_CONV_FL


    real   (kind=kreal),  intent(inout)::  B(:)
    real   (kind=kreal),  intent(inout)::  X(:)

    real   (kind=kreal),  intent(inout)::  AL(:)
    !C size is NPL
    real   (kind=kreal),  intent(inout)::  AU(:)
    !C size is NPU
    real   (kind=kreal),  intent(inout)::  D(:)

    integer(kind=kint ), intent(in) :: INU(0:), INL(0:)
    integer(kind=kint ), intent(in) :: IAL(:)
    integer(kind=kint ), intent(in) :: IAU(:)
    
    integer(kind=kint ) :: NEIBPETOT
    integer(kind=kint ) :: NEIBPE      (:)
    integer(kind=kint ) :: STACK_IMPORT(0:)
    integer(kind=kint ) :: NOD_IMPORT  (:)
    integer(kind=kint ) :: STACK_EXPORT(0:)
    integer(kind=kint ) :: NOD_EXPORT  (:)

    real(kind=kreal), allocatable, save ::  WS(:), WR(:)

    real(kind=kreal), allocatable, save ::  WW(:,:)

    real(kind=kreal)                    :: C0(2), CG(2), C01, C02

    integer(kind=kint ) :: IFLAG, R, RT, P, PT, S, ST, T, V, MAXIT,ierr
    integer(kind=kint ) :: i,j,k,l,m
    real   (kind=kreal) :: TOL, W, SS, WVAL, RHO0,RHO,OMEGA,ALPHA,RHO1
    real   (kind=kreal) :: BETA,C20,C2, DNRM20,DNRM2,BNRM20
    
    integer(kind=kint) :: WSIZE2, LEVEL_NUM, SISIZE, SESIZE, NP_max
    integer(kind=kint), save :: WSIZE, NP_old
    !C initialize flag shows the first time this subroutine is called
    logical, save :: initialize = .true.
    logical :: amgprecond, CONV_FL
    
    !C
    !C-- INIT.

    amgprecond = .true.

    if(SETUP_FL .or. initialize) then
       allocate (WW(NP,8))
       NP_old = NP
       !C- decision of the allocation size for WS, WR
       WSIZE = 0
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)
          end if
       end if
       if( WSIZE < NP ) WSIZE = NP
    
       ALLOCATE ( WS(WSIZE),WR(WSIZE) )
       initialize = .false.
    end if

    R=  1
    RT= 2
    P=  3
    PT= 4
    S=  5
    ST= 6
    T=  7
    V=  8

    MAXIT = ITER
!    TOL   = RESID
!C    conv_criterion_val = RESID

    SISIZE = STACK_IMPORT(NEIBPETOT)
    SESIZE = STACK_EXPORT(NEIBPETOT)       

    if(amgprecond .and. SETUP_FL ) then

       CALL data_creation_unsym_ssi_amg(N, NP, NPL, NPU, D, AL, &
            & INL, IAL, AU, INU, IAU, NEIBPETOT,                &
            & NEIBPE, STACK_IMPORT, NOD_IMPORT,                 &
            & STACK_EXPORT, NOD_EXPORT,                         &
            & SISIZE, SESIZE, SOLVER_COMM, LEVEL_NUM, WSIZE2)
       

    else if(LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA, LEVEL_NUM)
       
    else
       !C in the case, LEVEL_NUM == 0
       j = max(SISIZE, SESIZE)
       if(WSIZE < j) WSIZE2 = j
    end if
    
    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA, LEVEL_NUM)
    
    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2), WR(WSIZE2))
       WSIZE = WSIZE2
    END if
    if(NP_max > NP_old) then
       deallocate(WW)
       allocate (WW(NP_max,8))       
       NP_old = NP_max
    end if

    !C
    !C +-----------------------+
    !C | {r0}= {b} - [A]{xini} |
    !C +-----------------------+
    !C===


    !C
    !C-- INTERFACE data EXCHANGE
    if (NEIBPETOT > 0) then
       call SOLVER_SEND_RECV2                                                     &
            &   ( NP,SESIZE,SISIZE, NEIBPETOT,                                    &
            &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS, &
            &     WR, X , SOLVER_COMM, my_rank)
       
    end if



    !C
    !C-- BEGIN calculation

    !$omp parallel do private(k,WVAL,i)
    do j= 1, N
       WVAL = B(j) - D(j)*X(j)

       do k= INL(j-1)+1, INL(j)
          i= IAL(k)
          WVAL= WVAL - AL(k)*X(i)
       enddo

       do k= INU(j-1)+1, INU(j)
          i= IAU(k)
          WVAL= WVAL - AU(k)*X(i)
       enddo

       WW(j  ,R )= WVAL
       WW(j  ,RT)= WVAL

       WW(j  ,PT)= 0.d0
       WW(j  ,ST)= 0.d0
    enddo
    !$omp end parallel do
    
    iter= 0
    !C===

    !C
    !C*************************************************************** iterative procedures
    !C
    do iter= 1, MAXIT
       !C
       !C +-----------------+
       !C | RHO= {r}{r_tld} |
       !C +-----------------+
       !C===

       RHO0= 0.d0
       !$omp parallel do reduction(+:RHO0)
       do j= 1, N
          RHO0= RHO0+WW(j,RT)*WW(j,R)
       enddo
       !$omp end parallel do

       call MPI_allREDUCE (RHO0, RHO, 1, MPI_DOUBLE_PRECISION,           &
            &                    MPI_SUM, SOLVER_COMM, ierr)
       !C===

       !C
       !C +----------------------------------------+
       !C | BETA= (RHO/RHO1) * (ALPHA/OMEGA)       |
       !C | {p} = {r} + BETA * ( {p} - OMEGA*{v} ) |
       !C +----------------------------------------+
       !C===
       if (iter.gt.1) then
          BETA= (RHO/RHO1) * (ALPHA/OMEGA)
          !$omp parallel do
          do j= 1, N
             WW(j  ,P)= WW(j  ,R)+BETA*(WW(j  ,P)-OMEGA*WW(j  ,V))
          enddo
          !$omp end parallel do
       else
          !$omp parallel do
          do j= 1, N
             WW(j  ,P)= WW(j  ,R)
          enddo
          !$omp end parallel do
       endif
       !C===

       !C
       !C +--------------------+
       !C | {p_tld}= [Minv]{p} |
       !C +--------------------+
       !C===

       !C
       !C-- INTERFACE data EXCHANGE

!!$       call SOLVER_SEND_RECV2                                                   &
!!$            &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
!!$            &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,P) , SOLVER_COMM,      &
!!$            &     my_rank)


       if(amgprecond) then
          !C WW(:,V) is used as temporary space
          !$omp parallel do
          do j= 1, NP
             WW(j,PT) = 0.0d0
          end do
          !$omp end parallel do
          
          CALL amg_cycles(WW(:,P), WW(:,PT), LEVEL_NUM, SOLVER_COMM,&
               & my_rank, WS, WR, WW(:,V), PETOT, WSIZE, NP)


          
       else 
          !$omp parallel do          
          do j= 1, NP
             WW(j  ,PT)= WW(j  ,P)
          enddo
          !$omp end parallel do
       end if
       


       !C
       !C +-------------------+
       !C | {v} = [A] {p_tld} |
       !C +-------------------+
       !C===

       !C
       !C-- INTERFACE data EXCHANGE

       call SOLVER_SEND_RECV2                                                   &
            &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,PT) , SOLVER_COMM,     &
            &     my_rank)

       !C
       !C-- BEGIN calculation
       !$omp parallel do private(WVAL,k,i)
       do j= 1, N
          WVAL= D(j)*WW(j,PT)
          do k= INL(j-1)+1, INL(j)
             i= IAL(k)
             WVAL= WVAL + AL(k)*WW(i,PT)
          enddo

          do k= INU(j-1)+1, INU(j)
             i= IAU(k)
             WVAL= WVAL + AU(k)*WW(i,PT)
          enddo

          WW(j,V)= WVAL
       enddo
       !$omp end parallel do
       !C===

       !C
       !C-- calc. ALPHA
       C20= 0.d0
       !$omp parallel do reduction(+:C20)
       do j= 1, N
          C20= C20 + WW(j,RT)*WW(j,V) 
       enddo
       !$omp end parallel do

       call MPI_allREDUCE (C20, C2, 1, MPI_DOUBLE_PRECISION,             &
            &                    MPI_SUM, SOLVER_COMM, ierr) 
       ALPHA= RHO / C2

       !C
       !C-- {s}= {r} - ALPHA*{V}
       !$omp parallel do 
       do j= 1, N
          WW(j  ,S)= WW(j  ,R) - ALPHA*WW(j  ,V)
       enddo
       !$omp end parallel do 

       !C
       !C +--------------------+
       !C | {s_tld}= [Minv]{s} |
       !C +--------------------+
       !C===

       !C
       !C-- INTERFACE data EXCHANGE
       
!!$       call SOLVER_SEND_RECV2                                                   &
!!$            &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
!!$            &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,S) , SOLVER_COMM,      &
!!$            &     my_rank)


       if(amgprecond) then
          !C WW(:,T) is used as temporary space
          !$omp parallel do 
          do j = 1, NP
             WW(j, ST) = 0.0d0
          end do
          !$omp end parallel do 

          CALL amg_cycles(WW(:,S), WW(:,ST), LEVEL_NUM, SOLVER_COMM,&
               & my_rank, WS(1:WSIZE), WR(1:WSIZE), WW(:,T), PETOT, WSIZE, NP)

       else
          !$omp parallel do 
          do j= 1, NP
             WW(j  ,ST)= WW(j  ,S)
          enddo
          !$omp end parallel do 
       end if
       
          
       !C
       !C +------------------+
       !C | {t} = [A]{s_tld} |
       !C +------------------+
       !C===

       !C
       !C-- INTERFACE data EXCHANGE
       call SOLVER_SEND_RECV2                                                   &
            &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,ST) , SOLVER_COMM,     &
            &     my_rank)

       !C
       !C-- BEGIN calculation
       !$omp parallel do private(WVAL,i,k)
       do j= 1, N
          WVAL= D(j)*WW(j  ,ST)

          do k= INL(j-1)+1, INL(j)
             i= IAL(k)
             WVAL= WVAL + AL(k)*WW(i  ,ST)
          enddo

          do k= INU(j-1)+1, INU(j)
             i= IAU(k)
             WVAL= WVAL + AU(k)*WW(i  ,ST)
          enddo

          WW(j  ,T)= WVAL
       enddo
       !$omp end parallel do 
       !C===

       !C
       !C +----------------------------+
       !C | OMEGA= ({t}{s}) / ({t}{t}) |
       !C +----------------------------+
       !C===
       C01= 0.d0
       C02= 0.d0
       CG(1)= 0.d0
       CG(2)= 0.d0

       !$omp parallel do reduction(+:C01,C02)
       do j= 1, N
          C01= C01 + WW(j,T)*WW(j,S)
          C02= C02 + WW(j,T)*WW(j,T)
       enddo
       !$omp end parallel do
       
       C0(1) = C01; C0(2)=C02
       call MPI_allREDUCE (C0, CG, 2, MPI_DOUBLE_PRECISION,              &
            &                    MPI_SUM, SOLVER_COMM, ierr)
       OMEGA= CG(1) / CG(2)
       !C===

       !C
       !C +----------------+
       !C | update {x},{r} |
       !C +----------------+
       !C===
       !$omp parallel do
       do j= 1, N
          X (j  )= X(j  ) + ALPHA*WW(j  ,PT) + OMEGA*WW(j  ,ST)
          WW(j  ,R)= WW(j  ,S) - OMEGA*WW(j  ,T)
       enddo
       !$omp end parallel do

       RHO1= RHO
       
       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N, SOLVER_COMM)

       !C##### ITERATION HISTORY
#ifdef PRINT_REZ
!       if (my_rank.eq.0 .and. terminal_kind==0) write (*, 1000) ITER, RESID
       if (my_rank.eq.0 ) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif
       !C#####


       if(CONV_FL) then
          REACH_CONV_FL = .true.
          exit
       end if
       !C===

    enddo

    !C
    !C-- INTERFACE data EXCHANGE
    call SOLVER_SEND_RECV2                                                   &
         &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
         &     STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,my_rank)


    if(conv_check_fl) then
       !C differentiate divergence cases from non-convergent cases
       if(.not.CONV_FL) then
          call cal_relative_res_norm(RESID, X, B, WS, WR, WSIZE)
          if(RESID > 1.0) then
             iter = -1
          end if
       end if
    end if


    if(SETUP_FL) then
       deallocate (WW)
       deallocate (WR)
       deallocate (WS)
       initialize = .true.
    end if

    if(amgprecond .and. SETUP_FL) then
       call clear_matrix_ssi_amg(LEVEL_NUM)
    end if
    
  end subroutine        AMGBiCGSTAB
end module     solver_AMGBiCGSTAB

