!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.

#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE solver_AMGCG
CONTAINS
  SUBROUTINE AMGCG    (N, NP,  NPL, NPU, D,  AL, INL, IAL, AU, INU, IAU,  &
       & B, X, RESID, ITER, my_rank, NEIBPETOT, NEIBPE,                   &
       & STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
       & NPROCS, SETUP_FL, LEVEL_NUM, REACH_CONV_FL                       &
       & )

    USE solver_SR2
    USE data_structure_for_AMG
    USE solver_Gnumbering
    USE count_time_mod
    USE common_mod, only: clear_matrix_ssi_amg
    USE data_creation_AMGCG
    USE AMG_cycle
    USE terminal_condition, only: terminal_cond_check, BNRM2, terminal_kind
    USE residual_calculation
    USE amg_parameter, only: conv_check_fl
    !$ USE omp_lib

    IMPLICIT NONE
    include 'mpif.h'

    LOGICAL, intent(in) :: SETUP_FL    
    INTEGER(kind=kint ), intent(in)  :: N,NP,NPL,NPU
    REAL   (kind=kreal), INTENT(inout)::  RESID
    INTEGER(kind=kint ), INTENT(inout)::  ITER
    INTEGER(kind=kint ), INTENT(in   )::  my_rank
    LOGICAL, intent(inout) :: REACH_CONV_FL

    INTEGER(kind=kint ), INTENT(in)   :: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in)   :: NPROCS

    !C NP
    REAL   (kind=kreal) ::  D(:)
    !C N
    REAL   (kind=kreal) ::  B(:)

    !C NP
    REAL   (kind=kreal) ::  X(:)
    REAL   (kind=kreal) ::  AU(:)
    REAL   (kind=kreal) ::  AL(:)

    INTEGER(kind=kint ) ::  INU(0:)
    INTEGER(kind=kint ) ::  IAU(:)
    INTEGER(kind=kint ) ::  INL(0:)
    INTEGER(kind=kint ) ::  IAL(:)

    INTEGER(kind=kint ) :: NEIBPE      (:)
    INTEGER(kind=kint ) :: STACK_IMPORT(0:)
    INTEGER(kind=kint ) :: NOD_IMPORT  (:)

    INTEGER(kind=kint ) :: STACK_EXPORT(0:)
    INTEGER(kind=kint ) :: NOD_EXPORT  (:)

    INTEGER(kind=kint ) :: SISIZE,SESIZE

    REAL   (kind=kreal), ALLOCATABLE, save :: WS(:), WR(:)
    REAL   (kind=kreal), ALLOCATABLE, save :: WW(:,:)
    REAL   (kind=kreal), ALLOCATABLE, save :: Temp(:)

    REAL   (kind=kreal), ALLOCATABLE, SAVE :: DD(:)
    REAL   (kind=kreal), ALLOCATABLE, SAVE :: SCALE(:)

    INTEGER(kind=kint ) :: P, Q, R, Z, MAXIT, IFLAG=0
    REAL   (kind=kreal) :: TOL, W, SS

    INTEGER(kind=kint)  :: LEVEL_NUM,cnt,NEIBPETOT,ISU,IEU,ISL,IEL,WSIZE2
    INTEGER(kind=kint),save  :: WSIZE, NP_old
    INTEGER(kind=kint)  :: INOD, IERR, i,j,k,l,m, NP_max
    REAL   (kind=kreal) :: WVAL,BNRM20,RHO0,RHO,BETA,RHO1,C10,C1,ALPHA,DNRM20
    REAL   (kind=kreal) :: DNRM2,X1,X2
    REAL   (kind=kreal) :: STARTTIME,ENDTIME,RTIME,STARTTIME2,ENDTIME2,RTIME2
    logical, save :: initialize = .true.
    logical :: CONV_FL

    !C-- INIT. 
    SISIZE = STACK_IMPORT(NEIBPETOT) 
    SESIZE = STACK_EXPORT(NEIBPETOT) 

    if(initialize .or. SETUP_FL) then
       ALLOCATE (WW(NP,3))
       NP_old = NP

       !C- decision of the allocation size for WS, WR
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)
          end if
       end if
       if( WSIZE < NP ) WSIZE = NP

       ALLOCATE ( WS(WSIZE),WR(WSIZE) )
       ALLOCATE ( Temp(NP) )

       initialize = .false.
    end if

    R = 1
    Z = 2
    Q = 2
    P = 3

    MAXIT = ITER
    !C    conv_criterion_val = RESID 
    !C if terminal_condition isn't set,  RESID will 
    

    !C  D(1:NP), AL(1:NPL), INL(0:N), IAL(1:NPL), AU(1:NPU), INU(0:N), IAU(1:NPU)
    !C  NEIBPE(1:NEIBPETOT), STACK_IMPORT(0:NEIBPETOT), NOD_IMPORT(1:SISIZE)
    !C  STACK_EXPORT(0:NEIBPETOT), NOD_EXPORT(1:SESIZE)
    if(SETUP_FL) then
       
       CALL data_creation_unsym_ssi_amg(N, NP, NPL, NPU, D, AL, INL, &
!!$       CALL data_creation_ssi_amg(N, NP, NPL, NPU, D, AL, INL, &
            & IAL, AU, INU, IAU, NEIBPETOT,                          &
            & NEIBPE, STACK_IMPORT, NOD_IMPORT,                      &
            & STACK_EXPORT, NOD_EXPORT,                              &
            & SISIZE, SESIZE, SOLVER_COMM, LEVEL_NUM, WSIZE2)

    else if (LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA, LEVEL_NUM)

    else
       !C in the case, LEVEL_NUM == 0
       stop 'error. no multi-level data is available'

    end if

    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA, LEVEL_NUM)
    
    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2),WR(WSIZE2))
       WSIZE = WSIZE2
    END if
    if(NP_max > NP_old) then
       deallocate(WW)
       allocate(WW(NP_max,3))
       deallocate(Temp)
       allocate(Temp(NP_max))
       NP_old = NP_max
    END if
    !C-------- allocation of WS, WS, WW, temp is finished.
    

    if (NEIBPETOT > 0) then
       call SOLVER_SEND_RECV2                                                     &
            &   ( NP,SESIZE,SISIZE, NEIBPETOT,                                    &
            &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS, &
            &     WR, X , SOLVER_COMM, my_rank)

    end if
    
    !C
    !C +-----------------------+
    !C | {r0}= {b} - [A]{xini} |
    !C +-----------------------+
    !C===

    !$omp parallel do private(j,WVAL,isU,ieU,i,inod,isL,ieL)
    do j= 1, N
       WVAL= B(j) - D(j) * X(j)
       isU= INU(j-1) + 1
       ieU= INU(j  ) 

       do i= isU, ieU
          inod= IAU(i)+ZERO_ORIGIN
          WVAL= WVAL - AU(i) * X(inod)
       enddo

       isL= INL(j-1) + 1
       ieL= INL(j  ) 
       do i= isL, ieL
          inod= IAL(i)+ZERO_ORIGIN
          WVAL= WVAL - AL(i) * X(inod)
       enddo
       WW(j,R)= WVAL
    enddo
    !$omp end parallel do

!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(+:BNRM20)
!!$       do i= 1, N
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       !$omp end parallel do
!!$       
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$    else if(terminal_kind == 1) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(max:BNRM20)
!!$       do i= 1, N
!!$          WVAL = dabs(B(i))
!!$          if(BNRM20 < WVAL) BNRM20 = WVAL
!!$       enddo
!!$       !$omp end parallel do
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_MAX, SOLVER_COMM, ierr)
!!$    else 
!!$       stop 'unknown terminal_kind: solver_AMGCG.F90'
!!$    end if
!!$
!!$    if (BNRM2.eq.0.d0) BNRM2= 1.d0
    

    do iter= 1, MAXIT


       !C +----------------+
       !C | {z}= [Minv]{r} |
       !C +----------------+
       !$omp parallel do
       do i = 1, NP
          WW(i,Z) = 0.0
       end do
       !$omp end parallel do


       CALL amg_cycles(WW(:,R), WW(:,Z), LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, TEMP, &
            &          NPROCS, WSIZE, NP)

!!$       if(NEIBPETOT > 0) then
!!$          call SOLVER_SEND_RECV2                                                      &
!!$               &   ( NP, WSIZE, WSIZE, NEIBPETOT,                                     &
!!$               &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS,  &
!!$               &     WR, WW(:,Z), SOLVER_COMM, my_rank)
!!$       END if
       
       !C +---------------+
       !C | {RHO}= {r}{z} |
       !C +---------------+


       RHO0= 0.0
       !$omp parallel do reduction(+:RHO0)
       do i= 1, N
          RHO0= RHO0 + WW(i,R)*WW(i,Z)
       enddo
       !$omp end parallel do


       call MPI_allREDUCE (RHO0, RHO, 1, MPI_DOUBLE_PRECISION,           &
            &                    MPI_SUM, SOLVER_COMM, ierr)


       !C +-----------------------------+
       !C | {p} = {z} if      ITER=1    |
       !C | BETA= RHO / RHO1  otherwise |
       !C +-----------------------------+

       if ( ITER.eq.1 ) then
          !$omp parallel do
          do i= 1, N
             WW(i,P)= WW(i,Z)
          enddo
          !$omp end parallel do
       else

          BETA= RHO / RHO1

          !$omp parallel do
          do i= 1, N
             WW(i,P)= WW(i,Z) + BETA*WW(i,P)
          enddo
          !$omp end parallel do
       endif

       !C +-------------+
       !C | {q}= [A]{p} |
       !C +-------------+
       !C-- INTERFACE data EXCHANGE
       if(NEIBPETOT > 0) then

          call SOLVER_SEND_RECV2                                                     &
               &   ( NP, WSIZE, WSIZE, NEIBPETOT,                                    &
               &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS, &
               &     WR, WW(:,P), SOLVER_COMM, my_rank)

       end if


       !$omp parallel do private(WVAL,i,inod, isU,ieU,isL,ieL)
       do j= 1, N
          WVAL= D(j) * WW(j,P)

          isU= INU(j-1) + 1
          ieU= INU(j  ) 

          do i= isU, ieU
             inod= IAU(i)+ZERO_ORIGIN
             WVAL= WVAL + AU(i) * WW(inod,P)
          enddo

          isL= INL(j-1) + 1
          ieL= INL(j  ) 

          do i= isL, ieL
             inod= IAL(i)+ZERO_ORIGIN
             WVAL= WVAL + AL(i) * WW(inod,P)
          enddo
          WW(j,Q)= WVAL
       enddo
       !$omp end parallel do

       !C +---------------------+
       !C | ALPHA= RHO / {p}{q} |
       !C +---------------------+

       C10= 0.d0
       !$omp parallel do reduction(+:C10)
       do i= 1, N
          C10= C10 + WW(i,P)*WW(i,Q)
       enddo
       !$omp end parallel do

       call MPI_allREDUCE (C10, C1, 1, MPI_DOUBLE_PRECISION,             &
            &                    MPI_SUM, SOLVER_COMM, ierr)

       ALPHA= RHO / C1


       !C +----------------------+
       !C | {x}= {x} + ALPHA*{p} |
       !C | {r}= {r} - ALPHA*{q} |
       !C +----------------------+

       !$omp parallel do 
       do i= 1, N
          X(i)    = X(i)   + ALPHA * WW(i,P)
          WW(i,R) = WW(i,R) - ALPHA * WW(i,Q)
       enddo
       !$omp end parallel do

       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N, SOLVER_COMM)


#ifdef PRINT_REZ
       if (my_rank.eq.0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif

       if ( CONV_FL  ) then
          REACH_CONV_FL = .true.
          exit
       end if
       RHO1 = RHO                                                             
    end do
    !C===


#ifdef PRINT_INFO
    call matrix_counting(LEVEL_NUM,SOLVER_COMM,my_rank)
    call count_time(4,solver_comm,my_rank,1)
#endif


    if(conv_check_fl) then
       call cal_res_vec(WW(:,R), X, B)
       !C differentiate divergence cases from non-convergent cases
       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N, SOLVER_COMM)       
       if(.not.CONV_FL) then
          call cal_relative_res_norm(RESID, X, B, WS, WR, WSIZE)
          if(RESID > 1.0) then
             iter = -1
          end if
       end if
    end if
       
    if(SETUP_FL) then
       call clear_matrix_ssi_amg(LEVEL_NUM)
       DEALLOCATE (WW)
       DEALLOCATE (WR)
       DEALLOCATE (WS)
       DEALLOCATE (Temp)
       initialize = .true.
    end if

#ifdef PSPASES
    CALL PSPACEC(pspcommf,0)
#endif

  END SUBROUTINE AMGCG

END MODULE  solver_AMGCG
