
module solver_AMGCG_NB
contains
  subroutine AMGCG_NB                                                &
       &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, &
       &                  B,  X, RESID,  ITER, my_rank, NEIBPETOT, NEIBPE,&
       &                  STACK_IMPORT, NOD_IMPORT,                       &
       &                  STACK_EXPORT, NOD_EXPORT,                       &
       &                  SOLVER_COMM, NPROCS, SETUP_FL, LEVEL_NUM,       &
       &                  REACH_CONV_FL, NB, k_or_q_vec, NK)     

    use data_structure_for_AMG_NB, only: HIERARCHICAL_DATA_NB, max_comm_table_sz, &
         & max_NP_on_all_levs
    use data_creation_AMGCG_NB, only: data_creation_NB
    use solver_SR_NB
    use count_time_mod
    use common_mod, only: setup_phase, clear_matrix_ssi_amg
    USE terminal_condition, only: terminal_cond_check, BNRM2, terminal_kind
    USE amg_parameter, only: conv_check_fl,INDPDT_AGGRE
    use residual_calculation, only:cal_relative_res_norm_NB

    IMPLICIT NONE
    include  'precision.inc'
    include  'mpif.h'

    LOGICAL, intent(in) :: SETUP_FL    
    integer(kind=kint ), intent(in):: NB, N, NP, NPU, NPL, my_rank
    integer(kind=kint ), intent(in):: NEIBPETOT
    integer(kind=kint ), intent(in):: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in)   :: NPROCS
    LOGICAL, intent(inout) :: REACH_CONV_FL


    integer(kind=kint ), intent(inout):: ITER
    real   (kind=kreal), intent(inout):: RESID

!!$    real(kind=kreal), dimension(NB*NP)   , intent(inout):: B, X
!!$    real(kind=kreal), dimension(NB,NB,NPL), intent(inout):: AL
!!$    real(kind=kreal), dimension(NB,NB,NPU), intent(inout):: AU
!!$    real(kind=kreal), dimension(NB,NB,NP ), intent(inout):: D
    real(kind=kreal), dimension(:)   , intent(inout):: B, X
    real(kind=kreal), dimension(:,:,:), intent(inout):: AL
    real(kind=kreal), dimension(:,:,:), intent(inout):: AU
    real(kind=kreal), dimension(:,:,:), intent(inout):: D

!!$    integer(kind=kint ), dimension(0:NP) ,intent(in) :: INU, INL
!!$    integer(kind=kint ), dimension(  NPL),intent(in) :: IAL
!!$    integer(kind=kint ), dimension(  NPU),intent(in) :: IAU
    integer(kind=kint ), dimension(0:) ,intent(in) :: INU, INL
    integer(kind=kint ), dimension(:),intent(in) :: IAL
    integer(kind=kint ), dimension(:),intent(in) :: IAU

    INTEGER(kind=kint ), INTENT(in) :: NK
    real(kind=kreal), pointer :: k_or_q_vec(:,:)


    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:), NOD_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:), NOD_EXPORT(:)

    real(kind=kreal), dimension(:),    allocatable       :: WS, WR
    real(kind=kreal), dimension(:,:),  allocatable       :: WW
    real(kind=kreal), dimension(:,:,:), allocatable, target :: ALU

    real(kind=kreal), dimension(:,:,:), allocatable :: AUlu0, ALlu0
    real(kind=kreal), dimension(:,:,:), allocatable :: Dlu0

    integer(kind=kint), dimension(:), allocatable :: inumFI1L, FI1L
    integer(kind=kint), dimension(:), allocatable :: inumFI1U, FI1U

    integer(kind=kint ) :: P, Q, R, Z, ZP, MAXIT
    real   (kind=kreal) :: TOL, W, SS


    !C LEVEL_NUM is added..
    integer(kind=kint ) :: LEVEL_NUM,ierr,i,j,k,l,m,col,ip
    INTEGER(kind=kint),save  :: WSIZE, NP_old, NB_old
    integer(kind=kint ) :: isL,ieL,isU,ieU,iq1,iq2,iq3,inod 
    integer(kind=kint ) :: SISIZE, SESIZE, WSIZE2, NP_max
    real   (kind=kreal) :: BNRM20, X1,X2,X3,D11,D22,D33
    real   (kind=kreal) :: WVAL(NB), v
    real   (kind=kreal) :: DNRM20,DNRM2,ALPHA,C10,C1,RHO,RHO0,RHO1,BETA
    logical, save :: initialize = .true.
    logical :: CONV_FL
    
    CHARACTER(len=20) :: PRECOND
    

    !C-- INIT.
    SISIZE = STACK_IMPORT(NEIBPETOT)
    SESIZE = STACK_EXPORT(NEIBPETOT)


    
    if(initialize .or. SETUP_FL) then
!!$       allocate(WVAL(NB))
       
       allocate (WW(NB*NP,4))
       NP_old = NP
!!$       NB_old = NB

       !C- decision of the allocation size for WS, WR
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)*NB
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)*NB
          end if
       end if
       if( WSIZE < NP*NB ) WSIZE = NP*NB

       ALLOCATE ( WS(WSIZE),WR(WSIZE) )

       initialize = .false.
    end if

    R = 1
    Z = 2
    Q = 2
    P = 3
    ZP= 4

    MAXIT  = ITER
!!$    TOL   = RESID           


    if(SETUP_FL) then
       
       if(INDPDT_AGGRE) then
          PRECOND ="IND1"
       else
          PRECOND ="SHA2"
       end if
       call data_creation_NB                                                   &
            &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU,           &
            &                  INU, IAU,                                       &
            &                  LEVEL_NUM, my_rank, NEIBPETOT, NEIBPE,          &
            &                  STACK_IMPORT, NOD_IMPORT,                       &
            &                  STACK_EXPORT, NOD_EXPORT,                       &
            &                  SOLVER_COMM, PRECOND, NB, k_or_q_vec, NK)     


    end if

    if(LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA_NB, LEVEL_NUM)
       
    else
       !C in the case, LEVEL_NUM == 0
       j = max(SISIZE, SESIZE)
       if(WSIZE < j*NB) WSIZE2 = j*NB
    end if
    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA_NB, LEVEL_NUM)
    
    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2), WR(WSIZE2))
       WSIZE = WSIZE2
    END if
    if(NP_max > NP_old) then
       deallocate(WW)
       allocate (WW(NB*NP_max,4))
       NP_old = NP_max
    end if



!!$    WSIZE2 = 0
!!$    
!!$    if(NEIBPETOT>0) then
!!$       if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
!!$          WSIZE2 = STACK_IMPORT(NEIBPETOT)
!!$       else
!!$          WSIZE2 = STACK_EXPORT(NEIBPETOT)
!!$       end if
!!$    end if
!!$    if(WSIZE2 < NP) WSIZE2 = NP
!!$
!!$    !C level 2..
!!$    if(NEIBPETOT > 0) then
!!$       m = WSIZE2
!!$       DO i = 2, LEVEL_NUM
!!$          j = HIERARCHICAL_DATA_NB(i) % COMM_TABLE % NEIBPETOT 
!!$          k = HIERARCHICAL_DATA_NB(i) % COMM_TABLE % STACK_IMPORT(j)
!!$          l = HIERARCHICAL_DATA_NB(i) % COMM_TABLE % STACK_EXPORT(j)
!!$          m = max(k,l,m)
!!$
!!$          j = HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % NEIBPETOT 
!!$          k = HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % STACK_IMPORT(j)
!!$          l = HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % STACK_EXPORT(j)
!!$          m = max(k,l,m)
!!$       END DO
!!$       
!!$       if(m > WSIZE2) then
!!$          !C-- WS, WR must be reallocated
!!$          WSIZE2 = m
!!$       end if
!!$    end if

!!$    if(WSIZE2 > WSIZE) then
!!$       deallocate(WS,WR)
!!$       allocate(WS(NB*WSIZE2),WR(NB*WSIZE2))
!!$       WSIZE = WSIZE2
!!$    END if
!!$    if(NP > NP_old) then
!!$       deallocate(WW)
!!$       allocate(WW(NB*NP,4))
!!$       NP_old = NP
!!$    END if
    !C-------- allocation of WS, WS, WW is finished.      
    

    !C
    !C +-----------------------+
    !C | {r0}= {b} - [A]{xini} |
    !C +-----------------------+
    !C===
    !C
    !C-- INTERFACE data EXCHANGE
    if (NEIBPETOT > 0) then
       call SOLVER_SEND_RECV_NB                                        &
            &   ( WSIZE,WSIZE,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,       &
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,my_rank, NB)
    end if

    !C
    !C-- BEGIN calculation
    
    !$omp parallel do private(j,k,l,col,WVAL)
    do i= 1, N
       do j = 1, NB
          WVAL(j) = B(NB*(i-1)+j)
          do k = 1, NB
             WVAL(j) = WVAL(j) - D(j,k,i)*X(NB*(i-1)+k)
          end do
       end do

       do l = INL(i-1)+1, INL(i)
          col = IAL(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AL(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do l = INU(i-1)+1, INU(i)
          col = IAU(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AU(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do j = 1, NB
          WW(NB*(i-1)+j,R) = WVAL(j)
       end do

    enddo
    !$omp end parallel do

!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$
!!$       !$omp parallel do reduction(+:BNRM20)
!!$       do i= 1, N*NB
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       !$omp end parallel do
!!$       
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$    else if(terminal_kind == 1) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(max:BNRM20), private(v)
!!$       do i= 1, N*NB
!!$          v = dabs(B(i))
!!$          if(BNRM20 < v) BNRM20 = v
!!$       enddo
!!$       !$omp end parallel do
!!$       
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_MAX, SOLVER_COMM, ierr)
!!$    else 
!!$       stop 'unknown terminal_kind: solver_AMGCG_NB.F90'
!!$    end if
!!$
!!$    if (BNRM2.eq.0.d0) BNRM2= 1.d0
    !C===

    do iter= 1, MAXIT
       
       !C +----------------+
       !C | {z}= [Minv]{r} |
       !C +----------------+
       !C===
       !$omp parallel do
       do i= 1, NB*NP
          WW(i, Z) = 0.d0
       end do
       !$omp end parallel do
       
       call count_time(1,SOLVER_COMM,my_rank,9)
       CALL v_cycle(WW(:,R),WW(:,Z),LEVEL_NUM,SOLVER_COMM,my_rank,WS,WR,WW(:,ZP),WSIZE)
!!$       CALL sgs_NB(N,NP,D,AL,INL,IAL,AU,INU,IAU,ALU,WW(:,R),WW(:,Z),1,1,SOLVER_COMM,&
!!$            &         my_rank,WS,WR,NP,NB)
       call count_time(2,SOLVER_COMM,my_rank,9)


!!$           DO i=1,NP*NB
!!$              WW(i,Z)=WW(i,R)
!!$           END DO

       !C +---------------+
       !C | {RHO}= {r}{z} |
       !C +---------------+
       !C===
       RHO0= 0.d0
       !$omp parallel do reduction(+:RHO0)       
       do i = 1, N*NB
          RHO0 = RHO0 + WW(i,R)*WW(i,Z)
       end do
       !$omp end parallel do       

       call MPI_allREDUCE (RHO0, RHO, 1, MPI_DOUBLE_PRECISION,           &
            &                    MPI_SUM, SOLVER_COMM, ierr)
       !C===

       !C
       !C +-----------------------------+
       !C | {p} = {z} if      ITER=1    |
       !C | BETA= RHO / RHO1  otherwise |
       !C +-----------------------------+
       !C===
       if ( ITER.eq.1 ) then
          !$omp parallel do          
          do i = 1, NB*N
             WW(i, P) = WW(i, Z)
          end do
          !$omp end parallel do          
       else
          
          BETA= RHO / RHO1

          !$omp parallel do          
          do i = 1, NB*N
             WW(i,P) = WW(i,Z) + BETA*WW(i,P)
          end do
          !$omp end parallel do
       endif
       !C===

       !C +-------------+
       !C | {q}= [A]{p} |
       !C +-------------+
       !C===        

       !C
       !C-- INTERFACE data EXCHANGE
       if(NEIBPETOT > 0) then
          call SOLVER_SEND_RECV_NB                                        &
               &   ( NP,NP,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,     &
               &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,P) , SOLVER_COMM,   &
               &     my_rank, NB)
       end if

       !$omp parallel do private(WVAL,j,k,l,inod,isU,ieU,isL,ieL)       
       do i= 1, N
          WVAL = 0.d0
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j)+D(j,k,i)*WW(NB*i-(NB-k),P)
             end do
          end do

          do l = INL(i-1)+1, INL(i)
             inod = IAL(l)
             do j = 1, NB
                do k = 1, NB
                   WVAL(j) = WVAL(j) + AL(j,k,l)*WW(NB*inod-(NB-k),P)
                end do
             end do
          end do

          do l = INU(i-1)+1, INU(i)
             inod = IAU(l)
             do j = 1, NB
                do k = 1, NB
                   WVAL(j) = WVAL(j) + AU(j,k,l)*WW(NB*inod-(NB-k), P)
                end do
             end do
          end do

          do j = 1, NB
             WW(NB*i-(NB-j),Q) = WVAL(j)
          end do

       enddo
       !$omp end parallel do


       !C +---------------------+
       !C | ALPHA= RHO / {p}{q} |
       !C +---------------------+
       C10= 0.d0
       !$omp parallel do reduction(+:C10)       
       do i = 1, NB*N
          C10 = C10+WW(i,P)*WW(i,Q)
       end do
       !$omp end parallel do       

       call MPI_allREDUCE (C10, C1, 1, MPI_DOUBLE_PRECISION,             &
            &                    MPI_SUM, SOLVER_COMM, ierr)
       ALPHA= RHO / C1


       !C +----------------------+
       !C | {x}= {x} + ALPHA*{p} |
       !C | {r}= {r} - ALPHA*{q} |
       !C +----------------------+

       !$omp parallel do        
       do i = 1, NB*N
          X(i) = X(i) +ALPHA * WW(i,P)
          WW(i,R) = WW(i,R) - ALPHA * WW(i,Q)
       end do
       !$omp end parallel do       

       call terminal_cond_check(CONV_FL, RESID, WW(:,R), NB*N, SOLVER_COMM)
       

!!$       DNRM20= 0.d0
!!$       do i = 1, NB*N
!!$          DNRM20 = DNRM20 + WW(i, R)**2
!!$       end do
!!$
!!$
!!$
!!$       call MPI_allREDUCE (DNRM20, DNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$
!!$       RESID= dsqrt(DNRM2/BNRM2)

#ifdef PRINT_REZ       
       if (my_rank.eq.0 .and. terminal_kind == 0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif       


       if (CONV_FL) then
          REACH_CONV_FL = .true.
          exit
       end if
       RHO1 = RHO                                                             
    enddo

!!$    call SOLVER_SEND_RECV_NB                                                 &
!!$         &   ( NP,NP,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,        &
!!$         &     STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,my_rank,NB)
    
    
    !C differentiate divergence cases from non-convergent cases
    if(conv_check_fl)then
       if(.not.CONV_FL) then
          call cal_relative_res_norm_NB(RESID, X, B, WS, WR, WSIZE, NB)
          if(RESID > 1.0) then
             iter = -1
          end if
       end if
    end if
    
#ifdef PRINT_INFO
    call count_time(4,solver_comm,my_rank,1)      
#endif

    if(SETUP_FL) then
       call clear_matrix_ssi_amg(LEVEL_NUM)
       deallocate (WW)
       deallocate (WR)
       deallocate (WS)
       initialize = .true.
    end if

  end subroutine        AMGCG_NB


  SUBROUTINE v_cycle(problem_B,problem_X,LEVEL_NUM,SOLVER_COMM,my_rank,WS,WR,Temp,WSIZE)
    USE data_structure_for_AMG_NB
    USE comm_subs_NB
    USE solver_SR_NB
    USE count_time_mod
    USE amg_parameter, only: iteration_number_up, iteration_number_down, &
         & iteration_number_bottom, gs_kind, W_OUT_LU, &
         & iterative_parameter_set_at_each_lev

    IMPLICIT NONE
    include 'mpif.h'
    INTEGER(kind=kint ),  INTENT(in)              :: LEVEL_NUM
    INTEGER(kind=kint ),  INTENT(in)              :: WSIZE

    INTEGER(kind=kint )                           :: N, NP, NB, NB_c, NB_f
    REAL   (kind=kreal), DIMENSION(:,:,:), pointer:: D
    !C DIMENSION(3,3,NP)
    REAL   (kind=kreal), DIMENSION(: ), target    :: problem_B
    !C DIMENSION(3*N)
    REAL   (kind=kreal), DIMENSION(: ), target    :: problem_X   
    !C DIMENSION(3*N)

    REAL   (kind=kreal), DIMENSION(: ), pointer   :: B
    !C DIMENSION(3*N)
    REAL   (kind=kreal), DIMENSION(: ), pointer   :: X   
    !C DIMENSION(3*N)

    REAL   (kind=kreal), DIMENSION(:,:,:), pointer:: AU
    REAL   (kind=kreal), DIMENSION(:,:,:), pointer:: AL
    REAL   (kind=kreal), DIMENSION(:,:,:), pointer:: ALU

    INTEGER(kind=kint ), DIMENSION(:), pointer    :: INU
    !C DIMENSION(0:N)
    INTEGER(kind=kint ), DIMENSION(:), pointer    :: IAU
    !C DIMENSION(NPU)
    INTEGER(kind=kint ), DIMENSION(:), pointer    :: INL
    !C DIMENSION(0:N)
    INTEGER(kind=kint ), DIMENSION(:), pointer    :: IAL

    INTEGER, intent(in) :: SOLVER_COMM
    INTEGER(kind=kint), intent(in) :: my_rank
    INTEGER :: ierr

    REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_B
    REAL   (kind=kreal), DIMENSION(:), pointer :: coarser_X
    REAL   (kind=kreal), DIMENSION(:) :: Temp
    REAL   (kind=kreal), DIMENSION(:) :: WS, WR

    INTEGER(kind=kint ) :: nth_lev,i,inod,j,ieL,isL,is,ie,coarser_N,isU,ieU,k,coarser_NP,l
    REAL   (kind=kreal) :: R_norm,GR_norm
    REAL   (kind=kreal) :: R1,R2,R3,X1,X2,X3,B1,B2,B3
    REAL   (kind=kreal) :: RES_V(100)
    !C max number of NB is assumed to be 100.

    type(INTER_LEVEL_OPERATOR), pointer :: R, P

    INTEGER(kind=kint)          :: NEIBPETOT
    INTEGER(kind=kint), DIMENSION(:), pointer :: NEIBPE(:)
    INTEGER(kind=kint), DIMENSION(:), pointer :: STACK_IMPORT(:),STACK_EXPORT(:)
    INTEGER(kind=kint), DIMENSION(:), pointer :: NOD_IMPORT(:),NOD_EXPORT(:)
    INTEGER :: INFO


    HIERARCHICAL_DATA_NB(1) % B => problem_B
    HIERARCHICAL_DATA_NB(1) % X => problem_X

    DO nth_lev = 1, LEVEL_NUM-1

       call iterative_parameter_set_at_each_lev(nth_lev)

#ifdef MEASURING
       CALL count_time(1, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif


       NB = HIERARCHICAL_DATA_NB(nth_lev) % NB
       N  = HIERARCHICAL_DATA_NB(nth_lev) % N
       NP = HIERARCHICAL_DATA_NB(nth_lev) % NP
       INU=>HIERARCHICAL_DATA_NB(nth_lev) % INU 
       INL=>HIERARCHICAL_DATA_NB(nth_lev) % INL 
       X  =>HIERARCHICAL_DATA_NB(nth_lev) % X
       B  =>HIERARCHICAL_DATA_NB(nth_lev) % B
       D  =>HIERARCHICAL_DATA_NB(nth_lev) % D
       IAU=>HIERARCHICAL_DATA_NB(nth_lev) % IAU
       IAL=>HIERARCHICAL_DATA_NB(nth_lev) % IAL
       ALU=>HIERARCHICAL_DATA_NB(nth_lev) % ALU
       AU =>HIERARCHICAL_DATA_NB(nth_lev) % AU
       AL =>HIERARCHICAL_DATA_NB(nth_lev) % AL
       STACK_IMPORT => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % STACK_IMPORT
       STACK_EXPORT => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % STACK_EXPORT
       NOD_IMPORT   => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NOD_IMPORT
       NOD_EXPORT   => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NOD_EXPORT
       NEIBPE       => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPE
       NEIBPETOT    =  HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPETOT

       NEIBPETOT = HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPETOT

       coarser_B => HIERARCHICAL_DATA_NB(nth_lev+1) % B
       coarser_X => HIERARCHICAL_DATA_NB(nth_lev+1) % X


       CALL count_time(1,SOLVER_COMM,my_rank,8)
       CALL sgs_NB(N, NP, D, AL, INL, IAL, AU, INU, IAU, ALU, B, X, &
            &      iteration_number_down, nth_lev,SOLVER_COMM,      &
            &      my_rank, WS, WR, WSIZE, NB, 0)
       
       CALL count_time(2,SOLVER_COMM,my_rank,8)


       !C calculate residual
       DO j= 1, N
          DO k = 1, NB
             RES_V(k) = B((j-1)*NB+k)
             DO l = 1, NB
                RES_V(k) = RES_V(k) - D(k,l,j)*X((j-1)*NB+l)
             END DO
          END DO


          isU= INU(j-1)+1
          ieU= INU(j)
          DO i= isU, ieU
             inod=IAU(i)
             DO k = 1, NB
                DO l = 1, NB
                   RES_V(k) = RES_V(k) - AU(k,l,i)*X((inod-1)*NB+l)
                END DO
             END DO
          END DO

          isL= INL(j-1)+1
          ieL= INL(j)
          DO i= isL, ieL
             inod= IAL(i)
             DO k = 1, NB
                DO l = 1, NB
                   RES_V(k) = RES_V(k) - AL(k,l,i)*X((inod-1)*NB+l)
                END DO
             END DO
          END DO

          DO k = 1, NB
             Temp((j-1)*NB+k) = RES_V(k)
          END DO
       END DO
       
#ifdef MEASURING
       CALL count_time(2, SOLVER_COMM,  my_rank, 40+nth_lev)
       CALL count_time(1, SOLVER_COMM,  my_rank, 50+nth_lev)

#endif

       !C restrict the residual vector
       R => HIERARCHICAL_DATA_NB(nth_lev+1) % R
       coarser_N  = HIERARCHICAL_DATA_NB(nth_lev+1) % N
       coarser_NP = HIERARCHICAL_DATA_NB(nth_lev+1) % NP
       NB_c = HIERARCHICAL_DATA_NB(nth_lev+1) % NB
       NB_f = HIERARCHICAL_DATA_NB(nth_lev) % NB
       if(.not.HIERARCHICAL_DATA_NB(nth_lev+1) % PRO_AGGRE_OPT) then
          DO j= 1, coarser_NP
             is= R % IN(j-1)+1
             ie= R % IN(j)


             DO k = 1, NB_c
                RES_V(k) = 0
             END DO

             DO i = is, ie
                inod = R % CN(i)

                DO k = 1, NB_c
                   DO l = 1, NB_f
                      RES_V(k) = RES_V(k) + R % V((i-1)*NB_c*NB_f+(k-1)*NB_f+l)*Temp((inod-1)*NB_f+l)
                   END DO
                END DO

             END DO

             DO k = 1, NB_c
                coarser_B((j-1)*NB_c+k) = RES_V(k)
             END DO

          END DO
          
          CALL SOLVER_SR_REDUCE(coarser_NP, &
               & WSIZE,&
               & WSIZE,&
               & HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NEIBPETOT,    &
               & HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NEIBPE,       &
               & HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % STACK_IMPORT, &
               & HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NOD_IMPORT,   &
               & HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % STACK_EXPORT, &
               & HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NOD_EXPORT,   &
               & WS, WR, coarser_B, SOLVER_COMM, my_rank,NB_c)

       else
          call SOLVER_SEND_RECV_NB                                           &
               &( WSIZE,WSIZE,                                               &
               &  HIERARCHICAL_DATA_NB(nth_lev) % NP,                        &
               &  HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPETOT,    &
               &  HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPE,       &
               &  HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % STACK_IMPORT, &
               &  HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NOD_IMPORT,   &
               &  HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % STACK_EXPORT, &
               &  HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NOD_EXPORT,   &
               &  WS, WR, Temp, SOLVER_COMM, my_rank, NB_f)
          

          DO j= 1, R % ROW_SIZE
             is= R % IN(j-1)+1
             ie= R % IN(j)

             DO k = 1, NB_c
                RES_V(k) = 0
             END DO

             DO i = is, ie
                inod = R % CN(i)
                DO k = 1, NB_c
                   DO l = 1, NB_f
                      RES_V(k) = RES_V(k) + R % V((i-1)*NB_c*NB_f+(k-1)*NB_f+l)*Temp((inod-1)*NB_f+l)
                   END DO
                END DO
             END DO

             DO k = 1, NB_c
                coarser_B((j-1)*NB_c+k) = RES_V(k)
             END DO
          END DO

          call SOLVER_SEND_RECV_NB                                                &
               &( WSIZE,WSIZE,                                                    &
               &  HIERARCHICAL_DATA_NB(nth_lev+1) % NP,                           &
               &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NEIBPETOT,    &
               &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NEIBPE,       &
               &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % STACK_EXPORT, &
               &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NOD_EXPORT,   &
               &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % STACK_IMPORT, &
               &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NOD_IMPORT,   &
               &  WS, WR, coarser_B, SOLVER_COMM, my_rank, NB_c)

       end if
       
       DO j=1,coarser_NP*NB_c
          coarser_X(j)= 0.0
       END DO
#ifdef MEASURING
       CALL count_time(2, SOLVER_COMM,  my_rank, 50+nth_lev)

#endif

    END DO


#ifdef MEASURING
    CALL count_time(1, SOLVER_COMM,  my_rank, 40+LEVEL_NUM)
#endif

    call iterative_parameter_set_at_each_lev(LEVEL_NUM)
    
    NB =  HIERARCHICAL_DATA_NB(LEVEL_NUM) % NB
    N  =  HIERARCHICAL_DATA_NB(LEVEL_NUM) % N
    NP =  HIERARCHICAL_DATA_NB(LEVEL_NUM) % NP
    X  => HIERARCHICAL_DATA_NB(LEVEL_NUM) % X
    B  => HIERARCHICAL_DATA_NB(LEVEL_NUM) % B

    if(.not.W_OUT_LU) then
       CALL count_time(1,SOLVER_COMM,my_rank,8)
       if(N > 0) then
          X = B
          call DGETRS('N', N*NB, 1, coarsest_A_NB, N*NB, coarsest_IDX_NB, X, N*NB, INFO)
          if(INFO/=0) stop 'forward/backward substitution error!!'
       end if
       CALL count_time(2,SOLVER_COMM,my_rank,8)
       
    else

       NEIBPETOT    =  HIERARCHICAL_DATA_NB(LEVEL_NUM) % COMM_TABLE % NEIBPETOT 
       NEIBPE       => HIERARCHICAL_DATA_NB(LEVEL_NUM) % COMM_TABLE % NEIBPE
       STACK_IMPORT => HIERARCHICAL_DATA_NB(LEVEL_NUM) % COMM_TABLE % STACK_IMPORT
       STACK_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NUM) % COMM_TABLE % STACK_EXPORT
       NOD_IMPORT   => HIERARCHICAL_DATA_NB(LEVEL_NUM) % COMM_TABLE % NOD_IMPORT
       NOD_EXPORT   => HIERARCHICAL_DATA_NB(LEVEL_NUM) % COMM_TABLE % NOD_EXPORT

       INU=>HIERARCHICAL_DATA_NB(LEVEL_NUM) % INU 
       INL=>HIERARCHICAL_DATA_NB(LEVEL_NUM) % INL 
       D => HIERARCHICAL_DATA_NB(LEVEL_NUM) % D
       IAU=>HIERARCHICAL_DATA_NB(LEVEL_NUM) % IAU
       IAL=>HIERARCHICAL_DATA_NB(LEVEL_NUM) % IAL
       ALU=>HIERARCHICAL_DATA_NB(LEVEL_NUM) % ALU
       AU =>HIERARCHICAL_DATA_NB(LEVEL_NUM) % AU
       AL =>HIERARCHICAL_DATA_NB(LEVEL_NUM) % AL


       CALL count_time(1,SOLVER_COMM,my_rank,8)
       CALL sgs_NB(N, NP, D, AL, INL, IAL, AU, INU, IAU, ALU, B, X, &
            &      iteration_number_bottom, nth_lev,SOLVER_COMM,    &
            &      my_rank, WS, WR, WSIZE, NB, 0)
       CALL count_time(2,SOLVER_COMM,my_rank,8)
       !C barrier in ssor ########



    end if

#ifdef MEASURING
    CALL count_time(2, SOLVER_COMM,  my_rank, 40+LEVEL_NUM)
#endif



    
    DO nth_lev=LEVEL_NUM-1,1,-1
       call iterative_parameter_set_at_each_lev(nth_lev)
       
#ifdef MEASURING
       CALL count_time(1, SOLVER_COMM,  my_rank, 50+nth_lev)
#endif


       NB_f = HIERARCHICAL_DATA_NB(nth_lev) % NB
       N= HIERARCHICAL_DATA_NB(nth_lev) % N
       NP=HIERARCHICAL_DATA_NB(nth_lev) % NP

       INU          => HIERARCHICAL_DATA_NB(nth_lev) % INU
       INL          => HIERARCHICAL_DATA_NB(nth_lev) % INL
       X            => HIERARCHICAL_DATA_NB(nth_lev) % X
       B            => HIERARCHICAL_DATA_NB(nth_lev) % B
       D            => HIERARCHICAL_DATA_NB(nth_lev) % D
       IAU          => HIERARCHICAL_DATA_NB(nth_lev) % IAU
       IAL          => HIERARCHICAL_DATA_NB(nth_lev) % IAL
       ALU          => HIERARCHICAL_DATA_NB(nth_lev) % ALU
       AU           => HIERARCHICAL_DATA_NB(nth_lev) % AU
       AL           => HIERARCHICAL_DATA_NB(nth_lev) % AL
       STACK_IMPORT => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % STACK_IMPORT
       STACK_EXPORT => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % STACK_EXPORT
       NOD_IMPORT   => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NOD_IMPORT
       NOD_EXPORT   => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NOD_EXPORT
       NEIBPE       => HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPE
       NEIBPETOT    =  HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPETOT
       coarser_X    => HIERARCHICAL_DATA_NB(nth_lev+1) % X
       NB_c = HIERARCHICAL_DATA_NB(nth_lev+1) % NB
       P => HIERARCHICAL_DATA_NB(nth_lev+1) % P
          
       call SOLVER_SEND_RECV_NB                                                &
            &( WSIZE,WSIZE,                                                    &
            &  HIERARCHICAL_DATA_NB(nth_lev+1) % NP,                           &
            &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NEIBPETOT,    &
            &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NEIBPE,       &
            &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % STACK_IMPORT, &
            &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NOD_IMPORT,   &
            &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % STACK_EXPORT, &
            &  HIERARCHICAL_DATA_NB(nth_lev+1) % INT_LVL_TABLE % NOD_EXPORT,   &
            &  WS, WR, coarser_X, SOLVER_COMM,my_rank, NB_c)

       
       if(.not. HIERARCHICAL_DATA_NB(nth_lev+1) % PRO_AGGRE_OPT) then
          
          DO j= 1, N
             is= P % IN(j-1)+1
             ie= P % IN(j)

             DO k = 1, NB_f
                RES_V(k) = 0
             END DO

             DO i=is, ie
                inod= P % CN(i)
             
                DO k = 1, NB_f
                   DO l = 1, NB_c
                      RES_V(k) = RES_V(k) + P % V((i-1)*NB_f*NB_c+(k-1)*NB_c+l) * coarser_X((inod-1)*NB_c+l)
                   END DO
                END DO
             END DO

             DO k = 1, NB_f
                X((j-1)*NB_f+k) = X((j-1)*NB_f+k)+RES_V(k)
             END DO
          END DO
          
       else
          
          DO j= 1, NP
             is= P % IN(j-1)+1
             ie= P % IN(j)

             DO k = 1, NB_f
                RES_V(k) = 0
             END DO

             DO i=is, ie
                inod= P % CN(i)
             
                DO k = 1, NB_f
                   DO l = 1, NB_c
                      RES_V(k) = RES_V(k) + P % V((i-1)*NB_f*NB_c+(k-1)*NB_c+l) * coarser_X((inod-1)*NB_c+l)
                   END DO
                END DO
             END DO

             DO k = 1, NB_f
                Temp((j-1)*NB_f+k) = RES_V(k)
             END DO

          END DO
          
          CALL SOLVER_SR_REDUCE(NP, &
               & WSIZE,&
               & WSIZE,&
               & HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPETOT,    &
               & HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NEIBPE,       &
               & HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % STACK_IMPORT, &
               & HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NOD_IMPORT,   &
               & HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % STACK_EXPORT, &
               & HIERARCHICAL_DATA_NB(nth_lev) % COMM_TABLE % NOD_EXPORT,   &
               & WS, WR, Temp, SOLVER_COMM, my_rank, NB_f)


          DO j = 1, N*NB_f
             X(j) = X(j)+Temp(j)
          END DO

       end if
       
#ifdef MEASURING
       CALL count_time(2, SOLVER_COMM,  my_rank, 50+nth_lev)
       CALL count_time(1, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

       call SOLVER_SEND_RECV_NB                                              &
            & ( WSIZE,WSIZE,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
            &   STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,my_rank, NB_f)
       
       CALL count_time(1,SOLVER_COMM,my_rank,8)         
       CALL sgs_NB(N, NP, D, AL, INL, IAL, AU, INU, IAU, ALU, B, X, &
            &      iteration_number_up, nth_lev, SOLVER_COMM,       &
            &      my_rank, WS, WR, WSIZE, NB_f, 0)

       CALL count_time(2,SOLVER_COMM,my_rank,8)

#ifdef MEASURING
       CALL count_time(2, SOLVER_COMM,  my_rank, 40+nth_lev)
#endif

    END DO

  END SUBROUTINE v_cycle
  

  subroutine debug_norm(vec,size,id, SOLVER_COMM)
    implicit none
    include 'mpif.h'
    integer :: SOLVER_COMM
    integer(kind=4) :: id,size, my_rank
    REAL(kind=8) ::vec(:)
    
    integer(kind=4) :: i, ierr
    real(kind=8) ::v,gv

    call MPI_COMM_RANK(SOLVER_COMM, my_rank,ierr)
    
    v=0.0
    DO i =1, size
       v = v+vec(i)*vec(i)
    END DO
    
    call MPI_allREDUCE (v, gv, 1, MPI_DOUBLE_PRECISION,       &
         &                    MPI_SUM, SOLVER_COMM, ierr)
    
    if(my_rank==0) write(*,*) "###", id,":",dsqrt(gv) ,"###"
    
  end subroutine debug_norm

  

  SUBROUTINE sgs_NB(N,NP,D,AL,INL,IAL,AU,INU,IAU,ALU,B,X,count,LEVEL_NO,SOLVER_COMM,&
       &         my_rank,WS,WR,WSIZE,NB, kind)
    USE data_structure_for_AMG_NB
    USE solver_SR_NB
    USE amg_parameter, only: dump_gauss_seidel_weight
    USE count_time_mod
    IMPLICIT NONE
    !C kind = 0,1,2 ::   sgs, gs, gs in reverse order
    INTEGER(kind=kint ),INTENT(in )  :: N,NP,count, NB,kind

    REAL   (kind=kreal), DIMENSION(:,:,:):: D
    REAL   (kind=kreal), DIMENSION(:)    :: B
    REAL   (kind=kreal), DIMENSION(:)    :: X

    REAL   (kind=kreal), DIMENSION(:,:,:):: AU
    REAL   (kind=kreal), DIMENSION(:,:,:):: AL
    REAL   (kind=kreal), DIMENSION(:,:,:):: ALU

    INTEGER(kind=kint ), DIMENSION(0:)   :: INU
    INTEGER(kind=kint ), DIMENSION(:)    :: IAU
    INTEGER(kind=kint ), DIMENSION(0:)   :: INL
    INTEGER(kind=kint ), DIMENSION(:)    :: IAL

    INTEGER,             intent(in)      :: SOLVER_COMM
    INTEGER(kind=kint ), intent(in)      :: my_rank
    INTEGER                              :: ierr
    INTEGER(kind=kint ), intent(in)      :: LEVEL_NO
    INTEGER(kind=kint ), intent(in)      :: WSIZE

    INTEGER(kind=kint ), DIMENSION(:),pointer :: NEIBPE
    INTEGER(kind=kint ), DIMENSION(:),pointer :: STACK_IMPORT,STACK_EXPORT
    INTEGER(kind=kint ), DIMENSION(:),pointer :: NOD_EXPORT,NOD_IMPORT
    INTEGER(kind=kint )                       :: NEIBPETOT

    REAL   (kind=kreal), DIMENSION(:) :: WS,WR

    INTEGER(kind=kint ):: cnt,isL,ieL,isU,ieU,i,j,k,l,inod
    REAL   (kind=kreal):: R_val, R_norm,GR_norm
    REAL   (kind=kreal):: X1,X2,X3,SW1,SW2,SW3,omega
    REAL   (kind=kreal), allocatable :: SW(:)
    integer :: info


    omega = dump_gauss_seidel_weight

    NEIBPETOT    =  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPETOT 
    NEIBPE       => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPE
    STACK_IMPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_IMPORT
    STACK_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_EXPORT
    NOD_IMPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_IMPORT
    NOD_EXPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_EXPORT

    allocate(SW(NB))

    DO cnt=1,count
       if(kind==1 .or. kind==0) then
          call count_time(1,solver_comm,my_rank,10)
          DO i=1,N
             DO j = 1, NB
                SW(j) = B(NB*(i-1)+j)
             END DO

             isL=INL(i-1)+1
             ieL=INL(i)
             DO l=isL, ieL
                inod=IAL(l)
                DO j = 1, NB
                   DO k = 1, NB
                      SW(j) = SW(j)-AL(j,k,l)*X(NB*(inod-1)+k)
                   END DO
                END DO
             END DO

             isU=INU(i-1)+1
             ieU=INU(i)
             DO l=isU, ieU
                inod=IAU(l)
                DO j = 1, NB
                   DO k = 1, NB
                      SW(j) = SW(j)-AU(j,k,l)*X(NB*(inod-1)+k)
                   END DO
                END DO
             END DO


             !C            X(i) = X(i)/D(i)
             !C            SW(:) --- X(i)

             call DGETRS('N',NB,1,ALU(1:NB,1:NB,i),NB,ALU(:,NB+1,i),SW,NB,info)
             if(info/=0) stop "forward backward substitution error in sgs_NB()"

             DO j = 1, NB
                X(NB*(i-1)+j) = omega*SW(j)+(1-omega)*X(NB*(i-1)+j)
             END DO

          END DO
          
          call count_time(2,solver_comm,my_rank,10)                
          call SOLVER_SEND_RECV_NB                                                &
            &     ( WSIZE,WSIZE,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,  &
            &       STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,            &
            &       my_rank,NB)
       end if
          
       

       if(kind==2 .or. kind==0) then
          call count_time(1,solver_comm,my_rank,10)                


          DO i=N,1,-1
             !C           X(i)= B(i)
             DO j = 1, NB
                SW(j) = B(NB*(i-1)+j)
             END DO

             isL=INL(i-1)+1
             ieL=INL(i)
             DO l = isL, ieL
                inod = IAL(l)
                DO j = 1, NB
                   DO k = 1, NB
                      SW(j) = SW(j) - AL(j,k,l)*X(NB*(inod-1)+k)
                   END DO
                END DO
             END DO

             isU=INU(i-1)+1
             ieU=INU(i)
             DO l = isU, ieU
                inod = IAU(l)
                DO j = 1, NB
                   DO k = 1, NB
                      SW(j) = SW(j) - AU(j,k,l)*X(NB*(inod-1)+k)
                   END DO
                END DO
             END DO



            !C            X(i) = X(i)/D(i)
             call DGETRS('N',NB,1,ALU(1:NB,1:NB,i),NB,ALU(:,NB+1,i),SW,NB,info)
             if(info/=0) stop "forward backward substitution error in sgs_NB()"

             DO j = 1, NB
                X(NB*(i-1)+j) = omega*SW(j)+(1-omega)*X(NB*(i-1)+j)
             END DO



          END DO
          CALL count_time(2,solver_comm,my_rank,10)       
          call SOLVER_SEND_RECV_NB                                                &
               &     ( WSIZE,WSIZE,NP,NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,&
               &       STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,         &
               &       my_rank, NB)
          
       end if

       !C-- residual check
!!$         SW = 0.d0
!!$         R_val = 0.d0
!!$         DO i = 1, N
!!$            DO j = 1, NB
!!$               SW(j) = B(NB*i-(NB-j))
!!$            END DO
!!$            DO j = 1, NB
!!$               DO k = 1, NB
!!$                  SW(j) = SW(j) - D(j,k,i)*X(NB*i-(NB-k))
!!$               END DO
!!$            END DO
!!$            
!!$            isl=INL(i-1)+1;iel=INL(i)
!!$            DO l = isl, iel
!!$               inod = IAL(l)
!!$               DO j = 1, NB
!!$                  DO k = 1, NB
!!$                     SW(j) = SW(j) - AL(j,k,l)*X(NB*inod-(NB-k))
!!$                  END DO
!!$               END DO
!!$            END DO
!!$
!!$            isu=INU(i-1)+1;ieu=INU(i)
!!$            DO l = isu, ieu
!!$               inod = IAU(l)
!!$               DO j = 1, NB
!!$                  DO k = 1, NB
!!$                     SW(j) = SW(j) - AU(j,k,l)*X(NB*inod-(NB-k))
!!$                  END DO
!!$               END DO
!!$            END DO
!!$            
!!$            DO j = 1, NB
!!$               R_val = R_val + SW(j)**2
!!$            END DO
!!$            
!!$         END DO
!!$         
!!$         write(*,*) dsqrt(R_val), "##########"
!!$         
    END DO
    deallocate(SW)

  END SUBROUTINE sgs_NB


  SUBROUTINE jacobi(N,NP,D,AL,INL,IAL,AU,INU,IAU,ALU,B,X,count,LEVEL_NO,SOLVER_COMM,&
       &         my_rank,WS,WR,WSIZE,NB)
    USE data_structure_for_AMG_NB
    USE solver_SR_NB
    USE amg_parameter, only: dump_gauss_seidel_weight
    USE count_time_mod

    IMPLICIT NONE
    !C kind = 0,1,2 ::   sgs, gs, gs in reverse order
    INTEGER(kind=kint ),INTENT(in )  :: N,NP,count, NB

    REAL   (kind=kreal), DIMENSION(:,:,:):: D
    REAL   (kind=kreal), DIMENSION(:)    :: B
    REAL   (kind=kreal), DIMENSION(:)    :: X

    REAL   (kind=kreal), DIMENSION(:,:,:):: AU
    REAL   (kind=kreal), DIMENSION(:,:,:):: AL
    REAL   (kind=kreal), DIMENSION(:,:,:):: ALU

    INTEGER(kind=kint ), DIMENSION(0:)   :: INU
    INTEGER(kind=kint ), DIMENSION(:)    :: IAU
    INTEGER(kind=kint ), DIMENSION(0:)   :: INL
    INTEGER(kind=kint ), DIMENSION(:)    :: IAL

    INTEGER,             intent(in)      :: SOLVER_COMM
    INTEGER(kind=kint ), intent(in)      :: my_rank
    INTEGER                              :: ierr
    INTEGER(kind=kint ), intent(in)      :: LEVEL_NO
    INTEGER(kind=kint ), intent(in)      :: WSIZE

    INTEGER(kind=kint ), DIMENSION(:),pointer :: NEIBPE
    INTEGER(kind=kint ), DIMENSION(:),pointer :: STACK_IMPORT,STACK_EXPORT
    INTEGER(kind=kint ), DIMENSION(:),pointer :: NOD_EXPORT,NOD_IMPORT
    INTEGER(kind=kint )                       :: NEIBPETOT

    REAL   (kind=kreal), DIMENSION(:) :: WS,WR

    INTEGER(kind=kint ):: cnt,isL,ieL,isU,ieU,i,j,k,l,inod
    REAL   (kind=kreal):: R_val, R_norm,GR_norm
    REAL   (kind=kreal):: X1,X2,X3,SW1,SW2,SW3,omega
    REAL   (kind=kreal), allocatable :: SW(:), X_old(:)

    omega = dump_gauss_seidel_weight

    NEIBPETOT    =  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPETOT 
    NEIBPE       => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPE
    STACK_IMPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_IMPORT
    STACK_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_EXPORT
    NOD_IMPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_IMPORT
    NOD_EXPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_EXPORT

    allocate(SW(NB),X_old(NP*NB))

    DO cnt=1,count
       X_old(:) = X(:)
       call count_time(1,solver_comm,my_rank,10)
       DO i=1,N
          DO j = 1, NB
             SW(j) = B(NB*(i-1)+j)
          END DO

          isL=INL(i-1)+1
          ieL=INL(i)
          DO l=isL, ieL
             inod=IAL(l)
             DO j = 1, NB
                DO k = 1, NB
                   SW(j) = SW(j)-AL(j,k,l)*X_old(NB*(inod-1)+k)
                END DO
             END DO
          END DO

          isU=INU(i-1)+1
          ieU=INU(i)
          DO l=isU, ieU
             inod=IAU(l)
             DO j = 1, NB
                DO k = 1, NB
                   SW(j) = SW(j)-AU(j,k,l)*X_old(NB*(inod-1)+k)
                END DO
             END DO
          END DO
          
          DO j = 2, NB
             DO k = 1, j-1
                SW(j) = SW(j) - ALU(j,k,i)*SW(k)
             END DO
          END DO

          DO j = NB, 1, -1
             DO k = j+1, NB
                SW(j) = SW(j) - ALU(j,k,i)*SW(k)
             END DO
             SW(j) = ALU(j,j,i)*SW(j)
          END DO

          DO j = 1, NB
             X(NB*(i-1)+j) = omega*SW(j)+(1-omega)*X_old(NB*(i-1)+j)
          END DO

       END DO

       call count_time(2,solver_comm,my_rank,10)                
       call SOLVER_SEND_RECV_NB                                                &
            &     ( WSIZE,WSIZE,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,  &
            &       STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,            &
            &       my_rank,NB)

    END DO
    deallocate(SW)

  END SUBROUTINE jacobi

  subroutine cal_res_norm_NB_lev(X, B, WS, WR, WSIZE, NB, LEVEL_NO, &
       & my_rank,SOLVER_COMM)
    USE data_structure_for_AMG_NB
    use solver_SR_NB      

    implicit none
    include 'mpif.h'

    INTEGER(kind=kint ), intent(in )  :: NB
    INTEGER(kind=kint ), intent(in)   :: LEVEL_NO
    INTEGER(kind=kint ), intent(in)   :: WSIZE
    INTEGER,             intent(in)   :: SOLVER_COMM
    INTEGER,             intent(in)   :: my_rank
    REAL(kind=kreal) :: res_norm
    
    INTEGER(kind=kint ) :: N,NP
    REAL   (kind=kreal), DIMENSION(:) :: B
    REAL   (kind=kreal), DIMENSION(:) :: X

    REAL   (kind=kreal), DIMENSION(:,:,:), pointer:: D

    REAL   (kind=kreal), DIMENSION(:,:,:), pointer:: AU
    REAL   (kind=kreal), DIMENSION(:,:,:), pointer:: AL
    REAL   (kind=kreal), DIMENSION(:,:,:), pointer:: ALU

    INTEGER(kind=kint ), DIMENSION(:), pointer   :: INU
    INTEGER(kind=kint ), DIMENSION(:), pointer    :: IAU
    INTEGER(kind=kint ), DIMENSION(:), pointer   :: INL
    INTEGER(kind=kint ), DIMENSION(:), pointer    :: IAL

    INTEGER(kind=kint ), DIMENSION(:),pointer :: NEIBPE
    INTEGER(kind=kint ), DIMENSION(:),pointer :: STACK_IMPORT,STACK_EXPORT
    INTEGER(kind=kint ), DIMENSION(:),pointer :: NOD_EXPORT,NOD_IMPORT
    INTEGER(kind=kint )                       :: NEIBPETOT

    REAL   (kind=kreal), DIMENSION(:) :: WS,WR


    REAL(kind=kreal) :: WVAL(NB), res_norm0
    INTEGER(kind=kint) :: i,j,k,l,col, ierr

    N  =  HIERARCHICAL_DATA_NB(LEVEL_NO) % N
    NP =  HIERARCHICAL_DATA_NB(LEVEL_NO) % NP

    INU          => HIERARCHICAL_DATA_NB(LEVEL_NO) % INU
    INL          => HIERARCHICAL_DATA_NB(LEVEL_NO) % INL
    D            => HIERARCHICAL_DATA_NB(LEVEL_NO) % D
    IAU          => HIERARCHICAL_DATA_NB(LEVEL_NO) % IAU
    IAL          => HIERARCHICAL_DATA_NB(LEVEL_NO) % IAL
    ALU          => HIERARCHICAL_DATA_NB(LEVEL_NO) % ALU
    AU           => HIERARCHICAL_DATA_NB(LEVEL_NO) % AU
    AL           => HIERARCHICAL_DATA_NB(LEVEL_NO) % AL
    STACK_IMPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_IMPORT
    STACK_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_EXPORT
    NOD_IMPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_IMPORT
    NOD_EXPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_EXPORT
    NEIBPE       => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPE
    NEIBPETOT    =  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPETOT


    call SOLVER_SEND_RECV_NB                                      &
         &   ( WSIZE, WSIZE, NP, NEIBPETOT, NEIBPE, STACK_IMPORT,&
         &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
         &     WS, WR, X, SOLVER_COMM,  &
         &     my_rank, NB)


    res_norm0 = 0.0

    do i= 1, N
       do j = 1, NB
          WVAL(j) = B(NB*(i-1)+j)
          do k = 1, NB
             WVAL(j) = WVAL(j) - D(j,k,i)*X(NB*(i-1)+k)
          end do
       end do
       
       do l = INL(i-1)+1, INL(i)
          col = IAL(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AL(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do l = INU(i-1)+1, INU(i)
          col = IAU(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AU(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do
       
       do j = 1, NB
          res_norm0 = res_norm0 + WVAL(j)**2
       end do
    end do

    CALL MPI_ALLREDUCE(res_norm0, res_norm, 1, MPI_DOUBLE_PRECISION, &
         &             MPI_SUM, SOLVER_COMM, ierr)

    res_norm = dsqrt(res_norm)
    if(my_rank == 0) write(*,*) "***",res_norm
    
  end subroutine cal_res_norm_NB_lev



end module     solver_AMGCG_NB


