module solver_AMGGPBiCG
contains
  subroutine AMGGPBiCG                                                    &
       &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, &
       &                  B,  X, RESID,  ITER, my_rank, NEIBPETOT, NEIBPE,&
       &                  STACK_IMPORT, NOD_IMPORT,                       &
       &                  STACK_EXPORT, NOD_EXPORT,                       &
       &                  SOLVER_COMM, NPROCS, SETUP_FL, LEVEL_NUM,       &
       &                  REACH_CONV_FL)     
    
    use data_structure_for_AMG, only: HIERARCHICAL_DATA, max_comm_table_sz, &
         & max_NP_on_all_levs
    use data_creation_AMGCG, only: data_creation_unsym_ssi_amg
    use solver_SR2
    use common_mod, only: setup_phase, clear_matrix_ssi_amg
    USE terminal_condition, only: terminal_cond_check, BNRM2, terminal_kind
    USE amg_parameter, only: conv_check_fl,INDPDT_AGGRE
    use residual_calculation, only:cal_relative_res_norm
    USE AMG_cycle

    IMPLICIT NONE
    include  'precision.inc'
    include  'mpif.h'

    LOGICAL, intent(in) :: SETUP_FL    
    integer(kind=kint ), intent(in):: N, NP, NPU, NPL, my_rank
    integer(kind=kint ), intent(in):: NEIBPETOT
    integer(kind=kint ), intent(in):: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in)   :: NPROCS
    LOGICAL, intent(inout) :: REACH_CONV_FL


    integer(kind=kint ), intent(inout):: ITER
    real   (kind=kreal), intent(inout):: RESID

!!$    real(kind=kreal), dimension(NB*NP)   , intent(inout):: B, X
!!$    real(kind=kreal), dimension(NB,NB,NPL), intent(inout):: AL
!!$    real(kind=kreal), dimension(NB,NB,NPU), intent(inout):: AU
!!$    real(kind=kreal), dimension(NB,NB,NP ), intent(inout):: D
    real(kind=kreal), dimension(:)   , intent(inout):: B, X
    real(kind=kreal), dimension(:), intent(inout):: AL
    real(kind=kreal), dimension(:), intent(inout):: AU
    real(kind=kreal), dimension(:), intent(inout):: D

!!$    integer(kind=kint ), dimension(0:NP) ,intent(in) :: INU, INL
!!$    integer(kind=kint ), dimension(  NPL),intent(in) :: IAL
!!$    integer(kind=kint ), dimension(  NPU),intent(in) :: IAU
    integer(kind=kint ), dimension(0:) ,intent(in) :: INU, INL
    integer(kind=kint ), dimension(:),intent(in) :: IAL
    integer(kind=kint ), dimension(:),intent(in) :: IAU


    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:), NOD_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:), NOD_EXPORT(:)

    real(kind=kreal), dimension(:),    allocatable, save :: WS, WR
    real(kind=kreal), dimension(:,:),  allocatable, save :: WW

    integer(kind=kint ) :: MAXIT
    integer(kind=kint ) :: P,R,RINI,Y,T,U,Z,W,BVECT,AP,KINV_R,KINV_AP,KINV_T,KINV_TOLD,ZP
    real   (kind=kreal) :: TOL, SS


    !C LEVEL_NUM is added..
    integer(kind=kint ) :: LEVEL_NUM,ierr,i,j,k,l,m,col,ip
    INTEGER(kind=kint),save  :: WSIZE, NP_old
    integer(kind=kint ) :: isL,ieL,isU,ieU,iq1,iq2,iq3,inod 
    integer(kind=kint ) :: SISIZE, SESIZE, WSIZE2, NP_max
    real   (kind=kreal) :: BNRM20, X1,X2,X3,D11,D22,D33
    real   (kind=kreal) :: wval, v, SCALWVAL
    real   (kind=kreal) :: DNRM20,DNRM2,C10,C1,RHO,RHO0,RHO1
    logical, save :: initialize = .true.
    logical :: CONV_FL

    real   (kind=kreal) :: alpha, beta, zeta, eta
    !C- Temp space of (r,r)
    REAL   (kind=kreal) :: RR,RAPTemp,RRTemp, sendTemp_1, sendTemp_2
    REAL   (kind=kreal) :: sendTemp_3, sendTemp_4, sendTemp_5
    !C- Temp vector for culc of zeta and eta
    REAL   (kind=kreal), dimension(5) :: sendTemp, recvTemp

    
    !C-- INIT. 
    SISIZE = STACK_IMPORT(NEIBPETOT)
    SESIZE = STACK_EXPORT(NEIBPETOT)

    
    if(initialize .or. SETUP_FL) then
!!$       allocate(WVAL(NB))
       
       allocate (WW(NP,15))
       NP_old = NP
!!$       NB_old = NB

       !C- decision of the allocation size for WS, WR
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)
          end if
       end if
       if( WSIZE < NP ) WSIZE = NP

       ALLOCATE ( WS(WSIZE),WR(WSIZE) )

       initialize = .false.
    end if

    P = 1
    R = 2
    RINI = 3
    Y = 4
    T = 5
    U = 6
    Z = 7
    W = 8
    BVECT = 9
    AP = 10
    KINV_R = 11
    KINV_AP = 12
    KINV_T  = 13
    KINV_TOLD = 14
    ZP = 15

    MAXIT = ITER
    !C    conv_criterion_val = RESID 
    !C if terminal_condition isn't set,  RESID will 
    

    if(SETUP_FL) then
       
       CALL data_creation_unsym_ssi_amg(N, NP, NPL, NPU, D, AL, &
            & INL, IAL, AU, INU, IAU, NEIBPETOT,                &
            & NEIBPE, STACK_IMPORT, NOD_IMPORT,                 &
            & STACK_EXPORT, NOD_EXPORT,                         &
            & SISIZE, SESIZE, SOLVER_COMM, LEVEL_NUM, WSIZE2)
       
    else if (LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA, LEVEL_NUM)

    else
       !C in the case, LEVEL_NUM == 0
       stop 'error. no multi-level data is available'

    end if

    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA, LEVEL_NUM)
    
    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2),WR(WSIZE2))
       WSIZE = WSIZE2
    END if
    if(NP_max > NP_old) then
       deallocate(WW)
       allocate(WW(NP_max,15))
       NP_old = NP_max
    endif
    !C-------- allocation of WS, WS, WW, temp is finished.
    !C-
    !C Init T,W,P vector and beta

    !$omp parallel do 
    do i = 1, NP
       WW(i,T) = 0
       WW(i,W) = 0
       WW(i,P) = 0
       WW(i,Z) = 0
    enddo
    !$omp end parallel do
    alpha = 0
    beta = 0
    !C-

    
    !C
    !C +-----------------------+
    !C | {r0}= {b} - [A]{xini} |
    !C +-----------------------+
    !C===
    !C-- INTERFACE data EXCHANGE
    if (NEIBPETOT > 0) then
       call SOLVER_SEND_RECV2                                                        &
            &   (NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,       &
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,my_rank)
    end if

    !C
    !C-- BEGIN calculation

    !$omp parallel do private(l,col,WVAL)
    do i= 1, N
       WVAL = B(i)
       WVAL = WVAL - D(i)*X(i)
       
       do l = INL(i-1)+1, INL(i)
          col = IAL(l)
          WVAL = WVAL - AL(l)*X(col)
       end do

       do l = INU(i-1)+1, INU(i)
          col = IAU(l)
          WVAL = WVAL - AU(l)*X(col)
       end do

       WW(i,R)    = WVAL
       WW(i,RINI) = WVAL


    enddo
    !$omp end parallel do


    !C
    !C +-----------------------+
    !C | culc {r} * {r}        |
    !C +-----------------------+
    !C===
    SCALWVAL = 0
    !$omp parallel do reduction(+:SCALWVAL)
    do i = 1, N 
       SCALWVAL = SCALWVAL + WW(i,RINI) * WW(i,R)
    enddo
    !$omp end parallel do

    RR = 0
    call MPI_Allreduce(SCALWVAL, RR, 1, MPI_DOUBLE_PRECISION, MPI_SUM, SOLVER_COMM, &
            &          ierr)


    !C
    !C +-----------------------+
    !C | Iteration part start  |
    !C +-----------------------+
    !C===
    do iter= 1, MAXIT


       !C
       !C +------------------------------------+
       !C | {p0}= {Kinv}{r}                    |
       !C | {pn}= {Kinv}{r}+beta*{pn-1 - un-1} |
       !C +------------------------------------+
       !C===
       !C- culc {KINV_R} = {Kinv}{r}
       !$omp parallel do
       do i = 1, NP
          WW(i,KINV_R) = 0
       enddo
       !$omp end parallel do

       CALL amg_cycles(WW(:,R),WW(:,KINV_R),LEVEL_NUM,SOLVER_COMM,my_rank,WS,WR,WW(:,ZP),NPROCS,WSIZE,NP)

       !C- if iter number is zero, {p} = {Kinv}{r}, else culc beta * (p - u) and add {Kinv}{r}
       if(iter == 1) then
          !$omp parallel do          
          do i = 1, N
             WW(i,P) = WW(i,KINV_R) 
          enddo
          !$omp end parallel do
       else
          !$omp parallel do                    
          do i = 1, N
             WW(i,P) = WW(i,KINV_R) + beta * ( WW(i,P) - WW(i,U) )
          enddo
          !$omp end parallel do                    
       endif


       !C
       !C +------------------------------------+
       !C | alpha = {r}{r} / {r}{Ap}           |
       !C +------------------------------------+
       !C===

       !C- culc {Ap}
       !C
       !C-- INTERFACE data EXCHANGE
       if(NEIBPETOT > 0) then
          call SOLVER_SEND_RECV2                                                      &
               &   ( NP,WSIZE,WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,     &
               &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,P) , SOLVER_COMM,         &
               &     my_rank)
       end if

       !$omp parallel do private(WVAL,l,inod)       
       do i= 1, N
          WVAL = 0.d0
          WVAL = WVAL+D(i)*WW(i,P)

          do l = INL(i-1)+1, INL(i)
             inod = IAL(l)
             WVAL = WVAL + AL(l)*WW(inod,P)
          end do

          do l = INU(i-1)+1, INU(i)
             inod = IAU(l)
             WVAL = WVAL + AU(l)*WW(inod, P)
          end do

          WW(i,AP) = WVAL

       enddo
       !$omp end parallel do
       
       !C- culc {r}{Ap}
       SCALWVAL = 0
       !$omp parallel do reduction(+:SCALWVAL)
       do i = 1, N
          SCALWVAL = SCALWVAL + WW(i,RINI) * WW(i,AP)
       enddo
       !$omp end parallel do
       RAPTemp = 0
       call MPI_Allreduce(SCALWVAL,RAPTemp,1,MPI_DOUBLE_PRECISION,MPI_SUM,SOLVER_COMM, &
                   &      ierr)

       !C- culc alpha
       alpha = RR / RAPTemp 

       
       !C
       !C +--------------------------------------------+
       !C | {y} = {t} - {r} - {alpha}{w} + {alpha}{Ap} |
       !C +--------------------------------------------+
       !C===
       !$omp parallel do       
       do i = 1, N
          WW(i,Y) = WW(i,T) - WW(i,R) - alpha * WW(i,W) + alpha * WW(i,AP)
       enddo
       !$omp end parallel do       
       

       !C
       !C +----------------------------------+
       !C | {t} = {r} - {alpha}{Ap}          |
       !C +----------------------------------+
       !C===
       !$omp parallel do       
       do i = 1, N
          WW(i,T) = WW(i,R) - alpha * WW(i,AP)
       enddo
       !$omp end parallel do       

       !C
       !C +-----------------------------------------------+
       !C | {Kinv}{t] = {Kinv}{r] - {alpha}{Kinv}{Ap}     |
       !C +-----------------------------------------------+
       !C===
       !C-Init KINV_AP vector
       !$omp parallel do       
       do i = 1, NP
          WW(i,KINV_AP) = 0
       enddo
       !$omp end parallel do       

       !C- culc {KINV_AP} = {Kinv}{AP}

       call amg_cycles(WW(:,AP),WW(:,KINV_AP),LEVEL_NUM,SOLVER_COMM,my_rank,WS,WR,WW(:,ZP),NPROCS,WSIZE, NP)       
!!$       CALL v_cycle(WW(:,AP),WW(:,KINV_AP),LEVEL_NUM,SOLVER_COMM,my_rank,WS,WR,WW(:,ZP),NP)
!!$       CALL sgs_NB(N,NP,D,AL,INL,IAL,AU,INU,IAU,ALU,WW(:,R),WW(:,Z),1,1,SOLVER_COMM,&
!!$            &         my_rank,WS,WR,NP,NB)
      
       !C- culc {KINV_T} 
       !$omp parallel do       
       do i = 1, N
          WW(i,KINV_T) = WW(i,KINV_R) - alpha * WW(i,KINV_AP)
       enddo 
       !$omp end parallel do       


       !C
       !C +-----------------------+
       !C | culc zeta and eta     |
       !C +-----------------------+
       !C===
       !C culc {b} = {A}{KINV_T}
       !C
       !C-- INTERFACE data EXCHANGE
       if(NEIBPETOT > 0) then
          call SOLVER_SEND_RECV2                                                   &
               &   ( NP,WSIZE,WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,  &
               &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,KINV_T) , SOLVER_COMM, &
               &     my_rank)
       end if

       !$omp parallel do private(WVAL,l,inod)       
       do i= 1, N
          WVAL = 0.d0
          WVAL = WVAL+D(i)*WW(i,KINV_T)

          do l = INL(i-1)+1, INL(i)
             inod = IAL(l)
             WVAL = WVAL + AL(l)*WW(inod,KINV_T)
          end do

          do l = INU(i-1)+1, INU(i)
             inod = IAU(l)
             WVAL = WVAL + AU(l)*WW(inod, KINV_T)
          end do

          WW(i,BVECT) = WVAL
          
       enddo
       !$omp end parallel do
      
       !C- sendTemp(1) : (b,a)
       !C- sendTemp(2) : (b,b)
       !C- sendTemp(3) : (c,c)
       !C- sendTemp(4) : (c,a)
       !C- sendTemp(5) : (c,b), (b,c)
       !C- recvTemp is result of this culc
       !C- Init sendTemp and recvTemp
       do i = 1, 5
          sendTemp(i) = 0
          recvTemp(i) = 0
       enddo
       sendTemp_1=0
       sendTemp_2=0
       sendTemp_3=0
       sendTemp_4=0
       sendTemp_5=0


       !C- culc (b,a) and (b,b)
       !$omp parallel do reduction(+:sendTemp_1, sendTemp_2)
       do i = 1 , N
          sendTemp_1 = sendTemp_1 + WW(i,BVECT) * WW(i,T)
          sendTemp_2 = sendTemp_2 + WW(i,BVECT) * WW(i,BVECT)
       enddo
       !$omp end parallel do
       sendTemp(1) = sendTemp_1
       sendTemp(2) = sendTemp_2

       !C- culc zeta end eta (n = 0 and n != 0 are different)
       if(iter == 1) then
          call MPI_Allreduce(sendTemp, recvTemp, 2, MPI_DOUBLE_PRECISION, MPI_SUM, &
               &        SOLVER_COMM, ierr)
          
          !C- zeta = (b,a) / (b,b) , eta = 0
          zeta = recvTemp(1) / recvTemp(2)
          eta = 0
       else
          !C- culc (c,c), (c,a), (c,b) 
          !$omp parallel do reduction(+:sendTemp_3, sendTemp_4, sendTemp_5)
          do i = 1, N
             sendTemp_3 = sendTemp_3 + WW(i,Y) * WW(i,Y)
             sendTemp_4 = sendTemp_4 + WW(i,Y) * WW(i,T)
             sendTemp_5 = sendTemp_5 + WW(i,Y) * WW(i,BVECT)
          enddo
          !$omp end parallel do 
          sendTemp(3) = sendTemp_3
          sendTemp(4) = sendTemp_4
          sendTemp(5) = sendTemp_5

          call MPI_Allreduce(sendTemp, recvTemp, 5, MPI_DOUBLE_PRECISION, MPI_SUM, &
                &       SOLVER_COMM, ierr)

          !C- culc zeta
          zeta = (recvTemp(1) * recvTemp(3) - recvTemp(4) * recvTemp(5)) /    &
              &  (recvTemp(3) * recvTemp(2) - recvTemp(5) * recvTemp(5))
          !C- culc eta
          eta =  (recvTemp(2) * recvTemp(4) - recvTemp(5) * recvTemp(1)) /    &
              &  (recvTemp(2) * recvTemp(3) - recvTemp(5) * recvTemp(5))
       endif
       

       !C
       !C +---------------------------------------------------------------+
       !C | culc {u} (if n = 0, then only culc {u} = zeta * {Kinv}{AP})   |
       !C +---------------------------------------------------------------+
       !C===
       if(iter == 1) then
          !$omp parallel do 
          do i = 1, N
             WW(i,U) = zeta * WW(i,KINV_AP)
          enddo
          !$omp end parallel do 
       else
          !$omp parallel do 
          do i = 1, N
             WW(i,U) = zeta * WW(i,KINV_AP) + eta *  &
               &      ( WW(i,KINV_TOLD) - WW(i,KINV_R) + beta * WW(i,U) )
          enddo
          !$omp end parallel do 
       endif


       !C
       !C +----------------------------------------------------+
       !C | {z} = zeta * {KINV}{R} + eta * {z} - alpha * {u}   |
       !C +----------------------------------------------------+
       !C===
       !$omp parallel do 
       do i = 1, N
          WW(i,Z) = zeta * WW(i,KINV_R) + eta * WW(i,Z) - alpha * WW(i,U)
       enddo
       !$omp end parallel do 


       
       !C
       !C +---------------------------------+
       !C | {x} = {x} + alpha * {p} + {z}   |
       !C +---------------------------------+
       !C===
       !$omp parallel do 
       do i = 1, N
          X(i) = X(i) + alpha * WW(i,P) + WW(i,Z)
       enddo
       !$omp end parallel do 


       !C
       !C +-------------------------------------------+
       !C | {r} = {t} - eta * {y} - {zeta}{AKINV_T}   |
       !C | {AKINV_T} is {b} vector                   |
       !C +-------------------------------------------+
       !C===
       !$omp parallel do 
       do i = 1, N
          WW(i,R) = WW(i,T) - eta * WW(i,Y) - zeta * WW(i,BVECT)
       enddo
       !$omp end parallel do 


       !C
       !C +-----------------------------------------------------------+
       !C | beta = (alpha / zeta) * ( ({r0},{rn+1}) / ({r0},{rn}) )   |
       !C +-----------------------------------------------------------+
       !C===
       !C- culc ( {r0}, {rn+1} )
       SCALWVAL = 0
       !$omp parallel do reduction(+:SCALWVAL)
       do i = 1, N
          SCALWVAL = SCALWVAL + WW(i,RINI) * WW(i,R)
       enddo
       !$omp end parallel do
       RRTemp = 0
       call MPI_Allreduce(SCALWVAL, RRTemp, 1, MPI_DOUBLE_PRECISION,   &
                       &  MPI_SUM, SOLVER_COMM, ierr)

       !C- culc beta
       beta = ( alpha / zeta ) * ( RRTemp / RR )

       !C- Update RR (RRTemp is new RR, so RR = RRTemp)
       RR = RRTemp


       !C
       !C +-----------------------------------------------------------+
       !C | {w} = {AKINV_T} + beta * {AP}  ( {AKINV_T} is vector {b} )|
       !C +-----------------------------------------------------------+
       !C===
       !$omp parallel do 
       do i = 1, N
          WW(i,W) = WW(i,BVECT) + beta * WW(i,AP)
       enddo
       !$omp end parallel do 


       !C
       !C +---------------------------+
       !C | Check relative residual   |
       !C +---------------------------+
       !C===        
       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N, SOLVER_COMM)

#ifdef PRINT_REZ       
       if (my_rank.eq.0 .and. terminal_kind == 0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif       

       if ( CONV_FL  ) then
          REACH_CONV_FL = .true.
          exit
       end if

             
       !C
       !C +---------------------------+
       !C | Update {KINV_TOLD}        |
       !C +---------------------------+
       !C===        
       !$omp parallel do 
       do i = 1, N
          WW(i,KINV_TOLD) = WW(i,KINV_T)
       enddo
       !$omp end parallel do 


    end do
    !C===
    !C
    !C +----------------------+
    !C | Iteration part end   |
    !C +----------------------+
    !C===

    !C differentiate divergence cases from non-convergent cases
    if(conv_check_fl)then
       if(.not.CONV_FL) then
          call cal_relative_res_norm(RESID, X, B, WS, WR, WSIZE)
          if(RESID > 1.0) then
             iter = -1
          end if
       end if
    end if

    if(SETUP_FL) then
       call clear_matrix_ssi_amg(LEVEL_NUM)
       DEALLOCATE (WW)
       DEALLOCATE (WR)
       DEALLOCATE (WS)
       initialize = .true.
    end if

  end subroutine        AMGGPBICG


end module     solver_AMGGPBiCG


