module solver_AMGGPBiCG_NB
contains
  subroutine AMGGPBiCG_NB                                                &
       &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, &
       &                  B,  X, RESID,  ITER, my_rank, NEIBPETOT, NEIBPE,&
       &                  STACK_IMPORT, NOD_IMPORT,                       &
       &                  STACK_EXPORT, NOD_EXPORT,                       &
       &                  SOLVER_COMM, NPROCS, SETUP_FL, LEVEL_NUM,       &
       &                  REACH_CONV_FL, NB, k_or_q_vec, NK)     

    use data_structure_for_AMG_NB, only: HIERARCHICAL_DATA_NB, max_comm_table_sz, &
         & max_NP_on_all_levs
    use data_creation_AMGCG_NB, only: data_creation_NB
    use solver_SR_NB
    use count_time_mod
    use common_mod, only: setup_phase, clear_matrix_ssi_amg
    USE terminal_condition, only: terminal_cond_check, BNRM2, terminal_kind
    USE amg_parameter, only: conv_check_fl,INDPDT_AGGRE
    use residual_calculation, only:cal_relative_res_norm_NB
    USE solver_AMGCG_NB, only: v_cycle

    IMPLICIT NONE
    include  'precision.inc'
    include  'mpif.h'

    LOGICAL, intent(in) :: SETUP_FL    
    integer(kind=kint ), intent(in):: NB, N, NP, NPU, NPL, my_rank
    integer(kind=kint ), intent(in):: NEIBPETOT
    integer(kind=kint ), intent(in):: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in)   :: NPROCS
    LOGICAL, intent(inout) :: REACH_CONV_FL


    integer(kind=kint ), intent(inout):: ITER
    real   (kind=kreal), intent(inout):: RESID

!!$    real(kind=kreal), dimension(NB*NP)   , intent(inout):: B, X
!!$    real(kind=kreal), dimension(NB,NB,NPL), intent(inout):: AL
!!$    real(kind=kreal), dimension(NB,NB,NPU), intent(inout):: AU
!!$    real(kind=kreal), dimension(NB,NB,NP ), intent(inout):: D
    real(kind=kreal), dimension(:)   , intent(inout):: B, X
    real(kind=kreal), dimension(:,:,:), intent(inout):: AL
    real(kind=kreal), dimension(:,:,:), intent(inout):: AU
    real(kind=kreal), dimension(:,:,:), intent(inout):: D

!!$    integer(kind=kint ), dimension(0:NP) ,intent(in) :: INU, INL
!!$    integer(kind=kint ), dimension(  NPL),intent(in) :: IAL
!!$    integer(kind=kint ), dimension(  NPU),intent(in) :: IAU
    integer(kind=kint ), dimension(0:) ,intent(in) :: INU, INL
    integer(kind=kint ), dimension(:),intent(in) :: IAL
    integer(kind=kint ), dimension(:),intent(in) :: IAU

    INTEGER(kind=kint ), INTENT(in) :: NK
    real(kind=kreal), pointer :: k_or_q_vec(:,:)


    integer(kind=kint ), pointer :: NEIBPE(:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:), NOD_IMPORT(:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:), NOD_EXPORT(:)

    real(kind=kreal), dimension(:),    allocatable,save  :: WS, WR
    real(kind=kreal), dimension(:,:),  allocatable,save  :: WW
    real(kind=kreal), dimension(:,:,:), allocatable, target :: ALU

    real(kind=kreal), dimension(:,:,:), allocatable :: AUlu0, ALlu0
    real(kind=kreal), dimension(:,:,:), allocatable :: Dlu0

    integer(kind=kint), dimension(:), allocatable :: inumFI1L, FI1L
    integer(kind=kint), dimension(:), allocatable :: inumFI1U, FI1U

    integer(kind=kint ) :: MAXIT
    integer(kind=kint ) :: P,R,RINI,Y,T,U,Z,W,BVECT,AP,KINV_R,KINV_AP,KINV_T,KINV_TOLD,ZP
    real   (kind=kreal) :: TOL, SS


    !C LEVEL_NUM is added..
    integer(kind=kint ) :: LEVEL_NUM,ierr,i,j,k,l,m,col,ip
    INTEGER(kind=kint),save  :: WSIZE, NP_old, NB_old
    integer(kind=kint ) :: isL,ieL,isU,ieU,iq1,iq2,iq3,inod 
    integer(kind=kint ) :: SISIZE, SESIZE, WSIZE2, NP_max
    real   (kind=kreal) :: BNRM20, X1,X2,X3,D11,D22,D33
    real   (kind=kreal) :: WVAL(NB), v, SCALWVAL
    real   (kind=kreal) :: DNRM20,DNRM2,C10,C1,RHO,RHO0,RHO1
    logical, save :: initialize = .true.
    logical :: CONV_FL

    CHARACTER(len=20) :: PRECOND

    real   (kind=kreal) :: alpha, beta, zeta, eta
    !C- Temp space of (r,r)
    REAL   (kind=kreal) :: RR,RAPTemp,RRTemp
    !C- Temp vector for culc of zeta and eta
    REAL   (kind=kreal), dimension(5) :: sendTemp, recvTemp

    
    !C-- INIT. 
    SISIZE = STACK_IMPORT(NEIBPETOT)
    SESIZE = STACK_EXPORT(NEIBPETOT)

    
    if(initialize .or. SETUP_FL) then
!!$       allocate(WVAL(NB))
       
       allocate (WW(NB*NP,15))
       NP_old = NP
!!$       NB_old = NB

       !C- decision of the allocation size for WS, WR
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)*NB
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)*NB
          end if
       end if
       if( WSIZE < NP*NB ) WSIZE = NP*NB

       ALLOCATE ( WS(WSIZE),WR(WSIZE) )

       initialize = .false.
    end if

    P = 1
    R = 2
    RINI = 3
    Y = 4
    T = 5
    U = 6
    Z = 7
    W = 8
    BVECT = 9
    AP = 10
    KINV_R = 11
    KINV_AP = 12
    KINV_T  = 13
    KINV_TOLD = 14
    ZP = 15

    MAXIT = ITER
    !C    conv_criterion_val = RESID 
    !C if terminal_condition isn't set,  RESID will 
    

    if(SETUP_FL) then
       
       if(INDPDT_AGGRE) then
          PRECOND ="IND1"
       else
          PRECOND ="SHA2"
       end if
       call data_creation_NB                                                   &
            &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU,           &
            &                  INU, IAU,                                       &
            &                  LEVEL_NUM, my_rank, NEIBPETOT, NEIBPE,          &
            &                  STACK_IMPORT, NOD_IMPORT,                       &
            &                  STACK_EXPORT, NOD_EXPORT,                       &
            &                  SOLVER_COMM, PRECOND, NB, k_or_q_vec, NK)     


    end if

    if (LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA_NB, LEVEL_NUM)

    else
       !C in the case, LEVEL_NUM == 0
       j = max(SISIZE, SESIZE)
       if(WSIZE < j*NB) WSIZE2 = j*NB
    end if

    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA_NB, LEVEL_NUM)
    
    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2),WR(WSIZE2))
       WSIZE = WSIZE2
    END if
    if(NP_max > NP_old) then
       deallocate(WW)
       allocate(WW(NB*NP_max,15))
       NP_old = NP_max
    endif
    !C-------- allocation of WS, WS, WW, temp is finished.
    !C-
    !C Init T,W,P vector and beta
    do i = 1, NP * NB
       WW(i,T) = 0
       WW(i,W) = 0
       WW(i,P) = 0
       WW(i,Z) = 0
    enddo
    alpha = 0
    beta = 0
    !C-

    
    !C
    !C +-----------------------+
    !C | {r0}= {b} - [A]{xini} |
    !C +-----------------------+
    !C===
    !C-- INTERFACE data EXCHANGE
    if (NEIBPETOT > 0) then
       call SOLVER_SEND_RECV_NB                                        &
            &   ( WSIZE,WSIZE,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,       &
            &     STACK_EXPORT, NOD_EXPORT, WS, WR, X , SOLVER_COMM,my_rank, NB)
    end if

    !C
    !C-- BEGIN calculation

    !$omp parallel do private(j,k,l,col,WVAL)
    do i= 1, N
       do j = 1, NB
          WVAL(j) = B(NB*(i-1)+j)
          do k = 1, NB
             WVAL(j) = WVAL(j) - D(j,k,i)*X(NB*(i-1)+k)
          end do
       end do

       do l = INL(i-1)+1, INL(i)
          col = IAL(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AL(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do l = INU(i-1)+1, INU(i)
          col = IAU(l)
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j) - AU(j,k,l)*X(NB*(col-1)+k)
             end do
          end do
       end do

       do j = 1, NB
          WW(NB*(i-1)+j,R) = WVAL(j)
          WW(NB*(i-1)+j,RINI) = WVAL(j)
       end do

    enddo
    !$omp end parallel do


    !C
    !C +-----------------------+
    !C | culc {r} * {r}        |
    !C +-----------------------+
    !C===
    SCALWVAL = 0
    do i = 1, N * NB
       SCALWVAL = SCALWVAL + WW(i,RINI) * WW(i,R)
    enddo
    
    RR = 0
    call MPI_Allreduce(SCALWVAL, RR, 1, MPI_DOUBLE_PRECISION, MPI_SUM, SOLVER_COMM, &
            &          ierr)


    !C
    !C +-----------------------+
    !C | Iteration part start  |
    !C +-----------------------+
    !C===
    do iter= 1, MAXIT


       !C
       !C +------------------------------------+
       !C | {p0}= {Kinv}{r}                    |
       !C | {pn}= {Kinv}{r}+beta*{pn-1 - un-1} |
       !C +------------------------------------+
       !C===
       !C- culc {KINV_R} = {Kinv}{r}
       do i = 1, NP * NB
          WW(i,KINV_R) = 0
       enddo

       call count_time(1,SOLVER_COMM,my_rank,9)
       CALL v_cycle(WW(:,R),WW(:,KINV_R),LEVEL_NUM,SOLVER_COMM,my_rank,WS,WR,WW(:,ZP),NP)
!!$       CALL sgs_NB(N,NP,D,AL,INL,IAL,AU,INU,IAU,ALU,WW(:,R),WW(:,Z),1,1,SOLVER_COMM,&
!!$            &         my_rank,WS,WR,NP,NB)
       call count_time(2,SOLVER_COMM,my_rank,9)

       !C- if iter number is zero, {p} = {Kinv}{r}, else culc beta * (p - u) and add {Kinv}{r}
       if(iter == 1) then
          do i = 1, N * NB
             WW(i,P) = WW(i,KINV_R) 
          enddo
       else
          do i = 1, N * NB
             WW(i,P) = WW(i,KINV_R) + beta * ( WW(i,P) - WW(i,U) )
          enddo
       endif


       !C
       !C +------------------------------------+
       !C | alpha = {r}{r} / {r}{Ap}           |
       !C +------------------------------------+
       !C===

       !C- culc {Ap}
       !C
       !C-- INTERFACE data EXCHANGE
       if(NEIBPETOT > 0) then
          call SOLVER_SEND_RECV_NB                                        &
               &   ( NP,NP,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,     &
               &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,P) , SOLVER_COMM,   &
               &     my_rank, NB)
       end if

       !$omp parallel do private(WVAL,j,k,l,inod,isU,ieU,isL,ieL)       
       do i= 1, N
          WVAL = 0.d0
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j)+D(j,k,i)*WW(NB*i-(NB-k),P)
             end do
          end do

          do l = INL(i-1)+1, INL(i)
             inod = IAL(l)
             do j = 1, NB
                do k = 1, NB
                   WVAL(j) = WVAL(j) + AL(j,k,l)*WW(NB*inod-(NB-k),P)
                end do
             end do
          end do

          do l = INU(i-1)+1, INU(i)
             inod = IAU(l)
             do j = 1, NB
                do k = 1, NB
                   WVAL(j) = WVAL(j) + AU(j,k,l)*WW(NB*inod-(NB-k), P)
                end do
             end do
          end do

          do j = 1, NB
             WW(NB*i-(NB-j),AP) = WVAL(j)
          end do

       enddo
       !$omp end parallel do
       
       !C- culc {r}{Ap}
       SCALWVAL = 0
       do i = 1, N * NB
          SCALWVAL = SCALWVAL + WW(i,RINI) * WW(i,AP)
       enddo
       RAPTemp = 0
       call MPI_Allreduce(SCALWVAL,RAPTemp,1,MPI_DOUBLE_PRECISION,MPI_SUM,SOLVER_COMM, &
                   &      ierr)

       !C- culc alpha
       alpha = RR / RAPTemp 

       
       !C
       !C +--------------------------------------------+
       !C | {y} = {t} - {r} - {alpha}{w} + {alpha}{Ap} |
       !C +--------------------------------------------+
       !C===
       do i = 1, N * NB
          WW(i,Y) = WW(i,T) - WW(i,R) - alpha * WW(i,W) + alpha * WW(i,AP)
       enddo
       

       !C
       !C +----------------------------------+
       !C | {t} = {r} - {alpha}{Ap}          |
       !C +----------------------------------+
       !C===
       do i = 1, N * NB
          WW(i,T) = WW(i,R) - alpha * WW(i,AP)
       enddo

       !C
       !C +-----------------------------------------------+
       !C | {Kinv}{t] = {Kinv}{r] - {alpha}{Kinv}{Ap}     |
       !C +-----------------------------------------------+
       !C===
       !C-Init KINV_AP vector
       do i = 1, NP * NB
          WW(i,KINV_AP) = 0
       enddo

       !C- culc {KINV_AP} = {Kinv}{AP}
       call count_time(1,SOLVER_COMM,my_rank,9)
       CALL v_cycle(WW(:,AP),WW(:,KINV_AP),LEVEL_NUM,SOLVER_COMM,my_rank,WS,WR,WW(:,ZP),NP)
!!$       CALL sgs_NB(N,NP,D,AL,INL,IAL,AU,INU,IAU,ALU,WW(:,R),WW(:,Z),1,1,SOLVER_COMM,&
!!$            &         my_rank,WS,WR,NP,NB)
       call count_time(2,SOLVER_COMM,my_rank,9)
      
       !C- culc {KINV_T} 
       do i = 1, N * NB
          WW(i,KINV_T) = WW(i,KINV_R) - alpha * WW(i,KINV_AP)
       enddo 


       !C
       !C +-----------------------+
       !C | culc zeta and eta     |
       !C +-----------------------+
       !C===
       !C culc {b} = {A}{KINV_T}
       !C
       !C-- INTERFACE data EXCHANGE
       if(NEIBPETOT > 0) then
          call SOLVER_SEND_RECV_NB                                        &
               &   ( NP,NP,NP, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,     &
               &     STACK_EXPORT, NOD_EXPORT, WS, WR, WW(:,KINV_T) , SOLVER_COMM,   &
               &     my_rank, NB)
       end if

       !$omp parallel do private(WVAL,j,k,l,inod,isU,ieU,isL,ieL)       
       do i= 1, N
          WVAL = 0.d0
          do j = 1, NB
             do k = 1, NB
                WVAL(j) = WVAL(j)+D(j,k,i)*WW(NB*i-(NB-k),KINV_T)
             end do
          end do

          do l = INL(i-1)+1, INL(i)
             inod = IAL(l)
             do j = 1, NB
                do k = 1, NB
                   WVAL(j) = WVAL(j) + AL(j,k,l)*WW(NB*inod-(NB-k),KINV_T)
                end do
             end do
          end do

          do l = INU(i-1)+1, INU(i)
             inod = IAU(l)
             do j = 1, NB
                do k = 1, NB
                   WVAL(j) = WVAL(j) + AU(j,k,l)*WW(NB*inod-(NB-k), KINV_T)
                end do
             end do
          end do

          do j = 1, NB
             WW(NB*i-(NB-j),BVECT) = WVAL(j)
          end do

       enddo
       !$omp end parallel do
      
       !C- sendTemp(1) : (b,a)
       !C- sendTemp(2) : (b,b)
       !C- sendTemp(3) : (c,c)
       !C- sendTemp(4) : (c,a)
       !C- sendTemp(5) : (c,b), (b,c)
       !C- recvTemp is result of this culc
       !C- Init sendTemp and recvTemp
       do i = 1, 5
          sendTemp(i) = 0
          recvTemp(i) = 0
       enddo

       !C- culc (b,a) and (b,b)
       do i = 1 , N * NB
          sendTemp(1) = sendTemp(1) + WW(i,BVECT) * WW(i,T)
          sendTemp(2) = sendTemp(2) + WW(i,BVECT) * WW(i,BVECT)
       enddo

       !C- culc zeta end eta (n = 0 and n != 0 are different)
       if(iter == 1) then
          call MPI_Allreduce(sendTemp, recvTemp, 2, MPI_DOUBLE_PRECISION, MPI_SUM, &
                &        SOLVER_COMM, ierr)
          
          !C- zeta = (b,a) / (b,b) , eta = 0
          zeta = recvTemp(1) / recvTemp(2)
          eta = 0
       else
          !C- culc (c,c), (c,a), (c,b) 
          do i = 1, N * NB
             sendTemp(3) = sendTemp(3) + WW(i,Y) * WW(i,Y)
             sendTemp(4) = sendTemp(4) + WW(i,Y) * WW(i,T)
             sendTemp(5) = sendTemp(5) + WW(i,Y) * WW(i,BVECT)
          enddo

          call MPI_Allreduce(sendTemp, recvTemp, 5, MPI_DOUBLE_PRECISION, MPI_SUM, &
                &       SOLVER_COMM, ierr)

          !C- culc zeta
          zeta = (recvTemp(1) * recvTemp(3) - recvTemp(4) * recvTemp(5)) /    &
              &  (recvTemp(3) * recvTemp(2) - recvTemp(5) * recvTemp(5))
          !C- culc eta
          eta =  (recvTemp(2) * recvTemp(4) - recvTemp(5) * recvTemp(1)) /    &
              &  (recvTemp(2) * recvTemp(3) - recvTemp(5) * recvTemp(5))
       endif
       

       !C
       !C +---------------------------------------------------------------+
       !C | culc {u} (if n = 0, then only culc {u} = zeta * {Kinv}{AP})   |
       !C +---------------------------------------------------------------+
       !C===
       if(iter == 1) then
          do i = 1, N * NB
             WW(i,U) = zeta * WW(i,KINV_AP)
          enddo
       else
          do i = 1, N * NB
             WW(i,U) = zeta * WW(i,KINV_AP) + eta *  &
               &      ( WW(i,KINV_TOLD) - WW(i,KINV_R) + beta * WW(i,U) )
          enddo
       endif


       !C
       !C +----------------------------------------------------+
       !C | {z} = zeta * {KINV}{R} + eta * {z} - alpha * {u}   |
       !C +----------------------------------------------------+
       !C===
       do i = 1, N * NB
          WW(i,Z) = zeta * WW(i,KINV_R) + eta * WW(i,Z) - alpha * WW(i,U)
       enddo


       !C
       !C +---------------------------------+
       !C | {x} = {x} + alpha * {p} + {z}   |
       !C +---------------------------------+
       !C===
       do i = 1, N * NB
          X(i) = X(i) + alpha * WW(i,P) + WW(i,Z)
       enddo


       !C
       !C +-------------------------------------------+
       !C | {r} = {t} - eta * {y} - {zeta}{AKINV_T}   |
       !C | {AKINV_T} is {b} vector                   |
       !C +-------------------------------------------+
       !C===
       do i = 1, N * NB
          WW(i,R) = WW(i,T) - eta * WW(i,Y) - zeta * WW(i,BVECT)
       enddo


       !C
       !C +-----------------------------------------------------------+
       !C | beta = (alpha / zeta) * ( ({r0},{rn+1}) / ({r0},{rn}) )   |
       !C +-----------------------------------------------------------+
       !C===
       !C- culc ( {r0}, {rn+1} )
       SCALWVAL = 0
       do i = 1, N * NB
          SCALWVAL = SCALWVAL + WW(i,RINI) * WW(i,R)
       enddo
       RRTemp = 0
       call MPI_Allreduce(SCALWVAL, RRTemp, 1, MPI_DOUBLE_PRECISION,   &
                       &  MPI_SUM, SOLVER_COMM, ierr)

       !C- culc beta
       beta = ( alpha / zeta ) * ( RRTemp / RR )

       !C- Update RR (RRTemp is new RR, so RR = RRTemp)
       RR = RRTemp


       !C
       !C +-----------------------------------------------------------+
       !C | {w} = {AKINV_T} + beta * {AP}  ( {AKINV_T} is vector {b} )|
       !C +-----------------------------------------------------------+
       !C===
       do i = 1, N * NB
          WW(i,W) = WW(i,BVECT) + beta * WW(i,AP)
       enddo


       !C
       !C +---------------------------+
       !C | Check relative residual   |
       !C +---------------------------+
       !C===        
       call terminal_cond_check(CONV_FL, RESID, WW(:,R), N*NB, SOLVER_COMM)

#ifdef PRINT_REZ       
       if (my_rank.eq.0 .and. terminal_kind == 0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif       

       if ( CONV_FL  ) then
          REACH_CONV_FL = .true.
          exit
       end if

             
       !C
       !C +---------------------------+
       !C | Update {KINV_TOLD}        |
       !C +---------------------------+
       !C===        
       do i = 1, N * NB
          WW(i,KINV_TOLD) = WW(i,KINV_T)
       enddo


    end do
    !C===
    !C
    !C +----------------------+
    !C | Iteration part end   |
    !C +----------------------+
    !C===

    !C differentiate divergence cases from non-convergent cases
    if(conv_check_fl)then
       if(.not.CONV_FL) then
          call cal_relative_res_norm_NB(RESID, X, B, WS, WR, WSIZE, NB)
          if(RESID > 1.0) then
             iter = -1
          end if
       end if
    end if

#ifdef PRINT_INFO
    call count_time(4,solver_comm,my_rank,1)      
#endif
       
    if(SETUP_FL) then
       call clear_matrix_ssi_amg(LEVEL_NUM)
       DEALLOCATE (WW)
       DEALLOCATE (WR)
       DEALLOCATE (WS)
       initialize = .true.
    end if

  end subroutine        AMGGPBICG_NB


end module     solver_AMGGPBiCG_NB


