!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University. 
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.


#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif

MODULE solver_AMGRESCUT
CONTAINS
  !C SETUP_FL shows if this subroutine contains the setup phase
  !C and deletes the matrix
  subroutine AMGRESCUT                                                   &
       &                 (LSIZ, N, NP,  NPL, NPU,                        &
       &                  D,  AL, INL, IAL, AU, INU, IAU,                &
       &                  B,  X,                                         &
       &                  RESID,  ITER,                                  &
       &                  my_rank, NEIBPETOT, NEIBPE,                    &
       &                  STACK_IMPORT, NOD_IMPORT,                      &
       &                  STACK_EXPORT, NOD_EXPORT,                      &
       &                  SOLVER_COMM, PETOT, SETUP_FL, LEVEL_NUM,       &
       &                  REACH_CONV_FL)     

    USE solver_SR2
    USE common_mod, only: clear_matrix_ssi_amg
    USE data_creation_AMGCG
    USE AMG_cycle
    USE data_structure_for_AMG
    USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
         & terminal_cond_check, BNRM2, terminal_kind,conv_criterion_val
    USE residual_calculation
    USE amg_parameter, only: conv_check_fl
    !$ USE omp_lib

    implicit none
    include 'mpif.h'
    integer(kind=kint ), intent(in   )::  LSIZ, N, NP, NPU, NPL
    real   (kind=kreal), intent(inout)::  RESID
    integer(kind=kint ), intent(inout)::  ITER, LEVEL_NUM
    integer(kind=kint ), intent(in   )::  my_rank

    integer            , intent(in)   :: SOLVER_COMM
    integer(kind=kint ), intent(in   )::  PETOT
    logical, intent(in)  :: SETUP_FL
    logical, intent(out) :: REACH_CONV_FL

    real   (kind=kreal),  intent(inout)::  B(:)
    real   (kind=kreal),  intent(inout)::  X(:)

    real   (kind=kreal),  intent(inout)::  AL(:)
    !C size is NPL
    real   (kind=kreal),  intent(inout)::  AU(:)
    !C size is NPU
    real   (kind=kreal),  intent(inout)::  D(:)

    integer(kind=kint ), intent(in) :: INU(0:), INL(0:)
    integer(kind=kint ), intent(in) :: IAL(:)
    integer(kind=kint ), intent(in) :: IAU(:)

    integer(kind=kint ) :: NEIBPETOT
    integer(kind=kint ) :: NEIBPE      (:)
    integer(kind=kint ) :: STACK_IMPORT(0:)
    integer(kind=kint ) :: NOD_IMPORT  (:)
    integer(kind=kint ) :: STACK_EXPORT(0:)
    integer(kind=kint ) :: NOD_EXPORT  (:)

    real(kind=kreal), allocatable, save ::  WS(:), WR(:)
    INTEGER(kind=kint ), save :: WSIZE, NP_old, LSIZ_old
    logical, save :: initialize = .true.

    REAL   (kind=kreal),  ALLOCATABLE, save :: Temp(:), Residual_vec(:), psy(:,:), fai(:,:)
    REAL   (kind=kreal),  ALLOCATABLE, save :: alpha(:)
    REAL   (kind=kreal),  ALLOCATABLE, save :: AA(:,:), AA_bak(:,:), bb_bak(:)

    REAL   (kind=kreal) :: DNRM2, val, BNRM20, WVAL
    INTEGER(kind=kint ) :: MAXIT, WSIZE2, SESIZE, SISIZE


    INTEGER(kind=kint)  :: i,j,k,l,m, INOD, NOPT, zero_ind, ierr, NP_max
    logical :: CONV_FL, ERR_FL

    
!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(+:BNRM20)
!!$       do i= 1, N
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       !$omp end parallel do
!!$       
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$    else if(terminal_kind == 1) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(max:BNRM20)
!!$       do i= 1, N
!!$          WVAL = dabs(B(i))
!!$          if(BNRM20 < WVAL) BNRM20 = WVAL
!!$       enddo
!!$       !$omp end parallel do
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_MAX, SOLVER_COMM, ierr)
!!$    else 
!!$       stop 'unknown terminal_kind: solver_AMGRESCUT.F90'
!!$    end if
!!$
!!$    if(BNRM2 == 0) BNRM2 = 1.0



    MAXIT  = ITER
    SISIZE = STACK_IMPORT(NEIBPETOT)
    SESIZE = STACK_EXPORT(NEIBPETOT)       

    !C  if SETUP_FL == .true. then all matrices are cleared every iteration
    !C  if SETUP_FL == .false. then matrices are remained allocated.
    if(initialize) then
       ALLOCATE(Residual_vec(NP), Temp(NP),  psy(NP, 0:LSIZ), fai(NP, 0:LSIZ))
       ALLOCATE(AA(0:LSIZ, 0:LSIZ),AA_bak(0:LSIZ, 0:LSIZ), alpha(0:LSIZ+1) )       

       !C alpha(LSIZ+1) conatins the value of NOPT.
       NP_old = NP
       LSIZ_old = LSIZ

       !C- decision of the allocation size for WS, WR
       WSIZE = 0
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)
          end if
       end if
       if( WSIZE < NP ) WSIZE = NP

       ALLOCATE ( WS(WSIZE),WR(WSIZE) )
       initialize = .false.
    end if

    if(SETUP_FL) then

       !C data creation for AMG
       CALL data_creation_unsym_ssi_amg(N, NP, NPL, NPU, D, AL,       &
            & INL, IAL, AU, INU, IAU, NEIBPETOT, NEIBPE, STACK_IMPORT,&
            & NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SISIZE, SESIZE,   &
            & SOLVER_COMM, LEVEL_NUM, WSIZE2)
       
    else if (LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA, LEVEL_NUM)

    else
       !C in the case, LEVEL_NUM == 0
       stop 'error. no multi-level data is available'

    end if

    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA, LEVEL_NUM)

    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2), WR(WSIZE2))
       WSIZE = WSIZE2
    END if

    
    if(NP_max > NP_old .or. LSIZ > LSIZ_old) then
       deallocate(psy, fai)
       allocate(psy(NP_max, 0:LSIZ), fai(NP_max, 0:LSIZ))
    end if

    !C LSIZ is assumed to be fix during the execution.
    if(NP_max > NP_old) then
       deallocate(Residual_vec, Temp) 
       allocate(Residual_vec(NP_max), Temp(NP_max))
       NP_old  = NP_max
    end if

    if(LSIZ > LSIZ_old) then
       deALLOCATE(AA, AA_bak, alpha)       
       ALLOCATE(AA(0:LSIZ, 0:LSIZ),AA_bak(0:LSIZ, 0:LSIZ), alpha(0:LSIZ+1) ) 
       LSIZ_old = LSIZ
    end if

       




    !C +-----------------------------------+
    !C | {Residual_vec}= ({b} - [A]{xini}) |
    !C +-----------------------------------+
    !C===
    call cal_res(Residual_vec)

    


    zero_ind = 0


    NOPT = 0
    !C-- Begin iteration
    do iter = 1, MAXIT
       !C upper bound of NOPT is LSIZ
       if(NOPT >LSIZ) NOPT = LSIZ

       !C psy and fai have LSIZ+1 vectors. New one is stored at psy or fai(:,zero_ind)
       !C Three vertors  are stored at mod(zero_ind+2, LSIZ+1), mod(zero_ind+1, LSIZ+1),
       !C zero_ind in chronological order.


       !C +-----------------------------------+
       !C | {psy(:,0)} = M^-1 {Residual_vec}  |
       !C +-----------------------------------+
       !$omp parallel do
       do j = 1, NP
          psy(j, zero_ind) = 0       
       end do
       !$omp end parallel do
       CALL amg_cycles(Residual_vec, psy(:, zero_ind), LEVEL_NUM, SOLVER_COMM,&
            & my_rank, WS(1:WSIZE), WR(1:WSIZE), Temp, PETOT, WSIZE, NP)


       !C +----------------------------+
       !C | {alpha(:)} is determined   |
       !C +----------------------------+
       CALL setalpha()

       !C +-----------------------------------------------+
       !C | {psy(:,0) = sum_{i=0,NOPT} alpha(i)*psy(:,i)} |
       !C +-----------------------------------------------+
       !C change psy(:,zero_ind)

       !$omp parallel private(j, i)
       !$omp do
       DO j = 1, N
          psy(j, zero_ind) = alpha(0)*psy(j, zero_ind)
       END DO
       !$omp end do

       do j = 1, NOPT
          !$omp do
          do i = 1, N
             psy(i,zero_ind) = psy(i,zero_ind) + alpha(j)*psy(i,mod(zero_ind+j, LSIZ+1))
          end do
          !$omp end do
       end do
       !$omp end parallel

       !C +------------------------+
       !C | {X} = {X} + {psy(:,0)} |
       !C +------------------------+
       !$omp parallel do
       DO j = 1, N
          X(j) = X(j) + psy(j, zero_ind)
       END DO
       !$omp end parallel do 

       !C +---------------------------+
       !C | {psy(:,i+1)} = {psy(:,i)} |
       !C +---------------------------+
       !C shift psy and fai vector
       !C zero_ind is changed to zero_ind - 1
       if(zero_ind == 0) then
          zero_ind = LSIZ
       else if(zero_ind >0) then
          zero_ind = zero_ind-1
       else
          stop "error zero_ind < 0: solver_AMGRESCUT"
       end if


       !C +-----------------------------+
       !C | {Residual_vec} = {B}-{A}{x} |
       !C +-----------------------------+
       call cal_res(Residual_vec)


!!$       RESID= DNRM2/BNRM2
       call terminal_cond_check(CONV_FL, RESID, Residual_vec(:), N, SOLVER_COMM)

#ifdef PRINT_REZ
       if (my_rank == 0) write (*, 1000) ITER, RESID
1000   format (i5, 1pe16.6)
#endif

       if ( CONV_FL   ) then
          REACH_CONV_FL = .true.
          exit
       end if
       if(terminal_kind==0 .and.  .not. RESID < 1.0E5 ) then
          write(*,*) "Divergence!"
          exit
       end if

       NOPT = NOPT+1
    end do
    
    !C differentiate divergence cases from non-convergent cases
    if(conv_check_fl) then
       if(.not.CONV_FL) then
          call cal_relative_res_norm(RESID, X, B, WS, WR, WSIZE)
          if(RESID > 1.0) then
             iter = -1
          end if
       end if
    end if
    

    if(SETUP_FL) then
       deallocate(WS, WR, Residual_vec, Temp, psy, fai)
       deallocate(AA, AA_bak, alpha)       
       initialize = .true.

       call clear_matrix_ssi_amg(LEVEL_NUM)

    end if

  contains
    subroutine cal_res(res_vec)
      REAL(kind=kreal),intent(out) :: res_vec(:)
      REAL(kind=kreal) :: WVAL, res_norm0
      INTEGER(kind=kint) :: isU,ieU,isL,ieL, i,j,inod


      call SOLVER_SEND_RECV2                                       &
           &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT,&
           &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,             &
           &     WS(1:WSIZE), WR(1:WSIZE), X(1:NP) , SOLVER_COMM,  &
           &     my_rank)


      res_norm0 = 0.0

      !$omp parallel do private(WVAL, isU, ieU, isL, ieL, i, inod) reduction(+:res_norm0)
      do j= 1, N
         WVAL= B(j) - D(j) * X(j)
         isU= INU(j-1) + 1
         ieU= INU(j  ) 
         do i= isU, ieU
            inod= IAU(i)+ZERO_ORIGIN
            WVAL= WVAL - AU(i) * X(inod)
         enddo

         isL= INL(j-1) + 1
         ieL= INL(j  ) 
         do i= isL, ieL
            inod= IAL(i)+ZERO_ORIGIN
            WVAL= WVAL - AL(i) * X(inod)
         enddo
         !C res(1:N)
         res_vec(j)= WVAL
!!$         res_norm0 = res_norm0+WVAL*WVAL
      enddo
      !$omp end parallel do

!!$      CALL MPI_ALLREDUCE(res_norm0, res_norm, 1, MPI_DOUBLE_PRECISION, &
!!$           &             MPI_SUM, SOLVER_COMM, ierr)

!!$      res_norm = dsqrt(res_norm)
    end subroutine cal_res

    subroutine setalpha()
      !$ USE omp_lib
      integer(kind=kint), ALLOCATABLE   :: indx(:)
      integer(kind=kint) :: i,j,k
      integer(kind=kint) :: INFO
      real(kind=kreal) :: EPS

      alpha(0:LSIZ+1) = 0.


      !C Everytime psy(:,0) and psy(:,1) are new vectors.
      !C fai(:,0) is calculated as A x psy(:,0)
      call matvec_fai(zero_ind)
      !C fai(:,1)  is calculated as A x psy(:,1)
      if(NOPT > 0) then
         k = mod(zero_ind+1, LSIZ+1)
         call matvec_fai(k)
      end if


      !C rhs vector is calculated at rank 0 's alpha(:)
      do j = 0, NOPT
         k = mod(zero_ind+j, LSIZ+1)
         call my_dot_product(Residual_vec, fai(:,k), N, alpha(j))
      end do


      if(my_rank == 0) then
         !C shift AA_bak
         !C Only rank 0 has correct values of AA and AA_bak. 
         do j = NOPT, 2, -1
            do i = NOPT, 2, -1
               AA_bak(i,j) = AA_bak(i-1,j-1)
            end do
         end do
      end if


      !C calculate 0th row and column of AA_bak.
      do i = 0, NOPT
         k = mod(zero_ind+i, LSIZ+1)
         call my_dot_product(fai(:,zero_ind), fai(:,k), N, AA_bak(0,i))
         AA_bak(i,0) = AA_bak(0,i)
      end do

      !C calculate 1th row and column of AA_bak.
      do i = 1, NOPT
         j = mod(zero_ind+1, LSIZ+1) 
         k = mod(zero_ind+i, LSIZ+1)
         call my_dot_product(fai(:,j), fai(:,k), N, AA_bak(1,i))
         AA_bak(i,1) = AA_bak(1,i)
      end do


      if(my_rank == 0 ) then
         !C copy AA_bak to AA
         do i = 0, NOPT
            do j = 0, NOPT
               AA(j,i) = AA_bak(j,i)
            end do
         end do

         allocate(indx(NOPT+1))

         EPS = 0.0
         
         call DGETRF(NOPT+1, NOPT+1, AA(0,0), LSIZ+1, indx, INFO)

         if(INFO == 0 .and. NOPT >= 1) then
            call DGETRS('N', NOPT+1,1,AA(0,0), LSIZ+1, indx, alpha(0), LSIZ+2, INFO)
         end if

         deallocate(indx)

         if(INFO/=0 .or. NOPT < 1) then
!!$              if(my_rank ==0) write(*,*) "Reset alpha"
            alpha(0) = 1.0
            alpha(1:NOPT) = 0.
            NOPT = 1
            !C reset flag is recorded at alpha(LSIZ+1)
            alpha(LSIZ+1) = -1.
         else
            !C normal execution
            alpha(LSIZ + 1) = 1.
         end if

      end if

      call MPI_BCAST(alpha(0), LSIZ+2, MPI_DOUBLE_PRECISION, 0, SOLVER_COMM, ierr)
      if(alpha(LSIZ+1)<0) NOPT = 0

    end subroutine setalpha

    !C dot product and the result is collected at rank0 
    subroutine my_dot_product(vec1, vec2, n, v)
      !$ USE omp_lib

      real(kind=kreal), intent(in) :: vec1(:), vec2(:)
      integer(kind=kint), intent(in) :: n
      real(kind=kreal), intent(out) :: v
      real(kind=kreal) :: v0

      v0 = 0; v =0
      !$omp parallel do reduction(+:v0)
      do i = 1,n
         v0 = v0+vec1(i)*vec2(i)
      end do
      !$omp end parallel do

      call MPI_reduce(v0, v, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, SOLVER_COMM, ierr)

    end subroutine my_dot_product


    subroutine matvec_fai(k)
      integer(kind=kint), intent(in) :: k
      INTEGER(kind=kint) :: i,j,isU,ieU,isL,ieL, inod
      real(kind=kreal) :: WVAL


      call SOLVER_SEND_RECV2                                            &
           &   ( NP, WSIZE, WSIZE, NEIBPETOT, NEIBPE, STACK_IMPORT,     &
           &     NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,                  &
           &     WS(1:WSIZE), WR(1:WSIZE), psy(1:NP, k) , SOLVER_COMM,  &
           &     my_rank)

      !$omp parallel do private(WVAL, isU, ieU, isL, ieL, i, inod)
      do j= 1, N
         WVAL= D(j) * psy(j, k)

         isU= INU(j-1) + 1
         ieU= INU(j  ) 
         do i= isU, ieU
            inod= IAU(i)
            WVAL= WVAL + AU(i) * psy(inod, k)
         enddo

         isL= INL(j-1) + 1
         ieL= INL(j  ) 
         do i= isL, ieL
            inod= IAL(i)
            WVAL= WVAL + AL(i) * psy(inod, k)
         enddo
         fai(j, k) = WVAL
      end do
      !$omp end parallel do 
    end subroutine matvec_fai
  end subroutine AMGRESCUT
  
END MODULE  solver_AMGRESCUT
