!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.


#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif


MODULE solver_AMG_NB
CONTAINS

  SUBROUTINE AMG_NB    (N, NP,  NPL, NPU, D,  AL, INL, IAL, AU, INU, IAU,    &
       & B, X, RESID, ITER, my_rank, NEIBPETOT, NEIBPE,                   &
       & STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
       & NPROCS, SETUP_FL, LEVEL_NUM, REACH_CONV_FL, NB, k_or_q_vec, NK   &
       & )


    USE solver_SR_NB
    use data_structure_for_AMG_NB
    use data_creation_AMGCG_NB, only: data_creation_NB
    use common_mod, only: setup_phase, clear_matrix_ssi_amg
    USE count_time_mod
    USE solver_AMGCG_NB, only: v_cycle
    USE terminal_condition, only: terminal_cond_check, BNRM2, terminal_kind
    use residual_calculation, only: cal_res_vec_NB_W, cal_relative_res_norm_NB
    USE amg_parameter, only:conv_check_fl, INDPDT_AGGRE    

    IMPLICIT NONE
!C    include 'precision.inc'
    include 'mpif.h'

    LOGICAL, intent(in) :: SETUP_FL    
    INTEGER(kind=kint ), intent(in)  :: N,NP,NPL,NPU,NB
    REAL   (kind=kreal), INTENT(inout)::  RESID
    INTEGER(kind=kint ), INTENT(inout)::  ITER
    INTEGER(kind=kint ), INTENT(in   )::  my_rank

    INTEGER(kind=kint ), INTENT(in)   :: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in)   :: NPROCS
    LOGICAL, intent(inout) :: REACH_CONV_FL

    !C NP
    REAL   (kind=kreal) ::  D(:,:,:)
    !C N
    REAL   (kind=kreal) ::  B(:)

    !C NP
    REAL   (kind=kreal) ::  X(:)
    REAL   (kind=kreal) ::  AU(:,:,:)
    REAL   (kind=kreal) ::  AL(:,:,:)

    INTEGER(kind=kint ) ::  INU(0:)
    INTEGER(kind=kint ) ::  IAU(:)
    INTEGER(kind=kint ) ::  INL(0:)
    INTEGER(kind=kint ) ::  IAL(:)

    INTEGER(kind=kint ), INTENT(in) :: NK
    real(kind=kreal), pointer :: k_or_q_vec(:,:)
    
    INTEGER(kind=kint ) :: NEIBPE      (:)
    INTEGER(kind=kint ) :: STACK_IMPORT(0:)
    INTEGER(kind=kint ) :: NOD_IMPORT  (:)

    INTEGER(kind=kint ) :: STACK_EXPORT(0:)
    INTEGER(kind=kint ) :: NOD_EXPORT  (:)

    INTEGER(kind=kint ) :: SISIZE,SESIZE

    REAL   (kind=kreal), ALLOCATABLE, save :: WS(:), WR(:)
    REAL   (kind=kreal), ALLOCATABLE, save :: R(:)
    REAL   (kind=kreal), ALLOCATABLE, save :: Temp(:)

    INTEGER(kind=kint ) :: MAXIT

    INTEGER(kind=kint)  :: LEVEL_NUM,cnt,NEIBPETOT,ISU,IEU,ISL,IEL,WSIZE2, NP_max
    INTEGER(kind=kint),save  :: WSIZE, NP_old
    INTEGER(kind=kint)  :: INOD, IERR, i,j,k,l,m
    REAL   (kind=kreal) :: WVAL,BNRM20,BETA,RHO1,DNRM20
    REAL   (kind=kreal) :: DNRM2, v

    logical, save :: initialize = .true.
    logical :: CONV_FL

    CHARACTER(len=20) :: PRECOND

    !C-- INIT. 
    SISIZE = STACK_IMPORT(NEIBPETOT) 
    SESIZE = STACK_EXPORT(NEIBPETOT) 

    if(initialize .or. SETUP_FL) then
       ALLOCATE (R(NP*NB))
       ALLOCATE (Temp(NP*NB))
       NP_old = NP

       !C- decision of the allocation size for WS, WR
       WSIZE = 0
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)*NB
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)*NB
          end if
       end if
       if( WSIZE < NP*NB ) WSIZE = NP*NB
    
       ALLOCATE ( WS(WSIZE),WR(WSIZE) )
       initialize = .false.
    end if
    
    MAXIT  = ITER
!C    conv_criterion_val = RESID 
    !C if terminal_condition isn't set,  RESID will 
    
!C  D(1:NP), AL(1:NPL), INL(0:N), IAL(1:NPL), AU(1:NPU), INU(0:N), IAU(1:NPU)
!C  NEIBPE(1:NEIBPETOT), STACK_IMPORT(0:NEIBPETOT), NOD_IMPORT(1:SISIZE)
!C  STACK_EXPORT(0:NEIBPETOT), NOD_EXPORT(1:SESIZE)
    if(SETUP_FL) then
       
       !C data creation for AMG
       if(INDPDT_AGGRE) then
          PRECOND ="IND1"
       else
          PRECOND ="SHA2"
       end if
       call data_creation_NB                                                   &
            &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU,           &
            &                  INU, IAU,                                       &
            &                  LEVEL_NUM, my_rank, NEIBPETOT, NEIBPE,          &
            &                  STACK_IMPORT, NOD_IMPORT,                       &
            &                  STACK_EXPORT, NOD_EXPORT,                       &
            &                  SOLVER_COMM, PRECOND, NB, k_or_q_vec, NK)     


    end if
    
    WSIZE2 = 0
    if(LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA_NB, LEVEL_NUM)
    else
       !C in the case, LEVEL_NUM == 0
       stop 'error not setup part is created.. in solver_AMG_NB.F90'
    end if

    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA_NB, LEVEL_NUM)
    
    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2),WR(WSIZE2))
       WSIZE = WSIZE2
    END if
    if(NP_max > NP_old) then
       deallocate(R,TEMP)
       allocate(R(NP*NB), TEMP(NP*NB))
       NP_old = NP_max
    END if

    !C-------- allocation of WS, WS, WW, temp is finished.
    
    if (NEIBPETOT > 0) then
       call SOLVER_SEND_RECV_NB                                                   &
            &   ( SESIZE,SISIZE, NP, NEIBPETOT,                                    &
            &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS, &
            &     WR, X , SOLVER_COMM, my_rank,NB)
       
    end if
    

    !C
    !C +-----------------------+
    !C | {r0}= {b} - [A]{xini} |
    !C +-----------------------+
    !C===
    !C-- BEGIN calculation
!C    call cal_res_vec_NB(R, X, B, WS, WR, WSIZE, NB)

!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(+:BNRM20)
!!$       do i= 1, N
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       !$omp end parallel do
!!$       
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$    else if(terminal_kind == 1) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(max:BNRM20)
!!$       do i= 1, N
!!$          WVAL = dabs(B(i))
!!$          if(BNRM20 < WVAL) BNRM20 = WVAL
!!$       enddo
!!$       !$omp end parallel do
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_MAX, SOLVER_COMM, ierr)
!!$    else 
!!$       stop 'unknown terminal_kind: solver_AMG_NB.F90'
!!$    end if
!!$       
!!$    if(BNRM2.eq.0.d0) then
!!$       BNRM2= 1.d0
!!$    end if
    
    do iter = 1, MAXIT

       CALL v_cycle(B, X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, TEMP(:), &
            &          WSIZE)

!!$       if(NEIBPETOT > 0) then
!!$          call SOLVER_SEND_RECV_NB                                                    &
!!$               &   ( NP, WSIZE, WSIZE, NEIBPETOT,                                     &
!!$               &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS,  &
!!$               &     WR, X, SOLVER_COMM, my_rank)
!!$       END if
       
       !C residual calculation
       call cal_res_vec_NB_W(R, X, B, WS, WR, WSIZE, NB)

       !C calculate relative residual norm on RESID
       call terminal_cond_check(CONV_FL, RESID, R, N*NB, SOLVER_COMM)

#ifdef PRINT_REZ
       if (my_rank.eq.0 .and. terminal_kind==0) write (*, 1000) ITER, RESID
 1000   format (i5, 1pe16.6)
#endif
      
      if ( CONV_FL  ) then
         REACH_CONV_FL = .true.
         exit
      end if
   end do
   
#ifdef PRINT_INFO
   call matrix_counting(LEVEL_NUM,SOLVER_COMM,my_rank)
   call count_time(4,solver_comm,my_rank,1)
#endif
     
   !C differentiate divergence cases from non-convergent cases
   if(conv_check_fl .and..not.CONV_FL) then
      call cal_relative_res_norm_NB(RESID, X, B, WS, WR, WSIZE, NB)      
      
      if(RESID > 1.0) then
         iter = -1
      end if
   end if
   
   if(SETUP_FL) then
      call clear_matrix_ssi_amg(LEVEL_NUM)
      DEALLOCATE (R)
      DEALLOCATE (WR)
      DEALLOCATE (WS)
      DEALLOCATE (Temp)
      initialize = .true.
   end if

#ifdef PSPASES
   CALL PSPACEC(pspcommf,0)
#endif
     
 END SUBROUTINE AMG_NB
   
   
END MODULE  solver_AMG_NB
