
MODULE solver_GPU
CONTAINS
  
  !C SETUP_FL shows if this subroutine contains the setup phase
  !C and deletes the matrix
  SUBROUTINE solver_on_GPU (N, NP,  NPL, NPU, D,  AL, INL, IAL, AU, INU, IAU,    &
       & B, X, RESID, ITER, my_rank, NEIBPETOT, NEIBPE,                   &
       & STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
       & NPROCS, SETUP_FL, LEVEL_NUM, REACH_CONV_FL, sol, omega     &
       & )
    USE solver_SR2    
    USE data_structure_for_AMG
    USE data_creation_AMGCG
    USE terminal_condition, only: terminal_kind, CONV_CRITERION_VAL, CONV_CRITERION_VEC, reorder_conv_vec
    USE amg_parameter, only: conv_check_fl
    USE residual_calculation, only: cal_relative_res_norm
    USE common_mod, only: setup_phase
    
    !C CONV_CRITERION_VAL: criterion value for 2norm of relative residual with rhs vector
    !C CONV_CRITERION_VEC: criterion vector for residual
     
    IMPLICIT NONE    

    LOGICAL, intent(in) :: SETUP_FL
    INTEGER(kind=kint ), INTENT(in   )::  my_rank
    INTEGER(kind=kint ), INTENT(in)   :: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in)   :: NPROCS

    LOGICAL, intent(inout) :: REACH_CONV_FL
    INTEGER(kind=kint ), intent(inout) :: LEVEL_NUM
    INTEGER(kind=kint ), intent(in)    :: N, NP,  NPL, NPU
    REAL   (kind=kreal)                :: RESID, TOL
    INTEGER(kind=kint ), intent(inout) :: ITER

    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAU
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  INL
    INTEGER(kind=kint ), DIMENSION(:), pointer ::  IAL

    REAL   (kind=kreal), DIMENSION(: ), pointer ::  D
    REAL   (kind=kreal), DIMENSION(: )  ::  B
    REAL   (kind=kreal), DIMENSION(: )  ::  X

    REAL   (kind=kreal), DIMENSION(:), pointer ::  AU
    REAL   (kind=kreal), DIMENSION(:), pointer ::  AL

    INTEGER(kind=kint ) :: NEIBPE      (:), NEIBPETOT
    INTEGER(kind=kint ) :: STACK_IMPORT(0:)
    INTEGER(kind=kint ) :: NOD_IMPORT  (:)

    INTEGER(kind=kint ) :: STACK_EXPORT(0:)
    INTEGER(kind=kint ) :: NOD_EXPORT  (:)


    INTEGER(kind=kint),intent(in)  :: sol
    real(kind=kreal),intent(in)  :: omega

    INTEGER(kind=kint)  :: ROWSIZE 
    
    real(kind=kreal), allocatable :: WS(:), WR(:)
    integer(kind=kint) :: SISIZE, SESIZE, WSIZE2

    integer(kind=kint) :: i
    real(kind=kreal) :: v
    
    REACH_CONV_FL = .false.
    
!C    conv_criterion_val   = RESID           

    IF(SETUP_FL) THEN
       call setup_phase()
!!$       CALL data_creation_ssi_amg(N, NP, NPL, NPU, D, AL, INL, &
!!$       CALL data_creation_unsym_ssi_amg(N, NP, NPL, NPU, D, AL, INL, &
!!$            & IAL, AU, INU, IAU, NEIBPETOT,                    &
!!$            & NEIBPE, STACK_IMPORT, NOD_IMPORT,                &
!!$            & STACK_EXPORT, NOD_EXPORT,                        &
!!$            & SISIZE, SESIZE, SOLVER_COMM, LEVEL_NUM, WSIZE2)
!!$       
!!$       call mat_and_solver_set_on_GPU(LEVEL_NUM, sol, omega)
    END IF
    
!C  solverAMG_S.F90 170, 171 set call
    if(terminal_kind == 0) then
       call amgs_terminal_val_set_on_gpu(CONV_CRITERION_VAL)       
    else if(terminal_kind == 1) then
       ROWSIZE = HIERARCHICAL_DATA(1) % N
       call reorder_conv_vec(rowsize)
       call amgs_terminal_vec_set_on_gpu(CONV_CRITERION_VEC, ROWSIZE);
    else
       stop "No terminal condition is set in solver_GPU()"
    end if


!C  solverAMG_S.F90 39 set maxiter
    call solve_on_GPU(X, B, iter)

    if(conv_check_fl) then
       allocate(WS(NP), WR(NP))
       WSIZE2 = NP
       call cal_relative_res_norm(RESID, X, B, WS, WR, WSIZE2)
       if(RESID > 1.0) then
          iter = -1
       end if
       deallocate(WS,WR)
    end if

    if(SETUP_FL) then
       call clear_mat_on_GPU()
    end if

  END SUBROUTINE SOLVER_ON_GPU
  
  
  
END MODULE  solver_GPU
