module solver_SR_NB
contains

  subroutine  SOLVER_SEND_RECV_NB                                         &
       & ( WS_size, WR_size, X_size, NEIBPETOT, NEIBPE,                   &
       &   STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS, WR, X, &
       &   SOLVER_COMM, my_rank, NB)


    implicit none
    include  'mpif.h'
    include  'precision.inc'


    integer(kind=kint )                , intent(in)   ::  NB
    integer(kind=kint )                , intent(in)   ::  X_size
    integer(kind=kint )                , intent(in)   ::  WS_size,WR_size
    integer(kind=kint )                , intent(in)   ::  NEIBPETOT
    integer(kind=kint )  :: NEIBPE      (:)
    integer(kind=kint )  :: STACK_IMPORT(0:)
    integer(kind=kint )  :: NOD_IMPORT  (:)
    integer(kind=kint )  :: STACK_EXPORT(0:)
    integer(kind=kint )  :: NOD_EXPORT  (:)
!!$    real   (kind=kreal), dimension(NB*WS_size), intent(inout):: WS
!!$    real   (kind=kreal), dimension(NB*WR_size), intent(inout):: WR
!!$    real   (kind=kreal), dimension(NB*X_size), intent(inout):: X
    real   (kind=kreal), dimension(:), intent(inout):: WS
    real   (kind=kreal), dimension(:), intent(inout):: WR
    real   (kind=kreal), dimension(:), intent(inout):: X
    integer                            , intent(in)   ::SOLVER_COMM
    integer                            , intent(in)   :: my_rank

    integer(kind=kint ), dimension(:,:), save, allocatable :: sta1
    integer(kind=kint ), dimension(:,:), save, allocatable :: sta2
    integer(kind=kint ), dimension(:  ), save, allocatable :: req1
    integer(kind=kint ), dimension(:  ), save, allocatable :: req2  


    integer(kind=kint ), save :: NFLAG=0
    integer(kind=kint ), save :: EARLIER_NEIBPETOT=0
    integer(kind=kint ) :: i, j, k, neib, istart, inum, ii, ierr

    !C
    !C-- INIT.
    if (NFLAG.eq.0) then
       allocate (sta1(MPI_STATUS_SIZE,NEIBPETOT))
       allocate (sta2(MPI_STATUS_SIZE,NEIBPETOT))
       allocate (req1(NEIBPETOT))
       allocate (req2(NEIBPETOT))
       NFLAG= 1
       EARLIER_NEIBPETOT=NEIBPETOT
    else if(EARLIER_NEIBPETOT<NEIBPETOT)then
       deallocate(sta1);deallocate(sta2)
       deallocate(req1);deallocate(req2)
       allocate (sta1(MPI_STATUS_SIZE,NEIBPETOT))
       allocate (sta2(MPI_STATUS_SIZE,NEIBPETOT))
       allocate (req1(NEIBPETOT))
       allocate (req2(NEIBPETOT))
       EARLIER_NEIBPETOT=NEIBPETOT
    endif

    !C
    !C-- SEND

    do neib= 1, NEIBPETOT
       istart= STACK_EXPORT(neib-1)
       inum  = STACK_EXPORT(neib  ) - istart

       do k= istart+1, istart+inum
          ii   = NB*NOD_EXPORT(k)
          do j = NB-1, 0, -1
             WS(NB*k-j)= X(ii-j)
          end do
       enddo
       call MPI_ISEND (WS(NB*istart+1), NB*inum,MPI_DOUBLE_PRECISION,    &
            &                  NEIBPE(neib), 0, SOLVER_COMM, req1(neib), ierr)
    enddo

    !C
    !C-- RECEIVE
    do neib= 1, NEIBPETOT
       istart= STACK_IMPORT(neib-1)
       inum  = STACK_IMPORT(neib  ) - istart
       call MPI_IRECV (WR(NB*istart+1), NB*inum, MPI_DOUBLE_PRECISION,   &
            &                  NEIBPE(neib), 0, SOLVER_COMM, req2(neib), ierr)
    enddo

    call MPI_WAITALL (NEIBPETOT, req2, sta2, ierr)

    do neib= 1, NEIBPETOT
       istart= STACK_IMPORT(neib-1)
       inum  = STACK_IMPORT(neib  ) - istart
       do k= istart+1, istart+inum
          ii   = NB*NOD_IMPORT(k)
          do j = NB-1, 0, -1
             X(ii-j)= WR(NB*k-j)
          end do
       enddo
    enddo

    call MPI_WAITALL (NEIBPETOT, req1, sta1, ierr)

  end subroutine SOLVER_SEND_RECV_NB

  subroutine  SEND_RECV_buf                                               &
       &                ( WS_size,WR_size,NEIBPETOT, NEIBPE,              &
       &                  STACK_IMPORT, NOD_IMPORT, STACK_EXPORT,         &
       &                   NOD_EXPORT, WS, WR, SOLVER_COMM, my_rank)


    implicit none
    include  'mpif.h'
    include  'precision.inc'

    integer(kind=kint )                , intent(in)   ::  WS_size,WR_size
    integer(kind=kint )                , intent(in)   ::  NEIBPETOT
    integer(kind=kint ), pointer :: NEIBPE      (:)
    integer(kind=kint ), pointer :: STACK_IMPORT(:)
    integer(kind=kint ), pointer :: NOD_IMPORT  (:)
    integer(kind=kint ), pointer :: STACK_EXPORT(:)
    integer(kind=kint ), pointer :: NOD_EXPORT  (:)
!!$    integer(kind=kint ), dimension(WS_size), intent(inout):: WS
!!$    integer(kind=kint), dimension(WR_size), intent(inout):: WR
    integer(kind=kint ), dimension(:), intent(inout):: WS
    integer(kind=kint), dimension(:), intent(inout):: WR
    integer                            , intent(in)   ::SOLVER_COMM
    integer                            , intent(in)   :: my_rank

    integer(kind=kint ), dimension(:,:), allocatable :: sta1
    integer(kind=kint ), dimension(:,:), allocatable :: sta2
    integer(kind=kint ), dimension(:  ), allocatable :: req1
    integer(kind=kint ), dimension(:  ), allocatable :: req2  


    integer(kind=kint ) :: neib, istart, inum, ierr

    !C
    !C-- INIT.
    allocate (sta1(MPI_STATUS_SIZE,NEIBPETOT))
    allocate (sta2(MPI_STATUS_SIZE,NEIBPETOT))
    allocate (req1(NEIBPETOT))
    allocate (req2(NEIBPETOT))

    !C
    !C-- SEND
    do neib= 1, NEIBPETOT
       istart= STACK_EXPORT(neib-1)
       inum  = STACK_EXPORT(neib  ) - istart
       call MPI_ISEND (WS(istart+1), inum, MPI_INTEGER,    &
            &          NEIBPE(neib), 0, SOLVER_COMM, req1(neib), ierr)
    enddo

    !C
    !C-- RECEIVE
    do neib= 1, NEIBPETOT
       istart= STACK_IMPORT(neib-1)
       inum  = STACK_IMPORT(neib  ) - istart
       call MPI_IRECV (WR(istart+1), inum, MPI_INTEGER,   &
            &          NEIBPE(neib), 0, SOLVER_COMM, req2(neib), ierr)
    enddo

    call MPI_WAITALL (NEIBPETOT, req2, sta2, ierr)
    call MPI_WAITALL (NEIBPETOT, req1, sta1, ierr)

    deallocate(sta1,sta2,req1,req2)

  end subroutine SEND_RECV_BUF

  subroutine  SOLVER_SEND_RECV2I                                      &
       &                ( N, S,R,NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, &
       &                                        STACK_EXPORT, NOD_EXPORT, &
       &                  WS, WR, X, SOLVER_COMM,my_rank)
    
    implicit none
    include  'mpif.h'
    include  'precision.inc'

    integer(kind=kint )                , intent(in)   ::  N
    integer(kind=kint )                , intent(in)   ::  S
    integer(kind=kint )                , intent(in)   ::  R
    integer(kind=kint )                , intent(in)   ::  NEIBPETOT
    integer(kind=kint ) :: NEIBPE      (:)
    integer(kind=kint ) :: STACK_IMPORT(0:)
    integer(kind=kint ) :: NOD_IMPORT  (:)
    integer(kind=kint ) :: STACK_EXPORT(0:)
    integer(kind=kint ) :: NOD_EXPORT  (:)
!!$    integer(kind=kint ), dimension(S)  , intent(inout):: WS
!!$    integer(kind=kint ), dimension(R)  , intent(inout):: WR
!!$
!!$    INTEGER(kind=kint ), dimension(N)  , intent(inout):: X
    integer(kind=kint ), dimension(:)  , intent(inout):: WS
    integer(kind=kint ), dimension(:)  , intent(inout):: WR

    INTEGER(kind=kint ), dimension(:)  , intent(inout):: X
    integer                            , intent(in)   ::SOLVER_COMM
    integer                            , intent(in)   :: my_rank


    integer(kind=kint ), dimension(:,:), save, allocatable :: sta1
    integer(kind=kint ), dimension(:,:), save, allocatable :: sta2
    integer(kind=kint ), dimension(:  ), save, allocatable :: req1
    integer(kind=kint ), dimension(:  ), save, allocatable :: req2  
    
    
    integer(kind=kint) :: neib,istart,inum,k,ierr
    integer(kind=kint ), save :: NFLAG=0
    integer(kind=kint ), save :: EARLIER_NEIBPETOT=0
    !C
    !C-- INIT.
    if (NFLAG.eq.0) then
       allocate (sta1(MPI_STATUS_SIZE,NEIBPETOT))
       allocate (sta2(MPI_STATUS_SIZE,NEIBPETOT))
       allocate (req1(NEIBPETOT))
       allocate (req2(NEIBPETOT))
       NFLAG= 1
       EARLIER_NEIBPETOT=NEIBPETOT
    else if(EARLIER_NEIBPETOT<NEIBPETOT)then
       deallocate(sta1);deallocate(sta2)
       deallocate(req1);deallocate(req2)
       allocate (sta1(MPI_STATUS_SIZE,NEIBPETOT))
       allocate (sta2(MPI_STATUS_SIZE,NEIBPETOT))
       allocate (req1(NEIBPETOT))
       allocate (req2(NEIBPETOT))
       EARLIER_NEIBPETOT=NEIBPETOT
    endif

    
    !C
    !C-- SEND
    do neib= 1, NEIBPETOT
       istart= STACK_EXPORT(neib-1)
       inum  = STACK_EXPORT(neib  ) - istart
       do k= istart+1, istart+inum
          WS(k)= X(NOD_EXPORT(k))
       enddo
       call MPI_ISEND (WS(istart+1), inum, MPI_INTEGER,       &
            &                  NEIBPE(neib), 13, SOLVER_COMM,  &
            &                  req1(neib), ierr)
    enddo

    !C
    !C-- RECEIVE
    do neib= 1, NEIBPETOT
       istart= STACK_IMPORT(neib-1)
       inum  = STACK_IMPORT(neib  ) - istart
       call MPI_IRECV (WR(istart+1), inum, MPI_INTEGER,        &
            &                  NEIBPE(neib), 13, SOLVER_COMM,  &
            &                  req2(neib), ierr)
    enddo
    call MPI_WAITALL (NEIBPETOT, req2, sta2, ierr)
    
    do neib= 1, NEIBPETOT
       istart= STACK_IMPORT(neib-1)
       inum  = STACK_IMPORT(neib  ) - istart
       do k= istart+1, istart+inum
          X(NOD_IMPORT(k))= WR(k)
       enddo
    enddo

    call MPI_WAITALL (NEIBPETOT, req1, sta1, ierr)
  end subroutine solver_send_recv2I


end module     solver_SR_NB
