!C   Copyright (C) 2010 High Performance Computing Laboratory, Kogakuin University.
!C   All rights reserved.
!C
!C   Redistribution and use in source and binary forms, with or without
!C   modification, are permitted provided that the following conditions are met:
!C   1. Redistributions of source code must retain the above copyright
!C      notice, this list of conditions and the following disclaimer.
!C   2. Redistributions in binary form must reproduce the above copyright
!C      notice, this list of conditions and the following disclaimer in the
!C      documentation and/or other materials provided with the distribution.
!C   3. Neither the name of the project nor the names of its contributors
!C      may be used to endorse or promote products derived from this software
!C      without specific prior written permission.
!C
!C   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
!C   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
!C   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
!C   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
!C   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
!C   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
!C   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
!C   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
!C   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
!C   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
!C   POSSIBILITY OF SUCH DAMAGE.


#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif


MODULE solver_tuning_eval
CONTAINS

  SUBROUTINE tuning_eval    (N, NP,  NPL, NPU, D,  AL, INL, IAL, AU, INU, IAU,    &
       & B, X, RESID, ITER, my_rank, NEIBPETOT, NEIBPE,                   &
       & STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, SOLVER_COMM, &
       & NPROCS, SETUP_FL, LEVEL_NUM, REACH_CONV_FL                       &
       & )
    
    USE solver_SR2
    USE data_structure_for_AMG
    USE common_mod, only: clear_matrix_ssi_amg
    USE solver_Gnumbering
    USE count_time_mod
    USE data_creation_AMGCG
    USE AMG_cycle
    USE cycle_shape_tuning, only: initialize_record_data, v_cycle_at_each_lev, record_best_cycle, &
         & precision_arry, precision_sz, record_data
    USE terminal_condition, only:terminal_cond_vec_set, terminal_cond_val_set, &
         & terminal_cond_check,  terminal_kind, conv_criterion_val
    USE residual_calculation
    USE amg_parameter, only: conv_check_fl, cycle_kind, iteration_number_up, &
         & iteration_number_down, iteration_number_bottom, gs_kind,          &
         & dump_gauss_seidel_weight


    IMPLICIT NONE
    include 'mpif.h'

    LOGICAL, intent(in) :: SETUP_FL    
    INTEGER(kind=kint ), intent(in)  :: N,NP,NPL,NPU
    REAL   (kind=kreal), INTENT(inout)::  RESID
    INTEGER(kind=kint ), INTENT(inout)::  ITER
    INTEGER(kind=kint ), INTENT(in   )::  my_rank

    INTEGER(kind=kint ), INTENT(in)   :: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in)   :: NPROCS
    LOGICAL, intent(inout) :: REACH_CONV_FL

    !C NP
    REAL   (kind=kreal) ::  D(:)
    !C N
    REAL   (kind=kreal) ::  B(:)

    !C NP
    REAL   (kind=kreal) ::  X(:)
    REAL   (kind=kreal) ::  AU(:)
    REAL   (kind=kreal) ::  AL(:)

    INTEGER(kind=kint ) ::  INU(0:)
    INTEGER(kind=kint ) ::  IAU(:)
    INTEGER(kind=kint ) ::  INL(0:)
    INTEGER(kind=kint ) ::  IAL(:)
    
    INTEGER(kind=kint ) :: NEIBPE      (:)
    INTEGER(kind=kint ) :: STACK_IMPORT(0:)
    INTEGER(kind=kint ) :: NOD_IMPORT  (:)

    INTEGER(kind=kint ) :: STACK_EXPORT(0:)
    INTEGER(kind=kint ) :: NOD_EXPORT  (:)

    INTEGER(kind=kint ) :: SISIZE,SESIZE

    REAL   (kind=kreal), ALLOCATABLE, save :: WS(:), WR(:)
    REAL   (kind=kreal), ALLOCATABLE, save :: R(:)
    REAL   (kind=kreal), ALLOCATABLE, save :: Temp(:)

    INTEGER(kind=kint ) :: MAXIT

    INTEGER(kind=kint)  :: LEVEL_NUM,cnt,NEIBPETOT,ISU,IEU,ISL,IEL,WSIZE2
    INTEGER(kind=kint),save  :: WSIZE, NP_old
    INTEGER(kind=kint)  :: INOD, IERR, i,j,k,l,m, NP_max, try, pre_level
    REAL   (kind=kreal) :: WVAL,BETA,RHO1,DNRM20
    REAL   (kind=kreal) :: DNRM2, o

    logical, save :: initialize = .true.
    logical :: CONV_FL

    REAL   (kind=kreal),allocatable :: correct_vec(:)
    REAL   (kind=kreal) :: BNRM2, BNRM20, diff_NRM20, diff_NRM2
    
    !C-- INIT. 

    SISIZE = STACK_IMPORT(NEIBPETOT) 
    SESIZE = STACK_EXPORT(NEIBPETOT) 

    if(initialize .or. SETUP_FL) then
       ALLOCATE (R(NP))
       ALLOCATE ( Temp(NP) )
       NP_old = NP

       !C- decision of the allocation size for WS, WR
       WSIZE = 0
       if(NEIBPETOT > 0) then
          if( STACK_IMPORT(NEIBPETOT) > STACK_EXPORT(NEIBPETOT) ) then
             WSIZE = STACK_IMPORT(NEIBPETOT)
          else
             WSIZE = STACK_EXPORT(NEIBPETOT)
          end if
       end if
       if( WSIZE < NP ) WSIZE = NP
    
       ALLOCATE ( WS(WSIZE),WR(WSIZE) )
       initialize = .false.
    end if
    
    R = 1
    
    MAXIT  = ITER
!C    conv_criterion_val = RESID 
    !C if terminal_condition isn't set,  RESID will 
    
!C  D(1:NP), AL(1:NPL), INL(0:N), IAL(1:NPL), AU(1:NPU), INU(0:N), IAU(1:NPU)
!C  NEIBPE(1:NEIBPETOT), STACK_IMPORT(0:NEIBPETOT), NOD_IMPORT(1:SISIZE)
!C  STACK_EXPORT(0:NEIBPETOT), NOD_EXPORT(1:SESIZE)
    if(SETUP_FL) then

!!$       CALL data_creation_ssi_amg(N, NP, NPL, NPU, D, AL, INL, &
       CALL data_creation_unsym_ssi_amg(N, NP, NPL, NPU, D, AL, INL, &
            & IAL, AU, INU, IAU, NEIBPETOT,                    &
            & NEIBPE, STACK_IMPORT, NOD_IMPORT,                &
            & STACK_EXPORT, NOD_EXPORT,                        &
            & SISIZE, SESIZE, SOLVER_COMM, LEVEL_NUM, WSIZE2)
    else if (LEVEL_NUM > 0) then
       !C LEVEL_NUM is assumed to be > 0 in this sub.
       call max_comm_table_sz(WSIZE2, HIERARCHICAL_DATA, LEVEL_NUM)
    else
       !C in the case, LEVEL_NUM == 0
       stop 'error. no multi-level data is available'
    end if
    
    !C-- NP_max is equal to the max size of unknowns on all levels.
    call max_NP_on_all_levs(NP_max, NP, HIERARCHICAL_DATA, LEVEL_NUM)
    

    if(WSIZE2 > WSIZE) then
       deallocate(WS,WR)
       allocate(WS(WSIZE2),WR(WSIZE2))
       WSIZE = WSIZE2
    END if
    if(NP_max > NP_old) then
       deallocate(R,TEMP)
       allocate(R(NP_max), TEMP(NP_max))
       NP_old = NP_max
    END if


       !C-------- allocation of WS, WS, WW, temp is finished.
    

!!$    if (NEIBPETOT > 0) then
!!$       call SOLVER_SEND_RECV2                                                     &
!!$            &   ( NP,SESIZE,SISIZE, NEIBPETOT,                                    &
!!$            &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS, &
!!$            &     WR, X , SOLVER_COMM, my_rank)
!!$       
!!$    end if
    

    !C
    !C +-----------------------+
    !C | {r0}= {b} - [A]{xini} |
    !C +-----------------------+
    !C===
    !C-- BEGIN calculation
!!$    do j= 1, N
!!$       WVAL= B(j) - D(j) * X(j)
!!$       isU= INU(j-1) + 1
!!$       ieU= INU(j  ) 
!!$
!!$       do i= isU, ieU
!!$          inod= IAU(i)+ZERO_ORIGIN
!!$          WVAL= WVAL - AU(i) * X(inod)
!!$       enddo
!!$
!!$       isL= INL(j-1) + 1
!!$       ieL= INL(j  ) 
!!$       do i= isL, ieL
!!$          inod= IAL(i)+ZERO_ORIGIN
!!$          WVAL= WVAL - AL(i) * X(inod)
!!$       enddo
!!$       R(j)= WVAL
!!$    enddo

!!$    if(terminal_kind == 0) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(+:BNRM20)
!!$       do i= 1, N
!!$          BNRM20 = BNRM20 + B(i)**2
!!$       enddo
!!$       !$omp end parallel do
!!$       
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$    else if(terminal_kind == 1) then
!!$       BNRM20 = 0.d0
!!$       !$omp parallel do reduction(max:BNRM20)
!!$       do i= 1, N
!!$          WVAL = dabs(B(i))
!!$          if(BNRM20 < WVAL) BNRM20 = WVAL
!!$       enddo
!!$       !$omp end parallel do
!!$       call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_MAX, SOLVER_COMM, ierr)
!!$    else 
!!$       stop 'unknown terminal_kind: solver_AMG.F90'
!!$    end if
!!$       
!!$    if(BNRM2.eq.0.d0) then
!!$       BNRM2= 1.d0
!!$    end if


    call initialize_record_data(LEVEL_NUM)
    call record_best_cycle(LEVEL_NUM, SOLVER_COMM, Temp, WS, WR, WSIZE, my_rank)
    
    if(my_rank == 0) then
       do i = 1, LEVEL_NUM
          do j = 1, precision_sz
             write(*,*) "LEVEL", i, "precision", j
             write(*,*) record_data(j,i) % viter, record_data(j,i) % smiter, record_data(j,i) % coarser_level_precision, &
                  & record_data(j,i) % omega
          end do
       end do
    end if

    !C-- problem's correct solution, right hand side vec, sol vec are set.
    allocate(correct_vec(NP))
    call random_number(correct_vec)
    IF(NEIBPETOT > 0) then
       CALL SOLVER_SEND_RECV2(&
            & NP, &
            & STACK_EXPORT(NEIBPETOT),  &
            & STACK_IMPORT(NEIBPETOT),  &
            & NEIBPETOT,                &
            & NEIBPE,                   &
            & STACK_IMPORT,             &
            & NOD_IMPORT,               &
            & STACK_EXPORT,             &
            & NOD_EXPORT,               &
            & WS, WR, correct_vec, SOLVER_COMM, my_rank)
    end IF
    X = 0                
    call cal_res_vec_at_nth_lev(X, correct_vec, B, 1)
    B = -B 
    !C-- end: problem's correct solution, right hand side vec, sol vec are set.

    !C-- set BNRM2 for  new conv 
    BNRM20 = 0
    DO i = 1, N
       BNRM20 = BNRM20 + B(i) ** 2
    END DO
    call MPI_allREDUCE (BNRM20, BNRM2, 1, MPI_DOUBLE_PRECISION,       &
         &                    MPI_SUM, SOLVER_COMM, ierr)


    MAXIT = 100
    do try = 1, precision_sz
       
       !C----- From here, -----
       if(my_rank == 0) &
            & write(*,*)"AMG setting:", cycle_kind, iteration_number_up, &
            & iteration_number_down, iteration_number_bottom, gs_kind,   &
            & dump_gauss_seidel_weight
       X = 0
       
       call count_time(0,solver_comm,my_rank,21)       
       call count_time(1,solver_comm,my_rank,21)       
       do iter = 1, MAXIT

          CALL amg_cycles(B, X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, TEMP, &
            &          NPROCS, WSIZE, NP)

!!$          if(NEIBPETOT > 0) then
!!$             call SOLVER_SEND_RECV2                                                      &
!!$                  &   ( NP, WSIZE, WSIZE, NEIBPETOT,                                     &
!!$                  &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS,  &
!!$                  &     WR, X, SOLVER_COMM, my_rank)
!!$          END if

          !C residual calculation
!!$          !$omp parallel do private(j,k)
!!$          do i= 1, N
!!$             R(i)= B(i)
!!$             do j = INL(i-1)+1,INL(i)
!!$                k = IAL(j) 
!!$                R(i) = R(i)-AL(j)*X(k)
!!$             end do
!!$             do j = INU(i-1)+1,INU(i)
!!$                k = IAU(j) 
!!$                R(i) = R(i)-AU(j)*X(k)
!!$             end do
!!$             R(i) = R(i)-D(i)*X(i)
!!$          enddo
!!$          !$omp end parallel do

          !C calculate relative residual norm on RESID
!!$       DNRM20 = 0.0
!!$       do i= 1, N
!!$          DNRM20 = DNRM20 + R(i)**2
!!$       enddo
!!$       call MPI_allREDUCE (DNRM20, DNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$       RESID= dsqrt(DNRM2/BNRM2)

!!$          call terminal_cond_check(CONV_FL, RESID, R, N, SOLVER_COMM)


       diff_NRM20 = 0
       DO i = 1, N
          diff_NRM20 = diff_NRM20 + (X(i)-correct_vec(i)) ** 2
       END DO
       call MPI_allREDUCE (diff_NRM20, diff_NRM2, 1, MPI_DOUBLE_PRECISION,       &
            &                    MPI_SUM, SOLVER_COMM, ierr)
       RESID = dsqrt(diff_nrm2/BNRM2)
       conv_FL = (RESID < precision_arry(try))


#ifdef PRINT_REZ
!!$       if (my_rank.eq.0 .and. terminal_kind==0) write (*, 1000) ITER, RESID
       if (my_rank.eq.0) write (*, 1000) ITER, RESID, precision_arry(try)

1000      format (i5, 1pe16.6, 1pe16.6)
#endif

          if ( CONV_FL  ) then
             REACH_CONV_FL = .true.
             exit
          end if
       end do
       call count_time(2,solver_comm,my_rank,21)       
       if(my_rank == 0) write(*,*) "precision lev", try, precision_arry(try)
       call count_time(3,solver_comm,my_rank,21)       



!C--- from here, AT version 
       X = 0
       pre_level = try
       call count_time(0,solver_comm,my_rank,21)       
       call count_time(1,solver_comm,my_rank,21)       
       do iter = 1, MAXIT

!!$       CALL amg_cycles(B, X, LEVEL_NUM, SOLVER_COMM, my_rank, WS, WR, TEMP, &
!!$            &          NPROCS, WSIZE, NP)

          call v_cycle_at_each_lev(B, X, Temp, 1, record_data(pre_level, 1) % smiter, record_data(pre_level, 1) % omega, &
               & record_data(try, 1) % coarser_level_precision, 1, &
               & WS, WR, WSIZE, SOLVER_COMM, my_rank)
          
!!$          if(NEIBPETOT > 0) then
!!$             call SOLVER_SEND_RECV2                                                      &
!!$                  &   ( NP, WSIZE, WSIZE, NEIBPETOT,                                     &
!!$                  &     NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT, WS,  &
!!$                  &     WR, X, SOLVER_COMM, my_rank)
!!$          END if

!!$          !C residual calculation
!!$          !$omp parallel do
!!$          do i= 1, N
!!$             R(i)= B(i)
!!$             do j = INL(i-1)+1,INL(i)
!!$                k = IAL(j) 
!!$                R(i) = R(i)-AL(j)*X(k)
!!$             end do
!!$             do j = INU(i-1)+1,INU(i)
!!$                k = IAU(j) 
!!$                R(i) = R(i)-AU(j)*X(k)
!!$             end do
!!$             R(i) = R(i)-D(i)*X(i)
!!$          enddo
!!$          !$omp end parallel do
          

          !C calculate relative residual norm on RESID
!!$       DNRM20 = 0.0
!!$       do i= 1, N
!!$          DNRM20 = DNRM20 + R(i)**2
!!$       enddo
!!$       call MPI_allREDUCE (DNRM20, DNRM2, 1, MPI_DOUBLE_PRECISION,       &
!!$            &                    MPI_SUM, SOLVER_COMM, ierr)
!!$       RESID= dsqrt(DNRM2/BNRM2)

!!$          call terminal_cond_check(CONV_FL, RESID, R, N, SOLVER_COMM)

          diff_NRM20 = 0
          DO i = 1, N
             diff_NRM20 = diff_NRM20 + (X(i)-correct_vec(i)) ** 2
          END DO
          call MPI_allREDUCE (diff_NRM20, diff_NRM2, 1, MPI_DOUBLE_PRECISION,       &
               &                    MPI_SUM, SOLVER_COMM, ierr)
          RESID = dsqrt(diff_nrm2/BNRM2)
          conv_FL = (RESID < precision_arry(try))



#ifdef PRINT_REZ
!!$       if (my_rank.eq.0 .and. terminal_kind==0) write (*, 1000) ITER, RESID
          if (my_rank.eq.0) write (*, 1000) ITER, RESID
#endif

          if ( CONV_FL  ) then
             REACH_CONV_FL = .true.
             exit
          end if

          !C planed iteration is finished, then modify pre_level
          if(iter > record_data(try, 1) % viter) then
!!$             pre_level=2
!!$             RESID = precision_arry(try)/RESID
!!$             DO i = 1, precision_sz
!!$                if(precision_arry(i) < RESID) exit
!!$             end DO
!!$             
!!$             pre_level = i
!!$             if(pre_level > precision_sz) pre_level = precision_sz
          end if




       end do
       call count_time(2,solver_comm,my_rank,21)       
       if(my_rank == 0) write(*,*) "precision lev", try, precision_arry(try)
       call count_time(3,solver_comm,my_rank,21)       




    end do

#ifdef PRINT_INFO
   call matrix_counting(LEVEL_NUM,SOLVER_COMM,my_rank)
   call count_time(4,solver_comm,my_rank,1)
#endif
     
   call MPI_BARRIER  (SOLVER_COMM, ierr)

   !C differentiate divergence cases from non-convergent cases
   if(conv_check_fl) then
      if(.not.CONV_FL) then
         call cal_relative_res_norm(RESID, X, B, WS, WR, WSIZE)
         if(RESID > 1.0) then
            iter = -1
         end if
      end if
   end if



   if(SETUP_FL) then
      call clear_matrix_ssi_amg(LEVEL_NUM)
      DEALLOCATE (R)
      DEALLOCATE (WR)
      DEALLOCATE (WS)
      DEALLOCATE (Temp)
      initialize = .true.
   end if

#ifdef PSPASES
   CALL PSPACEC(pspcommf,0)
#endif
     
 END SUBROUTINE TUNING_EVAL
   
   
END MODULE  solver_tuning_eval
