#include "my_cuda.h"
#include "my_mpi_cuda.h"
#include "v_node.h"

#include <iostream>
#include <new>

using std::bad_alloc;
using std::cout;
using std::endl;

void smoother_setup_default(v_node&);
void smoother_kernel_default(v_node&, double);
void smoother_dummy_default(v_node&);

v_node::v_node()
{
	level = -1;

	n  = 0;
	np = 0;

	x = 0;
	b = 0;

	dn  = 0;
	dnp = 0;

	dx = 0;
	db = 0;
	de = 0;

	b_n   = 0;
	b_cnt = 0;
	b_ind = 0;
	order = 0;

	comm = 0;

	smoother_setup       = smoother_setup_default;
	smoother_kernel_init = smoother_dummy_default;
	smoother_kernel      = smoother_kernel_default;
	smoother_finalize    = smoother_dummy_default;

	scl_vec   = 0;
	d_scl_vec = 0;
}


v_node::~v_node()
{
	if( v_node_delete() ){
		cout << "exit in v_node destructor." << endl;
	}
}


int v_node::v_node_new(int n, int np)
{
	this->n     = n;
	this->np    = np;

	try{
		x = new double[np];
		b = new double[np];
	}
	catch(bad_alloc)
	{
		cout << "v_node bad alloc expection : " << n << endl;
		return 1;
	}
	catch(...)
	{
		cout << "v_node new expection." << endl;
		return 1;
	}
	return 0;
}


int v_node::v_node_delete()
{
	try{
		if(x) delete[] x;
		if(b) delete[] b;

		if(dx) my_cudaFree(dx);
		if(db) my_cudaFree(db);
		if(de) my_cudaFree(de);

		if(b_cnt) delete[] b_cnt;
		if(b_ind) delete[] b_ind;
		if(order) delete[] order;
		
		if(scl_vec) delete[] scl_vec;
		if(d_scl_vec) my_cudaFree(d_scl_vec);

		level = -1;

		n  = 0;
		np = 0;

		x = 0;
		b = 0;

		dn  = 0;
		dnp = 0;

		dx = 0;
		db = 0;
		de = 0;

		b_n   = 0;
		b_cnt = 0;
		b_ind = 0;
		order = 0;

		comm = 0;

		scl_vec   = 0;
		d_scl_vec = 0; 
	}
	catch(...)
	{
		cout << "v_node delete expection." << endl;
		return 1;
	}
	return 0;
}

int v_node::smoother(int iter, double omega)
{
	int i;

	this->smoother_kernel_init(*this);

	for(i=0; i<iter; i++){
		this->smoother_kernel(*this, omega);
		comm_margin<COMM_MARGIN_EQL>(this->dx, &this->area_outer, &this->area_inner);
	}
	return 0;
}


void smoother_setup_default(v_node&)
{
	cout << "please set smoother_setup function." << endl;
}

void smoother_kernel_default(v_node&, double)
{
	cout << "please set smoother_kernel function." << endl;
}


void smoother_dummy_default(v_node&)
{
}
