
#ifndef ZERO_ORIGIN
#define ZERO_ORIGIN 0
#endif
  
  
MODULE data_creation_AMGCG_NB
CONTAINS
  
  SUBROUTINE data_creation_unsym_ssi_amg_NB(N, NP, NB, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU,  &
       &                 NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,  &
       &                 k_vec, NumK, NOD_IMPORT_SIZE, NOD_EXPORT_SIZE, SOLVER_COMM, LEVEL_NUM, WSIZE)
    USE data_structure_for_AMG
    USE amg_parameter, only: INDPDT_AGGRE
    USE count_time_mod
    IMPLICIT NONE
    include 'mpif.h'


    !C my_rank,and PRECOND are set before calling data_creation_AMGCG_NB()
    
    INTEGER(kind=kint), intent(out)  :: WSIZE, LEVEL_NUM
    INTEGER(kind=kint), intent(in)   :: N, NP, NB, NPL, NPU, NOD_IMPORT_SIZE, NOD_EXPORT_SIZE
    INTEGER(kind=kint), intent(in)   :: NumK
    INTEGER(kind=kint), INTENT(in) :: NEIBPETOT
    INTEGER(kind=kint), INTENT(in) :: SOLVER_COMM


#ifdef C_INTERFACE_FOR_SSI

    !C NP
    REAL   (kind=kreal) :: D(NB,NB,NP)
    !C NPU
    REAL   (kind=kreal) :: AU(NB,NB,NPU)
    !C NPL
    REAL   (kind=kreal) :: AL(NB,NB,NPL)
    REAL   (kind=kreal) :: k_vec(NB*NP, NumK)
    
    !C 0:NP, 0:NP
    INTEGER(kind=kint ) :: INU(0:N), INL(0:N)
    !C NPU, NPL
    INTEGER(kind=kint ) :: IAU(NPU), IAL(NPL)
    
    !C-     NEIBPE      (NEIBPETOT) 
    !C-     STACK_IMPORT(0:NEIBPETOT)
    !C-     NOD_IMPORT  (:)
    !C-     STACK_EXPORT(0:NEIBPETOT)
    !C-     NOD_EXPORT  (:)
    INTEGER(kind=kint ) :: NEIBPE      (NEIBPETOT) 
    INTEGER(kind=kint ) :: STACK_IMPORT(0:NEIBPETOT)
    INTEGER(kind=kint ) :: NOD_IMPORT  (NB*NOD_IMPORT_SIZE)
    INTEGER(kind=kint ) :: STACK_EXPORT(0:NEIBPETOT)
    INTEGER(kind=kint ) :: NOD_EXPORT  (NB*NOD_EXPORT_SIZE)

#else

    !C NP
    REAL   (kind=kreal) :: D(:,:,:)
    !C NPU
    REAL   (kind=kreal) :: AU(:,:,:)
    !C NPL
    REAL   (kind=kreal) :: AL(:,:,:)
    REAL   (kind=kreal) :: k_vec(:,:)
    INTEGER(kind=kint ) :: INU(0:), INL(0:)
    INTEGER(kind=kint ) :: IAU(:), IAL(:)
    INTEGER(kind=kint ) :: NEIBPE      (:) 
    INTEGER(kind=kint ) :: STACK_IMPORT(0:)
    INTEGER(kind=kint ) :: NOD_IMPORT  (:)
    INTEGER(kind=kint ) :: STACK_EXPORT(0:)
    INTEGER(kind=kint ) :: NOD_EXPORT  (:)
    !C-     NEIBPE      (NEIBPETOT) 
    !C-     STACK_IMPORT(0:NEIBPETOT)
    !C-     NOD_IMPORT  (:)
    !C-     STACK_EXPORT(0:NEIBPETOT)
    !C-     NOD_EXPORT  (:)
#endif
    
    call  data_creation_NB_ssi_wrapper(N, NP, NB, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU,  &
       &                 NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,  &
       &                 k_vec, NumK, NOD_IMPORT_SIZE, NOD_EXPORT_SIZE, SOLVER_COMM, LEVEL_NUM, WSIZE)                                     

  END SUBROUTINE data_creation_unsym_ssi_amg_NB

  !C this subroutine is just for data_creation_unsym_ssi_amg_NB..
  !C this sub copy k_vec to k_or_q_vec in problem_matrix module.
  !C If k_or_q_vec is already generated by other routine, this becomes redundunt calculation
  SUBROUTINE data_creation_NB_ssi_wrapper(N, NP, NB, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU,  &
       &                 NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, NOD_EXPORT,  &
       &                 k_vec, NumK, NOD_IMPORT_SIZE, NOD_EXPORT_SIZE, SOLVER_COMM, LEVEL_NUM, WSIZE)
    USE data_structure_for_AMG
    USE amg_parameter, only: INDPDT_AGGRE
    USE problem_matrix, only: k_or_q_vec
    IMPLICIT NONE
    include 'mpif.h'

    !C my_rank,and PRECOND are set before calling data_creation_AMGCG_NB()
    
    INTEGER(kind=kint), intent(out)  :: WSIZE, LEVEL_NUM
    INTEGER(kind=kint), intent(in)   :: N, NP, NB, NPL, NPU, NOD_IMPORT_SIZE, NOD_EXPORT_SIZE
    INTEGER(kind=kint), intent(in)   :: NumK
    INTEGER(kind=kint), INTENT(in) :: NEIBPETOT
    INTEGER(kind=kint), INTENT(in) :: SOLVER_COMM


    !C NP
    REAL   (kind=kreal) :: D(:,:,:)
    !C NPU
    REAL   (kind=kreal) :: AU(:,:,:)
    !C NPL
    REAL   (kind=kreal) :: AL(:,:,:)
    REAL   (kind=kreal) :: k_vec(:,:)
    INTEGER(kind=kint ) :: INU(0:), INL(0:)
    INTEGER(kind=kint ) :: IAU(:), IAL(:)
    INTEGER(kind=kint ) :: NEIBPE      (:) 
    INTEGER(kind=kint ) :: STACK_IMPORT(0:)
    INTEGER(kind=kint ) :: NOD_IMPORT  (:)
    INTEGER(kind=kint ) :: STACK_EXPORT(0:)
    INTEGER(kind=kint ) :: NOD_EXPORT  (:)
    !C-     NEIBPE      (NEIBPETOT) 
    !C-     STACK_IMPORT(0:NEIBPETOT)
    !C-     NOD_IMPORT  (:)
    !C-     STACK_EXPORT(0:NEIBPETOT)
    !C-     NOD_EXPORT  (:)
    
    REAL   (kind=kreal),pointer :: k_or_q_vec_f90(:,:)
    INTEGER(kind=kint)  :: my_rank
    CHARACTER(len=20)   :: PRECOND
    INTEGER(kind=kint) :: ierr
    INTEGER(kind=kint) :: i,j,k,l,m
    
    if(NumK > 0) then
       if(associated(k_or_q_vec)) deallocate(k_or_q_vec)
       allocate(k_or_q_vec(NB*N,NumK))
       do j = 1, N*NB
          do i = 1, NumK
             k_or_q_vec(j,i) = k_vec(j,i)
          end do
       end do
    end if
       
    CALL MPI_COMM_RANK (SOLVER_COMM, my_rank, ierr )    

    if(INDPDT_AGGRE) then
       PRECOND = "IND1"
    else
       PRECOND = "SHA2"
    endif
    

    call  data_creation_NB                                                &
       &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, &
       &                  LEVEL_NUM,my_rank, NEIBPETOT, NEIBPE,           &
       &                  STACK_IMPORT, NOD_IMPORT,                       &
       &                  STACK_EXPORT, NOD_EXPORT,                       &
       &                  SOLVER_COMM, PRECOND, NB, k_or_q_vec, NumK)     

    call max_comm_table_sz(WSIZE, HIERARCHICAL_DATA, LEVEL_NUM)
    
  END SUBROUTINE data_creation_NB_ssi_wrapper


  SUBROUTINE  data_creation_NB                                            &
       &                 (N, NP, NPL, NPU, D, AL, INL, IAL, AU, INU, IAU, &
       &                  LEVEL_NUM,my_rank, NEIBPETOT, NEIBPE,           &
       &                  STACK_IMPORT, NOD_IMPORT,                       &
       &                  STACK_EXPORT, NOD_EXPORT,                       &
       &                  SOLVER_COMM, PRECOND, NB, k_or_q_vec, NK)     
    USE isort
    USE solver_SR_NB
    USE comm_subs_NB
    USE count_time_mod
    USE data_structure_for_AMG_NB
    USE aggregate_mod
    USE output_data, only: output_aggre_data, aggregate_transfer
    USE amg_parameter, only: theta, dump_jacobi_weight, MAX_LEVEL_SIZE, MIN_NODE_SIZE,    &
         & theta_reduction_rate, PROCESS_AGGRE, process_aggre_threshold, comm_opt_for_CGA,&
         & NOT_USE_AGGRE_TRANSFER, setup_parameter_set_at_each_lev
    USE int_vec_module, &
         & only: int_vec, int_vec_initialize, check_and_extend, deallocate_int_vec

#ifdef PARMETIS
    USE process_aggregation
#endif

    IMPLICIT NONE
    INCLUDE 'mpif.h'
    INTEGER(kind=kint ), INTENT(in) :: N,NP,NPL,NPU,NB
    INTEGER(kind=kint ), INTENT(out):: LEVEL_NUM
    INTEGER(kind=kint )             :: NPROCS
    
    INTEGER(kind=kint ), INTENT(in) :: NK
!!$    real(kind=kreal), INTENT(inout), TARGET :: k_or_q_vec(:,:)
    real(kind=kreal), pointer :: k_or_q_vec(:,:)


!!$    REAL   (kind=kreal), INTENT(inout),TARGET::  D(NB,NB,NP )
    REAL   (kind=kreal), INTENT(inout),TARGET::  D(:,:,: )
    REAL   (kind=kreal), ALLOCATABLE :: tilde_D(:,:,:),tilde_DLU(:,:,:)
    REAL   (kind=kreal), POINTER :: newALU(:,:,:)

!!$    REAL   (kind=kreal), INTENT(inout),TARGET::  AU(NB,NB,NPU)
    REAL   (kind=kreal), INTENT(inout),TARGET::  AU(:,:,:)

!!$    REAL   (kind=kreal), INTENT(inout),TARGET::  AL(NB,NB,NPL)
    REAL   (kind=kreal), INTENT(inout),TARGET::  AL(:,:,:)

!!$    INTEGER(kind=kint ), INTENT(in),   TARGET::  INU(0:NP), INL(0:NP)
    INTEGER(kind=kint ), INTENT(in),   TARGET::  INU(0:), INL(0:)

!!$    INTEGER(kind=kint ), INTENT(in),   TARGET::  IAL(NPL), IAU(NPU)
    INTEGER(kind=kint ), INTENT(in),   TARGET::  IAL(:), IAU(:)
    INTEGER(kind=kint ), ALLOCATABLE :: NI(:), PNI(:)


    INTEGER(kind=kint ), INTENT(in)          :: my_rank
    INTEGER(kind=kint ), INTENT(in)          :: NEIBPETOT
    INTEGER(kind=kint ), INTENT(in)          :: SOLVER_COMM
    CHARACTER(len=20)  , INTENT(in)          :: PRECOND

    INTEGER(kind=kint ), TARGET :: NEIBPE      (:) 
    INTEGER(kind=kint ), TARGET :: STACK_IMPORT(0:)
    INTEGER(kind=kint ), TARGET :: NOD_IMPORT  (:)
    INTEGER(kind=kint ), TARGET :: STACK_EXPORT(0:)
    INTEGER(kind=kint ), TARGET :: NOD_EXPORT  (:)
    INTEGER(kind=kint )          :: ierr

    INTEGER(kind=kint ), ALLOCATABLE :: sorted_fNeibPE(:)
    INTEGER(kind=kint )          :: finerNEIBPETOT
    INTEGER(kind=kint ), POINTER :: finerNEIBPE      (:) 
    INTEGER(kind=kint ), POINTER :: finerSTACK_IMPORT(:)
    INTEGER(kind=kint ), POINTER :: finerNOD_IMPORT  (:)
    INTEGER(kind=kint ), POINTER :: finerSTACK_EXPORT(:)
    INTEGER(kind=kint ), POINTER :: finerNOD_EXPORT  (:)

    INTEGER(kind=kint ), ALLOCATABLE :: node_index(:)

    INTEGER(kind=kint ) :: coarser_level_size, level, finer_level_size, count
    INTEGER(kind=kint ) :: finer_level_NP, i, j, k, agmtd
    INTEGER(kind=kint ) :: finer_level_NPU, finer_level_NPL, finer_level_NB

    INTEGER(kind=kint ) :: PE_num_size
    INTEGER(kind=kint ), POINTER :: PE_nums(:)
    INTEGER(kind=kint ), POINTER :: nodes_for_each_PE(:)
    INTEGER(kind=kint ), POINTER :: in_nodes_for_each_PE(:)

    INTEGER(kind=kint ), POINTER :: PE_list(:,:)
    INTEGER(kind=kint )              :: PE_list_size 
    REAL   (kind=kreal) :: theta_copy, reduction_rate

    REAL   (kind=kreal) :: val
    REAL   (kind=kreal), ALLOCATABLE :: WW(:,:), tmp_bl(:,:)

!!$    REAL   (kind=kreal) :: theta = 0.05

    INTEGER(kind=kint ), POINTER :: in_aggregates_result(:)
    INTEGER(kind=kint ), POINTER :: aggregates_result(:)
    INTEGER(kind=kint )          :: in_aggregates_result_size

    REAL   (kind=kreal),  POINTER :: Temp_P(:,:)
    INTEGER(kind=kint ),  POINTER :: Temp_P_row_size(:)
    INTEGER(kind=kint ),  POINTER :: Temp_PI(:,:)

    REAL   (kind=kreal), ALLOCATABLE :: WS(:), WR(:)
    INTEGER(kind=kint ), ALLOCATABLE :: WSI(:), WRI(:)

    INTEGER(kind=kint ) :: l,m
    LOGICAL :: finish_flag, global_finish_flag, local_flag, global_flag
    LOGICAL :: border_flag, SYMMETRIC_FLAG
    INTEGER(kind=kint ) :: global_CN, nfepe_size
    REAL   (kind=kreal) :: R11,R12,R13,R21,R22,R23,R31,R32,R33

    !C - begin: aggregate_table_array
    INTEGER(kind=kint) :: aggregate_table_size=12000
    !C two arguments: communication table,each PE
    TYPE(INT_VEC), POINTER :: aggregate_table_array(:)
    !C argument: the number of aggregates
    INTEGER(kind=kint), POINTER :: aggregate_number_in_table(:)
    !C - end: aggregate_table_array
    INTEGER(kind=kint), ALLOCATABLE :: GIN_aggregate(:)

    !C - begin: hash_table
    INTEGER(kind=kint)             :: hash_table_size
    PARAMETER(hash_table_size=280001)
    INTEGER(kind=kint),ALLOCATABLE :: global_local_hash_table(:,:)
    !C - end: hash_table
    INTEGER(kind=kint)             :: local_aggre_size,col
    TYPE(INTER_LEVEL_OPERATOR),POINTER :: P,R
    INTEGER(kind=kint),ALLOCATABLE :: Temp_R_IN(:)

    !C variables for lower_ext_matrix_crtn
    INTEGER(kind=kint ) :: EX_NPL
    INTEGER(kind=kint ), POINTER :: EX_INL (:), EX_IAL (:)
    REAL   (kind=kreal), POINTER :: EX_AL  (:,:,:)

    integer(kind=kint) :: coarser_lev_parallelism, start_nd_num, sum_nd_num
    integer(kind=kint) :: coarser_lev_N_bf_CGA
    logical :: PRO_AGGRE_OPT

    INTEGER(kind=kint), POINTER :: new_GIN_aggregate(:)

    
    CHARACTER(len= 4 )  :: penum
    
    integer(kind=kint) :: partner
    logical :: sender_fl, receiver_fl

#ifdef MATRIX_OUT
    CHARACTER(len=6) :: levnum,fn
#endif

    SYMMETRIC_FLAG = .false.



    !C-- theta means the critierion for strong connection.
    !C-- if theta=0 then all nonzero elements are recognized as
    !C-- strong connection.
    theta_copy = theta
    reduction_rate = theta_reduction_rate




#ifdef MEASURING
    call count_time(0,SOLVER_COMM,my_rank,19)
    call count_time(1,SOLVER_COMM,my_rank,19)
#endif


    CALL MPI_COMM_SIZE(SOLVER_COMM, NPROCS, IERR)
    
    if(NEIBPETOT > 0) then
       ALLOCATE(WW(NP*NB,NB))
       !$omp parallel do private(j,k)
       DO i = 1, N
          DO j = 1, NB
             DO k = 1, NB
                WW((i-1)*NB+k,j) = D(k,j,i)
             END DO
          END DO
       END DO
       !$omp end parallel do
       
       j = STACK_EXPORT(NEIBPETOT)
       k = STACK_IMPORT(NEIBPETOT)
       ALLOCATE ( WS(j * NB) )
       ALLOCATE ( WR(k * NB) )

       DO i = 1, NB
          CALL SOLVER_SEND_RECV_NB                                                        &
               &     ( j, k, NP,NEIBPETOT, NEIBPE,STACK_IMPORT, NOD_IMPORT, STACK_EXPORT, &
               &       NOD_EXPORT, WS, WR, WW(:,i), SOLVER_COMM, my_rank, NB )
       END DO
       DEALLOCATE(WS)
       DEALLOCATE(WR)

       !$omp parallel do private(j,k)
       DO i = N+1, NP
          DO j = 1, NB
             DO k = 1, NB
                D(k,j,i) = WW(NB*(i-1)+k,j)
             END DO
          END DO
       ENDDO
       !$omp end parallel do
       
       DEALLOCATE(WW)

    end if



    agmtd=0
    IF (       ((PRECOND(1:1).EQ.'I').OR.(PRECOND(1:1).EQ.'i')) .AND.       &
         &     ((PRECOND(2:2).EQ.'N').OR.(PRECOND(2:2).EQ.'n')) .AND.       &
         &     ((PRECOND(3:3).EQ.'D').OR.(PRECOND(3:3).EQ.'d')) .AND.       &
         &     (PRECOND(4:4).EQ.'1') ) THEN
       agmtd=1
    ELSE IF (  ((PRECOND(1:1).EQ.'I').OR.(PRECOND(1:1).EQ.'i')) .AND.       &
         &     ((PRECOND(2:2).EQ.'N').OR.(PRECOND(2:2).EQ.'n')) .AND.       &
         &     ((PRECOND(3:3).EQ.'D').OR.(PRECOND(3:3).EQ.'d')) .AND.       &
         &     (PRECOND(4:4).EQ.'2') ) THEN
       agmtd=2
    ELSE IF (  ((PRECOND(1:1).EQ.'S').OR.(PRECOND(1:1).EQ.'s')) .AND.       &
         &     ((PRECOND(2:2).EQ.'H').OR.(PRECOND(2:2).EQ.'h')) .AND.       &
         &     ((PRECOND(3:3).EQ.'A').OR.(PRECOND(3:3).EQ.'a')) .AND.       &
         &     (PRECOND(4:4).EQ.'1') ) THEN
       agmtd=3
    ELSE IF (  ((PRECOND(1:1).EQ.'S').OR.(PRECOND(1:1).EQ.'s')) .AND.       &
         &     ((PRECOND(2:2).EQ.'H').OR.(PRECOND(2:2).EQ.'h')) .AND.       &
         &     ((PRECOND(3:3).EQ.'A').OR.(PRECOND(3:3).EQ.'a')) .AND.       &
         &     (PRECOND(4:4).EQ.'2') ) THEN
       agmtd=4
    ELSE IF (  ((PRECOND(1:1).EQ.'S').OR.(PRECOND(1:1).EQ.'s')) .AND.       &
         &     ((PRECOND(2:2).EQ.'H').OR.(PRECOND(2:2).EQ.'h')) .AND.       &
         &     ((PRECOND(3:3).EQ.'A').OR.(PRECOND(3:3).EQ.'a')) .AND.       &
         &     (PRECOND(4:4).EQ.'3') ) THEN
       agmtd=5
    ELSE IF (  ((PRECOND(1:1).EQ.'M').OR.(PRECOND(1:1).EQ.'m')) .AND.       &
         &     ((PRECOND(2:2).EQ.'I').OR.(PRECOND(2:2).EQ.'i')) .AND.       &
         &     ((PRECOND(3:3).EQ.'X').OR.(PRECOND(3:3).EQ.'x')) ) THEN
       agmtd=6
    ENDIF
    IF(agmtd == 0)THEN 
       IF(my_rank == 0) THEN
          WRITE(*,*) "Select correct aggregate strategy"
          WRITE(*,*) "Aggregation strategies are ind1,ind2,sha1,sha2,sha3,and mix."
       END IF
       STOP 
    END IF

    !C-- switch aggregation strategy
    !C single case takes up independent aggregation
    if(NPROCS == 1) agmtd = 1
    
    IF(agmtd>=3 .and. associated(k_or_q_vec)) then
       IF(my_rank == 0)then
          write(*,*) "PARAMETER SETTING ERROR!!:"
          write(*,*) "Aggregation strategy must be independent aggregation, &
               & when you input the kernel vectors."
       END IF
       call MPI_BARRIER(SOLVER_COMM, ierr)
       STOP
    END IF
    

    HIERARCHICAL_DATA_NB(1) % NB  = NB
    HIERARCHICAL_DATA_NB(1) % N   = N
    HIERARCHICAL_DATA_NB(1) % NP  = NP
    HIERARCHICAL_DATA_NB(1) % NPL = NPL
    HIERARCHICAL_DATA_NB(1) % NPU = NPU
    HIERARCHICAL_DATA_NB(1) % D   => D
    HIERARCHICAL_DATA_NB(1) % INL => INL 
    HIERARCHICAL_DATA_NB(1) % INU => INU 
    HIERARCHICAL_DATA_NB(1) % IAL => IAL
    HIERARCHICAL_DATA_NB(1) % IAU => IAU
    HIERARCHICAL_DATA_NB(1) % AL  => AL
    HIERARCHICAL_DATA_NB(1) % AU  => AU

    HIERARCHICAL_DATA_NB(1) % COMM_TABLE % NEIBPETOT    =  NEIBPETOT
    HIERARCHICAL_DATA_NB(1) % COMM_TABLE % NEIBPE       => NEIBPE
    HIERARCHICAL_DATA_NB(1) % COMM_TABLE % STACK_IMPORT => STACK_IMPORT
    HIERARCHICAL_DATA_NB(1) % COMM_TABLE % STACK_EXPORT => STACK_EXPORT
    HIERARCHICAL_DATA_NB(1) % COMM_TABLE % NOD_IMPORT   => NOD_IMPORT
    HIERARCHICAL_DATA_NB(1) % COMM_TABLE % NOD_EXPORT   => NOD_EXPORT

    HIERARCHICAL_DATA_NB(1) % NK  = NK
    
    if(associated(k_or_q_vec)) HIERARCHICAL_DATA_NB(1) % k_or_q_vec => k_or_q_vec

    RATE_OF_SPACE=19.

#ifdef PRINT_REZ          
    if(my_rank == 0) then
       write(*,*) "Is kernel vectors registered?: ", associated(k_or_q_vec)
       write(*,*) "Number of kernel vectors: ", NK
    end if
#endif


    !C initial dof is printed
    i = 0
    CALL MPI_ALLREDUCE(N, i, 1, MPI_INTEGER, MPI_SUM, SOLVER_COMM, ierr)
#ifdef PRINT_REZ          
    if(my_rank == 0) write(*,*) i 
#endif

    
    !C PE_list(*,1): PE number
    !C PE_list(*,2): NEIBPE number|
    !C PE_list(*,3): PE_num number|these two has 0 if no correspondings.
    ALLOCATE(PE_list(NPROCS,3))


    DO level=2,MAX_LEVEL_SIZE
       call  setup_parameter_set_at_each_lev(level, theta_copy)



#ifdef MEASURING
       call count_time(0,SOLVER_COMM,my_rank,30+level-1)
       call count_time(0,SOLVER_COMM,my_rank,40+level)
       call count_time(0,SOLVER_COMM,my_rank,50+level-1)

       call count_time(1,SOLVER_COMM,my_rank,30+level-1)
#endif


       finer_level_NB   = HIERARCHICAL_DATA_NB(level-1) % NB
       finer_level_size = HIERARCHICAL_DATA_NB(level-1) % N
       finer_level_NP   = HIERARCHICAL_DATA_NB(level-1) % NP
       finer_level_NPL  = HIERARCHICAL_DATA_NB(level-1) % NPL
       finer_level_NPU  = HIERARCHICAL_DATA_NB(level-1) % NPU

       finerNEIBPETOT   =  HIERARCHICAL_DATA_NB(level-1) % COMM_TABLE % NEIBPETOT
       finerNEIBPE      => HIERARCHICAL_DATA_NB(level-1) % COMM_TABLE % NEIBPE 
       finerSTACK_IMPORT=> HIERARCHICAL_DATA_NB(level-1) % COMM_TABLE % STACK_IMPORT
       finerSTACK_EXPORT=> HIERARCHICAL_DATA_NB(level-1) % COMM_TABLE % STACK_EXPORT
       finerNOD_IMPORT  => HIERARCHICAL_DATA_NB(level-1) % COMM_TABLE % NOD_IMPORT
       finerNOD_EXPORT  => HIERARCHICAL_DATA_NB(level-1) % COMM_TABLE % NOD_EXPORT

       ALLOCATE(sorted_fNEIBPE(finerNEIBPETOT))
       IF(NPROCS > 1 .and. finerNEIBPETOT > 0) THEN
          CALL QSORTI(sorted_fNEIBPE, finerNEIBPETOT, finerNEIBPE)
       END IF


#ifdef DEBUG       
       allocate(Temp_R_IN(finer_level_size))
       Temp_R_IN=0

       if(finerNEIBPETOT > 0) then
          do i = 1, finerSTACK_EXPORT(finerNEIBPETOT)
             Temp_R_IN(finerNOD_EXPORT(i)+ZERO_ORIGIN) = Temp_R_IN(finerNOD_EXPORT(i)+ZERO_ORIGIN)+1       
          end do
          do i = 1, finer_level_size
             if(Temp_R_IN(i) >= NPROCS) then
                write(*,*) Temp_R_IN(i), i, "Error in export table!!"
                stop
             end if
          end do
       end if
       deallocate(Temp_R_IN)
#endif       

       CALL DEBUG_BARRIER_PRINT("## repeated_nds_tbl_crtn ##", SOLVER_COMM, my_rank)
       if(NPROCS > 1) then

          CALL repeated_nds_tbl_crtn(level, sorted_fNEIBPE, SOLVER_COMM, &
               & my_rank, PE_num_size,PE_nums,nodes_for_each_PE,         &
               & in_nodes_for_each_PE,NPROCS)
          
       else
          
          PE_num_size = 0 
          allocate(PE_nums(0))
          !C nodes_for_each_PE, in_nodes_for_each_PE are not associated
          
       END IF


       CALL DEBUG_BARRIER_PRINT("## lower_ext_matrix_crtn ##", SOLVER_COMM, my_rank)
       
       if(.not.SYMMETRIC_FLAG) then
          
          CALL lower_ext_matrix_crtn(EX_NPL, EX_INL, EX_IAL, EX_AL, level, my_rank,  &
               & SOLVER_COMM, NPROCS)
       else
          
          EX_NPL = 0
          
       end if





       CALL DEBUG_BARRIER_PRINT("## neighbors             ##", SOLVER_COMM, my_rank)

       ALLOCATE(NI        (finer_level_NPL+finer_level_NPU+EX_NPL))
       ALLOCATE(PNI       (0:2*finer_level_NP)             )
       ALLOCATE(node_index(finer_level_NP)                 )
       ALLOCATE(tilde_D   (finer_level_NB,finer_level_NB,finer_level_NP)           )


       CALL count_time(1,solver_comm,my_rank,5)                     
       if(SYMMETRIC_FLAG) then
          CALL neighbors (PNI, NI, theta_copy, level, node_index, tilde_D)
       else
          CALL neighbors_unsym(EX_INL, EX_IAL, EX_AL, PNI, NI, theta_copy, level, node_index, tilde_D, &
               & SOLVER_COMM, my_rank)
       end if

       CALL count_time(2,solver_comm,my_rank,5)                     

       theta_copy = theta_copy * reduction_rate
       

       CALL DEBUG_BARRIER_PRINT("## exchange node_index   ##", SOLVER_COMM, my_rank)

       
       if(NPROCS > 1) then
          ALLOCATE ( WSI(finerSTACK_EXPORT(finerNEIBPETOT)) )
          ALLOCATE ( WRI(finerSTACK_IMPORT(finerNEIBPETOT)) )

          CALL SOLVER_SEND_RECV2I                                             &
               &   ( finer_level_NP, finerSTACK_EXPORT(finerNEIBPETOT),       &
               &     finerSTACK_IMPORT(finerNEIBPETOT), finerNEIBPETOT,       &
               &     finerNEIBPE, finerSTACK_IMPORT, finerNOD_IMPORT,         &
               &     finerSTACK_EXPORT, finerNOD_EXPORT, WSI, WRI, node_index,&
               &     SOLVER_COMM, my_rank)

          DEALLOCATE(WSI)
          DEALLOCATE(WRI)
       end if

       CALL DEBUG_BARRIER_PRINT("## exchange tilde_D      ##", SOLVER_COMM, my_rank)

       ALLOCATE ( WW(finer_level_NB*finer_level_NP,finer_level_NB) )       

       
       !C-- exchanging filtered DIAGONAL components
       WW= 0.d0
       DO i= 1, finer_level_size
          DO j = 1, finer_level_NB
             DO k = 1, finer_level_NB
                WW(finer_level_NB*(i-1)+k,j) = tilde_D(k,j,i)
             END DO
          END DO
       ENDDO

       !C Identity matrix is set at neighboring halo unknowns, 
       !C because some halo elements are not communicated with the following routine.
       !C This unknown willnot be used in this level calculation.
       DO i= finer_level_size+1, finer_level_NP
          DO j = 1, finer_level_NB
             WW(finer_level_NB*(i-1)+j,j) = 1.0
          END DO
       ENDDO


       if(NPROCS > 1) then
          j = finerSTACK_EXPORT(finerNEIBPETOT)
          k = finerSTACK_IMPORT(finerNEIBPETOT)
          ALLOCATE ( WS(j * finer_level_NB) )
          ALLOCATE ( WR(k * finer_level_NB) )

          DO i = 1, finer_level_NB
             CALL SOLVER_SEND_RECV_NB                                               &
                  &     ( j, k, finer_level_NP,finerNEIBPETOT, finerNEIBPE,         &
                  &       finerSTACK_IMPORT, finerNOD_IMPORT, finerSTACK_EXPORT,    &
                  &       finerNOD_EXPORT, WS, WR, WW(:,i), SOLVER_COMM, my_rank, finer_level_NB )
          END DO
          DEALLOCATE(WS)
          DEALLOCATE(WR)

          DO i = finer_level_size+1, finer_level_NP
             DO j = 1, finer_level_NB
                DO k = 1, finer_level_NB
                   tilde_D(k,j,i) = WW(finer_level_NB*(i-1)+k,j)
                END DO
             END DO

          ENDDO

       end if
       
       DEALLOCATE(WW)

       ALLOCATE ( tilde_DLU(finer_level_NB,finer_level_NB+1,finer_level_NP) )
       DO i=1,finer_level_NP
          call ILU_NB(tilde_DLU(:,:,i),tilde_D(:,:,i),finer_level_NB)
       END DO
       DEALLOCATE(tilde_D)
       
       
       CALL count_time(1,solver_comm,my_rank,6)
       IF(agmtd==1 .or. agmtd==2) THEN

          CALL DEBUG_BARRIER_PRINT("## independent aggregate ##", SOLVER_COMM, my_rank)

          border_flag = .true.
          if(agmtd == 1) border_flag = .false.

          CALL indpdt_agrgt (finer_level_size, &
               & HIERARCHICAL_DATA_NB(level - 1) % INL, &
               & HIERARCHICAL_DATA_NB(level - 1) % INU, &
               & HIERARCHICAL_DATA_NB(level - 1) % IAL, &
               & HIERARCHICAL_DATA_NB(level - 1) % IAU, &
               & finerNEIBPETOT, &
               & finerSTACK_EXPORT, &
               & finerNOD_EXPORT, &
               & PNI,NI,level,node_index, in_aggregates_result,&
               & aggregates_result, in_aggregates_result_size, border_flag )
          
          CALL make_PE_lists (level,sorted_fNEIBPE,PE_num_size,PE_nums,PE_list_size, &
               &              PE_list, NPROCS, finerNEIBPE, finerNEIBPETOT)

          DEALLOCATE(sorted_fNeibPE)


          !C-- k_or_q_vec is only dealt by independent aggregation..
          if(associated(k_or_q_vec)) then          
             j = in_aggregates_result_size
             ALLOCATE(HIERARCHICAL_DATA_NB(level) % k_or_q_vec(j*NK,NK))
             call assign_kernel_vectors(&
                  & in_aggregates_result, aggregates_result, in_aggregates_result_size, &
                  & HIERARCHICAL_DATA_NB(level-1) % k_or_q_vec, finer_level_size, finer_level_NB,&
                  & HIERARCHICAL_DATA_NB(level)   % k_or_q_vec, NK)
          end if


#ifdef PARMETIS
          if(PROCESS_AGGRE) then
             call decide_parallelism(in_aggregates_result_size, coarser_lev_parallelism, &
                  & SOLVER_COMM, process_aggre_threshold)
             if(my_rank == 0) write(*,*) "coarser_lev_parallelism:", coarser_lev_parallelism

             call decide_send_recv_allocation(in_aggregates_result_size, finerNEIBPETOT, finerNEIBPE, &
                  & coarser_lev_parallelism, sender_fl, start_nd_num, partner, receiver_fl, &
                  & sum_nd_num, SOLVER_COMM, my_rank, NPROCS, &
                  & HIERARCHICAL_DATA_NB(level) % k_or_q_vec, NK)

             
             coarser_lev_N_bf_CGA = in_aggregates_result_size
    
          else
#endif
             call decide_sender_and_receiver(SOLVER_COMM, finer_level_size, finerNEIBPETOT,    &
                  & finerNEIBPE, finerSTACK_IMPORT, finerSTACK_EXPORT, sender_fl, receiver_fl, &
                  & partner, NPROCS, my_rank)
#ifdef PARMETIS
          end if
#endif



          ALLOCATE(aggregate_table_array(NPROCS))
          DO i = 1, NPROCS
             call int_vec_initialize(aggregate_table_array(i))
          END DO
          ALLOCATE(aggregate_number_in_table(NPROCS))
          ALLOCATE(GIN_aggregate(0:NPROCS))
#ifdef PARMETIS
          if(PROCESS_AGGRE) then
             if(comm_opt_for_CGA/=0) then
                PRO_AGGRE_OPT = .true.
             else
                PRO_AGGRE_OPT = .false.
             end if
             HIERARCHICAL_DATA_NB(level) % PRO_AGGRE_OPT = PRO_AGGRE_OPT
             
             call set_GIN_aggre_and_aggre_tbl(in_aggregates_result_size, sender_fl, start_nd_num, partner,  &
                  & receiver_fl, sum_nd_num, GIN_aggregate, NPROCS, aggregate_number_in_table, &
                  & aggregate_table_array, aggregate_table_size, PE_list, PE_list_size,&
                  & my_rank, SOLVER_COMM, level, PRO_AGGRE_OPT, .true.)
          else
#endif
          
             !C--initialize aggregate_table and set GIN_aggregate(:) for indpdt_agrgt
             !C j, k are dummy. We don't use them later, here.
             k = in_aggregates_result_size
             call cordinate_GIN_aggregate(sender_fl, receiver_fl, partner, j, &
                  & k, GIN_aggregate, NPROCS, my_rank, SOLVER_COMM, &
                  & aggregate_number_in_table, aggregate_table_array, aggregate_table_size, &
                  & PE_list, PE_list_size)
             
             !C--end:initialize aggregate_table and set GIN_aggregate(:) for indpdt_agrgt

#ifdef PARMETIS
          endif
#endif



       ELSE IF(agmtd==3 .or. agmtd==4 .or. agmtd==5) THEN

          CALL DEBUG_BARRIER_PRINT("## border_aggregate      ##", SOLVER_COMM, my_rank)

          CALL border_aggregate (&
                  & finer_level_NP, finer_level_size, &
                  & HIERARCHICAL_DATA_NB(level - 1) % IAL, &
                  & HIERARCHICAL_DATA_NB(level - 1) % IAU, &
                  & finerNEIBPETOT, finerNEIBPE, finerSTACK_IMPORT, finerNOD_IMPORT, &
                  & finerSTACK_EXPORT, finerNOD_EXPORT, &
                  & PNI, NI, level, node_index, sorted_fNEIBPE,      &
                  &                 my_rank, PE_num_size,PE_nums,nodes_for_each_PE,  &
                  &                 in_nodes_for_each_PE,PE_list_size,PE_list,NPROCS,&
                  &                 SOLVER_COMM)
          DEALLOCATE(sorted_fNeibPE)


          CALL DEBUG_BARRIER_PRINT("## aggregate             ##", SOLVER_COMM, my_rank)

          ALLOCATE(aggregate_table_array(NPROCS))
          DO i = 1, NPROCS
             call int_vec_initialize(aggregate_table_array(i))
          END DO
          
          ALLOCATE(aggregate_number_in_table(NPROCS))
          ALLOCATE(GIN_aggregate(0:NPROCS))

          CALL aggregate (&
               & finer_level_NP, finer_level_size, &
               & HIERARCHICAL_DATA_NB(level - 1) % INL, &
               & HIERARCHICAL_DATA_NB(level - 1) % INU, &
               & HIERARCHICAL_DATA_NB(level - 1) % IAL, &
               & HIERARCHICAL_DATA_NB(level - 1) % IAU, &
               & finerNEIBPETOT, finerNEIBPE, finerSTACK_IMPORT, finerNOD_IMPORT, &
               & finerSTACK_EXPORT, finerNOD_EXPORT, &
               & PNI, NI, level,node_index,in_aggregates_result,            &
               &          aggregates_result ,in_aggregates_result_size,my_rank,      &
               &          aggregate_table_size,aggregate_table_array,                &
               &          aggregate_number_in_table,SOLVER_COMM,NPROCS,GIN_aggregate,&
               &          PE_num_size,PE_nums,nodes_for_each_PE,in_nodes_for_each_PE,&
               &          PE_list_size,PE_list,agmtd-2)

       ELSE IF(agmtd==6) THEN
          local_flag=finer_level_size > SWITCH_SIZE
          CALL MPI_ALLREDUCE(local_flag,global_flag,1,MPI_LOGICAL,MPI_LAND,SOLVER_COMM,ierr)
          if(global_flag) then

             CALL DEBUG_BARRIER_PRINT("## independent aggregate ##", SOLVER_COMM, my_rank)

             CALL indpdt_agrgt (finer_level_size, &
               & HIERARCHICAL_DATA_NB(level - 1) % INL, &
               & HIERARCHICAL_DATA_NB(level - 1) % INU, &
               & HIERARCHICAL_DATA_NB(level - 1) % IAL, &
               & HIERARCHICAL_DATA_NB(level - 1) % IAU, &
               & finerNEIBPETOT, &
               & finerSTACK_EXPORT, &
               & finerNOD_EXPORT, &
               & PNI,NI,level,node_index, in_aggregates_result,&
                  & aggregates_result, in_aggregates_result_size, .false. )


             CALL make_PE_lists (level,sorted_fNEIBPE,PE_num_size,PE_nums,PE_list_size, &
                  &              PE_list, NPROCS, finerNEIBPE, finerNEIBPETOT)

             DEALLOCATE(sorted_fNeibPE)
             ALLOCATE(aggregate_table_array(NPROCS))
             DO i = 1, NPROCS
                call int_vec_initialize(aggregate_table_array(i))
             END DO
             
             ALLOCATE(aggregate_number_in_table(NPROCS))
             ALLOCATE(GIN_aggregate(0:NPROCS))

             !C--initialize aggregate_table and set GIN_aggregate(:) for indpdt_agrgt
             aggregate_number_in_table=0
             CALL MPI_ALLGATHER(in_aggregates_result_size, 1, MPI_INTEGER, &
                  & GIN_aggregate(1),1,MPI_INTEGER,SOLVER_COMM,ierr)
             GIN_aggregate(0)=0
             DO i=2,NPROCS
                GIN_aggregate(i)=GIN_aggregate(i-1)+GIN_aggregate(i)
             END DO
             !C--end:initialize aggregate_table and set GIN_aggregate(:) for indpdt_agrgt

          else

             CALL DEBUG_BARRIER_PRINT("## border_aggregate      ##", SOLVER_COMM, my_rank)

             CALL border_aggregate (& 
                  & finer_level_NP, finer_level_size, &
                  & HIERARCHICAL_DATA_NB(level - 1) % IAL, &
                  & HIERARCHICAL_DATA_NB(level - 1) % IAU, &
                  & finerNEIBPETOT, finerNEIBPE, finerSTACK_IMPORT, finerNOD_IMPORT, &
                  & finerSTACK_EXPORT, finerNOD_EXPORT, &
                  & PNI, NI, level, node_index, sorted_fNEIBPE,      &
                  &                 my_rank, PE_num_size,PE_nums,nodes_for_each_PE,  &
                  &                 in_nodes_for_each_PE,PE_list_size,PE_list,NPROCS,&
                  &                  SOLVER_COMM)
             DEALLOCATE(sorted_fNeibPE)

             CALL DEBUG_BARRIER_PRINT("## aggregate             ##", SOLVER_COMM, my_rank)

             ALLOCATE(aggregate_table_array(NPROCS))
             DO i = 1, NPROCS
                call int_vec_initialize(aggregate_table_array(i))
             END DO

             ALLOCATE(aggregate_number_in_table(NPROCS))
             ALLOCATE(GIN_aggregate(0:NPROCS))

             CALL aggregate (&
               & finer_level_NP, finer_level_size, &
               & HIERARCHICAL_DATA_NB(level - 1) % INL, &
               & HIERARCHICAL_DATA_NB(level - 1) % INU, &
               & HIERARCHICAL_DATA_NB(level - 1) % IAL, &
               & HIERARCHICAL_DATA_NB(level - 1) % IAU, &
               & finerNEIBPETOT, finerNEIBPE, finerSTACK_IMPORT, finerNOD_IMPORT, &
               & finerSTACK_EXPORT, finerNOD_EXPORT, &
               & PNI, NI, level,node_index,in_aggregates_result,            &
               &          aggregates_result ,in_aggregates_result_size,my_rank,      &
               &          aggregate_table_size,aggregate_table_array,                &
               &          aggregate_number_in_table,SOLVER_COMM,NPROCS,GIN_aggregate,&
               &          PE_num_size,PE_nums,nodes_for_each_PE,in_nodes_for_each_PE,&
               &          PE_list_size,PE_list,2)
          end if

       END IF

       CALL count_time(2,solver_comm,my_rank,6)


       coarser_level_size = GIN_aggregate(my_rank+1)-GIN_aggregate(my_rank)

       global_CN = 0
       CALL MPI_ALLREDUCE(coarser_level_size, global_CN, 1, MPI_INTEGER, MPI_SUM, SOLVER_COMM, ierr)
       
#ifdef PRINT_REZ          
       if(my_rank == 0) write(*,*) global_CN 
#endif
       global_finish_flag = (global_CN < MIN_NODE_SIZE) .OR. (level >= MAX_LEVEL_SIZE) 

       
       if(.not.NOT_USE_AGGRE_TRANSFER) then
          if(global_finish_flag) then
          
!!$          call output_aggre_data(my_rank, in_aggregates_result_size, in_aggregates_result, &
!!$               & aggregates_result, NPROCS, GIN_aggregate, aggregate_table_array,          &
!!$               & aggregate_number_in_table, PE_list, PE_list_size)
!!$          stop 'normal finish'
          
             !C most coarsest level is transferred to single process
             allocate(new_GIN_aggregate(0:NPROCS))
             new_GIN_aggregate(0) = 0
             new_GIN_aggregate(1:NPROCS) = GIN_aggregate(NPROCS)
             call aggregate_transfer(my_rank, in_aggregates_result_size, &
                  & in_aggregates_result, aggregates_result, NPROCS, GIN_aggregate, &
                  & aggregate_table_array, aggregate_number_in_table, PE_list, &
                  & PE_list_size, new_GIN_aggregate)
             
             deallocate(new_GIN_aggregate)
             
             
          end if
       end if
       
       call MPI_barrier(SOLVER_COMM, ierr)

       DEALLOCATE(node_index)

       coarser_level_size = GIN_aggregate(my_rank + 1) - GIN_aggregate(my_rank)

       
#ifdef DEBUG
       if(coarser_level_size == 0) write(*,*) "rank:",my_rank,"has no unknowns at lev",level
#endif

       

       RATE_OF_SPACE = RATE_OF_SPACE+80*(level-2)
       IF(RATE_OF_SPACE > global_CN) RATE_OF_SPACE = global_CN
       
       HIERARCHICAL_DATA_NB(level) % NB = NK
       HIERARCHICAL_DATA_NB(level) % NK = NK

       ALLOCATE(Temp_P (INT(RATE_OF_SPACE)*finer_level_NB*NK, finer_level_NP))
       ALLOCATE(Temp_PI(INT(RATE_OF_SPACE),   finer_level_NP))
       ALLOCATE(Temp_P_row_size(finer_level_NP))
          
       
       CALL DEBUG_BARRIER_PRINT("## smooth_aggregate      ##", SOLVER_COMM, my_rank)

       CALL count_time(1,solver_comm,my_rank,7)              

!!$       if(level == 2) then
!!$          call debug_output_aggre(my_rank, coarser_level_size, in_aggregates_result_size, &
!!$               & in_aggregates_result, aggregates_result)
!!$       end if
           


       if(associated(k_or_q_vec)) then
          
          CALL smooth_aggregate_unsym_vec (PNI, NI, dump_jacobi_weight, level, &
               &   in_aggregates_result_size, in_aggregates_result,            &
               &   aggregates_result, Temp_P, Temp_PI, Temp_P_row_size,        &
               &   tilde_DLU, INT(RATE_OF_SPACE), EX_INL, EX_IAL, EX_AL,         &
               &   finer_level_NB, NK, SOLVER_COMM, my_rank,                   &
               &   HIERARCHICAL_DATA_NB(level-1) % k_or_q_vec)
          
       else
          
          CALL  smooth_aggregate_unsym_vec_NB (PNI, NI, dump_jacobi_weight, level,  &
               & in_aggregates_result_size, in_aggregates_result, aggregates_result,&
               & Temp_P, Temp_PI,Temp_P_row_size, tilde_DLU, INT(RATE_OF_SPACE), EX_INL, &
               & EX_IAL, EX_AL, finer_level_NB, SOLVER_COMM, my_rank)
          
       end if
       
       CALL count_time(2,solver_comm,my_rank,7)              

       DEALLOCATE(tilde_DLU)
       DEALLOCATE(NI)
       DEALLOCATE(PNI)
       DEALLOCATE(aggregates_result) 
       DEALLOCATE(in_aggregates_result)

       CALL DEBUG_BARRIER_PRINT("## sr_aggregates         ##", SOLVER_COMM, my_rank)       
       
       ALLOCATE(global_local_hash_table(hash_table_size,2))

       IF(NPROCS > 1) then
          CALL count_time(1,solver_comm,my_rank,2)       
          nfepe_size = in_nodes_for_each_PE(PE_num_size)
          
          CALL solver_SR_aggregates &
               &( PE_num_size,PE_nums, nfepe_size, nodes_for_each_PE, &
               &  in_nodes_for_each_PE,  &
               &  SOLVER_COMM, INT(RATE_OF_SPACE), finer_level_NP, &
               &  Temp_P,Temp_PI,Temp_P_row_size,my_rank,          &
               &  GIN_aggregate, NPROCS, hash_table_size,global_local_hash_table,&
               &  aggregate_table_size, aggregate_table_array,                  &
               &  aggregate_number_in_table,local_aggre_size,PE_list,&
               &  PE_list_size, level, finer_level_NB, NK) 
          
          CALL count_time(2,solver_comm,my_rank,2)       

          DEALLOCATE(PE_nums)
          DEALLOCATE(nodes_for_each_PE)
          DEALLOCATE(in_nodes_for_each_PE)
       ELSE
          !C for single CPU case
          local_aggre_size = in_aggregates_result_size
          DEALLOCATE(PE_nums)
       END IF


       !--C check Temp_P
       DO i=1,finer_level_NP
          IF(Temp_P_row_size(i) > RATE_OF_SPACE) THEN
             WRITE(*,*) "RATE_OF_SPACE should be enlarged! "
             STOP 'The allocation error : Temp_P in make_rap'
          END IF
       END DO



       !C-- Interpolation data is recorded in HIERARCHICAL_DATA_NB()..
       ALLOCATE(HIERARCHICAL_DATA_NB(level) % P)
       P => HIERARCHICAL_DATA_NB(level) % P
       P % ROW_SIZE = HIERARCHICAL_DATA_NB(level-1) % NP
       k            = P % ROW_SIZE
       ALLOCATE(P % IN(0:k))
       P % IN(0) = 0
       DO i = 1,k
          P % IN(i) = P % IN(i-1) + Temp_P_row_size(i)
       END DO
       j = P % IN(k)
       ALLOCATE(P % CN ( j     ))
       ALLOCATE(P % V  ( j * finer_level_NB*NK ))
       k=1
       DO i=1,P % ROW_SIZE

          DO j=1,Temp_P_row_size(i)
             P % CN(k)    = Temp_PI(j,i)
             !C transposing Temp_P  column major to row major
             DO l = 1, finer_level_NB
                DO m = 1, NK
                   P % V((k-1)*finer_level_NB*Nk+(l-1)*NK+m) = Temp_P((j-1)*finer_level_NB*NK+l+(m-1)*finer_level_NB ,i)
                end DO
             END DO
             
             k = k+1
          END DO
       END DO
       
!!$       if(level==2) then
!!$          write(*,*) my_rank,finer_level_size, coarser_level_size,"debug!!"
!!$          call debug_output(my_rank,finer_level_NB, NK, finer_level_NP, P)
!!$       end if


       DEALLOCATE(Temp_P)
       DEALLOCATE(Temp_PI)
       DEALLOCATE(Temp_P_row_size)


       !C-- end:Temp_P is recorded in HIERARCHICAL_DATA_NB()..


       !C-- make R : 1..finer_level_NP -> 1..local_aggre_size
       !C   after RAP calculation,
       !C   later chanremake R:  1..finer_level_size -> 1..local_aggre_size
       ALLOCATE( HIERARCHICAL_DATA_NB(level) % R )
       R => HIERARCHICAL_DATA_NB(level) % R
       P => HIERARCHICAL_DATA_NB(level) % P
       R % ROW_SIZE = local_aggre_size
       ALLOCATE( Temp_R_IN( 0:local_aggre_size ) )
       Temp_R_IN( 0:local_aggre_size ) = 0
       
       count = 0
       DO i = 1,finer_level_NP
          DO j = P % IN(i-1)+1,P % IN(i)
             col = P % CN(j)
             count = count+1
             IF(col < local_aggre_size)  &
                  & Temp_R_IN(col+1) = Temp_R_IN(col+1)+1
          END DO
       END DO

       ALLOCATE(R % CN(count)  )
       ALLOCATE(R % V (count*Nk*finer_level_NB))

       DO i = 2,R % ROW_SIZE
          Temp_R_IN(i) = Temp_R_IN(i) + Temp_R_IN(i-1)
       END DO

       !C    -8 -7 -6
       !C    -5 -4 -3
       !C    -2 -1 -0
       !C row major access.
       DO i = 1,finer_level_NP
          DO j = P % IN(i-1)+1,P % IN(i)
             col = P % CN(j)
             Temp_R_IN(col) = Temp_R_IN(col)+1
             k = Temp_R_IN(col)
             R % CN(k) = i

             DO l = 1, NK
                DO m = 1, finer_level_NB
                   R % V((k-1)*finer_level_NB*NK+(l-1)*finer_level_NB+m) = P % V((j-1)*finer_level_NB*NK+(m-1)*NK+l)
                END DO
             END DO

          END DO
       END DO
       !C--end: make R




       CALL DEBUG_BARRIER_PRINT("## make_rap              ##", SOLVER_COMM, my_rank)       

       !C Temp_P,Temp_PI,Temp_P_row_size are deallocated..
       !C non use of pspases
       finish_flag = .FALSE.
       CALL make_rap(level, my_rank,SOLVER_COMM,NPROCS,               &
            & coarser_level_size, global_CN,                          &
            & global_finish_flag, GIN_aggregate,PE_list_size,         &
            & PE_list, aggregate_table_size,aggregate_table_array,    &
            & aggregate_number_in_table, hash_table_size,             &
            & global_local_hash_table,local_aggre_size,               &
            & Temp_R_IN, finer_level_NB, NK, PRO_AGGRE_OPT, coarser_lev_N_bf_CGA)


       
       DEALLOCATE(Temp_R_IN)
       DEALLOCATE(global_local_hash_table)
       DO i = 1, NPROCS
          call deallocate_int_vec(aggregate_table_array(i))
       END DO
       DEALLOCATE(aggregate_table_array)
       DEALLOCATE(aggregate_number_in_table)
       DEALLOCATE(GIN_aggregate)

       if(.not.SYMMETRIC_FLAG) then       
          deallocate(EX_INL, EX_IAL, EX_AL)       
       end if
       

#ifdef DEBUG
       IF(my_rank==0)THEN
          WRITE(*,*) '############',level,'############'
       END IF
#endif



#ifdef MEASURING
       call count_time(2,SOLVER_COMM,my_rank,30+level-1)
#endif
       
       IF(global_finish_flag) EXIT

    END DO

    LEVEL_NUM = level

    !C--ALU is level1's diagonal inverse
    ALLOCATE( HIERARCHICAL_DATA_NB(1) % ALU(NB,NB+1,N))
    newALU => HIERARCHICAL_DATA_NB(1) % ALU


    DO i=1,N
       CALL ILU_NB (newALU(:,:,i), D(:,:,i), NB)
    END DO



#ifdef MEASURING
    call count_time(2,SOLVER_COMM,my_rank,19)
    !C lev: NEIBPETOT at eeach lev
    DO i =1, LEVEL_NUM
       call store_and_disp_data(0, SOLVER_COMM, my_rank, i, 0.d0)
       call store_and_disp_data(1, SOLVER_COMM, my_rank, i, &
            & dble(HIERARCHICAL_DATA_NB(i) % COMM_TABLE % NEIBPETOT))
    END DO

    !C comm table size at eeach lev
    DO i =1, LEVEL_NUM
       k = HIERARCHICAL_DATA_NB(i) % NB
       call store_and_disp_data(0, SOLVER_COMM, my_rank, i+10, 0.d0)
       j = HIERARCHICAL_DATA_NB(i) % COMM_TABLE % NEIBPETOT
       if(j > 0)  call store_and_disp_data(1, SOLVER_COMM, my_rank, i+10, &
            & dble(HIERARCHICAL_DATA_NB(i) % COMM_TABLE % STACK_EXPORT(j) * k))
    END DO

    !C NEIBPETOT between levs
    DO i =2, LEVEL_NUM
       call store_and_disp_data(0, SOLVER_COMM, my_rank, i-1+20, 0.d0)
       call store_and_disp_data(1, SOLVER_COMM, my_rank, i-1+20, &
            & dble(HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % NEIBPETOT))
    END DO
    
    !C comm table size between levs
    DO i = 2, LEVEL_NUM
       k = HIERARCHICAL_DATA_NB(i) % NB
       call store_and_disp_data(0, SOLVER_COMM, my_rank, i-1+30, 0.d0)
       j = HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % NEIBPETOT
       if(j > 0) call store_and_disp_data(1, SOLVER_COMM, my_rank, i-1+30, &
            & dble(HIERARCHICAL_DATA_NB(i) % INT_LVL_TABLE % STACK_EXPORT(j) * k))
       
    END DO

    DO i = 1, LEVEL_NUM
       call disp_nonzero_data(0, SOLVER_COMM, my_rank, i, dble(HIERARCHICAL_DATA_NB(i) % N))
       j =    HIERARCHICAL_DATA_NB(i) % INL(HIERARCHICAL_DATA_NB(i) % N) + &
            & HIERARCHICAL_DATA_NB(i) % INU(HIERARCHICAL_DATA_NB(i) % N) + &
            & HIERARCHICAL_DATA_NB(i) % N 
       call disp_nonzero_data(1, SOLVER_COMM, my_rank, i, dble(j))
    END DO

#endif

  END SUBROUTINE data_creation_NB
  
  subroutine debug_output( my_rank, NB, NK, NP, P)
    USE data_structure_for_AMG_NB, only: INTER_LEVEL_OPERATOR
    implicit none
    include 'precision.inc'
    integer(kind=kint) :: my_rank, NP,NK, NB
    TYPE(INTER_LEVEL_OPERATOR),POINTER :: P

    integer(kind=kint) :: i,j,k,l,m,col
    character(5) :: rank_s, rank_s_l
    character(10) :: fname
    real(kind=kreal) :: norm
    
    write(rank_s, '(i5)') my_rank
    rank_s_l=adjustl(rank_s)
    fname=trim("debug"//rank_s_l)
    
    write(*,*) fname
    open(23, file=fname, form='formatted')

    DO i = 1, NP
       DO j = P % IN(i-1)+1, P % IN(i)
          col = P % CN(j)
          DO l = 1, NK
             DO m = 1, NB
                write(23,*) (i-1)*NB+m,(col-1)*NK+l,P %V((j-1)*NB*NK+(m-1)*NK+l)
             END DO
          END DO
       END DO
    END DO
    close(23)
    
  end subroutine debug_output
  
  subroutine debug_output_aggre(my_rank, coarser_level_size, in_aggregates_result_size, &
               & in_aggregates_result, aggregates_result)
    implicit none
    include 'precision.inc'
    integer(kind=kint) :: my_rank, coarser_level_size,in_aggregates_result_size
    integer(kind=kint) :: in_aggregates_result(0:)
    integer(kind=kint) :: aggregates_result(:)


    integer(kind=kint) :: i,j,k,l,m,col
    character(5) :: rank_s, rank_s_l
    character(10) :: fname
    
    write(rank_s, '(i5)') my_rank
    rank_s_l=adjustl(rank_s)
    fname=trim("aggre"//rank_s_l)
    
    write(*,*) fname
    open(24, file=fname, form='formatted')


    write(24, *) coarser_level_size
    DO i = 1, in_aggregates_result_size
       DO j = in_aggregates_result(i-1)+1, in_aggregates_result(i)
          col = aggregates_result(j)
          write(24, *) i,col
       end DO
    end DO
    close(24)
    
  end subroutine debug_output_aggre

  !C this kernel supposed to be used to orthogonalization
  !C independent aggregation is assumed.
  !C k_or_q_vec_f(:,:)
  subroutine assign_kernel_vectors(in_aggregates_result, aggregates_result, &
       & in_aggregates_result_size, k_or_q_vec_f, N_f, NB_f, k_or_q_vec_c, NK)
    implicit none
    INCLUDE  'precision.inc'
    !C k_or_q_vec(N*NB, NK)
    real(kind=kreal), intent(inout) :: k_or_q_vec_f(:,:),k_or_q_vec_c(:,:)
    integer(kind=kint), intent(in) :: N_f, NB_f, NK, in_aggregates_result_size
    integer(kind=kint) :: in_aggregates_result(0:), aggregates_result(:) 
    
    integer(kind=kint) :: i, j, k, maxm, LWORK
    
    real(kind=kreal), allocatable :: temp_A(:,:), temp_TAU(:), temp_WORK(:)
        
    maxm=0
    do i = 1, in_aggregates_result_size
       j = in_aggregates_result(i)-in_aggregates_result(i-1)
       if(maxm < j) maxm = j
    end do
    
    !C omp parallel
    !C --
    !C prepare arrays for each thread
    LWORK = NK*20
    allocate(temp_A(maxm*NB_f,NK),temp_TAU(NK),temp_WORK(LWORK))
    
    !C omp do
    do k = 1, in_aggregates_result_size
       call qr_for_aggregate_k(k,in_aggregates_result, aggregates_result, &
            & k_or_q_vec_f, N_f, NB_f, k_or_q_vec_c, in_aggregates_result_size, NK, maxm,       &
            & temp_A, temp_TAU, temp_WORK, LWORK)
    end do
    !C
    !C deallocate arrays
    !C
    deallocate(temp_A, temp_TAU, temp_WORK)
    
!!$    k_or_q_vec_c(1:in_aggregates_result_size*NK,1:NK) = k_or_q_vec_f(1:in_aggregates_result_size*NK,1:NK)
     
  end subroutine assign_kernel_vectors

  
  !C this subroutine considers kth aggregates
  subroutine qr_for_aggregate_k(k,in_aggregates_result, aggregates_result, &
       & k_or_q_vec_f, N_f, NB_f, k_or_q_vec_c, N_c, NK, maxm, &
       & temp_A, temp_TAU, temp_WORK, LWORK)
!C    use lapack95
    implicit none
    include 'precision.inc'
    integer(kind=kint), intent(in) :: k, maxm, in_aggregates_result(0:), aggregates_result(:)
    integer(kind=kint), intent(in) :: N_f, N_c, NB_f, NK, LWORK
    real(kind=kreal) :: k_or_q_vec_f(:,:), k_or_q_vec_c(:,:), temp_A(:,:), temp_TAU(:), temp_WORK(:)
    
    integer(kind=kint) :: i,j,row,col,rowt, INFO, rown
    
    !C k_or_q_vec_f(N_f*NB_f,NK)
    !C k_or_q_vec_c(N_c*NK,  NK)
    !C temp_A(maxm*NB_f)
    !C temp_TAU(NK)
    !C temp_WORK(LWORK)
    
    rowt   = 0
    DO i = in_aggregates_result(k-1)+1, in_aggregates_result(k)
       j = aggregates_result(i)
       DO row = (j-1)*NB_f+1,j*NB_f
          rowt = rowt+1
          DO col = 1, NK
             temp_A(rowt, col) = k_or_q_vec_f(row,col)
          END DO
       END DO
    end DO
    
    call dgeqrf(rowt, NK, temp_A, maxm*NB_f, temp_TAU, temp_WORK, LWORK,INFO)
    if(INFO/=0) stop 'error in qr factorization!!'
    
    !C copy back to R at k_or_q_vec_c((k-1)*NK+1-3)
    DO i = 1, min(rowt,NK)
       DO j = 1,i-1
          k_or_q_vec_c((k-1)*NK+i,j)=0.0
       end DO
       DO j = i,NK
          k_or_q_vec_c((k-1)*NK+i,j)=temp_A(i,j)
       end DO
    end DO

    DO i = min(rowt,NK)+1, NK
       DO j = 1,NK
          k_or_q_vec_c((k-1)*NK+i,j)=0.0
       end DO
    end DO

    
    call dorgqr(rowt, min(NK,rowt), min(rowt, NK), temp_A, maxm*NB_f, temp_TAU, temp_WORK, LWORK, INFO)
    if(INFO/=0) stop 'error in qr factorization!! 2'

    !C copy back to Q at k_or_q_vec_f()
    rown   = 0
    DO i = in_aggregates_result(k-1)+1, in_aggregates_result(k)
       j = aggregates_result(i)
       DO row = (j-1)*NB_f+1,j*NB_f
          rown = rown+1
          DO col = 1, min(NK,rowt)
             k_or_q_vec_f(row,col) = temp_A(rown, col)
          END DO
          DO col = min(NK,rowt)+1, NK
             k_or_q_vec_f(row,col) = 0.0
          END DO
       END DO
    end DO

  end subroutine qr_for_aggregate_k
  

  !C create the lower external part of the matrix. 
  !C Row numbers of that part are from N+1 to NP on COMM tables.
  SUBROUTINE lower_ext_matrix_crtn(EX_NPL, EX_INL, EX_IAL, EX_AL, LEVEL_NO, my_rank, SOLVER_COMM, PETOT)
    USE data_structure_for_AMG_NB
    IMPLICIT NONE
    INCLUDE 'mpif.h'
    
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO, PETOT

    !C These three arrays are allocated
    INTEGER(kind=kint ) :: EX_NPL
    INTEGER(kind=kint ), POINTER :: EX_INL (:), EX_IAL (:)
    REAL   (kind=kreal), POINTER :: EX_AL  (:,:,:)

    INTEGER(kind=kint ) :: N, NP, NB, NBsq
    REAL   (kind=kreal), POINTER :: AU (:,:,:)
    REAL   (kind=kreal), POINTER :: AL (:,:,:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)
    !C communication data structures
    INTEGER(kind=kint ), INTENT(in) :: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in) :: my_rank

    INTEGER(kind=kint )          :: NEIBPETOT
    INTEGER(kind=kint ), POINTER :: NEIBPE    (:)
    INTEGER(kind=kint ), POINTER :: NOD_IMPORT(:), STACK_IMPORT(:)
    INTEGER(kind=kint ), POINTER :: NOD_EXPORT(:), STACK_EXPORT(:)
    

    INTEGER(kind=kint ), allocatable :: Temp_N_NP(:), in_send_buf(:), send_buf(:)
    !C debug
!!$    INTEGER(kind=kint ), allocatable :: Temp(:)
    REAL   (kind=kreal), allocatable :: send_buf_dbl(:)
!!$    INTEGER(kind=kint ), allocatable :: localglobal_vec(:), index(:)
    INTEGER(kind=kint ) :: i, j, k, l, m, r, c, send_buf_size, o, q
    INTEGER(kind=kint ) :: send_buf_base
!!$    INTEGER(kind=kint ), ALLOCATABLE  :: WSI(:), WRI(:)

    INTEGER(kind=kint ), allocatable :: in_recv_buf(:), recv_buf(:)
    REAL   (kind=kreal), allocatable :: recv_buf_dbl(:)
    INTEGER(kind=kint ) :: recv_buf_size, recv_buf_base

    INTEGER(kind=kint ) :: ierr, istart, inum
    INTEGER(kind=kint ), allocatable :: req1(:),req2(:),sta1(:,:),sta2(:,:)
    INTEGER(kind=kint ), allocatable :: req3(:),req4(:),sta3(:,:),sta4(:,:)

    NB = HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % NB     
    NBsq = NB*NB
    NP = HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % NP     
    N  = HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % N 
    
    INL => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % INL
    INU => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % INU
    IAL => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % IAL
    IAU => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % IAU
    AL  => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % AL
    AU  => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % AU
    NEIBPETOT    =  HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % COMM_TABLE % NEIBPETOT

    NEIBPE       => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % COMM_TABLE % NEIBPE
    STACK_IMPORT => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % COMM_TABLE % STACK_IMPORT
    NOD_IMPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % COMM_TABLE % NOD_IMPORT
    STACK_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % COMM_TABLE % STACK_EXPORT
    NOD_EXPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % COMM_TABLE % NOD_EXPORT

    !C- count up the communication size for each PE
       
    allocate(Temp_N_NP(NP - N), in_send_buf(STACK_EXPORT(NEIBPETOT)))
    
    Temp_N_NP = 0    
    in_send_buf = 0
    DO i = 1, NEIBPETOT
       
       !C numbering the order in RECV table
       DO j = STACK_IMPORT(i - 1) + 1, STACK_IMPORT(i)
          r = NOD_IMPORT(j)+ZERO_ORIGIN
          Temp_N_NP(r - N) = j - STACK_IMPORT(i - 1)
       END DO

       !C count up the nonzeros for each row in the transposed upper external matrix.
       DO j = STACK_EXPORT(i - 1) + 1, STACK_EXPORT(i)
          r = NOD_EXPORT(j)+ZERO_ORIGIN
          DO k = INU(r - 1) + 1, INU(r)
             c = IAU(k) + ZERO_ORIGIN
             if(c > N) then
                l = Temp_N_NP(c - N)
                if(l > 0) then
                   !C count up l
                   in_send_buf(j) = in_send_buf(j) + 1
                end if
             END if
          END DO
       END DO
       
       !C reset the Temp_N_NP
       DO j = STACK_IMPORT(i - 1) + 1, STACK_IMPORT(i)
          r = NOD_IMPORT(j)+ZERO_ORIGIN
          Temp_N_NP(r - N) = 0
       END DO
    END DO
    !C- count up the communication size for each PE
    

    !C- create the send_buf and send_buf_dbl
    !C indexing for each PE
    DO i = 1, NEIBPETOT
       DO j = STACK_EXPORT(i - 1) + 2, STACK_EXPORT(i)
          in_send_buf(j) = in_send_buf(j - 1) + in_send_buf(j)
       END DO
    END DO
    
    !C send_buf_size
    send_buf_size = 0
    DO i = 1, NEIBPETOT
       if(STACK_EXPORT(i) > STACK_EXPORT(i-1)) then
          send_buf_size = send_buf_size + in_send_buf(STACK_EXPORT(i))
       end if
    END DO
	
    allocate(send_buf(send_buf_size), send_buf_dbl(send_buf_size*NBsq))
    send_buf_base = 0
    DO i = 1, NEIBPETOT

       !C numbering the order in RECV table
       DO j = STACK_IMPORT(i - 1) + 1, STACK_IMPORT(i)
          r = NOD_IMPORT(j)+ZERO_ORIGIN
          Temp_N_NP(r - N) = j - STACK_IMPORT(i - 1)
       END DO


       !C shift the in_send_buf array
       DO j = STACK_EXPORT(i), STACK_EXPORT(i - 1) + 2, -1
          in_send_buf(j) = in_send_buf(j - 1)
       END DO
       
       if(STACK_EXPORT(i)>STACK_EXPORT(i-1)) in_send_buf(STACK_EXPORT(i - 1) + 1) = 0
       
       !C assign the nonzeros for each row in the upper external matrix.
       DO j = STACK_EXPORT(i - 1) + 1, STACK_EXPORT(i)
          r = NOD_EXPORT(j)+ZERO_ORIGIN
          DO k = INU(r - 1) + 1, INU(r)
             c = IAU(k) + ZERO_ORIGIN
             if(c > N) then
                l = Temp_N_NP(c - N)
                if(l > 0) then
                   m = in_send_buf(j) + 1
                   in_send_buf(j) = m
                   send_buf    (send_buf_base + m) = l
                   DO o = 1, NB
                      q = (send_buf_base+m-1)*NBsq+(o-1)*NB
                      send_buf_dbl(q+1:q+NB) = AU(1:NB,o,k)
                   END DO
                end if
             END if
          END DO
       END DO
       
       if(STACK_EXPORT(i) > STACK_EXPORT(i - 1)) then
          send_buf_base = send_buf_base + in_send_buf(STACK_EXPORT(i))
       end if
       
       !C reset the Temp_N_NP
       DO j = STACK_IMPORT(i - 1) + 1, STACK_IMPORT(i)
          r = NOD_IMPORT(j)+ZERO_ORIGIN
          Temp_N_NP(r - N) = 0
       END DO
    END DO
    !C- create the send_buf and send_buf_dbl

    
    !C   in_send_buf(1:STACK_EXPORT(NEIBPETOT)) stores each row's size of 
    !C the external upper part.
    !C   send_buf(1:send_buf_size) stores the numbers in IMPORT TABLE 
    !C in order to show the column number on the other PE.
    !C   send_buf_dbl(1:send_buf_size) stores the value of the nonzero element

    !C- sendrecv in_send_buf
    
    allocate(in_recv_buf(STACK_IMPORT(NEIBPETOT)))
    allocate(sta1(MPI_STATUS_SIZE, NEIBPETOT), sta2(MPI_STATUS_SIZE, NEIBPETOT))
    allocate(sta3(MPI_STATUS_SIZE, NEIBPETOT), sta4(MPI_STATUS_SIZE, NEIBPETOT))
    allocate(req1(NEIBPETOT), req2(NEIBPETOT))
    allocate(req3(NEIBPETOT), req4(NEIBPETOT))

    req1 = MPI_REQUEST_NULL
    req2 = MPI_REQUEST_NULL


    DO i = 1, NEIBPETOT
       istart = STACK_EXPORT(i - 1)
       inum   = STACK_EXPORT(i) - istart
       CALL MPI_ISEND(in_send_buf(istart + 1), inum, MPI_INTEGER, NEIBPE(i), 0, &
            & SOLVER_COMM, req1(i), ierr)
    END DO
    DO i = 1, NEIBPETOT
       istart = STACK_IMPORT(i - 1)
       inum = STACK_IMPORT(i) - istart
       CALL MPI_IRECV(in_recv_buf(istart + 1), inum, MPI_INTEGER, NEIBPE(i), 0, &
            & SOLVER_COMM, req2(i), ierr)
    END DO

    if(NEIBPETOT > 0) then
       call MPI_WAITALL (NEIBPETOT, req2, sta2, ierr)
       call MPI_WAITALL (NEIBPETOT, req1, sta1, ierr)
    end if
    !C- sendrecv in_send_buf

    !C- sendrecv send_buf, send_buf_dbl
    recv_buf_size = 0
    DO i = 1, NEIBPETOT
       if(STACK_IMPORT(i) > STACK_IMPORT(i-1)) then
          recv_buf_size = recv_buf_size + in_recv_buf(STACK_IMPORT(i))
       end if

    END DO
    
    allocate(recv_buf(recv_buf_size), recv_buf_dbl(recv_buf_size*NBsq))

    !C in_send_buf elements means accumulated numbers of nonzero elements of 
    !C the corresponding node on export table.
    !C So in the case that there are no export nodes for a PE, there are no
    !C elements to be sent.
    send_buf_base = 0
    DO i = 1, NEIBPETOT
       istart = send_buf_base
       if(STACK_EXPORT(i)==STACK_EXPORT(i-1)) then
          inum = 0
       else 
          inum = in_send_buf(STACK_EXPORT(i))
       end if
       if(inum>0) then
          call MPI_ISEND(send_buf(istart + 1),     inum, MPI_INTEGER,          &
               & NEIBPE(i), 0, SOLVER_COMM, req1(i), ierr)
          call MPI_ISEND(send_buf_dbl(istart*NBsq + 1), inum*NBsq, MPI_DOUBLE_PRECISION, &
               & NEIBPE(i), 0, SOLVER_COMM, req2(i), ierr)
          send_buf_base = send_buf_base + inum
       end if
    END DO

    recv_buf_base = 0
    DO i = 1, NEIBPETOT
       istart = recv_buf_base
       if(STACK_IMPORT(i)==STACK_IMPORT(i-1)) then
          inum = 0
       else 
          inum = in_recv_buf(STACK_IMPORT(i))
       end if
       if(inum>0) then
          call MPI_IRECV(recv_buf(istart + 1),     inum, MPI_INTEGER,          &
               & NEIBPE(i), 0, SOLVER_COMM, req3(i), ierr)
          call MPI_IRECV(recv_buf_dbl(istart*NBsq + 1), inum*NBsq, MPI_DOUBLE_PRECISION, &
               & NEIBPE(i), 0, SOLVER_COMM, req4(i), ierr)
          recv_buf_base = recv_buf_base + inum
       end if
    END DO
    if(NEIBPETOT > 0) then
       call MPI_WAITALL (NEIBPETOT, req3, sta3, ierr)
       call MPI_WAITALL (NEIBPETOT, req4, sta4, ierr)
       call MPI_WAITALL (NEIBPETOT, req2, sta2, ierr)
       call MPI_WAITALL (NEIBPETOT, req1, sta1, ierr)
    end if
    
    deallocate(in_send_buf, send_buf, send_buf_dbl, Temp_N_NP)
    !C- sendrecv send_buf, send_buf_dbl

    !C- input the external 
    allocate(EX_INL(0:NP - N))
    EX_INL(0:NP-N) = 0

    
    DO i = 1, NEIBPETOT
       !C front node in IMPORT table from each PE
       if(STACK_IMPORT(i-1) < STACK_IMPORT(i)) then
          j = STACK_IMPORT(i - 1) + 1
          r = NOD_IMPORT(j)+ZERO_ORIGIN - N
          EX_INL(r) = in_recv_buf(j)
       end if
       !C the other nodes in IMPORT table from each PE
       DO j = STACK_IMPORT(i - 1) + 2, STACK_IMPORT(i)
          r = NOD_IMPORT(j)+ZERO_ORIGIN - N
          EX_INL(r) = in_recv_buf(j) - in_recv_buf(j - 1)
       END DO
    END DO

    !C 1. EX_INL is indexed
    DO i = 1, NP - N
       EX_INL(i) = EX_INL(i) + EX_INL(i - 1)
    END DO
    
    
    !C 2. EX_IAL, EX_AL is allocated
    EX_NPL = EX_INL(NP - N)
    allocate(EX_IAL(EX_NPL), EX_AL(NB,NB,EX_NPL))


    !C 3. EX_INL is shifted
    DO i = NP - N, 1, -1
       EX_INL(i) = EX_INL(i - 1)
    END DO
    
    !C 4. assigning the data in EX_IAL, EX_AL with counting up in EX_INL
    recv_buf_base = 0
    DO i = 1, NEIBPETOT
       !C front node
       if(STACK_IMPORT(i-1)<STACK_IMPORT(i)) then
          j = STACK_IMPORT(i - 1) + 1
          r = NOD_IMPORT(j)+ZERO_ORIGIN - N
          l = EX_INL(r)
          DO k = 1, in_recv_buf(j)
             EX_IAL(l + k) = NOD_EXPORT(recv_buf(recv_buf_base + k) + STACK_EXPORT(i - 1))
             DO o = 1, NB
                q = (recv_buf_base+k-1)*NBsq+(o-1)*NB
                !C transposition was not needed.
                EX_AL (:,o,l + k) = recv_buf_dbl(q+1:q+NB)
             end DO
          END DO
          EX_INL(r) = EX_INL(r) + in_recv_buf(j)
          

          !C rest nodes
          DO j = STACK_IMPORT(i - 1) + 2, STACK_IMPORT(i)
             r = NOD_IMPORT(j)+ZERO_ORIGIN - N

             l = EX_INL(r)
             DO k = in_recv_buf(j - 1) + 1, in_recv_buf(j)
                EX_IAL(l + k - in_recv_buf(j - 1)) = NOD_EXPORT(recv_buf(recv_buf_base + k) + STACK_EXPORT(i - 1))
                DO o = 1, NB
                   q = (recv_buf_base + k-1)*NBsq+(o-1)*NB
                   !C transposition is not needed! 
                   EX_AL (:,o,l + k - in_recv_buf(j - 1)) = recv_buf_dbl(q+1:q+NB)

                END DO
             END DO
             EX_INL(r) = EX_INL(r) + in_recv_buf(j) - in_recv_buf(j - 1)
          END DO
          
          recv_buf_base = recv_buf_base + in_recv_buf(STACK_IMPORT(i))
       end if
    END DO
    
    deallocate(in_recv_buf, sta1, sta2, sta3, sta4, req1, req2, req3, req4)
    deallocate(recv_buf, recv_buf_dbl)

  END SUBROUTINE lower_ext_matrix_crtn



  SUBROUTINE repeated_nds_tbl_crtn(LEVEL_NO,sorted_NEIBPE,SOLVER_COMM,my_rank, &
       &       PE_num_size,PE_nums,nodes_for_each_PE,in_nodes_for_each_PE,NPROCS)
    USE data_structure_for_AMG_NB
    USE count_time_mod
    USE comm_subs_NB
    IMPLICIT NONE

    INTEGER(kind=kint ) :: N,NEIBPETOT

    INTEGER(kind=kint ), INTENT(in) :: sorted_NEIBPE(:)
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO,NPROCS
    INTEGER(kind=kint ), INTENT(in) :: SOLVER_COMM
    INTEGER(kind=kint ), INTENT(in) :: my_rank

    INTEGER(kind=kint ), POINTER    :: NEIBPE      (:)
    INTEGER(kind=kint ), POINTER    :: NOD_IMPORT  (:), STACK_IMPORT(:)
    INTEGER(kind=kint ), POINTER    :: NOD_EXPORT  (:), STACK_EXPORT(:)

    INTEGER(kind=kint ), ALLOCATABLE:: export_node_list(:,:)
    INTEGER(kind=kint ), ALLOCATABLE:: repeated_nodes_at_boundary(:,:)

    INTEGER(kind=kint ) :: i,j,k,l,js,je,count
    INTEGER(kind=kint ) :: current_node,repeated_nodes_tot

    INTEGER(kind=kint ) :: REPEATED_NODE_SIZE
    INTEGER(kind=kint ), ALLOCATABLE   :: export_repeated(:,:)
    INTEGER(kind=kint ), ALLOCATABLE   :: in_export_repeated(:)
    INTEGER(kind=kint ), ALLOCATABLE   :: repeated_nodes_send_buf(:)

    INTEGER(kind=kint ),INTENT(out) :: PE_num_size
    INTEGER(kind=kint ),  POINTER :: PE_nums(:)
    INTEGER(kind=kint ),  POINTER :: nodes_for_each_PE(:)
    INTEGER(kind=kint ),  POINTER :: in_nodes_for_each_PE(:)

    N            = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % N
    NEIBPE       => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % NEIBPE
    STACK_IMPORT => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % STACK_IMPORT
    NOD_IMPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % NOD_IMPORT
    STACK_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % STACK_EXPORT
    NOD_EXPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % NOD_EXPORT
    NEIBPETOT    =  HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % NEIBPETOT     


    k=STACK_EXPORT(NEIBPETOT)
    ALLOCATE(export_node_list(3,k))

    !C export_node_list 1: node 2: PE NO 3: next
    DO i=1,k
       export_node_list(3,i)=0
    END DO
    ALLOCATE(repeated_nodes_at_boundary(3,N))

    !C repeated_nodes_at_boundary 1:TAIL 2:SIZE,3:HEAD
    DO i=1,N
       repeated_nodes_at_boundary(1,i)=0
       repeated_nodes_at_boundary(2,i)=0
    END DO

    DO i=1,NEIBPETOT
       js=STACK_EXPORT(i-1)+1
       je=STACK_EXPORT(i)
       DO j=js,je
          k=NOD_EXPORT(j)
          repeated_nodes_at_boundary(2,k)=repeated_nodes_at_boundary(2,k)+1
          current_node=repeated_nodes_at_boundary(1,k)
          IF(current_node == 0) THEN
             repeated_nodes_at_boundary(3,k)=j
          ELSE
             export_node_list(3,current_node)=j
          END IF
          repeated_nodes_at_boundary(1,k)=j
          !C node is the number from  STACK_EXPORT's associated PE js
          export_node_list(1,j)=j-js+1
          export_node_list(2,j)=i
       END DO
    END DO

    !C-- REPEATED_NODE_SIZE is max size of send table
    REPEATED_NODE_SIZE=0
    DO i=1,NEIBPETOT
       j=STACK_EXPORT(i)-STACK_EXPORT(i-1)
       IF(REPEATED_NODE_SIZE<j) REPEATED_NODE_SIZE=j
    END DO
    !C--end: REPEATED_NODE_SIZE is calculated.

    ALLOCATE(export_repeated(REPEATED_NODE_SIZE,NEIBPETOT))
    ALLOCATE(in_export_repeated(0:NEIBPETOT))

    repeated_nodes_tot=0
    in_export_repeated(0:NEIBPETOT)=0

    DO i=1,N
       l=repeated_nodes_at_boundary(2,i)
       IF(l >= 2) THEN
          repeated_nodes_tot = repeated_nodes_tot + l*(l+1)
          !C l:number of PEs  *  l+1:node_NO,PE_list_size,PE_list  
          current_node=repeated_nodes_at_boundary(3,i)
          DO WHILE(current_node>0)
             k=export_node_list(2,current_node)
             in_export_repeated(k) = in_export_repeated(k)+1
             j=in_export_repeated(k)
             !C node_number
             export_repeated(j,k) = i
             current_node=export_node_list(3,current_node)
          END DO
       END IF
    END DO

    !C No nodes are repeated.-> assign 0 
    DO i=1,NEIBPETOT
       IF(in_export_repeated(i)==0)THEN
          repeated_nodes_tot=repeated_nodes_tot+1
       END IF
    END DO

    ALLOCATE(repeated_nodes_send_buf(repeated_nodes_tot))
    count=0
    DO i=1,NEIBPETOT
       IF(in_export_repeated(i)==0) THEN
          !C No nodes are repeated for that PE.
          count=count+1
          repeated_nodes_send_buf(count)=0
       ELSE
          DO j=1,in_export_repeated(i)
             k=export_repeated(j,i)
             current_node=repeated_nodes_at_boundary(3,k)
             count=count+1
             repeated_nodes_send_buf(count)=repeated_nodes_at_boundary(2,k)
             count=count+1
             !C l is the place for node number in NOD_EXPORT.
             l=count
             DO WHILE(current_node>0)
                !C PE in NEIBPE(i) is in the stack_export.
                !C export_node_list:: 1:node  2:pe  3:next
                IF(export_node_list(2,current_node)==i) THEN
                   repeated_nodes_send_buf(l)=export_node_list(1,current_node) 
                ELSE
                   count=count+1
                   repeated_nodes_send_buf(count)=NEIBPE(export_node_list(2,current_node))
#ifdef DEBUG
                   IF(NEIBPE(export_node_list(2,current_node))>NPROCS) THEN
                      STOP "repeated_nodes creation send buffer"
                   END IF
#endif
                END IF
                current_node=export_node_list(3,current_node)
             END DO
          END DO
       END IF
       !C in_export_repeated is changed to the index for each PE in 
       !C repeated_nodes_send_buf.
       in_export_repeated(i)=count
    END DO
    in_export_repeated(0)=0

    !C deallocate repeated_nodes_at_boundary
    DEALLOCATE(repeated_nodes_at_boundary)
    DEALLOCATE(export_node_list)
    DEALLOCATE(export_repeated)

    CALL count_time(1,solver_comm,my_rank,1)    
    !C repeated nodes information is exchanged
    !C PE_num_size,PE_nums,in_nodes_for_each_PE,nodes_for_each_PE is created

    Call solver_SR_repeated_nodes &
         &    ( repeated_nodes_tot,     &
         &      NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,                            &
         &      STACK_EXPORT, NOD_EXPORT, in_export_repeated,repeated_nodes_send_buf,   &
         &      PE_num_size,PE_nums,nodes_for_each_PE,in_nodes_for_each_PE,SOLVER_COMM, &
         &      sorted_NEIBPE,my_rank,NPROCS) 

    CALL count_time(2,solver_comm,my_rank,1)    

    DEALLOCATE(repeated_nodes_send_buf)
    DEALLOCATE(in_export_repeated)

  CONTAINS
    SUBROUTINE  solver_SR_repeated_nodes &
         &     ( SIZE, NEIBPETOT, NEIBPE, STACK_IMPORT, NOD_IMPORT,STACK_EXPORT,&
         &       NOD_EXPORT, in_export_repeated,repeated_nodes_send_buf,                &
         &       PE_num_size,PE_nums,nodes_for_each_PE,in_nodes_for_each_PE,SOLVER_COMM,&
         &       sorted_recv_order_rank,my_rank,NPROCS)

      USE isort
      IMPLICIT NONE
      INCLUDE  'mpif.h'
      INCLUDE  'precision.inc'


      INTEGER(kind=kint ), INTENT(in) :: SIZE,NEIBPETOT,NPROCS
      INTEGER(kind=kint ), INTENT(in) :: NEIBPE(NEIBPETOT)
      INTEGER(kind=kint ), INTENT(in) :: STACK_IMPORT(0:NEIBPETOT),STACK_EXPORT(0:NEIBPETOT)
      INTEGER(kind=kint ), INTENT(in) :: NOD_IMPORT(:),NOD_EXPORT(:)
      INTEGER(kind=kint ), INTENT(in) :: sorted_recv_order_rank(NEIBPETOT)

      INTEGER(kind=kint ), INTENT(inout) :: in_export_repeated(0:NEIBPETOT)
      INTEGER(kind=kint ), INTENT(in)  :: repeated_nodes_send_buf(SIZE)
      INTEGER(kind=kint ), INTENT(in)  :: SOLVER_COMM
      INTEGER(kind=kint ), INTENT(in)  :: my_rank

      INTEGER(kind=kint ),  ALLOCATABLE :: status(:)
      INTEGER(kind=kint ),  ALLOCATABLE :: sta1(:,:)
      INTEGER(kind=kint ),  ALLOCATABLE :: sta2(:,:)
      INTEGER(kind=kint ),  ALLOCATABLE :: req1(:)
      INTEGER(kind=kint ),  ALLOCATABLE :: req2(:)

      INTEGER(kind=kint ) :: neib,istart,inum,message_size,buf_size,count,w,s,e,hn,hc,hs
      INTEGER(kind=kint ) :: i,j,k,l,m,counter_PE_nums,counter_nodes_for_each_PE,base
      INTEGER(kind=kint ) :: PE_list_size,src_PE,PE_NO,Local_node_NO
      INTEGER :: ierr

      INTEGER(kind=kint), ALLOCATABLE :: buf(:)
      INTEGER(kind=kint), ALLOCATABLE :: cnt_nd_for_echPE(:)
      INTEGER(kind=kint), ALLOCATABLE :: in_recv_order_rank(:)

      INTEGER(kind=kint ), INTENT(out) :: PE_num_size
      INTEGER(kind=kint ), POINTER :: PE_nums(:)
      INTEGER(kind=kint ), POINTER :: nodes_for_each_PE(:)
      INTEGER(kind=kint ), POINTER :: in_nodes_for_each_PE(:)

      ALLOCATE (status(MPI_STATUS_SIZE))
      ALLOCATE (sta1(MPI_STATUS_SIZE,NEIBPETOT))
      ALLOCATE (sta2(MPI_STATUS_SIZE,NEIBPETOT))
      ALLOCATE (req1(NEIBPETOT))
      ALLOCATE (req2(NEIBPETOT))      
      ALLOCATE (in_recv_order_rank(0:NEIBPETOT))

      !C change index to size
      do i= NEIBPETOT, 1, -1
         in_export_repeated(i)= in_export_repeated(i)-in_export_repeated(i-1)
      end do

      !C-- message size is communicated
      DO neib= 1, NEIBPETOT
         CALL MPI_ISEND (in_export_repeated(neib), 1, MPI_INTEGER, NEIBPE(neib), 0, &
              &          SOLVER_COMM, req1(neib), ierr)
      END DO
      DO neib= 1, NEIBPETOT      
         CALL MPI_IRECV(in_recv_order_rank(neib),1,MPI_INTEGER, &
              &         NEIBPE(neib),0,SOLVER_COMM,req2(neib), ierr)
      END DO
      CALL MPI_WAITALL (NEIBPETOT, req2, sta2, ierr)
      CALL MPI_WAITALL (NEIBPETOT, req1, sta1, ierr)

      !C change size to index
      do i= 2, NEIBPETOT
         in_export_repeated(i)= in_export_repeated(i)+in_export_repeated(i-1)
      end do


      in_recv_order_rank(0)=0
      DO i=1,NEIBPETOT
         in_recv_order_rank(i)=in_recv_order_rank(i)+in_recv_order_rank(i-1)
      END DO

      ALLOCATE( buf( in_recv_order_rank(NEIBPETOT) ) )
      buf=0
      !C-- end: message size is communicated


      !C-- repeated node and PE numbers are communicated
      DO neib= 1, NEIBPETOT
         istart= in_export_repeated(neib-1)
         inum  = in_export_repeated(neib  ) - istart
         CALL MPI_ISEND (repeated_nodes_send_buf(istart+1), inum, MPI_INTEGER,     &
              &                NEIBPE(neib), 0, SOLVER_COMM,                       &
              &                req1(neib), ierr)
      END DO
      DO neib= 1, NEIBPETOT      
         istart=in_recv_order_rank(neib-1)
         inum=in_recv_order_rank(neib)-istart
         CALL MPI_IRECV(buf(istart+1),inum,MPI_INTEGER,NEIBPE(neib),0,SOLVER_COMM, &
              &         req2(neib),ierr)
      END DO

      CALL MPI_WAITALL (NEIBPETOT, req2, sta2, ierr)
      CALL MPI_WAITALL (NEIBPETOT, req1, sta1, ierr)
      !C--end: repeated node and PE numbers are communicated


      !C-- determine node sized for all PEs on cnt_nd_for_echPE
      ALLOCATE(cnt_nd_for_echPE(0:NPROCS-1))
      cnt_nd_for_echPE=0

      DO i=1,NEIBPETOT
         s=in_recv_order_rank(i-1)+1
         e=in_recv_order_rank(i)
         k=s
         IF(buf(k)>0) THEN
            DO WHILE(k < e)
               PE_list_size=buf(k)
               DO j=k+2,k+PE_list_size
                  PE_NO=buf(j)
#ifdef DEBUG
                  IF(PE_NO>NPROCS) THEN 
                     WRITE(*,*) "PE_NO:",PE_NO
                     STOP "PE_NO in repeated tables creation"
                  END IF
#endif
                  cnt_nd_for_echPE(PE_NO)=cnt_nd_for_echPE(PE_NO)+1
               END DO
               k=k+PE_list_size+1
            END DO
         END IF
      END DO
      !C--end: determine node size for all PEs on cnt_nd_for_echPE


      !C--PE_num_size and in_nodes_for_each_PE are created from cnt_nd_for_echPE
      !C--cnt_nd_for_echPE is used for transpose PE_nums in lateral part.
      PE_num_size=0
      DO i=0,NPROCS-1
         IF(cnt_nd_for_echPE(i)>0) PE_num_size=PE_num_size+1
      END DO
      ALLOCATE(PE_nums(PE_num_size))
      ALLOCATE(in_nodes_for_each_PE(0:PE_num_size+1))

      in_nodes_for_each_PE(0)=0
      in_nodes_for_each_PE(1)=0
      j=0
      DO i=0,NPROCS-1
         IF(cnt_nd_for_echPE(i)>0) THEN
            j=j+1
            PE_nums(j)=i
            in_nodes_for_each_PE(j+1)=in_nodes_for_each_PE(j)+cnt_nd_for_echPE(i)

            !C PE_nums index can be retrieved from PE number 
            cnt_nd_for_echPE(i)=j
         END IF
      END DO
      !C--end: PE_num_size and in_nodes_for_each_PE are created from cnt_nd_for_echPE    


      !C--nodes_for_each_PE is determined
      ALLOCATE(nodes_for_each_PE(in_nodes_for_each_PE(PE_num_size+1)))
      DO i=1,NEIBPETOT
         j=sorted_recv_order_rank(i)                
         base=STACK_IMPORT(j-1)
         s=in_recv_order_rank(j-1)+1
         e=in_recv_order_rank(j)
         k=s
         IF(buf(k)>0) THEN
            DO WHILE(k < e)
               PE_list_size=buf(k)
               Local_node_NO=NOD_IMPORT(buf(k+1)+base)
               DO j=k+2,k+PE_list_size
                  PE_NO=buf(j)
                  l=cnt_nd_for_echPE(PE_NO)

                  m=in_nodes_for_each_PE(l)+1
                  in_nodes_for_each_PE(l)=m
                  nodes_for_each_PE(m)=Local_node_NO

               END DO
               k=k+PE_list_size+1
            END DO
         END IF
      END DO
      DEALLOCATE(cnt_nd_for_echPE)
      !C--end: nodes_for_each_PE is determined

      DEALLOCATE (in_recv_order_rank)
      DEALLOCATE (buf)

      DEALLOCATE (status)
      DEALLOCATE (sta1)
      DEALLOCATE (sta2)
      DEALLOCATE (req1)
      DEALLOCATE (req2)
    END SUBROUTINE solver_SR_repeated_nodes

  END SUBROUTINE repeated_nds_tbl_crtn


  SUBROUTINE RAP_old(Temp_N,Temp_CN,Temp_V,hash_size,&
       &         N,ownaggre_size,LEVEL_NO,local_aggre_size)

    USE data_structure_for_AMG_NB
    IMPLICIT NONE

    INTEGER(kind=kint ),               INTENT(IN)   :: hash_size,N,ownaggre_size
    INTEGER(kind=kint ),               INTENT(IN)   :: LEVEL_NO,local_aggre_size
!!$    INTEGER(kind=kint ),DIMENSION(local_aggre_size),  INTENT(inout):: Temp_N
!!$    INTEGER(kind=kint ),DIMENSION(hash_size,local_aggre_size),INTENT(inout):: Temp_CN
    INTEGER(kind=kint ),DIMENSION(:),  INTENT(inout):: Temp_N
    INTEGER(kind=kint ),DIMENSION(:,:),INTENT(inout):: Temp_CN
    TYPE(INTER_LEVEL_OPERATOR),POINTER :: P

!!$    REAL   (kind=kreal),DIMENSION(hash_size*9,local_aggre_size),&
!!$         &                             INTENT(inout):: Temp_V
    REAL   (kind=kreal),DIMENSION(:,:), INTENT(inout):: Temp_V

    !C pointers for matrix
    REAL   (kind=kreal),POINTER, DIMENSION(:,:,:) ::  D
    REAL   (kind=kreal),POINTER, DIMENSION(:,:,:) ::  AU
    REAL   (kind=kreal),POINTER, DIMENSION(:,:,:) ::  AL
    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  INU
    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  IAU
    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  INL
    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  IAL

    INTEGER(kind=kint )::i,j,k,rowins,rowine,colins,coline,row,col
    INTEGER(kind=kint )::js,je,jin,rowin,colin,count,hash_no
    LOGICAL :: flag
    REAL   (kind=kreal) ::L11,L12,L13,L21,L22,L23,L31,L32,L33 
    REAL   (kind=kreal) ::M11,M12,M13,M21,M22,M23,M31,M32,M33 
    REAL   (kind=kreal) ::R11,R12,R13,R21,R22,R23,R31,R32,R33

    D            => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % D 
    INL          => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INL  
    INU          => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INU 
    IAL          => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAL
    IAU          => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAU
    AL           => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AL
    AU           => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AU
    P            => HIERARCHICAL_DATA_NB(LEVEL_NO  ) % P

    !C Calculation of RAP. 
    !C ------
    DO i=1,N
       rowins=P % IN(i-1)+1
       rowine=P % IN(i)

       !C A's diagonal part
       DO rowin = rowins, rowine
          row   = P % CN(rowin)
          colins=rowins
          coline=rowine
          DO colin=colins,coline
             col=P % CN (colin)
             IF(row<= ownaggre_size .AND. col<row ) CYCLE

             !C search (row col) in new matrix
             hash_no=MOD(col,hash_size)+1
             count=hash_no

             DO WHILE( count<=hash_size )
                IF(Temp_CN(count,row)==0 .OR. Temp_CN(count,row)==col) EXIT
                count=count+1
             END DO
             IF(count>hash_size) THEN
                count=1
                DO WHILE( count<hash_no .AND. Temp_CN(count,row)/=0    &
                     &                  .AND. Temp_CN(count,row)/=col )
                   count=count+1
                END DO
                IF(count>=hash_no) STOP "hash overflow..:Temp_CN,_V"
             END IF
             !C---three matrix(3,3) multiply
             R11=P % V(rowin*9-8)
             R21=P % V(rowin*9-7)
             R31=P % V(rowin*9-6)
             R12=P % V(rowin*9-5)
             R22=P % V(rowin*9-4)
             R32=P % V(rowin*9-3)
             R13=P % V(rowin*9-2)
             R23=P % V(rowin*9-1)
             R33=P % V(rowin*9  )
             M11 = R11*D(1,1,i)+R12*D(2,1,i)+R13*D(3,1,i)
             M12 = R11*D(1,2,i)+R12*D(2,2,i)+R13*D(3,2,i)
             M13 = R11*D(1,3,i)+R12*D(2,3,i)+R13*D(3,3,i)
             M21 = R21*D(1,1,i)+R22*D(2,1,i)+R23*D(3,1,i)
             M22 = R21*D(1,2,i)+R22*D(2,2,i)+R23*D(3,2,i)
             M23 = R21*D(1,3,i)+R22*D(2,3,i)+R23*D(3,3,i)
             M31 = R31*D(1,1,i)+R32*D(2,1,i)+R33*D(3,1,i)
             M32 = R31*D(1,2,i)+R32*D(2,2,i)+R33*D(3,2,i)
             M33 = R31*D(1,3,i)+R32*D(2,3,i)+R33*D(3,3,i)
             R11=P % V(colin*9-8)
             R12=P % V(colin*9-7)
             R13=P % V(colin*9-6)
             R21=P % V(colin*9-5)
             R22=P % V(colin*9-4)
             R23=P % V(colin*9-3)
             R31=P % V(colin*9-2)
             R32=P % V(colin*9-1)
             R33=P % V(colin*9  )
             L11 = M11*R11 + M12*R21 + M13*R31
             L12 = M11*R12 + M12*R22 + M13*R32
             L13 = M11*R13 + M12*R23 + M13*R33
             L21 = M21*R11 + M22*R21 + M23*R31
             L22 = M21*R12 + M22*R22 + M23*R32
             L23 = M21*R13 + M22*R23 + M23*R33
             L31 = M31*R11 + M32*R21 + M33*R31
             L32 = M31*R12 + M32*R22 + M33*R32
             L33 = M31*R13 + M32*R23 + M33*R33

             !C assign the value
             Temp_V(count*9-8,row) = Temp_V(count*9-8,row)+L11
             Temp_V(count*9-7,row) = Temp_V(count*9-7,row)+L12
             Temp_V(count*9-6,row) = Temp_V(count*9-6,row)+L13
             Temp_V(count*9-5,row) = Temp_V(count*9-5,row)+L21
             Temp_V(count*9-4,row) = Temp_V(count*9-4,row)+L22
             Temp_V(count*9-3,row) = Temp_V(count*9-3,row)+L23
             Temp_V(count*9-2,row) = Temp_V(count*9-2,row)+L31
             Temp_V(count*9-1,row) = Temp_V(count*9-1,row)+L32
             Temp_V(count*9,  row) = Temp_V(count*9,  row)+L33

             IF(Temp_CN(count,row)/=col) THEN
                Temp_N(row)=Temp_N(row)+1
                Temp_CN(count,row)=col
             END IF
             !C---
          END DO
       END DO

       !C A's lower part
       js=INL(i-1)+1
       je=INL(i)
       DO jin=js,je
          j=IAL(jin)

          DO rowin=rowins,rowine
             row=P % CN(rowin)
             colins=P % IN(j-1)+1
             coline=P % IN(j)
             DO colin=colins,coline
                col=P % CN(colin)
                IF(row<= ownaggre_size .AND. col<row ) CYCLE

                !C search (row col) in new matrix
                hash_no=MOD(col,hash_size)+1
                count=hash_no

                DO WHILE( count<=hash_size)
                   IF(Temp_CN(count,row)==0 .OR. Temp_CN(count,row)==col) EXIT
                   count=count+1
                END DO
                IF(count>hash_size) THEN
                   count=1
                   DO WHILE( count<hash_no .AND. Temp_CN(count,row)/=0 &
                        &                  .AND. Temp_CN(count,row)/=col )
                      count=count+1
                   END DO
                   IF(count>=hash_no) STOP "hash overflow.."
                END IF

                !C---three matrix(3,3) multiply
                R11=P % V(rowin*9-8)
                R21=P % V(rowin*9-7)
                R31=P % V(rowin*9-6)
                R12=P % V(rowin*9-5)
                R22=P % V(rowin*9-4)
                R32=P % V(rowin*9-3)
                R13=P % V(rowin*9-2)
                R23=P % V(rowin*9-1)
                R33=P % V(rowin*9  )
                M11 = R11*AL(1,1,jin)+R12*AL(2,1,jin)+R13*AL(3,1,jin)
                M12 = R11*AL(1,2,jin)+R12*AL(2,2,jin)+R13*AL(3,2,jin)
                M13 = R11*AL(1,3,jin)+R12*AL(2,3,jin)+R13*AL(3,3,jin)
                M21 = R21*AL(1,1,jin)+R22*AL(2,1,jin)+R23*AL(3,1,jin)
                M22 = R21*AL(1,2,jin)+R22*AL(2,2,jin)+R23*AL(3,2,jin)
                M23 = R21*AL(1,3,jin)+R22*AL(2,3,jin)+R23*AL(3,3,jin)
                M31 = R31*AL(1,1,jin)+R32*AL(2,1,jin)+R33*AL(3,1,jin)
                M32 = R31*AL(1,2,jin)+R32*AL(2,2,jin)+R33*AL(3,2,jin)
                M33 = R31*AL(1,3,jin)+R32*AL(2,3,jin)+R33*AL(3,3,jin)
                R11=P % V(colin*9-8)
                R12=P % V(colin*9-7)
                R13=P % V(colin*9-6)
                R21=P % V(colin*9-5)
                R22=P % V(colin*9-4)
                R23=P % V(colin*9-3)
                R31=P % V(colin*9-2)
                R32=P % V(colin*9-1)
                R33=P % V(colin*9  )
                L11 = M11*R11 + M12*R21 + M13*R31
                L12 = M11*R12 + M12*R22 + M13*R32
                L13 = M11*R13 + M12*R23 + M13*R33
                L21 = M21*R11 + M22*R21 + M23*R31
                L22 = M21*R12 + M22*R22 + M23*R32
                L23 = M21*R13 + M22*R23 + M23*R33
                L31 = M31*R11 + M32*R21 + M33*R31
                L32 = M31*R12 + M32*R22 + M33*R32
                L33 = M31*R13 + M32*R23 + M33*R33

                !C assign the value
                Temp_V(count*9-8,row) = Temp_V(count*9-8,row)+L11
                Temp_V(count*9-7,row) = Temp_V(count*9-7,row)+L12
                Temp_V(count*9-6,row) = Temp_V(count*9-6,row)+L13
                Temp_V(count*9-5,row) = Temp_V(count*9-5,row)+L21
                Temp_V(count*9-4,row) = Temp_V(count*9-4,row)+L22
                Temp_V(count*9-3,row) = Temp_V(count*9-3,row)+L23
                Temp_V(count*9-2,row) = Temp_V(count*9-2,row)+L31
                Temp_V(count*9-1,row) = Temp_V(count*9-1,row)+L32
                Temp_V(count*9  ,row) = Temp_V(count*9  ,row)+L33

                IF(Temp_CN(count,row)/=col) THEN
                   Temp_N(row)=Temp_N(row)+1
                   Temp_CN(count,row)=col
                END IF
                !C---
             END DO
          END DO
       END DO

       !C A's upper part      
       js=INU(i-1)+1
       je=INU(i)
       DO jin=js,je
          j=IAU(jin)

          DO rowin=rowins,rowine
             row=P % CN(rowin)

             colins=P % IN(j-1)+1
             coline=P % IN(j)
             DO colin=colins,coline
                col=P % CN(colin)

                IF(row<=ownaggre_size .AND.col<row ) CYCLE

                !C search (row col) in new matrix
!!$                CALL hash(col,hash_size,hash_no)
!!$                CALL FIND(hash_no,hash_size,count,Temp_CN(:,row),col)
                hash_no=MOD(col,hash_size)+1
                count=hash_no

                DO WHILE(count<=hash_size)
                   IF(Temp_CN(count,row)==0 .OR. Temp_CN(count,row)==col) EXIT
                   count=count+1
                END DO
                IF(count>hash_size) THEN
                   count=1
                   DO WHILE( count<hash_no .AND. Temp_CN(count,row)/=0  &
                        &                  .AND. Temp_CN(count,row)/=col )
                      count=count+1
                   END DO
                   IF(count>=hash_no) STOP "hash overflow.."
                END IF

                !C---three matrix(3,3) multiply
                R11=P % V(rowin*9-8)
                R21=P % V(rowin*9-7)
                R31=P % V(rowin*9-6)
                R12=P % V(rowin*9-5)
                R22=P % V(rowin*9-4)
                R32=P % V(rowin*9-3)
                R13=P % V(rowin*9-2)
                R23=P % V(rowin*9-1)
                R33=P % V(rowin*9  )
                M11 = R11*AU(1,1,jin)+R12*AU(2,1,jin)+R13*AU(3,1,jin)
                M12 = R11*AU(1,2,jin)+R12*AU(2,2,jin)+R13*AU(3,2,jin)
                M13 = R11*AU(1,3,jin)+R12*AU(2,3,jin)+R13*AU(3,3,jin)
                M21 = R21*AU(1,1,jin)+R22*AU(2,1,jin)+R23*AU(3,1,jin)
                M22 = R21*AU(1,2,jin)+R22*AU(2,2,jin)+R23*AU(3,2,jin)
                M23 = R21*AU(1,3,jin)+R22*AU(2,3,jin)+R23*AU(3,3,jin)
                M31 = R31*AU(1,1,jin)+R32*AU(2,1,jin)+R33*AU(3,1,jin)
                M32 = R31*AU(1,2,jin)+R32*AU(2,2,jin)+R33*AU(3,2,jin)
                M33 = R31*AU(1,3,jin)+R32*AU(2,3,jin)+R33*AU(3,3,jin)
                R11=P % V(colin*9-8)
                R12=P % V(colin*9-7)
                R13=P % V(colin*9-6)
                R21=P % V(colin*9-5)
                R22=P % V(colin*9-4)
                R23=P % V(colin*9-3)
                R31=P % V(colin*9-2)
                R32=P % V(colin*9-1)
                R33=P % V(colin*9  )
                L11 = M11*R11 + M12*R21 + M13*R31
                L12 = M11*R12 + M12*R22 + M13*R32
                L13 = M11*R13 + M12*R23 + M13*R33
                L21 = M21*R11 + M22*R21 + M23*R31
                L22 = M21*R12 + M22*R22 + M23*R32
                L23 = M21*R13 + M22*R23 + M23*R33
                L31 = M31*R11 + M32*R21 + M33*R31
                L32 = M31*R12 + M32*R22 + M33*R32
                L33 = M31*R13 + M32*R23 + M33*R33

                !C assign the value
                Temp_V(count*9-8,row) = Temp_V(count*9-8,row)+L11
                Temp_V(count*9-7,row) = Temp_V(count*9-7,row)+L12
                Temp_V(count*9-6,row) = Temp_V(count*9-6,row)+L13
                Temp_V(count*9-5,row) = Temp_V(count*9-5,row)+L21
                Temp_V(count*9-4,row) = Temp_V(count*9-4,row)+L22
                Temp_V(count*9-3,row) = Temp_V(count*9-3,row)+L23
                Temp_V(count*9-2,row) = Temp_V(count*9-2,row)+L31
                Temp_V(count*9-1,row) = Temp_V(count*9-1,row)+L32
                Temp_V(count*9  ,row) = Temp_V(count*9  ,row)+L33

                IF(Temp_CN(count,row)/=col) THEN
                   Temp_N(row)=Temp_N(row)+1
                   Temp_CN(count,row)=col
                END IF
                !C---

             END DO
          END DO
       END DO
    END DO
  END SUBROUTINE RAP_old

  !C symmetricity is assumed to cut off local lower part's calculation.
  SUBROUTINE RAP_NB_old(Temp_N,Temp_CN,Temp_V,hash_size,N,ownaggre_size,LEVEL_NO, &
       & local_aggre_size, Temp_R_IN)

    USE data_structure_for_AMG_NB
    IMPLICIT NONE

    INTEGER(kind=kint),intent(in)   :: hash_size,N,ownaggre_size,LEVEL_NO,local_aggre_size
    INTEGER(kind=kint),intent(in)   :: Temp_R_IN(0:)
    INTEGER(kind=kint),intent(inout):: Temp_N(local_aggre_size)
    INTEGER(kind=kint),intent(inout):: Temp_CN(hash_size,local_aggre_size)
    type(INTER_LEVEL_OPERATOR),pointer :: P, R
!!$    REAL(kind=kreal),intent(inout):: Temp_V(hash_size*9,local_aggre_size)
    REAL(kind=kreal),intent(inout):: Temp_V(:,:)

    !C pointers for matrix
    REAL   (kind=kreal),POINTER ::  D(:,:,:)
    REAL   (kind=kreal),POINTER ::  AU(:,:,:)
    REAL   (kind=kreal),POINTER ::  AL(:,:,:)
    INTEGER(kind=kint ),POINTER ::  INU(:)
    INTEGER(kind=kint ),POINTER ::  IAU(:)
    INTEGER(kind=kint ),POINTER ::  INL(:)
    INTEGER(kind=kint ),POINTER ::  IAL(:)

    INTEGER(kind=kint ),allocatable :: Temp_agr(:),Temp_sz(:),Temp_nd(:)
    INTEGER(kind=kint ),allocatable :: Temp_int(:,:)
    REAL   (kind=kreal),allocatable :: Temp(:,:),Temp_1_col(:)

    integer(kind=kint) ::i,j,k,l,l1,m,m1,m2,rowins,rowine,colins,coline,row,col,row2
    integer(kind=kint) ::do_whl_cnt,is,ie
    integer(kind=kint) ::js,je,jin,rowin,colin,count,hash_no,NP, Temp_SIZE, NB
    logical :: cancel_flag,finish_flag,logi

    INTEGER(kind=kint ) :: in1,in2,P_col,T_col,p_width,t_width,cancel_sz
    REAL   (kind=kreal) :: R11,R12,R13,R21,R22,R23,R31,R32,R33, v
    REAL   (kind=kreal) :: M11,M12,M13,M21,M22,M23,M31,M32,M33



    NB =  HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NB 
    NP   =  HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NP 
    D    => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % D 
    INL  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INL  
    INU  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INU 
    IAL  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAL
    IAU  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAU
    AL   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AL
    AU   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AU
    P    => HIERARCHICAL_DATA_NB(LEVEL_NO  ) % P
    R    => HIERARCHICAL_DATA_NB(LEVEL_NO  ) % R

    !C Calculation of RAP. This uses R information. 
    !C R % IN must be rewrite @ the latter part.
    !C So Temp_R_IN is used for R % IN(:).
    !C ------

    Temp_SIZE = 8

    allocate( Temp_sz (NP)    )
    allocate( Temp_int(NP,Temp_SIZE) )
    allocate( Temp    (NP*NB*NB,Temp_SIZE) )
    allocate( Temp_agr(local_aggre_size) )
    allocate( Temp_1_col(local_aggre_size*NB*NB) )
    allocate( Temp_nd(NP) )


    Temp_agr = 0
    !C-- RA
    finish_flag = .true.
    do_whl_cnt = 0    
    DO while( finish_flag  )
       finish_flag = .false.
       do_whl_cnt = do_whl_cnt+1
       Temp_sz = 0


       DO col = 1,local_aggre_size
          if(Temp_agr(col) > 0) cycle
          cancel_flag = .false.
          cancel_sz = 0

          i = Temp_R_IN(col-1)+1          
          DO while( i <= Temp_R_IN(col))
             if(R % CN(i) > N ) exit
             row = R % CN(i)

!!$             R11 = R % V(i*9-8)
!!$             R21 = R % V(i*9-7)
!!$             R31 = R % V(i*9-6)
!!$             R12 = R % V(i*9-5)
!!$             R22 = R % V(i*9-4)
!!$             R32 = R % V(i*9-3)
!!$             R13 = R % V(i*9-2)
!!$             R23 = R % V(i*9-1)
!!$             R33 = R % V(i*9  )

             k = Temp_sz( row )

             logi = .false.
             if(k == 0) then
                logi = .true. 
             else if(k > 0)then
                if(Temp_int(row, k) /= col) logi = .true.
             end if

             if(logi) then
                if(k < Temp_SIZE .and. .not.cancel_flag) then
                   k = k+1
                   Temp_sz ( row ) = k
                   Temp_int(row,k) = col

                   Temp((row-1)*NB*NB+1:row*NB*NB,k)= 0.0
                   
                   call block_product_R_A(NB, R % V((i-1)*NB*NB+1:i*NB*NB), D(:,:,row), Temp((row-1)*NB*NB+1:row*NB*NB,k))

!!$                   DO l = 1, NB
!!$                      DO m1 = 1, NB
!!$                         DO m2 = 1, NB
!!$                            Temp((row-1)*NB*NB+(l-1)*NB+m1,k) = Temp((row-1)*NB*NB+(l-1)*NB+m1,k) + R % V((i-1)*NB*NB+(l-1)*NB+m2)*D(m2,m1,row)
!!$                         END DO
!!$                      END DO
!!$                   END DO




!!$                   M11 = R11*D(1,1,row)+R12*D(2,1,row)+R13*D(3,1,row)
!!$                   M12 = R11*D(1,2,row)+R12*D(2,2,row)+R13*D(3,2,row)
!!$                   M13 = R11*D(1,3,row)+R12*D(2,3,row)+R13*D(3,3,row)
!!$                   M21 = R21*D(1,1,row)+R22*D(2,1,row)+R23*D(3,1,row)
!!$                   M22 = R21*D(1,2,row)+R22*D(2,2,row)+R23*D(3,2,row)
!!$                   M23 = R21*D(1,3,row)+R22*D(2,3,row)+R23*D(3,3,row)
!!$                   M31 = R31*D(1,1,row)+R32*D(2,1,row)+R33*D(3,1,row)
!!$                   M32 = R31*D(1,2,row)+R32*D(2,2,row)+R33*D(3,2,row)
!!$                   M33 = R31*D(1,3,row)+R32*D(2,3,row)+R33*D(3,3,row)
!!$                   Temp(row*9-8,k) = M11
!!$                   Temp(row*9-7,k) = M12
!!$                   Temp(row*9-6,k) = M13
!!$                   Temp(row*9-5,k) = M21
!!$                   Temp(row*9-4,k) = M22
!!$                   Temp(row*9-3,k) = M23
!!$                   Temp(row*9-2,k) = M31
!!$                   Temp(row*9-1,k) = M32
!!$                   Temp(row*9,  k) = M33

!!$                   Temp    (row,k) = v * D(row)

                   cancel_sz = cancel_sz + 1
                   Temp_nd(cancel_sz) = row
                else 
                   cancel_flag = .true.
                end if
             else 

                call block_product_R_A(NB, R % V((i-1)*NB*NB+1:i*NB*NB), &
                     & D(:,:,row), Temp((row-1)*NB*NB+1:row*NB*NB,k))                

!!$                DO l = 1, NB
!!$                   DO m1 = 1, NB
!!$                      DO m2 = 1, NB
!!$                         Temp((row-1)*NB*NB+(l-1)*NB+m1,k) = Temp((row-1)*NB*NB+(l-1)*NB+m1,k) + R % V((i-1)*NB*NB+(l-1)*NB+m2)*D(m2,m1,row)
!!$                      END DO
!!$                   END DO
!!$                END DO


!!$                M11 = R11*D(1,1,row)+R12*D(2,1,row)+R13*D(3,1,row)
!!$                M12 = R11*D(1,2,row)+R12*D(2,2,row)+R13*D(3,2,row)
!!$                M13 = R11*D(1,3,row)+R12*D(2,3,row)+R13*D(3,3,row)
!!$                M21 = R21*D(1,1,row)+R22*D(2,1,row)+R23*D(3,1,row)
!!$                M22 = R21*D(1,2,row)+R22*D(2,2,row)+R23*D(3,2,row)
!!$                M23 = R21*D(1,3,row)+R22*D(2,3,row)+R23*D(3,3,row)
!!$                M31 = R31*D(1,1,row)+R32*D(2,1,row)+R33*D(3,1,row)
!!$                M32 = R31*D(1,2,row)+R32*D(2,2,row)+R33*D(3,2,row)
!!$                M33 = R31*D(1,3,row)+R32*D(2,3,row)+R33*D(3,3,row)
!!$                Temp(row*9-8,k) = Temp(row*9-8,k)+M11
!!$                Temp(row*9-7,k) = Temp(row*9-7,k)+M12
!!$                Temp(row*9-6,k) = Temp(row*9-6,k)+M13
!!$                Temp(row*9-5,k) = Temp(row*9-5,k)+M21
!!$                Temp(row*9-4,k) = Temp(row*9-4,k)+M22
!!$                Temp(row*9-3,k) = Temp(row*9-3,k)+M23
!!$                Temp(row*9-2,k) = Temp(row*9-2,k)+M31
!!$                Temp(row*9-1,k) = Temp(row*9-1,k)+M32
!!$                Temp(row*9,  k) = Temp(row*9,  k)+M33

!!$                Temp(row,k) = Temp(row,k) + v * D(row)
             end if

             if(.not.cancel_flag) then
                DO j = INL(row-1)+1,INL(row)
                   row2 = IAL( j )
                   k = Temp_sz( row2 )
                   logi=.false.
                   if(k==0) then
                      logi = .true. 
                   else if(k>0)then
                      if(Temp_int(row2,k) /= col) logi=.true.
                   end if
                   if(logi) then
                      if(k < Temp_SIZE .and. .not.cancel_flag) then
                         k = k+1
                         Temp_sz( row2 )  = k
                         Temp_int(row2,k) = col

                         Temp((row2-1)*NB*NB+1:row2*NB*NB, k) = 0.0

                         call block_product_R_A(NB, R % V((i-1)*NB*NB+1:i*NB*NB), AL(:,:,j), Temp((row2-1)*NB*NB+1:row2*NB*NB,k))

!!$                         DO l = 1, NB
!!$                            DO m1 = 1, NB
!!$                               DO m2 = 1, NB
!!$                                  Temp((row2-1)*NB*NB+(l-1)*NB+m1,k) = Temp((row2-1)*NB*NB+(l-1)*NB+m1,k) + R % V((i-1)*NB*NB+(l-1)*NB+m2)*AL(m2,m1,j)
!!$                               END DO
!!$                            END DO
!!$                         END DO


!!$                         M11 = R11*AL(1,1,j)+R12*AL(2,1,j)+R13*AL(3,1,j)
!!$                         M12 = R11*AL(1,2,j)+R12*AL(2,2,j)+R13*AL(3,2,j)
!!$                         M13 = R11*AL(1,3,j)+R12*AL(2,3,j)+R13*AL(3,3,j)
!!$                         M21 = R21*AL(1,1,j)+R22*AL(2,1,j)+R23*AL(3,1,j)
!!$                         M22 = R21*AL(1,2,j)+R22*AL(2,2,j)+R23*AL(3,2,j)
!!$                         M23 = R21*AL(1,3,j)+R22*AL(2,3,j)+R23*AL(3,3,j)
!!$                         M31 = R31*AL(1,1,j)+R32*AL(2,1,j)+R33*AL(3,1,j)
!!$                         M32 = R31*AL(1,2,j)+R32*AL(2,2,j)+R33*AL(3,2,j)
!!$                         M33 = R31*AL(1,3,j)+R32*AL(2,3,j)+R33*AL(3,3,j)
!!$                         Temp(row2*9-8,k) = M11
!!$                         Temp(row2*9-7,k) = M12
!!$                         Temp(row2*9-6,k) = M13
!!$                         Temp(row2*9-5,k) = M21
!!$                         Temp(row2*9-4,k) = M22
!!$                         Temp(row2*9-3,k) = M23
!!$                         Temp(row2*9-2,k) = M31
!!$                         Temp(row2*9-1,k) = M32
!!$                         Temp(row2*9,  k) = M33


!!$                         Temp    (row2,k) = v * AL(j)
                         cancel_sz = cancel_sz + 1
                         Temp_nd(cancel_sz) = row2
                      else 
                         cancel_flag = .true.
                      end if
                   else

                      call block_product_R_A(NB, R % V((i-1)*NB*NB+1:i*NB*NB), AL(:,:,j), Temp((row2-1)*NB*NB+1:row2*NB*NB,k))

!!$                      DO l = 1, NB
!!$                         DO m1 = 1, NB
!!$                            DO m2 = 1, NB
!!$                               Temp((row2-1)*NB*NB+(l-1)*NB+m1,k) = Temp((row2-1)*NB*NB+(l-1)*NB+m1,k) + R % V((i-1)*NB*NB+(l-1)*NB+m2)*AL(m2,m1,j)
!!$                            END DO
!!$                         END DO
!!$                      END DO


!!$                      M11 = R11*AL(1,1,j)+R12*AL(2,1,j)+R13*AL(3,1,j)
!!$                      M12 = R11*AL(1,2,j)+R12*AL(2,2,j)+R13*AL(3,2,j)
!!$                      M13 = R11*AL(1,3,j)+R12*AL(2,3,j)+R13*AL(3,3,j)
!!$                      M21 = R21*AL(1,1,j)+R22*AL(2,1,j)+R23*AL(3,1,j)
!!$                      M22 = R21*AL(1,2,j)+R22*AL(2,2,j)+R23*AL(3,2,j)
!!$                      M23 = R21*AL(1,3,j)+R22*AL(2,3,j)+R23*AL(3,3,j)
!!$                      M31 = R31*AL(1,1,j)+R32*AL(2,1,j)+R33*AL(3,1,j)
!!$                      M32 = R31*AL(1,2,j)+R32*AL(2,2,j)+R33*AL(3,2,j)
!!$                      M33 = R31*AL(1,3,j)+R32*AL(2,3,j)+R33*AL(3,3,j)
!!$                      Temp(row2*9-8,k) = Temp(row2*9-8,k)+M11
!!$                      Temp(row2*9-7,k) = Temp(row2*9-7,k)+M12
!!$                      Temp(row2*9-6,k) = Temp(row2*9-6,k)+M13
!!$                      Temp(row2*9-5,k) = Temp(row2*9-5,k)+M21
!!$                      Temp(row2*9-4,k) = Temp(row2*9-4,k)+M22
!!$                      Temp(row2*9-3,k) = Temp(row2*9-3,k)+M23
!!$                      Temp(row2*9-2,k) = Temp(row2*9-2,k)+M31
!!$                      Temp(row2*9-1,k) = Temp(row2*9-1,k)+M32
!!$                      Temp(row2*9,  k) = Temp(row2*9,  k)+M33

!!$                      Temp(row2,k) = Temp(row2,k) + v * AL(j)
                   end if
                END DO
             end if

             if(.not.cancel_flag) then
                DO j = INU(row-1)+1,INU(row)
                   row2 = IAU( j )
                   k = Temp_sz( row2 )
                   logi=.false.
                   if(k==0) then
                      logi = .true. 
                   else if(k>0)then
                      if(Temp_int(row2,k) /= col) logi=.true.
                   end if
                   if(logi) then
                      if(k < Temp_SIZE .and. .not.cancel_flag) then
                         k = k+1
                         Temp_sz( row2 ) = k
                         Temp_int(row2,k) = col

                         Temp((row2-1)*NB*NB+1:row2*NB*NB,k) = 0.0
                         call block_product_R_A(NB, R % V((i-1)*NB*NB+1:i*NB*NB), AU(:,:,j), Temp((row2-1)*NB*NB+1:row2*NB*NB,k))

!!$                         DO l = 1, NB
!!$                            DO m1 = 1, NB
!!$                               DO m2 = 1, NB
!!$                                  Temp((row2-1)*NB*NB+(l-1)*NB+m1,k) = Temp((row2-1)*NB*NB+(l-1)*NB+m1,k) + R % V((i-1)*NB*NB+(l-1)*NB+m2)*AU(m2,m1,j)
!!$                               END DO
!!$                            END DO
!!$                         END DO



!!$                         M11 = R11*AU(1,1,j)+R12*AU(2,1,j)+R13*AU(3,1,j)
!!$                         M12 = R11*AU(1,2,j)+R12*AU(2,2,j)+R13*AU(3,2,j)
!!$                         M13 = R11*AU(1,3,j)+R12*AU(2,3,j)+R13*AU(3,3,j)
!!$                         M21 = R21*AU(1,1,j)+R22*AU(2,1,j)+R23*AU(3,1,j)
!!$                         M22 = R21*AU(1,2,j)+R22*AU(2,2,j)+R23*AU(3,2,j)
!!$                         M23 = R21*AU(1,3,j)+R22*AU(2,3,j)+R23*AU(3,3,j)
!!$                         M31 = R31*AU(1,1,j)+R32*AU(2,1,j)+R33*AU(3,1,j)
!!$                         M32 = R31*AU(1,2,j)+R32*AU(2,2,j)+R33*AU(3,2,j)
!!$                         M33 = R31*AU(1,3,j)+R32*AU(2,3,j)+R33*AU(3,3,j)
!!$                         Temp(row2*9-8,k) = M11
!!$                         Temp(row2*9-7,k) = M12
!!$                         Temp(row2*9-6,k) = M13
!!$                         Temp(row2*9-5,k) = M21
!!$                         Temp(row2*9-4,k) = M22
!!$                         Temp(row2*9-3,k) = M23
!!$                         Temp(row2*9-2,k) = M31
!!$                         Temp(row2*9-1,k) = M32
!!$                         Temp(row2*9,  k) = M33

!!$                         Temp    (row2,k) = v * AU(j)

                         cancel_sz = cancel_sz + 1
                         Temp_nd(cancel_sz) = row2
                      else 
                         cancel_flag = .true.
                      end if
                   else

                      call block_product_R_A(NB, R % V((i-1)*NB*NB+1:i*NB*NB), AU(:,:,j), Temp((row2-1)*NB*NB+1:row2*NB*NB,k))

!!$                      DO l = 1, NB
!!$                         DO m1 = 1, NB
!!$                            DO m2 = 1, NB
!!$                               Temp((row2-1)*NB*NB+(l-1)*NB+m1,k) = Temp((row2-1)*NB*NB+(l-1)*NB+m1,k) + R % V((i-1)*NB*NB+(l-1)*NB+m2)*AU(m2,m1,j)
!!$                            END DO
!!$                         END DO
!!$                      END DO


!!$                      M11 = R11*AU(1,1,j)+R12*AU(2,1,j)+R13*AU(3,1,j)
!!$                      M12 = R11*AU(1,2,j)+R12*AU(2,2,j)+R13*AU(3,2,j)
!!$                      M13 = R11*AU(1,3,j)+R12*AU(2,3,j)+R13*AU(3,3,j)
!!$                      M21 = R21*AU(1,1,j)+R22*AU(2,1,j)+R23*AU(3,1,j)
!!$                      M22 = R21*AU(1,2,j)+R22*AU(2,2,j)+R23*AU(3,2,j)
!!$                      M23 = R21*AU(1,3,j)+R22*AU(2,3,j)+R23*AU(3,3,j)
!!$                      M31 = R31*AU(1,1,j)+R32*AU(2,1,j)+R33*AU(3,1,j)
!!$                      M32 = R31*AU(1,2,j)+R32*AU(2,2,j)+R33*AU(3,2,j)
!!$                      M33 = R31*AU(1,3,j)+R32*AU(2,3,j)+R33*AU(3,3,j)
!!$                      Temp(row2*9-8,k) = Temp(row2*9-8,k)+M11
!!$                      Temp(row2*9-7,k) = Temp(row2*9-7,k)+M12
!!$                      Temp(row2*9-6,k) = Temp(row2*9-6,k)+M13
!!$                      Temp(row2*9-5,k) = Temp(row2*9-5,k)+M21
!!$                      Temp(row2*9-4,k) = Temp(row2*9-4,k)+M22
!!$                      Temp(row2*9-3,k) = Temp(row2*9-3,k)+M23
!!$                      Temp(row2*9-2,k) = Temp(row2*9-2,k)+M31
!!$                      Temp(row2*9-1,k) = Temp(row2*9-1,k)+M32
!!$                      Temp(row2*9,  k) = Temp(row2*9,  k)+M33

!!$                      Temp(row2,k) = Temp(row2,k) + v * AU(j)
                   end if
                END DO
             end if
             i=i+1
          END DO


          finish_flag = finish_flag .or. cancel_flag

          !C cancel and regist the aggregate.
          !C maybe more effective cancel process by recording 
          !C the selected nodes 
          if(cancel_flag) then
             DO i = 1,cancel_sz
                j = Temp_nd( i )
                Temp_sz( j ) = Temp_sz( j )-1 
             END DO
          else
             Temp_agr(col) = 1
          end if
       END DO
       !C-- end: RA

       !C-- temp * P
       DO col = 1,R % ROW_SIZE
          Temp_1_col = 0      

          js = Temp_R_IN(col-1)+1
          je = Temp_R_IN(col)
          DO j = js, je
             k = R % CN(j)

!!$             R11 = R % V(j*9-8)
!!$             R21 = R % V(j*9-7)
!!$             R31 = R % V(j*9-6)
!!$             R12 = R % V(j*9-5)
!!$             R22 = R % V(j*9-4)
!!$             R32 = R % V(j*9-3)
!!$             R13 = R % V(j*9-2)
!!$             R23 = R % V(j*9-1)
!!$             R33 = R % V(j*9  )

!!$             v = R % V(j)
             DO l = 1, Temp_sz(k)
                row = Temp_int(k,l)
!!$                Temp_1_col(row) = Temp_1_col(row) + v * Temp(k,l)

                call block_product_Tmp_P(NB, Temp((k-1)*NB*NB+1:k*NB*NB,l), &
                     & R % V((j-1)*NB*NB+1:j*NB*NB), Temp_1_col((row-1)*NB*NB+1:row*NB*NB))                
!!$                DO l1 = 1, NB
!!$                   DO m1 = 1, NB
!!$                      DO m2 = 1, NB
!!$                         Temp_1_col((row-1)*NB*NB+(l1-1)*NB+m1) = Temp_1_col((row-1)*NB*NB+(l1-1)*NB+m1) + Temp((k-1)*NB*NB+(l1-1)*NB+m2,l)*R % V((j-1)*NB*NB+(m1-1)*NB+m2)
!!$                      END DO
!!$                   END DO
!!$                END DO


!!$                M11 = Temp(k*9-8,l)*R11+Temp(k*9-7,l)*R12+Temp(k*9-6,l)*R13
!!$                M12 = Temp(k*9-8,l)*R21+Temp(k*9-7,l)*R22+Temp(k*9-6,l)*R23
!!$                M13 = Temp(k*9-8,l)*R31+Temp(k*9-7,l)*R32+Temp(k*9-6,l)*R33
!!$                M21 = Temp(k*9-5,l)*R11+Temp(k*9-4,l)*R12+Temp(k*9-3,l)*R13
!!$                M22 = Temp(k*9-5,l)*R21+Temp(k*9-4,l)*R22+Temp(k*9-3,l)*R23
!!$                M23 = Temp(k*9-5,l)*R31+Temp(k*9-4,l)*R32+Temp(k*9-3,l)*R33
!!$                M31 = Temp(k*9-2,l)*R11+Temp(k*9-1,l)*R12+Temp(k*9  ,l)*R13
!!$                M32 = Temp(k*9-2,l)*R21+Temp(k*9-1,l)*R22+Temp(k*9  ,l)*R23
!!$                M33 = Temp(k*9-2,l)*R31+Temp(k*9-1,l)*R32+Temp(k*9  ,l)*R33
!!$                
!!$                Temp_1_col(9*row-8) = Temp_1_col(9*row-8)+M11
!!$                Temp_1_col(9*row-7) = Temp_1_col(9*row-7)+M12
!!$                Temp_1_col(9*row-6) = Temp_1_col(9*row-6)+M13
!!$                Temp_1_col(9*row-5) = Temp_1_col(9*row-5)+M21
!!$                Temp_1_col(9*row-4) = Temp_1_col(9*row-4)+M22
!!$                Temp_1_col(9*row-3) = Temp_1_col(9*row-3)+M23
!!$                Temp_1_col(9*row-2) = Temp_1_col(9*row-2)+M31
!!$                Temp_1_col(9*row-1) = Temp_1_col(9*row-1)+M32
!!$                Temp_1_col(9*row  ) = Temp_1_col(9*row  )+M33
             END DO
          END DO


          !C-- input
          DO row = 1,local_aggre_size
!!$             if( &
!!$                  & ( Temp_1_col(9*row-8) == 0 .and.  &
!!$                  &   Temp_1_col(9*row-7) == 0 .and.  &
!!$                  &   Temp_1_col(9*row-6) == 0 .and.  &
!!$                  &   Temp_1_col(9*row-5) == 0 .and.  &
!!$                  &   Temp_1_col(9*row-4) == 0 .and.  &
!!$                  &   Temp_1_col(9*row-3) == 0 .and.  &
!!$                  &   Temp_1_col(9*row-2) == 0 .and.  &
!!$                  &   Temp_1_col(9*row-1) == 0 .and.  &
!!$                  &   Temp_1_col(9*row  ) == 0 ) .or. &
!!$                  & (col < row .and. row <= ownaggre_size)) cycle
             if(col < row .and. row <= ownaggre_size) cycle

             v = 0.0
             do l = 1, NB*NB
                v = v+Temp_1_col(NB*NB*(row-1)+l)**2
             end do

             if(v == 0) cycle


             hash_no = mod(col,hash_size)+1
             count = hash_no
             DO while(count <= hash_size)
                if(Temp_CN(count,row) == 0) exit
                count = count+1
             end DO
             if(count > hash_size) then
                count = 1
                DO while( count < hash_no .and. Temp_CN(count,row) /= 0 )
                   count = count+1
                end DO
                if(count >= hash_no) stop "hash overflow.."
             end if

             !C assign the value
             Temp_N(row) = Temp_N(row)+1
             Temp_CN(count,row) = col

             DO l = 1, NB*NB
                Temp_V((count-1)*NB*NB+l,row) = Temp_1_col((row-1)*NB*NB+l)
             END DO
!!$             Temp_V(count*9-8,row)  = Temp_1_col(row*9-8)
!!$             Temp_V(count*9-7,row)  = Temp_1_col(row*9-7)
!!$             Temp_V(count*9-6,row)  = Temp_1_col(row*9-6)
!!$             Temp_V(count*9-5,row)  = Temp_1_col(row*9-5)
!!$             Temp_V(count*9-4,row)  = Temp_1_col(row*9-4)
!!$             Temp_V(count*9-3,row)  = Temp_1_col(row*9-3)
!!$             Temp_V(count*9-2,row)  = Temp_1_col(row*9-2)
!!$             Temp_V(count*9-1,row)  = Temp_1_col(row*9-1)
!!$             Temp_V(count*9,  row)  = Temp_1_col(row*9  )
          END DO
          !C-- end: input

       END DO
       !C-- end:temp * P
    END DO


    deallocate( Temp_nd    )
    deallocate( Temp_sz    )
    deallocate( Temp_int   )
    deallocate( Temp       )
    deallocate( Temp_agr   )
    deallocate( Temp_1_col )

  contains
    subroutine block_product_R_A(NB, int_lev, block_mat, out)
      implicit none
      integer(kind=kint) :: l, m1, m2
      integer(kind=kint) :: NB
      real(kind=kreal)   :: int_lev(:), block_mat(:,:), out(:)

      DO l = 1, NB
         DO m1 = 1, NB
            DO m2 = 1, NB
               out((l-1)*NB+m1) = out((l-1)*NB+m1) + int_lev((l-1)*NB+m2)*block_mat(m2,m1)
            END DO
         END DO
      END DO

    end subroutine block_product_R_A

    !C int_lev_t get Restriction operator.
    !C int_lev_t is transposed in calculation of matrix product.
    subroutine block_product_Tmp_P(NB, temp, int_lev_t, out)
      implicit none
      integer(kind=kint) :: l, m1, m2
      integer(kind=kint) :: NB
      real(kind=kreal)   :: int_lev_t(:), temp(:), out(:)

      DO l = 1, NB
         DO m1 = 1, NB
            DO m2 = 1, NB
               out((l-1)*NB+m1) = out((l-1)*NB+m1) + temp((l-1)*NB+m2)*int_lev_t((m1-1)*NB+m2)
            END DO
         END DO
      END DO

    end subroutine block_product_Tmp_P

  END SUBROUTINE RAP_NB_old


  SUBROUTINE RAP_NB(Temp_N,Temp_CN,Temp_V,hash_size,N,ownaggre_size,LEVEL_NO, &
       & local_aggre_size, Temp_R_IN, NB, NK)
    USE hash_mod
    USE data_structure_for_AMG_NB
    !$ USE omp_lib
    
    IMPLICIT NONE

    INTEGER(kind=kint),intent(in)   :: hash_size,N,ownaggre_size,LEVEL_NO,local_aggre_size
    INTEGER(kind=kint),intent(in)   :: Temp_R_IN(0:), NB, NK
!!$    INTEGER(kind=kint),intent(inout):: Temp_N(local_aggre_size)
!!$    INTEGER(kind=kint),intent(inout):: Temp_CN(hash_size,local_aggre_size)
    INTEGER(kind=kint),intent(inout):: Temp_N(:)
    INTEGER(kind=kint),intent(inout):: Temp_CN(:,:)
    type(INTER_LEVEL_OPERATOR),pointer :: P, R

    REAL(kind=kreal),intent(inout):: Temp_V(:,:)

    !C pointers for matrix
    REAL   (kind=kreal),POINTER ::  D(:,:,:)
    REAL   (kind=kreal),POINTER ::  AU(:,:,:)
    REAL   (kind=kreal),POINTER ::  AL(:,:,:)
    INTEGER(kind=kint ),POINTER ::  INU(:)
    INTEGER(kind=kint ),POINTER ::  IAU(:)
    INTEGER(kind=kint ),POINTER ::  INL(:)
    INTEGER(kind=kint ),POINTER ::  IAL(:)

    INTEGER(kind=kint ),allocatable :: Temp_agr(:),Temp_sz(:,:),Temp_nd(:,:)
    INTEGER(kind=kint ),allocatable :: Temp_int(:,:), nd_sz(:), agr_num_list(:)
    REAL   (kind=kreal),allocatable :: Temp(:,:),Temp_1_col(:)

    integer(kind=kint) ::i,j,k,l,m,rowins,rowine,colins,coline,row,col,row2
    integer(kind=kint) ::do_whl_cnt,is,ie 
    integer(kind=kint) ::js,je,jin,rowin,colin,count,hash_no,NP, Temp_SIZE
    logical :: cancel_flag,finish_flag,logi, flag

    INTEGER(kind=kint ) :: cancel_sz
    REAL   (kind=kreal) :: v
    
    INTEGER(kind=kint) :: tsz, tid, tstart, tend, q, tidp1, tidNP, agr_num, agr_num_index
    
!!$    NB =  HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NB     
!!$    NK =  HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NK     
    NP   =  HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NP 
    D    => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % D 
    INL  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INL  
    INU  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INU 
    IAL  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAL
    IAU  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAU
    AL   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AL
    AU   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AU
    P    => HIERARCHICAL_DATA_NB(LEVEL_NO  ) % P
    R    => HIERARCHICAL_DATA_NB(LEVEL_NO  ) % R

    !C Calculation of RAP. This uses R information. 
    !C R % IN must be rewrite @ the latter part.
    !C So Temp_R_IN is used for R % IN(:).
    !C ------

    tsz = 1
    tid = 0
    !$ tsz = OMP_GET_MAX_THREADS()

    Temp_SIZE = 1

    allocate(Temp_sz (NP, tsz)    )
    allocate(Temp_int(Temp_SIZE, NP*tsz) )
    allocate(Temp    (Temp_SIZE*NK*NB, NP*tsz) )
    allocate(Temp_agr(local_aggre_size) )
    allocate(Temp_1_col(local_aggre_size*NK*NK) )
    allocate(Temp_nd(NP, tsz) )
    allocate(nd_sz(tsz) )
    allocate(agr_num_list(local_aggre_size))

    !C set agr_num_list
    agr_num=local_aggre_size
    !$omp parallel do
    do i = 1, local_aggre_size
       agr_num_list(i) = i
       Temp_agr(i) = 0
    end do
    !$omp end parallel do 
    
    !$omp parallel do
    do i = 1, local_aggre_size*NK*NK
       Temp_1_col(i) = 0 
    end do
    !$omp end parallel do 



    !C-- RA
    finish_flag = .true.
    do_whl_cnt = 0
    DO while( finish_flag  )
       finish_flag = .false.
       do_whl_cnt = do_whl_cnt+1

       tstart = 1
       tend = agr_num

       
!C -- parallel for each aggregate block --        
       !$omp  parallel private(agr_num_index,row,row2,col,cancel_flag,  &
       !$omp& cancel_sz,flag,i,j,k,l,m,js,je,tstart,tend,tid,v,hash_no, &
       !$omp& count,tidp1,tidNP,tsz) reduction(.or.:finish_flag)
       !$ tid = OMP_GET_THREAD_NUM()
       !$ tsz = OMP_GET_NUM_THREADS()
       !$ tstart = tid*agr_num/tsz+1; tend=(tid+1)*agr_num/tsz
       tidp1 = tid+1
       tidNP = tid*NP
       Temp_sz(:,tidp1) = 0
       
       DO agr_num_index = tstart, tend
          col = agr_num_list(agr_num_index)
          cancel_flag = .false.
          cancel_sz = 0

          DO i = Temp_R_IN(col-1)+1, Temp_R_IN(col)
             row = R % CN(i) 
             if(row <= N .AND. Temp_sz(row, tidp1) >= Temp_SIZE) then
                cancel_flag = .TRUE.
                exit
             end if
          END DO
          if(.not.cancel_flag) then
             DO i = Temp_R_IN(col-1)+1, Temp_R_IN(col)
                row = R % CN(i)
                if(row <= N) then
                   DO j = INL(row-1)+1, INL(row)
                      row2 = IAL(j)
                      if(Temp_sz(row2, tidp1) >= Temp_SIZE) then
                         cancel_flag = .true.
                         exit
                      end if
                   END DO
                   if(.not.cancel_flag) then
                      DO j = INU(row-1)+1, INU(row)
                         row2 = IAU(j)
                         if(Temp_sz(row2, tidp1) >= Temp_SIZE) then
                            cancel_flag = .true.
                            exit
                         end if
                      END DO
                   end if
                end if
             END DO
          end if

          finish_flag = finish_flag .or. cancel_flag          
          if(cancel_flag) cycle
          Temp_agr(col) = 1

          
          DO i = Temp_R_IN(col-1)+1, Temp_R_IN(col)
             row = R % CN(i)
             !C v   = R % V(i)
             if(row <= N ) then
                k = Temp_sz( row, tidp1 )

                flag = (k==0)
                if(k > 0)then
                   if(Temp_int(k, tidNP+row) /= col) flag = .true.
                end if
                
                if(flag) then
                   k = k+1
                   Temp_sz ( row, tidp1 )  = k
                   Temp_int(k, tidNP+row) = col
                   Temp((k-1)*NK*NB+1:k*NK*NB, tidNP+row) = 0.0
                   
                end if
                
                call block_product_R_A(NB, NK, R % V((i-1)*NK*NB+1:i*NK*NB), D(:,:,row), Temp((k-1)*NK*NB+1:k*NK*NB,tidNP+row))
                
                DO j = INL(row-1)+1,INL(row)
                   row2 = IAL( j )
                   if(row2 > N) CYCLE
                   k = Temp_sz( row2, tidp1 )
                   
                   flag = (k==0)
                   if(k > 0) then
                      if(Temp_int(k, tidNP+row2) /= col) flag = .true.
                   end if
                   
                   if(flag) then
                      k = k+1
                      Temp_sz( row2, tidp1 )  = k
                      Temp_int(k,tidNP+row2) = col
                      Temp    ((k-1)*NK*NB+1:k*NK*NB,tidNP+row2) = 0.0

                   end if
                   
                   call block_product_R_A(NB, NK, R % V((i-1)*NK*NB+1:i*NK*NB), AL(:,:,j), Temp((k-1)*NK*NB+1:k*NK*NB,tidNP+row2))

                   !C Temp(k,tidNP+row2) = Temp(k,tidNP+row2) + v * AL(j)

                END DO
                
                
                DO j = INU(row-1)+1,INU(row)
                   row2 = IAU( j )
                   k = Temp_sz( row2, tidp1 )

                   flag = (k==0)
                   if(k>0) then
                      if(Temp_int(k,tidNP+row2) /= col) flag = .true.
                   end if

                   if(flag) then
                      k = k+1
                      Temp_sz( row2, tidp1 ) = k
                      Temp_int(k,tidNP+row2) = col
!C                      Temp    (k,tidNP+row2) = v * AU(j)
                      Temp    ((k-1)*NK*NB+1:k*NK*NB,tidNP+row2) = 0.0
                   end if

                   call block_product_R_A(NB, NK, R % V((i-1)*NK*NB+1:i*NK*NB), &
                        & AU(:,:,j), Temp((k-1)*NK*NB+1:k*NK*NB,tidNP+row2))                   
!C                      Temp(k,tidNP+row2) = Temp(k,tidNP+row2) + v * AU(j)

                END DO
             end if
                
          END DO
          
       END DO
       !C-- end: RA

       !C-- temp * P
       DO col = 1,R % ROW_SIZE
          nd_sz(tidp1) = 0
          
          js = Temp_R_IN(col-1)+1
          je = Temp_R_IN(col)
          DO j = js, je
             k = R % CN(j)
!C             v = R % V(j)

             DO l = 1, Temp_sz(k,tidp1)
                row = Temp_int(l,tidNP+k)
                DO m = 1, nd_sz(tidp1)
                   if(Temp_nd(m,tidp1) == row) exit
                end DO
                
                if(m > nd_sz(tidp1)) then
                   nd_sz(tidp1) = nd_sz(tidp1) + 1
                   Temp_nd(nd_sz(tidp1),tidp1) = row
                end if
                
                call block_product_Tmp_P(NB, NK, Temp((l-1)*NK*NB+1:l*NK*NB,tidNP+k), &
                     & R % V((j-1)*NK*NB+1:j*NK*NB), Temp_1_col((row-1)*NK*NK+1:row*NK*NK))
!C                Temp_1_col(row) = Temp_1_col(row) + v * Temp(l,tidNP+k)
             END DO
          END DO
          

          !C-- input
          DO j = 1, nd_sz(tidp1)
             row = Temp_nd(j,tidp1)

             !C search (row col) in new matrix
             CALL hash(hash_size, Temp_CN(:, row), col, count, "NONZEROS_PER_ROW")

!!$             hash_no = mod(col,hash_size)+1
!!$             count = hash_no
!!$             DO while(count <= hash_size)
!!$                if(Temp_CN(count,row) == 0) exit
!!$                count = count+1
!!$             end DO
!!$             if(count > hash_size) then
!!$                count = 1
!!$                DO while( count < hash_no .and. Temp_CN(count,row) /= 0 )
!!$                   count = count+1
!!$                end DO
!!$                if(count >= hash_no) stop "hash overflow.."
!!$             end if

             !C assign the value
             Temp_N(row) = Temp_N(row)+1
             Temp_CN(count,row) = col

             !C                Temp_V(count,row)  = Temp_1_col(row)             
             DO l=1, NK*NK
                Temp_V((count-1)*NK*NK+l,row) = Temp_1_col((row-1)*NK*NK+l)
             END DO
             !C             end if
             
             Temp_1_col((row-1)*NK*NK+1:row*NK*NK)= 0
          END DO
          !C-- end: input
          
       END DO
       !C-- end:temp * P
       
       !$omp end parallel
       
       !C update agr_num, agr_num_liist
       if(finish_flag) then
          j =0
          DO i = 1, agr_num
             if(Temp_agr(agr_num_list(i)) == 0) then
                j=j+1
                agr_num_list(j) = agr_num_list(i)
             end if
          END DO
          agr_num = j
       end if
       
    END DO
    


    deallocate( Temp_nd    )
    deallocate( Temp_sz    )
    deallocate( Temp_int   )
    deallocate( Temp       )
    deallocate( Temp_agr   )
    deallocate( Temp_1_col )
    deallocate( nd_sz      )

  contains
    subroutine block_product_R_A(NB, NK, int_lev, block_mat, out)
      implicit none
      integer(kind=kint) :: l, m1, m2
      integer(kind=kint) :: NB, NK
      real(kind=kreal)   :: int_lev(:), block_mat(:,:), out(:)

      DO l = 1, NK
         DO m1 = 1, NB
            DO m2 = 1, NB
               out((l-1)*NB+m1) = out((l-1)*NB+m1) + int_lev((l-1)*NB+m2)*block_mat(m2,m1)
            END DO
         END DO
      END DO
      
    end subroutine block_product_R_A
    
    !C int_lev_t get Restriction operator.
    !C int_lev_t is transposed in calculation of matrix product.
    subroutine block_product_Tmp_P(NB, NK, temp, int_lev_t, out)
      implicit none
      integer(kind=kint) :: l, m1, m2
      integer(kind=kint) :: NB, NK
      real(kind=kreal)   :: int_lev_t(:), temp(:), out(:)

      DO l = 1, NK
         DO m1 = 1, NK
            DO m2 = 1, NB
               out((l-1)*NK+m1) = out((l-1)*NK+m1) + temp((l-1)*NB+m2)*int_lev_t((m1-1)*NB+m2)
            END DO
         END DO
      END DO

    end subroutine block_product_Tmp_P

  END SUBROUTINE RAP_NB



  SUBROUTINE make_rap(LEVEL_NO,my_rank,SOLVER_COMM,NPROCS,          &
       & coarser_level_size,global_coarser_size,global_finish_flag, &
       & GIN_aggregate,PE_list_size,PE_list,aggregate_table_size,   &
       & aggregate_table_array,aggregate_number_in_table,           &
       & hash_table_size,global_local_hash_table,local_aggre_size,  &
       & Temp_R_IN, NB, NK, PRO_AGGRE_OPT, coarser_lev_N_bf_CGA)

    USE data_structure_for_AMG_NB
    USE count_time_mod
    USE comm_subs_NB
    USE solver_sr_NB
    USE amg_parameter, only: W_OUT_LU
    USE int_vec_module, &
         & only: int_vec, int_vec_initialize, check_and_extend, deallocate_int_vec
    USE isort, only: qsorti



    IMPLICIT NONE
    INCLUDE 'mpif.h'

    INTEGER(kind=kint), INTENT(in) :: LEVEL_NO,global_coarser_size
    INTEGER(kind=kint), INTENT(in) :: coarser_level_size, coarser_lev_N_bf_CGA
    LOGICAL,            INTENT(in) :: global_finish_flag, PRO_AGGRE_OPT

    !C temporary space of RAP
    REAL   (kind=kreal), DIMENSION(:,:),ALLOCATABLE :: Temp_V
    INTEGER(kind=kint ), DIMENSION(:,:),ALLOCATABLE :: Temp_CN
    INTEGER(kind=kint ), DIMENSION(:),  ALLOCATABLE :: Temp_N

    INTEGER(kind=kint ), DIMENSION(0:) :: Temp_R_IN    


    !C pointers for matrix
    REAL   (kind=kreal),POINTER, DIMENSION(:,:,:) ::  newD,newAU,newAL,newALU
    INTEGER(kind=kint ),POINTER, DIMENSION(:) ::  newINU,newIAU,newINL,newIAL

    !C communication data structures
    INTEGER                               , INTENT(in) :: NPROCS,SOLVER_COMM
    INTEGER(kind=kint )                   , INTENT(in) :: my_rank

    INTEGER(kind=kint )          :: NEIBPETOT
    INTEGER(kind=kint ), POINTER :: NEIBPE      (:)
    INTEGER(kind=kint ), POINTER :: NOD_IMPORT  (:),STACK_IMPORT(:)
    INTEGER(kind=kint ), POINTER :: NOD_EXPORT  (:),STACK_EXPORT(:)
    INTEGER                      :: ierr

    REAL   (kind=kreal), DIMENSION(:,:),POINTER :: Temp_P
    INTEGER(kind=kint ), DIMENSION(:,:),POINTER :: Temp_PI
    INTEGER(kind=kint ), DIMENSION(:),  POINTER :: Temp_P_row_size 

    !C GIN_aggregate is used as rowdistbx on toplevel.
!!$    INTEGER(kind=kint ), INTENT(in)  :: GIN_aggregate(0:NPROCS)
    INTEGER(kind=kint ), INTENT(in)  :: GIN_aggregate(0:)

    INTEGER(kind=kint ), DIMENSION(:),   ALLOCATABLE   :: node_aggre_index_receive(:)
    INTEGER(kind=kint ), DIMENSION(2)                  :: node_aggre_size_for_PE

    INTEGER(kind=kint ), DIMENSION(:),   ALLOCATABLE   :: repeated_nodes_send_buf      
    INTEGER(kind=kint ), DIMENSION(:),   POINTER       :: in_repeated_send_buf      

    INTEGER(kind=kint ) :: N,NP,i,j,k,l,m,g,j_ex,j_im, d_i, n_max, js, je
    INTEGER(kind=kint ) :: REPEATED_NODE_SIZE,n_lower,n_upper,repeated_nodes_tot
    INTEGER(kind=kint ) :: index,row,col,colins,coline,colin,count,Temp_COLSIZE

    REAL   (kind=kreal), ALLOCATABLE :: WS(:), WR(:)
    REAL   (kind=kreal), ALLOCATABLE :: WW(:,:)

    INTEGER(kind=kint ) :: PE_num_size
    INTEGER(kind=kint ), DIMENSION(:),   POINTER :: PE_nums
    INTEGER(kind=kint ), DIMENSION(:),   POINTER :: nodes_for_each_PE
    INTEGER(kind=kint ), DIMENSION(:),   POINTER :: in_nodes_for_each_PE

!!$    INTEGER(kind=kint ), INTENT(inout) :: PE_list(NPROCS,3)
    INTEGER(kind=kint ), INTENT(inout) :: PE_list(:,:)
    INTEGER(kind=kint ), INTENT(inout) :: PE_list_size 

    !C - begin: hash_table
    INTEGER(kind=kint), INTENT(in)    :: hash_table_size
!!$    INTEGER(kind=kint), INTENT(inout) :: global_local_hash_table(hash_table_size,2)
    INTEGER(kind=kint), INTENT(inout) :: global_local_hash_table(:,:)
    !C - end: hash_table

    !C - begin: aggregate_table_array
    INTEGER(kind=kint), INTENT(in)    :: aggregate_table_size
    !C two arguments: communication table,each PE
    type(int_vec), pointer :: aggregate_table_array(:)
    !C argument: the number of aggregates
!!$    INTEGER(kind=kint), INTENT(inout) :: aggregate_number_in_table(NPROCS)
    INTEGER(kind=kint), INTENT(inout) :: aggregate_number_in_table(:)
    !C - end: aggregate_table_array

    INTEGER(kind=kint) :: local_aggre_size

    INTEGER(kind=kint) :: newNEIBPETOT, newNP, newN, NB, NK
    INTEGER(kind=kint),POINTER,DIMENSION(:) :: newNOD_IMPORT,newNOD_EXPORT
    INTEGER(kind=kint),POINTER,DIMENSION(:) :: newSTACK_IMPORT,newSTACK_EXPORT
    INTEGER(kind=kint),POINTER,DIMENSION(:) :: newNEIBPE

    INTEGER(kind=kint),ALLOCATABLE :: export_global_buffer(:)
    INTEGER(kind=kint) :: idx_PE_INT,idx_NOD_INT,idx_PE,idx_NOD
    INTEGER(kind=kint) :: newNEIBPETOT_INT
    INTEGER(kind=kint),POINTER,DIMENSION(:) :: newNOD_IMPORT_INT,newNOD_EXPORT_INT
    INTEGER(kind=kint),POINTER,DIMENSION(:) :: newSTACK_IMPORT_INT,newSTACK_EXPORT_INT
    INTEGER(kind=kint),POINTER,DIMENSION(:) :: newNEIBPE_INT
    INTEGER(kind=kint),ALLOCATABLE :: export_int_lvl_buffer(:)
    INTEGER(kind=kint) :: m1,m2

    INTEGER(kind=kint), ALLOCATABLE :: comm_recv_node(:), tmp_int(:)
    REAL   (kind=kreal), ALLOCATABLE :: vec(:)

    REAL   (kind=kreal) ::  val
    TYPE(INTER_LEVEL_OPERATOR), POINTER :: P,R

    INTEGER(kind=kint) :: rin,trin
    integer :: INFO


    N=HIERARCHICAL_DATA_NB(LEVEL_NO-1) % N 
    NP=HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NP 

    NEIBPE       => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % NEIBPE
    STACK_IMPORT => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % STACK_IMPORT
    NOD_IMPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % NOD_IMPORT
    STACK_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % STACK_EXPORT
    NOD_EXPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % NOD_EXPORT
    NEIBPETOT    =  HIERARCHICAL_DATA_NB(LEVEL_NO-1) % COMM_TABLE % NEIBPETOT     

    !C the number of non-zero elements in one row of RAP = TEMP_COLSIZE
    !C new matrix rap's allocation. 
    TEMP_COLSIZE=(1001+501*(LEVEL_NO-2))
    
    IF(TEMP_COLSIZE > global_coarser_size) TEMP_COLSIZE=global_coarser_size
    ALLOCATE(Temp_N(local_aggre_size))
    ALLOCATE(Temp_V(TEMP_COLSIZE*NK*NK,local_aggre_size))
    ALLOCATE(Temp_CN(TEMP_COLSIZE,local_aggre_size))

    !$omp parallel do 
    do i = 1, local_aggre_size
       Temp_N(i)=0
       Temp_CN(1:TEMP_COLSIZE,i)=0
       Temp_V(1:TEMP_COLSIZE*NK*NK,i)=0
    end do
    !$omp end parallel do 
    
    CALL count_time(1,solver_comm,my_rank,4)       
    
    CALL RAP_NB(Temp_N,Temp_CN,Temp_V,TEMP_COLSIZE,N,coarser_level_size, &
         &     LEVEL_NO,local_aggre_size,Temp_R_IN, NB, NK)

    CALL count_time(2,solver_comm,my_rank,4)       


    DO i=1,local_aggre_size
       IF(Temp_N(i) > TEMP_COLSIZE) STOP 'TEMP_COL_SIZE should be enlarged: RAP'
    END DO


#ifdef DEBUG
    CALL MPI_Barrier(solver_comm,ierr)
    IF (my_rank==0) THEN
       WRITE(*,*) "### sr_elements         ###"
    END IF
    CALL MPI_Barrier(solver_comm,ierr)
#endif


    !C send the not-own aggregates and receive own aggregates in matrix(Temp_CN,_N,_V)
    CALL count_time(1,solver_comm,my_rank,3)       

    CALL solver_SR_elements &
         &    ( SOLVER_COMM,my_rank,GIN_aggregate,NPROCS,hash_table_size,          &
         &      global_local_hash_table,aggregate_table_size,aggregate_table_array,&
         &      aggregate_number_in_table,local_aggre_size,PE_list,PE_list_size,   &
         &      Temp_CN,Temp_N,Temp_V,TEMP_COLSIZE, NK) 

    CALL count_time(2,solver_comm,my_rank,3)       

    DO i=1,coarser_level_size
       IF(Temp_N(i) > TEMP_COLSIZE) THEN
          STOP 'Temp_col_size should be enlarged: sr_elements'
       END IF
    END DO



    !C copy Temp matrix to correct location
    !C ------
!!$    if(global_finish_flag)then
!!$       if(my_rank==0) write(*,*) "in"
!!$       call MPI_Barrier(solver_comm,ierr)
!!$
!!$
!!$       CALL make_PSPSE_MAT(Temp_CN,Temp_N,Temp_V,TEMP_COLSIZE,GIN_aggregate,&
!!$            & aggregate_table_array,aggregate_number_in_table,PE_list_size,PE_list,&
!!$            & local_aggre_size,R % ROW_SIZE,GIN_aggregate(my_rank),SOLVER_COMM)
!!$
!!$       if(my_rank==0) write(*,*) "out"
!!$       call MPI_Barrier(solver_comm,ierr)
!!$
!!$
!!$       allocate(rowdistbx(0:NPROCS))
!!$       rowdistbx=GIN_aggregate
!!$       HIERARCHICAL_DATA_NB(LEVEL_NO) % X=>coarsest_X(:,1)
!!$       HIERARCHICAL_DATA_NB(LEVEL_NO) % B=>coarsest_B(:,1)

!!$    else
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % INU(0:local_aggre_size))
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % INL(0:local_aggre_size))
    newINL=>HIERARCHICAL_DATA_NB(LEVEL_NO) % INL 
    newINU=>HIERARCHICAL_DATA_NB(LEVEL_NO) % INU 

    !$omp parallel do 
    do i = 0, local_aggre_size
       newINL(i)=0
       newINU(i)=0
    end do
    !$omp end parallel do

    !$omp parallel do private(colin, col, i)
    DO row=1,coarser_level_size
       i = 0
       DO colin=1,TEMP_COLSIZE
          col=Temp_CN(colin,row)
          if(col>0) then
             IF(col > row) THEN
                newINU(row)=newINU(row)+1
                
             else if(col < row) then
                newINL(row) = newINL(row) + 1
             END IF

             !C collect the distributed data on each row.
             i=i+1
             Temp_CN(i,row) = Temp_CN(colin, row)  
             Temp_V((i-1)*NK*NK+1:i*NK*NK, row) = Temp_V((colin-1)*NK*NK+1:colin*NK*NK, row)
             
          end if
       END DO
    END DO
    !$omp end parallel do

    DO i=1,coarser_level_size
       newINU(i)=newINU(i-1)+newINU(i)
       newINL(i)=newINL(i-1)+newINL(i)
    END DO

    HIERARCHICAL_DATA_NB(LEVEL_NO) % NPU = newINU(coarser_level_size)
    HIERARCHICAL_DATA_NB(LEVEL_NO) % NPL = newINL(coarser_level_size)
    !C----- NB blocking is done from start to here!! -----

    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % IAL(     HIERARCHICAL_DATA_NB(LEVEL_NO) % NPL))
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % IAU(     HIERARCHICAL_DATA_NB(LEVEL_NO) % NPU))
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % AL(NK,NK,HIERARCHICAL_DATA_NB(LEVEL_NO) % NPL))
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % AU(NK,NK,HIERARCHICAL_DATA_NB(LEVEL_NO) % NPU))
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % D(NK,NK,local_aggre_size))
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % X(local_aggre_size*NK))
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % B(local_aggre_size*NK))

    newD   => HIERARCHICAL_DATA_NB(LEVEL_NO) % D 
    newIAL => HIERARCHICAL_DATA_NB(LEVEL_NO) % IAL
    newIAU => HIERARCHICAL_DATA_NB(LEVEL_NO) % IAU
    newAL  => HIERARCHICAL_DATA_NB(LEVEL_NO) % AL
    newAU  => HIERARCHICAL_DATA_NB(LEVEL_NO) % AU

    !C substitution Temp array to new Matrix
    allocate(comm_recv_node(local_aggre_size - coarser_level_size))
    comm_recv_node = 0


    !C calculation of max number of Temp_N(:)
    n_max = 0
    do i = 1, coarser_level_size
       if(n_max < Temp_N(i)) n_max = Temp_N(i)
    end do


    !C val record the max nonzero element for debug
#ifdef DEBUG
    val = 0.0
#endif

    
    !$omp parallel private(tmp_int,row,d_i,i,j,k,colin,col,n_upper,n_lower)
    allocate(tmp_int(n_max))
    
    !$omp do
    do row = 1, coarser_level_size
       !C tmp_int(i) has i th smallest number's index.
       call qsorti(tmp_int,Temp_N(row),Temp_CN(1:Temp_N(row), row))
       
       !C diagonal element is found at d_i th element in the sorted non-zero elements. 
       do d_i = 1, Temp_N(row)
          if(row==Temp_CN(tmp_int(d_i),row)) exit
       end do
       
#ifdef DEBUG
       if(d_i > Temp_N(row)) stop 'error in make_rap()'
#endif

       !C lower part
       do colin = 1, d_i-1
          !C k is index of Temp_CN(:,row), Temp_V(:,row)
          k = tmp_int(colin)
          col = Temp_CN(k,row)
          
          n_lower = newINL(row-1)+colin
          newIAL(n_lower) = col
          
          do i = 1, NK
             do j = 1, NK
                newAL(i,j,n_lower) = Temp_V((k-1)*NK*NK+(i-1)*NK+j,row)

#ifdef DEBUG
                if(val < dabs(newAL(i,j,n_lower))) then
                   val = newAL(i,j,n_lower)
                   l = row
                end if
#endif
                
             end do
          end do
       end do

       !C diagonal part
       !C k is index of Temp_CN(:,row), Temp_V(:,row)
       k = tmp_int(d_i)
       do i = 1, NK
          do j = 1, NK
             newD(i,j,row) = Temp_V((k-1)*NK*NK+(i-1)*NK+j,row)
          end do
       end do

       !C upper part
       do colin = d_i+1, Temp_N(row)
          !C k is index of Temp_CN(:,row), Temp_V(:,row)
          k = tmp_int(colin)
          col = Temp_CN(k,row)
          
          n_upper = newINU(row-1)+colin-d_i
          newIAU(n_upper) = col
          
          do i = 1, NK
             do j = 1, NK
                newAU(i,j,n_upper) = Temp_V((k-1)*NK*NK+(i-1)*NK+j,row)
#ifdef DEBUG
                if(val < dabs(newAU(i,j,n_upper))) then
                   val = newAU(i,j,n_upper)
                   l = row
                end if
#endif


             end do
          end do

          !C record for recv table
          if(col > coarser_level_size) then
             comm_recv_node(col - coarser_level_size) = 1
          end if
       end do
    end do
    !$omp end do
    
    deallocate(tmp_int)
    !$omp end parallel

!C this debugging assumes single thread execution
#ifdef DEBUG    
    write(*,*) "my_rank,max(newAL,and newAU),row, rowsize:", my_rank, val, l, coarser_level_size
#endif

    DEALLOCATE(Temp_V)
    DEALLOCATE(Temp_CN)
    DEALLOCATE(Temp_N)
    !C ------

    !C remake R
    R => HIERARCHICAL_DATA_NB(LEVEL_NO) % R
    P => HIERARCHICAL_DATA_NB(LEVEL_NO) % P
    ALLOCATE(R % IN(0:local_aggre_size))    
    if(.not.PRO_AGGRE_OPT) then
       R % IN = 0
       DO i = 1,N
          DO j = P % IN(i-1)+1,P % IN(i)
             col = P % CN(j)
             R % IN(col) = R % IN(col)+1
          END DO
       END DO
       DO i = 2,local_aggre_size
          R % IN(i) = R % IN(i)+R % IN(i-1)
       END DO

       DO i = 2,R % ROW_SIZE
          rin = R % IN(i-1)
          trin = Temp_R_IN(i-1)
          j = trin-rin
          k = R % IN(i)-rin

          DO l = trin+1,trin+k
             R % CN(l-j) = R % CN(l)
             R % V((l-j-1)*NK*NB+1:(l-j)*NK*NB) = R % V((l-1)*NK*NB+1:l*NK*NB)
          END DO
       END DO

       R % ROW_SIZE = local_aggre_size    
       !C-- end:remake R
    else
       R % ROW_SIZE = coarser_lev_N_bf_CGA
       DO i = 0, R % ROW_SIZE
          R % IN(i) = Temp_R_IN(i)
       END DO

       k = 0
       if(P % ROW_SIZE > 0) then
          je = 0
          DO i = 1, P % ROW_SIZE
             js = je+1
             je = P % IN(i)
             DO j = js, je
                col = P % CN(j)
                if(col > R % ROW_SIZE) then
                   k = k+1
                else
                   P % CN(j-k) = P % CN(j)
                   DO l=1,NK*NB
                      P % V((j-k-1)*NK*NB+l) = P % V((j-1)*NK*NB+l)
                   END DO
                end if
             END DO
             P % IN(i) = je-k
          END DO
       end if
    end if

#ifdef DEBUG
    CALL MPI_Barrier(solver_comm,ierr)
    IF (my_rank==0) THEN
       WRITE(*,*) "### make communication tables ###"
    END IF
    CALL MPI_Barrier(solver_comm,ierr)
#endif


!!$    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % R)
!!$    R=>HIERARCHICAL_DATA_NB(LEVEL_NO) % R
!!$    R % ROW_SIZE=coarser_level_size
!!$    ALLOCATE(R % IN(0:coarser_level_size))
!!$    R % IN(0:coarser_level_size) = 0
!!$
!!$    count=0
!!$    DO i=1,P % ROW_SIZE
!!$       DO j=P % IN(i-1)+1,P % IN(i)
!!$          col=P % CN(j)
!!$          if(col<=coarser_level_size) then
!!$             count=count+1
!!$             if(col<coarser_level_size) R % IN(col+1)=R % IN(col+1)+1
!!$          end if
!!$       END DO
!!$    END DO
!!$
!!$    ALLOCATE(R % CN(count))
!!$    ALLOCATE(R % V(count*9))
!!$
!!$    DO i=2,R % ROW_SIZE
!!$       R % IN(i)=R % IN(i)+R % IN(i-1)
!!$    END DO
!!$
!!$!C    -8 -7 -6
!!$!C    -5 -4 -3
!!$!C    -2 -1 -0
!!$    DO i=1,P % ROW_SIZE
!!$       DO j=P % IN(i-1)+1,P % IN(i)
!!$          col=P % CN(j)
!!$          if(col<=coarser_level_size) then 
!!$             R % IN(col)=R % IN(col)+1
!!$             k=R % IN(col)
!!$             R % CN(k)=i
!!$             R % V (k*9-8)=P % V (j*9-8)
!!$             R % V (k*9-7)=P % V (j*9-5)
!!$             R % V (k*9-6)=P % V (j*9-2)
!!$             R % V (k*9-5)=P % V (j*9-7)
!!$             R % V (k*9-4)=P % V (j*9-4)
!!$             R % V (k*9-3)=P % V (j*9-1)
!!$             R % V (k*9-2)=P % V (j*9-6)
!!$             R % V (k*9-1)=P % V (j*9-3)
!!$             R % V (k*9)  =P % V (j*9)
!!$          END if
!!$       END DO
!!$    END DO

    !C-- communication table creation

    j=0
    DO i=1,PE_list_size
       j=j+aggregate_number_in_table(i)
    END DO

    newN =coarser_level_size
    newNP=local_aggre_size
!!$    HIERARCHICAL_DATA_NB(LEVEL_NO) % NB = NK
    HIERARCHICAL_DATA_NB(LEVEL_NO) % N  = newN
    HIERARCHICAL_DATA_NB(LEVEL_NO) % NP = newNP

    ALLOCATE(export_global_buffer(j))    
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_IMPORT(j))    
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_IMPORT(0:NPROCS))    
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_EXPORT(0:NPROCS))    
    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPE(NPROCS))    
    newNOD_IMPORT   => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_IMPORT
    newSTACK_IMPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_IMPORT
    newSTACK_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_EXPORT
    newNEIBPE       => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPE
    newSTACK_IMPORT(0)=0
    newSTACK_EXPORT(0)=0

    if(.not.PRO_AGGRE_OPT) then
       ALLOCATE(export_int_lvl_buffer(j+1))    
       ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT(j+1))    
       ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT(0:NPROCS))
       ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT(0:NPROCS))    
       ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPE(NPROCS))
       newNOD_IMPORT_INT   => HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NOD_IMPORT
       newSTACK_IMPORT_INT => HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT
       newSTACK_EXPORT_INT => HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT
       newNEIBPE_INT       => HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPE
       newSTACK_IMPORT_INT(0)=0
       newSTACK_EXPORT_INT(0)=0
    end if

    !C-- check for comm_table and select important nodes
    idx_NOD = 0
    idx_PE  = 0
    idx_NOD_INT = 0
    idx_PE_INT  = 0
    DO i = 1,PE_list_size
       k = aggregate_number_in_table(i)
       DO j = 1,k
          l = aggregate_table_array(i) % int_array(j*2)
          g = aggregate_table_array(i) % int_array(j*2-1)
!C          m1 = newINL(l)-newINL(l-1)+newINU(l)-newINU(l-1)
          m1 = comm_recv_node(l - coarser_level_size)

          !C we ignore coefficients among external nodes,so m1 /=0 means 
          !C l is connected to own node.
          IF(m1 > 0) THEN
             idx_NOD = idx_NOD+1
             export_global_buffer(idx_NOD) = g
             newNOD_IMPORT(idx_NOD)        = l
          END IF
          if(.not.PRO_AGGRE_OPT) then
             m2 = R % IN(l)-R % IN(l-1)
             IF(m2 > 0) THEN
                idx_NOD_INT = idx_NOD_INT+1
                export_int_lvl_buffer(idx_NOD_INT) = g
                newNOD_IMPORT_INT(idx_NOD_INT)     = l
             END IF
          end if
       END DO
       IF(idx_NOD > newSTACK_IMPORT(idx_PE)) THEN
          idx_PE = idx_PE+1
          newNEIBPE(idx_PE) = PE_list(i,1)
          newSTACK_IMPORT(idx_PE) = idx_NOD
       END IF
       if(.not.PRO_AGGRE_OPT) then       
          IF(idx_NOD_INT > newSTACK_IMPORT_INT(idx_PE_INT)) THEN
             idx_PE_INT = idx_PE_INT+1
             newNEIBPE_INT(idx_PE_INT) = PE_list(i,1)
             newSTACK_IMPORT_INT(idx_PE_INT) = idx_NOD_INT
          END IF
       end if
    END DO


    newNEIBPETOT    = idx_PE
    HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE    % NEIBPETOT = newNEIBPETOT
    if(.not.PRO_AGGRE_OPT) then           
       newNEIBPETOT_INT= idx_PE_INT
       HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT = newNEIBPETOT_INT
       !C--end: check for comm_table and select important nodes

       CALL make_export_NOD &
            & (SOLVER_COMM, export_int_lvl_buffer,GIN_aggregate,my_rank,NPROCS,    &
            &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPETOT,            &
            &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NEIBPE,               &
            &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % NOD_EXPORT,           &
            &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_IMPORT,         &
            &  HIERARCHICAL_DATA_NB(LEVEL_NO) % INT_LVL_TABLE % STACK_EXPORT)

       CALL mpi_barrier(solver_comm,ierr)
    end if

    CALL make_export_NOD &
         & (SOLVER_COMM, export_global_buffer,GIN_aggregate,my_rank,NPROCS, &
         &  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPETOT,           &
         &  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPE,              &
         &  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_EXPORT,          &
         &  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_IMPORT,        &
         &  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % STACK_EXPORT)

    newNOD_EXPORT => HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NOD_EXPORT
    newNEIBPETOT  =  HIERARCHICAL_DATA_NB(LEVEL_NO) % COMM_TABLE % NEIBPETOT 


    DEALLOCATE(export_global_buffer)
    if(.not.PRO_AGGRE_OPT) then               
       DEALLOCATE(export_int_lvl_buffer)
    end if
    !C-- end:communication table creation

    if(global_finish_flag .and. .not.W_OUT_LU) then
       !C LU factorization here..
       !C all data is assumed to be 1 process
       if(newN > 0) then
          if(newN /= GIN_aggregate(NPROCS)) then
             stop "coarsest level is distributed on more than 1 process. try set W_OUT_LU."
          end if

          !C copy mat data to coarsest_A_NB(:,:)
          allocate(coarsest_A_NB(newN*NK, newN*NK), coarsest_IDX_NB(newN*NK))
          allocate(vec(NK))
          coarsest_A_NB = 0.
          do i = 1, newN
             vec= 0
             do j = newINL(i-1)+1, newINL(i)
                k = newIAL(j)
                do l = 1, NK
                   do m = 1,NK
                      coarsest_A_NB((i-1)*NK+l,(k-1)*NK+m) = newAL(l,m,j)
                      if(vec(l) < dabs(newAL(l,m,j))) vec(l) = dabs(newAL(l,m,j))
                   end do
                end do
             end do
             do j = newINU(i-1)+1, newINU(i)
                k = newIAU(j)
                do l = 1, NK
                   do m = 1, NK
                      coarsest_A_NB((i-1)*NK+l,(k-1)*NK+m) = newAU(l,m,j)
                      if(vec(l) < dabs(newAU(l,m,j))) vec(l) = dabs(newAU(l,m,j))
                   end do
                end do
             end do

             do l = 1, NK
                do m = 1, NK
                   coarsest_A_NB((i-1)*NK+l,(i-1)*NK+m) = newD(l,m,i)
                   if(vec(l) < dabs(newD(l,m,i))) vec(l) = dabs(newD(l,m,i))
                end do
             end do

             !C if there is zero row, then set 1.0 of the diagonal element.
             do l = 1, NK
                if(vec(l) == 0.0) coarsest_A_NB((i-1)*NB+l,(i-1)*NB+l) = 1.0
             end do

          end do
          deallocate(vec)
          

          call DGETRF(newN*NK, newN*NK, coarsest_A_NB, newN*NK, coarsest_IDX_NB, INFO)
          
          if(INFO/=0) stop "lu factorization!"

       end if
    end if
 
    do i = 1, newN
       do j = 1, NK
          if(newD(j,j,i) == 0) then
             !C- Check row value if dinagonal is zero
             do k = 1, NK
                if(newD(j,k,i) /= 0) exit
             enddo

             if(k > NK) then
                write(*,*) "Interpolate diagonal value because of zero in diagonal, rank:", my_rank, " i:", i
                newD(j,j,i) = 1.0
             endif
          endif
       enddo
    enddo
    
    !C diagonal part is to have correct values in overlapped region.
    IF( .NOT. global_finish_flag) THEN
#ifdef DEBUG
       CALL MPI_BARRIER(SOLVER_COMM,ierr)

       IF(my_rank == 0) WRITE(*,*) "newD exchange"

       CALL MPI_BARRIER(SOLVER_COMM,ierr)
#endif


       j_ex = newSTACK_EXPORT( newNEIBPETOT )
       j_im = newSTACK_IMPORT( newNEIBPETOT )
       ALLOCATE ( WS(j_ex * NK) )
       ALLOCATE ( WR(j_im * NK) )
       ALLOCATE ( WW(newNP * NK, NK) )       

       DO i= 1, newN
          DO j = 1, NK
             DO k = 1, NK
                WW(NK*(i-1)+k,j) = newD(k,j,i)
             END DO
          END DO

       ENDDO

       if(newNEIBPETOT > 0) then
          DO i = 1, NK
             CALL SOLVER_SEND_RECV_NB                                        &
                  &     (j_ex,j_im , newNP,newNEIBPETOT, newNEIBPE,          &
                  &      newSTACK_IMPORT, newNOD_IMPORT,newSTACK_EXPORT,     &
                  &      newNOD_EXPORT, WS, WR, WW(:,i), SOLVER_COMM,my_rank, NK )
          END DO
       end if


       DO i= newN+1, newNP
          DO j = 1,NK
             DO k = 1,NK
                newD(k,j,i) = WW(NK*(i-1)+k,j)
             END DO
          END DO
       ENDDO

       DEALLOCATE(WS)
       DEALLOCATE(WR)
       DEALLOCATE(WW)
    END IF

    !C ------
#ifdef DEBUG
    CALL MPI_BARRIER(SOLVER_COMM,ierr)
    IF(my_rank == 0) WRITE(*,*) "newALU is creation"
    CALL MPI_BARRIER(SOLVER_COMM,ierr)
#endif

    ALLOCATE(HIERARCHICAL_DATA_NB(LEVEL_NO) % ALU(NK,NK+1,newN))
    newALU => HIERARCHICAL_DATA_NB(LEVEL_NO) % ALU
    DO i=1,newN
       call ILU_NB(newALU(:,:,i), newD(:,:,i), NK)
    END DO





  CONTAINS
    !C--
    !C this routine specifies internal shared nodes by looking into NOD_EXPORT.
    !C
    !C Output
    !C NEIBPETOT:
    !C NEIBPE:
    !C in_repeated_send_buf: index
    !C repeated_nodes_send_buf: buffer of lists of shared PEs for each
    !C                         internal node
    SUBROUTINE make_repeated_nodes_send_buffer 
      !C intent(in ) N,STACK_EXPORT,NOD_EXPORT,NEIBPETOT,NEIBPE
      !C intent(out) repeated_nodes_tot
      !C intent(out) repeated_nodes_send_buf(repeated_nodes_tot)
      !C intent(out) in_repeated_nodes_send_buf
      INTEGER(kind=kint) :: debug1,i,j,k,l,js,je,current_node,count      
      INTEGER(kind=kint), DIMENSION(:,:),ALLOCATABLE :: repeated_nodes_at_boundary
      INTEGER(kind=kint), DIMENSION(:,:),ALLOCATABLE :: export_node_list
      INTEGER(kind=kint ), DIMENSION(:,:), ALLOCATABLE :: export_repeated
      INTEGER(kind=kint ), DIMENSION(:),  ALLOCATABLE :: in_export_repeated

      REPEATED_NODE_SIZE=100+(LEVEL_NO-1)*200
      ALLOCATE(export_repeated(REPEATED_NODE_SIZE,NEIBPETOT))
      ALLOCATE(in_export_repeated(0:NEIBPETOT))

      !C search for repeated nodes in NOD_EXPORT
      !C -----
      k=STACK_EXPORT(NEIBPETOT)
      ALLOCATE(export_node_list(3,k))
      !C export_node_list 1: node 2: PE NO 3: next
      DO i=1,k
         export_node_list(3,i)=0
      END DO
      ALLOCATE(repeated_nodes_at_boundary(3,N))
      DO i=1,N
         repeated_nodes_at_boundary(1,i)=0
         repeated_nodes_at_boundary(2,i)=0
      END DO

      DO i=1,NEIBPETOT
         js=STACK_EXPORT(i-1)+1
         je=STACK_EXPORT(i)
         DO j=js,je
            k=NOD_EXPORT(j)
            !C repeated_nodes_at_boundary 1:TAIL 2:SIZE,3:HEAD
            repeated_nodes_at_boundary(2,k)=repeated_nodes_at_boundary(2,k)+1
            current_node=repeated_nodes_at_boundary(1,k)
            IF(current_node == 0) THEN
               repeated_nodes_at_boundary(3,k)=j
            ELSE
               export_node_list(3,current_node)=j
            END IF
            repeated_nodes_at_boundary(1,k)=j          
            export_node_list(1,j)=j-js+1
            export_node_list(2,j)=i
         END DO
      END DO
      !C node is the number from  STACK_EXPORT's associated PE js

      repeated_nodes_tot=0
      
      in_export_repeated(0:NEIBPETOT)=0

      DO i=1,N
         l=repeated_nodes_at_boundary(2,i)
         IF(l >= 2) THEN
            repeated_nodes_tot = repeated_nodes_tot + l*(l+1)
            !C l:number of PEs  *  l+1:node_NO,PE_list_size,PE_list  
            current_node=repeated_nodes_at_boundary(3,i)
            DO WHILE(current_node>0)
               k=export_node_list(2,current_node)

               in_export_repeated(k) = in_export_repeated(k)+1
               j=in_export_repeated(k)
               !C node_number
               !C debug
#ifdef  DEBUG
               IF(j>REPEATED_NODE_SIZE) STOP    &
                    &  "export_repeated(*,) is overflow:make_repeated_nodes_send_buffer"
#endif
               export_repeated(j,k) = i
               current_node=export_node_list(3,current_node)
            END DO
         END IF
      END DO

      !C No nodes is repeated.-> assign 0 
      DO i=1,NEIBPETOT
         IF(in_export_repeated(i)==0)THEN
            repeated_nodes_tot=repeated_nodes_tot+1
         END IF
      END DO

      ALLOCATE(repeated_nodes_send_buf(repeated_nodes_tot))
      ALLOCATE(in_repeated_send_buf(0:NEIBPETOT))
      count=0
      DO i=1,NEIBPETOT
         IF(in_export_repeated(i)==0) THEN
            !C No nodes is repeated for that PE.
            count=count+1
            repeated_nodes_send_buf(count)=0
         ELSE
            DO j=1,in_export_repeated(i)
               k=export_repeated(j,i)
               current_node=repeated_nodes_at_boundary(3,k)
               count=count+1
               repeated_nodes_send_buf(count)=repeated_nodes_at_boundary(2,k)
               count=count+1
               !C l is the place for node number in NOD_EXPORT.
               l=count

               !C debug
#ifdef DEBUG               
               debug1=0
#endif               
               DO WHILE(current_node>0)
                  !C PE in NEIBPE(i) is in the stack_export.
                  !C export_node_list:: 1:node  2:pe  3:next
                  IF(export_node_list(2,current_node)==i) THEN
                     repeated_nodes_send_buf(l)=export_node_list(1,current_node) 
                     !C debug
#ifdef DEBUG                     
                     debug1=debug1+1
#endif                     
                  ELSE
                     count=count+1
                     repeated_nodes_send_buf(count)=NEIBPE(export_node_list(2,current_node))
                  END IF
                  current_node=export_node_list(3,current_node)
               END DO
               !C debug
#ifdef DEBUG               
               IF(debug1>1) WRITE(*,*)"error in making repeated_nodes_send_buf"
#endif
            END DO
         END IF
         in_repeated_send_buf(i)=count
      END DO
      in_repeated_send_buf(0)=0

      !C debug
#ifdef DEBUG
      IF(repeated_nodes_tot/=count)THEN
         WRITE(*,*)"error in making repeteated_nodes_send_buf"
         WRITE(*,*)repeated_nodes_tot,count
         STOP "unexpected error"
      END IF
#endif

      !C deallocate repeated_nodes_at_boundary
      DEALLOCATE(repeated_nodes_at_boundary)
      DEALLOCATE(export_node_list)
      DEALLOCATE(in_export_repeated)

    END SUBROUTINE make_repeated_nodes_send_buffer

    SUBROUTINE make_export_NOD (SOLVER_COMM,export_global_buffer,         &
         &     GIN_aggregate,my_rank,NPROCS,NEIBPETOT,NEIBPE,NOD_EXPORT,  &
         &     STACK_IMPORT,STACK_EXPORT)

      IMPLICIT NONE
!C      INCLUDE  'mpif.h'
      INTEGER(kind=kint), INTENT(in)     :: NPROCS
      INTEGER(kind=kint), INTENT(inout)  :: NEIBPETOT
      INTEGER(kind=kint), INTENT(inout)  :: NEIBPE(:)
      INTEGER(kind=kint), INTENT(inout)  :: STACK_IMPORT(0:),STACK_EXPORT(0:)
      INTEGER(kind=kint), POINTER        :: NOD_EXPORT(:)

      INTEGER(kind=kint),  INTENT(in)    :: export_global_buffer(:)
      INTEGER(kind=kint), ALLOCATABLE    :: pes(:),in_recv_buf(:),recv_buf(:)

      INTEGER(kind=kint) :: pes_s(NPROCS),pes_r(NPROCS)

      INTEGER(kind=kint), INTENT(in)     :: GIN_aggregate(0:)

      INTEGER :: ierr
      INTEGER, INTENT(in) :: SOLVER_COMM
      INTEGER, INTENT(in) :: my_rank
      INTEGER(kind=kint), ALLOCATABLE :: req1(:),req2(:)
      INTEGER(kind=kint), ALLOCATABLE :: sta1(:,:),sta2(:,:)


      INTEGER(kind=kint) :: i,j,k,dest,inum,istart,recv_buf_size,message_size,l
      INTEGER(kind=kint) :: GLOBAL,LOCAL,base_index,n_pes,neib,from_pe,msg_sz
      PARAMETER(GLOBAL=1,LOCAL=2)
      LOGICAL :: hash_flag

      neib=NEIBPETOT
      ALLOCATE(sta1(MPI_STATUS_SIZE,neib))
      ALLOCATE(req1(neib))    

      !C-- determine number of recv
      pes_s=-1
      DO i=1,neib
         dest=NEIBPE(i)
         pes_s(dest+1)     = STACK_IMPORT(i)-STACK_IMPORT(i-1)
#ifdef DEBUG
         IF(pes_s(dest+1)==0) STOP "0 size tables in make_export_NOD()"
#endif
      END DO
      pes_r=-1
      CALL MPI_ALLTOALL(pes_s, 1, MPI_INTEGER, pes_r, 1, MPI_INTEGER, &
           &            SOLVER_COMM, ierr)

      !C-- determine number of recv
      !C RECEIVE
      n_pes=0
      ALLOCATE(in_recv_buf(0:NPROCS))
      ALLOCATE(pes(NPROCS))
      in_recv_buf(0)=0
      DO i=0,NPROCS-1
         from_pe = i
         msg_sz  = pes_r(i+1)
         IF(msg_sz>=0) THEN
            n_pes              = n_pes+1
            pes(n_pes)         = from_pe
            in_recv_buf(n_pes) = in_recv_buf(n_pes-1)+msg_sz
         END IF
      END DO
      ALLOCATE(req2(n_pes))
      ALLOCATE(sta2(MPI_STATUS_SIZE,n_pes))    
      recv_buf_size = in_recv_buf(n_pes)
      ALLOCATE(recv_buf(recv_buf_size))
      ALLOCATE(NOD_EXPORT(recv_buf_size))
      DO i=1,n_pes
         istart = in_recv_buf(i-1)
         inum   = in_recv_buf(i)-in_recv_buf(i-1)
         dest   = pes(i)
         CALL MPI_IRECV(recv_buf(istart+1), inum, MPI_INTEGER, dest ,11,&
              &        SOLVER_COMM, req2(i),ierr)
      END DO

      !C SEND
      DO i=1,neib
         istart=STACK_IMPORT(i-1)
         inum  =STACK_IMPORT(i)-istart
         dest=NEIBPE(i)
         CALL MPI_ISEND(export_global_buffer(istart+1),inum,MPI_INTEGER,dest,11,  &
              &         SOLVER_COMM,req1(i),ierr)
      END DO


      IF(n_pes >0) CALL MPI_WAITALL (n_pes, req2, sta2, ierr)
      IF(neib  >0) CALL MPI_WAITALL (neib,  req1, sta1, ierr)


      base_index=GIN_aggregate(my_rank)    
      STACK_EXPORT(0)=0
      DO i=1,neib
         dest=NEIBPE(i)
         j=1
         DO WHILE(j<=n_pes .AND. pes(j)/=dest)
            j=j+1
         END DO
         IF(j>n_pes) THEN
            !C there is no nodes to send.
            STACK_EXPORT(i)=STACK_EXPORT(i-1)
         ELSE
            STACK_EXPORT(i)=STACK_EXPORT(i-1)+in_recv_buf(j)-in_recv_buf(j-1)
            l=STACK_EXPORT(i-1)-in_recv_buf(j-1)
            DO k=in_recv_buf(j-1)+1,in_recv_buf(j)
               NOD_EXPORT(l+k)=recv_buf(k)-base_index
#ifdef DEBUG
               IF(pes(j)==-1) STOP "error!!! in make_export_table"
               IF(NOD_EXPORT(k+l)<0 .OR.NOD_EXPORT(k+l)>GIN_aggregate(my_rank+1)-base_index) THEN
                  WRITE(*,*) "unexpected error in make_export_nod():", NOD_EXPORT(k+l),my_rank
                  STOP
               END IF
#endif            
            END DO
            !C mark which means un-needed 
            pes(j)=-1
         END IF
      END DO
      DO i=1,n_pes
         IF(pes(i)/=-1) THEN
            dest=pes(i)
            NEIBPETOT=NEIBPETOT+1
            NEIBPE(NEIBPETOT)=dest
            STACK_IMPORT(NEIBPETOT)=STACK_IMPORT(NEIBPETOT-1)
            STACK_EXPORT(NEIBPETOT)=STACK_EXPORT(NEIBPETOT-1)+in_recv_buf(i)-in_recv_buf(i-1)
            l=STACK_EXPORT(NEIBPETOT-1)-in_recv_buf(i-1)
            DO k=in_recv_buf(i-1)+1,in_recv_buf(i)
               NOD_EXPORT(k+l)=recv_buf(k)-base_index
#ifdef DEBUG
               IF(NOD_EXPORT(k+l)<0 .OR. NOD_EXPORT(k+l)>GIN_aggregate(my_rank+1)-base_index) THEN
                  WRITE(*,*) "unexpected error in make_export_nod():" ,NOD_EXPORT(k+l),my_rank
                  STOP
               END IF
#endif            
            END DO
         END IF
      END DO

      DEALLOCATE(pes)    
      DEALLOCATE(in_recv_buf)    
      DEALLOCATE(recv_buf)    

      DEALLOCATE(req1)
      DEALLOCATE(req2)
      DEALLOCATE(sta2)
      DEALLOCATE(sta1)
    END SUBROUTINE make_export_NOD


  END SUBROUTINE make_rap


  SUBROUTINE ILU_NB (ALU, D, NB)
    IMPLICIT none
    include 'precision.inc'
    integer(kind=kint), intent(in) :: NB
    REAL(kind=kreal), intent(in) :: D(NB,NB)
    REAL(kind=kreal), intent(out) :: ALU(NB,NB+1)

    integer(kind=kint) :: i,j,k
    integer :: info


    DO i = 1, NB
       DO j = 1, NB
          ALU(i,j) = D(i,j)
       END DO
    END DO

    call DGETRF(NB,NB,ALU(1:NB,1:NB),NB,ALU(:,NB+1),info)
    
    if(info/=0) then
       write(*,*) "lu factorization error in ILU_NB(), and reset the ALU to I. D(:,:) is ", D(:,:)
       
       ALU(:,:)=0.0
       DO i = 1, NB
          ALU(i,i) = 1.0
       END DO
       call DGETRF(NB,NB,ALU(1:NB,1:NB),NB,ALU(:,NB+1),info)

    end if
    
  END SUBROUTINE ILU_NB


  SUBROUTINE smooth_aggregate_NB (PNI,NI,omega,LEVEL_NO, in_aggregates_result_size, &
       & in_aggregates_result, aggregates_result,Temp_P,Temp_PI,Temp_P_row_size,    &
       & tilde_D)
    USE data_structure_for_AMG_NB
    IMPLICIT NONE

    INTEGER(kind=kint ), INTENT(in)  :: PNI(0:)
    INTEGER(kind=kint ), INTENT(in)  :: NI(:)
    REAL   (kind=kreal), INTENT(in)  :: omega
    INTEGER(kind=kint ), INTENT(in)  :: LEVEL_NO
    INTEGER(kind=kint ), INTENT(in)  :: in_aggregates_result_size
    REAL   (kind=kreal), INTENT(in)  ::  tilde_D(:,:,:)

    INTEGER(kind=kint ), INTENT(in)  :: in_aggregates_result(0:)
    INTEGER(kind=kint ), INTENT(in)  :: aggregates_result(:)
    REAL   (kind=kreal), INTENT(out) :: Temp_P(:,:)
    INTEGER(kind=kint ), INTENT(out) :: Temp_PI(:,:)
    INTEGER(kind=kint ), INTENT(out) :: Temp_P_row_size(:)

    REAL   (kind=kreal), ALLOCATABLE :: Temp(:,:)

    REAL   (kind=kreal), POINTER ::  D (:,:,:)
    REAL   (kind=kreal), POINTER ::  AU(:,:,:)
    REAL   (kind=kreal), POINTER ::  AL(:,:,:)
    INTEGER(kind=kint ), POINTER ::  INU(:)
    INTEGER(kind=kint ), POINTER ::  IAU(:)
    INTEGER(kind=kint ), POINTER ::  INL(:)
    INTEGER(kind=kint ), POINTER ::  IAL(:)

    REAL   (kind=kreal) :: omega2,T1,T2,T3,SW(10)
    !C NB is assumed to be 10 at most.
    INTEGER(kind=kint ) :: i,j,k,l,row,index,isL,ieL,isU,ieU,is,ie,NP,N,m1,m2
    INTEGER(kind=kint ) :: connected_node,connected_node_start,connected_node_end

    INTEGER(kind=kint)  :: row_no,row_size,A_row,A_col,P_col, NB, count, i_ni
    INTEGER(kind=kint), ALLOCATABLE :: Temp_int(:)

    NB = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NB
    NP = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NP
    N  = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % N 

    ALLOCATE(Temp_int(NP))

    D   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % D 
    INL => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INL 
    INU => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INU 
    IAL => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AU

    Temp_P_row_size(1:NP)=0

    !C-- Temp_int() has aggregate NO for each node.
    Temp_int = 0
    DO k = 1, in_aggregates_result_size
       is = in_aggregates_result(k-1)+1
       ie = in_aggregates_result(k)
       DO i = is,ie
          row_no = aggregates_result(i)
          Temp_int(row_no) = k
       END DO
    END DO

    !C-- Temp_P's space is calculated for smoothed aggregations
    !C   and input column and value( initial values 0 )
    DO k = 1, in_aggregates_result_size
       is = in_aggregates_result(k-1)+1
       ie = in_aggregates_result(k)

       DO i = is,ie
          row_no   = aggregates_result(i)
          row_size = Temp_P_row_size(row_no)
          DO j = 1, row_size
             if(Temp_PI(j,row_no) == k) exit
          end DO
          if(j > row_size) then
             Temp_P_row_size(row_no) = j
             Temp_PI(j,row_no) = k

             DO l = 1, NB*NB
                Temp_P((j-1)*NB*NB+l, row_no) = 0.0
             END DO

          end if

          !C lower part
          connected_node_start = PNI(2*row_no-2)+1
          connected_node_end   = PNI(2*row_no-1)
          DO connected_node = connected_node_start,connected_node_end
             i_ni = NI(connected_node)
             A_col = IAL(i_ni)
             row_size = Temp_P_row_size(A_col)
             DO j = 1, row_size
                if(Temp_PI(j,A_col) == k) exit
             END DO
             if(j > row_size) then
                Temp_P_row_size(A_col) = j
                Temp_PI(j,A_col) = k
                DO l = 1, NB*NB
                   Temp_P((j-1)*NB*NB+l, A_col) = 0.0
                END DO
             end if
             !C transposed value is used : AL(m2,m1,l)
             DO m1 = 1, NB
                DO m2 = 1, NB
                   Temp_P((j-1)*NB*NB+NB*(m1-1)+m2, A_col) = Temp_P((j-1)*NB*NB+NB*(m1-1)+m2, A_col) - AL(m2,m1,i_ni)
                END DO
             END DO

          END DO
          !C upper part
          connected_node_start = PNI(2*row_no-1)+1
          connected_node_end = PNI(2*row_no)
          DO connected_node = connected_node_start,connected_node_end
             i_ni = NI(connected_node)
             A_col = IAU(i_ni)
             row_size = Temp_P_row_size(A_col)
             DO j = 1, row_size
                if(Temp_PI(j,A_col) == k) exit
             END DO

             if(j > row_size) then
                Temp_P_row_size(A_col) = j
                Temp_PI(j,A_col) = k

                DO l = 1, NB*NB
                   Temp_P((j-1)*NB*NB+l,A_col) = 0.0
                END DO

             end if
             !C transposed value is used : AU(m2,m1,l)
             DO m1 = 1, NB
                DO m2 = 1, NB
                   Temp_P((j-1)*NB*NB+NB*(m1-1)+m2, A_col) = Temp_P((j-1)*NB*NB+NB*(m1-1)+m2, A_col) - AU(m2,m1,i_ni)
                END DO
             END DO
          END DO
       END DO
    END DO

    !C-- diagonal part
    DO A_row = 1, NP
       DO i = 1, Temp_P_row_size(A_row)

!!$          Temp_P(i,A_row) % value = Temp_P(i,A_row) % value * omega/tilde_D(A_row)

          !C 1 2 3 -> j
          !C 4 5 6
          !C 7 8 9
          !C
          !C 3 vectors are applied to forward and backward substitution
          DO m1 = 1, NB
             !C set SW(:)
             DO m2 = 1, NB
                SW(m2) = Temp_P((i-1)*NB*NB+m1+(m2-1)*NB, A_row)
             END DO

             DO j = 2, NB
                DO k = 1, j-1
                   SW(j) = SW(j) - tilde_D(j,k,A_row)*SW(k)
                END DO
             END DO

             DO j = NB, 1, -1
                DO k = j+1, NB
                   SW(j) = SW(j) - tilde_D(j,k,A_row)*SW(k)
                END DO
                SW(j) = tilde_D(j,j,A_row)*SW(j)
             END DO

             DO m2 = 1, NB
                Temp_P((i-1)*NB*NB+m1+(m2-1)*NB, A_row) = omega * SW(m2)
             END DO

          END DO

          !C -8 -7 -6
          !C -5 -4 -3
          !C -2 -1 -0

       END DO
    END DO


    omega2 = 1. - omega
    DO j = 1, in_aggregates_result_size

       !C  element's  value of the aggregate is 1 in array P      
       is = in_aggregates_result(j - 1) + 1
       ie = in_aggregates_result(j)

       DO i = is, ie
          row_no = aggregates_result(i)
          DO count = 1, Temp_P_row_size(row_no)
             if(Temp_PI(count, row_no) == j) exit
          END DO
          if(count > Temp_P_row_size(row_no)) then
             stop 'unexpected error in smooth_aggregate()'
          else
             !C             Temp_P(count, row_no) % value = Temp_P(count, row_no) % value + omega2
             DO k = 1, NB
                Temp_P((count-1)*NB*NB+(k-1)*NB+k, row_no) = Temp_P((count-1)*NB*NB+(k-1)*NB+k, row_no) + omega2
             END DO
          end if
       END DO
    END DO

  END SUBROUTINE smooth_aggregate_NB



!C This routine uses symmetricity structure of PNI in order to treat 
!C unstructured problem matrix.
!C This routine uses PNI, NI and no option other than SMOOTH_TILDE_A
!C NB is set as 50
  SUBROUTINE smooth_aggregate_unsym_vec_NB (PNI, NI, omega, LEVEL_NO, in_aggregates_result_size, &
       & in_aggregates_result, aggregates_result, Temp_P, Temp_PI,Temp_P_row_size, tilde_DLU,   &
       & RATE_OF_SPACE_i, EX_INL, EX_IAL, EX_AL, NB, SOLVER_COMM, my_rank)
    use count_time_mod
    use data_structure_for_AMG_NB
    !$ USE omp_lib
    
    IMPLICIT NONE

    INTEGER(kind=kint ), INTENT(in) :: PNI(0:)
    INTEGER(kind=kint ), INTENT(in) :: NI(:)
    REAL   (kind=kreal), INTENT(in) :: omega 
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO, RATE_OF_SPACE_i
    INTEGER(kind=kint ), INTENT(in) :: in_aggregates_result_size, SOLVER_COMM, my_rank
    REAL   (kind=kreal), INTENT(in) ::  tilde_DLU(:,:,:)
    INTEGER(kind=kint ), INTENT(in) :: EX_INL (0:), EX_IAL (:)
    REAL   (kind=kreal), INTENT(in) :: EX_AL  (:,:,:)
    
    INTEGER(kind=kint ), INTENT(in) :: in_aggregates_result(0:)
    INTEGER(kind=kint ), INTENT(in) :: aggregates_result(:)

    real   (kind=kreal), INTENT(out) :: Temp_P(:,:)
    INTEGER(kind=kint ), INTENT(out) :: Temp_PI(:,:)
    INTEGER(kind=kint ), INTENT(out) :: Temp_P_row_size(:)
    INTEGER(kind=kint ), allocatable :: Temp_agr_num(:)

    REAL   (kind=kreal), POINTER ::  D (:,:,:)
    REAL   (kind=kreal), POINTER ::  AU(:,:,:)
    REAL   (kind=kreal), POINTER ::  AL(:,:,:)
    INTEGER(kind=kint ), POINTER ::  INU(:)
    INTEGER(kind=kint ), POINTER ::  IAU(:)
    INTEGER(kind=kint ), POINTER ::  INL(:)
    INTEGER(kind=kint ), POINTER ::  IAL(:)
    
    INTEGER(kind=kint ), INTENT(in) :: NB
    
    Real   (kind=kreal) :: omega2, v
    INTEGER(kind=kint ) :: N, i, j, k, l, row, index, is, ie, NP, c, m1, m2
    INTEGER(kind=kint ) :: connected_node, connected_node_start, connected_node_end
    INTEGER(kind=kint ) :: P_CN, count, node_num

    INTEGER(kind=kint), allocatable :: temp(:,:), temp_list(:,:), temp_sz(:),  temp_index(:,:)
    INTEGER(kind=kint) ::  tsz=1, tid = 0, tstart, tend, stid=1, ms, me, m, min_k
    
    REAL(kind=kreal), allocatable :: SW(:)
    integer :: info
    

    NP = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NP
    N  = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % N 

    D   => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % D 
    INL => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % INL
    INU => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % INU
    IAL => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % IAL
    IAU => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % IAU
    AL  => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % AL
    AU  => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % AU
    
    !C assign the aggregate number in Temp_agr_num
    allocate(Temp_agr_num(N))

    !$omp parallel private(i,j,k)
    !$omp do
    DO i = 1, N
       Temp_agr_num(i) = 0
    END DO
    !$omp end do

    !$omp do 
    DO i = 1, in_aggregates_result_size
       DO j = in_aggregates_result(i - 1) + 1, in_aggregates_result(i)
          k = aggregates_result(j)
          Temp_agr_num(k) = i
       END DO
    END DO
    !$omp end do 
    
    !$omp do
    DO i = 1, NP
       Temp_P_row_size(i) = 0
    end DO
    !$omp end  do
    !$omp end parallel

    !C determine nonzero structure which utilizes the symmetricity of SMOOTH_TILDE_A
    
    !C----- parallel ----
    !C aggregates are allocated to threads cyclicly.
    !$omp parallel  private(count, i, j, k, m, l, tstart, tend, ms, me, stid, tid, &
    !$omp is, ie, P_CN, connected_node, connected_node_start, connected_node_end,  &
    !$omp node_num)
    !$ tid = OMP_GET_THREAD_NUM()
    !$ tsz = OMP_GET_NUM_THREADS()

    !$omp single
    allocate(temp(NP, tsz), temp_list(NP, tsz), temp_sz(tsz), temp_index(0:in_aggregates_result_size, tsz))
    !$omp end single

    
    stid = tid+1

    !C initialize
    DO j = 1, NP
       temp(j, stid) = 0
       temp_list(j, stid) = 0
    END DO

    !C temp_sz has last aggregate number which is dealt with for tid.
    temp_sz(stid) = 0  

    !C k is the head of aggregate block for threads.
    !C k is increasing by tsz, as loop iterates.
    k = 1 
    DO while(k <= in_aggregates_result_size)    
       !C temp_index  and temp_sz is set.. 
       !C Here k-1 aggregates data is stored in Temp_PI already, 
       !C so temp_index(0:(k-1)/tsz, stid) is set 0
       temp_index(0:(k-1)/tsz, stid) = 0

       temp_sz(stid) = 0  
       count = 0 !C index for temp_list(:, stid)

       DO while(k <= in_aggregates_result_size)
          if(k+tid <= in_aggregates_result_size) then
             !C aggregate number is k+tid
             !C  element's  value of the aggregate is 1 in array P      
             is = in_aggregates_result((k+tid) - 1) + 1
             ie = in_aggregates_result(k+tid)
             DO i = is, ie
                P_CN = aggregates_result(i)          
                !C temp(:, stid) is redundancy check vector
                if(temp(P_CN, stid) /= k+tid) then
                   count = count+1
                   if(count > NP) exit !C temp_list buffer is full
                   temp_list(count, stid) = P_CN
                   temp(P_CN, stid) = k+tid
                end if
                connected_node_start = PNI(2 * P_CN - 2) + 1
                connected_node_end   = PNI(2 * P_CN - 1)
                DO connected_node = connected_node_start, connected_node_end
                   node_num = IAL(NI(connected_node)) + ZERO_ORIGIN
                   if(temp(node_num, stid) /= k+tid) then
                      count = count+1
                      if(count > NP) exit
                      temp_list(count, stid) = node_num
                      temp(node_num, stid) = k+tid
                   end if
                END DO
                if(count > NP) exit !C temp_list buffer is full
                connected_node_start = PNI(2 * P_CN - 1) + 1
                connected_node_end = PNI(2 * P_CN)
                DO connected_node = connected_node_start, connected_node_end
                   node_num = IAU(NI(connected_node)) + ZERO_ORIGIN
                   if(temp(node_num, stid) /= k+tid) then
                      count = count+1
                      if(count > NP) exit
                      temp_list(count, stid) = node_num
                      temp(node_num, stid) = k+tid
                   end if
                END DO
                if(count > NP) exit !C temp_list buffer is full
             END DO
          END if

          if(count <= NP) then
             !C record the index information every loop.
             temp_index((k-1)/tsz+1, stid) = count
             temp_sz(stid) = k
             k = k + tsz
          else 
             !C temp_list buffer is full.
             !C cancel the recorded info on redundancy check vector, and exit out of DO loop
             DO i = temp_index((k-1)/tsz, stid)+1, count-1
                temp(temp_list(i,stid), stid) = 0
             END DO
             exit
          end if
       END DO

       !$omp barrier

       !$omp single
       min_k = temp_sz(1)
       DO i = 2, tsz
          if(min_k > temp_sz(i)) min_k = temp_sz(i)
       END DO
       !$omp end single
       !C --syncronizaiton..--

       !C cancel the recorded info
       if(min_k < temp_sz(stid)) then
          is = temp_index((min_k-1)/tsz+1, stid)+1
          ie = temp_index((temp_sz(stid)-1)/tsz+1, stid)
          DO i = is, ie
             temp(temp_list(i,stid), stid) = 0
          END DO
          temp_sz(stid) = min_k
       end if

       tstart = tid*NP/tsz+1; tend = (tid+1)*NP/tsz

       DO m = 1, (min_k-1)/tsz+1 !C m: thread block number 
          DO l = 1, tsz !C l: thread number from 1..tsz
             is = temp_index(m-1,l)+1; ie = temp_index(m,l)
             DO i = is, ie
                node_num = temp_list(i, l)
                !$ if(node_num >= tstart .and. node_num <= tend) then
                      count = Temp_P_row_size(node_num)+1
                      Temp_P_row_size(node_num) = count
                      !C for vector operations here's modified.
                      Temp_PI(count, node_num) = (m-1)*tsz+l
                      Temp_P((count-1)*NB*NB+1:count*NB*NB, node_num) = 0.
                !$ end if
             END DO
          END DO
       END DO


       !$omp barrier
       k = min_k+tsz
    END DO
    !$omp end parallel

    !C----- parallel ----       

    
    !C Dumped yacobi method
    !$omp parallel private(i,j,k,count,c,m1,m2,SW,is,ie,P_CN) 
    
    !$omp do 
    DO i = 1, N
       !C lower part
       DO j = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
!!$          v = AL(NI(j))
          k = Temp_agr_num(IAL(NI(j))+ZERO_ORIGIN)

          DO count = 1, Temp_P_row_size(i)
             if(Temp_PI(count, i) == k) exit
          END DO
          IF(count > Temp_P_row_size(i)) THEN
             if(k > 0) stop '0:error in smooth_aggregate_unsym: non-symmetric data structure'
          else
             !C for vector cases, v becomes AL(:,:,NI(j)) * orthogonalized kernel vecs(:,:) of IAL(NI(j))
             !C             Temp_P(count, i) % value  = Temp_P(count, i) % value - v
             !C columne direction is assumed to be sequence access.
             
             DO m1 = 1, NB
                DO m2 = 1, NB
                   Temp_P((count-1)*NB*NB+m1+(m2-1)*NB, i) = &
                        & Temp_P((count-1)*NB*NB+m1+(m2-1)*NB, i) - AL(m1,m2,NI(j))
                END DO
             END DO
             
          END IF
       END DO

       !C upper part
       DO j = PNI(2 * i - 1) + 1, PNI(2 * i)
!!$          v = AU(NI(j))
          c = IAU(NI(j)) + ZERO_ORIGIN
          if(c <= N) then
             k = Temp_agr_num(c)

             DO count = 1, Temp_P_row_size(i)
                if(Temp_PI(count, i) == k) exit
             END DO
             !C we must consider the case of k==0. In that case, following stop statement will be executed.
             IF(count > Temp_P_row_size(i)) THEN
                if(k > 0) stop '1:error in smooth_aggregate_unsym: non-symmetric data structure'
             else
                !C for vector cases, v becomes AU(NI(j)) x orthogonalized kernel vecs(:,:) of IAU(NI(j))
!!$                Temp_P(count, i) % value  = Temp_P(count, i) % value - v

                DO m1 = 1, NB
                   DO m2 = 1, NB
                      Temp_P((count-1)*NB*NB+m1+(m2-1)*NB, i) = &
                           & Temp_P((count-1)*NB*NB+m1+(m2-1)*NB, i) -  AU(m1,m2,NI(j))
                   END DO
                END DO

             END IF
          end if
       END DO
    END DO
    !$omp end do

    !C external lower part
    !$omp do 
    DO i = N + 1, NP
       DO j = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
!!!$          v = EX_AL(NI(j))
          c = EX_IAL(NI(j))+ZERO_ORIGIN
          if(c <= N) then
             k = Temp_agr_num(c)

             DO count = 1, Temp_P_row_size(i)
                if(Temp_PI(count, i) == k) exit
             END DO

             IF(count > Temp_P_row_size(i)) THEN
                if(k > 0) stop '2:error in smooth_aggregate_unsym: non-symmetric data structure'
             else
                !C for vector cases, v becomes EX_AL(:,:,NI(j)) x orthogonalized kernel vecs(:,:) of EX_IAL(NI(j))
!!!$                Temp_P(count, i) % value  = Temp_P(count, i) % value - v

                DO m1 = 1, NB
                   DO m2 = 1, NB
                      Temp_P((count-1)*NB*NB+m1+(m2-1)*NB, i) = &
                           & Temp_P((count-1)*NB*NB+m1+(m2-1)*NB, i) - EX_AL(m1,m2,NI(j))
                   END DO
                END DO
             END IF
          END if
       end DO
    END DO
    !$omp end do

    !C-- diagonal part
!!$    !$omp do private(i,j)
!!$    DO j = 1, NP
!!$       DO i = 1, Temp_P_row_size(j)
!!$          Temp_P(i, j) % value = Temp_P(i, j) % value * omega / &
!!$               & tilde_D(j)
!!$       END DO
!!$    END DO
!!$    !$omp end do

    !C 1 4 7 -> j
    !C 2 5 8
    !C 3 6 9
    !C-- diagonal part
    allocate(SW(NB))
    !$omp do 
    DO j = 1, NP
       DO i = 1, Temp_P_row_size(j)
!!$          Temp_P(i, j) % value = Temp_P(i, j) % value * omega / &
!!$               & tilde_D(j)
          !C 3 vectors are applied to forward and backward substitution
          DO m2 = 1, NB
             !C set SW(:)
             DO m1 = 1, NB
                SW(m1) = Temp_P((i-1)*NB*NB+m1+(m2-1)*NB, j)
             END DO

             call DGETRS('N',NB,1,tilde_DLU(1:NB,1:NB,j),NB,tilde_DLU(:,NB+1,j),SW,NB,info)
             if(info/=0) stop "forward backward substitution error in smooth_aggregate_unsym_vec_NB()"

             DO m1 = 1, NB
                Temp_P((i-1)*NB*NB+m1+(m2-1)*NB, j) = omega * SW(m1)
             END DO
          END DO
       END DO
    END DO
    !$omp end do
    deallocate(SW)
    
    !$omp single
    omega2 = 1. - omega
    !$omp end single
    

    !$omp do 
    DO k = 1, in_aggregates_result_size
       !C  element's  value of the aggregate is 1 in array P      
       is = in_aggregates_result(k - 1) + 1
       ie = in_aggregates_result(k)

       DO i = is, ie
          P_CN = aggregates_result(i)
          DO count = 1, Temp_P_row_size(P_CN)
             if(Temp_PI(count, P_CN) == k) exit
          END DO
          if(count > Temp_P_row_size(P_CN)) then
             stop 'unexpected error in smooth_aggregate()'
          else
!!$             Temp_P(count, P_CN) % value = Temp_P(count, P_CN) % value + omega2
             DO j = 1, NB
                Temp_P((count-1)*NB*NB+j+(j-1)*NB, P_CN) = Temp_P((count-1)*NB*NB+j+(j-1)*NB, P_CN) + omega2
             END DO
          end if
       END DO
    END DO
    !$omp end do
    !$omp end parallel

    deallocate(Temp_agr_num, temp, temp_list, temp_sz, temp_index)
    
  END SUBROUTINE smooth_aggregate_unsym_vec_NB


!C This routine uses symmetricity structure of PNI in order to treat 
!C unstructured problem matrix.
!C This routine uses PNI, NI and no option other than SMOOTH_TILDE_A
!C NB is set as 50
  SUBROUTINE smooth_aggregate_unsym_vec (PNI, NI, omega, LEVEL_NO, in_aggregates_result_size, &
       & in_aggregates_result, aggregates_result, Temp_P, Temp_PI,Temp_P_row_size, tilde_DLU,   &
       & RATE_OF_SPACE_i, EX_INL, EX_IAL, EX_AL, NB, NK, SOLVER_COMM, my_rank, k_or_q_vec)
    use count_time_mod
    use data_structure_for_AMG_NB
    !$ USE omp_lib
    
    IMPLICIT NONE

    INTEGER(kind=kint ), INTENT(in) :: PNI(0:)
    INTEGER(kind=kint ), INTENT(in) :: NI(:)
    REAL   (kind=kreal), INTENT(in) :: omega 
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO, RATE_OF_SPACE_i
    INTEGER(kind=kint ), INTENT(in) :: in_aggregates_result_size, SOLVER_COMM, my_rank
    REAL   (kind=kreal), INTENT(in) ::  tilde_DLU(:,:,:)
    INTEGER(kind=kint ), INTENT(in) :: EX_INL (0:), EX_IAL (:)
    REAL   (kind=kreal), INTENT(in) :: EX_AL  (:,:,:), k_or_q_vec(:,:)
    
    INTEGER(kind=kint ), INTENT(in) :: in_aggregates_result(0:)
    INTEGER(kind=kint ), INTENT(in) :: aggregates_result(:)

    real   (kind=kreal), INTENT(out) :: Temp_P(:,:)
    INTEGER(kind=kint ), INTENT(out) :: Temp_PI(:,:)
    INTEGER(kind=kint ), INTENT(out) :: Temp_P_row_size(:)
    INTEGER(kind=kint ), allocatable :: Temp_agr_num(:)

    REAL   (kind=kreal), POINTER ::  D (:,:,:)
    REAL   (kind=kreal), POINTER ::  AU(:,:,:)
    REAL   (kind=kreal), POINTER ::  AL(:,:,:)
    INTEGER(kind=kint ), POINTER ::  INU(:)
    INTEGER(kind=kint ), POINTER ::  IAU(:)
    INTEGER(kind=kint ), POINTER ::  INL(:)
    INTEGER(kind=kint ), POINTER ::  IAL(:)
    
    INTEGER(kind=kint ), INTENT(in) :: NB, NK
    
    Real   (kind=kreal) :: omega2, v
    INTEGER(kind=kint ) :: N, i, j, k, l, row, index, is, ie, NP, c, m1, m2
    INTEGER(kind=kint ) :: connected_node, connected_node_start, connected_node_end
    INTEGER(kind=kint ) :: P_CN, count, node_num

    INTEGER(kind=kint), allocatable :: temp(:,:), temp_list(:,:), temp_sz(:),  temp_index(:,:)
    INTEGER(kind=kint) ::  tsz=1, tid = 0, tstart, tend, stid=1, ms, me, m, min_k
    integer :: info
    
    REAL(kind=kreal), allocatable :: SW(:)
    REAL(kind=kreal), allocatable :: tmp(:,:)

    NP = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NP
    N  = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % N 

    D   => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % D 
    INL => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % INL
    INU => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % INU
    IAL => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % IAL
    IAU => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % IAU
    AL  => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % AL
    AU  => HIERARCHICAL_DATA_NB(LEVEL_NO - 1) % AU
    
    !C assign the aggregate number in Temp_agr_num
    allocate(Temp_agr_num(N))


    !$omp parallel private(j,k)
    !$omp do
    DO i = 1, N
       Temp_agr_num(i) = 0
    END DO
    !$omp end do

    !$omp do 
    DO i = 1, in_aggregates_result_size
       DO j = in_aggregates_result(i - 1) + 1, in_aggregates_result(i)
          k = aggregates_result(j)
          Temp_agr_num(k) = i
       END DO
    END DO
    !$omp end do 
    
    !$omp do
    DO i = 1, NP
       Temp_P_row_size(i) = 0
    end DO
    !$omp end  do
    !$omp end parallel

    !C determine nonzero structure which utilizes the symmetricity of SMOOTH_TILDE_A



    
    !C----- parallel ----
    !C aggregates are allocated to threads cyclicly.
    !$omp parallel  private(count, i, j, k, m, l, tstart, tend, ms, me, stid, tid, &
    !$omp is, ie, P_CN, connected_node, connected_node_start, connected_node_end, &
    !$omp node_num)
    !$ tid = OMP_GET_THREAD_NUM()
    !$ tsz = OMP_GET_NUM_THREADS()

    !$omp single
    allocate(temp(NP, tsz), temp_list(NP, tsz), temp_sz(tsz), temp_index(0:in_aggregates_result_size, tsz))
    !$omp end single
    
    stid = tid+1

    !C initialize
    DO j = 1, NP
       temp(j, stid) = 0
       temp_list(j, stid) = 0
    END DO

    !C temp_sz has last aggregate number which is dealt with for tid.
    temp_sz(stid) = 0  
    
    !C k is the head of aggregate block for threads.
    !C k is increasing by tsz, as loop iterates.
    k = 1 
    DO while(k <= in_aggregates_result_size)    
       !C temp_index  and temp_sz is set.. 
       !C Here k-1 aggregates data is stored in Temp_PI already, 
       !C so temp_index(0:(k-1)/tsz, stid) is set 0
       temp_index(0:(k-1)/tsz, stid) = 0

       temp_sz(stid) = 0  
       count = 0 !C index for temp_list(:, stid)

       DO while(k <= in_aggregates_result_size)
          if(k+tid <= in_aggregates_result_size) then
             !C aggregate number is k+tid
             !C  element's  value of the aggregate is 1 in array P      
             is = in_aggregates_result((k+tid) - 1) + 1
             ie = in_aggregates_result(k+tid)
             DO i = is, ie
                P_CN = aggregates_result(i)          
                !C temp(:, stid) is redundancy check vector
                if(temp(P_CN, stid) /= k+tid) then
                   count = count+1
                   if(count > NP) exit !C temp_list buffer is full
                   temp_list(count, stid) = P_CN
                   temp(P_CN, stid) = k+tid
                end if
                connected_node_start = PNI(2 * P_CN - 2) + 1
                connected_node_end   = PNI(2 * P_CN - 1)
                DO connected_node = connected_node_start, connected_node_end
                   node_num = IAL(NI(connected_node)) + ZERO_ORIGIN
                   if(temp(node_num, stid) /= k+tid) then
                      count = count+1
                      if(count > NP) exit
                      temp_list(count, stid) = node_num
                      temp(node_num, stid) = k+tid
                   end if
                END DO
                if(count > NP) exit !C temp_list buffer is full
                connected_node_start = PNI(2 * P_CN - 1) + 1
                connected_node_end = PNI(2 * P_CN)
                DO connected_node = connected_node_start, connected_node_end
                   node_num = IAU(NI(connected_node)) + ZERO_ORIGIN
                   if(temp(node_num, stid) /= k+tid) then
                      count = count+1
                      if(count > NP) exit
                      temp_list(count, stid) = node_num
                      temp(node_num, stid) = k+tid
                   end if
                END DO
                if(count > NP) exit !C temp_list buffer is full
             END DO
          END if

          if(count <= NP) then
             !C record the index information every loop.
             temp_index((k-1)/tsz+1, stid) = count
             temp_sz(stid) = k
             k = k + tsz
          else 
             !C temp_list buffer is full.
             !C cancel the recorded info on redundancy check vector, and exit out of DO loop
             DO i = temp_index((k-1)/tsz, stid)+1, count-1
                temp(temp_list(i,stid), stid) = 0
             END DO
             exit
          end if
       END DO

       !$omp barrier

       !$omp single
       min_k = temp_sz(1)
       DO i = 2, tsz
          if(min_k > temp_sz(i)) min_k = temp_sz(i)
       END DO
       !$omp end single
       !C --syncronizaiton..--

       !C cancel the recorded info
       if(min_k < temp_sz(stid)) then
          is = temp_index((min_k-1)/tsz+1, stid)+1
          ie = temp_index((temp_sz(stid)-1)/tsz+1, stid)
          DO i = is, ie
             temp(temp_list(i,stid), stid) = 0
          END DO
          temp_sz(stid) = min_k
       end if

       tstart = tid*NP/tsz+1; tend = (tid+1)*NP/tsz

       DO m = 1, (min_k-1)/tsz+1 !C m: thread block number 
          DO l = 1, tsz !C l: thread number from 1..tsz
             is = temp_index(m-1,l)+1; ie = temp_index(m,l)
             DO i = is, ie
                node_num = temp_list(i, l)
                !$ if(node_num >= tstart .and. node_num <= tend) then
                      count = Temp_P_row_size(node_num)+1
                      Temp_P_row_size(node_num) = count
                      !C for vector operations here's modified.
                      Temp_PI(count, node_num) = (m-1)*tsz+l
                      Temp_P((count-1)*NB*NK+1:count*NB*NK, node_num) = 0.
                !$ end if
             END DO
          END DO
       END DO

       !$omp barrier
       k = min_k+tsz
    END DO
    !$omp end parallel
    !C----- parallel ----       
    


    !C Dumped yacobi method
    !$omp parallel private(i,j,k,count,c,m1,m2,SW,tmp,is,ie,P_CN) 
    allocate(tmp(NB,NK))
    
    !$omp do 
    DO i = 1, N
       !C lower part
       DO j = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
!!$          v = AL(NI(j))
          k = Temp_agr_num(IAL(NI(j))+ZERO_ORIGIN)

          DO count = 1, Temp_P_row_size(i)
             if(Temp_PI(count, i) == k) exit
          END DO
          IF(count > Temp_P_row_size(i)) THEN
             if(k > 0) stop '0:error in smooth_aggregate_unsym: non-symmetric data structure'
          else
             !C for vector cases, v becomes AL(:,:,NI(j)) * orthogonalized kernel vecs(:,:) of IAL(NI(j))
             !C             Temp_P(count, i) % value  = Temp_P(count, i) % value - v
             !C columne direction is assumed to be sequence access.
             
             call  product_kernel_vec_A(NB, NK, k_or_q_vec, tmp, AL(:,:,NI(j)), IAL(NI(j)))
             
             DO m1 = 1, NB
                DO m2 = 1, NK
                   Temp_P((count-1)*NB*Nk+m1+(m2-1)*NB, i) = &
!!$                        & Temp_P((count-1)*NB*NK+m1+(m2-1)*NB, i) - AL(m1,m2,NI(j))
                        & Temp_P((count-1)*NB*NK+m1+(m2-1)*NB, i) - tmp(m1,m2)
                END DO
             END DO
             
          END IF
       END DO
       
       !C upper part
       DO j = PNI(2 * i - 1) + 1, PNI(2 * i)
!!$          v = AU(NI(j))
          c = IAU(NI(j)) + ZERO_ORIGIN
          if(c <= N) then
             k = Temp_agr_num(c)

             DO count = 1, Temp_P_row_size(i)
                if(Temp_PI(count, i) == k) exit
             END DO
             !C we must consider the case of k==0. In that case, following stop statement will be executed.
             IF(count > Temp_P_row_size(i)) THEN
                if(k > 0) stop '1:error in smooth_aggregate_unsym: non-symmetric data structure'
             else
                !C for vector cases, v becomes AU(NI(j)) x orthogonalized kernel vecs(:,:) of IAU(NI(j))
!!$                Temp_P(count, i) % value  = Temp_P(count, i) % value - v

                call  product_kernel_vec_A(NB, NK, k_or_q_vec, tmp, AU(:,:,NI(j)), IAU(NI(j)))

                DO m1 = 1, NB
                   DO m2 = 1, NK
                      Temp_P((count-1)*NB*Nk+m1+(m2-1)*NB, i) = &
!!$                           & Temp_P((count-1)*NB*Nk+m1+(m2-1)*NB, i) -  AU(m1,m2,NI(j))
                           & Temp_P((count-1)*NB*Nk+m1+(m2-1)*NB, i) -  tmp(m1,m2)
                   END DO
                END DO

             END IF
          end if
       END DO
    END DO
    !$omp end do nowait


    !C external lower part
    !$omp do 
    DO i = N + 1, NP
       DO j = PNI(2 * i - 2) + 1, PNI(2 * i - 1)
!!$          v = EX_AL(NI(j))
          c = EX_IAL(NI(j))+ZERO_ORIGIN
          if(c <= N) then
             k = Temp_agr_num(c)

             DO count = 1, Temp_P_row_size(i)
                if(Temp_PI(count, i) == k) exit
             END DO

             IF(count > Temp_P_row_size(i)) THEN
                if(k > 0) stop '2:error in smooth_aggregate_unsym: non-symmetric data structure'
             else
                !C for vector cases, v becomes EX_AL(:,:,NI(j)) x orthogonalized kernel vecs(:,:) of EX_IAL(NI(j))
!!$                Temp_P(count, i) % value  = Temp_P(count, i) % value - v

                call  product_kernel_vec_A(NB, NK, k_or_q_vec, tmp, EX_AL(:,:,NI(j)), EX_IAL(NI(j)))

                DO m1 = 1, NB
                   DO m2 = 1, NK
                      Temp_P((count-1)*NB*NK+m1+(m2-1)*NB, i) = &
!!$                           & Temp_P((count-1)*NB*Nk+m1+(m2-1)*NB, i) - EX_AL(m1,m2,NI(j))
                           & Temp_P((count-1)*NB*NK+m1+(m2-1)*NB, i) - tmp(m1,m2)
                   END DO
                END DO
             END IF
          END if
       end DO
    END DO
    !$omp end do

    !C-- diagonal part
!!$    !$omp do private(i,j)
!!$    DO j = 1, NP
!!$       DO i = 1, Temp_P_row_size(j)
!!$          Temp_P(i, j) % value = Temp_P(i, j) % value * omega / &
!!$               & tilde_D(j)
!!$       END DO
!!$    END DO
!!$    !$omp end do

    !C 1 4 7 -> j
    !C 2 5 8
    !C 3 6 9
    !C-- diagonal part
    allocate(SW(NB))
    !$omp do 
    DO j = 1, NP
       DO i = 1, Temp_P_row_size(j)
!!$          Temp_P(i, j) % value = Temp_P(i, j) % value * omega / &
!!$               & tilde_D(j)
          !C NK vectors are applied to forward and backward substitution
          DO m2 = 1, NK
             !C set SW(:)
             DO m1 = 1, NB
                SW(m1) = Temp_P((i-1)*NB*NK+m1+(m2-1)*NB, j)
             END DO
             
             
             call DGETRS('N',NB,1,tilde_DLU(1:NB,1:NB,j),NB,tilde_DLU(:,NB+1,j),SW,NB,info)
             if(info/=0) stop "forward backward substitution error in smooth_aggregate_unsym_vec()"
             
             DO m1 = 1, NB
                Temp_P((i-1)*NB*NK+m1+(m2-1)*NB, j) = omega * SW(m1)
             END DO
          END DO
       END DO
    END DO
    !$omp end do
    deallocate(SW)
    
    !$omp single
    omega2 = 1. - omega
    !$omp end single

    !$omp do 
    DO k = 1, in_aggregates_result_size
       !C  element's  value of the aggregate is 1 in array P      
       is = in_aggregates_result(k - 1) + 1
       ie = in_aggregates_result(k)

       DO i = is, ie
          P_CN = aggregates_result(i)
          DO count = 1, Temp_P_row_size(P_CN)
             if(Temp_PI(count, P_CN) == k) exit
          END DO
          if(count > Temp_P_row_size(P_CN)) then
             stop 'unexpected error in smooth_aggregate()'
          else
             !C for vector version
!!$             Temp_P(count, P_CN) % value = Temp_P(count, P_CN) % value + omega2
             DO m1 = 1, NB
                DO m2 = 1, NK
                   Temp_P((count-1)*NB*NK+m1+(m2-1)*NB, P_CN) = &
                        & Temp_P((count-1)*NB*Nk+m1+(m2-1)*NB, P_CN) + &
                        & omega2 * k_or_q_vec((P_CN-1)*NB+m1,m2)
                END DO
             END DO
          end if
       END DO
    END DO
    !$omp end do
    !$omp end parallel

    deallocate(Temp_agr_num, temp, temp_list, temp_sz, temp_index)
    
  contains
    
    subroutine product_kernel_vec_A(NB, NK, k_or_q_vec, out, mat, col)
      implicit none
      integer(kind=kint) :: NB, NK, col
      real(kind=kreal)   :: k_or_q_vec(:,:), out(:,:), mat(:,:)
      
      integer(kind=kint) :: i,j,k
      
      out = 0
      
      DO i = 1, NB
         DO j = 1, NK
            do k = 1, NB
               out(i,j) = out(i,j) + mat(i,k)*k_or_q_vec((col-1)*NB+k,j)
            end do
         end DO
      end DO
      
    end subroutine product_kernel_vec_A
    
  END SUBROUTINE smooth_aggregate_unsym_vec

  SUBROUTINE neighbors(PNI, NI, theta, level, node_index, tilde_D)
    USE data_structure_for_AMG_NB

    IMPLICIT NONE
    INTEGER(kind=kint ),  INTENT(out) :: PNI(0:)
    INTEGER(kind=kint ),  INTENT(out) :: NI(:)
    INTEGER(kind=kint ),  INTENT(out) :: node_index(:)
    REAL   (kind=kreal),  INTENT(out) :: tilde_D(:,:,:)
    REAL   (kind=kreal),  INTENT(in ) :: theta
    INTEGER(kind=kint ),  INTENT(in ) :: level

    INTEGER(kind=kint) :: N, NP, NPL, NPU
    REAL   (kind=kreal), POINTER ::  D(:,:,:)
    REAL   (kind=kreal), POINTER ::  AU(:,:,:)
    REAL   (kind=kreal), POINTER ::  AL(:,:,:)
    INTEGER(kind=kint ), POINTER ::  INU(:)
    INTEGER(kind=kint ), POINTER ::  IAU(:)
    INTEGER(kind=kint ), POINTER ::  INL(:)
    INTEGER(kind=kint ), POINTER ::  IAL(:)

    INTEGER(kind=kint ) :: i,j,k,l,isL,ieL,isU,ieU,inod
    INTEGER(kind=kint ) :: nni,nni_before,size,NB
    REAL   (kind=kreal) :: dii,djj,aij

    NB  = HIERARCHICAL_DATA_NB(level-1) % NB 
    NP  = HIERARCHICAL_DATA_NB(level-1) % NP 
    N   = HIERARCHICAL_DATA_NB(level-1) % N 

    NPL = HIERARCHICAL_DATA_NB(level-1) % NPL
    NPU = HIERARCHICAL_DATA_NB(level-1) % NPU

    D   => HIERARCHICAL_DATA_NB(level-1) % D
    INL => HIERARCHICAL_DATA_NB(level-1) % INL
    INU => HIERARCHICAL_DATA_NB(level-1) % INU
    IAL => HIERARCHICAL_DATA_NB(level-1) % IAL
    IAU => HIERARCHICAL_DATA_NB(level-1) % IAU
    AL  => HIERARCHICAL_DATA_NB(level-1) % AL
    AU  => HIERARCHICAL_DATA_NB(level-1) % AU

    PNI(0) = 0


    nni    = 0

    DO i = 1, NP
       DO j = 1, NB
          DO k = 1, NB
             tilde_D(k,j,i) = D(k,j,i)
          end DO
       end DO
    END DO

    DO i = 1,NP
       node_index(i) = 0
    END DO

    DO j = 1,N

       isU = INU(j-1)+1
       ieU = INU(j  )
       isL = INL(j-1)+1
       ieL = INL(j  )

       nni_before = nni
       DO i = isL, ieL

          inod = IAL(i)+ZERO_ORIGIN
          call norm_NB(D(:,:,j),    dii, NB)
          call norm_NB(D(:,:,inod), djj, NB)
          call norm_NB(AL(:,:,i),   aij, NB)

          !C diagonal part is positive number.
          IF(aij*aij > dabs(dii*djj*theta*theta) .AND. dii*djj*aij<0) THEN
!!$          IF(aij*aij > dabs(dii*djj*theta*theta)) THEN
             nni = nni + 1
             NI(nni) = i
          ELSE
             DO k = 1, NB
                DO l = 1, NB
                   tilde_D(k,k,j) = tilde_D(k,k,j) +AL(k,l,i)
                END DO
             END DO
          ENDIF
       ENDDO

       PNI(2*j-1)=nni
       DO i= isU, ieU
          inod= IAU(i)

          call norm_NB(D(:,:,j),    dii, NB)
          call norm_NB(D(:,:,inod), djj, NB)
          call norm_NB(AU(:,:,i),   aij, NB)

          !C overlap nodes are ignored
          IF(aij*aij > dabs(dii*djj*theta*theta) .AND. dii*djj*aij<0) THEN
!!$          IF(aij*aij > dabs(dii*djj*theta*theta)) THEN
             nni = nni + 1
             NI(nni)=i
          ELSE
             DO k = 1, NB
                DO l = 1, NB
                   tilde_D(k,k,j) = tilde_D(k,k,j) + AU(k,l,i)
                END DO
             END DO
             
          END IF
       ENDDO
       PNI(2*j)=nni

       size = nni-nni_before
       IF(size ==0 .and. j <= N) node_index(j)=-1
    ENDDO

  end SUBROUTINE neighbors
  
  
  !C this subroutine considers the case of unsymmetric problem matrix.
  !C But symmetricity of nonzero structure of the matrix is assumed.
  !C 1..N rows are calculated parallel by omp. N+1..NP rows are not parallelized.
  SUBROUTINE neighbors_unsym(EX_INL, EX_IAL, EX_AL, PNI, NI, theta, LEVEL_NO,  &
       & node_index, tilde_D, SOLVER_COMM, my_rank)
    USE data_structure_for_AMG_NB
    USE count_time_mod
    !$ USE omp_lib
    IMPLICIT NONE
    
    INTEGER(kind=kint ) :: N, NP, NB
    REAL   (kind=kreal), INTENT(in) :: theta
    INTEGER(kind=kint ), INTENT(in) :: EX_INL (0:), EX_IAL (:), SOLVER_COMM, my_rank
    REAL   (kind=kreal), INTENT(in) :: EX_AL  (:,:,:)

    INTEGER(kind=kint ), INTENT(out) :: NI  (:)
    INTEGER(kind=kint ), INTENT(out) :: PNI(0:)
    INTEGER(kind=kint ), INTENT(out) :: node_index(:)
    
    INTEGER(kind=kint ), INTENT(in) :: LEVEL_NO
    REAL   (kind=kreal), POINTER :: D  (:,:,:)
    REAL   (kind=kreal), POINTER :: AU (:,:,:)
    REAL   (kind=kreal), POINTER :: AL (:,:,:)
    INTEGER(kind=kint ), POINTER :: INU(:)
    INTEGER(kind=kint ), POINTER :: IAU(:)
    INTEGER(kind=kint ), POINTER :: INL(:)
    INTEGER(kind=kint ), POINTER :: IAL(:)

    REAL(kind=kreal)  :: tilde_D(:,:,:)

    INTEGER(kind=kint), ALLOCATABLE :: check_U(:)

    INTEGER(kind=kint) :: i, j, k, isU, ieU, isL, ieL, inod, nni
    INTEGER(kind=kint) :: start, goal, nni_ini
    INTEGER(kind=kint) :: tid, tsz=1 
    REAL   (kind=kreal),allocatable:: tl(:,:), tu(:,:)


    NB  = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NB     
    allocate(tl(NB,NB), tu(NB,NB))
    
    NP  = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % NP     
    N   = HIERARCHICAL_DATA_NB(LEVEL_NO-1) % N 
    
    D   => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % D
    INL => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INL
    INU => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % INU
    IAL => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAL
    IAU => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % IAU
    AL  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AL
    AU  => HIERARCHICAL_DATA_NB(LEVEL_NO-1) % AU
    
    PNI(0) = 0
    nni    = 0
    
    !$omp parallel do private(j,k)
    DO i = 1, NP
       DO j = 1, NB
          DO k = 1, NB
             tilde_D(k,j,i) = D(k,j,i)
          end DO
       end DO
    END DO
    !$omp end parallel do


    !$omp parallel do
    DO i = 1, NP
       node_index(i) = 0
    END DO
    !$omp end parallel do

    start = 1; goal = N

    !C parallelization by row block.
    !$omp parallel private(tid, tsz, start, goal, nni, i, j, isL, ieL, inod, isU, ieU)
    !$ tid = OMP_GET_THREAD_NUM()
    !$ tsz = OMP_GET_NUM_THREADS()
    !$ start = (tid*N)/tsz +1; goal = (tid+1)*N/tsz 
    !$ nni = INL(start-1)+INU(start-1)
    
    DO j = start, goal
       isL = INL(j - 1) + 1; ieL = INL(j)
       DO i = isL, ieL
          inod = IAL(i) + ZERO_ORIGIN
          !C determining the cordinate position of upper matrix
          
          isU = INU(inod - 1) + 1; ieU = INU(inod)
          call cordinating_upper_lower_mat(NI, nni, D, theta, tilde_D, IAU, AU, isU, ieU, &
               & i, j, AL(:,:,i), inod, NB)
          
       END DO
       PNI(2 * j - 1) = nni
       
       isU = INU(j - 1) + 1; ieU = INU(j)
       DO i = isU, ieU
          inod = IAU(i) + ZERO_ORIGIN

          
          !C determining the cordinate position of lower matrix          
          if(inod > N) then
             !C This part uses external lower part matrix
             isL = EX_INL(inod - N - 1) + 1; ieL = EX_INL(inod - N)
             call cordinating_upper_lower_mat(NI, nni, D, theta, tilde_D, EX_IAL, EX_AL, &
                  & isL, ieL, i, j, AU(:,:,i), inod,NB)
          ELSE
             !C determining the cordinate position of lower matrix
             isL = INL(inod - 1) + 1; ieL = INL(inod)
             call cordinating_upper_lower_mat(NI, nni, D, theta, tilde_D, IAL, AL, isL, &
                  & ieL, i, j, AU(:,:,i), inod, NB)
          ENDIF
       ENDDO
       PNI(2 * j) = nni
    END DO
    !$omp barrier

    !$omp single
    !C neat PNI
    DO i = 2, tsz
      start = (i-1)*N/tsz+1; goal = (i)*N/tsz
      nni_ini = INL(start-1)+INU(start-1)
      k = nni_ini-PNI(2*start-2)
      DO j = start, goal
        PNI(2*j-1) = PNI(2*j-1) - k
        PNI(2*j) = PNI(2*j) - k
      end DO
    END DO
    
    !C neat NI(:)
    DO i = 2, tsz
      start = (i-1)*N/tsz+1; goal = (i)*N/tsz
      nni_ini = INL(start-1)+INU(start-1)
      k = nni_ini-PNI(2*start -2)
      DO j = PNI(2*start-2)+1, PNI(2*start-2)+PNI(2*goal)-PNI(2*start-2)
         NI(j) = NI(j+k)
      end DO
    END DO
    !$omp end single
    !$omp end parallel
    
    !C reset nni
    nni = PNI(2*N)

    
    !C This part is for external lower submatrix. serial part.
    DO j = N + 1, NP
       isL = EX_INL(j - N - 1) + 1
       ieL = EX_INL(j - N)
       DO i = isL, ieL
          inod = EX_IAL(i)+ZERO_ORIGIN
          tl = EX_AL(:,:,i)
          !C determining the cordinate position of upper matrix
          isU = INU(inod - 1) + 1; ieU = INU(inod)
          call cordinating_upper_lower_mat(NI, nni, D, theta, tilde_D, IAU, AU,  &
               & isU, ieU, i, j, tl, inod, NB)
       ENDDO
       PNI(2 * j - 1) = nni
       PNI(2 * j) = nni
    END DO


    !$omp parallel do
    DO j = 1, N
       IF(PNI(2 * j) - PNI(2 * j - 2) == 0) node_index(j) = -1
    END DO
    !$omp end parallel do

  contains
    !C search the cordinating nonzero element in matrix(cord_*A) and
    !C record chosen nodes in NI(:).
    subroutine cordinating_upper_lower_mat(NI, nni, D, theta, tilde_D, cord_IA, &
         & cord_A, isc, iec, i, r, tl, inod, NB)
      implicit none
      
      INTEGER(kind=kint ), intent(in)    :: i, r, inod, isc, iec, NB
      INTEGER(kind=kint ), intent(inout) :: nni,  NI(:)
      INTEGER(kind=kint ), intent(in)    :: cord_IA(:)
      REAL   (kind=kreal), intent(in)    :: cord_A(:,:,:), D(:,:,:), theta, tl(:,:)
      REAL   (kind=kreal), intent(inout) :: tilde_D(:,:,:)

      REAL   (kind=kreal) :: tcnrm, tlnrm, dr, dinod
      INTEGER(kind=kint ) :: k, l, m 
      
      DO k = isc, iec
         IF(cord_IA(k)+ZERO_ORIGIN == r) EXIT
      END DO
      IF(k > iec) then
         write(*,*) "Strange case", r, inod
!!$         tilde_D(:,:,r) = tilde_D(:,:,r) + tl(:,:)
         DO l = 1, NB
            DO m = 1, NB
               tilde_D(l,l,r) = tilde_D(l,l,r) + tl(l,m)
            END DO
         END DO

      ELSE
         
         call norm_NB(D(:,:,r), dr, NB)
         call norm_NB(D(:,:,inod), dinod, NB)
         call norm_NB(tl, tlnrm, NB)
         call norm_NB(cord_A(:,:,k) , tcnrm, NB)

!!$         IF(tlnrm * tlnrm > dabs(dr * dinod * theta * theta) .AND. dr * dinod * tlnrm < 0 .OR. &
!!$              & tcnrm * tcnrm > dabs(dr * dinod * theta * theta) .AND. dr * dinod * tcnrm < 0) THEN
         IF(tlnrm * tlnrm > dabs(dr * dinod * theta * theta).OR. &
              & tcnrm * tcnrm > dabs(dr * dinod * theta * theta) ) THEN
            nni = nni + 1
            NI(nni) = i
         ELSE
!C            tilde_D(r) = tilde_D(r) + tl
            DO l = 1, NB
               DO m = 1, NB
                  tilde_D(l,l,r) = tilde_D(l,l,r) +tl(l,m)
               END DO
            END DO
         ENDIF
      
      END IF
      
    END subroutine cordinating_upper_lower_mat
  END SUBROUTINE neighbors_unsym
  

  SUBROUTINE norm_NB(a,val, NB)
    IMPLICIT NONE
    INCLUDE 'precision.inc'
    INTEGER(kind=kint), intent(in ) :: NB
    REAL(kind=kreal),   intent(in ) :: a(1:NB, 1:NB)
    REAL(kind=kreal),   INTENT(out) :: val
    INTEGER(kind=kint) :: i,j,k
    REAL(kind=kreal) :: ab_val(100), mv
    !C assume max NB is 100..

    if(.true.)  then
       val=0.0
       DO i = 1, NB
          DO j = 1, NB
             val = val+a(i,j)*a(i,j)
          END DO
       end DO
       val = dsqrt(val)
       
    else 
       DO i = 1, NB
          ab_val(i) = dabs(a(i,i))
       END DO

       val = a(1,1)
       mv = dabs(a(1,1))

       DO i = 2, NB
!!$       DO j = 1, NB
          if(mv < dabs(a(i,i))) then
             mv = dabs(a(i,i))
             val = a(i,i)
          end if
!!$       END DO
       end DO
    end if

  END SUBROUTINE norm_NB



  SUBROUTINE  DEBUG_BARRIER_PRINT(string, comm, my_rank)
    IMPLICIT NONE
    INCLUDE 'mpif.h'
    INCLUDE 'precision.inc'

    CHARACTER(len=*), intent(in) :: string
    INTEGER(kind=kint), intent(in) :: comm, my_rank
    INTEGER(kind=kint) :: ierr

#ifdef DEBUG    
    CALL MPI_BARRIER(comm, ierr)
    IF (my_rank == 0) THEN
       WRITE(*,*) string
    END IF
    CALL MPI_Barrier(comm, ierr)
#endif    


  END SUBROUTINE DEBUG_BARRIER_PRINT

END MODULE  data_creation_AMGCG_NB
